/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1986 Wayne A. Christopher
         1994 Stephen R. Whiteley
****************************************************************************/

/*
 * The main routine for the help system in stand-alone mode.
 */
#include "spice.h"
#include "cpdefs.h"
#ifdef HAVE_GETPWUID
#include <pwd.h>
#endif

#ifdef HAVE_X11

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

Display *Xdisplay;
static XtAppContext app_con;
static Widget toplevel;
static char *hcopydev;

static String fallback_resources[] = {
    
    "help.help.background: lightgrey",
    "help.help.Box.background: lightblue",
    "help.help.Box.Command.background: yellow",
    "help.help.Box.Toggle.background: yellow",
    "help.help.Box.Label.background: pink",
    "help.help.Label.background: pink",
    
    "help.popup.form.background: lightblue",
    "help.popup.form.Command.background: yellow",
    "help.popup.form.Label.background: pink",

    "help.popup_m.form_m.background: pink",
    "help.popup_m.form_m.Command.background: yellow",
    NULL
};


static int
errorhandler(display, errorev)
    
Display *display;
XErrorEvent *errorev;
{   
    char ErrorMessage[1024];

    XGetErrorText(display, errorev->error_code, ErrorMessage, 1024);
    printf(ErrorMessage);
    return;
}   


static void
CRaction(caller, call_data, params, numparams)

Widget caller;
XEvent *call_data;
String *params;
Cardinal *numparams;
{
    /* simulate a button press when CR entered in text of popup */
    Widget action = XtNameToWidget(XtParent(caller), "action");

    if (action)
        XtCallCallbacks(action, XtNcallback, NULL);
}


static void
QUaction(caller, call_data, params, numparams)
    
Widget caller;
XEvent *call_data;
String *params;
Cardinal *numparams;
{           
    /* simulate a Quit button press when DeleteWindow requested */
    Widget quit;
    
    quit = XtNameToWidget(caller, "help.buttonbox.delete");
    if (quit)
        XtCallCallbacks(quit, XtNcallback, NULL);
}       

static XtActionsRec actions[] = {
    {"cr_action", CRaction},
    {"quit_action", QUaction}
};

#endif

/* dummy declaration so CP.a doesn't pull in lexical.o and other objects */
bool cp_interactive = true;
FILE *cp_in, *cp_out, *cp_err;

extern bool hlp_usex, hlp_killx;
char *Help_Path;
char *kw_hcopydev = "hcopydev";

#ifdef __STDC__
static char *tilde_expand(char*);
#else
static char *tilde_expand();
#endif


/* ARGSUSED */
int
main(argc, argv)

int argc;
char **argv;
{
    wordlist *wl = NULL;
    char *s;

    /* SPICE_HELP defined with -D cmdline option */
    Help_Path = copy(SPICE_HELP);
    if ((s = getenv("SPICE_HELP")) != NULL)
        Help_Path = copy(s);
    else if ((s = getenv("SPICE_LIB_DIR")) != NULL) {
        Help_Path = tmalloc(strlen(s) + 10);
        sprintf(Help_Path,"%s/helpdir",s);
    }
    if ((s = getenv("PRINTER")) != NULL)
        hcopydev = copy(s);
    else
        hcopydev = "lp";
    cp_in = stdin;
    cp_out = stdout;
    cp_err = stderr;

#ifdef HAVE_X11
    hlp_usex = 1;
    hlp_killx = 1;

    /* initialize X toolkit */
    XtToolkitInitialize();
    app_con = XtCreateApplicationContext();
    XtAppSetFallbackResources(app_con, fallback_resources);
    Xdisplay = XtOpenDisplay(app_con, NULL, NULL, "Help", NULL, 0,
        &argc, argv);
    if (!Xdisplay) {
        fprintf(stderr, "Error: Can't open X display");
        return  (1);
    }
    XSetErrorHandler(errorhandler);
    XtAppAddActions(app_con, actions, XtNumber(actions));
#endif

out:
    if (argc > 1)
        wl = wl_build(argv + 1);
    hlp_main(Help_Path, wl);

#ifdef HAVE_X11
    if (hlp_usex) {
        XtAppMainLoop(app_con);
    }
#endif

    return (EXIT_NORMAL);
}


void
fatal()

{
    fprintf(stderr, "fatal error\n");
    exit(1);
}


/* ARGSUSED */
bool
cp_getvar(n, t, r)

char *n, *r;
int t;
{
    if (!strcmp(n, kw_hcopydev)) {
        strcpy(r, hcopydev);
        return (True);
    }
    return (false);
}


void
cp_pathfix(buf)
            
char *buf;
{       
#ifdef MSDOS
    char *s, *t;
    int bcnt = 0, ecnt = 0;

    s = t = buf;
    while (*t != '\0') {
        if (*t == '/' || *t == '\\') {
            *s++ = '\\';
            t++;
            bcnt = 0;
            ecnt = 0;
        }
        else if (*t == '.') {
            *s++ = *t++;
            ecnt = 1;
        }
        else if (!ecnt) {
            if (bcnt++ < 8) *s++ = *t++;
            else t++;
        }
        else {
            if (ecnt++ < 4) *s++ = *t++;
            else t++;
        }
    }
    *s = '\0';

#else
    char *s;

    if (index(buf,'~')) {
        s = tilde_expand(buf);
        if (s) {
            strcpy(buf,s);
            txfree(s);
        }
    }
#endif
}


static char *
tilde_expand(string)

char *string;
{
#ifdef HAVE_GETPWUID
    struct passwd *pw;
    char    *tail;
    char    buf[BSIZE_SP];
    char    *k, c;

    if (!string)
    return (NULL);

    while (*string && isspace(*string))
    string++;

    if (*string != '~')
        return copy(string);

    string += 1;

    if (!*string || *string == '/') {
        pw = getpwuid(getuid());
        *buf = 0;
    }
    else {
        k = buf;
        while ((c = *string) && c != '/')
            *k++ = c, string++;
        *k = 0;
        pw = getpwnam(buf);
    }

    if (pw) {
        strcpy(buf, pw->pw_dir);
        if (*string)
            strcat(buf, string);
    }
    else
        return (NULL);

    return (copy(buf));

#else
    return (copy(string));
#endif
}

#ifdef HAVE_X11

/* ARGSUSED */
void
ScedESC()
{ return; }

void
ToTopOfSced()
{ return; }

#endif
