/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * The main routine for Jspice3 and nutmeg.
 */

#include "spice.h"
#include "misc.h"
#include "ftedefs.h"
#include "inpdefs.h"
#include "iferrmsg.h"
#include "plotdev.h"
#include "ftedebug.h"
#include "const.h"
#include "scedio.h"
#include <setjmp.h>

#ifdef HAVE_GETPWUID
#include <pwd.h>
#endif

#ifdef HAVE_SIGNAL
#include <sys/types.h>
#include <signal.h>
#endif

#ifndef HAVE_GETRUSAGE
#ifdef HAVE_FTIME
#include <sys/types.h>
#include <sys/timeb.h>
#endif
#endif

char *ft_rawfile = "rawspice.raw";

char *errRtn;
char *errMsg;
char *cp_program;
char *cp_display;

bool ft_servermode = false;
bool ft_intrpt = false;     /* Set by the signal handlers. */
bool ft_setflag = false;    /* Don't abort after an interrupt. */

#ifdef __STDC__
struct variable *(*if_getparam)(char*,char**,char*,int,wordlist**);
#else
struct variable *(*if_getparam)();
#endif

#ifdef BATCH

DISPDEVICE device[ ] = {
    {"", 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0 }
    };
DISPDEVICE *dispdev = device;
bool ft_nopage = false;
bool ft_parsedb = false;
bool ft_evdb = false;
bool ft_vecdb = false;
bool ft_grdb = false;
bool ft_gidb = false;
bool ft_controldb = false;
bool ft_asyncdb = false;

bool ft_debug = false;
bool ft_batchmode = true;
bool ft_nospiceadd = true;
bool ft_simdb = false;
static char *usage = "Usage: %s [ - ] [ -r rawfile ] [ file ... ]\n";
FILE *cp_curin = 0;
FILE *cp_curout = 0;
FILE *cp_curerr = 0;
FILE *cp_in = 0;
FILE *cp_out = 0;
FILE *cp_err = 0;
bool gi_endpause = true;

#ifndef SPICE2
bool ft_listprint = true;
bool ft_optsprint = false;
bool ft_nodesprint = false;
bool ft_acctprint = false;
struct plot *plot_list = NULL;
#endif

#else

bool ft_batchmode = false;
jmp_buf jbuf;
static char *usage = 
"Usage: %s [-] [-b] [-i] [-s] [-n] [-o outfile] [-r rawfile]\n\
\t[-t term] [file ...]\n";
struct options *exitoption;
struct options *helpoption;
extern RETSIGTYPE ft_sigintr(), sigfloat(), sigstop(), sigquit();
extern RETSIGTYPE sigill(), sigbus(), sigsegv(), sig_sys();

#endif

static started = false;
IFsimulator *ft_sim;

extern IFsimulator SIMinfo;

/* for HLP */
char *hlp_filelist[] = {"spice", "sced", 0 };

#ifdef SIMULATOR

bool ft_nutmeg = false;

#ifndef BATCH
extern struct comm spcp_coms[ ];
struct comm *cp_coms = spcp_coms;
#endif
extern int OUTstopnow(), OUTerror();
extern int OUTbeginPlot(), OUTdata(), OUTsetDims(), OUTendPlot();
extern int OUTendit;
static IFfrontEnd nutmeginfo = {
    IFnewUid,
    OUTstopnow,
    seconds,
    OUTerror,
    OUTbeginPlot,
    OUTdata,
    OUTsetDims,
    OUTendPlot,
    &OUTendit
};

static void clear_previous();
extern int  InProgress(), IsIplot();
extern void ft_dosim(), inp_decksource();
extern void com_resume(), com_iplot(), com_delete(); 

/* for SCED */
static struct sSCEDitf sc_itf =
{
    InProgress,
    ft_dosim,
    inp_decksource,
    clear_previous,
    com_resume,
    IsIplot,
    com_iplot,
    com_delete
};

#else

bool ft_nutmeg = true;
extern struct comm nutcp_coms[ ];
struct comm *cp_coms = nutcp_coms;
static IFfrontEnd nutmeginfo;


/* for SCED */
static struct sSCEDitf sc_itf =
{
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL
};

/*
 * Misc globals we have to fake for nutmeg.
 */

struct circ *ft_curckt;
struct circ *ft_circuits;
struct dbcomm *dbiplot = NULL;      /* export for iplot */

bool ft_listprint;
bool ft_optsprint;
bool ft_nodesprint;
bool ft_acctprint;

/* ARGSUSED */
void
inp_spsource(fp,i,s)  /* main.c */
FILE* fp;
bool i;
char *s;
{ inp_nutsource(fp,i,s); }

/* ARGSUSED */
void
inp_source(s)  /* main.c */
char *s;
{
    wordlist *wl;

    wl = cp_lexer(s);
    nutcom_source(wl);
    wl_free(wl);
}

/* ARGSUSED */
void
ft_dorun(s)  /* main.c */
char *s;
{ return; }

/* ARGSUSED */
void
ft_dotsaves()  /* main.c */
{ return; }

/* ARGSUSED */
void
ft_savedotargs()  /* main.c */
{ return; }

/* ARGSUSED */
void
ft_cktcoms(i)  /* main.c */
bool i;
{ return; }

/* ARGSUSED */
char *
if_errstring(code)  /* error.c */
{ return ("spice error"); }

/* ARGSUSED */
void
if_option(ckt, name, type, value)  /* options.c */
char *name, *ckt, *value;
int type; { }

/* ARGSUSED */
struct variable *
if_getstat(n, c, w)  /* resource.c */
char *n, *c;
wordlist **w;
{ return (NULL);}

/* ARGSUSED */
wordlist *
GetAnalysisFromDeck()  /* scedintr.c */
{ return (NULL); }

/* ARGSUSED */
void
inp_list(fp,l1,l2,i)  /* argaf.c */
FILE *fp;
struct line *l1, *l2;
int i;
{ return; }

/* screen line editor for nutmeg, when not using X */
/* cursor height - 1 */
#define CURHT 0
#define FontWidth currentgraph->fontwidth

static void
textcursor(x,y)

int x,y;
{
    DevBox(x,y-1,x+FontWidth,y+CURHT);
}

char *
KbEdit(s,x,y,bg,fg,cc)

/*
 * s      Initial string to edit.
 * x,y    Lower left coordinates.
 * bg,fg  Background and foreground colors.
 * cc     Background color at cursor location.
 * Returns edited string (static!).
 */
char *s;
int x, y, bg, fg, cc;
{
    char tbuf[128];
    int i, k;
    char *end, ctmp[2];
    int editX, editY, editFg, editCc;
    char *editC;
    static char editBuf[128];

    editFg = fg;
    editCc = cc;
    editX = x;
    editY = y;

    DevSetColor(fg);
    *editBuf = '\0';
    if (s) {
        /* s can be NULL */
        DevText(s, editX, editY);
        strcpy(editBuf,s);
    }
    ctmp[1] = '\0';
    editC = editBuf;
    end = strchr(editBuf,'\0');
    *(end+1) = '\0';
    ctmp[0] = *editC;
    DevSetColor(cc);
    textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
    DevSetColor(fg);
    DevText(ctmp, editX + (int)(editC-editBuf)*FontWidth, editY);
    for (;;) {
        k = DevGetchar(NULL);
        if ((char) k == '\r') break;
        switch (k) {

        case 333:    /* right arrow */
            if (editC >= end) continue;
            ctmp[0] = *editC;
            DevSetColor(bg);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevSetColor(fg);
            DevText(ctmp,editX + (int)(editC-editBuf)*FontWidth,editY);
            editC++;
            ctmp[0] = *editC;
            DevSetColor(cc);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevSetColor(fg);
            DevText(ctmp,editX + (int)(editC-editBuf)*FontWidth,editY);
            continue;

        case 331:    /* left arrow */
            if (editC <= editBuf) continue;
            ctmp[0] = *editC;
            DevSetColor(bg);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevSetColor(fg);
            DevText(ctmp,editX + (int)(editC-editBuf)*FontWidth,editY);
            editC--;
            ctmp[0] = *editC;
            DevSetColor(cc);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevSetColor(fg);
            DevText(ctmp,editX + (int)(editC-editBuf)*FontWidth,editY);
            continue;

        case '\b':
            if (editC == editBuf) continue;
            DevSetColor(bg);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            editC--;
            DevText(editC,editX + (int)(editC-editBuf)*FontWidth,editY);
            DevSetColor(fg);
            *editC = '\0';
            strcat(editBuf,++editC);
            DevText(editC,editX + (int)(editC-editBuf)*FontWidth,editY);
            editC--;
            if (end > editBuf) end--;
            ctmp[0] = *editC;
            DevSetColor(cc);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevSetColor(fg);
            DevText(ctmp,editX + (int)(editC-editBuf)*FontWidth,editY);
            continue;

        case 339: /* DEL, in DOS */
        case 127:
            if (editC == end) continue;
            DevSetColor(bg);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevText(editC,editX + (int)(editC-editBuf)*FontWidth,editY);
            DevSetColor(fg);
            *editC = '\0';
            strcat(editBuf,++editC);
            DevText(editC,editX + (int)(editC-editBuf)*FontWidth,editY);
            editC--;
            if (end > editBuf) end--;
            ctmp[0] = *editC;
            DevSetColor(cc);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevSetColor(fg);
            DevText(ctmp,editX + (int)(editC-editBuf)*FontWidth,editY);
            continue;

        case '\025': /* ^U */
        case '\030': /* ^X */
        case '\033': /* ESC */
            DevSetColor(bg);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevText(editBuf,editX,editY);
            editC = end = editBuf;
            *editC = '\0';
            if ((char) k == '\033') {
                return (NULL);
            }
            DevSetColor(cc);
            textcursor(editX,editY);
            DevSetColor(fg);
            continue;
        default:
            if (k > 255) continue;
            if (((char) k < ' ') || ((char) k > '~'))  continue;
            DevSetColor(bg);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevText(editC,editX + (int)(editC-editBuf)*FontWidth,editY);
            strcpy(tbuf,editC);
            *editC = (char) k;
            sprintf(++editC,tbuf);
            ctmp[0] = *editC;
            DevSetColor(fg);
            DevText(editC-1,
                editX+(int)(editC-1-editBuf)*FontWidth,editY);
            DevSetColor(cc);
            textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
            DevSetColor(fg);
            DevText(ctmp,editX + (int)(editC-editBuf)*FontWidth,editY);
            end++;
            *(end+1) = '\0';

        }
    }
    DevSetColor(bg);
    textcursor(editX+(int)(editC-editBuf)*FontWidth,editY);
    DevSetColor(fg);
    ctmp[0] = *editC;
    DevText(ctmp,editX + (int)(editC-editBuf)*FontWidth,editY);
    return (editBuf);
}
#endif

typedef RETSIGTYPE (*SigType)();

struct sSCEDitf *ft_sced = &sc_itf;

/* at last! */
int
main(ac, av)

int ac;
char **av;
{
    char    **tv;
    int     tc, i;
    int     err;
    bool    gotone = false;
    char    *p;
    char    *cmd_line_term = 0, term_1stch;

#ifdef BATCH
    bool    st = false;
#else

#ifdef HAVE_GETPWUID
    struct passwd *pw;
#endif
    char    buf[BSIZE_SP];
    bool    rflag = false, ciprefix();
    bool    istty = true, iflag = false, qflag = false;
    bool    readinit = true;
    bool    gdata = true;
    bool    read_err = false;
    FILE    *fp;
    extern char *kw_mplot_cur;

#endif

    /* MFB tends to jump to 0 on errors.  This tends to catch it. */
    if (started) {
        fprintf(cp_err, "main: Internal Error: jump to zero\n");
        fatal();
    }
    started = true;

    ivars( );

    cp_in = stdin;
    cp_out = stdout;
    cp_err = stderr;

#ifdef BATCH

    cp_curin = stdin;
    cp_curout = stdout;
    cp_curerr = stderr;

#else

#ifdef MALLOCTRACE
    mallocTraceInit("malloc.out");
#endif
#ifdef HAVE_ISATTY
    istty = (bool) isatty(fileno(stdin));
#endif

    rusage_init();
#endif

    err = SIMinit(&nutmeginfo,&ft_sim);
    if(err != OK) {
        ft_sperror(err,"SIMinit");
        return (EXIT_BAD);
    }
    cp_program = ft_sim->simulator;

#ifdef HAVE_GETPID
    srandom(getpid());
#else
    srandom(17);
#endif

    tv = av;
    tc = ac;

    /* Pass 1 -- get options. */
    while (--tc > 0) {
        tv++;
        if (**tv == Spice_OptChar) {   /* Option argument. */
            switch ((*tv)[1]) {

#ifndef BATCH
                case '\0':  /* No raw file. */
                        gdata = false;
                        break;

                case 'b':   /* Batch mode. */
                case 'B':
                        ft_batchmode = true;
                        break;

                case 's':   /* Server mode. */
                case 'S':
                        ft_servermode = true;
                        break;

                case 'i':   /* Interactive mode. */
                case 'I':
                        iflag = true;
                        break;

                case 'q':   /* No command completion. */
                case 'Q':
                        qflag = true;
                        break;

                case 'n':   /* Don't read .spiceinit. */
                case 'N':
                        readinit = false;
                        break;

                case 't':   /* Terminal type. */
                case 'T':
                        if (tc > 1) {
                            tc--;
                            tv++;
                            cmd_line_term = *tv;
                            term_1stch = **tv;
                            **tv = Spice_OptChar;
                        }
                        else {
                            fprintf(cp_err, usage, cp_program);
                            return (EXIT_BAD);
                        }
                        break;
                case 'd':   /* X display */
                case 'D':
                        if (tc > 1) {
                            tc--;
                            tv++;
                            cp_display = copy(*tv);
                            **tv = Spice_OptChar;
                        }
                        else {
                            fprintf(cp_err, usage, cp_program);
                            return (EXIT_BAD);
                        }
                        break;
                case 'r':   /* The rawfile. */
                case 'R':
                        if (tc > 1) {
                            tc--;
                            tv++;
                            ft_rawfile = copy(*tv);
                            cp_vset("rawfile", VT_STRING, *tv);
                            **tv = Spice_OptChar;
                        }
                        rflag = true;
                        break;

#else /* if BATCH */
                case 'r':   /* The rawfile. */
                case 'R':
                        if (tc > 1) {
                            tc--;
                            tv++;
                            ft_rawfile = copy(*tv);
                            **tv = Spice_OptChar;
                        }
                        else {
                            fprintf(cp_err, usage, cp_program);
                            return (EXIT_BAD);
                        }
#endif

                case 'o':   /* Output file. */
                case 'O':
                        if (tc > 1) {
                            tc--;
                            tv++;
                            if (!(freopen(*tv, "w", stdout))) {
                                perror(*tv);
                                return (EXIT_BAD);
                            }
                            **tv = Spice_OptChar;
                        }
                        else {
                            fprintf(cp_err, usage, av[0]);
                            return (EXIT_BAD);
                        }
                        break;

                default:
                        fprintf(cp_err, "Error: bad option %s\n", *tv);
                        fprintf(cp_err, usage, cp_program);
                        return (EXIT_BAD);
            }
        }
    }

    if (!cp_display) {
        cp_display = getenv("DISPLAY");
        if (cp_display)
            cp_display = copy(cp_display);
    }

#ifdef SIMULATOR
    if_getparam = spif_getparam;
#else
    if_getparam = nutif_getparam;
#endif

#ifndef BATCH

    if ((!iflag && !istty) || ft_servermode)
        ft_batchmode = true;
    if ((iflag && !istty) || qflag)
        cp_nocc = true;
    if (!istty || ft_batchmode)
        out_moremode = false;
    if (ft_servermode)
        ft_rawfile = "";
    cp_interactive = false;

#ifdef HAVE_SIGNAL
    /* Set up signal handling */
    if (!ft_batchmode) {
        (void) signal(SIGINT, (SigType)ft_sigintr);
        (void) signal(SIGFPE, (SigType)sigfloat);
#ifdef SIGTSTP
        (void) signal(SIGTSTP, (SigType)sigstop);
#endif
    }
    /* Set up signal handling for fatal errors. */
    (void) signal(SIGILL, (SigType)sigill);

#ifdef SIGBUS
    (void) signal(SIGBUS, (SigType)sigbus);
#endif
#ifdef SIGSEGV
    (void) signal(SIGSEGV, (SigType)sigsegv);
#endif
#ifdef SIGSYS
    (void) signal(SIGSYS, (SigType)sig_sys);
#endif
#endif

    /* To catch interrupts during .spiceinit... */
    if (setjmp(jbuf) == 1) {
        read_err = true;
        goto bot;
    }

    /* init graphics, so we can setrdb in init file */
    if (!ft_batchmode)
        DevInit( );

    /* Have to initialize cp now. */
    ft_cpinit();  /* this sources the spinit file */

    if (!ft_servermode && readinit) {
#ifdef HAVE_GETPWUID
        /* Try to source either .spiceinit or ~/.spiceinit. */
        if (access(".spiceinit", 0) == 0)
            inp_source(".spiceinit");
        else {
            pw = getpwuid(getuid());
            (void) strcpy(buf, pw->pw_dir);
            (void) strcat(buf, "/.spiceinit");
            if (access(buf, 0) == 0)
                inp_source(buf);
        }
#else
        /* Try to source the file "spice.rc" in the current directory.  */
        if ((fp = fopen("spice.rc", "r")) != NULL) {
            (void) fclose(fp);
            inp_source("spice.rc");
        }
#endif
    }

bot:

    if (cmd_line_term) {
        *cmd_line_term = term_1stch;
        /* XXX oh, gross!
         * First char got squashed scanning the
         * command line
         */
        cp_vset("term", VT_STRING, cmd_line_term);
        *cmd_line_term = Spice_OptChar;
    }

    if (!ft_batchmode) {
        com_version(NULL);
        if (News_File && *News_File) {
            fp = fopen(cp_tildexpand(News_File), "r");
            if (fp) {
                while (fgets(buf, BSIZE_SP, fp))
                    fputs(buf, stdout);
                (void) fclose(fp);
            }
        }
    }
    if (read_err)
        fprintf(cp_err, "Warning: error executing init file.\n");

    /* Pass 2 -- get the filenames. If we are spice, then this means
     * build a circuit for this file. If this is in server mode, don't
     * process any of these args.
     */

    if (setjmp(jbuf) == 1) {
        goto evl;
    }

#endif /* not BATCH */

    if (!ft_servermode && !ft_nutmeg) {
        FILE *file = NULL, *tp = NULL;
        char *tempfile = NULL, buf[BSIZE_SP], *smktemp();

        for (tv = av + 1, i = 0; *tv; tv++)
            if (**tv != Spice_OptChar)
                i++;
        if (i == 1) {
            for (tv = av + 1, i = 0; *tv; tv++)
                if (**tv != Spice_OptChar)
                    break;
            if (!(file = fopen(*tv, "r"))) {
                perror(*tv);
                i = 0;
            }
        }
        else if (i) {
            tempfile = smktemp("sp");
            if (!(file = fopen(tempfile, "w+"))) {
                perror(tempfile);
                return (EXIT_BAD);
            }
            for (tv = av + 1, i = 0; *tv; tv++)
                if (**tv != Spice_OptChar) {
                    if (!(tp = fopen(*tv, "r"))) {
                        perror(*tv);
                        continue;
                    }
                    while ((i = fread(buf, 1, BSIZE_SP, tp)) > 0)
                        (void) fwrite(buf, i, 1, file);
                    (void) fclose(tp);
                }
            (void) fseek(file, (long) 0, 0);
        }
        if (file) {
            inp_spsource(file, false, tempfile ? (char *) NULL : *tv);
            (void) fclose(file);
            gotone = true;
            if (tempfile)
                (void) unlink(tempfile);
        }
    }

    if (!gotone && ft_batchmode && !ft_nutmeg)
        inp_spsource(stdin, false, (char *) NULL);

#ifndef BATCH
    if (ft_nutmeg && gdata) {
        /* Read in the rawfiles */
        for (av++; *av; av++)
            if (**av != Spice_OptChar) {
                ft_loadfile(*av);
                gotone = true;
            }
        if (!gotone)
            ft_loadfile(ft_rawfile);
    }

evl:
    if (ft_batchmode) {
        /* If we get back here in batch mode then something is
         * wrong, so exit.
         */
        bool st = false;

        (void) setjmp(jbuf);

        if (st == true)
            return (EXIT_BAD);
        st = true;

        if (ft_curckt == NULL) {
            char p;
            if (cp_getvar(kw_mplot_cur,VT_STRING,&p))
                /* margin analysis performed */
                return (EXIT_NORMAL);
            fprintf(cp_err, "Error: no circuit loaded!\n");
            return (EXIT_BAD);
        }
        else if (ft_curckt->ci_runonce)
            /* already did it */
            return (EXIT_NORMAL);

        if (ft_servermode) {
            ft_dorun(ft_rawfile);
            return (EXIT_NORMAL);
        }
        /* If -r is specified, then we don't bother with the dot
         * cards. Otherwise, we use wrd_run, but we are careful
         * not to save too much.
         */
        cp_interactive = false;
        if (rflag) {
            ft_dotsaves();
            ft_dorun(ft_rawfile);
            ft_cktcoms(true);
        }
        else {
            ft_savedotargs();
            ft_dorun((char *) NULL);
            ft_cktcoms(false);
        }
    }
    else {
        (void) setjmp(jbuf);
        cp_interactive = true;
        while (cp_evloop((char *) NULL) == 1) ;
    }

#else /* if BATCH */

    if (st == true)
        return (EXIT_BAD);
    st = true;
    if (ft_curckt == NULL) {
        char p;
        if (cp_getvar(kw_mplot_cur,VT_STRING,&p))
            /* margin analysis performed */
            return (EXIT_NORMAL);
        fprintf(cp_err, "Error: no circuit loaded!\n");
        return (EXIT_BAD);
    }
    else if (ft_curckt->ci_runonce)
        /* already did it */
        return (EXIT_NORMAL);
#ifdef SPICE2
    ft_savedotargs( );
    ft_dorun(NULL);
    ft_cktcoms(false);
#else
    ft_dotsaves( );
    ft_dorun(ft_rawfile);
    ft_cktcoms(false);

    fprintf(stderr,
    "Note: \".plot\" or \".print\" lines ignored (rawfile only)\n");
#endif

#endif

    return (EXIT_NORMAL);
}


static void
clear_previous()

{
    /* for SCED, we don't save old circuits or plots */
#ifdef SIMULATOR
    if_cktclear();
    com_destroy((wordlist *)NULL);
#endif
}


/* allocate space for global constants in 'CONST.h' */

double CONSTroot2;
double CONSTvt0;
double CONSTKoverQ;
double CONSTe;
IFfrontEnd *SPfrontEnd = NULL;

int
SPIinit(frtEnd,description)

IFfrontEnd *frtEnd;
IFsimulator **description;
{

    SPfrontEnd = frtEnd;
    *description = &SIMinfo;
    CONSTroot2 = sqrt(2.);
    CONSTvt0 = CONSTboltz * (27 /* deg c */ + CONSTCtoK ) / CHARGE;
    CONSTKoverQ = CONSTboltz / CHARGE;
    CONSTe = exp((double)1.0);
    return (OK);
}

/* XXX SIMinit and SPIinit ?? */

int
SIMinit(frontEnd,simulator)

IFfrontEnd *frontEnd;
IFsimulator **simulator;
{
    return (SPIinit(frontEnd,simulator));
}

#ifdef BATCH
/*
 * Incredibly Ugly
 */

/* Now some misc junk that we need to fake */

bool cp_nocc = true;
bool cp_debug = false;
int cp_maxhistlength;
char cp_chars[128];

MFBHalt( ) { }

#ifndef SPICE2
/* ARGSUSED */ char *cp_tildexpand(s) char *s; { return (s); }
/* ARGSUSED */ struct dvec *vec_fromplot(w, p) char *w; struct plot *p;
        { return (NULL); }
void vec_new() {}
void plot_new() {}
void plot_setcur() {}
/* ARGSUSED */ struct plot *plot_alloc(name) char *name; { return (NULL); }
struct plot *plot_cur = NULL;
/* ARGSUSED */ void plot_docoms(wl) wordlist *wl; {}
void out_init() {}
/* PC defines out_printf to printf since vsprintf doesn't exist */
/* ARGSUSED */ /* VARARGS1 */ void out_printf(c) char *c; {}
/* ARGSUSED */ void out_send(s) char *s; {}
char out_pbuf[1];
/* ARGSUSED */ struct dvec *vec_get(word) char *word; { return (NULL); }
#endif

/* ARGSUSED */ int cp_usrset(v, i) struct variable *v; bool i; {return(US_OK);}
void cp_pushcontrol() {}
void cp_popcontrol() {}
/* ARGSUSED */ void cp_addkword(class, word) int class; char *word; {}
struct circ *ft_circuits = 0, *ft_curckt = 0;
/* ARGSUSED */ char *cp_kwswitch(c, t) int c; char *t; { return (NULL); }
/* ARGSUSED */ bool ft_bpcheck(r, i) struct plot *r; int i; { return (true); }
/* ARGSUSED */ void cp_ccon(b) bool b; {}

int currentgraph = 0;
/* ARGSUSED */ void Input(p, q) char *p, *q; {}
/* ARGSUSED */ void DevSwitch(i) int i; {}
/* ARGSUSED */ int *CopyGraph(i) int i; {}
/* ARGSUSED */ void DestroyGraph(i) int i; {}
/* ARGSUSED */ void NewViewport(i) int i; {}

/* ARGSUSED */ void cp_remkword() { }
/* ARGSUSED */ void cp_resetcontrol() { }

bool gr_init() {}
void gr_redraw() {}
/* ARGSUSED */ void gr_iplot(pl) struct plot *pl; {}
/* ARGSUSED */ void gr_end_iplot(pl) struct plot *pl; {}
/* ARGSUSED */ void gr_pmsg(text, more) char *text; bool more; {}
/* ARGSUSED */ void gr_clean(intr) bool intr; {}
/* ARGSUSED */ int cp_evloop(string) char *string; { return (0); }
void ft_graf() {}
void ft_trquery() {}
/* ARGSUSED */ void ft_newcirc(ckt) struct circ *ckt; {}
void cp_doquit() { exit(0); }
/* ARGSUSED */ void cp_usrvars(v1, v2) struct variable **v1, **v2; { return; }
/* ARGSUSED */ struct variable * cp_enqvar(word) char *word; { return (NULL); }
/* ARGSUSED */ void cp_ccom(w, b, e) wordlist *w; char *b; bool e; { return; }

void
com_save(wl)
wordlist *wl;
{
    return;
}

int
ft_getSaves(savesp)
char ***savesp;
{
    return (0);
}

/* This is from options.c */
/* Extract the .option cards from the deck... */

struct line *
inp_getopts(deck)

struct line *deck;
{
    struct line *last = NULL, *opts = NULL, *dd, *next = NULL;

    for (dd = deck->li_next; dd; dd = next) {
        next = dd->li_next;
        if (ciprefix(".opt", dd->li_line)) {
            inp_casefix(dd->li_line);
            if (last)
                last->li_next = dd->li_next;
            else
                deck->li_next = dd->li_next;
            dd->li_next = opts;
            opts = dd;
        }
        else
            last = dd;
    }
    return (opts);
}

#ifndef SPICE2

static wordlist *gettoks();

/* This is from dotcards.c -- we don't want to include the whole file. */
/* Extract all the .save cards */

void
ft_dotsaves()

{
    wordlist *com, *wl = NULL;
    char *s;

    if (!ft_curckt) /* Shouldn't happen. */
        return;

    for (com = ft_curckt->ci_commands; com; com = com->wl_next) {
        if (ciprefix(".save", com->wl_word)) {
            s = com->wl_word;
            (void) gettok(&s);
            wl = wl_append(wl, gettoks(s));
        }
    }
/*  wrd_saves = wl; */

    com_save(wl);
/* fprintf(stderr, "error, .save card code is broken\n"); */
    return;
}

/* This is also from dotcards.c.  We should probably include the file. */
/* Stripped down version. */

/* Execute the .whatever cards found in the deck, after we are done running.
 * We'll be cheap and use cp_lexer to get the words... This should make us
 * spice-2 compatible.  Is terse is true then there was a rawfile, so don't
 * print lots of junk.
 */

void
ft_cktcoms(terse)
bool terse;

{
    wordlist *coms, *command;
    char *plottype, *s;
    struct dvec *v;
    static wordlist twl = { "col", NULL, NULL } ;
    int i;

    if (!ft_curckt)
        return;
    if (!ft_curckt->ci_commands)
    goto nocmds;
    coms = ft_curckt->ci_commands;
    cp_interactive = false;

nocmds:
    /* Circuit name */
    fprintf(cp_out, "Circuit: %s\nDate: %s\n\n", ft_curckt->ci_name,
            datestring());
    fprintf(cp_out, "\n");

    /* Listing if necessary... */
    if (ft_listprint)
        inp_list(cp_out, ft_curckt->ci_deck, ft_curckt->ci_options,
                LS_DECK);

    /* And finally the accounting info. */
    if (ft_acctprint) {
        static wordlist ww = { "everything", NULL, NULL } ;
        com_rusage(&ww);
    } else
        com_rusage((wordlist *) NULL);

    (void) putc('\n', cp_out);
    return;

bad:    fprintf(cp_err, "Internal Error: ft_cktcoms: bad commands\n");
    return;
}


static wordlist *
gettoks(s)

char *s;
{
    char *t, *r, buf[64];
    wordlist *wl = NULL, *end = NULL;
    bool iflag;

    while (t = gettok(&s)) {
        if (*t == '(' /* ) */) {
            /* This is a (upper, lower) thing -- ignore. */
            continue;
        }
        else if (!index(t, '(' /*)*/ )) {
            if (end) {
                end->wl_next = alloc(struct wordlist);
                end->wl_next->wl_prev = end;
                end = end->wl_next;
            }
            else
                wl = end = alloc(struct wordlist);
            end->wl_word = copy(t);
        }
        else if (!index(t, ',')) {
            iflag = ((*t == 'i') || (*t == 'I')) ? true : false;
            while (*t != '(' /*)*/)
                t++;
            t++;
            for (r = t; *r && *r != /*(*/ ')'; r++)
                ;
            *r = '\0';
            if (end) {
                end->wl_next = alloc(struct wordlist);
                end->wl_next->wl_prev = end;
                end = end->wl_next;
            }
            else
                wl = end = alloc(struct wordlist);
            if (iflag) {
                (void) sprintf(buf, "%s#branch", t);
                t = buf;
            }
            end->wl_word = copy(t);
        }
        else {
            /* The painful case... */
            while (*t != '(' /*)*/)
                t++;
            t++;
            for (r = t; *r && *r != ','; r++)
                ;
            *r = '\0';
            if (end) {
                end->wl_next = alloc(struct wordlist);
                end->wl_next->wl_prev = end;
                end = end->wl_next;
            }
            else
                wl = end = alloc(struct wordlist);
            end->wl_word = copy(t);
            t = r + 1;
            for (r = t; *r && *r != /*(*/ ')'; r++)
                ;
            *r = '\0';
            if (end) {
                end->wl_next = alloc(struct wordlist);
                end->wl_next->wl_prev = end;
                end = end->wl_next;
            }
            else
                wl = end = alloc(struct wordlist);
            end->wl_word = copy(t);
        }
    }
    return (wl);
}

#endif

#endif
