/**********
Copyright 1990 Regents of the University of California.  All rights reserved.
**********/

#include "spice.h"
#define CONFIG

#include <stdio.h>
#include "devdefs.h"
#include "cktext.h"

#include "jobdefs.h"
#include "acdefs.h"
#include "distodef.h"
#include "dcodefs.h"
#include "optdefs.h"
#include "noisedef.h"
#include "pzdefs.h"
#include "tfdefs.h"
#include "trandefs.h"
#include "dctdefs.h"

#ifndef AN_dc

#ifdef AN_ac
#define AN_dc
#endif
#ifdef AN_pz
#define AN_dc
#endif
#ifdef AN_tran
#define AN_dc
#endif
#ifdef AN_noise
#define AN_dc
#endif
#ifdef AN_disto
#define AN_dc
#endif
#ifdef AN_sense
#define AN_dc
#endif

#endif

extern SPICEanalysis OPTinfo;
extern SPICEanalysis ACinfo;
extern SPICEanalysis DCTinfo;
extern SPICEanalysis DCOinfo;
extern SPICEanalysis TRANinfo;
extern SPICEanalysis PZinfo;
extern SPICEanalysis SENSinfo;
extern SPICEanalysis TFinfo;
extern SPICEanalysis DISTOinfo;
extern SPICEanalysis NOISEinfo;

SPICEanalysis *analInfo[] = {
    &OPTinfo,
#ifdef AN_ac
    &ACinfo,
#endif
#ifdef AN_dc
    &DCTinfo,
#endif
#ifdef AN_op
    &DCOinfo,
#endif
#ifdef AN_tran
    &TRANinfo,
#endif
#ifdef AN_pz
    &PZinfo,
#endif
#ifdef AN_sense
    &SENSinfo,
#endif
#ifdef AN_tf
    &TFinfo,
#endif
#ifdef AN_disto
    &DISTOinfo,
#endif
#ifdef AN_noise
    &NOISEinfo,
#endif
};

int ANALmaxnum = sizeof(analInfo)/sizeof(SPICEanalysis*);

extern SPICEdev BJTinfo;
extern SPICEdev B1info;
extern SPICEdev B2info;
extern SPICEdev CAPinfo;
extern SPICEdev DIOinfo;
extern SPICEdev INDinfo;
extern SPICEdev JFETinfo;
extern SPICEdev JJinfo;
extern SPICEdev LTRAinfo;
extern SPICEdev MESinfo;
extern SPICEdev MOSinfo;
extern SPICEdev MUTinfo;
extern SPICEdev RESinfo;
extern SPICEdev SRCinfo;
extern SPICEdev SWinfo;
extern SPICEdev TRAinfo;
extern SPICEdev URCinfo;
extern SPICEdev SFFTinfo;

SPICEdev *DEVices[] = {

#ifdef DEV_bjt
        &BJTinfo,
#endif
#ifdef DEV_bsim1
        &B1info,
#endif
#ifdef DEV_bsim2
        &B2info,
#endif
#ifdef DEV_cap
        &CAPinfo,
#endif
#ifdef DEV_dio
        &DIOinfo,
#endif
#ifdef DEV_ind
        &INDinfo,
        &MUTinfo,
#endif
#ifdef DEV_jfet
        &JFETinfo,
#endif
#ifdef DEV_jj
        &JJinfo,
#endif
#ifdef DEV_ltra
        &LTRAinfo,
#endif
#ifdef DEV_mes
        &MESinfo,
#endif
#ifdef DEV_mos
        &MOSinfo,
#endif
#ifdef DEV_res
        &RESinfo,
#endif
#ifdef DEV_src
        &SRCinfo,
#endif
#ifdef DEV_sw
        &SWinfo,
#endif
#ifdef DEV_tra
        &TRAinfo,
#endif
#ifdef DEV_urc
        &URCinfo,
#endif
#ifdef DEV_sfft
        &SFFTinfo,
#endif
};

/* my internal global constant for number of device types */
int DEVmaxnum = sizeof(DEVices)/sizeof(SPICEdev *);

static char * specSigList[] = {
    "time"
};

static IFparm nodeParms[] = {
    IP( "nodeset",PARM_NS ,IF_REAL,"suggested initial voltage"),
    IP( "ic",PARM_IC ,IF_REAL,"initial voltage"),
    IP( "type",PARM_NODETYPE ,IF_INTEGER,"output type of equation")
};

IFsimulator SIMinfo = {
    "jspice3",        /* name */
    "jspice3 circuit level simulation program",  /* more about me */
    Spice_Version,  /* version */

    CKTinit,        /* newCircuit function */
    CKTdestroy,     /* deleteCircuit function */

    CKTnewNode,     /* newNode function */
    CKTground,      /* groundNode function */
    CKTbindNode,    /* bindNode function */
    CKTfndNode,     /* findNode function */
    CKTinst2Node,   /* instToNode function */
    CKTsetNodPm,    /* setNodeParm function */
    CKTaskNodQst,   /* askNodeQuest function */
    CKTdltNod,      /* deleteNode function */

    CKTcrtElt,      /* newInstance function */
    CKTparam,       /* setInstanceParm function */
    CKTask,         /* askInstanceQuest function */
    CKTfndDev,      /* findInstance funciton */
    CKTdltInst,     /* deleteInstance function */

    CKTmodCrt,      /* newModel function */
    CKTmodParam,    /* setModelParm function */
    CKTmodAsk,      /* askModelQuest function */
    CKTfndMod,      /* findModel function */
    CKTdltMod,      /* deleteModel function */

    CKTnewTask,     /* newTask function */
    CKTnewAnal,     /* newAnalysis function */
    CKTsetAnalPm,   /* setAnalysisParm function */
    CKTaskAnalQ,    /* askAnalysisQuest function */
    CKTfndAnal,     /* findAnalysis function */
    CKTfndTask,     /* findTask function */
    CKTdelTask,     /* deleteTask function */

    CKTdoJob,       /* doAnalyses function */
    CKTtrouble,     /* non-convergence message function */

    sizeof(DEVices)/sizeof(SPICEdev *),
    (IFdevice**)DEVices,

    sizeof(analInfo)/sizeof(SPICEanalysis *),
    (IFanalysis **)analInfo,

    sizeof(nodeParms)/sizeof(IFparm),
    nodeParms,

    sizeof(specSigList)/sizeof(char *),
    specSigList,

};
