/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

#define ANAL_EXT

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "srcdefs.h"
#include "sperror.h"
#include "outdata.h"
#include "acdefs.h"
#include "util.h"
#include "cktext.h"
#include "niext.h"

#ifdef __STDC__
static int ac_init(CKTcircuit*,struct sOUTdata*);
static int ac_dcoperation(CKTcircuit*,struct sOUTdata*,int);
static int ac_operation(CKTcircuit*,struct sOUTdata*,int);
#else
static int ac_init();
static int ac_dcoperation();
static int ac_operation();
#endif


int
ACan(cktp,restart)

GENERIC *cktp;
int restart;
{
    CKTcircuit *ckt = (CKTcircuit *)cktp;
    static struct sOUTdata outd;
    int error;

    if (ckt->CKTjjPresent) {
        (*(SPfrontEnd->IFerror))(ERR_FATAL,
            "AC analysis not possible with Josephson junctions", NULL);
            return (OK);
    }

    if (restart) {
        error = ac_init(ckt,&outd);
        if (error)
            return (error);
    }
    error = DCTloop(ac_dcoperation,ckt,restart,
        &((ACAN*)ckt->CKTcurJob)->DC,&outd);
    if (error == E_PAUSE)
        return (error);

    (*(SPfrontEnd->OUTendPlot))(*outd.plotPtr);
    return (error);
}


static int
ac_dcoperation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    int error;

    error = CKTic(ckt);
    if (error)
        return (error);

    error = CKTop(ckt,
        MODEDCTRANCURVE | MODEINITJCT,
        MODEDCTRANCURVE | MODEINITFLOAT,
        ckt->CKTdcMaxIter);
    if (error)
        return (error);

    ckt->CKTmode = MODEDCOP | MODEINITSMSIG;

    error = CKTload(ckt);
    if (error)
        return (error);

    error = ACloop(ac_operation,ckt,restart,
        &((ACAN*)ckt->CKTcurJob)->AC,outd);
    if (error)
        return (error);

    return (OK);
}


static int
ac_init(ckt,outd)

CKTcircuit *ckt;
struct sOUTdata *outd;
{
    int i;
    int code;
    int error;
    ACAN *job = (ACAN*)ckt->CKTcurJob;

    if (job->DC.eltName[0] != NULL) {
        /* DC source was given */
        code = CKTtypelook("Source");

        for (i = 0; i <= job->DC.nestLevel; i++) {
            SRCinstance *here;

            here = (SRCinstance*)NULL;
            error = CKTfndDev((GENERIC*)ckt,&code,(GENERIC**)&here,
                job->DC.eltName[i], (GENERIC *)NULL,(IFuid)NULL);

            if (error) {
                (*(SPfrontEnd->IFerror))(ERR_FATAL, 
                    "DCtrCurv: source %s not in circuit",
                    &(job->DC.eltName[i]));
                return (E_NOTFOUND);
            }
            job->DC.elt[i] = (GENinstance*)here;
            job->DC.vsave[i] = here->SRCdcValue;
        }
    }

    error = CKTnames(ckt,&outd->numNames,&outd->dataNames);
    if (error)
        return (error);

    (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&outd->refName,(IFuid)NULL,
            "frequency", UID_OTHER,(GENERIC **)NULL);

    outd->circuitPtr = (GENERIC *)ckt;
    outd->analysisPtr = (GENERIC*)ckt->CKTcurJob;
    outd->analName = ckt->CKTcurJob->JOBname;
    outd->refType = IF_REAL;
    outd->dataType = IF_COMPLEX;
    outd->plotPtr = &job->ACplot;
    outd->numPts = 1;
    outd->initValue = ((ACAN*)ckt->CKTcurJob)->AC.fstart;
    outd->finalValue = ((ACAN*)ckt->CKTcurJob)->AC.fstop;
    outd->step = 0;
    outd->count = 0;
    (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)outd);
    return (OK);
}


/* ARGSUSED */
static int
ac_operation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    int error;

    error = NIacIter(ckt);
    if (error)
        return (error);

    error = CKTacDump(ckt,ckt->CKTomega/(2*M_PI),*outd->plotPtr);
    if (error)
        return (error);
    outd->count++;
    return (OK);
}


int
ACloop(func,ckt,restart,ac,outd)

#ifdef __STDC__
int (*func)(CKTcircuit*,struct sOUTdata*,int);
#else
int (*func)();
#endif
CKTcircuit *ckt;
int restart;
struct sACprms *ac;
struct sOUTdata *outd;
{
    double freq;
    double freqTol, freqDel;
    int error;

    switch (ac->stepType) {
        case DECADE:
            if (ac->fstart <= 0.0)
                ac->fstart = 1e-3;
            if (ac->fstop <= ac->fstart)
                ac->fstop = 10.0 * ac->fstart;
            freqDel = exp(log(10.0)/ac->numSteps);
            freqTol = freqDel * ac->fstop * ckt->CKTreltol;
        case OCTAVE:
            if (ac->fstart <= 0.0)
                ac->fstart = 1e-3;
            if (ac->fstop <= ac->fstart)
                ac->fstop = 2.0 * ac->fstart;
            freqDel = exp(log(2.0)/ac->numSteps);
            freqTol = freqDel * ac->fstop * ckt->CKTreltol;
            break;
        case LINEAR:
            freqDel = (ac->fstop - ac->fstart)/(ac->numSteps+1);
            freqTol = freqDel * ckt->CKTreltol;
            break;
        case DCSTEP:
            return ( (*func)(ckt,outd,restart) );
        default:
            return (E_BADPARM);
    }

    if (restart) {
        /* reset static variables */
        ac->fsave = ac->fstart;
    }

    freq = ac->fsave;

    while (freq <= ac->fstop + freqTol) {

        if ( (*(SPfrontEnd->IFpauseTest))() ) { 
            /* user asked us to pause via an interrupt */
            ac->fsave = freq;
            return (E_PAUSE);
        }
        ckt->CKTomega = 2.0 * M_PI * freq;
        ckt->CKTmode = MODEAC;

        error = (*func)(ckt,outd,restart);
        if (error) {
            ac->fsave = freq;
            return (error);
        }
        if (*SPfrontEnd->OUTendit) {
            *SPfrontEnd->OUTendit = 0;
            return (OK);
        }

        /*  increment frequency */

        switch (ac->stepType) {
            case DECADE:
            case OCTAVE:
                freq *= freqDel;
                if (freqDel == 1)
                    return (OK);
                break;
            case LINEAR:
                freq += freqDel;
                if (freqDel == 0)
                    return (OK);
                break;
            default:
                return (E_INTERN);
        }
    }

    return (OK);
}


int
CKTacLoad(ckt)

CKTcircuit *ckt;
/* CKTacLoad(ckt)
 * this is a driver program to iterate through all the various
 * ac load functions provided for the circuit elements in the
 * given circuit 
 */
{
    extern SPICEdev *DEVices[];
    int i;
    int size;
    int error;
    struct sCKTmodHead *mh;
    int (*func)();

    size = spGetSize(ckt->CKTmatrix,1);
    for (i = 0; i <= size; i++) {
        *(ckt->CKTrhs+i) = 0;
        *(ckt->CKTirhs+i) = 0;
    }
    spSetComplex(ckt->CKTmatrix);
    spClear(ckt->CKTmatrix);

    for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
        if ((func = DEVices[mh->type]->DEVacLoad) != NULL) {
            error = (*func)(mh->head,ckt);
            if (error)
                return (error);
        }
    }
    return (OK);
}
