/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "ftedefs.h"
#include "inpdefs.h"
#include "inpmacs.h"
#include "misc.h"


/* ARGSUSED */
int
ACparse(ckt,tab,which,currentp,line,task,gnode)

/*
    .ac {DEC OCT LIN} NP FSTART FSTOP [ dc SRC1NAME Vstart1 [Vstop1 [Vinc1]]
        [SRC2NAME Vstart2 [Vstop2 [Vinc2]]] ]
*/
GENERIC *ckt;
INPtables *tab;
int which;
GENERIC *currentp;
char **line;
GENERIC *task;
GENERIC *gnode;
{
    GENERIC *foo;   /* pointer to analysis */
    int error;      /* error code temporary */
    char *token;    /* a token from the line */
    card *current = (card *)currentp;

    IFC(newAnalysis,(ckt,which,"acan",&foo,task))
    GCA(ParseAC,(ckt,line,current,which,foo,tab))
    if (**line) {
        INPgetTok(line, &token, 1);
        strtolower(token);
        if (!strcmp(token,"dc")) {
            txfree(token);
            GCA(ParseDC,(ckt,line,current,which,foo,tab,1))
            if (**line)
                GCA(ParseDC,(ckt,line,current,which,foo,tab,2))
        }
        else {
            LITERR("Syntax error: 'dc' expected.")
        }
    }

    return (0);
}


int
ParseAC(ckt,line,current,which,aptr,tab)

GENERIC *ckt;
char **line;
card *current;
int which;
GENERIC *aptr;
INPtables *tab;
{
    char *steptype; /* ac analysis, type of stepping function */
    IFvalue ptemp;  /* a value structure to package stuff into */
    IFvalue *parm;  /* a pointer to a value struct for function returns */
    int error;      /* error code temporary */

    INPgetTok(line,&steptype,1); /* get DEC, OCT, or LIN */
    strtolower(steptype);
    ptemp.iValue = 1;
    GCA(INPapName,(ckt,which,aptr,steptype,&ptemp))
    txfree(steptype);

    parm = INPgetValue(ckt,line,IF_INTEGER,tab); /* number of points */
    GCA(INPapName,(ckt,which,aptr,"numsteps",parm))

    parm = INPgetValue(ckt,line,IF_REAL,tab);    /* fstart */
    GCA(INPapName,(ckt,which,aptr,"start",parm))

    parm = INPgetValue(ckt,line,IF_REAL,tab);    /* fstop */
    GCA(INPapName,(ckt,which,aptr,"stop",parm))

    return (0);
}
