/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "acdefs.h"
#include "jobdefs.h"
#include "iferrmsg.h"


/* ARGSUSED */
int 
ACsetParm(cktp,anal,which,value)

GENERIC *cktp;
GENERIC *anal;
int which;
IFvalue *value;
{
    ACAN *job = (ACAN *) anal;

    if (ACsetp(&job->AC,which,value) == OK)
        return (OK);
    if (DCTsetp(&job->DC,which,value) == OK)
        return (OK);
    return (E_BADPARM);
}


int 
ACsetp(ac,which,value)

struct sACprms *ac;
int which;
IFvalue *value;
{
    switch (which) {

        case AC_START:
            ac->fstart = value->rValue;
            break;

        case AC_STOP:
            ac->fstop = value->rValue;
            break;

        case AC_STEPS:
            ac->numSteps = value->iValue;
            break;

        case AC_DEC:
            if (value->iValue) {
                ac->stepType = DECADE;
            }
            else {
                if (ac->stepType == DECADE) {
                    ac->stepType = 0;
                }
            }
            break;

        case AC_OCT:
            if (value->iValue) {
                    ac->stepType = OCTAVE;
            }
            else {
                if (ac->stepType == OCTAVE) {
                    ac->stepType = 0;
                }
            }
            break;

        case AC_LIN:
            if (value->iValue) {
                ac->stepType = LINEAR;
            }
            else {
                if (ac->stepType == LINEAR) {
                    ac->stepType = 0;
                }
            }
            break;

        default:
            return (E_BADPARM);
    }
    return (OK);
}



static IFparm ACparms[] = {
 { "start",    AC_START,  IFP|IF_REAL,     "starting frequency" },
 { "stop",     AC_STOP,   IFP|IF_REAL,     "ending frequency" },
 { "numsteps", AC_STEPS,  IFP|IF_INTEGER,  "number of frequencies"},
 { "dec",      AC_DEC,    IFP|IF_FLAG,     "step by decades" },
 { "oct",      AC_OCT,    IFP|IF_FLAG,     "step by octaves" },
 { "lin",      AC_LIN,    IFP|IF_FLAG,     "step linearly" },
 { "name1",    DC_NAME1,  IFP|IF_INSTANCE, "name of source to step" },
 { "start1",   DC_START1, IFP|IF_REAL,     "starting voltage/current"},
 { "stop1",    DC_STOP1,  IFP|IF_REAL,     "ending voltage/current" },
 { "step1",    DC_STEP1,  IFP|IF_REAL,     "voltage/current step" },
 { "name2",    DC_NAME2,  IFP|IF_INSTANCE, "name of source to step" },
 { "start2",   DC_START2, IFP|IF_REAL,     "starting voltage/current"},
 { "stop2",    DC_STOP2,  IFP|IF_REAL,     "ending voltage/current" },
 { "step2",    DC_STEP2,  IFP|IF_REAL,     "voltage/current step" }
};

SPICEanalysis ACinfo  = {
    { 
        "AC",
        "A.C. Small signal analysis",

        sizeof(ACparms)/sizeof(IFparm),
        ACparms
    },
    sizeof(ACAN),
    FREQUENCYDOMAIN,
    ACparse,
    ACsetParm,
    ACaskQuest,
    ACan
};
