/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "devdefs.h"
#include "tskdefs.h"
#include "dcodefs.h"
#include "sperror.h"
#include "outdata.h"
#include "util.h"
#include "cktext.h"


/* ARGSUSED */
int
DCOan(cktp,restart)

GENERIC *cktp;
int restart;
{
    CKTcircuit *ckt = (CKTcircuit *)cktp;
    struct sOUTdata outd;
    int converged;
    int error;
    GENERIC *plot;
    
    if (ckt->CKTjjPresent) {
        (*(SPfrontEnd->IFerror))(ERR_FATAL,
            "DC analysis not possible with Josephson junctions", NULL);
            return (OK);
    }
    error = CKTnames(ckt,&outd.numNames,&outd.dataNames);
    if (error)
        return (error);

    outd.circuitPtr = (GENERIC *)ckt;
    outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
    outd.analName = ckt->CKTcurJob->JOBname;
    outd.refName = (IFuid)NULL;
    outd.refType = IF_REAL;
    outd.dataType = IF_REAL;
    outd.plotPtr = &plot;
    outd.numPts = 1;
    outd.initValue = 0;
    outd.finalValue = 0;
    outd.step = 0;
    (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

    error = CKTic(ckt);
    if (error)
        return (error);

    converged = CKTop(ckt,
            MODEDCOP | MODEINITJCT,
            MODEDCOP | MODEINITFLOAT,
            ckt->CKTdcMaxIter);

    if (converged != 0)
        return (converged);

    ckt->CKTmode = MODEDCOP | MODEINITSMSIG;
    converged = CKTload(ckt);
    CKTdump(ckt,(double)0,plot);
    (*(SPfrontEnd->OUTendPlot))(plot);

    return (converged);
}


int
CKTop(ckt,firstmode, continuemode, iterlim)

CKTcircuit *ckt;
long firstmode;    
long continuemode;    
int iterlim;
{
    int converged;
    int i;

    ckt->CKTmode = firstmode;
    if (!ckt->CKTnoOpIter) {
        converged = NIiter(ckt,iterlim);
    }
    else {
        converged = 1;  /* the 'go directly to gmin stepping' option */
    }
    if (converged != 0) {
        /* no convergence on the first try, so we do something else */
        /* first, check if we should try gmin stepping */
        /* note that no path out of this code allows ckt->CKTdiagGmin to be 
         * anything but 0.000000000
         */
        if (ckt->CKTnumGminSteps >1) {
            ckt->CKTmode = firstmode;
            (*(SPfrontEnd->IFerror))(ERR_INFO,
                    "starting Gmin stepping",(IFuid *)NULL);
            ckt->CKTdiagGmin = ckt->CKTgmin;
            for (i = 0; i < ckt->CKTnumGminSteps; i++) {
                ckt->CKTdiagGmin *= 10;
            }
            for (i = 0; i <= ckt->CKTnumGminSteps; i++) {
                ckt->CKTnoncon = 1;
                converged = NIiter(ckt,iterlim);
                if (converged != 0) {
                    ckt->CKTdiagGmin = 0;
                    (*(SPfrontEnd->IFerror))(ERR_WARNING,
                            "Gmin step failed",(IFuid *)NULL);
                    break;
                }
                ckt->CKTdiagGmin /= 10;
                ckt->CKTmode=continuemode;
                (*(SPfrontEnd->IFerror))(ERR_INFO,
                        "One sucessful Gmin step",(IFuid *)NULL);
            }
            ckt->CKTdiagGmin = 0;
            converged = NIiter(ckt,iterlim);
            if (converged == 0) {
                (*(SPfrontEnd->IFerror))(ERR_INFO,
                        "Gmin stepping completed",(IFuid *)NULL);
                return (0);
            }
            (*(SPfrontEnd->IFerror))(ERR_WARNING,
                    "Gmin stepping failed",(IFuid *)NULL);

        }
        /* now, we'll try source stepping - we scale the sources
         * to 0, converge, then start stepping them up until they
         * are at their normal values
         *
         * note that no path out of this code allows ckt->CKTsrcFact to be 
         * anything but 1.000000000
         */
        if (ckt->CKTnumSrcSteps > 1) {
            ckt->CKTmode = firstmode;
            (*(SPfrontEnd->IFerror))(ERR_INFO,
                    "starting source stepping",(IFuid *)NULL);
            for (i = 0; i <= ckt->CKTnumSrcSteps; i++) {
                ckt->CKTsrcFact = ((double)i)/((double)ckt->CKTnumSrcSteps);
                converged = NIiter(ckt,iterlim);
                ckt->CKTmode = continuemode;
                if (converged != 0) {
                    ckt->CKTsrcFact = 1;
                    ckt->CKTcurrentAnalysis = DOING_TRAN;
                    (*(SPfrontEnd->IFerror))(ERR_WARNING,
                            "source stepping failed",(IFuid *)NULL);
                    return (converged);
                }
                (*(SPfrontEnd->IFerror))(ERR_INFO,
                        "One successful source step",(IFuid *)NULL);
            }
            (*(SPfrontEnd->IFerror))(ERR_INFO,
                    "Source stepping completed",(IFuid *)NULL);
            ckt->CKTsrcFact = 1;
            return (0);
        }
        else {
            return (converged);
        }
    } 
    return (0);
}


/* CKTconvTest(ckt)
 *    this is a driver program to iterate through all the various
 *    convTest functions provided for the circuit elements in the
 *    given circuit 
 */

int
CKTconvTest(ckt)

CKTcircuit *ckt;
{
    extern SPICEdev *DEVices[];
    int error;
    struct sCKTmodHead *mh;
    int (*func)();

    for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
        if ((func = DEVices[mh->type]->DEVconvTest) != NULL) {
            error = (*func)(mh->head,ckt);
            if (error) return (error);
        }
        if (ckt->CKTnoncon) {
            /*
            printf("convTest: device %s failed\n",
                (*DEVices[mh->type]).DEVpublic.name);
            */
            return (OK);
        }
    }
    return (OK);
}
