/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

#define ANAL_EXT

#include "spice.h"
#include "misc.h"
#include <stdio.h>
#include "srcdefs.h"
#include "sperror.h"
#include "outdata.h"
#include "dctdefs.h"
#include "util.h"
#include "cktext.h"
#include "niext.h"

#ifdef __STDC__
static int dct_init(CKTcircuit*,struct sOUTdata*);
static int dct_operation(CKTcircuit*,struct sOUTdata*,int);
#else
static int dct_init();
static int dct_operation();
#endif


int
DCTan(cktp,restart)

GENERIC *cktp;
int restart;    /* forced restart flag */
{
    CKTcircuit *ckt = (CKTcircuit *)cktp;
    DCTAN *job = (DCTAN*)ckt->CKTcurJob;
    static struct sOUTdata outd;
    int i, error;

    if (ckt->CKTjjPresent) {
        (*(SPfrontEnd->IFerror))(ERR_FATAL,
            "DC analysis not possible with Josephson junctions", NULL);
            return (OK);
    }
    if (restart) {
        error = dct_init(ckt,&outd);
        if (error)
            return (error);
    }

    error = DCTloop(dct_operation,ckt,restart,&job->DC,&outd);

    if (error == E_PAUSE)
        return (error);

    (*(SPfrontEnd->OUTendPlot))(*outd.plotPtr);

    return (error);
}


static int
dct_init(ckt,outd)

CKTcircuit *ckt;
struct sOUTdata *outd;
{
    DCTAN *job = (DCTAN*)ckt->CKTcurJob;
    int i;
    int code;
    int error;
    char *swp;

    code = CKTtypelook("Source");
    ckt->CKTtime = 0;
    ckt->CKTdelta = job->DC.vstep[0];
    ckt->CKTmode = MODEDCTRANCURVE | MODEINITJCT;
    ckt->CKTorder = 1;

    for (i = 0; i < 7; i++) {
        ckt->CKTdeltaOld[i] = ckt->CKTdelta;
    }
    for (i = 0; i <= job->DC.nestLevel; i++) {
        SRCinstance *here;

        here = (SRCinstance*)NULL;
        error = CKTfndDev((GENERIC*)ckt,&code,(GENERIC**)&here,
            job->DC.eltName[i], (GENERIC *)NULL,(IFuid)NULL);

        if (error) {
            (*(SPfrontEnd->IFerror))(ERR_FATAL, 
                "DCtrCurv: source %s not in circuit",
                &(job->DC.eltName[i]));
            return (E_NOTFOUND);
        }
        job->DC.elt[i] = (GENinstance*)here;
        job->DC.vsave[i] = here->SRCdcValue;
    }
    error = CKTnames(ckt,&outd->numNames,&outd->dataNames);
    if (error)
        return (error);

    if (*(char*)(job->DC.eltName[0]) == 'v' ||
        *(char*)(job->DC.eltName[0]) == 'V')
        swp = "vsweep";
    else
        swp = "isweep";

    (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&outd->refName,(IFuid )NULL,
        swp, UID_OTHER, (GENERIC **)NULL);

    outd->circuitPtr = (GENERIC *)ckt;
    outd->analysisPtr = (GENERIC*)ckt->CKTcurJob;
    outd->analName = ckt->CKTcurJob->JOBname;
    outd->refType = IF_REAL;
    outd->dataType = IF_REAL;
    outd->plotPtr = &job->DCTplot;
    outd->numPts = 1;
    outd->initValue = job->DC.vstart[0];
    outd->finalValue = job->DC.vstop[0];
    outd->step = job->DC.vstep[0];
    outd->count = 0;
    (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)outd);
    return (OK);
}


/* ARGSUSED */
static int
dct_operation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    DCTAN *job = (DCTAN*)ckt->CKTcurJob;
    int i, converged = 1;
    double *temp;


    if (ckt->CKTmode & MODEINITPRED) {
        temp = ckt->CKTstates[ckt->CKTmaxOrder+1];
        for (i = ckt->CKTmaxOrder; i >= 0; i--) {
            ckt->CKTstates[i+1] = ckt->CKTstates[i];
        }
        ckt->CKTstate0 = temp;
        converged = NIiter(ckt,ckt->CKTdcTrcvMaxIter);
    }
    if (converged) {

        converged = CKTic(ckt);
        if (converged)
            return (converged);

        converged = CKTop(ckt,
            MODEDCTRANCURVE | MODEINITJCT,
            MODEDCTRANCURVE | MODEINITFLOAT,
            ckt->CKTdcMaxIter);
        if (converged) {
            return (converged);
        }
        bcopy((char *)ckt->CKTstate0,(char *)ckt->CKTstate1,
            ckt->CKTnumStates*sizeof(double));
    }
    ckt->CKTmode = MODEDCTRANCURVE | MODEINITPRED;
    ckt->CKTtime = ((SRCinstance *)(job->DC.elt[0]))->SRCdcValue;

    CKTdump(ckt,ckt->CKTtime,*outd->plotPtr);
    outd->count++;

    return (OK);
}


int
DCTloop(func,ckt,restart,dc,outd)

#ifdef __STDC__
int (*func)(CKTcircuit*,struct sOUTdata*,int);
#else
int (*func)();
#endif
CKTcircuit *ckt;
int restart;
struct sDCTprms *dc;
struct sOUTdata *outd;
{
    double tt;
    int i, j;
    int error;

    if (dc->elt[0] == NULL) {
        /* no dc sources specified */
        return ( (*func)(ckt,outd,restart) );
    }

    if (restart) {
        /* reset static variables */
        dc->dims[0] = dc->dims[1] = 0;
        dc->nestSave = 0;
        dc->skip = 0;
        for (i = 0; i <= dc->nestLevel; i++)
            ((SRCinstance *)dc->elt[i])->SRCdcValue = dc->vstart[i];
    }
    else {
        for (i = 0; i <= dc->nestLevel; i++)
            ((SRCinstance *)dc->elt[i])->SRCdcValue = dc->vstate[i];
    }

    i = dc->nestSave;

    if (dc->skip) {
        /* paused in (*func)() */
        dc->skip = 0;
        goto didpause;
    }

    for (;;) {

        tt = ((SRCinstance*)dc->elt[i])->SRCdcValue - dc->vstop[i];

        if ((dc->vstep[i] > 0 && tt > 1e-8) ||
            (dc->vstep[i] < 0 && tt < 1e-8) ||
            dc->vstep[i] == 0) {

            i++; 
            ckt->CKTmode =
                (ckt->CKTmode & MODEUIC) | MODEDCTRANCURVE | MODEINITJCT;
            if (i > dc->nestLevel)
                break;
            dc->dims[0]++;
            if (dc->dims[2] <= 1) {
                (*(SPfrontEnd->OUTsetDims))(*outd->plotPtr,dc->dims,2);
            }
        }
        else {
            if (!dc->dims[0])
                dc->dims[1]++;
        
            while (i > 0) { 
                i--; 
                ((SRCinstance *)dc->elt[i])->SRCdcValue = dc->vstart[i];
            }

            /* do operation */
            outd->count = 0;
didpause:
            error = (*func)(ckt,outd,restart);
            if (error) {
                dc->nestSave = i;
                dc->skip = 1;
                /* save state, put everything back to normal */
                for (i = 0; i <= dc->nestLevel; i++) {
                    dc->vstate[i] = ((SRCinstance *)dc->elt[i])->SRCdcValue;
                    ((SRCinstance *)dc->elt[i])->SRCdcValue = dc->vsave[i];
                }
                return (error);
            }
            /* so next sub-analysis starts properly */
            restart = true;

            /* store block size in dims[] */
            dc->dims[2] = outd->count;

            if (dc->dims[2] > 1) {

                if (dc->dims[0])
                    (*(SPfrontEnd->OUTsetDims))(*outd->plotPtr,dc->dims,3);
                else
                    (*(SPfrontEnd->OUTsetDims))(*outd->plotPtr,dc->dims+1,2);
            }
            if (*SPfrontEnd->OUTendit) {
                *SPfrontEnd->OUTendit = false;
                for (i = 0; i <= dc->nestLevel; i++) {
                    dc->vstate[i] = ((SRCinstance *)dc->elt[i])->SRCdcValue;
                    ((SRCinstance *)dc->elt[i])->SRCdcValue = dc->vsave[i];
                }
                return (OK);
            }
        }

        ((SRCinstance*)dc->elt[i])->SRCdcValue += dc->vstep[i];

        if ( (*(SPfrontEnd->IFpauseTest))() ) {
            /* user asked us to pause, so save state */
            dc->nestSave = i;
            dc->skip = 0;
            /* save state, put everything back to normal */
            for (i = 0; i <= dc->nestLevel; i++) {
                dc->vstate[i] = ((SRCinstance *)dc->elt[i])->SRCdcValue;
                ((SRCinstance *)dc->elt[i])->SRCdcValue = dc->vsave[i];
            }
            return (E_PAUSE);
        }
    }

    /* done, put everything back to normal */
    for (i = 0; i <= dc->nestLevel; i++)
        ((SRCinstance *)dc->elt[i])->SRCdcValue = dc->vsave[i];
    return (OK);
}
