/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "ftedefs.h"
#include "inpdefs.h"
#include "inpmacs.h"
#include "misc.h"

/* ARGSUSED */
int
DCTparse(ckt,tab,which,currentp,line,task,gnode)

/*
    .dc SRC1NAME Vstart1 [Vstop1 [Vinc1]]
        [SRC2NAME Vstart2 [Vstop2 [Vinc2]]]
*/
GENERIC *ckt;
INPtables *tab;
int which;
GENERIC *currentp;
char **line;
GENERIC *task;
GENERIC *gnode;
{
    GENERIC *foo;   /* pointer to analysis */
    int error;      /* error code temporary */
    card *current = (card *)currentp;

    IFC(newAnalysis,(ckt,which,"DCtransferCurve",&foo, task))

    GCA(ParseDC,(ckt,line,current,which,foo,tab,1))

    if (**line) {
        GCA(ParseDC,(ckt,line,current,which,foo,tab,2))
    }
    return (0);
}


int
ParseDC(ckt,line,current,which,aptr,tab,index)

GENERIC *ckt;
char **line;
card *current;
int which;
GENERIC *aptr;
INPtables *tab;
int index;
{
    char *name;     /* the device's name */
    IFvalue ptemp;  /* a value structure to package stuff into */
    int error;      /* error code temporary */
    double vstart, vstop, vstep;
    char buf[32];

    INPgetTok(line,&name,1);
    strtolower(name);
    INPinsert(&name,tab);
    ptemp.uValue = name;
    sprintf(buf,"name%d",index);
    GCA(INPapName,(ckt,which,aptr,buf,&ptemp))

    vstart = INPevaluate(line,&error,1);          /* vstart */
    if (error == 0) {
        ptemp.rValue = vstart;
        sprintf(buf,"start%d",index);
        GCA(INPapName,(ckt,which,aptr,buf,&ptemp))
    }
    else {
        LITERR("Bad vstart parameter.")
    }

    vstop = INPevaluate(line,&error,1);           /* vstop? */
    if (error == 0) {
        ptemp.rValue = vstop;
        sprintf(buf,"stop%d",index);
        GCA(INPapName,(ckt,which,aptr,buf,&ptemp))
        vstep = INPevaluate(line,&error,1);       /* vstep? */
        if (error == 0)
            ptemp.rValue = vstep;
        else
            ptemp.rValue = vstop - vstart;
        sprintf(buf,"step%d",index);
        GCA(INPapName,(ckt,which,aptr,buf,&ptemp))
    }
    else {
        ptemp.rValue = vstart;
        sprintf(buf,"stop%d",index);
        GCA(INPapName,(ckt,which,aptr,buf,&ptemp))
        ptemp.rValue = 0.0;
        sprintf(buf,"step%d",index);
        GCA(INPapName,(ckt,which,aptr,buf,&ptemp))
    }

    return (0);
}
