/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "dctdefs.h"
#include "jobdefs.h"
#include "iferrmsg.h"
#include "util.h"


/* ARGSUSED */
int 
DCTsetParm(cktp,anal,which,value)

GENERIC *cktp;
GENERIC *anal;
int which;
IFvalue *value;
{
    DCTAN *job = (DCTAN *) anal;

    if (DCTsetp(&job->DC,which,value) == OK)
        return (OK);
    return (E_BADPARM);
}


int 
DCTsetp(dc,which,value)

struct sDCTprms *dc;
int which;
IFvalue *value;
{
    switch (which) {

        case DC_NAME1:
            dc->nestLevel = 0;
            dc->eltName[0] = value->uValue;
            break;

        case DC_START1:
            dc->vstart[0] = value->rValue;
            break;

        case DC_STOP1:
            dc->vstop[0] = value->rValue;
            break;

        case DC_STEP1:
            dc->vstep[0] = value->rValue;
            break;

        case DC_NAME2:
            dc->eltName[1] = value->uValue;
            break;

        case DC_START2:
            dc->vstart[1] = value->rValue;
            break;

        case DC_STOP2:
            dc->vstop[1] = value->rValue;
            break;

        case DC_STEP2:
            dc->vstep[1] = value->rValue;
            dc->nestLevel = 1;
            break;
    
        default:
            return (E_BADPARM);
    }
    return (OK);
}


static IFparm DCTparms[] = {
 { "name1",  DC_NAME1,  IFP|IF_INSTANCE, "name of source to step" },
 { "start1", DC_START1, IFP|IF_REAL,     "starting voltage/current"},
 { "stop1",  DC_STOP1,  IFP|IF_REAL,     "ending voltage/current" },
 { "step1",  DC_STEP1,  IFP|IF_REAL,     "voltage/current step" },
 { "name2",  DC_NAME2,  IFP|IF_INSTANCE, "name of source to step" },
 { "start2", DC_START2, IFP|IF_REAL,     "starting voltage/current"},
 { "stop2",  DC_STOP2,  IFP|IF_REAL,     "ending voltage/current" },
 { "step2",  DC_STEP2,  IFP|IF_REAL,     "voltage/current step" }
};


SPICEanalysis DCTinfo  = {
    { 
        "DCTransfer",
        "D.C. Transfer curve analysis",

        sizeof(DCTparms)/sizeof(IFparm),
        DCTparms
    },
    sizeof(DCTAN),
    SWEEPDOMAIN,
    DCTparse,
    DCTsetParm,
    DCTaskQuest,
    DCTan
};
