/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1988 Jaijeet S. Roychowhury
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "srcdefs.h"
#include "distodef.h"
#include "sperror.h"
#include "outdata.h"
#include "util.h"
#include "cktext.h"
#include "distoext.h"
#include "niext.h"


static void
DISswap(a,b)

double **a, **b;
{
    double *c;

    c = *a;
    *a = *b;
    *b = c;
}


static void
DmemAlloc(a,size)

double **a;
int size;

{
    *a = (double *) MALLOC( sizeof(double) * size );
}


static void
DstorAlloc(header,size)

double ***header;
int size;
{
    *header = (double **) MALLOC( sizeof(double *)*size);
}


#ifdef STDC
extern int CKTdisto(CKTcircuit*, int);
extern int DkerProc(int,double*,double*,int,DISTOan*);
#else
extern int CKTdisto();
extern int DkerProc();
#endif


int
DISTOan(cktp,restart)

GENERIC *cktp;
int restart;
{

    CKTcircuit *ckt = (CKTcircuit *)cktp;
    struct sOUTdata outd;
    double freq;
    static double freqTol;    /* tolerence for finding final frequency */
    static int NoOfPoints;
    static int size;
    static int displacement;
    int error;
    double omegadelta;
    long save;
    int save1;
    int i;
    int numNames;
    IFuid *nameList;
    IFuid freqUid;
    GENERIC *acPlot = NULL;
    DISTOAN* job = (DISTOAN *) (ckt->CKTcurJob);
    static char *nof2src = "No source with f2 distortion input";
#ifdef DISTODEBUG
    double time,time1;
#endif

    if (ckt->CKTjjPresent) {
        (*(SPfrontEnd->IFerror))(ERR_FATAL,
            "Distortion analysis not possible with Josephson junctions",
            NULL);
        return (OK);
    }

    /* start at beginning */

#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))();
#endif
    switch(job->DstepType) {

        case DECADE:
            job->DfreqDelta =  exp(log(10.0)/job->DnumSteps);
            freqTol = job->DfreqDelta * 
                job->DstopF1 * ckt->CKTreltol;
            NoOfPoints = 1 + floor((job->DnumSteps) / log(10.0) *
                log((job->DstopF1+freqTol)/(job->DstartF1)));
            break;

        case OCTAVE:
            job->DfreqDelta = exp(log(2.0)/job->DnumSteps);
            freqTol = job->DfreqDelta * 
                job->DstopF1 * ckt->CKTreltol;
            NoOfPoints = 1 + floor((job->DnumSteps) / log(2.0) *
                log((job->DstopF1+freqTol)/(job->DstartF1)));
            break;

        case LINEAR:
            job->DfreqDelta = (job->DstopF1 - job->DstartF1)/
                    (job->DnumSteps+1);
            freqTol = job->DfreqDelta * ckt->CKTreltol;
            NoOfPoints = job->DnumSteps+ 1 + floor(freqTol/(job->DfreqDelta));
            break;

        default:
            return(E_BADPARM);
    }

    /* freqUid is used as temporary storage */
    freqUid = job->JOBname;
    job->JOBname = "DISTORTION - operating pt.";
    error = DCOan(ckt,0); 
    if (error) return(error);

    job->JOBname = freqUid;

#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))() - time1;
printf("Time for initial work (including op. pt.): %g seconds \n", time1);
#endif

#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))();
#endif

    error = CKTdisto(ckt,D_SETUP);

#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))() - time1;
printf("Time outside D_SETUP: %g seconds \n", time1);
#endif

    if (error) return(error);

    displacement = 0;

#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))();
#endif

    freq = job->DstartF1;
    if (job->Df2wanted) {
        /*
        omegadelta = 2.0 * M_PI * freq *(1. - job->Df2ovrF1);
        */
        /* keeping f2 const to be compatible with spectre */
        job->Domega2 = 2.0 * M_PI * freq * job->Df2ovrF1;
    }
    DstorAlloc(&(job->r1H1stor),NoOfPoints+2);
    DstorAlloc(&(job->r2H11stor),NoOfPoints+2);
    DstorAlloc(&(job->i1H1stor),NoOfPoints+2);
    DstorAlloc(&(job->i2H11stor),NoOfPoints+2);
    size = spGetSize(ckt->CKTmatrix,1);

    if (!job->r1H1ptr) {
        DmemAlloc( &(job->r1H1ptr) , size+2);
        if (!job->r1H1ptr) return (E_NOMEM);
    }
    
    if (!job->r2H11ptr) {
        DmemAlloc( &(job->r2H11ptr) , size+2);
        if (!job->r2H11ptr) return (E_NOMEM);
    }

    if (!job->i1H1ptr) {
        DmemAlloc( &(job->i1H1ptr) , size+2);
        if (!job->i1H1ptr) return (E_NOMEM);
    }
    
    if (!job->i2H11ptr) {
        DmemAlloc( &(job->i2H11ptr) , size+2);
        if (!job->i2H11ptr) return (E_NOMEM);
    }

    if (!job->Df2wanted) {
        DstorAlloc(&(job->r3H11stor),NoOfPoints+2);
        DstorAlloc(&(job->i3H11stor),NoOfPoints+2);
        if (!job->r3H11ptr) {
            DmemAlloc(&(job->r3H11ptr), size+2);
            if (!job->r3H11ptr) return (E_NOMEM);
        }
        if (!job->i3H11ptr) {
            DmemAlloc(&(job->i3H11ptr), size+2);
            if (!job->i3H11ptr) return (E_NOMEM);
        }
    }
    else {
        DstorAlloc(&(job->r1H2stor),NoOfPoints+2);
        DstorAlloc(&(job->i1H2stor),NoOfPoints+2);
        DstorAlloc(&(job->r2H12stor),NoOfPoints+2);
        DstorAlloc(&(job->i2H12stor),NoOfPoints+2);
        DstorAlloc(&(job->r2H1m2stor),NoOfPoints+2);
        DstorAlloc(&(job->i2H1m2stor),NoOfPoints+2);
        DstorAlloc(&(job->r3H1m2stor),NoOfPoints+2);
        DstorAlloc(&(job->i3H1m2stor),NoOfPoints+2);
        if (!job->r1H2ptr) {
            DmemAlloc(&(job->r1H2ptr), size+2);
            if (!job->r1H2ptr) return (E_NOMEM);
        }
        if (!job->r2H12ptr) {
            DmemAlloc(&(job->r2H12ptr) , size+2);
            if (! job->r2H12ptr) return (E_NOMEM);
        }
        if (!job->r2H1m2ptr) {
            DmemAlloc(&(job->r2H1m2ptr), size+2);
            if (!job->r2H1m2ptr) return (E_NOMEM);
        }
        if (!job->r3H1m2ptr) {
            DmemAlloc(&(job->r3H1m2ptr), size+2);
            if (!job->r3H1m2ptr) return (E_NOMEM);
        }
        if (!job->i1H2ptr) {
            DmemAlloc(&(job->i1H2ptr), size+2);
            if (!job->i1H2ptr) return (E_NOMEM);
        }
        if (! job->i2H12ptr) {
            DmemAlloc(&(job->i2H12ptr), size+2);
            if (!job->i2H12ptr) return (E_NOMEM);
        }
        if (!job->i2H1m2ptr) {
            DmemAlloc(&(job->i2H1m2ptr), size+2);
            if (!job->i2H1m2ptr) return (E_NOMEM);
        }
        if (!job->i3H1m2ptr) {
            DmemAlloc(&(job->i3H1m2ptr), size+2);
            if (!job->i3H1m2ptr) return (E_NOMEM);
        }
    }
        
#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))() - time1;
printf("Time for other setup (storage allocation etc.): %g seconds \n", time1);
#endif

#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))();
#endif

    while(freq <= job->DstopF1+freqTol) {

        /*
        if( (*(SPfrontEnd->IFpauseTest))() ) { 
            job->DsaveF1 = freq;
            return(E_PAUSE);
        }
        */

        ckt->CKTomega = 2.0 * M_PI * freq;
        job->Domega1 = ckt->CKTomega;
        ckt->CKTmode = MODEAC;

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))();
#endif

        error = CKTacLoad(ckt);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))() - time;
printf("Time for CKTacLoad: %g seconds \n", time);
#endif

        if (error) return(error);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))();
#endif

        error = CKTdisto(ckt,D_RHSF1);
        /* sets up the RHS vector
            for all inputs corresponding to F1 */

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))() - time;
printf("Time outside DISTO_RHSFIX: %g seconds \n", time);
#endif

        if (error) return(error);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))();
#endif

        error = NIdIter(ckt);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))() - time;
printf("Time for NIdIter: %g seconds \n", time);
#endif

        if (error) return(error);
        DISswap(&(ckt->CKTrhsOld),&(job->r1H1ptr));
        DISswap(&(ckt->CKTirhsOld),&(job->i1H1ptr));

        ckt->CKTomega *= 2;
        error = CKTacLoad(ckt);
        if (error) return(error);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))();
#endif

        error = CKTdisto(ckt,D_TWOF1);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))() - time;
printf("Time outside D_TWOF1: %g seconds \n", time);
#endif

        if (error) return(error);
        error = NIdIter(ckt);
        if (error) return(error);
        DISswap(&(ckt->CKTrhsOld),&(job->r2H11ptr));
        DISswap(&(ckt->CKTirhsOld),&(job->i2H11ptr));

        if (!job->Df2wanted) {

            ckt->CKTomega = 3 * job->Domega1;
            error = CKTacLoad(ckt);
            if (error) return(error);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))();
#endif

            error = CKTdisto(ckt,D_THRF1);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))() - time;
printf("Time outside D_THRF1: %g seconds \n", time);
#endif

            if (error) return(error);
            error = NIdIter(ckt);
            if (error) return(error);
            DISswap(&(ckt->CKTrhsOld),&(job->r3H11ptr));
            DISswap(&(ckt->CKTirhsOld),&(job->i3H11ptr));
        }
        else if (job->Df2given) {

            /*
            ckt->CKTomega = job->Domega1 - omegadelta;
            job->Domega2 = ckt->CKTomega;
            */
            ckt->CKTomega = job->Domega2;
            error = CKTacLoad(ckt);
            if (error) return(error);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))();
#endif

            error = CKTdisto(ckt,D_RHSF2);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))() - time;
printf("Time outside DISTO_RHSFIX: %g seconds \n", time);
#endif

            if (error) return(error);
            error = NIdIter(ckt);
            if (error) return(error);
            DISswap(&(ckt->CKTrhsOld),&(job->r1H2ptr));
            DISswap(&(ckt->CKTirhsOld),&(job->i1H2ptr));

            ckt->CKTomega = job->Domega1 + job->Domega2;
            error = CKTacLoad(ckt);
            if (error) return(error);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))();
#endif

            error = CKTdisto(ckt,D_F1PF2);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))() - time;
printf("Time outside D_F1PF2: %g seconds \n", time);
#endif

            if (error) return(error);
            error = NIdIter(ckt);
            if (error) return(error);
            DISswap(&(ckt->CKTrhsOld),&(job->r2H12ptr));
            DISswap(&(ckt->CKTirhsOld),&(job->i2H12ptr));

            ckt->CKTomega = job->Domega1 - job->Domega2;
            error = CKTacLoad(ckt);
            if (error) return(error);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))();
#endif

            error = CKTdisto(ckt,D_F1MF2);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))() - time;
printf("Time outside D_F1MF2: %g seconds \n", time);
#endif

            if (error) return(error);
            error = NIdIter(ckt);
            if (error) return(error);
            DISswap(&(ckt->CKTrhsOld),&(job->r2H1m2ptr));
            DISswap(&(ckt->CKTirhsOld),&(job->i2H1m2ptr));

            ckt->CKTomega = 2*job->Domega1 - job->Domega2;
            error = CKTacLoad(ckt);
            if (error) return(error);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))();
#endif

            error = CKTdisto(ckt,D_2F1MF2);

#ifdef D_DBG_SMALLTIMES
time = (*(SPfrontEnd->IFseconds))() - time;
printf("Time outside D_2F1MF2: %g seconds \n", time);
#endif

            if (error) return(error);
            error = NIdIter(ckt);
            if (error) return(error);
            DISswap(&(ckt->CKTrhsOld),&(job->r3H1m2ptr));
            DISswap(&(ckt->CKTirhsOld),&(job->i3H1m2ptr));
        }
        else {
            errMsg = MALLOC(strlen(nof2src)+1);
            strcpy(errMsg,nof2src);
            return(E_NOF2SRC);
        }

        DmemAlloc( &(job->r1H1stor[displacement]),size);
        DISswap(&(job->r1H1stor[displacement]),&(job->r1H1ptr));
        job->r1H1stor[displacement][0]=freq;
        DmemAlloc( &(job->r2H11stor[displacement]),size);
        DISswap(&(job->r2H11stor[displacement]),&((job->r2H11ptr)));
        job->r2H11stor[displacement][0]=freq;
        DmemAlloc( &(job->i1H1stor[displacement]),size);
        DISswap(&(job->i1H1stor[displacement]),&((job->i1H1ptr)));
        job->i1H1stor[displacement][0]=0.0;
        DmemAlloc( &(job->i2H11stor[displacement]),size);
        DISswap(&(job->i2H11stor[displacement]),&((job->i2H11ptr)));
        job->i2H11stor[displacement][0]=0.0;

        if (! (job->Df2wanted)) {
            DmemAlloc( &(job->r3H11stor[displacement]),size); 
            DISswap(&(job->r3H11stor[displacement]),&((job->r3H11ptr)));
            job->r3H11stor[displacement][0]=freq;
            DmemAlloc( &(job->i3H11stor[displacement]),size); 
            DISswap(&(job->i3H11stor[displacement]),&((job->i3H11ptr)));
            job->i3H11stor[displacement][0]=0.0;
        }
        else {
            DmemAlloc( &(job->r1H2stor[displacement]),size);
            DISswap(&(job->r1H2stor[displacement]),&((job->r1H2ptr)));
            job->r1H2stor[displacement][0]=freq;
            DmemAlloc( &(job->r2H12stor[displacement]),size);
            DISswap(&(job->r2H12stor[displacement]),&((job->r2H12ptr)));
            job->r2H12stor[displacement][0]=freq;
            DmemAlloc( &(job->r2H1m2stor[displacement]),size);
            DISswap(&(job->r2H1m2stor[displacement]),&((job->r2H1m2ptr)));
            job->r2H1m2stor[displacement][0]=freq;
            DmemAlloc( &(job->r3H1m2stor[displacement]),size);
            DISswap(&(job->r3H1m2stor[displacement]),&((job->r3H1m2ptr)));
            job->r3H1m2stor[displacement][0]=freq;

            DmemAlloc( &(job->i1H2stor[displacement]),size);
            DISswap(&(job->i1H2stor[displacement]),&((job->i1H2ptr)));
            job->i1H2stor[displacement][0]=0.0;
            DmemAlloc( &(job->i2H12stor[displacement]),size);
            DISswap(&(job->i2H12stor[displacement]),&((job->i2H12ptr)));
            job->i2H12stor[displacement][0]=0.0;
            DmemAlloc( &(job->i2H1m2stor[displacement]),size);
            DISswap(&(job->i2H1m2stor[displacement]),&((job->i2H1m2ptr)));
            job->i2H1m2stor[displacement][0]=0.0;
            DmemAlloc( &(job->i3H1m2stor[displacement]),size);
            DISswap(&(job->i3H1m2stor[displacement]),&((job->i3H1m2ptr)));
            job->i3H1m2stor[displacement][0]=0.0;
        }
        displacement++;

        switch(job->DstepType) {
            case DECADE:
            case OCTAVE:
                freq *= job->DfreqDelta;
                if (job->DfreqDelta == 1) goto endsweep;
                break;

            case LINEAR:
                freq += job->DfreqDelta;
                if (job->DfreqDelta == 0) goto endsweep;
                break;

            default:
                return(E_INTERN);
        }
    }

#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))() - time1;
printf("Time inside frequency loop: %g seconds \n", time1);
#endif
    
endsweep:


    /* output routines to process the H's and output actual ckt variable
       values */

#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))();
#endif

    if (!job->Df2wanted) {
        error = CKTnames(ckt,&numNames,&nameList);
        if (error) return (error);
        (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&freqUid,(IFuid)NULL,
            "frequency", UID_OTHER,(GENERIC **)NULL);

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = "DISTORTION - 2nd harmonic";
        outd.refName = freqUid;
        outd.refType = IF_REAL;
        outd.numNames = numNames;
        outd.dataNames = nameList;
        outd.dataType = IF_COMPLEX;
        outd.plotPtr = &acPlot;
        outd.numPts = 1;
        outd.initValue = 0;
        outd.finalValue = 0;
        outd.step = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

        for (i=0; i< displacement ; i++) {
            DkerProc(D_TWOF1,*(job->r2H11stor + i),
                 *(job->i2H11stor + i),
                 size, job);
            ckt->CKTrhsOld = *((job->r2H11stor) + i);
            ckt->CKTirhsOld = *((job->i2H11stor) + i);
            error = CKTacDump(ckt,ckt->CKTrhsOld[0],acPlot);
            if (error) return(error);
        }
        (*(SPfrontEnd->OUTendPlot))(acPlot);

        error = CKTnames(ckt,&numNames,&nameList);
        if (error) return(error);
        (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&freqUid,(IFuid)NULL,
            "frequency", UID_OTHER,(GENERIC **)NULL);

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = "DISTORTION - 3rd harmonic";
        outd.refName = freqUid;
        outd.refType = IF_REAL;
        outd.numNames = numNames;
        outd.dataNames = nameList;
        outd.dataType = IF_COMPLEX;
        outd.plotPtr = &acPlot;
        outd.numPts = 1;
        outd.initValue = 0;
        outd.finalValue = 0;
        outd.step = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

        for (i=0; i< displacement ; i++) {
            DkerProc(D_THRF1,*(job->r3H11stor + i),
                 *(job->i3H11stor + i),
                 size, job);
            ckt->CKTrhsOld = *((job->r3H11stor) + i);
            ckt->CKTirhsOld = *((job->i3H11stor) + i);
            error = CKTacDump(ckt,ckt->CKTrhsOld[0],acPlot);
        }
        (*(SPfrontEnd->OUTendPlot))(acPlot);
      
    }
    else {

        error = CKTnames(ckt,&numNames,&nameList);
        if (error) return(error);
        (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&freqUid,(IFuid)NULL,
            "frequency", UID_OTHER,(GENERIC **)NULL);

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = "DISTORTION - IM: f1+f2";
        outd.refName = freqUid;
        outd.refType = IF_REAL;
        outd.numNames = numNames;
        outd.dataNames = nameList;
        outd.dataType = IF_COMPLEX;
        outd.plotPtr = &acPlot;
        outd.numPts = 1;
        outd.initValue = 0;
        outd.finalValue = 0;
        outd.step = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

        for (i=0; i< displacement ; i++) {
            DkerProc(D_F1PF2,*(job->r2H12stor + i),
                 *(job->i2H12stor + i),
                 size, job);
            ckt->CKTrhsOld = *((job->r2H12stor) + i);
            ckt->CKTirhsOld = *((job->i2H12stor) + i);
            error = CKTacDump(ckt,ckt->CKTrhsOld[0],acPlot);
            if (error) return (error);
        }
        (*(SPfrontEnd->OUTendPlot))(acPlot);

        error = CKTnames(ckt,&numNames,&nameList);
        if (error) return (error);
        (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&freqUid,(IFuid)NULL,
            "frequency", UID_OTHER,(GENERIC **)NULL);

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = "DISTORTION - IM: f1-f2";
        outd.refName = freqUid;
        outd.refType = IF_REAL;
        outd.numNames = numNames;
        outd.dataNames = nameList;
        outd.dataType = IF_COMPLEX;
        outd.plotPtr = &acPlot;
        outd.numPts = 1;
        outd.initValue = 0;
        outd.finalValue = 0;
        outd.step = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

        for (i=0; i< displacement ; i++) {
            DkerProc(D_F1MF2,
                *(job->r2H1m2stor + i),
                *(job->i2H1m2stor + i),
                 size, job);
            ckt->CKTrhsOld = *((job->r2H1m2stor) + i);
            ckt->CKTirhsOld = *((job->i2H1m2stor) + i);
            error = CKTacDump(ckt,ckt->CKTrhsOld[0],acPlot);
            if (error) return(error);
        }
        (*(SPfrontEnd->OUTendPlot))(acPlot);

        error = CKTnames(ckt,&numNames,&nameList);
        if (error) return (error);
        (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&freqUid,(IFuid)NULL,
            "frequency", UID_OTHER,(GENERIC **)NULL);

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = "DISTORTION - IM: 2f1-f2";
        outd.refName = freqUid;
        outd.refType = IF_REAL;
        outd.numNames = numNames;
        outd.dataNames = nameList;
        outd.dataType = IF_COMPLEX;
        outd.plotPtr = &acPlot;
        outd.numPts = 1;
        outd.initValue = 0;
        outd.finalValue = 0;
        outd.step = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

        for (i=0; i< displacement ; i++) {
            DkerProc(D_2F1MF2,
                *(job->r3H1m2stor + i),
                *(job->i3H1m2stor + i),
                 size, job);
            ckt->CKTrhsOld = *((job->r3H1m2stor) + i);
            ckt->CKTirhsOld = *((job->i3H1m2stor) + i);
            error = CKTacDump(ckt,ckt->CKTrhsOld[0],acPlot);
            if (error) return (error);
        }
        (*(SPfrontEnd->OUTendPlot))(acPlot);

    }
    FREE(job->r1H1ptr);
    FREE(job->i1H1ptr);
    FREE(job->r2H11ptr);
    FREE(job->i2H11ptr);

    FREE(job->r1H1stor);
    FREE(job->i1H1stor);
    FREE(job->r2H11stor);
    FREE(job->i2H11stor);

    if (! (job->Df2wanted)) {
        FREE(job->r3H11ptr);
        FREE(job->i3H11ptr);

        FREE(job->i3H11stor);
        FREE(job->r3H11stor);
    }
    else {

        FREE(job->r2H1m2ptr);
        FREE(job->r3H1m2ptr);
        FREE(job->r1H2ptr);
        FREE(job->i1H2ptr);
        FREE(job->r2H12ptr);
        FREE(job->i2H12ptr);
        FREE(job->i2H1m2ptr);
        FREE(job->i3H1m2ptr);

        FREE(job->r1H2stor);
        FREE(job->r2H12stor);
        FREE(job->r2H1m2stor);
        FREE(job->r3H1m2stor);
        FREE(job->i1H2stor);
        FREE(job->i2H12stor);
        FREE(job->i2H1m2stor);
        FREE(job->i3H1m2stor);
    }

#ifdef D_DBG_BLOCKTIMES
time1 = (*(SPfrontEnd->IFseconds))() - time1;
printf("Time for output and deallocation: %g seconds \n", time1);
#endif

    return(OK);
}



/*
 * CKTdisto (ckt, mode)
 */


#ifdef __STDC__
static void dset(CKTcircuit*,int,DISTOAN*);
#else
static void dset();
#endif

int
CKTdisto (ckt, mode)

CKTcircuit *ckt;
int mode;
{
    extern SPICEdev *DEVices[];
    DISTOAN* cv = (DISTOAN*) (ckt->CKTcurJob);
    int i;
    int error;
    int size;
    struct sCKTmodHead *mh;
    int (*func)();

    switch (mode) {

        case D_TWOF1:
        case D_THRF1:
        case D_F1PF2:
        case D_F1MF2:
        case D_2F1MF2:

            size = spGetSize(ckt->CKTmatrix,1);
            for (i = 1; i <= size; i++) {
                *(ckt->CKTrhs + i) = 0;
                *(ckt->CKTirhs + i) = 0;
            }

        case D_SETUP:
            for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
                if ((func = DEVices[mh->type]->DEVdisto) != NULL) {
                    error = (*func)(mode,mh->head,ckt);
                    if (error)
                        return (error);
                }
            }
            break;

        case D_RHSF1:

             cv->Df2given = 0; /* will change if any F2 source is found */

        case D_RHSF2:

            dset(ckt,mode,cv);
            error = 0;
            break;

        default: 

            error = E_BADPARM;
            break;
    }

    return (error);
}


static void
dset(ckt,mode,cv)

CKTcircuit *ckt;
int mode;
DISTOAN* cv;
{
    SRCinstance *here;
    SRCmodel *model;
    double mag, phase;
    int i, code, size;

    size = spGetSize(ckt->CKTmatrix,1);
    for (i = 0; i <= size; i++) {
        *(ckt->CKTrhs+i) = 0;
        *(ckt->CKTirhs+i) = 0;
    }

    code = CKTtypelook("Source");

    if (code < 0)
        return;

    for (model = (SRCmodel *)ckt->CKThead[code]; model != NULL;
            model = model->SRCnextModel) {
        for (here = model->SRCinstances; here!=NULL;
                here = here->SRCnextInstance) {

            /* check if the source has a distortion input*/

            if (here->SRCdGiven) {
                if (here->SRCdF2given)
                    cv->Df2given = 1;
                if ((here->SRCdF1given) && (mode == D_RHSF1)) {
                    mag = here->SRCdF1mag;
                    phase = here->SRCdF1phase;
                }
                else if ((here->SRCdF2given) && (mode == D_RHSF2)) {
                    mag = here->SRCdF2mag;
                    phase = here->SRCdF2phase;
                }
                if (((here->SRCdF1given) && (mode == D_RHSF1)) || 
                        ((here->SRCdF2given) && (mode == D_RHSF2))) {
                    if (here->SRCtype == SRC_V) {
                        *(ckt->CKTrhs + here->SRCbranch) =
                            0.5*mag* cos(M_PI*phase/180.0);
                        *(ckt->CKTirhs + here->SRCbranch) =
                            0.5*mag*sin(M_PI*phase/180.0); 
                    }
                    else {
                        *(ckt->CKTrhs + here->SRCposNode) =
                            - 0.5*mag*cos(M_PI*phase/180.0);
                        *(ckt->CKTrhs + here->SRCnegNode) =
                            0.5*mag*cos(M_PI*phase/180.0);
                        *(ckt->CKTirhs + here->SRCposNode) =
                            - 0.5*mag*sin(M_PI*phase/180.0);
                        *(ckt->CKTirhs + here->SRCnegNode) =
                            0.5*mag*sin(M_PI*phase/180.0);
                    }
                }
            }
        }
    }
}
