/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1988 Jaijeet S Roychowdhury
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "jobdefs.h"
#include "distodef.h"
#include "iferrmsg.h"
#include "distoext.h"


/* ARGSUSED */
int 
DISTOsetParm(cktp,anal,which,value)

GENERIC *cktp;
GENERIC *anal;
int which;
IFvalue *value;
{
    switch (which) {

        case D_START:
            ((DISTOAN*)anal)->DstartF1 = value->rValue;
            break;

        case D_STOP:
            ((DISTOAN*)anal)->DstopF1 = value->rValue;
            break;

        case D_STEPS:
            ((DISTOAN*)anal)->DnumSteps = value->iValue;
            break;

        case D_DEC:
            ((DISTOAN*)anal)->DstepType = DECADE;
            break;

        case D_OCT:
            ((DISTOAN*)anal)->DstepType = OCTAVE;
            break;

        case D_LIN:
            ((DISTOAN*)anal)->DstepType = LINEAR;
            break;

        case D_F2OVRF1:
            ((DISTOAN*)anal)->Df2ovrF1 = value->rValue;
            ((DISTOAN*)anal)->Df2wanted = 1;
            break;

        default:
            return (E_BADPARM);
    }
    return (OK);
}


static IFparm Dparms[] = {
    { "start",      D_START,   IF_SET|IF_REAL, "starting frequency" },
    { "stop",       D_STOP,    IF_SET|IF_REAL, "ending frequency" },
    { "numsteps",   D_STEPS,   IF_SET|IF_INTEGER,  "number of frequencies" },
    { "dec",        D_DEC,     IF_SET|IF_FLAG, "step by decades" },
    { "oct",        D_OCT,     IF_SET|IF_FLAG, "step by octaves" },
    { "lin",        D_LIN,     IF_SET|IF_FLAG, "step linearly" },
    { "f2overf1",   D_F2OVRF1, IF_SET|IF_REAL, "ratio of F2 to F1" },
};

SPICEanalysis DISTOinfo  = {
    { 
        "DISTO",
        "Small signal distortion analysis",

        sizeof(Dparms)/sizeof(IFparm),
        Dparms
    },
    sizeof(DISTOAN),
    FREQUENCYDOMAIN,
    DISTOparse,
    DISTOsetParm,
    DISTOaskQuest,
    DISTOan
};
