/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Gary W. Ng
         1992 Stephen R. Whiteley
****************************************************************************/

#define ANAL_EXT

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "srcdefs.h"
#include "fteconst.h"
#include "iferrmsg.h"
#include "sperror.h"
#include "util.h"
#include "outdata.h"
#include "noisedef.h"
#include "cktext.h"
#include "niext.h"

#ifdef __STDC__
static int noi_operation(CKTcircuit*,struct sOUTdata*,int);
static int noi_dcoperation(CKTcircuit*,struct sOUTdata*,int);
#else
static int noi_operation();
static int noi_dcoperation();
#endif


int
NOISEan(cktp, restart)

GENERIC *cktp;
int restart;
{
    CKTcircuit *ckt = (CKTcircuit *)cktp;
    Ndata *data;
    struct sOUTdata outd;
    int error, error1;
    int i, code;
    CKTnode *node;
    NOISEAN *job = (NOISEAN*) (ckt->CKTcurJob);
    static char *nodundef = "noise output node %s is not defined";
    static char *noacinput = "no AC input source %s for noise analysis";
    SRCinstance *here;

    if (ckt->CKTjjPresent) {
        (*(SPfrontEnd->IFerror))(ERR_FATAL,
            "Noise analysis not possible with Josephson junctions", NULL);
            return (OK);
    }
    /* make sure all the nodes/sources referenced are actually
     * in the circuit
     */
    if (restart) {
        job->NposOutNode = -1;
        if (job->Noutput) {
            for (node = ckt->CKTnodes; node; node = node->next) {
                if (strcmp(node->name, job->Noutput) == 0) {
                    job->NposOutNode = node->number;
                    break;
                }
            }
        }
        if (job->NposOutNode == -1) {
            (*(SPfrontEnd->IFerror))(ERR_FATAL,nodundef,
                (IFuid)&(job->Noutput));
            return (E_NODUNDEF);
        }
        if (job->NoutputRef) {     /* was an output reference specified? */
            job->NnegOutNode = -1;
            for (node = ckt->CKTnodes; node; node = node->next) {
                if (strcmp(node->name, job->NoutputRef) == 0) {
                    job->NnegOutNode = node->number;
                    break;
                }
            }
            if (job->NnegOutNode == -1) {
                (*(SPfrontEnd->IFerror))(ERR_FATAL,nodundef,
                    (IFuid)&(job->NoutputRef));
                return (E_NODUNDEF);
            }
        }
        else {
            job->NnegOutNode = 0;
            /* if no output reference node was specified, we assume ground */
        }

        /* see if the source specified is AC */

        code = CKTtypelook("Source");
        here = (SRCinstance *)NULL;
        error = CKTfndDev((GENERIC*)ckt,&code,(GENERIC**)&here,
            job->Ninput, (GENERIC *)NULL,(IFuid)NULL);

        if (error) {
            (*(SPfrontEnd->IFerror))(ERR_FATAL,noacinput,
                (IFuid)&(job->Ninput));
            return (E_NOACINPUT);
        }

        if (job->DC.eltName[0] != NULL) {
            /* DC source was given */

            for (i = 0; i <= job->DC.nestLevel; i++) {

                here = (SRCinstance *)NULL;
                error = CKTfndDev((GENERIC*)ckt,&code,(GENERIC**)&here,
                    job->DC.eltName[i], (GENERIC *)NULL,(IFuid)NULL);

                if (error) {
                    (*(SPfrontEnd->IFerror))(ERR_FATAL, 
                        "DCtrCurv: source %s not in circuit",
                        &(job->DC.eltName[i]));
                    return (E_NOTFOUND);
                }
                job->DC.elt[i] = (GENinstance*)here;
                job->DC.vsave[i] = here->SRCdcValue;
            }
        }

        data = (Ndata*)MALLOC(sizeof(Ndata));
        job->NdataPtr = data;

        /* the current front-end needs the namelist to be fully
         * declared before an OUTbeginplot
         */

        (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&outd.refName,(IFuid)NULL,
            "frequency", UID_OTHER,(GENERIC **)NULL);

        data->numPlots = 0;    /* we don't have any plots yet */
        error = CKTnoise(ckt,N_DENS,N_OPEN,data);
        if (error)
            return(error);

        /*
         * all names in the namelist have been declared. now start the
         * plot
         */

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = "Noise Spectral Density Curves - (V^2 or A^2)/Hz";
        outd.refType = IF_REAL;
        outd.numNames = data->numPlots;
        outd.dataNames = data->namelist;
        outd.dataType = IF_REAL;
        outd.plotPtr = &(data->NplotPtr);
        outd.numPts = 1;
        outd.initValue = 0;
        outd.finalValue = 0;
        outd.step = 0;
        outd.count = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

    }

    /* do the noise analysis over all frequencies */
    error = DCTloop(noi_dcoperation,ckt,restart,&job->DC,&outd);
    if (error == E_PAUSE)
        return (error);
    error1 = CKTnoise(ckt,N_DENS,N_CLOSE,data);
    if (error)
        return (error);
    if (error1)
        return (error1);

    data->numPlots = 0;
    data->outNumber = 0;

    if (job->AC.fstart != job->AC.fstop) {
        error = CKTnoise(ckt,INT_NOIZ,N_OPEN,data);
        if (error)
            return(error);

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = "Integrated Noise - V^2 or A^2";
        outd.refName = (IFuid)NULL;
        outd.refType = (int)0;
        outd.numNames = data->numPlots;
        outd.dataNames = data->namelist;
        outd.dataType = IF_REAL;
        outd.plotPtr = &(data->NplotPtr);
        outd.numPts = 1;
        outd.initValue = 0;
        outd.finalValue = 0;
        outd.step = 0;
        outd.count = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

        error = CKTnoise(ckt,INT_NOIZ,N_CALC,data);
        if (error)
            return(error);

        error = CKTnoise(ckt,INT_NOIZ,N_CLOSE,data);
        if (error)
            return(error);
    }

    FREE (data);
    return (OK);
}

static int
noi_dcoperation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    NOISEAN *job = (NOISEAN*) ckt->CKTcurJob;
    Ndata *data = job->NdataPtr;
    int error;

    if (restart) {
        data->freq = job->AC.fstart;
        data->lstFreq = data->freq;
        data->outNoiz = 0.0;
        data->inNoise = 0.0;
        job->Nstep = 0;
    }
    else {
        /* we must have paused before.  pick up where we left off */
        data->outNoiz = job->NsavOnoise;
        data->inNoise = job->NsavInoise;
    }

    error = CKTic(ckt);
    if (error)
        return (error);

    error = CKTop(ckt,
        MODEDCOP | MODEINITJCT,
        MODEDCOP | MODEINITFLOAT,
        ckt->CKTdcMaxIter);

    if (error)
        return (error);

    error = ACloop(noi_operation,ckt,restart,&job->AC,outd);
    if (error == E_PAUSE) {
        job->NsavOnoise = data->outNoiz; /* up until now */
        job->NsavInoise = data->inNoise;
        return (error);
    }
    return (OK);
}


static int
noi_operation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    NOISEAN *job = (NOISEAN*) ckt->CKTcurJob;
    Ndata *data = job->NdataPtr;
    double realVal, imagVal;
    int error;

    data->freq = ckt->CKTomega/(2*M_PI);
    realVal = *((ckt->CKTrhsOld) + job->NposOutNode) -
              *((ckt->CKTrhsOld) + job->NnegOutNode);
    imagVal = *((ckt->CKTirhsOld) + job->NposOutNode) -
              *((ckt->CKTirhsOld) + job->NnegOutNode);
    data->GainSqInv = 1.0 / MAX(((realVal*realVal)
        + (imagVal*imagVal)),N_MINGAIN);
    data->lnGainInv = log(data->GainSqInv);

    /* set up a block of "common" data so we don't have to
     * recalculate it for every device
     */

    data->delFreq = data->freq - data->lstFreq;
    data->lnFreq = log(MAX(data->freq,N_MINLOG));
    data->lnLastFreq = log(MAX(data->lstFreq,N_MINLOG));
    data->delLnFreq = data->lnFreq - data->lnLastFreq;

    if ((job->NStpsSm != 0) && ((job->Nstep % (job->NStpsSm)) == 0)) {
        data->prtSummary = TRUE;
    }
    else {
        data->prtSummary = FALSE;
    }

    data->outNumber = 0;
    /* the frequency will NOT be stored in array[0]  as before; instead,
     * it will be given in refVal.rValue (see later)
     */

    /* solve the adjoint system */
    NInzIter(ckt,job->NposOutNode,job->NnegOutNode);

    /* now we use the adjoint system to calculate the noise
     * contributions of each generator in the circuit
     */

    error = CKTnoise(ckt,N_DENS,N_CALC,data);
    if (error)
        return (error);
    outd->count++;

    data->lstFreq = data->freq;
    job->Nstep++;

    return (OK);
}


/*
 * CKTnoise (ckt, mode, operation, data)
 *
 *   This routine is responsible for naming and evaluating all of the
 *   noise sources in the circuit.  It uses a series of subroutines to
 *   name and evaluate the sources associated with each model, and then
 *   it evaluates the noise for the entire circuit.
 */


int
CKTnoise (ckt, mode, operation, data)

CKTcircuit *ckt;
int mode;
int operation;
Ndata *data;
{
    double outNdens;
    extern SPICEdev *DEVices[];
    IFvalue outData; /* output variable (points to list of outputs) */
    IFvalue refVal;  /* reference variable (always 0) */
    int error;
    struct sCKTmodHead *mh;
    int (*func)();

    outNdens = 0.0;

    /* let each device decide how many and what type of noise sources
     * it has
     */
    for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
        if ((func = DEVices[mh->type]->DEVnoise) != NULL) {
            error = (*func)(mode,operation,mh->head,ckt,data,&outNdens);
            if (error)
                return (error);
        }
    }

    if (operation == N_OPEN) {

        /* take care of the noise for the circuit as a whole */

        if (mode == N_DENS) {

            data->namelist = (IFuid *)trealloc((char *)data->namelist,
                (data->numPlots + 1)*sizeof(IFuid));

            (*(SPfrontEnd->IFnewUid))(ckt,&(data->namelist[data->numPlots++]),
                (IFuid)NULL,"onoise_spectrum",UID_OTHER,(GENERIC **)NULL);

            data->namelist = (IFuid *)trealloc((char *)data->namelist,
                (data->numPlots + 1)*sizeof(IFuid));

            (*(SPfrontEnd->IFnewUid))(ckt,&(data->namelist[data->numPlots++]),
                (IFuid)NULL,"inoise_spectrum",UID_OTHER,(GENERIC **)NULL);

            /* we've added two more plots */

            data->outpVector =
                (double *)MALLOC(data->numPlots * sizeof(double));

            return (OK);
        }

        if (mode == INT_NOIZ) {

            data->namelist = (IFuid *)trealloc((char *)data->namelist,
                (data->numPlots + 1)*sizeof(IFuid));

            (*(SPfrontEnd->IFnewUid))(ckt,&(data->namelist[data->numPlots++]),
                (IFuid)NULL,"onoise_total",UID_OTHER,(GENERIC **)NULL);

            data->namelist = (IFuid *)trealloc((char *)data->namelist,
                (data->numPlots + 1)*sizeof(IFuid));

            (*(SPfrontEnd->IFnewUid))(ckt,&(data->namelist[data->numPlots++]),
                (IFuid)NULL,"inoise_total",UID_OTHER,(GENERIC **)NULL);

            /* we've added two more plots */

            data->outpVector =
                (double *) MALLOC(data->numPlots * sizeof(double));

            return (OK);
        }

        return (E_INTERN);
    }

    if (operation == N_CALC) {

        if (mode == N_DENS) {

            if ((((NOISEAN*)ckt->CKTcurJob)->NStpsSm == 0)
                    || data->prtSummary) {
                data->outpVector[data->outNumber++] = outNdens;
                data->outpVector[data->outNumber++] =
                    (outNdens * data->GainSqInv);

                refVal.rValue = data->freq; /* the reference is the freq */
                outData.v.numValue = data->outNumber; /* vector number */
                outData.v.vec.rVec = data->outpVector; /* vector of outputs */
                (*(SPfrontEnd->OUTdata))(data->NplotPtr,&refVal,&outData);
            }
            return (OK);
        }

        if (mode == INT_NOIZ) {

            data->outpVector[data->outNumber++] =  data->outNoiz; 
            data->outpVector[data->outNumber++] =  data->inNoise;
            outData.v.vec.rVec = data->outpVector; /* vector of outputs */
            outData.v.numValue = data->outNumber;  /* vector number */
            (*(SPfrontEnd->OUTdata))(data->NplotPtr,&refVal,&outData);
            return (OK);
        }

        return (E_INTERN);
    }

    if (operation == N_CLOSE) {

        (*(SPfrontEnd->OUTendPlot))(data->NplotPtr);
        FREE(data->namelist);
        FREE(data->outpVector);
        return (OK);
    }

    return (E_INTERN);
}
