/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Gary W. Ng
         1993 Stephen R. Whiteley
****************************************************************************/

/*
 * NevalSrc (noise, lnNoise, ckt, type, node1, node2, param)
 *   This routine evaluates the noise due to different physical
 *   phenomena.  This includes the "shot" noise associated with dc
 *   currents in semiconductors and the "thermal" noise associated with
 *   resistance.  Although semiconductors also display "flicker" (1/f)
 *   noise, the lack of a unified model requires us to handle it on a
 *   "case by case" basis.  What we CAN provide, though, is the noise
 *   gain associated with the 1/f source.
 */


#include "spice.h"
#include <math.h>
#include "cktdefs.h"
#include "noisedef.h"
#include "const.h"
#include "util.h"


void
NevalSrc(noise, lnNoise, ckt, type, node1, node2, param)

double *noise;
double *lnNoise;
CKTcircuit *ckt;
int type;
int node1;
int node2;
double param;
{
    double realVal;
    double imagVal;
    double gain;

    realVal = *(ckt->CKTrhs + node1) - *(ckt->CKTrhs + node2);
    imagVal = *(ckt->CKTirhs + node1) - *(ckt->CKTirhs + node2);
    gain = realVal*realVal + imagVal*imagVal;

    switch (type) {

        case SHOTNOISE:
            /* param is the dc current in a semiconductor */
            *noise = gain * 2 * CHARGE * FABS(param);
            *lnNoise = log( MAX(*noise,N_MINLOG) ); 
            break;

        case THERMNOISE:
            /* param is the conductance of a resistor */
            *noise = gain * 4 * CONSTboltz * ckt->CKTtemp * param;
            *lnNoise = log( MAX(*noise,N_MINLOG) );
            break;

        case N_GAIN:
            *noise = gain;
            break;
    }
}
