/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Gary W. Ng
         1993 Stephen R. Whiteley
****************************************************************************/

/*
 * Nintegrate.c (noizDens, lnNdens, lnNlstDens, data)
 *
 *    This subroutine evaluates the integral of the function
 *
 *                                             EXPONENT
 *                      NOISE = a * (FREQUENCY)
 *
 *   given two points from the curve.  If EXPONENT is relatively close
 *   to 0, the noise is simply multiplied by the change in frequency.
 *   If it isn't, a more complicated expression must be used.  Note that
 *   EXPONENT = -1 gives a different equation than EXPONENT <> -1.
 *   Hence, the reason for the constant 'N_INTUSELOG'.
 */

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "cktdefs.h"
#include "noisedef.h"
#include "util.h"


double
Nintegrate(noizDens, lnNdens, lnNlstDens, dataptr)

double noizDens;
double lnNdens;
double lnNlstDens;
GENERIC *dataptr;
{
    Ndata *data = (Ndata*)dataptr;
    double exponent;
    double a;

    exponent = (lnNdens - lnNlstDens) / data->delLnFreq;

    if ( FABS(exponent) < N_INTFTHRESH ) {
        return (noizDens * data->delFreq);
    }
    else {
        a = exp(lnNdens - exponent*data->lnFreq);
        exponent += 1.0;

        if (FABS(exponent) < N_INTUSELOG) {
            return (a * (data->lnFreq - data->lnLastFreq));
        }
        else {
            return (a * ((exp(exponent * data->lnFreq) -
                exp(exponent * data->lnLastFreq)) / exponent));
        }
    }
}
