/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Gary W. Ng
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "noisedef.h"
#include "jobdefs.h"
#include "iferrmsg.h"


int 
NOISEsetParm(cktp,anal,which,value)

GENERIC *cktp;
GENERIC *anal;
int which;
IFvalue *value;
{
    NOISEAN *job = (NOISEAN *) anal;

    switch (which) {

        case N_OUTPUT:
            job->Noutput = value->sValue;
            break;

        case N_OUTREF:
            job->NoutputRef = value->sValue;
            break;

        case N_INPUT:
            job->Ninput = value->sValue;
            break;

        case N_PTSPERSUM:
            job->NStpsSm = value->iValue;
            break;

        default:
            if (ACsetp(&job->AC,which,value) == OK)
                return (OK);
            if (DCTsetp(&job->DC,which,value) == OK)
                return (OK);
            return (E_BADPARM);
    }
    return (OK);
}


static IFparm Nparms[] = {
 { "output",    N_OUTPUT,    IF_SET|IF_STRING,  "output noise summation node" },
 { "outputref", N_OUTREF,    IF_SET|IF_STRING,  "output noise reference node" },
 { "input",     N_INPUT,     IF_SET|IF_STRING,  "input noise source" },
 { "ptspersum", N_PTSPERSUM, IF_SET|IF_INTEGER, "frequency points per summary report" },
 { "dec",       AC_DEC,      IF_SET|IF_FLAG,    "step by decades" },
 { "oct",       AC_OCT,      IF_SET|IF_FLAG,    "step by octaves" },
 { "lin",       AC_LIN,      IF_SET|IF_FLAG,    "step linearly" },
 { "numsteps",  AC_STEPS,    IF_SET|IF_INTEGER, "number of frequencies" },
 { "start",     AC_START,    IF_SET|IF_REAL,    "starting frequency" },
 { "stop",      AC_STOP,     IF_SET|IF_REAL,    "ending frequency" },
 { "name1",     DC_NAME1,    IF_SET|IF_INSTANCE, "name of source to step" },
 { "start1",    DC_START1,   IF_SET|IF_REAL,     "starting voltage/current"},
 { "stop1",     DC_STOP1,    IF_SET|IF_REAL,     "ending voltage/current" },
 { "step1",     DC_STEP1,    IF_SET|IF_REAL,     "voltage/current step" },
 { "name2",     DC_NAME2,    IF_SET|IF_INSTANCE, "name of source to step" },
 { "start2",    DC_START2,   IF_SET|IF_REAL,     "starting voltage/current"},
 { "stop2",     DC_STOP2,    IF_SET|IF_REAL,     "ending voltage/current" },
 { "step2",     DC_STEP2,    IF_SET|IF_REAL,     "voltage/current step" }
};

SPICEanalysis NOISEinfo  = {
    { 
        "NOISE",
        "Noise analysis",

        sizeof(Nparms)/sizeof(IFparm),
        Nparms
    },
    sizeof(NOISEAN),
    FREQUENCYDOMAIN,
    NOISEparse,
    NOISEsetParm,
    NOISEaskQuest,
    NOISEan
};
