/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "optdefs.h"
#include "const.h"
#include "iferrmsg.h"


/* ARGSUSED */
int
OPTaskQuest(cktp,anal,which,value)

GENERIC *cktp;
GENERIC *anal;
int which;
IFvalue *value;
{
    CKTcircuit *ckt = (CKTcircuit*)cktp;
    STATistics *stat = (STATistics*)ckt->CKTstat;

    switch (which) {
        
        case OPT_EQNS:
            value->iValue = ckt->CKTmaxEqNum;
            break;
        case OPT_ITERS:
            value->iValue = stat->STATnumIter;
            break;
        case OPT_TRANIT:
            value->iValue = stat->STATtranIter;
            break;
        case OPT_TRANCURITER:
            value->iValue = stat->STATnumIter - stat->STAToldIter;
            break;
        case OPT_TRANPTS:
            value->iValue = stat->STATtimePts;
            break;
        case OPT_TRANACCPT:
            value->iValue = stat->STATaccepted;
            break;
        case OPT_TRANRJCT:
            value->iValue = stat->STATrejected;
            break;
        case OPT_TOTANALTIME:
            value->rValue = stat->STATtotAnalTime;
            break;
        case OPT_TRANTIME:
            value->rValue = stat->STATtranTime;
            break;
        case OPT_LOADTIME:
            value->rValue = stat->STATloadTime;
            break;
        case OPT_REORDTIME:
            value->rValue = stat->STATreorderTime;
            break;
        case OPT_DECOMP:
            value->rValue = stat->STATdecompTime;
            break;
        case OPT_SOLVE:
            value->rValue = stat->STATsolveTime;
            break;
        case OPT_TRANDECOMP:
            value->rValue = stat->STATtranDecompTime;
            break;
        case OPT_TRANSOLVE:
            value->rValue = stat->STATtranSolveTime;
            break;
        case OPT_TEMP:
            value->rValue = ckt->CKTtemp - CONSTCtoK;
            break;
        case OPT_TNOM:
            value->rValue = ckt->CKTnomTemp - CONSTCtoK;
            break;
        case OPT_MATSIZ:
            value->iValue = stat->STATmatSize;
            break;
        case OPT_NONZ:
            value->iValue = stat->STATnonZero;
            break;
        case OPT_FILLIN:
            value->iValue = stat->STATfillIn;
            break;
        default:
            return (E_BADPARM);
    }
    return (OK);
}
