/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: UCB CAD Group
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "devdefs.h"
#include "pzdefs.h"
#include "sperror.h"
#include "cktext.h"

extern SPICEdev *DEVices[];


int
CKTpzLoad(ckt, s)

CKTcircuit *ckt;
SPcomplex *s;
{
    PZAN *pzan = (PZAN *) (ckt->CKTcurJob);
    int solution_col;
    int error;
    struct sCKTmodHead *mh;
    int i, size;
    int (*func)();

    size = spGetSize(ckt->CKTmatrix,1);
    for (i = 0; i <= size; i++) {
        ckt->CKTrhs[i] = 0.0;
        ckt->CKTirhs[i] = 0.0;
    }

    spSetComplex(ckt->CKTmatrix);
    spClear(ckt->CKTmatrix);
    for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
        if ((func = DEVices[mh->type]->DEVpzLoad) != NULL) {
            error = (*func)(mh->head, ckt, s);
            if (error) return (error);
        }
    }

#ifdef notdef
    printf("*** Before PZ adjustments *\n");
    spPrint(ckt->CKTmatrix,0,1,1);
#endif

    if (pzan->PZbalance_col && pzan->PZsolution_col) {
        spAddCol(ckt->CKTmatrix, pzan->PZbalance_col, pzan->PZsolution_col);
        /* AC sources ?? XXX */
    }

    if (pzan->PZsolution_col) {
        spZeroCol(ckt->CKTmatrix, pzan->PZsolution_col);
    }

    /* Driving function (current source) */
    if (pzan->PZdrive_pptr)
        *pzan->PZdrive_pptr = 1.0;
    if (pzan->PZdrive_nptr)
        *pzan->PZdrive_nptr = -1.0;

#ifdef notdef
    printf("*** After PZ adjustments *\n");
    spPrint(ckt->CKTmatrix,0,1,1);
#endif

    return(OK);
}
