/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "ftedefs.h"
#include "inpdefs.h"
#include "inpmacs.h"
#include "misc.h"


/* ARGSUSED */
int
SENSparse(ckt,tab,which,currentp,line,task,gnode)

/*
    .sens output [ac {dec lin oct} NP FSTART FSTOP]
        [ dc SRC1NAME Vstart1 [Vstop1 [Vinc1]]
        [SRC2NAME Vstart2 [Vstop2 [Vinc2]]] ]
*/
GENERIC *ckt;
INPtables *tab;
int which;
GENERIC *currentp;
char **line;
GENERIC *task;
GENERIC *gnode;
{
    char *nname1;   /* the first node's name */
    char *nname2;   /* the second node's name */
    GENERIC *node;  /* node pointer */
    GENERIC *foo;   /* pointer to analysis */
    IFvalue ptemp;  /* a value structure to package stuff into */
    IFvalue *parm;  /* a pointer to a value struct for function returns */
    int error;      /* error code temporary */
    char *token;    /* a token from the line */
    card *current = (card *)currentp;

    IFC(newAnalysis,(ckt,which,"Sensitivity Analysis",&foo, task))

    /* Get the output voltage or current */
    INPgetTok(line,&token,1);
    /* token is now either V or I or a serious error */
    if ((*token == 'v' || *token == 'V') && strlen(token) == 1) {
        txfree(token);
        if (**line != '(' /* match) */) { 
            LITERR("Syntax error: '(' expected after 'v'");
            return (0);
        }
        INPgetTok(line, &nname1, 0);
        INPtermInsert(ckt, &nname1, tab, &node);
        ptemp.nValue = node;
        GCA(INPapName,(ckt,which,foo,"outpos",&ptemp))

        if (**line != /* match ( */ ')') {
            INPgetTok(line, &nname2, 1);
            INPtermInsert(ckt, &nname2, tab, &node);
            ptemp.nValue = node;
            GCA(INPapName,(ckt,which,foo,"outneg",&ptemp))
            ptemp.sValue = (char *)
                tmalloc(sizeof(char)*(5+strlen(nname1)+strlen(nname2)));
            (void)sprintf(ptemp.sValue,"v(%s,%s)",nname1,nname2);
            GCA(INPapName,(ckt,which,foo,"outname",&ptemp))
        }
        else {
            ptemp.nValue = gnode;
            GCA(INPapName,(ckt,which,foo,"outneg",&ptemp))
            ptemp.sValue = (char *)tmalloc(sizeof(char)*(4+strlen(nname1)));
            (void)sprintf(ptemp.sValue,"v(%s)",nname1);
            GCA(INPapName,(ckt,which,foo,"outname",&ptemp))
        }
    }
    else if (*token == 'i' && strlen(token) == 1) {
        txfree(token);
        INPgetTok(line,&token,1);
        strtolower(token);
        INPinsert(&token,tab);
        ptemp.uValue = token;
        GCA(INPapName,(ckt,which,foo,"outsrc",&ptemp))
    }
    else {
        txfree(token);
        LITERR("Syntax error: voltage or current expected.")
        return (0);
    }

    INPgetTok(line,&token,1);
    if (token && !strcmp(token, "pct")) {
        txfree(token);
        ptemp.iValue = 1;
        GCA(INPapName,(ckt,which,foo,"pct",&ptemp))
            INPgetTok(line,&token,1);
    }
    if (token && !strcmp(token, "ac")) {
        txfree(token);
        GCA(ParseAC,(ckt,line,current,which,foo,tab))
    }

    if (token && *token && strcmp(token, "dc")) {
        /* Bad flag */
        LITERR("Syntax error: 'ac' or 'dc' expected.")
    }
    txfree(token);
    return (0);
}
