/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1988 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#define ANAL_EXT

/* subroutine to do Transfer Function analysis */

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "srcdefs.h"
#include "util.h"
#include "sperror.h"
#include "outdata.h"
#include "tfdefs.h"
#include "cktext.h"

#ifdef __STDC__
static int tf_acoperation(CKTcircuit*,struct sOUTdata*,int);
static int tf_dcoperation(CKTcircuit*,struct sOUTdata*,int);
#else
static int tf_acoperation();
static int tf_dcoperation();
#endif


/* ARGSUSED */
int
TFan(cktp,restart)

GENERIC *cktp;
int restart;    /* forced restart flag */
{
    CKTcircuit *ckt = (CKTcircuit *)cktp;
    static struct sOUTdata outd;
    int error;
    int i;
    IFuid uids[3];
    int code;
    TFAN *job = (TFAN*)ckt->CKTcurJob;

    if (ckt->CKTjjPresent) {
        (*(SPfrontEnd->IFerror))(ERR_FATAL,
            "Transfer function not possible with Josephson junctions", NULL);
            return (OK);
    }

    if (restart) {

        /* make sure sources are present */
        code = CKTtypelook("Source");
        error = CKTfndDev((GENERIC*)ckt,&code,&job->TFinSrcDev,
            job->TFinSrc, (GENERIC *)NULL,(IFuid)NULL);
        if (error != 0) {
            (*(SPfrontEnd->IFerror))(ERR_FATAL,
                "Transfer function source %s not in circuit",
                &job->TFinSrc);
            return (E_NOTFOUND);
        }
        if (((SRCinstance*)job->TFinSrcDev)->SRCtype == SRC_V) {
            job->TFinIsV = 1;
            job->TFinIsI = 0;
        }
        else {
            job->TFinIsI = 1;
            job->TFinIsV = 0;
        }

        if (job->TFoutIsI) {
            error = CKTfndDev((GENERIC*)ckt,&code,&job->TFoutSrcDev,
                job->TFoutSrc,(GENERIC *)NULL,(IFuid)NULL);
            if (error != 0) {
                (*(SPfrontEnd->IFerror))(ERR_FATAL,
                    "Transfer function source %s not in circuit",
                    &job->TFoutSrc);
                return (E_NOTFOUND);
            }
            if (((SRCinstance*)job->TFoutSrcDev)->SRCtype != SRC_V) {
                (*(SPfrontEnd->IFerror))(ERR_FATAL,
                    "Transfer function source %s not a voltage source",
                    &job->TFoutSrc);
                return (E_NOTFOUND);
            }
        }

        if (job->DC.eltName[0] != NULL) {
            /* DC source was given */
            SRCinstance *here;

            here = (SRCinstance*)NULL;
            for (i = 0; i <= job->DC.nestLevel; i++) {
                error = CKTfndDev((GENERIC*)ckt,&code,(GENERIC**)&here,
                    job->DC.eltName[i], (GENERIC *)NULL,(IFuid)NULL);

                if (error) {
                    (*(SPfrontEnd->IFerror))(ERR_FATAL, 
                        "DCtrCurv: source %s not in circuit",
                        &(job->DC.eltName[i]));
                    return (E_NOTFOUND);
                }
                job->DC.elt[i] = (GENinstance*)here;
                job->DC.vsave[i] = here->SRCdcValue;
            }
        }

        /* make a UID for the transfer function output */
        (*(SPfrontEnd->IFnewUid))(ckt,&uids[0],(IFuid)NULL,"tranfunc",
                UID_OTHER,(GENERIC **)NULL);

        /* make a UID for the input impedance */
        (*(SPfrontEnd->IFnewUid))(ckt,&uids[1],job->TFinSrc,
                "Zi", UID_OTHER,(GENERIC **)NULL);

        /* make a UID for the output impedance */
        if (job->TFoutIsI) {
            (*(SPfrontEnd->IFnewUid))
                (ckt,&uids[2],job->TFoutSrc,
                "Zo", UID_OTHER,(GENERIC **)NULL);
        }
        else {
            char *name = (char *)
                MALLOC(sizeof(char)*(strlen(job->TFoutName)+22));
            char *s = strchr(job->TFoutName,'(');
            if (s)
                (void)sprintf(name,"Zo%s",s);
            else
                (void)sprintf(name,"Zo(%s)",job->TFoutName);
            (*(SPfrontEnd->IFnewUid))(ckt,&uids[2],(IFuid)NULL,
                    name, UID_OTHER,(GENERIC **)NULL);
        }

        if (job->AC.stepType == DCSTEP) {
            outd.dataType = IF_REAL;
            outd.refName = (IFuid)NULL;
            outd.refType = 0;
        }
        else {
            outd.dataType = IF_COMPLEX;
            (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&outd.refName,
                (IFuid)NULL, "frequency", UID_OTHER,(GENERIC **)NULL);
            outd.refType = IF_REAL;
        }

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = ckt->CKTcurJob->JOBname;
        outd.numNames = 3;
        outd.dataNames = uids;
        outd.plotPtr = &job->TFplot;
        outd.numPts = 1;
        outd.initValue = 0;
        outd.finalValue = 0;
        outd.step = 0;
        outd.count = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);
    }
    error = DCTloop(tf_dcoperation,ckt,restart,&job->DC,&outd);

    if (error == E_PAUSE)
        return (error);

    (*(SPfrontEnd->OUTendPlot))(*outd.plotPtr);
    return (OK);
}


static int
tf_dcoperation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    TFAN *job = (TFAN*)ckt->CKTcurJob;
    int error;
    int size;
    int i;
    double outputs[3];
    double A;
    IFvalue outdata;
    IFvalue refval;
    SRCinstance *ptr;

    /* find the operating point */

    error = CKTic(ckt);
    if (error)
        return (error);

    error = CKTop(ckt,
            MODEDCOP | MODEINITJCT,
            MODEDCOP | MODEINITFLOAT,
            ckt->CKTdcMaxIter);
    if (error)
        return (error);

    if (job->AC.stepType != DCSTEP) {
        error = ACloop(tf_acoperation,ckt,restart,&job->AC,outd);
        return (error);
    }

    size = spGetSize(ckt->CKTmatrix,1);

    for (i = 0; i <= size; i++) {
        ckt->CKTrhs[i] = 0;
    }

    ptr = (SRCinstance*)job->TFinSrcDev;
    if (job->TFinIsI) {
        ckt->CKTrhs[ptr->SRCposNode] -= 1;
        ckt->CKTrhs[ptr->SRCnegNode] += 1;
    }
    else {
        ckt->CKTrhs[ptr->SRCbranch] += 1;
    }

    spSolve(ckt->CKTmatrix,ckt->CKTrhs,ckt->CKTrhs,NULL,NULL);
    ckt->CKTrhs[0] = 0;

    /* find transfer function */
    if (job->TFoutIsV) {
        outputs[0] =
            ckt->CKTrhs[job->TFoutPos->number] -
                ckt->CKTrhs[job->TFoutNeg->number] ;
    }
    else {
        ptr = (SRCinstance*)job->TFoutSrcDev;
        outputs[0] = ckt->CKTrhs[ptr->SRCbranch];
    }

    /* now for input impedance */
    ptr = (SRCinstance*)job->TFinSrcDev;
    if (job->TFinIsI) {
        outputs[1] =
            ckt->CKTrhs[ptr->SRCnegNode] - ckt->CKTrhs[ptr->SRCposNode];
    }
    else {
        A = ckt->CKTrhs[ptr->SRCbranch];
        if (FABS(A) < 1e-20)
            outputs[1] = 1e20;
        else
            outputs[1] = -1/A;
    }

    if (job->TFoutIsI && (job->TFoutSrc == job->TFinSrc)) {
        outputs[2] = outputs[1];
        /* no need to compute output impedance when it is the same as 
         * the input
         */
    }
    else {
        /* now for output resistance */
        for (i = 0; i <= size; i++) {
            ckt->CKTrhs[i] = 0;
        }
        if (job->TFoutIsV) {
            ckt->CKTrhs[job->TFoutPos->number] -= 1;
            ckt->CKTrhs[job->TFoutNeg->number] += 1;
        }
        else {
            ptr = (SRCinstance*)job->TFoutSrcDev;
            ckt->CKTrhs[ptr->SRCbranch] += 1;
        }
        spSolve(ckt->CKTmatrix,ckt->CKTrhs,ckt->CKTrhs,
            ckt->CKTirhs,ckt->CKTirhs);
        ckt->CKTrhs[0] = 0;

        if (job->TFoutIsV) {
            outputs[2] =
                ckt->CKTrhs[job->TFoutNeg->number] -
                ckt->CKTrhs[job->TFoutPos->number];
        }
        else {
            A = ckt->CKTrhs[ptr->SRCbranch];
            if (FABS(A) < 1e-20)
                outputs[2] = 1e20;
            else
                outputs[2] = 1/A;
        }
    }

    outdata.v.numValue = 3;
    outdata.v.vec.rVec = outputs;
    refval.rValue = ckt->CKTomega/(2*M_PI);
    (*(SPfrontEnd->OUTdata))(*outd->plotPtr,&refval,&outdata);
    outd->count++;
    return (OK);
}


static int
tf_acoperation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    int size;
    double A, B, MAG;
    IFcomplex coutputs[3];
    IFvalue outdata;
    IFvalue refval;
    int error;
    int i;
    SRCinstance *ptr;
    TFAN *job = (TFAN*)ckt->CKTcurJob;

    ckt->CKTmode = MODEDCOP | MODEINITSMSIG;
    error = CKTload(ckt);
    if (error)
        return (error);
    ckt->CKTmode = MODEAC;
    error = NIacIter(ckt);
    if (error)
        return (error);

    size = spGetSize(ckt->CKTmatrix,1);

    for (i = 0; i <= size; i++) {
        ckt->CKTrhs[i] = 0;
        ckt->CKTirhs[i] = 0;
    }

    ptr = (SRCinstance*)job->TFinSrcDev;
    if (job->TFinIsI) {
        ckt->CKTrhs[ptr->SRCposNode] -= 1;
        ckt->CKTrhs[ptr->SRCnegNode] += 1;
    }
    else {
        ckt->CKTrhs[ptr->SRCbranch] += 1;
    }

    spSolve(ckt->CKTmatrix,ckt->CKTrhs,ckt->CKTrhs,
        ckt->CKTirhs,ckt->CKTirhs);
    ckt->CKTrhs[0] = 0;
    ckt->CKTirhs[0] = 0;

    /* find transfer function */
    if (job->TFoutIsV) {
        coutputs[0].real =
            ckt->CKTrhs[job->TFoutPos->number] -
                ckt->CKTrhs[job->TFoutNeg->number] ;
        coutputs[0].imag =
            ckt->CKTirhs[job->TFoutPos->number] -
                ckt->CKTirhs[job->TFoutNeg->number] ;
    }
    else {
        ptr = (SRCinstance*)job->TFoutSrcDev;
        coutputs[0].real = ckt->CKTrhs[ptr->SRCbranch];
        coutputs[0].imag = ckt->CKTirhs[ptr->SRCbranch];
    }

    /* now for input impedance */
    ptr = (SRCinstance*)job->TFinSrcDev;
    if (job->TFinIsI) {
        coutputs[1].real =
            ckt->CKTrhs[ptr->SRCnegNode] - ckt->CKTrhs[ptr->SRCposNode];
        coutputs[1].imag =
            ckt->CKTirhs[ptr->SRCnegNode] - ckt->CKTirhs[ptr->SRCposNode];
    }
    else {
        A = ckt->CKTrhs[ptr->SRCbranch];
        B = ckt->CKTirhs[ptr->SRCbranch];
        MAG = A*A + B*B;
        if (MAG < 1e-40)
            MAG = 1e-40;
        coutputs[1].real = -A/MAG;
        coutputs[1].imag = B/MAG;
    }

    if (job->TFoutIsI && (job->TFoutSrc == job->TFinSrc)) {
        coutputs[2].real = coutputs[1].real;
        coutputs[2].imag = coutputs[1].imag;
        /* no need to compute output impedance when it is the same as 
         * the input
         */
    }
    else {
        /* now for output resistance */
        for (i = 0; i <= size; i++) {
            ckt->CKTrhs[i] = 0;
            ckt->CKTirhs[i] = 0;
        }
        if (job->TFoutIsV) {
            ckt->CKTrhs[job->TFoutPos->number] -= 1;
            ckt->CKTrhs[job->TFoutNeg->number] += 1;
        }
        else {
            ptr = (SRCinstance*)job->TFoutSrcDev;
            ckt->CKTrhs[ptr->SRCbranch] += 1;
        }
        spSolve(ckt->CKTmatrix,ckt->CKTrhs,ckt->CKTrhs,
            ckt->CKTirhs,ckt->CKTirhs);
        ckt->CKTrhs[0] = 0;
        ckt->CKTirhs[0] = 0;

        if (job->TFoutIsV) {
            coutputs[2].real =
                ckt->CKTrhs[job->TFoutNeg->number] -
                ckt->CKTrhs[job->TFoutPos->number];
            coutputs[2].imag =
                ckt->CKTirhs[job->TFoutNeg->number] -
                ckt->CKTirhs[job->TFoutPos->number];
        }
        else {
            A = ckt->CKTrhs[ptr->SRCbranch];
            B = ckt->CKTirhs[ptr->SRCbranch];
            MAG = A*A + B*B;
            if (MAG < 1e-40)
                MAG = 1e-40;
            coutputs[2].real = A/MAG;
            coutputs[2].imag = -B/MAG;
        }
    }

    outdata.v.numValue = 3;
    outdata.v.vec.cVec = coutputs;
    refval.rValue = ckt->CKTomega/(2*M_PI);
    (*(SPfrontEnd->OUTdata))(*outd->plotPtr,&refval,&outdata);
    outd->count++;
    return (OK);
}
