/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "ftedefs.h"
#include "inpdefs.h"
#include "inpmacs.h"
#include "misc.h"


/* ARGSUSED */
int
TRANparse(ckt,tab,which,currentp,line,task,gnode)

/*
    .tran Tstep Tstop [Tstart] [Tmax] [UIC]
        [ dc SRC1NAME Vstart1 [Vstop1 [Vinc1]]
        [SRC2NAME Vstart2 [Vstop2 [Vinc2]]] ]
*/
GENERIC *ckt;
INPtables *tab;
int which;
GENERIC *currentp;
char **line;
GENERIC *task;
GENERIC *gnode;
{
    GENERIC *foo;   /* pointer to analysis */
    IFvalue ptemp;  /* a value structure to package stuff into */
    int error;      /* error code temporary */
    char *token;    /* a token from the line */
    card *current = (card *)currentp;

    IFC(newAnalysis,(ckt,which,"transient",&foo,task))
    GCA(ParseTRAN,(ckt,line,current,which,foo,tab))
    while (**line) {
        INPgetTok(line, &token, 1);
        strtolower(token);
        if (!strcmp(token,"dc")) {
            txfree(token);
            GCA(ParseDC,(ckt,line,current,which,foo,tab,1))
            if (**line)
                GCA(ParseDC,(ckt,line,current,which,foo,tab,2))
            continue;
        }
        if (!strcmp(token,"uic")) {
            txfree(token);
            ptemp.iValue = 1;
            GCA(INPapName,(ckt,which,foo,"uic",&ptemp))
            continue;
        }
        LITERR("Unknown parameter on .tran - ignored.")
        txfree(token);
    }

    return (0);
}


int
ParseTRAN(ckt,line,current,which,aptr,tab)

GENERIC *ckt;
char **line;
card *current;
int which;
GENERIC *aptr;
INPtables *tab;
{
    IFvalue ptemp;  /* a value structure to package stuff into */
    IFvalue *parm;  /* a pointer to a value struct for function returns */
    double dtemp;   /* random double precision temporary */
    int error;      /* error code temporary */

    parm = INPgetValue(ckt,line,IF_REAL,tab);    /* Tstep */
    GCA(INPapName,(ckt,which,aptr,"tstep",parm))

    parm = INPgetValue(ckt,line,IF_REAL,tab);    /* Tstop */
    GCA(INPapName,(ckt,which,aptr,"tstop",parm))

    if (**line) {
        dtemp = INPevaluate(line,&error,1);      /* tstart? */
        if (error == 0) {
            ptemp.rValue = dtemp;
            GCA(INPapName,(ckt,which,aptr,"tstart",&ptemp))
            dtemp = INPevaluate(line,&error,1);  /* tmax? */
            if (error == 0) {
                ptemp.rValue = dtemp;
                GCA(INPapName,(ckt,which,aptr,"tmax",&ptemp))
            }
        }
    }
    return (0);
}
