/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

    /* CKTdestroy(ckt)
     * this is a driver program to iterate through all the various
     * destroy functions provided for the circuit elements in the
     * given circuit 
     */

#include "spice.h"
#include <stdio.h>
#include "devdefs.h"
#include "sperror.h"
#include "util.h"
#include "cktext.h"

extern SPICEdev *DEVices[];

int
CKTdestroy(inCkt)

GENERIC *inCkt;
{
    CKTcircuit *ckt = (CKTcircuit *)inCkt;
    int i;
    CKTnode *node;
    CKTnode *nnode;
    struct sCKTmodHead *mh, *mhnext;


    for (i = 0; i < DEVmaxnum; i++) {
        if ((DEVices[i]->DEVdestroy != NULL) &&
                (ckt->CKThead[i] != NULL)) {
            (*DEVices[i]->DEVdestroy)(&ckt->CKThead[i]);
        }
    }

    for (mh = ckt->CKTheadList; mh != NULL; mh = mhnext) {
        mhnext = mh->next;
        FREE(mh);
    }

    for (i = 0; i <= ckt->CKTmaxOrder+1; i++) {
        FREE(ckt->CKTstates[i]);
    }

    if (ckt->CKTmatrix) spDestroy(ckt->CKTmatrix);
    if (ckt->CKTbreaks) FREE(ckt->CKTbreaks);

    for (node = ckt->CKTnodes; node; ) {
        nnode = node->next;
        FREE(node);
        node = nnode;
    }

    FREE(ckt->CKTtimePoints);

    INPtablFree((GENERIC*)ckt);

    FREE(ckt);
    return (OK);
}
