/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "optdefs.h"
#include "jobdefs.h"
#include "tskdefs.h"
#include "sperror.h"
#include "util.h"
#include "cktext.h"

extern int ANALmaxnum;
extern FILE *cp_err;


int
CKTdoJob(inCkt,reset,inTask)

GENERIC *inCkt;
int reset;
GENERIC *inTask;
{
    CKTcircuit *ckt = (CKTcircuit *)inCkt;
    TSKtask    *task = (TSKtask *)inTask;
    JOB        *job;
    double     startTime;
    int        error, i;

    startTime = (*(SPfrontEnd->IFseconds))( );

    ckt->CKTtemp  = task->TSKtemp;
    ckt->CKTnomTemp  = task->TSKnomTemp;
    ckt->CKTmaxOrder  = task->TSKmaxOrder;
    ckt->CKTintegrateMethod  = task->TSKintegrateMethod;
    ckt->CKTbypass  = task->TSKbypass;
    ckt->CKTdcMaxIter  = task->TSKdcMaxIter;
    ckt->CKTdcTrcvMaxIter  = task->TSKdcTrcvMaxIter;
    ckt->CKTtranMaxIter  = task->TSKtranMaxIter;
    ckt->CKTnumSrcSteps  = task->TSKnumSrcSteps;
    ckt->CKTnumGminSteps  = task->TSKnumGminSteps;
    ckt->CKTminBreak  = task->TSKminBreak;
    ckt->CKTabstol  = task->TSKabstol;
    ckt->CKTpivotAbsTol  = task->TSKpivotAbsTol;
    ckt->CKTpivotRelTol  = task->TSKpivotRelTol;
    ckt->CKTreltol  = task->TSKreltol;
    ckt->CKTchgtol  = task->TSKchgtol;
    ckt->CKTvoltTol  = task->TSKvoltTol;
    ckt->CKTgmin  = task->TSKgmin;
    ckt->CKTdelmin  = task->TSKdelmin;
    ckt->CKTtrtol  = task->TSKtrtol;
    ckt->CKTdefaultMosL  = task->TSKdefaultMosL;
    ckt->CKTdefaultMosW  = task->TSKdefaultMosW;
    ckt->CKTdefaultMosAD  = task->TSKdefaultMosAD;
    ckt->CKTdefaultMosAS  = task->TSKdefaultMosAS;
    ckt->CKTfixLimit  = task->TSKfixLimit;
    ckt->CKTnoOpIter  = task->TSKnoOpIter;
    ckt->CKTtryToCompact = task->TSKtryToCompact;
    ckt->CKTtroubleNode  = 0;
    ckt->CKTtroubleElt  = NULL;
#ifdef NEWTRUNC
    ckt->CKTlteReltol = task->TSKlteReltol;
    ckt->CKTlteAbstol = task->TSKlteAbstol;
#endif /* NEWTRUNC */


    if (ckt->CKTmaxEqNum <= 1)
        /* Circuit has nothing in it! Just return. */ 
        return (OK);

    for (i = 0; i < ANALmaxnum; i++) {

        for (job = task->jobs; job; job = job->JOBnextJob) {

            if (job->JOBtype == i) {

                if (reset) {

                    ckt->CKTdelta = 0.0;
                    ckt->CKTtime = 0.0;
                    ckt->CKTcurrentAnalysis = 0;
                    ckt->CKTjjPresent = 0;

                    ckt->CKTpreload = 1; /* do preload */
                    /* normal reset */
                    error = CKTsetup(ckt);
                    if (error) return (error);
                    error = CKTtemp(ckt);
                    if (error) return (error);

                    spRtoI((char*)ckt->CKTmatrix);

                    /* The unchanging (real) constants in the matrix
                     * are preloaded during setup.  For DC and TRAN
                     * analysis (including operating point for AC), we
                     * copy the real part of the matrix into an
                     * otherwise unused field.  The (dc) load
                     * function, rather than clearing the real matrix,
                     * loads it with this field.  This saves a
                     * bit of computation.
                     */

                    ckt->CKTcurJob = job;
                }
                else {
                    if (ckt->CKTcurJob != job)
                        continue;
                }
                error = (*(analInfo[i]->an_func))((GENERIC*)ckt, reset);
                if (error) {
                    ((STATistics*)ckt->CKTstat)->STATtotAnalTime +=
                        (*(SPfrontEnd->IFseconds))( ) - startTime;
                    return (error);
                }
                if (!reset)
                    reset = TRUE;
            }
        }
    }

    ((STATistics*)ckt->CKTstat)->STATtotAnalTime +=
        (*(SPfrontEnd->IFseconds))( ) - startTime;

    return (OK);
}
