/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

    /* CKTload(ckt)
     * this is a driver program to iterate through all the various
     * load functions provided for the circuit elements in the
     * given circuit 
     */

#include "spice.h"
#include <stdio.h>
#include "devdefs.h"
#include "optdefs.h"
#include "sperror.h"
#include "cktext.h"


int
CKTload(ckt)

CKTcircuit *ckt;
{
    extern SPICEdev *DEVices[];
    int i;
    int size;
    int error;
    int (*func)();
    double startTime;
    CKTnode *node;
    struct sCKTmodHead *mh;
#ifdef STEPDEBUG
    int noncon;
#endif /* STEPDEBUG */

    startTime = (*(SPfrontEnd->IFseconds))();
    size = spGetSize(ckt->CKTmatrix,1);
    for (i = 0; i <= size; i++) {
        *(ckt->CKTrhs + i) = 0;
    }

    /* Reset the real part of the matrix by loading the imaginary
     * part, in which things might be cached.
     */
    spItoR((char*)ckt->CKTmatrix);

#ifdef STEPDEBUG
    noncon = ckt->CKTnoncon;
#endif /* STEPDEBUG */

    for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
        if ((func = DEVices[mh->type]->DEVload) != NULL) {
            error = (*func)(mh->head,ckt);

#ifdef STEPDEBUG
            if(noncon != ckt->CKTnoncon) {
                printf("device type %s nonconvergence\n",
                    DEVices[mh->type]->DEVpublic.name);
                noncon = ckt->CKTnoncon;
            }
#endif /* STEPDEBUG */

            if (error) return (error);
        }
    }
    if (ckt->CKTmode & MODEDC) {
        /* consider doing nodeset & ic assignments */
        if (ckt->CKTmode & (MODEINITJCT | MODEINITFIX)) {
            /* do nodesets */
            for (node = ckt->CKTnodes; node; node = node->next) {
                if (node->nsGiven) {
                    *(ckt->CKTrhs + node->number) += node->nodeset;
                    *(node->ptr) += 1;
                }
            }
        }
        if ((ckt->CKTmode & MODETRANOP) && (!(ckt->CKTmode & MODEUIC))) {
            for (node = ckt->CKTnodes; node; node = node->next) {
                if (node->icGiven) {
                    *(ckt->CKTrhs + node->number) += node->ic;
                    *(node->ptr) += 1;
                }
            }
        }
    }
    ((STATistics*)ckt->CKTstat)->STATloadTime +=
        (*(SPfrontEnd->IFseconds))() - startTime;
    return (OK);
}
