/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

    /* CKTsetup(ckt)
     * this is a driver program to iterate through all the various
     * setup functions provided for the circuit elements in the
     * given circuit 
     */

#include "spice.h"
#include <stdio.h>
#include "devdefs.h"
#include "sperror.h"
#include "util.h"
#include "cktext.h"
#include "niext.h"


#define CKALLOC(var,size,type) \
    if(size && (!(var =(type *)MALLOC((size)*sizeof(type))))){\
            return(E_NOMEM);\
}


int
CKTsetup(ckt)

CKTcircuit *ckt;
{
    extern SPICEdev *DEVices[];
    int i;
    int error;
    struct sCKTmodHead *mh;
    int (*func)();
    SMPmatrix *matrix;


    ckt->CKTnumStates = 0;
    matrix = ckt->CKTmatrix;

    /* we use the imag part for caching constants */
    spSetComplex(matrix);
    spClear(matrix);
    spSetReal(matrix);

    for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
        if ((func = DEVices[mh->type]->DEVsetup) != NULL) {
            error = (*func)(matrix,mh->head,ckt,&ckt->CKTnumStates);
            if (error) return (error);
        }
    }

    for (i = 0; i <= ckt->CKTmaxOrder+1; i++) {
        FREE(ckt->CKTstates[i]);
        CKALLOC(ckt->CKTstates[i],ckt->CKTnumStates,double);
    }
    if (ckt->CKTniState & NIUNINITIALIZED) {
        error = NIreinit(ckt);
        if (error) return (error);
    }

    return (OK);
}
