/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "util.h"
#include "cktext.h"

#define ccap (qcap+1)


void
CKTterr(qcap,ckt,timeStep)

int qcap;
CKTcircuit *ckt;
double *timeStep;
{ 
    double volttol;
    double chargetol;
    double tol;
    double del;
    double y0,y1,y2,y3;
    double *dd = ckt->CKTsols[0];
    int i;
    int j;
    static double gearCoeff[] = {
        .5,
        .2222222222,
        .1363636364,
        .096,
        .07299270073,
        .05830903790
    };

    y0 = *(ckt->CKTstate0 + ccap);
    y1 = *(ckt->CKTstate1 + ccap);
    if (y0 < 0) y0 = -y0;
    if (y1 < 0) y1 = -y1;
    volttol = ckt->CKTabstol + ckt->CKTreltol*MAX(y0,y1);
            
    y0 = *(ckt->CKTstate0 + qcap);
    y1 = *(ckt->CKTstate1 + qcap);
    if (y0 < 0) y0 = -y0;
    if (y1 < 0) y1 = -y1;
    chargetol = MAX(y0,y1);
    if (chargetol < ckt->CKTchgtol)
        chargetol = ckt->CKTchgtol;
    chargetol *= dd[5]; /* reltol/delta */

    tol = MAX(volttol,chargetol);

    if (ckt->CKTintegrateMethod == TRAPEZOIDAL) {

        if (ckt->CKTorder == 1) {
            y0 = *(ckt->CKTstate0 + qcap);
            y1 = *(ckt->CKTstate1 + qcap);
            y2 = *(ckt->CKTstate2 + qcap);
            del = (y0-y1)*dd[0] - (y1-y2)*dd[1];
        }
        else {
            y0 = *(ckt->CKTstate0 + qcap);
            y1 = *(ckt->CKTstate1 + qcap);
            y2 = *(ckt->CKTstate2 + qcap);
            y3 = *(ckt->CKTstate3 + qcap);
            del = (y0-y1)*dd[0] - (y1-y2)*dd[1] + (y2-y3)*dd[2];
        }
    }
    else {
        double diff[8];
        double deltmp[8];

        /* now divided differences */
        for (i = ckt->CKTorder + 1; i >= 0; i--) {
            diff[i] = *(ckt->CKTstates[i] + qcap);
        }
        for (i = 0; i <= ckt->CKTorder; i++) {
            deltmp[i] = ckt->CKTdeltaOld[i];
        }
        j = ckt->CKTorder;
        for (;;) {
            for (i = 0; i <= j; i++) {
                diff[i] = (diff[i] - diff[i+1])/deltmp[i];
            }
            if (--j < 0) break;
            for (i = 0; i <= j; i++) {
                deltmp[i] = deltmp[i+1] + ckt->CKTdeltaOld[i];
            }
        }
        del = diff[0]*gearCoeff[ckt->CKTorder-1];
    }
    if (del < 0) del = -del;
    del = ckt->CKTtrtol*tol/MAX(ckt->CKTabstol,del);

    /* normalization now done in CKTtrunc() */
    /*
    if (ckt->CKTorder == 2) {
        del = sqrt(del);
    }
    else if (ckt->CKTorder > 2) {
        del = exp(log(del)/ckt->CKTorder);
    }
    */

    if (del < *timeStep)
        *timeStep = del;
    return;
}
