/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

    /* CKTtrunc(ckt)
     * this is a driver program to iterate through all the various
     * truncation error functions provided for the circuit elements in the
     * given circuit 
     */

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "devdefs.h"
#include "util.h"
#include "sperror.h"
#include "cktext.h"


extern SPICEdev *DEVices[];

int
CKTtrunc(ckt,timeStep)

CKTcircuit *ckt;
double *timeStep;
{
    double timetemp;
    double temp;
    struct sCKTmodHead *mh;
    int error;
    int (*func)();
    double dd[6];
    double *dtmp;

    timetemp = HUGE;

    /* For trapezoid integration, cache the parameters for
     * fast evaluation.
     */
    if (ckt->CKTintegrateMethod == TRAPEZOIDAL) {
        if (ckt->CKTorder == 1) {

            dd[0] = 1/ckt->CKTdeltaOld[0];
            dd[1] = 1/ckt->CKTdeltaOld[1];
            dd[2] = .5/(ckt->CKTdeltaOld[0] + ckt->CKTdeltaOld[1]);

            dd[0] *= dd[2];
            dd[1] *= dd[2];
        }
        else {
            dd[0] = 1/ckt->CKTdeltaOld[0];
            dd[1] = 1/ckt->CKTdeltaOld[1];
            dd[2] = 1/ckt->CKTdeltaOld[2];
            dd[3] = 1/(ckt->CKTdeltaOld[0] + ckt->CKTdeltaOld[1]);
            dd[4] = 1/(ckt->CKTdeltaOld[1] + ckt->CKTdeltaOld[2]);
            dd[5] = .08333333333/(ckt->CKTdeltaOld[0] + ckt->CKTdeltaOld[1] +
                ckt->CKTdeltaOld[2]);

            dd[3] *= dd[5];
            dd[4] *= dd[5];
            dd[0] *= dd[3];
            dd[1] *= dd[3]+dd[4];
            dd[2] *= dd[4];
        }
    }
    /* cache this */
    dd[5] = ckt->CKTreltol/ckt->CKTdelta;

    /* temporary use of CKTsols[0] to pass coeffs (to CKTterr()) */
    dtmp = ckt->CKTsols[0];
    ckt->CKTsols[0] = dd;

    for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
        if ((func = DEVices[mh->type]->DEVtrunc) != NULL) {

#ifdef STEPDEBUG
            temp = timetemp;
#endif /* STEPDEBUG */
            error = (*func)(mh->head,ckt,&timetemp);
            if (error) {
                return (error);
            }
#ifdef STEPDEBUG
            if (temp != timetemp) {
                printf("timestep cut by device type %s from %g to %g\n",
                    DEVices[mh->type]->DEVpublic.name, temp,timetemp);
            }
#endif /* STEPDEBUG */
        }
    }

    /* Here we normalize to integration order ( not in CKTterr() ) to
     * avoid the math operation at each call.  This works directly for
     * all devices which simply call CKTterr() in the trunc functions.
     *
     * !!!> The LTRAtrunc() code needs fixing if you change this <!!!
     *
     */
    if (ckt->CKTorder == 2) {
        timetemp = sqrt(timetemp);
    }
    else if (ckt->CKTorder > 2) {
        timetemp = exp(log(timetemp)/ckt->CKTorder);
    }

    ckt->CKTsols[0] = dtmp;
    temp = *timeStep;
    temp += temp;
    *timeStep = MIN(temp,timetemp);
    return(OK);
}

