/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Do backquote substitution on a word list. 
 */

#include "spice.h"
#include "misc.h"
#include "cpdefs.h"
#include "suffix.h"

#ifdef __STDC__
static wordlist *backeval(char*);
#else
static wordlist *backeval();
#endif

char cp_back = '`';


void
cp_bquote(list)

wordlist **list;
{
    wordlist *wl, *nwl, *owl, *wlist;
    char *s, *t, buf[BSIZE_SP], wbuf[BSIZE_SP];
    int i;

    if (list == NULL)
        return;
    wlist = *list;

    for (wl = wlist; wl; wl = wl->wl_next) {
        t = wl->wl_word;
        if (!t)
            continue;
        i = 0;

        while (s = strchr(t, cp_back)) {
            while (t < s)
                wbuf[i++] = *t++;
            wbuf[i] = '\0';
            (void) strcpy(buf, ++s);
            s = buf;
            t++;
            while (*s && (*s != cp_back)) {
                /* Get s and t past the next backquote. */
                t++;
                s++;
            }
            /* What the heck, let "echo `foo" work... */
            *s = '\0';
            t++;    /* Get past the second ` */
            if (!(nwl = backeval(buf)) || !nwl->wl_word) {
                wl_free(wlist);
                *list = NULL;
                return;
            }
            if (i) {
                (void) strcpy(buf, wbuf);
                if (nwl->wl_word) {
                    (void) strcat(buf, nwl->wl_word);
                    tfree(nwl->wl_word);
                }
                nwl->wl_word = copy(buf);
            }
            owl = wl;
            wl = wl_splice(wl, nwl);
            if (owl == wlist)
                wlist = nwl;
            (void) strcpy(buf, wl->wl_word);
            i = strlen(buf);
            (void) strcat(buf, t);
            tfree(wl->wl_word);
            wl->wl_word = copy(buf);
            t = &wl->wl_word[i];
            s = wl->wl_word;
            for (i = 0; s < t; s++)
                wbuf[i++] = *s;
        }
    }
    *list = wlist;
    return;
}


/* Do a popen with the string, and then reset the file pointers so that
 * we can use the first pass of the parser on the output.
 */

static wordlist *
backeval(string)

char *string;
{
    FILE *proc, *old;
    wordlist *wl;
    bool intv;
    extern FILE *popen( );

    proc = popen(string, "r");
    if (proc == NULL) {
        fprintf(cp_err, "Error: can't evaluate %s.\n", string);
        return (NULL);
    }
    old = cp_inp_cur;
    cp_inp_cur = proc;
    intv = cp_interactive;
    cp_interactive = false;
    cp_bqflag = true;
    wl = cp_lexer((char *) NULL);
    cp_bqflag = false;
    cp_inp_cur = old;
    cp_interactive = intv;
    (void) pclose(proc);
    return (wl);
}
