/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1986 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Routines to handle "more"d output.  There are some serious system
 * dependencies in here, and it isn't clear that versions of this stuff
 * can be written for every possible machine...
 */

#include "spice.h"
#include "cpdefs.h"
#define SHORT_SCEDIO
#include "scedio.h"
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif


#ifdef HAVE_SGTTY_H
#include <sgtty.h>
#endif

#include "suffix.h"

#define DEF_SCRHEIGHT   24
#define DEF_SCRWIDTH    80

#ifdef __STDC__
static void outbufputc(void);
static int  fbufputc(int);
static void promptreturn(void);
#else
static void outbufputc();
static int  fbufputc();
static void promptreturn();
#endif

bool out_moremode = true;
bool out_isatty = true;
int out_width;
int out_height;

static int xsize, ysize;
static int xpos, ypos;
static bool noprint, nopause;


void
out_winsize(cols,rows)

int *cols, *rows;
{
    int x, y;
#ifdef TIOCGWINSZ
    struct winsize ws;
#endif
    char tbuf[4096], *s;

    x = y = 0;

    /* Figure out the screen size.  We try, in order, TIOCGSIZE,
     * tgetent(), and cp_getvar(height).  Default is 24 x 80.
     */

#ifdef TIOCGWINSZ
    if (!x || !y) {
        (void) ioctl(fileno(stdout), TIOCGWINSZ, (char *) &ws);
        x = ws.ws_col;
        y = ws.ws_row;
    }
#endif

#ifdef HAVE_TGETENT
    if (!x || !y) {
        if ((s = getenv("TERM")) && (tgetent(tbuf, s) != -1)) {
            x = tgetnum("co");
            y = tgetnum("li");
            if ((x <= 0) || (y <= 0))
                x = y = 0;
        }
    }
#endif

    if (!x) {
        if ((s = getenv("COLS"))) {
            x = atoi(s);
            if (x <= 0)
                x = 0;
        }
    }

    if (!y) {
        if ((s = getenv("LINES"))) {
            y = atoi(s);
            if (y <= 0)
                y = 0;
        }
    }

    if (!x)
        if (out_width > 0) x = out_width;
    if (!y)
        if (out_height > 0) y = out_height;

    if (!x)
        x = DEF_SCRWIDTH;
    if (!y)
        y = DEF_SCRHEIGHT;

    if (cols != NULL)
        *cols = x;
    if (rows != NULL)
        *rows = y;

    return;
}


/* Start output... */

void
out_init()
{
    bool moremode;

    noprint = nopause = false;

    if (!out_moremode || !cp_interactive)
        out_isatty = false;

    if (!out_isatty)
        return;

    out_winsize(&xsize,&ysize);
    ysize -= 2; /* Fudge room... */
    xpos = ypos = 0;

    return;
}


/* Putc may not be buffered (sp?), so we do it ourselves. */

static char staticbuf[BUFSIZ];
struct {
    int count;
    char *ptr;
} ourbuf = { BUFSIZ, staticbuf };


/* send buffer out */
static void
outbufputc()
{
    if (ourbuf.count != BUFSIZ) {
      fputs(staticbuf, cp_out);
      bzero(staticbuf, BUFSIZ-ourbuf.count);
      ourbuf.count = BUFSIZ;
      ourbuf.ptr = staticbuf;
    }
}

#define bufputc(c)  ( --ourbuf.count >= 0 ? ((int) \
    (*ourbuf.ptr++ = (unsigned)(c))) : fbufputc((unsigned) (c)))


static int
fbufputc(c)

unsigned char c;
{

    ourbuf.count = 0;
    outbufputc();
    ourbuf.count = BUFSIZ;
    ourbuf.ptr = staticbuf;
    bufputc(c);

}


static void
promptreturn()
{
    char buf[16];
    int attr;
    char *xx = "        -- hit return for more, ? for help -- ";
    char *menu =
"\nPossible responses:\n\
    q      : Discard the rest of the output.\n\
    c      : Continuously print the rest of the output.\n\
    h,?    : Print this help message.\n\
    other  : Print the next page of output.\n";

moe:
    fprintf(cp_out,"\n");
    attr = getattr();
    if (cp_out == stdout)
        cprint(15,xx);
    else {
        (void) fprintf(cp_out,xx);
        (void) fflush(cp_out);
    }
    if (cp_in == stdin)
        *buf = cp_inchar(cp_in);
    else if (!fgets(buf, 16, cp_in)) {
        clearerr(cp_in);
        *buf = 'q';
    }
    if (cp_out == stdout)
        cprint(attr,"\r                                                \r");

    switch (*buf) {
        case 'q':
            noprint = true;
            break;
        case 'c':
            nopause = true;
            break;
        case '?':
        case 'h':
            if (cp_out = stdout) {
                cprint(4,menu);
                cprint(attr," \n"); /* keeps the cursor the attr color */
            }
            else {
                fprintf(cp_out,menu);
            }
            goto moe;
    }
}


/* Print a string to the output.  If this would cause the screen to scroll,
 * print "more".
 */

void
out_send(string)

char *string;
{

    if (noprint)
        return;

    if (SCEDline(string))
        return;

    if (!out_isatty || nopause) {
        fputs(string, cp_out);
        return;
    }
    while (*string) {
        switch (*string) {
            case '\n':
                xpos = 0;
                ypos++;
                break;
            case '\f':
                ypos = ysize;
                xpos = 0;
                break;
            case '\t':
                xpos = xpos / 8 + 1;
                xpos *= 8;
                break;
            default:
                xpos++;
                break;
        }
        while (xpos >= xsize) {
            xpos -= xsize;
            ypos++;
        }
        if (ypos >= ysize) {
            outbufputc();       /* out goes buffer */
            promptreturn();
            (void) fflush(cp_out);
            ypos = xpos = 0;
        }
        bufputc(*string);   /* we need to buffer these */
        string++;
    }
    (void) outbufputc();
    return;
}

/* Printf some stuff using more mode. */

#define MAXLEN 4096


#ifdef __STDC__

void
out_printf(char *fmt, ...)

{
    va_list args;
    char buf[MAXLEN];

    va_start(args,fmt);
    (void) vsprintf(buf, fmt, args);
    va_end(args);

    out_send(buf);
    return;
}

#else


void
out_printf(va_alist)
va_dcl
{
    va_list args;
    char *fmt;
    char buf[MAXLEN];

    va_start(args);
    fmt = va_arg(args,char *);
    (void) vsprintf(buf, fmt, args);
    va_end(args);

    out_send(buf);
    return;
}

#endif  /* __STDC__ */


void
out_wlprint(wl)

/* print a wordlist */
wordlist* wl;
{
    wordlist *wlc = wl_copy(wl);

    cp_striplist(wlc);

    for (wl = wlc; wl; wl = wl->wl_next) {
        out_send(wl->wl_word);
        if (wl->wl_next)
            out_send(" ");
    }
    wl_free(wlc);
}
