/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Routines to do execution of operating system commands.
 */

#include "spice.h"
#include "cpdefs.h"
#define SHORT_SCEDIO
#include "scedio.h"

#include <sys/types.h>
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#ifndef direct
#define direct dirent
#endif
#else
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#endif

#ifdef __STDC__
static void upd_hash(bool);
static void add_dir(char*,bool,void**);
#else
static void upd_hash();
static void add_dir();
#endif

static void *hashtabp; /* before . in path */
static void *hashtabc; /* . */
static void *hashtabs; /* after . in path */
static char *dirbuffer;
static int dirlength, dirpos;


/* Create the hash table for the given search path. pathlist is a : seperated
 * list of directories. If docc is true, then all the commands found are
 * added to the command completion lists.
 */

void
cp_rehash(pathlist, docc)

char *pathlist;
bool docc;
{
    int i, tabp;
    char buf[BSIZE_SP];

    if (!pathlist) {
        /* called from the cd command */
        upd_hash(docc);
        return;
    }

    /* Clear out the old hash table. */
    htab_free(hashtabp,true);
    htab_free(hashtabc,true);
    htab_free(hashtabs,true);

    if (docc) {
        struct comm *c;
        cp_ccrestart(false);
        for (c = cp_coms; c->co_func; c++)
            cp_addcomm(c->co_comname, c->co_cctypes[0],
                c->co_cctypes[1], c->co_cctypes[2],c->co_cctypes[3]);
    }

    tabp = 0;
    while (pathlist && *pathlist) {
        i = 0;
        while (*pathlist && *pathlist != ';' && *pathlist != ':')
            buf[i++] = *pathlist++;
        while (*pathlist == ';' || *pathlist == ':')
            pathlist++;
        buf[i] = '\0';
        if (!strcmp(buf,".")) {
            if (tabp)
                continue;
            tabp++;
        }
        if (tabp == 0)
            add_dir(buf,docc,&hashtabp);
        else if (tabp == 1) {
            add_dir(buf,docc,&hashtabc);
            tabp++;
        }
        else
            add_dir(buf,docc,&hashtabs);
    }
    return;
}


static void
upd_hash(docc)

/* update hash tables - quicker than rebuilding after cd */
bool docc;
{
    wordlist *wl, *wl0;

    wl0 = wl = (wordlist*)htab_wl(hashtabc);
    while (wl) {
        if (!htab_get(wl->wl_word,hashtabp) &&
                !htab_get(wl->wl_word,hashtabs)) {
            cp_remcomm(wl->wl_word);
        }
        wl = wl->wl_next;
    }
    wl_free(wl0);

    htab_free(hashtabc,true);
    hashtabc = (void*)NULL;
    add_dir(".",docc,&hashtabc);
}


bool
cp_unixcom(wl)

/* The return value is false if no command was found, and true if it was. */
wordlist *wl;
{
    int i;
    bool ret;
    char *name;
    char *path;
    char **argv;
    char buf[BSIZE_SP];

    if (SCEDactive()) {
        ShowPrompt("Can't do this from sced.");
        return (true);
    }

    if (!wl)
        return (false);
    name = wl->wl_word;
    argv = wl_mkvec(wl);
    if (cp_debug) {
        printf("name: %s, argv: ", name);
        wl_print(wl, stdout);
        printf("\n");
    }
    if (strchr(name, DIR_TERM)) {
        ret = tryexec(name, argv);
        goto end;
    }
    path = (char*) htab_get(name,hashtabp);
    if (path) {
        (void) sprintf(buf,"%s%c%s",path,DIR_TERM,name);
        ret = tryexec(buf, argv);
        goto end;
    }
    path = (char*) htab_get(name,hashtabc);
    if (path) {
        (void) sprintf(buf,"%s",name);
        ret = tryexec(buf, argv);
        goto end;
    }
    path = (char*) htab_get(name,hashtabs);
    if (path) {
        (void) sprintf(buf,"%s%c%s",path,DIR_TERM,name);
        ret = tryexec(buf, argv);
        goto end;
    }
    ret = false;
end:
    for (i = 0; argv[i]; i++)
        tfree(argv[i]);
    tfree(argv);
    return (ret);
}


static void
add_dir(dir,docc,htab)

char *dir;
bool docc;
void **htab;
{
    DIR *wdir;
    struct direct *de;

    if (*htab == NULL)
        *htab = htab_init();

    if (!(wdir = opendir(dir))) return;
    while (de = readdir(wdir)) {
        if (!is_exec(de->d_name,dir))
            continue;
        htab_add(de->d_name,(void*)copy(dir),*htab);
        if (docc) {
            /* Add to completion hash table. */
            cp_addcomm(de->d_name,0L,0L,0L,0L);
        }
    }
    closedir(wdir);
}


/* Debugging. */

void
cp_hstat()
{
    htab_print(hashtabp,"i = %d, name = %s, path = %s\n");
    htab_print(hashtabc,"i = %d, name = %s, path = %s\n");
    htab_print(hashtabs,"i = %d, name = %s, path = %s\n");
}


#ifdef notdef
/* if above doesn't work, use this */

void
cp_rehash(pathlist, docc)

char *pathlist;
bool docc;
{ }


bool
cp_unixcom(wl)

wordlist *wl;
{
    char *s = wl_flatten(wl);

    if (system(s))
        return (false);
    else
        return (true);
}

#endif
