/*************************************************************************
 MFB graphics and miscellaneous library for protected mode MSDOS.
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

/* routines for digitizing pad interface.
 * In protected mode, the serial port interrupt handler and cursor drawing
 * routines, located in the TSR "pointer", are accessed via int 62.  In real
 * mode, the TSR is not used, and the interrupt handler and drawing routines
 * are included within the MFB library.  The protected mode programs had
 * problems servicing the hardware interupts from protected mode, therefor
 * a real mode handler seemed to be required.
 */

#include "mfb.h"
#include "mfbP.h"
#include <dos.h>

#ifdef __STDC__
static int pointer_on(void);
#else
static int pointer_on();
#endif


void
dp_start(irq,addr)

/* initialize pointer */
int addr, irq;
{
    union REGS r;

    /* presence test */
    if (!pointer_on()) {
        fprintf(stderr,"Error: \"pointer\" TSR not loaded.\n");
        fprintf(stderr,"Hit any key to exit.\n");
        (void)MFBGetchar();
        pc.pointertype = KBRD;
        MFBPointerInit();
        MFBClose();
        exit(0);
    }
    r.x.ax = 5;
    /* in real mode, pass irq,irq+8 */
#if __NDPC__
    /* Phar Lap */
    r.x.bx = (irq << 8) | get_hwvec(irq);
#else
    r.x.bx = (irq << 8) | irq + 8;
#endif
    r.x.cx = addr;
    r.x.dx = pc.xsize;
    r.x.di = pc.ysize;
    r.x.si = pc.mfbMODE;
    int86(0x62,&r,&r);
}


void
dp_end()

/* retire pointer */
{
    union REGS r;

    r.x.ax = 6;
    int86(0x62,&r,&r);
}


void
dp_pointer_on()

/* make marker visible */
{
    union REGS r;

    r.x.ax = 0;
    r.x.bx = pc.cursor_color;
    int86(0x62,&r,&r);
}


void
dp_pointer_off()

/* make pointer invisible */
{
    union REGS r;

    r.x.ax = 1;
    int86(0x62,&r,&r);
}


void
dp_status(x,y,b)

/* return coordinates, button press info */
int *x, *y, *b;
{
    union REGS r;

    r.x.ax = 4;
    int86(0x62,&r,&r);
    *x = r.x.bx;
    *y = r.x.cx;
    *b = r.x.dx;
}


static int
pointer_on()

/* return 1 if pointer installed on int 62, 0 otherwise */
{
#define NAME_OFF 18

#if __NDPC__
#define RP_OFF(rp) ((rp) & 0xFFFF)
#define RP_SEG(rp) ((rp) >> 16)
    unsigned long r;
    char c[8];

    blk_mb(&r,0x34,0,0x62*4,4);  /* getvect(0x62) */
    blk_mb(c,0x34,(RP_SEG(r) << 4) + RP_OFF(r) + NAME_OFF,8);
#else
#if __GNUC__
    char *c;
    
    c = (char*)(0xe0000000 + 0x62);
    c = (char*)(16*(*(short*)(c+2)) + *(short*)c + NAME_OFF);
#else
    return (I_cant_convert_real_address);
#endif
#endif

    if (c[0] != 'p') return (0);
    if (c[1] != 'o') return (0);
    if (c[2] != 'i') return (0);
    if (c[3] != 'n') return (0);
    if (c[4] != 't') return (0);
    if (c[5] != 'e') return (0);
    if (c[6] != 'r') return (0);
    return (1);
}


void
dp_point(x,y,k,b)
int *x, *y, *k, *b;

{
    int xx, yy, bb;
    int xlst, ylst;

    dp_status(&xx,&yy,&bb);
    yy = pc.ysize - yy - 1;
    MFBDrawCursor(xx-8,yy+7);
    for (;;) {
        if (kbhit()) {
            *k = 0xff & getch();
            if (!*k)
                *k = 256 + getch();
            *b = 0;
            break;
        }
        if (bb) {
            *k = 0;
            *b = bb;
            break;
        }
        xlst = xx;
        ylst = yy;
        dp_status(&xx,&yy,&bb);
        yy = pc.ysize - yy - 1;
        if (xx == xlst && yy == ylst)
            continue;
        MFBEraseCursor(xlst-8,ylst+7);
        MFBDrawCursor(xx-8,yy+7);
    }
    *x = xx;
    *y = yy;
    MFBEraseCursor(xx-8,yy+7);
}

#if __NDPC__

int
get_hwvec(num)

/* get relocated hardware interrupt vector */
/* Phar Lap call */
int num;
{
    union REGS r;

    r.x.ax = 0x250c;
    int86(0x21,&r,&r);
    if (num <= 7) return (num + (r.x.ax & 0xff));
    return (r.x.ax >> 8);
}

#endif
