/*************************************************************************
 MFB graphics and miscellaneous library
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include "mfb.h"
#include "mfbP.h"
#include <dos.h>

#define ror(x,n) ((x >> n)  | (x << (8-n)))

#ifdef __STDC__
static void mode_1_box(int,int,int,int);
static void mode_2_box(int,int,int,int);
#else
static void mode_1_box();
static void mode_2_box();
#endif


void
MFBBox(x1,y1,x2,y2)

int x1,y1,x2,y2;
{
    if (pc.mfbMODE & 2)
        mode_2_box(x1,y1,x2,y2);
    else
        mode_1_box(x1,y1,x2,y2);
}


static void
mode_1_box(xl,yl,xu,yu)

int xl, yl, xu, yu;
{
    int lnum, dx, tx, dy, next, *st = NULL;
    vidmptr rgen;
    unsigned char left, right, cbuf;
    union { unsigned short color2; unsigned char c[2]; } c;

    if (yu < yl) MFBSwapInt(yu,yl);
    if (xu < xl) MFBSwapInt(xu,xl);

    /* expand box so that it covers a box formed of lines with the
     * same coordinates
     */
    xu++;
    yl--;

    if (pc.curfillpatt)
        st = pc.stipples[pc.curfillpatt];

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];
    lnum = (pc.ysize-1-yu);
    dy = yu - yl;

    left  =  (0xff >> (xl & 7));
    right = ~(0xff >> (xu & 7));
    dx = (xu >> 3) - (xl >>= 3) - 1;
    if (dx < 0) { left &= right; dx = 0; right = 0; }
    rgen = pc.base + xl + (long) lnum*pc.bytpline;
    next = pc.bytpline - 1 - dx;

    outpw(0x3ce,0x0b05); /* read mode 1, write mode 3 */
    outpw(0x3ce,c.color2 & 0xff00); /* set/reset */
    outpw(0x3ce,0x7); /* zero color dont care */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    cbuf = 0xff;

    if (!st) {

        while (dy--) {
            mfb_trash = *rgen;
            *rgen = left;
            rgen++;
            tx = dx;
            while (tx--) {
                mfb_trash = *rgen;
                *rgen = cbuf;
                rgen++;
            }
            mfb_trash = *rgen;
            *rgen = right;
            rgen += next;
        }
    }
    else {
        while (dy--) {
            cbuf = st[lnum++ & 7];
            mfb_trash = *rgen;
            *rgen = left & cbuf;
            rgen++;
            tx = dx;
            while (tx--) {
                mfb_trash = *rgen;
                *rgen = cbuf;
                rgen++;
            }
            mfb_trash = *rgen;
            *rgen = right & cbuf;
            rgen += next;
        }
    }

    outpw(0x3ce,0x5);
    outpw(0x3ce,0xff07);
}


#ifndef __GNUC__


static void
mode_2_box(xl,yl,xu,yu)

int xl, yl, xu, yu;
{
    int lnum, dy, next, *st = NULL;
    unsigned short dx, tx;
    unsigned char left, right, cbuf;
    union { unsigned short o[2]; long l; } p1, p2;
    union { unsigned short color2; unsigned char c[2]; } c;
    vidmptr rgen;

    if (yu < yl) MFBSwapInt(yu,yl);
    if (xu < xl) MFBSwapInt(xu,xl);

    outpw(0x3ce,0xff08);     /* set bit mask */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    /* expand box so that it covers a box formed of lines with the
     * same coordinates
     */
    xu++;
    yl--;

    if (pc.curfillpatt)
        st = pc.stipples[pc.curfillpatt];

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];
    lnum = (pc.ysize-1-yu);
    dy = yu - yl;
    dx = xu - xl;
    p1.l = xl + (long) lnum*pc.xsize;
    p2.l = p1.l + dx;

    cbuf = inp(0x3cd);
    cbuf |= (cbuf << 4);
    next = pc.xsize;

    if (!st) {

        while (dy--) {
            if ((cbuf & 0xf) != p1.o[1]) {
                cbuf = p1.o[1];
                cbuf |= (cbuf << 4);
                outp(0x3cd,cbuf);
            }
            if (p1.o[1] != p2.o[1])
                tx = ~(p1.o[0]) + 1;
            else
                tx = dx;
            rgen = pc.base + p1.o[0];

            if (pc.alumode == 0x3) {
                if ((unsigned)rgen & 1) {
                    *rgen = c.c[0];
                    rgen++;
                    tx--;
                }
                if (tx & 1)
                    *(rgen + tx - 1) = c.c[0];
                tx >>= 1;
                while (tx--) {
                    *(short far *)rgen = c.color2;
                    rgen += 2;
                }
            }
            else {
                while (tx--) {
                    mfb_trash = *rgen;
                    *rgen = c.c[0];
                    rgen++;
                }
            }
            if (p1.o[1] != p2.o[1]) {
                cbuf += 0x11;
                outp(0x3cd,cbuf);
                tx = p2.o[0];
                rgen = pc.base;
                if (pc.alumode == 0x3) {
                    if (tx & 1)
                        *(rgen + tx - 1) = c.c[0];
                    tx >>= 1;
                    while (tx--) {
                        *(short far *)rgen = c.color2;
                        rgen += 2;
                    }
                }
                else {
                    while (tx--) {
                        mfb_trash = *rgen;
                        *rgen = c.c[0];
                        rgen++;
                    }
                }
            }
            p2.l += next;
            p1.l += next;
        }
    }

    else {

        while (dy--) {
            left = 0x80 >> (xl & 7);
            right = st[lnum++ & 7];

            if ((cbuf & 0xf) != p1.o[1]) {
                cbuf = p1.o[1];
                cbuf |= (cbuf << 4);
                outp(0x3cd,cbuf);
            }
            if (p1.o[1] != p2.o[1])
                tx = ~(p1.o[0]) + 1;
            else
                tx = dx;
            rgen = pc.base + p1.o[0];

            while (tx--) {
                if (left & right) {
                    if (pc.alumode != 3)
                        mfb_trash = *rgen;
                    *rgen = c.c[0];
                }
                left = ror(left,1);
                rgen++;
            }

            if (p1.o[1] != p2.o[1]) {
                cbuf += 0x11;
                outp(0x3cd,cbuf);
                tx = p2.o[0];
                rgen = pc.base;
                while (tx--) {
                    if (left & right) {
                        if (pc.alumode != 3)
                            mfb_trash = *rgen;
                        *rgen = c.c[0];
                    }
                    left = ror(left,1);
                    rgen++;
                }
            }
            p1.l += next;
            p2.l += next;
        }
    }
}


#else /* __GNUC__ */


static void
mode_2_box(xl,yl,xu,yu)

int xl, yl, xu, yu;
{
    int lnum, dx, tx, dy, next, *st = NULL;
    unsigned char left, right;
    union { unsigned short color2; unsigned char c[2]; } c;
    vidmptr rgen;

    if (yu < yl) MFBSwapInt(yu,yl);
    if (xu < xl) MFBSwapInt(xu,xl);

    outpw(0x3ce,0xff08);     /* set bit mask */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    /* expand box so that it covers a box formed of lines with the
     * same coordinates
     */
    xu++;
    yl--;

    if (pc.curfillpatt)
        st = pc.stipples[pc.curfillpatt];

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];
    lnum = (pc.ysize-1-yu);
    dy = yu - yl;
    dx = xu - xl;
    rgen = pc.base + xl + (long) lnum*pc.xsize;
    next = pc.xsize - dx;

    if (!st) {

        if (pc.alumode == 0x3) {

            while (dy--) {
                tx = dx;
                if ((unsigned)rgen & 1) {
                    *rgen = c.c[0];
                    rgen++;
                    tx--;
                }
                left = tx & 1;
                tx >>= 1;
                while (tx--) {
                    *(short*)rgen = c.color2;
                    rgen += 2;
                }
                if (left) {
                    *rgen = c.c[0];
                    rgen++;
                }
                rgen += next;
            }
        }
        else {

            while (dy--) {
                tx = dx;
                while (tx--) {
                    mfb_trash = *rgen;
                    *rgen = c.c[0];
                    rgen++;
                }
                rgen += next;
            }
        }
    }

    else {

        while (dy--) {
            left = 0x80 >> (xl & 7);
            right = st[lnum++ & 7];

            tx = dx;
            while (tx--) {
                if (left & right) {
                    if (pc.alumode != 3)
                        mfb_trash = *rgen;
                    *rgen = c.c[0];
                }
                left = ror(left,1);
                rgen++;
            }
            rgen += next;
        }
    }
}

#endif /* __GNUC__ */
