/*************************************************************************
 MFB graphics and miscellaneous library
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include "mfb.h"
#include "mfbP.h"
#include <ctype.h>
#include <time.h>
#if __GNUC__
#include <limits.h> /* for CLK_TCK */
#endif

/* Undefine below to use the function key selection feature */
/*
#define SELECT
*/

#ifdef __STDC__
static int  kb_point(int*,int*,int*,int*);
static void newmark(int,int,int,int);
static int  mouse_point(int*,int*,int*,int*);
static void get_config(int*,int*,int*);
static void dummy(void);
static void (*pointer_status)(int*,int*,int*);
static void (*pointer_on)(void);
static void (*pointer_off)(void);
static void (*pointer_close)(void);
#else
static int  kb_point();
static void newmark();
static int  mouse_point();
static void get_config();
static void dummy();
static void (*pointer_status)();
static void (*pointer_on)();
static void (*pointer_off)();
static void (*pointer_close)();
#endif

int functkey_dx;      /* should move this to pc structure */


int
MFBPoint(x,y,key,button)

int *x, *y, *key, *button;
{
    switch (pc.pointertype) {
        case MSMOUSE:
        case MMTAB:
            return (mouse_point(x,y,key,button));
        case KBRD:
            return (kb_point(x,y,key,button));
    }
    return (0);
}


static int
kb_point(x,y,key,button)

/* pointer routine for keyboard arrow keys */
int *x, *y, *key, *button;
{
    int k;
    int xo = 8, yo = 7, x2o, y2o;
    int xs = pc.xsize - 1;
    int ys = pc.ysize - 1;

    MFBDrawCursor(pc.cursx,pc.cursy);

    if (functkey_dx == 0)
        functkey_dx = pc.xsize/32;

    x2o = xo << 1;
    y2o = yo << 1;

    *key = '\0';

    for (;;) {
        k = MFBGetchar();
        switch (k) {

        /* arrow keys */
        case 333: /* right */
            if (pc.cursx < xs - pc.curstep - xo) {
                newmark(pc.cursx,pc.cursy,pc.cursx + pc.curstep,pc.cursy);
            }
            break;
        case 331: /* left */
            if (pc.cursx > pc.curstep - xo) {
                newmark(pc.cursx,pc.cursy,pc.cursx - pc.curstep,pc.cursy);
            }
            break;
        case 328: /* up */
            if (pc.cursy < ys - pc.curstep + yo) {
                newmark(pc.cursx,pc.cursy,pc.cursx,pc.cursy + pc.curstep);
            }
            break;
        case 336: /* down */
            if (pc.cursy > pc.curstep + yo) {
                newmark(pc.cursx,pc.cursy,pc.cursx,pc.cursy - pc.curstep);
            }
            break;
        case 372: /* control right */
            if (pc.cursx < xs - xo) {
                newmark(pc.cursx,pc.cursy,pc.cursx+1,pc.cursy);
            }
            break;
        case 371: /* control left */                   
            if (pc.cursx > -xo) {
                newmark(pc.cursx,pc.cursy,pc.cursx-1,pc.cursy);
            }
            break;
        case 397: /* control up */
            if (pc.cursy < ys + yo) {
                newmark(pc.cursx,pc.cursy,pc.cursx,pc.cursy+1);
            }
            break;
        case 401: /* control down */
            if (pc.cursy > yo) {
                newmark(pc.cursx,pc.cursy,pc.cursx,pc.cursy-1);
            }
            break;
        case 413: /* alt right */
            if (pc.cursx < xs - pc.curleap - xo) {
                newmark(pc.cursx,pc.cursy,pc.cursx + pc.curleap,pc.cursy);
            }
            break;
        case 411: /* alt left */
            if (pc.cursx > pc.curleap - xo) {
                newmark(pc.cursx,pc.cursy,pc.cursx - pc.curleap,pc.cursy);
            }
            break;
        case 408: /* alt up */
            if (pc.cursy < ys - pc.curleap + yo) {
                newmark(pc.cursx,pc.cursy,pc.cursx,pc.cursy + pc.curleap);
            }
            break;
        case 416: /* alt down */
            if (pc.cursy > pc.curleap + yo) {
                newmark(pc.cursx,pc.cursy,pc.cursx,pc.cursy - pc.curleap);
            }
            break;

    /* function keys */
#ifdef SELECT
        case 315: /* F1 */
            *x = functkey_dx;
            goto select;
        case 316: /* F2 */
            *x = 3*functkey_dx;
            goto select;
        case 317: /* F3 */
            *x = 5*functkey_dx;
            goto select;
        case 318: /* F4 */
            *x = 7*functkey_dx;
            goto select;
        case 319: /* F5 */
            *x = 9*functkey_dx;
            goto select;
        case 320: /* F6 */
            *x = 11*functkey_dx;
            goto select;
        case 321: /* F7 */
            *x = 13*functkey_dx;
            goto select;
        case 322: /* F8 */
            *x = 15*functkey_dx;
            goto select;
        case 323: /* F9 */
            *x = 17*functkey_dx;
            goto select;
        case 324: /* F10 */
            *x = 19*functkey_dx;
            goto select;
        case 389: /* F11 */
            *x = 21*functkey_dx;
            goto select;
        case 390: /* F12 */
            *x = 23*functkey_dx;
            goto select;
#endif

    /* page keys */
            case 327: /* HOME */
                newmark(pc.cursx,pc.cursy,1,ys-1);
                break;
        case 335: /* END */
                newmark(pc.cursx,pc.cursy,xs-x2o-1,y2o+1);
                break;
        case 329: /* PgUp */
                newmark(pc.cursx,pc.cursy,pc.cursx,ys-1);
                break;
        case 337: /* PgDn */
                newmark(pc.cursx,pc.cursy,pc.cursx,y2o+1);
                break;
        case 338: /* INS */
                newmark(pc.cursx,pc.cursy,xs/2 - xo,ys/2 + yo);
                break;


        case 13: /* CR */ /* yellow button */
            *x = pc.cursx + xo;
            *y = pc.cursy - yo;
            *button = 1;
            goto terminate;
        case 10: /* LF */ /* white button */
            *x = pc.cursx + xo;
            *y = pc.cursy - yo;
            *button = 2;
            goto terminate;
        case 339: /*DEL */ /* green button */
            *x = pc.cursx + xo;
            *y = pc.cursy - yo;
            *button = 8;
            goto terminate;
        case 284: /* alt ENTER */ /* blue button */
            *x = pc.cursx + xo;
            *y = pc.cursy - yo;
            *button = 4;
            goto terminate;
        default:
            *key = k;
            *button = 0;
            *x = pc.cursx + xo;
            *y = pc.cursy - yo;
            goto terminate;
        }
    }

#ifdef SELECT
select:
    *y = pc.functkey_yband;
    *button = 1;
#endif

terminate:
    MFBEraseCursor(pc.cursx,pc.cursy);
    return (MFBOK);
}


static void
newmark(xold,yold,x,y)

int xold,yold,x,y;
{
    MFBEraseCursor(xold,yold);
    pc.cursx = x;
    pc.cursy = y;
    MFBDrawCursor(x,y);
}


/*
                       ` ^ ^ '
   3-button mouse      *     *     returns:  (l)     1
                     *         *             (c)     4
                      (l)(c)(r)              (r)     2
                     *         *             (l+c)   5
                       *     *               (r+c)   6
                          *                  (l+r)   3
                          |_____             (l+c+r) 7
                                |



   digitizing pad:         (f)    returns:   (f)     1
                                             (r)     2
                                             (f+r)   3
                       (l)     (r)           (l)     4
                                             (b)     5 
                           (b)               (f+l)   5
                                             (l+r)   6
                                             etc.
*/


static int
mouse_point(x,y,key,button)

/* pointer routine for Microsoft mouse or equivalent */
int *x, *y, *key, *button;
{
    int b = 0;
    int k;
    int xlast, ylast;

    *key = '\0';

    do {
        ms_status(&pc.cursx,&pc.cursy,&b);
    } while (b);
    ms_point(&pc.cursx,&pc.cursy,&k,&b);

    switch (b & 7) {

        case 0:
#ifdef SELECT
            if (functkey_dx == 0)
                functkey_dx = pc.xsize/32;

            switch (k) {

                /* function keys */

                case 315: /* F1 */
                    *x = functkey_dx;
                    goto select;

                case 316: /* F2 */
                    *x = 3*functkey_dx;
                    goto select;

                case 317: /* F3 */
                    *x = 5*functkey_dx;
                    goto select;

                case 318: /* F4 */
                    *x = 7*functkey_dx;
                    goto select;

                case 319: /* F5 */
                    *x = 9*functkey_dx;
                    goto select;

                case 320: /* F6 */
                    *x = 11*functkey_dx;
                    goto select;

                case 321: /* F7 */
                    *x = 13*functkey_dx;
                    goto select;

                case 322: /* F8 */
                    *x = 15*functkey_dx;
                    goto select;

                case 323: /* F9 */
                    *x = 17*functkey_dx;
                    goto select;

                case 324: /* F10 */
                    *x = 19*functkey_dx;
                    goto select;

                case 389: /* F11 */
                    *x = 21*functkey_dx;
                    goto select;

                case 390: /* F12 */
                    *x = 23*functkey_dx;
                    goto select;

                /* arrow keys */
                /* case 333: right */
                /* case 331: left */
                case 328: /* up */
                case 336: /* down */
                    *x = pc.xsize - functkey_dx;
                    goto select;

                case 13: /* CR */ /* yellow button */
                    *button = 1;
                    goto terminate;

                case 10: /* LF */ /* white button */
                    *button = 2;
                    goto terminate;

                case 339: /*DEL */ /* green button */
                    *button = 3;
                    goto terminate;

                case 284: /* alt ENTER */ /* blue button */
                    *button = 4;
                    goto terminate;

                default:
                    *key = k;
                    *button = 0;
                    goto terminate;
            }
#else /*SELECT */
            *key = k;
            *button = 0;
            goto terminate;
#endif

        case 1:    /* center, yellow */
            *button = 1;
            goto terminate;

        case 2:
            /* right, blue */
            *button = 4;
            goto terminate;

        case 3: /* center and right */
            *button = 1;
            goto terminate;

        case 4:    /* left, white */
            *button = 2;
            goto terminate;

        case 5: /* left and center, or lower center, green */
            *button = 8;
            goto terminate;

        case 6: /* rignt and left */
            *button = 1;
            goto terminate;

        case 7:
            *button = 1;
            goto terminate;
    }

#ifdef SELECT
select:
    *y = pc.functkey_yband;
    if (*x > pc.xsize-functkey_dx) *x = pc.xsize-functkey_dx;
    *button = 1;
    return (MFBOK);
#endif

terminate:
    *x = pc.cursx;
    *y = pc.cursy;
    return (MFBOK);
}


void
MFBPointerInit()

{
    int irq = 4, serport = 0x3f8;  /* default to com1 */
    int type;

    /* Pointer type can be coerced by setting pc.pointertype
     * before calling this routine.
     */
    type = pc.pointertype;
    get_config(&irq,&serport,&type);

    if (!pc.pointertype) pc.pointertype = type;
    switch (pc.pointertype) {
        case MSMOUSE:
            ms_start();
            pointer_status = ms_status;
            pointer_on = ms_pointer_on;
            pointer_off = ms_pointer_off;
            pointer_close = ms_end;
            break;

#ifdef DPAD_SUPPORT
        case MMTAB:
            dp_start(irq,serport);
            pointer_status = dp_status;
            pointer_on = dp_pointer_on;
            pointer_off = dp_pointer_off;
            pointer_close = dp_end;
            break;
#endif

        default:
        case KBRD:
            pointer_close = dummy;
            break;
    }
}


void
MFBPointerClose()

{
    pointer_close();
}


static void
get_config(irq,adrs,pointer)

int *irq, *adrs, *pointer;
{
    FILE *fp;
    char s[128], *getenv();
    int i = 0;
    static int savirq,savadrs,savtype;

    if (savtype) {
        *irq = savirq;
        *adrs = savadrs;
        *pointer = savtype;
        return;
    }

    fp = fopen("mfb.rc","r");
    if (!fp) {
        strcpy(s,MFBRCD);
        strcat(s,"\\mfb.rc");
        fp = fopen(s,"r");
    }
    if (!fp) {
        /* No mfb.rc file found.  Is there a mouse driver? */
        if (ms_hreset())
            *pointer = MSMOUSE;
        else
            *pointer = KBRD;
        savtype = *pointer;
        return;
    }
    while (fgets(s,128,fp) != NULL) {
        if (*s == '#') continue;
        switch (i) {
            case 0:
                sscanf(s,"%d",irq);
                break;
            case 1:
                sscanf(s,"%x",adrs);
                break;
            case 2:
                if (!strncmp(s,"pad",3))
                    *pointer = MMTAB;
                else if (!strncmp(s,"mouse",5))
                    *pointer = MSMOUSE;
                else if (!strncmp(s,"keys",4))
                    *pointer = KBRD;
                break;
            default:
                break;
        }
        i++;
    }
    fclose(fp);
    savirq = *irq;
    savadrs = *adrs;
    savtype = *pointer;
}


static void
dummy()    {}
