/*************************************************************************
 MFB graphics and miscellaneous library
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include "mfb.h"
#if defined(MSDOS) || __MSDOS__
#include "mfbP.h"
#endif


/* Line clipping routines.
 * Note that these routines operate with long integers */


#define CODELEFT 1
#define CODEBOTTOM 2
#define CODERIGHT 4
#define CODETOP 8
#define CODE(x,y,c)  c = 0;\
                     if(x < l)\
                         c = CODELEFT;\
                     else if(x > r)\
                         c = CODERIGHT;\
                     if(y < b)\
                         c |= CODEBOTTOM;\
                     else if(y > t)\
                         c |= CODETOP;

void
MFB_Y_Intercept(x1,y1,x2,y2,e,yi)

long x1,y1,x2,y2;                /* two points on line */
long e;                          /* vertical line of intercept */
long *yi;                        /* y coordinate of intercept */
{
    /*
     * MFB_Y_Intercept will return the value 'yi' where the the coordinate
     * (e,yi) is the intersection of the vertical line x = e and the line
     * determined by the coordinates (x1,y1) and (x2,y2).
     */
    *yi = y1;
    if (x1 == x2) return;            /* vertical line */
    *yi = y1 + ((e - x1) * (y2 - y1))/(x2 - x1);
}


void
MFB_X_Intercept(x1,y1,x2,y2,e,xi)

long x1,y1,x2,y2;                /* two polongs on line */
long e;                          /* horizontal line of longercept */
long *xi;                        /* x coordinate of longercept */
{
    /*
     * MFB_X_Intercept will return the value 'xi' where the the coordinate
     * (xi,e) is the intersection of the horizontal line y = e and the line
     * determined by the coordinates (x1,y1) and (x2,y2).
     */
    *xi = x1;
    if (y1 == y2) return;            /* horizontal line */
    *xi = x1 + ((e - y1) * (x2 - x1))/(y2 - y1);
}


Bool
MFBLineClip(pX1,pY1,pX2,pY2,l,b,r,t)

long *pX1,*pY1,*pX2,*pY2,l,b,r,t;
{
    /*
     * MFBLineClip will clip a line to a rectangular area.  The returned
     * value is 'true' if the line is out of the AOI (therefore does not
     * need to be displayed) and 'false' if the line is in the AOI.
     */
    long x1 = *pX1;
    long y1 = *pY1;
    long x2 = *pX2;
    long y2 = *pY2;
    long x,y;
    int  c,c1,c2;

    CODE(x1,y1,c1)
    CODE(x2,y2,c2)
    while (c1 != 0 || c2 != 0) {
        if ((c1 & c2) != 0) return (true); /* Line is invisible. */
        if ((c = c1) == 0) c = c2;
        if (c & CODELEFT) {
            y = y1+(y2-y1)*(l-x1)/(x2-x1);
            x = l;
        }
        else if (c & CODERIGHT) {
            y = y1+(y2-y1)*(r-x1)/(x2-x1);
            x = r;
        }
        else if (c & CODEBOTTOM) {
            x = x1+(x2-x1)*(b-y1)/(y2-y1);
            y = b;
        }
        else if (c & CODETOP) {
            x = x1+(x2-x1)*(t-y1)/(y2-y1);
            y = t;
        }
        if (c == c1) {
            x1 = x; y1 = y; CODE(x,y,c1)
        }
        else {
            x2 = x; y2 = y; CODE(x,y,c2)
        }
    }
    *pX1 = x1; *pY1 = y1;
    *pX2 = x2; *pY2 = y2;
    return (false); /* Line is at least partially visible. */
}
