/*************************************************************************
 MFB graphics and miscellaneous library
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include "mfb.h"
#include "mfbP.h"
#include <dos.h>

#define ror(x,n) ((x >> n)  | (x << (8-n)))

#ifdef __STDC__
static void mode_1_line(int,int);
static void mode_2_line(int,int);
#else
static void mode_1_line();
static void mode_2_line();
#endif


void
MFBLine(x0,y0,x1,y1)

int x0, y0, x1, y1;
{
    pc.curx = x0;
    pc.cury = y0;
    if (pc.mfbMODE & 2)
        mode_2_line(x1,y1);
    else
        mode_1_line(x1,y1);
}


void
MFBDrawLineTo(x,y)

int x, y;
{
    if (pc.mfbMODE & 2)
        mode_2_line(x,y);
    else
        mode_1_line(x,y);
}


void
MFBMoveTo(x,y)

int x, y;
{
    pc.curx = x;
    pc.cury = y;
}


int
MFBSetLineStyle(style)

int style;
{
    if ((style < 0) || (style >= NSTYLES))
        return (MFBBADLST);
    pc.curstyle = style;
    return (MFBOK);
}


int
MFBDefineLineStyle(num, mask)

int num, mask;
{
    if ((num < 0) || (num >= NSTYLES))
        return (MFBBADLST);
    if (num == 0) return (MFBOK);
    pc.linestyles[num] = (unsigned char) mask;
    pc.curstyle = num;
    return (MFBOK);
}


static void
mode_1_line(x,y)
int x, y;
{
    int xl, yl, dx, dy, dy2, errterm = 0, next, lcnt;
    vidmptr rgen;
    unsigned char cbuf, left, right;
    union { unsigned short color2; unsigned char c[2]; } c;

    xl = pc.curx;
    yl = pc.cury;
    pc.curx = x;
    pc.cury = y;

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];

    outpw(0x3ce,0x0b05); /* read mode 1, write mode 3 */
    outpw(0x3ce,c.color2 & 0xff00); /* set/reset */
    outpw(0x3ce,0x7); /* zero color dont care */
    outpw(0x3ce,pc.alumode); /* set alu mode */
        
    if (y == yl) {
        if (x < xl) MFBSwapInt(x,xl);
        x++;
        left  =  (0xff >> (xl & 7));
        right = ~(0xff >> (x & 7));
        dx    = (x >> 3) - (xl >>= 3) - 1;
        if (dx < 0) { left &= right; dx = 0; right = 0; }
        rgen = pc.base + xl + (long) (pc.ysize-1-y)*pc.bytpline;

        cbuf = 0xff;
        if (pc.curstyle) {
            cbuf = pc.linestyles[pc.curstyle];
            left &= cbuf;
            right &= cbuf;
        }
        mfb_trash = *rgen;
        *rgen = left;
        rgen++;
        while (dx--) {
            mfb_trash = *rgen;
            *rgen = cbuf;
            rgen++;
        }
        mfb_trash = *rgen;
        *rgen = right;
        
        goto done;
    }

    if (x == xl) {

        if (y < yl) MFBSwapInt(y,yl);
        yl--;
        dy = y - yl;

        cbuf = 0x80 >> (xl & 7);

        next = pc.bytpline;

        if (!pc.curstyle) {
            rgen = pc.base + (x >> 3) + (long) (pc.ysize-1-y)*next;
            while (dy--) {
                mfb_trash = *rgen;
                *rgen = cbuf;
                rgen += next;
            }
        }
        else {
            lcnt = pc.ysize-1-y;
            rgen = pc.base + (x >> 3) + (long) lcnt*next;
            left = 0x80 >> ((lcnt+xl) & 7);
            right = pc.linestyles[pc.curstyle];

            while (dy--) {
                if (left & right) {
                    mfb_trash = *rgen;
                    *rgen = cbuf;
                }
                left = ror(left,1);
                rgen += next;
            }
        }
        goto done;
    }

    if (x < xl) {
        MFBSwapInt(xl,x);
        MFBSwapInt(yl,y);
    }
    dx = x - xl;

    next = pc.bytpline;
    dy = yl - y;

    if (!pc.curstyle) {
        rgen = pc.base + (xl >> 3) + (long) (pc.ysize-1-yl)*next;
        if (yl < y) {
            next = -next;
            dy = -dy;
        }

        cbuf = 0x80 >> (xl & 7);

        dy2 = dy;
        dy++;
        while (dy--) {
            errterm += dx;
            if (errterm <= 0) {
                mfb_trash = *rgen;
                *rgen = cbuf;
                rgen += next;
                continue;
            }
            while (errterm > 0 && xl <= x) {
                mfb_trash = *rgen;
                *rgen = cbuf;
                cbuf = ror(cbuf,1);
                if (cbuf & 0x80) rgen++;
                xl++;
                errterm -= dy2;
            }
            rgen += next;
        }
    }
    else {
        lcnt = pc.ysize-1-yl;
        rgen = pc.base + (xl >> 3) + (long) lcnt*next;
        left = 0x80 >> ((lcnt+xl) & 7);
        right = pc.linestyles[pc.curstyle];

        if (yl < y) {
            next = -next;
            dy = -dy;
        }

        cbuf = 0x80 >> (xl & 7);

        dy2 = dy;
        dy++;
        while (dy--) {
            errterm += dx;
            if (errterm <= 0) {
                if (left & right) {
                    mfb_trash = *rgen;
                    *rgen = cbuf;
                }
                rgen += next;
                left = ror(left,1);
                continue;
            }
            while (errterm > 0 && xl <= x) {
                if (left & right) {
                    mfb_trash = *rgen;
                    *rgen = cbuf;
                }
                left = ror(left,1);
                cbuf = ror(cbuf,1);
                if (cbuf & 0x80) rgen++;
                xl++;
                errterm -= dy2;
            }
            rgen += next;
            left = ror(left,1);
        }
    }
done:
    outpw(0x3ce,0xff07);
    outpw(0x3ce,0x5);
}


#ifndef __GNUC__


static void
mode_2_line(x,y)
int x, y;
{
    int xl, yl, dy, dy2, errterm = 0, next, lcnt;
    unsigned short dx;
    unsigned char cbuf, left, right;
    union { unsigned short o[2]; long l; } p1, p2;
    union { unsigned short color2; unsigned char c[2]; } c;
    vidmptr rgen;

    xl = pc.curx;
    yl = pc.cury;
    pc.curx = x;
    pc.cury = y;

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];

    outpw(0x3ce,0xff08);     /* set bit mask */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    if (y == yl) {
        if (x < xl) MFBSwapInt(x,xl);
        x++;
        dx = x - xl;
        p1.l = xl + (long) (pc.ysize-1-y)*pc.xsize;
        p2.l = p1.l + dx;
        if (p1.o[1] != p2.o[1])
            dx = ~(p1.o[0]) + 1;
        cbuf = p1.o[1];
        cbuf |= (cbuf << 4);
        outp(0x3cd,cbuf);
        rgen = pc.base + p1.o[0];

        if (!pc.curstyle) {

            if (pc.alumode == 0x3) {
                if ((unsigned)rgen & 1) {
                    *rgen = c.c[0];
                    rgen++;
                    dx--;
                }
                if (dx & 1)
                    *(rgen + dx - 1) = c.c[0];
                dx >>= 1;
                while (dx--) {
                    *(short far *)rgen = c.color2;
                    rgen += 2;
                }
            }
            else {
                while (dx--) {
                    mfb_trash = *rgen;
                    *rgen = c.c[0];
                    rgen++;
                }
            }
            
            if (p1.o[1] != p2.o[1]) {
                outp(0x3cd,cbuf + 0x11);
                dx = p2.o[0];
                rgen = pc.base;
                if (pc.alumode == 0x3) {
                    if (dx & 1)
                        *(rgen + dx - 1) = c.c[0];
                    dx >>= 1;
                    while (dx--) {
                        *(short far *)rgen = c.color2;
                        rgen += 2;
                    }
                }
                else {
                    while (dx--) {
                        mfb_trash = *rgen;
                        *rgen = c.c[0];
                        rgen++;
                    }
                }
            }
        }
        else {

            lcnt = pc.ysize-1-y;
            p1.l = xl + (long) lcnt*pc.xsize;
            left = 0x80 >> ((lcnt+xl) & 7);
            right = pc.linestyles[pc.curstyle];

            while (dx--) {
                if (left & right) {
                    if (pc.alumode != 3)
                        mfb_trash = *rgen;
                    *rgen = c.c[0];
                }
                left = ror(left,1);
                rgen++;
            }

            if (p1.o[1] != p2.o[1]) {
                outp(0x3cd,cbuf + 0x11);
                dx = p2.o[0];
                rgen = pc.base;
                while (dx--) {
                    if (left & right) {
                        if (pc.alumode != 3)
                            mfb_trash = *rgen;
                        *rgen = c.c[0];
                    }
                    left = ror(left,1);
                    rgen++;
                }
            }
        }
        return;
    }


    if (x == xl) {
        if (y < yl) MFBSwapInt(y,yl);
        yl--;
        dy = y - yl;
        next = pc.xsize;
        lcnt = pc.ysize-1-y;
        p1.l = xl + (long) lcnt*next;
        cbuf = p1.o[1];
        cbuf |= (cbuf << 4);
        outp(0x3cd,cbuf);

        if (!pc.curstyle) {
            while (dy--) {
                rgen = pc.base + p1.o[0];
                if (pc.alumode != 3)
                    mfb_trash = *rgen;
                *rgen = c.c[0];
                p1.l += next;
                if ((cbuf & 0xf) != p1.o[1]) {
                    cbuf += 0x11;
                    outp(0x3cd,cbuf);
                }
            }
        }
        else {
            left = 0x80 >> ((lcnt+xl) & 7);
            right = pc.linestyles[pc.curstyle];

            while (dy--) {
                rgen = pc.base + p1.o[0];
                if (left & right) {
                    if (pc.alumode != 3)
                        mfb_trash = *rgen;
                    *rgen = c.c[0];
                }
                left = ror(left,1);
                p1.l += next;
                if ((cbuf & 0xf) != p1.o[1]) {
                    cbuf += 0x11;
                    outp(0x3cd,cbuf);
                }
            }
        }
        return;
    }

    cbuf = inp(0x3cd);
    cbuf |= (cbuf << 4);

    if (x < xl) {
        MFBSwapInt(xl,x);
        MFBSwapInt(yl,y);
    }
    dx = x - xl;

    next = pc.xsize;
    dy = yl - y;

    if (!pc.curstyle) {
        p1.l = xl + (long) (pc.ysize-1-yl)*next;

        if (yl < y) {
            next = -next;
            dy = -dy;
        }

        dy2 = dy;
        dy++;
        while (dy--) {
            errterm += dx;
            if (errterm <= 0) {
                if ((cbuf & 0xf) != p1.o[1]) {
                    cbuf = p1.o[1];
                    cbuf |= (cbuf << 4);
                    outp(0x3cd,cbuf);
                }
                if (pc.alumode != 3)
                    mfb_trash = *(pc.base + p1.o[0]);
                *(pc.base + p1.o[0]) = c.c[0];
                p1.l += next;
                continue;
            }
            while (errterm > 0 && xl <= x) {
                if ((cbuf & 0xf) != p1.o[1]) {
                    cbuf = p1.o[1];
                    cbuf |= (cbuf << 4);
                    outp(0x3cd,cbuf);
                }
                if (pc.alumode != 3)
                    mfb_trash = *(pc.base + p1.o[0]);
                *(pc.base + p1.o[0]) = c.c[0];
                p1.l++;
                xl++;
                errterm -= dy2;
            }
            p1.l += next;
    
        }
    }
    else {
        lcnt = pc.ysize-1-yl;
        p1.l = xl + (long) lcnt*next;
        left = 0x80 >> ((lcnt+xl) & 7);
        right = pc.linestyles[pc.curstyle];

        if (yl < y) {
            next = -next;
            dy = -dy;
        }

        dy2 = dy;
        dy++;
        while (dy--) {

            errterm += dx;
            if (errterm <= 0) {
                if (left & right) {
                    if ((cbuf & 0xf) != p1.o[1]) {
                        cbuf = p1.o[1];
                        cbuf |= (cbuf << 4);
                        outp(0x3cd,cbuf);
                    }
                    if (pc.alumode != 3)
                        mfb_trash = *(pc.base + p1.o[0]);
                    *(pc.base + p1.o[0]) = c.color2;
                }
                p1.l += next;
                left = ror(left,1);
                continue;
            }
            while (errterm > 0 && xl <= x) {
                if (left & right) {
                    if ((cbuf & 0xf) != p1.o[1]) {
                        cbuf = p1.o[1];
                        cbuf |= (cbuf << 4);
                        outp(0x3cd,cbuf);
                    }
                    if (pc.alumode != 3)
                        mfb_trash = *(pc.base + p1.o[0]);
                    *(pc.base + p1.o[0]) = c.color2;
                }
                left = ror(left,1);
                p1.l++;
                xl++;
                errterm -= dy2;
            }
            p1.l += next;
            left = ror(left,1);
    
        }
    }
}


#else /* __GNUC__ */


static void
mode_2_line(x,y)
int x, y;
{
    int xl, yl, dx, dy, dy2, errterm = 0, next, lcnt;
    unsigned left, right;
    union { unsigned short color2; unsigned char c[2]; } c;
    vidmptr rgen;

    xl = pc.curx;
    yl = pc.cury;
    pc.curx = x;
    pc.cury = y;

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];

    outpw(0x3ce,0xff08);     /* set bit mask */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    if (y == yl) {
        if (x < xl) MFBSwapInt(x,xl);
        x++;
        dx = x - xl;
        lcnt = pc.ysize-1-y;
        rgen = pc.base + xl + (long) lcnt*pc.xsize;

        if (!pc.curstyle) {

            if (pc.alumode == 0x3) {
                if ((unsigned)rgen & 1) {
                    *rgen = c.c[0];
                    rgen++;
                    dx--;
                }
                if (dx & 1)
                    *(rgen + dx - 1) = c.c[0];
                dx >>= 1;
                while (dx--) {
                    *(short*)rgen = c.color2;
                    rgen += 2;
                }
            }
            else {
                while (dx--) {
                    mfb_trash = *rgen;
                    *rgen = c.c[0];
                    rgen++;
                }
            }
        }
        else {

            left = 0x80 >> ((lcnt+xl) & 7);
            right = pc.linestyles[pc.curstyle];

            while (dx--) {
                if (left & right) {
                    if (pc.alumode != 3)
                        mfb_trash = *rgen;
                    *rgen = c.c[0];
                }
                left = ror(left,1);
                rgen++;
            }
        }
        return;
    }


    if (x == xl) {
        if (y < yl) MFBSwapInt(y,yl);
        yl--;
        dy = y - yl;
        next = pc.xsize;
        lcnt = pc.ysize-1-y;
        rgen = pc.base + xl + (long) lcnt*next;

        if (!pc.curstyle) {
            while (dy--) {
                if (pc.alumode != 3)
                    mfb_trash = *rgen;
                *rgen = c.c[0];
                rgen += next;
            }
        }
        else {
            left = 0x80 >> ((lcnt+xl) & 7);
            right = pc.linestyles[pc.curstyle];

            while (dy--) {
                if (left & right) {
                    if (pc.alumode != 3)
                        mfb_trash = *rgen;
                    *rgen = c.c[0];
                }
                left = ror(left,1);
                rgen += next;
            }
        }
        return;
    }

    if (x < xl) {
        MFBSwapInt(xl,x);
        MFBSwapInt(yl,y);
    }
    dx = x - xl;

    next = pc.xsize;
    dy = yl - y;
    lcnt = pc.ysize-1-yl;
    rgen = pc.base + xl + (long) lcnt*next;

    if (!pc.curstyle) {

        if (yl < y) {
            next = -next;
            dy = -dy;
        }

        dy2 = dy;
        dy++;
        while (dy--) {
            errterm += dx;
            if (errterm <= 0) {
                if (pc.alumode != 3)
                    mfb_trash = *rgen;
                *rgen = c.c[0];
                rgen += next;
                continue;
            }
            while (errterm > 0 && xl <= x) {
                if (pc.alumode != 3)
                    mfb_trash = *rgen;
                *rgen = c.c[0];
                rgen++;
                xl++;
                errterm -= dy2;
            }
            rgen += next;
    
        }
    }
    else {
        left = 0x80 >> ((lcnt+xl) & 7);
        right = pc.linestyles[pc.curstyle];

        if (yl < y) {
            next = -next;
            dy = -dy;
        }

        dy2 = dy;
        dy++;
        while (dy--) {

            errterm += dx;
            if (errterm <= 0) {
                if (left & right) {
                    if (pc.alumode != 3)
                        mfb_trash = *rgen;
                    *rgen = c.color2;
                }
                rgen += next;
                left = ror(left,1);
                continue;
            }
            while (errterm > 0 && xl <= x) {
                if (left & right) {
                    if (pc.alumode != 3)
                        mfb_trash = *rgen;
                    *rgen = c.color2;
                }
                left = ror(left,1);
                rgen++;
                xl++;
                errterm -= dy2;
            }
            rgen += next;
            left = ror(left,1);
    
        }
    }
}

#endif /* __GNUC__ */
