/*************************************************************************
 MFB graphics and miscellaneous library
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include "mfb.h"
#include "mfbP.h"
#include <dos.h>

/* in real mode, the mode_?_... functions are used in the dp driver */
#ifdef __STDC__
extern void mode_1_mark(int,int);
extern void mode_2_mark(int,int);
extern void mode_1_clear(int,int);
extern void mode_2_clear(int,int);
#else
extern void mode_1_mark();
extern void mode_2_mark();
extern void mode_1_clear();
extern void mode_2_clear();
#endif


void
MFBDrawCursor(x,y)

/* draw cursor */
int x,y;
{
    int oldalu;

    if (pc.drawghost) {
        /* draw any rubber-banding */
        oldalu = pc.alumode;
        pc.alumode = 0x1803;
        (*pc.drawghost)(x+8,y-7,pc.refx,pc.refy);
        pc.alumode = oldalu;
    }
    if (pc.mfbMODE & 2)
        mode_2_mark(x,y);
    else
        mode_1_mark(x,y);
}


void
MFBEraseCursor(x,y)

/* erase cursor */
int x,y;
{
    int oldalu;

    if (pc.mfbMODE & 2)
        mode_2_clear(x,y);
    else
        mode_1_clear(x,y);
    if (pc.drawghost) {
        /* undraw any rubber-banding */
        oldalu = pc.alumode;
        pc.alumode = 0x1803;
        (*pc.drawghost)(x+8,y-7,pc.refx,pc.refy);
        pc.alumode = oldalu;
    }
}


void
mode_1_mark(x,y)

/* display cursor 16 color modes */
int x, y;
{
    int i, j, next;
    unsigned short *amap, *xmap;
    unsigned char bitmask, mode, datarot, esetreset, cdontcare;
    unsigned char readplane, planemask, rem, *px;
    vidmptr rgen, rgen0;
    union {long l; unsigned short i; char s[4]; } afld, xfld;
    int stop;

    outp(0x3ce,0x5);        /* read mode 0, write mode 0 */
    mode = inp(0x3cf);
    outp(0x3cf,0);

    outp(0x3ce,0x4);        /* read plane */
    readplane = inp(0x3cf);

    amap = (unsigned short *) vgacur.map;
    xmap = amap + 16;

    px = vgacur.pbs;

    rem = x & 7;
    x >>= 3;
    next = pc.bytpline;
    rgen0 = pc.base + x + (long) (pc.ysize-1-y)*next;
    stop = 16;
    while (rgen0 < pc.base) {
        rgen0 += next;
        amap++;
        xmap++;
        stop--;
    }
    next -= 2; 

    /* store background */
    for (j = 0; j < 4; j++) {
        rgen = rgen0;
        outp(0x3cf,j);     /* read plane select */
        for (i = 0; i < stop; i++) {
            *px++ = *rgen++;
            *px++ = *rgen++;
            *px++ = *rgen;
            rgen += next;
        }
        px += 16;
    }
    next += 2;

    outp(0x3c4,0x2);        /* write plane enable */
    planemask = inp(0x3c5);

    outp(0x3ce,0x1);        /* enable set reset */
    esetreset = inp(0x3cf);
    outp(0x3cf,0x0);

    outp(0x3ce,0x8);        /* bit mask */
    bitmask = inp(0x3cf);
    outp(0x3cf,0xff);

    outp(0x3ce,0x7);        /* color dont care */
    cdontcare = inp(0x3cf);
    outp(0x3cf,0x00);

    outp(0x3ce,0x5);        /* read mode 1, write mode 0 */
    outp(0x3cf,8);

    outp(0x3ce,0x3);        /* data rotate */
    datarot = inp(0x3cf);

    rgen = rgen0;

    for (i = 0; i < stop; i++) {
        afld.i = *(amap++);
        afld.s[2] = 0xff;
        afld.l <<=  8-rem;
        afld.s[0] |= 0xff >> rem;

        xfld.i = *(xmap++);
        xfld.s[2] = 0;
        xfld.l <<= 8-rem;

        /* xcnt is unsigned */
        if ((unsigned) x+2 >= next) {
            afld.s[0] = 0xff;
            xfld.s[0] = 0x0;
        }
        if ((unsigned) x+1 >= next) {
            afld.s[1] = 0xff;
            xfld.s[1] = 0x0;
        }
        if ((unsigned) x   >= next) {
            afld.s[2] = 0xff;
            xfld.s[2] = 0x0;
        }

        outp(0x3c5,0xf);
        outp(0x3cf,0x08); /* AND */
        *rgen &= afld.s[2];
        rgen++;
        *rgen &= afld.s[1];
        rgen++;
        *rgen &= afld.s[0];
        rgen -= 2;

        outp(0x3c5,pc.cursor_color);
        outp(0x3cf,0x18); /* XOR */
        *rgen &= xfld.s[2];
        rgen++;
        *rgen &= xfld.s[1];
        rgen++;
        *rgen &= xfld.s[0];
        rgen += next - 2;
    }
    outp(0x3c5,planemask);
    outp(0x3ce,0x1); outp(0x3cf,esetreset); /* enable set reset */
    outp(0x3ce,0x3); outp(0x3cf,datarot);   /* data rotate */
    outp(0x3ce,0x4); outp(0x3cf,readplane); /* read plane */
    outp(0x3ce,0x5); outp(0x3cf,mode);      /* write mode */
    outp(0x3ce,0x7); outp(0x3cf,cdontcare); /* color dont care */
    outp(0x3ce,0x8); outp(0x3cf,bitmask);   /* bit mask */
}


#ifndef __GNUC__


void
mode_2_mark(x,y)

/* display cursor 256 color modes */
int x, y;
{
    int i, j, next;
    unsigned short xcnt, *amap, *xmap;
    unsigned char cbuf, pix, color, segsel, *px;
    union {long l; unsigned short i; char s[4]; } afld, xfld;
    union { unsigned short o[2]; long l; } p1;

    segsel = inp(0x3cd);        /* Tseng Labs segment selector */

    amap = (unsigned short *) vgacur.map;
    xmap = amap + 16;

    px = vgacur.pbs;
    next = pc.xsize;
    p1.l = x + (long) (pc.ysize-1-y)*next;

    cbuf = p1.o[1];
    outp(0x3cd,(cbuf << 4) | cbuf);
    color = pc.cursor_color;
    outpw(0x3ce,0x3); /* set alu mode */

    j = x;
    i = 16;
    while (i--) {

        afld.i = *(amap++);
        xfld.i = *(xmap++);

        xcnt = 0x8000;
        while (xcnt) {
            if (p1.o[1] != cbuf) {
                cbuf = p1.o[1];
                outp(0x3cd,(cbuf << 4) | cbuf);
            }
            pix = *(pc.base + p1.o[0]);
            *px++ = pix;
            if (j < 0 || j >= next)
                goto skip;
            if (afld.i & xfld.i & xcnt)
                *(pc.base+p1.o[0]) = color ^ pix;
            else if (~afld.i & xfld.i & xcnt)
                *(pc.base+p1.o[0]) = color;
            else if (~afld.i & ~xfld.i & xcnt)
                *(pc.base+p1.o[0]) = 0;
skip:
            p1.l++;
            j++;
            xcnt >>= 1;
        }
        j -= 16;
        p1.l += next - 16;
    }
    outp(0x3cd,segsel);
}


#else /* __GNUC__ */


void
mode_2_mark(x,y)

/* display cursor 256 color modes */
int x, y;
{
    int i, next;
    unsigned short xcnt, *amap, *xmap;
    unsigned char pix, color, *px;
    union {long l; unsigned short i; char s[4]; } afld, xfld;
    vidmptr rgen;

    amap = (unsigned short *) vgacur.map;
    xmap = amap + 16;

    px = vgacur.pbs;
    next = pc.xsize;
    rgen = pc.base + x + (long) (pc.ysize-1-y)*next;

    color = pc.cursor_color;
    outpw(0x3ce,0x3); /* set alu mode */

    i = 16;
    while (i--) {

        afld.i = *(amap++);
        xfld.i = *(xmap++);

        xcnt = 0x8000;
        while (xcnt) {

            if (rgen >= pc.base) {
                pix = *rgen;
                if (x >= 0 && x < next) {
                    if (afld.i & xfld.i & xcnt)
                        *rgen = color ^ pix;
                    else if (~afld.i & xfld.i & xcnt)
                        *rgen = color;
                    else if (~afld.i & ~xfld.i & xcnt)
                        *rgen = 0;
                }
                *px++ = pix;
            }
            else
                *px++ = 0;
            rgen++;
            x++;
            xcnt >>= 1;
        }
        x -= 16;
        rgen += next - 16;
    }
}


#endif /* __GNUC__ */


void
mode_1_clear(x,y)

/* overwrite cursor 16 color modes */
int x, y;
{
    int i, next;
    unsigned char bitmask, mode, datarot, esetreset, planemask, rem, *px;
    vidmptr rgen, rgen0;
    int stop;

    outp(0x3ce,0x5);
    mode = inp(0x3cf);
    outp(0x3cf,0x0);         /* write mode 0 */

    outp(0x3ce,0x3);
    datarot = inp(0x3cf);
    outp(0x3cf,0x0);         /* straight thru */

    outp(0x3ce,0x1);
    esetreset = inp(0x3cf);
    outp(0x3cf,0x0);         /* enable set reset */

    outp(0x3ce,0x8);
    bitmask = inp(0x3cf);
    outp(0x3cf,0xff);        /* bit mask */

    outp(0x3c4,0x2);
    planemask = inp(0x3c5);

    px = vgacur.pbs;

    next = pc.bytpline;
    rgen0 = pc.base + (x >> 3) + (long) (pc.ysize-y-1)*next;
    stop = 16;
    while (rgen0 < pc.base) {
        rgen0 += next;
        stop--;
    }
    next -= 2;

    rem = 1;
    while (rem & 0xf) {
        rgen = rgen0;
        outp(0x3c5,rem); /* write plane select */
        for (i = 0; i < stop; i++) {

            *rgen++ = *px++;
            *rgen++ = *px++;
            *rgen   = *px++;

            rgen += next;
        }
        px += 16;
        rem <<= 1;
    }

    outp(0x3c5,planemask);                  /* write plane */
    outp(0x3ce,0x1); outp(0x3cf,esetreset); /* enable set reset */
    outp(0x3ce,0x3); outp(0x3cf,datarot);   /* data rotate */
    outp(0x3ce,0x5); outp(0x3cf,mode);      /* write mode */
    outp(0x3ce,0x8); outp(0x3cf,bitmask);   /* bit mask */
}


#ifndef __GNUC__


void
mode_2_clear(x,y)

/* overwrite cursor 256 color modes */
int x, y;
{
    int i, j, next;
    unsigned char segsel, cbuf, *px;
    union { unsigned short o[2]; long l; } p1;

    segsel = inp(0x3cd);        /* Tseng Labs segment selector */

    px = vgacur.pbs;

    next = pc.xsize;
    p1.l = x + (long) (pc.ysize-1-y)*next;
    next -= 16;

    outp(0x3cd,cbuf = p1.o[1]);
    outpw(0x3ce,0x3); /* set alu mode */
    for (i = 0; i < 16; i++) {
        for (j = 0; j < 16; j++) {
            if (cbuf != p1.o[1])
                outp(0x3cd,cbuf = p1.o[1]);
            *(pc.base + p1.o[0]) = *px++;
            p1.l++;
        }
        p1.l += next;
    }
    outp(0x3cd,segsel);
}


#else /* __GNUC__ */


void
mode_2_clear(x,y)

/* overwrite cursor 256 color modes */
int x, y;
{
    int i, j, next;
    unsigned char *px;
    vidmptr rgen;

    px = vgacur.pbs;
    next = pc.xsize;
    rgen = pc.base + x + (long) (pc.ysize-1-y)*next;
    next -= 16;

    outpw(0x3ce,0x3); /* set alu mode */

    i = 16;
    while (i--) {
        j = 16;
        while (j--) {
            if (rgen >= pc.base)
                *rgen = *px;
            px++;
            rgen++;
        }
        rgen += next;
    }
}

#endif /* __GNUC__ */
