/*************************************************************************
 MFB graphics and miscellaneous library
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include "mfb.h"
#include "mfbP.h"
#include <dos.h>
#include <ctype.h>


MFB *MFBCurrent;        /* current MFB structure, not used */
struct mfbpc pc;        /* special structure to store stuff in */
char mfb_trash;         /* return from latch read */

/* So we can nest calls to MFBOpen(), we save the mfbpc structure. */
struct ctext {
    struct mfbpc *pc;
    struct ctext *next;
};
static struct ctext *Clist;

static int mfb_vmode;

#ifdef __STDC__
static int  get_new_mode(char*,int*,int*,int*);
static void set_default_colors(void);
static void set_vmode(int,int*,int*,int);
static void reset_vmode(int);
static int  get_vmode();
static int  check_drv(void);
#else
extern int  get_new_mode();
extern void set_default_colors();
static void set_vmode();
static void reset_vmode();
static int  get_vmode();
static int  check_drv();
#endif


#define alloc(zzz)   (struct zzz *)malloc(sizeof(struct zzz))


MFB *
MFBOpen(name, device, error)

char *name, *device;
int *error;
{
    char *malloc();
    char palette[16];
    int i;
    static MFB thismfb;
    struct ctext *cl;

    if (error) *error = MFBOK;

    MFBCurrent = &thismfb;
    MFBCurrent->fileDesc = fileno(stdout);
    MFBCurrent->buttonMask[0] = 1;
    MFBCurrent->buttonMask[1] = 2;
    MFBCurrent->buttonMask[2] = 3;
    MFBCurrent->buttonMask[3] = 4;

    if (pc.mfbMODE != 0) {
        /* already in a graphics mode, so save context */
        cl = alloc(ctext);
        if (cl == NULL) {
            *error = MFBNOMEM;
            return (NULL);
        }
        cl->pc = alloc(mfbpc);
        if (cl->pc == NULL) {
            *error = MFBNOMEM;
            return (NULL);
        }
        *cl->pc = pc;
        cl->next = Clist;
        Clist = cl;
    }

    /* default font */
    pc.chartab = mfb_8X14font;
    pc.charheight = 14;
    pc.charwidth = 8;

    pc.video_mode = get_vmode() & 0x7f;
    mfb_vmode = get_new_mode(name,&pc.xsize,&pc.ysize,&pc.mfbMODE);

#ifdef __GNUC__
    switch (check_drv()) {
        case 0:
            /* internal vga driver being used */
            if (mfb_vmode > 0x13) {
                pc.xsize = 640;
                pc.ysize = 480;
                pc.mfbMODE = 1;
                mfb_vmode = 0x12;
            }
            reset_vmode(mfb_vmode);
            break;

        case 1:
            /* old style driver, force 256 color mode */
            set_vmode(6,&pc.xsize,&pc.ysize,256);
            mfb_vmode = 6;
            pc.mfbMODE = 2;
            break;

        case 2:
            /* new driver, assume driver sets color mode properly */
            set_vmode(9,&pc.xsize,&pc.ysize,pc.mfbMODE==2 ? 256 : 16);
            mfb_vmode = 9;
            break;
    }
#else
    set_vmode(mfb_vmode,0,0,0);
#endif

    /* store current font height */
    outp(0x3d4,0x9);
    pc.scan_height = inp(0x3d5) & 0x0f;

    pc.linestyles[0] = (unsigned char) 0xff;
    pc.cursor_color = 1;
    pc.alumode = 3; /* MFBALUJAM */
    pc.curstep = 8;
    pc.curleap = 64;
    pc.pointerdelay = 250;    /* 250 msec */
    pc.functkey_yband = 10;

    if (!pc.base)
#if __NDPC__
        pc.base = (char *) mapdev(0xa0000,0x10000);
#else
#if __GNUC__
        pc.base = (char *) 0xd0000000;
#else
#if __TURBOC__
        pc.base = (vidmptr) MK_FP(0xa000,0x0000);
#else
        pc.base = I_cant_map_video_memory;
#endif
#endif
#endif
        
    pc.cursx = pc.xsize >> 1;
    pc.cursy = pc.ysize >> 1;

    pc.bytpline = pc.xsize >> 3;

    if (pc.mfbMODE != 4)
        MFBPointerInit();

    for (i = 0; i < 16; i++) palette[i] = (char) i;
    vga_set_cpblock(palette);
    set_default_colors();
    vga_set_cblock((char (*)[3])pc.colormap,0,16);
    vga_get_cblock((char (*)[3])pc.colormap[16],16,240);

    return (MFBCurrent);
}


int
MFBInitialize()

{
    char palette[16];
    int i;

    if (Clist != NULL)
        return (MFBBADDEV);
    if (mfb_vmode == 6 || mfb_vmode == 9)
        set_vmode(mfb_vmode,&pc.xsize,&pc.ysize,pc.mfbMODE==2 ? 256 : 16);
    else
        reset_vmode(mfb_vmode);

    if (pc.mfbMODE != 4)
        MFBPointerInit();
    for (i = 0; i < 16; i++) palette[i] = (char) i;
    vga_set_cpblock(palette);
    set_default_colors();
    vga_set_cblock((char(*)[3])pc.colormap,0,16);
    vga_get_cblock((char(*)[3])pc.colormap[16],16,240);
    return (MFBOK);
}


int
MFBClose()

{
    int i;
    struct ctext *cl;
    char palette[16];

    if (pc.mfbMODE != 4)
        MFBPointerClose();
    MFBSetColor(0);
    MFBFlood();
    reset_vmode(pc.video_mode);
    if (pc.video_mode < 4 || pc.video_mode == 7
#ifdef PARADISE
        || (pc.video_mode < 0x58 && pc.video_mode > 0x53)
#endif
        )
        load_font(pc.scan_height);

    if (Clist != NULL) {
        /* have to reset some things for old graphics mode */
        pc = *Clist->pc;
        free(Clist->pc);
        for (i = 0; i < 16; i++) palette[i] = (char) i;
        vga_set_cpblock(palette);
        if (pc.mfbMODE != 4)
            MFBPointerInit();
        cl = Clist;
        Clist = Clist->next;
        free(cl);
    }
    else
        pc.mfbMODE = 0;

    return (MFBOK);
}


int
MFBHalt()

{
    if (Clist != NULL)
        return (MFBBADDEV);
    if (pc.mfbMODE != 4)
        MFBPointerClose();
    reset_vmode(pc.video_mode);
    if (pc.video_mode < 4 || pc.video_mode == 7
#ifdef PARADISE
        || (pc.video_mode < 0x58 && pc.video_mode > 0x53)
#endif
        )
        load_font(pc.scan_height);

    return (MFBOK);
}


int
MFBUpdate()
{
    return (MFBOK);
}


int
MFBGetchar()

/* returns raw key */
{
    int k;

    k = (getch() & 0xff);
    if (!k) k = 256 + getch();
    return (k);
}


static int
get_new_mode(name,xsize,ysize,mfbmode)

/* returns TSENG ET4000 super-VGA modes */
char *name;
int *xsize,*ysize,*mfbmode;
{
    /* pc.mfbMODE:
     * 0, text mode
     * 1, 16 color (4 bit planes) modes
     * 2, 256 color (byte per pixel) modes
     * 4, rasterizing
     */

    if (!name) {              /* 640 X 480 16 colors    */
        *mfbmode = 1;
        *xsize = 640;
        *ysize = 480;
        return (0x12);
    }
    if (!strcmp(name,"vga0")) { /* 640 X 350 16 colors */
        *mfbmode = 1;
        *xsize = 640;
        *ysize = 350;
        return (0x10);
    }
    if (!strcmp(name,"vga1")) { /* 640 X 480 16 colors */
        *mfbmode = 1;
        *xsize = 640;
        *ysize = 480;
        return (0x12);
    }
    if (!strcmp(name,"vga2")) { /* 320 X 200 256 colors */
        *mfbmode = 2;
        *xsize = 320;
        *ysize = 200;
        return (0x13);
    }
    if (!strcmp(name,"vga3")) { /* 800 X 600 16 colors */
        *mfbmode = 1;
        *xsize = 800;
        *ysize = 600;
#ifdef PARADISE
        return (0x58);
    }
#else
        return (0x29);
    }
    if (!strcmp(name,"vga4")) { /* 1024 X 768 16 colors */
        *mfbmode = 1;
        *xsize = 1024;
        *ysize = 768;
        return (0x37);
    }
    if (!strcmp(name,"vga5")) { /* 640 X 480 256 colors */
        *mfbmode = 2;
        *xsize = 640;
        *ysize = 480;
        return (0x2e);
    }
    if (!strcmp(name,"vga6")) { /* 800 X 600 256 colors */
        *mfbmode = 2;
        *xsize = 800;
        *ysize = 600;
        return (0x30);
    }
    if (!strcmp(name,"vga7")) { /* 1024 X 768 256 colors */
        *mfbmode = 2;
        *xsize = 1024;
        *ysize = 768;
        return (0x38);
    }
#endif
    if (!strcmp(name,"core")) { /* 640 X 480 16 colors */

/* Rasterizing for 16 color modes, for HP LasetJet.
 * The HP Laserjet has a graphics map 2400 X 3180.  This is written in
 * segments from vga memory, which has 64K size.  This works out to 15
 * segments, each 300 chars X 208 lines.
 */
        *mfbmode = 4;
        *xsize = 2400;
        *ysize = 208;
                /* 300 X 208, 15 pages
                 * 3180 is physical page length, 208
                 * set by 64K vga memory page size and
                 * mult of 16 for fillpattern
                 * continuity. Thus 3120 lines are used.
                 */
        return (0x12);
    }

    /* default -  640 X 480 16 colors */
    *mfbmode = 1;
    *xsize = 640;
    *ysize = 480;
    return (0x12);
}


static void
set_default_colors()

{
    unsigned char (*c)[3];

    c = pc.colormap;
    (*c)[0] = (unsigned char) 0;
    (*c)[1] = (unsigned char) 0;
    (*c)[2] = (unsigned char) 0;
    c++;
    (*c)[0] = (unsigned char) 60;
    (*c)[1] = (unsigned char) 60;
    (*c)[2] = (unsigned char) 60;
    c++;
    (*c)[0] = (unsigned char) 60;
    (*c)[1] = (unsigned char) 0;
    (*c)[2] = (unsigned char) 0;
    c++;
    (*c)[0] = (unsigned char) 20;
    (*c)[1] = (unsigned char) 60;
    (*c)[2] = (unsigned char) 20;

    c++;
    (*c)[0] = (unsigned char) 35;
    (*c)[1] = (unsigned char) 25;
    (*c)[2] = (unsigned char) 60;
    c++;
    (*c)[0] = (unsigned char) 60;
    (*c)[1] = (unsigned char) 60;
    (*c)[2] = (unsigned char) 30;
    c++;
    (*c)[0] = (unsigned char) 20;
    (*c)[1] = (unsigned char) 60;
    (*c)[2] = (unsigned char) 60;
    c++;
    (*c)[0] = (unsigned char) 60;
    (*c)[1] = (unsigned char) 30;
    (*c)[2] = (unsigned char) 60;

    c++;
    (*c)[0] = (unsigned char) 40;
    (*c)[1] = (unsigned char) 40;
    (*c)[2] = (unsigned char) 40;
    c++;
    (*c)[0] = (unsigned char) 45;
    (*c)[1] = (unsigned char) 45;
    (*c)[2] = (unsigned char) 15;
    c++;
    (*c)[0] = (unsigned char) 0;
    (*c)[1] = (unsigned char) 40;
    (*c)[2] = (unsigned char) 0;
    c++;
    (*c)[0] = (unsigned char) 0;
    (*c)[1] = (unsigned char) 0;
    (*c)[2] = (unsigned char) 40;

    c++;
    (*c)[0] = (unsigned char) 50;
    (*c)[1] = (unsigned char) 30;
    (*c)[2] = (unsigned char) 20;
    c++;
    (*c)[0] = (unsigned char) 25;
    (*c)[1] = (unsigned char) 35;
    (*c)[2] = (unsigned char) 40;
    c++;
    (*c)[0] = (unsigned char) 40;
    (*c)[1] = (unsigned char) 0;
    (*c)[2] = (unsigned char) 40;
    c++;
    (*c)[0] = (unsigned char) 63;
    (*c)[1] = (unsigned char) 63;
    (*c)[2] = (unsigned char) 63;
}


void
GetScreenSizeParameters(x,y,bytpline)
int **x,**y,**bytpline;
{
    *x = &pc.xsize;
    *y = &pc.ysize;
    *bytpline = &pc.bytpline;
}


int
MFBSetGhost(callback,x,y)

/* Nonstandard routine: when called with callback not NULL, the callback
 * function will be called after cursor motion, and any drawing done in the
 * callback will be xor'ed to the screen.  This enables rubber banding,
 * and "ghost" images to be attached to the cursor.
 */
#ifdef __STDC__
void (*callback)(int,int,int,int);
#else
void (*callback)();
#endif
int x, y;
{
    if (callback) {
        pc.refx = x;
        pc.refy = y;
        pc.drawghost = callback;
        return (MFBOK);
    }
    pc.drawghost = NULL;
    return (MFBOK);
}


#if __NDPC__
#define REGS REGS16
#endif


/* ARGSUSED */
static void
set_vmode(mode,width,height,numcolors)

int mode, *width, *height, numcolors;
{
    union REGS r;

#ifdef __GNUC__
    r.x.ax = (mode & 0xff) | 0xff00;
    r.x.bx = numcolors;
    r.x.cx = *width;
    r.x.dx = *height;
#else
    r.x.ax = mode & 0xff;
#endif
    int86(0x10,&r,&r);
#ifdef __GNUC__
    *width = r.x.cx;
    *height = r.x.dx;
#endif
}


static void
reset_vmode(mode)
{
    union REGS r;

    r.x.ax = mode & 0xff;
    int86(0x10,&r,&r);
}


static int
get_vmode()
{
    union REGS r;

    r.x.ax = 0x0f00;
    int86(0x10,&r,&r);
    return (r.x.ax & 0xff);
}


#ifdef __GNUC__

static int
check_drv()

/* See if we have a new driver (.grn), an old driver (.grd),
 * or no driver by parsing the GO32 environment variable.
 * Returns 2 for new, 1 for old, 0 if none or unrecognized.
 */
{
    char *s, *t, buf[128];
    char *getenv(), *strchr(), *strrchr();

    s = getenv("GO32");
    if (s) {
        strcpy(buf,s);
        strlwr(buf);
        s = buf;
        while ((s = strchr(s,'d')) != NULL) {
            if (s[1] == 'r' && s[2] == 'i' && s[3] == 'v')
                break;
            s++;
        }
        if (*s == 'd') {
            while (*s && !isspace(*s)) s++;
            if (*s) while (isspace(*s)) s++;
            if (*s) {
                t = s;
                while (*t && !isspace(*t)) t++;
                *t = '\0';
                t = strrchr(s,'.');
                if (t) {
                    if (!strncmp(t,".grd",4)) {
                        return (1);
                    }
                    else if (!strncmp(t,".grn",4)) {
                        return (2);
                    }
                    else
                        return (0);
                }
            }
        }
    }
    return (0);
}

#endif
