/*************************************************************************
 MFB graphics and miscellaneous library
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include "mfb.h"
#include "mfbP.h"
#include <dos.h>

#define ror(x,n) ((x >> n)  | (x << (8-n)))

#define pswap(tab,i,j) {int temp; temp = *((tab)+(i));\
            *((tab)+(i)) = *((tab)+(j));\
            *((tab)+(j)) = temp; }

long mfb_polybuffer[2*MFBPOLYGONBUFSIZE];

#ifdef __STDC__
static void sort(int*,int,int);
static void mode_1_zoids(int*,int*,int,int,int);
static void mode_2_zoids(int*,int*,int,int,int);
#else
static void sort();
static void mode_1_zoids();
static void mode_2_zoids();
#endif


void
MFBPolygon(poly)
MFBPOLYGON *poly;
{
    int i, twonvert, *line, stored = 0;
    int *xup, *ll, ny;
    int *xlp, *xip;
    int yl, yu;
    int sto1, sto2;
    int *ylist = (int *) mfb_polybuffer;
    int *xulist = ylist + MFBMAXPOLYGONVERTICES;
    int *xllist = xulist + MFBMAXPOLYGONVERTICES;
    int *xilist = xllist + MFBMAXPOLYGONVERTICES;
    
    twonvert = poly->nvertices << 1;

    line = (int *) poly->xy;

    if ((line[0] != line[twonvert-2]) || (line[1] != line[twonvert-1])) {

        sto1 = line[twonvert];       /* who knows what lives here!  */
        sto2 = line[twonvert+1];     /* might generate gp fault     */
        line[twonvert] = line[0];    /* on extremely rare occasions */
        line[twonvert+1] = line[1];
        twonvert += 2;
        stored = 1;
    }

    xip = line + twonvert;
    ll = ylist;
    while (line < xip) {
        *ll++ = *(line+1);
        line += 2;
    }
    ny = ll-ylist-1;
    sort(ylist,0,ny);

    xip = xilist;
    ll = ylist;
    while (ny--) {
        yl = *ll;
        yu = *(++ll);
        while (yu == yl) {
            if (!ny) goto done;
            ny--;
            yu = *(++ll);
        }
        memcpy(xllist,xilist,(int)((char *) xip - (char *) xilist));
        xlp = xllist + (int) (xip - xilist);
        xup = xulist;
        xip = xilist;
        
        for (i = twonvert - 4; i >= 0; i -= 2) {
            line = (int *) poly->xy + i;
            if (((line[1] > yu) && (line[3] < yu)) ||
                ((line[3] > yu) && (line[1] < yu))) {

                *xup++ = *xip++ = line[0] +
                    ((long)(yu-line[1])*(line[2]-line[0])) /
                    (line[3]-line[1]);
            }
            if (line[1] == yu) {

                if (line[3] < yu)
                    *xup++ = line[0];
                
                if (i) {
                    if (*(line-1) < yu)
                        *xup++ = line[0];
                }
                else
                    if (line[twonvert-3] < yu)
                        *xup++ = line[0];
            }
            if (line[1] == yl) {

                if (line[3] > yl)
                    *xlp++ = line[0];
                
                if (i) {
                    if (*(line-1) > yl)
                        *xlp++ = line[0];
                }
                else
                    if (line[twonvert-3] > yl)
                        *xlp++ = line[0];
            }
        }

        sort(xulist,0,(int)(xup-xulist-1));
        sort(xllist,0,(int)(xlp-xllist-1));

        if (pc.mfbMODE & 2)
            mode_2_zoids(xllist,xulist,(int)(xlp-xllist),yl,yu);
        else
            mode_1_zoids(xllist,xulist,(int)(xlp-xllist),yl,yu);
    }
done:
    if (stored) {
        line = (int *) poly->xy;
        line[twonvert-2] = sto1;
        line[twonvert-1] = sto2;
    }
}


static void
sort(tab,left,right)

int left,right, *tab;
{
    int i, last;

    if (right <= left) return;
    i = (left + right) >> 1;
    pswap(tab,left,i);
    last = left;
    for (i = left + 1; i <= right; i++)
        if (tab[i] < tab[left]) {
            last++;
            pswap(tab,last,i);
        }
    pswap(tab,left,last);
    sort(tab,left,last-1);
    sort(tab,last+1,right);
}


static void
mode_1_zoids(xll,xul,num,yl,yu)

int *xll, *xul, num, yl, yu;
{
    int x1, x2, x3, x4;
    int dx, dy, dy2, dx1, dx2, errterm1, errterm2;
    int i, s1, s2, lnum, lcnt, next, *st = NULL;
    vidmptr rgen, rgen0, rgen1;
    unsigned char left, right, cbuf;
    union { unsigned short color2; unsigned char c[2]; } c;

    if (pc.curfillpatt)
        st = pc.stipples[pc.curfillpatt];

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];
    lnum = (pc.ysize-1-yu);

    dy2 = yu - yl;

    next = pc.bytpline;
    rgen1 = pc.base + (long) lnum*next;

    outpw(0x3ce,0x0b05); /* read mode 1, write mode 3 */
    outpw(0x3ce,c.color2 & 0xff00); /* set/reset */
    outpw(0x3ce,0x7); /* zero color dont care */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    for (i = 0; i < num; i += 2) {

        rgen0 = rgen1;

        x1 = xll[i];
        x4 = xll[i+1];
        x2 = xul[i];
        x3 = xul[i+1];

        dy = dy2 + 1;

        dx1 = x1 - x2;
        dx2 = x3 - x4;
        s1 = 1;
        s2 = -1;
        if (x2 >= x1) {    dx1 = -dx1; s1 = -1; }
        if (x4 >= x3) {    dx2 = -dx2; s2 = 1; }
        errterm1 = errterm2 = 0;
        x3++;
        x4++;

        if (!st)
            while (dy--) {
                left  =  (0xff >> (x2 & 7));
                right = ~(0xff >> (x3 & 7));
                dx = (x3 >> 3) - (x2 >> 3) - 1;
                if (dx < 0) { left &= right; dx = 0; right = 0; }
                rgen = rgen0 + (x2 >> 3);
                mfb_trash = *rgen;
                *rgen = left;
                rgen++;
                while (dx--) {
                    mfb_trash = *rgen;
                    *rgen = 0xff;
                    rgen++;
                }
                mfb_trash = *rgen;
                *rgen = right;
                rgen0 += next;
                errterm1 += dx1;
                errterm2 += dx2;
                while (errterm1 > 0 && x1 != x2) {
                    errterm1 -= dy2;
                    x2 += s1;
                }
                while (errterm2 > 0 && x3 != x4) {
                    errterm2 -= dy2;
                    x3 += s2;
                }
            }
        else {
            lcnt = lnum;
            while (dy--) {
                left  =  (0xff >> (x2 & 7));
                right = ~(0xff >> (x3 & 7));
                dx = (x3 >> 3) - (x2 >> 3) - 1;
                if (dx < 0) { left &= right; dx = 0; right = 0; }
                rgen = rgen0 + (x2 >> 3);
                cbuf = st[lcnt++ & 7];
                mfb_trash = *rgen;
                *rgen = cbuf & left;
                rgen++;
                while (dx--) {
                    mfb_trash = *rgen;
                    *rgen = cbuf;
                    rgen++;
                }
                mfb_trash = *rgen;
                *rgen = cbuf & right;
                rgen0 += next;

                errterm1 += dx1;
                errterm2 += dx2;
                while (errterm1 > 0 && x1 != x2) {
                    errterm1 -= dy2;
                    x2 += s1;
                }
                while (errterm2 > 0 && x3 != x4) {
                    errterm2 -= dy2;
                    x3 += s2;
                }
            }
        }
    }
    outpw(0x3ce,0x5);
    outpw(0x3ce,0xff07);
}


#ifndef __GNUC__

static void
mode_2_zoids(xll,xul,num,yl,yu)

int *xll, *xul, num, yl, yu;
{
    int x1, x2, x3, x4;
    int dy, dy2, dx1, dx2, errterm1, errterm2;
    unsigned short dx;
    int i, s1, s2, lnum, lcnt, next, *st = NULL;
    unsigned char left, right, cbuf;
    union { unsigned short o[2]; long l; } p1, p2, p3, p4;
    union { unsigned short color2; unsigned char c[2]; } c;
    vidmptr rgen;

    if (pc.curfillpatt)
        st = pc.stipples[pc.curfillpatt];

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];
    lnum = (pc.ysize-1-yu);

    outpw(0x3ce,0xff08);     /* set bit mask */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    dy2 = yu - yl;

    next = pc.xsize;
    p4.l = (long) lnum*next;
    cbuf = inp(0x3cd);
    cbuf |= (cbuf << 4);
    for (i = 0; i < num; i += 2) {

        p3.l = p4.l;

        x1 = xll[i];
        x4 = xll[i+1];
        x2 = xul[i];
        x3 = xul[i+1];

        dy = dy2 + 1;

        dx1 = x1 - x2;
        dx2 = x3 - x4;
        s1 = 1;
        s2 = -1;
        if (x2 >= x1) {    dx1 = -dx1; s1 = -1; }
        if (x4 >= x3) {    dx2 = -dx2; s2 = 1; }
        errterm1 = errterm2 = 0;
        x3++;
        x4++;

        if (!st) {

            while (dy--) {
                dx = x3 - x2;
                p1.l = p3.l + x2;
                p2.l = p3.l + x3;
                if (p1.o[1] != p2.o[1])
                    dx = ~(p1.o[0]) + 1;
                if ((cbuf & 0xf) != p1.o[1]) {
                    cbuf = p1.o[1];
                    cbuf |= (cbuf << 4);
                    outp(0x3cd,cbuf);
                }
                rgen = pc.base + p1.o[0];
                if (pc.alumode == 3) {
                    if ((unsigned)rgen & 1) {
                        *rgen = c.c[0];
                        rgen++;
                        dx--;
                    }
                    if (dx & 1)
                        *(rgen + dx - 1) = c.c[0];
                    dx >>= 1;
                    while (dx--) {
                        *(short far *)rgen = c.color2;
                        rgen += 2;
                    }
                }
                else {
                    while (dx--) {
                        mfb_trash = *rgen;
                        *rgen = c.c[0];
                        rgen++;
                    }
                }
                
                if (p1.o[1] != p2.o[1]) {
                    outp(0x3cd,cbuf + 0x11);
                    dx = p2.o[0];
                    rgen = pc.base;
                    if (pc.alumode == 0x3) {
                        if (dx & 1)
                            *(rgen + dx - 1) = c.c[0];
                        dx >>= 1;
                        while (dx--) {
                            *(short far *)rgen = c.color2;
                            rgen += 2;
                        }
                    }
                    else {
                        while (dx--) {
                            mfb_trash = *rgen;
                            *rgen = c.c[0];
                            rgen++;
                        }
                    }

                }
                p3.l += next;

                errterm1 += dx1;
                errterm2 += dx2;
                while (errterm1 > 0 && x1 != x2) {
                    errterm1 -= dy2;
                    x2 += s1;
                }
                while (errterm2 > 0 && x3 != x4) {
                    errterm2 -= dy2;
                    x3 += s2;
                }
            }
        }
        else {
            lcnt = lnum;
            while (dy--) {
                right = st[lcnt++ & 7];
                left = 0x80 >> (x2 & 7);
                dx = x3 - x2;
                p1.l = p3.l + x2;

                if ((cbuf & 0xf) != p1.o[1]) {
                    cbuf = p1.o[1];
                    cbuf |= (cbuf << 4);
                    outp(0x3cd,cbuf);
                }
                rgen = pc.base + p1.o[0];

                while (dx--) {
                    if (left & right) {
                        if (pc.alumode != 3)
                            mfb_trash = *rgen;
                        *rgen = c.c[0];
                    }
                    left = ror(left,1);
                    rgen++;
                }
                if (p1.o[1] != p2.o[1]) {
                    outp(0x3cd,cbuf + 0x11);
                    dx = p2.o[0];
                    rgen = pc.base;
                    while (dx--) {
                        if (left & right) {
                            if (pc.alumode != 3)
                                mfb_trash = *rgen;
                            *rgen = c.c[0];
                        }
                        left = ror(left,1);
                        rgen++;
                    }
                }

                p3.l += next;
                errterm1 += dx1;
                errterm2 += dx2;
                while (errterm1 > 0 && x1 != x2) {
                    errterm1 -= dy2;
                    x2 += s1;
                }
                while (errterm2 > 0 && x3 != x4) {
                    errterm2 -= dy2;
                    x3 += s2;
                }
            }
        }
    }
}


#else /* __GNUC__ */


static void
mode_2_zoids(xll,xul,num,yl,yu)

int *xll, *xul, num, yl, yu;
{
    int x1, x2, x3, x4;
    int dx, dy, dy2, dx1, dx2, errterm1, errterm2;
    int i, s1, s2, lnum, lcnt, next, *st = NULL;
    unsigned char left, right;
    int p0, p1;
    union { unsigned short color2; unsigned char c[2]; } c;
    vidmptr rgen;

    if (pc.curfillpatt)
        st = pc.stipples[pc.curfillpatt];

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];
    lnum = (pc.ysize-1-yu);

    outpw(0x3ce,0xff08);     /* set bit mask */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    dy2 = yu - yl;
    next = pc.xsize;
    p0 = (long) lnum*next;

    for (i = 0; i < num; i += 2) {

        p1 = p0;

        x1 = xll[i];
        x4 = xll[i+1];
        x2 = xul[i];
        x3 = xul[i+1];

        dy = dy2 + 1;

        dx1 = x1 - x2;
        dx2 = x3 - x4;
        s1 = 1;
        s2 = -1;
        if (x2 >= x1) {    dx1 = -dx1; s1 = -1; }
        if (x4 >= x3) {    dx2 = -dx2; s2 = 1; }
        errterm1 = errterm2 = 0;
        x3++;
        x4++;

        if (!st) {

            while (dy--) {
                dx = x3 - x2;
                rgen = pc.base + p1 + x2;
                if (pc.alumode == 3) {
                    if ((unsigned)rgen & 1) {
                        *rgen = c.c[0];
                        rgen++;
                        dx--;
                    }
                    if (dx & 1)
                        *(rgen + dx - 1) = c.c[0];
                    dx >>= 1;
                    while (dx--) {
                        *(short*)rgen = c.color2;
                        rgen += 2;
                    }
                }
                else {
                    while (dx--) {
                        mfb_trash = *rgen;
                        *rgen = c.c[0];
                        rgen++;
                    }
                }
                p1 += next;
                errterm1 += dx1;
                errterm2 += dx2;
                while (errterm1 > 0 && x1 != x2) {
                    errterm1 -= dy2;
                    x2 += s1;
                }
                while (errterm2 > 0 && x3 != x4) {
                    errterm2 -= dy2;
                    x3 += s2;
                }
            }
        }
        else {
            lcnt = lnum;
            while (dy--) {
                right = st[lcnt++ & 7];
                left = 0x80 >> (x2 & 7);
                dx = x3 - x2;
                rgen = pc.base + p1 + x2;

                while (dx--) {
                    if (left & right) {
                        if (pc.alumode != 3)
                            mfb_trash = *rgen;
                        *rgen = c.c[0];
                    }
                    left = ror(left,1);
                    rgen++;
                }
                p1 += next;
                errterm1 += dx1;
                errterm2 += dx2;
                while (errterm1 > 0 && x1 != x2) {
                    errterm1 -= dy2;
                    x2 += s1;
                }
                while (errterm2 > 0 && x3 != x4) {
                    errterm2 -= dy2;
                    x3 += s2;
                }
            }
        }
    }
}

#endif /* __GNUC__ */
