/*************************************************************************
 MFB graphics and miscellaneous library
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include "mfb.h"
#include "mfbP.h"
#include <dos.h>

static  Xl,Yl, Xu = 0x7fff, Yu = 0x7fff;

#ifdef __STDC__
static void mode1_render_text(char*,int,int,int,int);
static void mode2_render_text(char*,int,int,int,int);
static int setcode(int);
static void xform_cell(unsigned short*,unsigned short*,int,int,int,int);
#else
static void mode1_render_text();
static void mode2_render_text();
static int setcode();
static void xform_cell();
#endif


void
MFBSetTextClip(xl,yl,xu,yu)

/* set up a clipping window for text */
int xl,yl,xu,yu;
{
    Xl = xl;
    Yl = yl;
    Xu = xu;
    Yu = yu;
}


void
MFBTextBB(string,len,hei)

char *string;
int *len,*hei;
{
    *len = 0;
    *hei = 0;
    if (string == NULL) return;
    *len = strlen(string)*pc.charwidth;
    *hei = pc.charheight;
}


void
MFBText(s,x,y,r)

char *s;
int x, y, r;
{
    if (pc.mfbMODE & 2)
        mode2_render_text(s,x,y,r,1);
    else
        mode1_render_text(s,x,y,r,1);
}


void
MFBScaledText(s,x,y,r,sc)

char *s;
int x,y,r,sc;
{
    if (pc.mfbMODE & 2)
        mode2_render_text(s,x,y,r,sc);
    else
        mode1_render_text(s,x,y,r,sc);
}


static void
mode1_render_text(text,x,y,degrees,scale)

char *text;
int x, y, degrees, scale;

/* Print the text, where each pixel expands to a (scale X scale) pixel field.
 * Supports fonts up to 16 X 16.
 * Assumes bit field in degrees:
 * 0x8: mirror x
 * 0x4: mirror y
 * 0x3: 0-no rotation, 1-90, 2-180, 3-270
 *
 * if > 15, assume rotation in degrees: 90,180,270.
 *
 * For mode 1 (16 colors).
 */

{
    int i, j, k, l, tx;
    int next, next1, bwidth, cwidth;
    int fheight, fwidth;
    int ynow, xw, cx, xn, bcnt;
    unsigned short rotbuf[16], *ctab;
    unsigned short z, z1, imask;
    unsigned char cbuf, omask;
    unsigned char mask[22];
    vidmptr rgen;
    char *str0, *str1;
    union { unsigned short color2; unsigned char c[2]; } c;

    if (scale < 1 || scale > 10) return;
    if (text == NULL) return;

    if (degrees >= 90) degrees /= 90;

    c.c[0] = pc.curcolor;
    c.c[1] = c.c[0];

    outpw(0x3ce,0x0b05);     /* read mode 1, write mode 3 */
    outpw(0x3ce,0x0007);     /* clear color dont care reg */
    outpw(0x3ce,c.color2 & 0xff00);  /* set/reset */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    if (degrees & 1) {
        fheight = pc.charwidth;
        fwidth = pc.charheight;
        cwidth = fwidth*scale;
        if (degrees & 2) {
            if (!(degrees & 4))
                y -= strlen(text)*fheight*scale;
            if (degrees & 8)
                x -= cwidth;
        }
        else {
            if (!(degrees & 8))
                x -= cwidth;
            if (degrees & 4)
                y -= strlen(text)*fheight*scale;
        }
    }
    else {
        fheight  = pc.charheight;
        fwidth = pc.charwidth;
        next1 = pc.bytpline*fheight*scale;
        cwidth = fwidth*scale;
        if (degrees & 2) {
            if (!(degrees & 8))
                x -= strlen(text)*cwidth;
            if (!(degrees & 4))
                y -= fheight*scale;
        }
        else {
            if (degrees & 8)
                x -= strlen(text)*cwidth;
            if (degrees & 4)
                y -= fheight*scale;
        }
    }
    degrees = setcode(degrees);

    rgen = pc.base + (long) ((x >> 3) + (pc.ysize-1-y)*pc.bytpline);
    cx = x & 7;
    xn = cx + cwidth;
    bwidth = xn >> 3;
    next = pc.bytpline + bwidth;
    str0 = text;
    str1 = text + strlen(text);

    while (str0 != str1) {
        if (x + cwidth < Xl || x > Xu) {
            if (degrees & 4) break;
            if (degrees & 8)
                str1--;
            else
                str0++;
        }
        else {
            memset(mask,0xff,bwidth+1);
            if (x < Xl) {
                z = (Xl >> 3) - (x >> 3);
                memset(mask,0,z);
                mask[z] >>= (Xl & 7);
            }
            if (Xu < x + cwidth) {
                z = (Xu >> 3) - (x >> 3) + 1;
                memset(mask+z,0,bwidth+1-z);
                mask[z-1] <<= 8 - (Xu & 7);
            }

            if (degrees & 8) {
                if ((tx = (*--str1 & 0x7f) - ' ') < 0) tx = 0;
            }
            else
                if ((tx = (*str0++ & 0x7f) - ' ') < 0) tx = 0;

            ctab = pc.chartab + tx*pc.charheight;
            xform_cell(rotbuf,ctab,fwidth,fheight,degrees,fwidth);
            ctab = rotbuf + fheight - 1;

            if (scale > 1) {
                ynow = y;
                for (i = fheight; i; i--,ctab--) {
                    for (j = scale; j; j--) {
                        if (ynow >= Yl && ynow <= Yu) {
                            cbuf = 0;
                            omask = 0x80 >> cx;
                            imask = 0x8000;
                            bcnt = 0;
                            for (k = fwidth; k; k--) {
                                for (l = scale; l; l--) {
                                    if (imask & *ctab) cbuf |= omask;
                                    if (!(omask >>= 1)) {
                                        omask = 0x80;
                                        cbuf &= mask[bcnt++];
                                        mfb_trash = *rgen;
                                        *rgen++ = cbuf;
                                        cbuf = 0;

                                    }
                                }
                                imask >>= 1;
                            }
                            if (omask != 0x80) {
                                mfb_trash = *rgen;
                                *rgen = cbuf;
                            }
                        }
                        else
                            rgen += bwidth;
                        rgen -= next;
                        ynow++;
                    }
                }
                if (degrees & 4) {
                    y += fheight*scale;
                    continue;
                }
                rgen += next1;
            }
            else {
                ynow = y;
                for (i = fheight; i; i--, ctab--) {
                    if (ynow >= Yl && ynow <= Yu) {
                        imask = *ctab >> cx;
                        mfb_trash = *rgen;
                        *rgen = (*((char *) &imask + 1) & mask[0]);
                        if (bwidth) {
                            rgen++;
                            mfb_trash = *rgen;
                            *rgen = ((char) imask & mask[1]);
                        }
                        if (bwidth & 2) {
                            rgen++;
                            mfb_trash = *rgen;
                            *rgen =
                                ((char) (*ctab << (8 - cx)) & mask[2]);
                        }
                    }
                    else
                        rgen += bwidth;
                    rgen -= next;
                    ynow++;
                }
                if (degrees & 4) {
                    y += fheight;
                    continue;
                }
                rgen += next1;
            }
        }
        next -= bwidth;
        rgen += bwidth;
        cx = xn & 7;
        xn = cx + cwidth;
        bwidth = xn >> 3;
        next += bwidth;
        x += cwidth;
    }
    outpw(0x3ce,0xff07);
    outpw(0x3ce,0x5);
}


#ifndef __GNUC__


static void
mode2_render_text(text,x,y,degrees,scale)

/* Print text, 256 color modes.  Supports up to 16 X 16 fonts */
char *text;
int x, y, degrees, scale;
{


    int i, ii, j, k, tx, next;
    int xnow, ynow;
    long next1;
    int fheight, fwidth;
    unsigned short rotbuf[16], *ctab, mask;
    union {long l; unsigned short o[2];} p;
    unsigned char cbuf;
    char *str0, *str1;
    union { unsigned short color2; unsigned char c[2]; } c;

    outpw(0x3ce,0xff08);     /* set bit mask */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    c.c[0] = pc.curcolor;

    if (scale < 1 || scale > 10) return;
    if (text == NULL) return;

    if (degrees >= 90) degrees /= 90;

    cbuf = inp(0x3cd);
    cbuf |= (cbuf << 4);
    if (degrees & 1) {
        fheight = pc.charwidth;
        fwidth = pc.charheight;
        next1 = 0;
        if (degrees & 2) {
            if (!(degrees & 4))
                y -= strlen(text)*fheight*scale;
            if (degrees & 8)
                x -= fwidth*scale;
        }
        else {
            if (!(degrees & 8))
                x -= fwidth*scale;
            if (degrees & 4)
                y -= strlen(text)*fheight*scale;
        }
    }
    else {
        fheight  = pc.charheight;
        fwidth = pc.charwidth;
        next1 = (pc.xsize*fheight + fwidth)*scale;
        if (degrees & 2) {
            if (!(degrees & 8))
                x -= strlen(text)*fwidth*scale;
            if (!(degrees & 4))
                y -= fheight*scale;
        }
        else {
            if (degrees & 8)
                x -= strlen(text)*fwidth*scale;
            if (degrees & 4)
                y -= fheight*scale;
        }
    }
    degrees = setcode(degrees);

    p.l = x + (long) (pc.ysize-y-1)*pc.xsize;
    next = pc.xsize + fwidth*scale;
    str0 = text;
    str1 = text + strlen(text);
    while (str0 != str1) {

        if (degrees & 8) {
            if ((tx = (*--str1 & 0x7f) - ' ') < 0) tx = 0;
        }
        else
            if ((tx = (*str0++ & 0x7f) - ' ') < 0) tx = 0;

        ctab = pc.chartab + tx*pc.charheight;
        xform_cell(rotbuf,ctab,fwidth,fheight,degrees,fwidth);
        ctab = rotbuf + fheight - 1;

        ynow = y;
        for (i = fheight; i; i--, ctab--) {
            for (j = scale; j; j--) {
                if (ynow >= Yl && ynow <= Yu) {
                    xnow = x;
                    for (mask = 0x8000,ii = fwidth; ii; mask >>= 1,ii--) {
                        for (k = scale; k; k--) {
                            if (xnow >= Xl && xnow <= Xu) {
                                if (*ctab & mask) {
                                    if (cbuf != p.o[1]) {
                                        cbuf = p.o[1];
                                        cbuf |= (cbuf << 4);
                                        outp(0x3cd,cbuf);
                                    }
                                    if (pc.alumode != 3)
                                        mfb_trash = *(pc.base + p.o[0]);
                                    *(pc.base + p.o[0]) = c.c[0];
                                }
                            }
                            p.l++;
                            xnow++;
                        }
                    }
                }
                else
                    p.l += fwidth*scale;
                p.l -= next;
                ynow++;
            }
        }
        if (degrees & 4) {
            y += fheight*scale;
            continue;
        }
        x += fwidth*scale;
        p.l += next1;
    }
}


#else /* __GNUC__ */


static void
mode2_render_text(text,x,y,degrees,scale)

/* Print text, 256 color modes.  Supports up to 16 X 16 fonts */
char *text;
int x, y, degrees, scale;
{


    int i, ii, j, k, tx, next;
    int xnow, ynow;
    long next1;
    int fheight, fwidth;
    unsigned short rotbuf[16], *ctab, mask;
    char *str0, *str1;
    union { unsigned short color2; unsigned char c[2]; } c;
    vidmptr rgen;

    outpw(0x3ce,0xff08);     /* set bit mask */
    outpw(0x3ce,pc.alumode); /* set alu mode */

    c.c[0] = pc.curcolor;

    if (scale < 1 || scale > 10) return;
    if (text == NULL) return;

    if (degrees >= 90) degrees /= 90;

    if (degrees & 1) {
        fheight = pc.charwidth;
        fwidth = pc.charheight;
        next1 = 0;
        if (degrees & 2) {
            if (!(degrees & 4))
                y -= strlen(text)*fheight*scale;
            if (degrees & 8)
                x -= fwidth*scale;
        }
        else {
            if (!(degrees & 8))
                x -= fwidth*scale;
            if (degrees & 4)
                y -= strlen(text)*fheight*scale;
        }
    }
    else {
        fheight  = pc.charheight;
        fwidth = pc.charwidth;
        next1 = (pc.xsize*fheight + fwidth)*scale;
        if (degrees & 2) {
            if (!(degrees & 8))
                x -= strlen(text)*fwidth*scale;
            if (!(degrees & 4))
                y -= fheight*scale;
        }
        else {
            if (degrees & 8)
                x -= strlen(text)*fwidth*scale;
            if (degrees & 4)
                y -= fheight*scale;
        }
    }
    degrees = setcode(degrees);

    rgen = pc.base + x + (long) (pc.ysize-y-1)*pc.xsize;
    next = pc.xsize + fwidth*scale;
    str0 = text;
    str1 = text + strlen(text);
    while (str0 != str1) {

        if (degrees & 8) {
            if ((tx = (*--str1 & 0x7f) - ' ') < 0) tx = 0;
        }
        else
            if ((tx = (*str0++ & 0x7f) - ' ') < 0) tx = 0;

        ctab = pc.chartab + tx*pc.charheight;
        xform_cell(rotbuf,ctab,fwidth,fheight,degrees,fwidth);
        ctab = rotbuf + fheight - 1;

        ynow = y;
        for (i = fheight; i; i--, ctab--) {
            for (j = scale; j; j--) {
                if (ynow >= Yl && ynow <= Yu) {
                    xnow = x;
                    for (mask = 0x8000,ii = fwidth; ii; mask >>= 1,ii--) {
                        for (k = scale; k; k--) {
                            if (xnow >= Xl && xnow <= Xu) {
                                if (*ctab & mask) {
                                    if (rgen >= pc.base) {
                                        if (pc.alumode != 3)
                                            mfb_trash = *rgen;
                                        *rgen = c.c[0];
                                    }
                                }
                            }
                            rgen++;
                            xnow++;
                        }
                    }
                }
                else
                    rgen += fwidth*scale;
                rgen -= next;
                ynow++;
            }
        }
        if (degrees & 4) {
            y += fheight*scale;
            continue;
        }
        x += fwidth*scale;
        rgen += next1;
    }
}

#endif /* __GNUC__ */

static int
setcode(deg)

int deg;
{
    /*
     * returned bit field:
     * 0x8: load text backward
     * 0x4: rotation
     * 0x2: reverse y data
     * 0x1: reverse x data
     *
     * Assumes input bit field:
     * 0x8: mirror x
     * 0x4: mirror y
     * 0x3: 0-no rotation, 1-90, 2-180, 3-270
     */

    int mx = 0,my = 0;
    int rotn;
    int xf = 0;

    rotn = deg & 3;
    if (deg & 8) mx = 1;
    if (deg & 4) my = 1;

    if ((mx && rotn == 0) || (!mx && rotn == 2) ||
        (my && rotn == 1) || (!my && rotn == 3))
        xf |= 8;
    if (rotn & 1)
        xf |= 4;
    if ((my && (rotn == 0 || rotn == 1)) ||
        (!my && (rotn == 2 || rotn == 3)))
        xf |= 2;
    if ((mx && (rotn == 0 || rotn == 3)) ||
        (!mx && (rotn == 1 || rotn == 2)))
        xf |= 1;
    return (xf);
}


static void
xform_cell(out,in,w,h,code,shift)

unsigned short *out,*in;
int w,h,code,shift;
{
    int i, j, k, l;

    for (i = 0; i < h; i++) {
        k = ((code & 2) ? h-i-1 : i);
        if (code & 4) {
            for (j = 0,out[i] = 0; j < w; j++) {
                l = ((code & 1) ? (w-j-1) : j);
                out[i] |= ((in[j] >> k) & 1) << l;
            }
        }
        else {
            if (code & 1)
                for (j = 0,out[i] = 0; j < w; j++)
                    out[i] |= ((in[k] >> j) & 1) << (w-j-1);
            else
                 out[i] = in[k];
        }
    }
    if (shift)
        for (i = 0,j = 16-shift; i < h; i++)
            out[i] <<= j;
}
