/*************************************************************************
 MFB graphics and miscellaneous library for protected mode MSDOS.
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include "mfb.h"
#include "mfbP.h"
#include <dos.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef __STDC__
static Bool dump(long,FILE*);
#else
static Bool dump();
#endif
extern char *malloc();

static char *hmsg = "hit any key to continue";
static char hc_device[10] = {'p','r','n'};
static int  hc_port;


Bool
makehpplot()

/* print the screen on the laserjet printer */
{
    FILE *fp;
    struct stat b;
    int er;


    fp = fopen("hpplot","wb");

    if (!fp) {
        PrinterErrmsg("Error: Can't open output file.",hmsg);
        (void)MFBGetchar();
        return (true);
    }

    if (rasterize(fp)) {
        PrinterErrmsg("Error: Rasterization failed.",hmsg);
        (void)MFBGetchar();
        return (true);
    }
    fclose(fp);

    if (hc_port >= 0) {
        while (PrinterPortStatus(hc_port)) {
            PrinterErrmsg("Error: Printer is not accessible.",
                "Hit any key to continue, q to abort:");
            if ((char)MFBGetchar() == 'q')
                return (true);
        }
    }
    if (stat("hpplot",&b)) {
        PrinterErrmsg("Error: File \"hpplot\" not found.\n",hmsg);
        (void)MFBGetchar();
        return (true);
    }

    fp = fopen("hpplot","rb");

    if (!fp) {
        PrinterErrmsg("Error: File \"hpplot\" not found.\n",hmsg);
        (void)MFBGetchar();
        return (true);
    }
    er = dump((long) b.st_size,fp);
    fclose(fp);
    return (er);
}


Bool
rasterize(fp)

/* create HP Laserjet plotfile of screen */
FILE *fp;
{
    int i, j, k, kk, len, isdev = 0;
    char *rgen;
    union { unsigned short o[2]; long l; } p1;
    char cbuf, *buf, *c;
    unsigned char mask, chr;
    union REGS r;
    if (!fp) return (true);

    r.x.bx = fileno(fp);
    r.x.ax = 0x4400;
    int86(0x21,&r,&r);
    i = r.x.dx;
    if (i & 0x80) { /* output is a device */

        isdev = 1;
        /* put driver in raw mode */
        r.x.bx = fileno(fp);
        r.x.dx = 0xa0;
        r.x.ax = 0x4401;
        int86(0x21,&r,&r);
    }

    buf = malloc(pc.bytpline + 8);
    if (!buf) {
        PrinterErrmsg("Out of memory!",NULL);
        return (true);
    }

    if (pc.mfbMODE != 4)

        /* X cursor position 250
         * Y cursor position 400
         * resolution 100 dpi
         * start raster graphics at current position
         */

        fprintf(fp,"\033*p250X\033*p400Y\033*t100R\033*r1A");

    /* raster line header */
    sprintf(buf,"\033*b%dW",pc.bytpline);
    len = strlen(buf);

    if (pc.mfbMODE & 2) {
        cbuf = inp(0x3cd);
        cbuf >>= 4;
        p1.l = 0L;
        for (i = 0; i < pc.ysize; i++) {
            c = buf + len;
            mask = 0x80;
            chr = 0;
            for (j = 0; j < pc.xsize; j++) {
                if (cbuf != p1.o[1])
                    outp(0x3cd,(cbuf = p1.o[1]) << 4);
                if (*(pc.base+p1.o[0]))
                    chr |= mask;
                mask = (mask >> 1) | (mask << 7);
                if (mask & 0x80) {
                    *c++ = chr;
                    chr = 0;
                }
                p1.l++;
            }
            if (isdev) {
                for (kk = 0; kk < pc.bytpline+len; kk++)
                    putc(buf[kk],fp);
            }
            else {
                if (fwrite(buf,1,pc.bytpline+len,fp) != pc.bytpline+len) {
                    free(buf);
                    return (true);
                }
            }
        }
        if (pc.mfbMODE != 4)
            /* end raster graphics, form feed */
            fprintf(fp,"\033*rB\014");
        fflush(fp);
        free(buf);
        return (false);
    }

    rgen = pc.base;
    for (i = 0; i < pc.ysize; i++) {
        c = buf + len;
        for (j = 0; j < pc.bytpline; j++) {
            *c = '\0';
            for (k = 0; k < 4; k++) {
                outpw(0x3ce,k << 8 | 4);
                *c |= *rgen;
            }
            rgen++;
            c++;
        }
        if (isdev) {
            for (kk = 0; kk < pc.bytpline+len; kk++)
                putc(buf[kk],fp);
        }
        else {
            if (fwrite(buf,1,pc.bytpline+len,fp) != pc.bytpline+len) {
                free(buf);
                return (true);
            }
        }
    }
    if (pc.mfbMODE != 4)
        fprintf(fp,"\033*rB\014");
    fflush(fp);
    free(buf);
    return (false);
}


void
PrinterErrmsg(msg1,msg2)

char *msg1, *msg2;
{
    int x,y,w,h;
    int l,r,t,b;

    w = MFBInfo(FONTWIDTH);
    h = MFBInfo(FONTHEIGHT);
    x = MFBInfo(MAXX)/2;
    y = MFBInfo(MAXY)/2;
    l = x - 19*w;
    r = x + 19*w;
    t = y + 2*h;
    b = y - 2*h;

    MFBSetColor(0);
    MFBBox(l,b,r,t);
    MFBSetColor(1);
    MFBLine(l,b,r,b);
    MFBDrawLineTo(r,t);
    MFBDrawLineTo(l,t);
    MFBDrawLineTo(l,b);
    MFBSetColor(2);
    if (msg1)
        MFBText(msg1,l+w,t-2*h,0);
    if (msg2)
        MFBText(msg2,l+w,t-3*h,0);
}


Bool
PrinterPortStatus(which)
int which;
{
    int i;
    union REGS r;

    r.x.ax = 0x0200;
    r.x.dx = which;
    int86(0x17,&r,&r);
    i = r.x.ax;
    i >>= 8;
    /* bit 5 out of paper, bit 3 io error, bit 0 timed out */
    if (i & 0x29) return (true);
    return (false);
}


static Bool
dump(size,fp)

/* dump size bytes from fp to parallel port */
long size;
FILE *fp;
{
    int i;
    unsigned char *c;
    static char *buf;
    int bsize = 8192;
    FILE *op;
    union REGS r;

    if (!buf) buf = malloc(bsize);
    if (!buf) {
        PrinterErrmsg("Error: Out of memory.",hmsg);
        (void)MFBGetchar();
        return (true);
    }

    op = fopen(hc_device,"wb");
    if (!op) {
        sprintf(buf,"Error: Printer driver %s not found.",hc_device);
        PrinterErrmsg(buf,hmsg);
        (void)MFBGetchar();
        return (true);
    }
    PrinterErrmsg("Printing ...",NULL);

    if (bsize > size) bsize = size;

    /* put driver in raw mode */
    r.x.bx = fileno(op);
    r.x.dx = 0xa0;
    r.x.ax = 0x4401;
    int86(0x21,&r,&r);

    while (size >= bsize) {
        fread(buf,bsize,1,fp);
        for (i = 0; i < bsize; i++)
            putc(buf[i],op);
        size -= bsize;
    }
    if (size) {
        fread(buf,size,1,fp);
        for (i = 0; i < size; i++)
            putc(buf[i],op);
    }
    PrinterErrmsg("Done",NULL);
    fclose(op);
    return (false);
}


void
SetPrinterDevice(device)

char *device;
{
    strlwr(device);
    if (device) {
        if (!strcmp(device,"prn") || !strcmp(device,"lpt1"))
            hc_port = 0;
        else if (!strcmp(device,"lpt2"))
            hc_port = 1;
        else if (!strcmp(device,"lpt3"))
            hc_port = 2;
        else if (!strcmp(device,"lpt4"))
            hc_port = 3;
        else
            hc_port = -1;
        strncpy(hc_device,device,8);
        hc_device[8] = '\0';
    }
    else {
        strcpy(hc_device,"prn");
        hc_port = 0;
    }
}

