/*************************************************************************
 MFB graphics and miscellaneous library for protected mode MSDOS.
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

/* Microsoft mouse (or compatible) interface. */

#include "mfb.h"
#include "mfbP.h"
#include <dos.h>


void
ms_start()

/* initialize mouse pointing device */
{
    union REGS r;

    if (!ms_hreset()) {
        fprintf(stderr,"Error: Mouse driver not loaded.\n");
        fprintf(stderr,"Using keyboard arrow keys.  Hit any key to continue.\n");
        (void)MFBGetchar();
        pc.pointertype = KBRD;
        MFBSetColor(0);
        MFBFlood();
        MFBPointerInit();
    }
    /* This driver is designed to work in "strange" super-VGA modes.
     * The mouse driver thinks these are text mode, and quantizes the
     * returned mouse position in units of 8.  To avoid this, the logical
     * screen size is multiplied by 8, and the return coordinates
     * ( in ms_status() ) are divided by 8.
     */

    /* set logical screen size */
    r.x.ax = 7;
    r.x.cx = 0;
    r.x.dx = (pc.xsize-1) << 3;
    int86(0x33,&r,&r);
    r.x.ax = 8;
    r.x.cx = 0;
    r.x.dx = (pc.ysize-1) << 3;
    int86(0x33,&r,&r);

    /* set mickeys/pixels ratio (1/8 default) */
    r.x.ax = 0xf;
    r.x.cx = 1;
    r.x.dx = 2;

    /* move pointer to center of logical screen */
    int86(0x33,&r,&r);
    r.x.ax = 0x4;
    r.x.cx = (pc.xsize-1) << 2;
    r.x.dx = (pc.ysize-1) << 2;
    int86(0x33,&r,&r);
}


void
ms_end()

/* end mouse pointing device activity */
{
    (void)ms_hreset();
}


int
ms_hreset()
/* reset mouse, return number of buttons, or 0 if error */
{
    union REGS r;

    r.x.ax = 0x0;
    int86(0x33,&r,&r);
    if (r.x.ax == 0xffff)
        return (r.x.bx);
    return (0);
}


void
ms_pointer_on()
/* display mouse pointer */
{
    union REGS r;

    /* don't use!
    r.x.ax = 0x1;
    int86(0x33,&r,&r);
    */
}


void
ms_pointer_off()
/* hide mouse pointer */
{
    union REGS r;

    r.x.ax = 0x2;
    int86(0x33,&r,&r);
}


void
ms_status(x,y,b)

/* get mouse position and button status */
int *x, *y, *b;
{
    union REGS r;

    r.x.ax = 0x3;
    int86(0x33,&r,&r);
    /* physical screen = 1/8 logical screen */
    *x = r.x.cx >> 3;
    *y = r.x.dx >> 3;
    *b = r.x.bx;
}


void
ms_point(x,y,k,b)
int *x, *y, *k, *b;

{
    int xx, yy, bb;
    int xlst, ylst;

    ms_status(&xx,&yy,&bb);
    yy = pc.ysize - yy - 1;
    MFBDrawCursor(xx-8,yy+7);
    for (;;) {
        if (kbhit()) {
            *k = 0xff & getch();
            if (!*k)
                *k = 256 + getch();
            *b = 0;
            break;
        }
        if (bb) {
            *k = 0;
            *b = bb;
            break;
        }
        xlst = xx;
        ylst = yy;
        ms_status(&xx,&yy,&bb);
        yy = pc.ysize - yy - 1;
        if (xx == xlst && yy == ylst)
            continue;
        MFBEraseCursor(xlst-8,ylst+7);
        MFBDrawCursor(xx-8,yy+7);
    }
    *x = xx;
    *y = yy;
    MFBEraseCursor(xx-8,yy+7);
}
