/*************************************************************************
 MFB graphics and miscellaneous library
 Copyright (c) Stephen R. Whiteley 1992
 Author: Stephen R. Whiteley
 *************************************************************************/

#include <dos.h>

#if __GNUC__
#define inp inportb
#define outp outportb
#define outpw outportw
#endif

void
vga_get_color(num, red, green, blue)

/* get DAC register */
int num, *red, *green, *blue;
{
    if (num > 255 || num < 0) return;
    outp(0x3c7,num);                 /* write index */
    while (!(inp(0x3da) & 0x8)) ;    /* wait for V retrace */
    *red   = inp(0x3c9);             /* read rgb */
    *green = inp(0x3c9);
    *blue  = inp(0x3c9);
}


void
vga_set_color(num, red, green, blue)

/* set DAC register */
int num, red, green, blue;
{
    if (num > 255 || num < 0) return;
    outp(0x3c8,num);                 /* write index */
    while (!(inp(0x3da) & 0x8)) ;    /* wait for V retrace */
    outp(0x3c9,red);                 /* write rgb */
    outp(0x3c9,green);
    outp(0x3c9,blue);
}


void
vga_get_cblock(p,i,n)

/* get block of DAC registers */
char (*p)[3];
int i, n;
{
    int j;

    if (i+n > 256 || i < 0 || n < 0) return;
    outp(0x3c4,1);             /* sequencer clock mode */
    outp(0x3c5,0x21);          /* blank the display */
    outp(0x3c7,i);             /* write index */
    for (j = 0; j < n; j++) {
        p[j][0] = inp(0x3c9);  /* read rgb */
        p[j][1] = inp(0x3c9);
        p[j][2] = inp(0x3c9);
    }
    outp(0x3c5,0x1);           /* turn display on */
}


void
vga_set_cblock(p,i,n)
/* set block of DAC registers */

char (*p)[3];
int i, n;
{
    int j;

    if (i+n > 256 || i < 0 || n < 0) return;
    outp(0x3c4,1);             /* sequencer clock mode */
    outp(0x3c5,0x21);          /* blank the display */
    outp(0x3c8,i);             /* write index */
    for (j = 0; j < n; j++) {
        outp(0x3c9,p[j][0]);   /* write rgb */
        outp(0x3c9,p[j][1]);
        outp(0x3c9,p[j][2]);
    }
    outp(0x3c5,0x1);           /* turn display on */
}


void
vga_get_cpblock(p)

/* get 16 entry palette register block */
char *p;
{
    int j;

    while (!(inp(0x3da) & 0x8)) ;  /* wait for V retrace    */
    for (j = 0; j < 16; j++) {
        outp(0x3c0,j);             /* index, enable xfer    */
        p[j] =  inp(0x3c1);        /* read data             */
        inp(0x3da);                /* reset to address mode */
        outp(0x3c0,j | 0x20);      /* index, enable EGA     */
        inp(0x3da);                /* reset to address mode */
    }
}


void
vga_set_cpblock(p)

/* set 16 entry palette register block */
char *p;
{
    int j;
    
    while (!(inp(0x3da) & 0x8)) ;  /* wait for V retrace    */
    for (j = 0; j < 16; j++) {
        outp(0x3c0,j);             /* index, enable xfer    */
        outp(0x3c0,p[j]);          /* write data            */
        inp(0x3da);                /* reset to address mode */
        outp(0x3c0,j | 0x20);      /* index, enable EGA     */
        inp(0x3da);                /* reset to address mode */
    }
}
