/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * For dealing with input decks and command scripts
 */

#include "spice.h"
#include "ftedefs.h"
#include "fteinp.h"

#define STRGROW 256


char *
readline(fd)

/*  This routine reads a line (of arbitrary length), up to a '\n' or 'EOF'
 *  and returns a pointer to the resulting null terminated string.
 *  The '\n' if found, is included in the returned string.
 *  From: jason@ucbopal.BERKELEY.EDU (Jason Venner)
 *  Newsgroups: net.sources
 */
FILE *fd;
{
    int c;
    int memlen;
    char *strptr;
    int strlen;
    
    strptr = NULL;
    strlen = 0;
    memlen = STRGROW; 
    strptr = tmalloc(memlen);
    memlen -= 1;          /* Save constant -1's in while loop */
    while((c = getc(fd)) != EOF) {
        strptr[strlen] = c;
        strlen++;
        if( strlen >= memlen ) {
            memlen += STRGROW;
            if( !(strptr = trealloc(strptr, memlen + 1))) {
                return (NULL);
            }
        }
        if (c == '\n') {
            break;
        }
    }
    if (!strlen) {
        txfree(strptr);
        return (NULL);
    }
    strptr[strlen] = '\0'; 
    /* Trim the string */
    strptr = trealloc(strptr, strlen + 1);
    return (strptr);
}


FILE *
inp_pathopen(name, mode)

/* Look up the variable sourcepath and try everything in the list in order
 * if the file isn't in . and it isn't an abs path name.
 */
char *name, *mode;
{
    FILE *fp;
    char buf[BSIZE_SP];
    struct variable *v = NULL;
    extern char *kw_sourcepath;

    /* If this is an abs pathname, or there is no sourcepath var, just
     * do an fopen.
     */
    if (strchr(name, DIR_TERM) || strchr(name,'/')
            || !cp_getvar(kw_sourcepath, VT_LIST, (char *) &v)) {
        strcpy(buf,name);
        cp_pathfix(buf);
        return (fopen(buf, mode));
    }

    while (v) {
        switch (v->va_type) {
            case VT_STRING:
                cp_wstrip(v->va_string);
                (void) sprintf(buf, "%s/%s", v->va_string, name);
                break;
            case VT_NUM:
                (void) sprintf(buf, "%d/%s", v->va_num, name);
                break;
            case VT_REAL:   /* This is foolish */
                (void) sprintf(buf, "%lg/%s", v->va_real, name);
                break;
        }
        cp_pathfix(buf);
        if (fp = fopen(buf, mode))
            return (fp);
        v = v->va_next;
    }
    return (NULL);
}


void
inp_readall(fp, data, title)

/* Read the entire input file and return  a pointer to the first line of 
 * the linked list of 'card' records in data.
 */
FILE *fp;
struct line **data;
char *title;
{
    struct line *cc, *end = NULL, *prev = NULL, *working, *newcard;
    char *buffer, *s, *t;
    int line = 1;
    FILE *newfp;

    for (;;) {
        if (title) {
            buffer = title;
            title = NULL;
        }
        else if ((buffer = readline(fp)) == NULL)
            break;

        if (*buffer == '@')
            break;

        /* Zap the newline. */
        for (s = buffer; *s && (*s != '\n'); s++)
            ;
        if (!*s) {
            fprintf(cp_err, "Warning: premature EOF\n");
        }
        *s = '\0';

        /* strip any leading white space */
        s = buffer;
        while (*s == ' ' || *s == '\t')
            s++;
        if (s != buffer) {
            t = buffer;
            while (*s)
                *t++ = *s++;
            *t = '\0';
        }

        if (prefix(".include", buffer)) {
            for (s = buffer; *s && !isspace(*s); s++)
                ;
            while (isspace(*s))
                s++;
            if (!*s) {
                fprintf(cp_err, 
                "Error: .include filename missing\n");
                continue;
            }
            for (t = s; *t && !isspace(*t); t++)
                ;
            *t = '\0';
            if (*s == '~')
                s = cp_tildexpand(s);
            if (!(newfp = inp_pathopen(s, "r"))) {
                perror(s);
                continue;
            }
            inp_readall(newfp, &newcard,NULL);
            (void) fclose(newfp);

            /* Make the .include a comment */
            *buffer = '*';
            if (end) {
                end->li_next = alloc(struct line);
                end = end->li_next;
            }
            else {
                end = cc = alloc(struct line);
            }
            end->li_line = copy(buffer);
            end->li_linenum = line++;
            end->li_next = newcard;

            /* Renumber the cards */
            for (end = newcard; end && end->li_next; end = end->li_next)
                end->li_linenum = line++;

            /* Fix the buffer up a bit. */
            (void) strncpy(buffer + 1, "end of:", 7);
        }

        if (end) {
            end->li_next = alloc(struct line);
            end = end->li_next;
        }
        else {
            end = cc = alloc(struct line);
        }
        end->li_line = buffer;
        end->li_linenum = line++;
    }
    if (!end) { /* No stuff here */
        *data = NULL;
        return;
    }

    /* Now make logical lines. */
    working = cc->li_next;      /* Skip title. */

    while (working) {
        switch (*working->li_line) {
            case '*':
            case '\0':
                prev = NULL;
                working = working->li_next;
                break;
            case '+':
                if (!prev) {
                    working->li_error = copy(
                    "Illegal continuation card: ignored.");
                    working = working->li_next;
                    break;
                }
                buffer = tmalloc(strlen(prev->li_line) + 
                        strlen(working->li_line) + 2);
                (void) sprintf(buffer, "%s %s", prev->li_line,
                        working->li_line + 1);
                s = prev->li_line;
                prev->li_line = buffer;
                prev->li_next = working->li_next;
                working->li_next = NULL;
                if (prev->li_actual) {
                    for (end = prev->li_actual;
                        end->li_next; end = end->li_next)
                        ;
                    end->li_next = working;
                    tfree(s);
                }
                else {
                    newcard = alloc(struct line);
                    newcard->li_linenum = prev->li_linenum;
                    newcard->li_line = s;
                    newcard->li_next = working;
                    prev->li_actual = newcard;
                }
                working = prev->li_next;
                break;
            default:
                prev = working;
                working = working->li_next;
                break;
        }
    }

    *data = cc;
    return;
}


void
inp_casefix(string)

char *string;
{
    if (string)
        while (*string) {
            *string = strip(*string);
            if (!isspace(*string) && !isprint(*string))
                *string = '_';
            if (isupper(*string))
                *string = tolower(*string);
            string++;
        }
    return;
}


void
inp_deckfree(deck)

struct line *deck;
{
    struct line *d, *dd;

    d = deck;
    while (d) {
        tfree(d->li_line);
        tfree(d->li_error);
        if (d->li_actual) inp_deckfree(d->li_actual);
        dd = d;
        d = d->li_next;
        tfree(dd);
    }
}


struct line *
inp_deckcopy(deck)

/* Copy a deck, including the actual lines. */
struct line *deck;
{
    struct line *d = NULL, *nd = NULL;

    while (deck) {
        if (nd) {
            d->li_next = alloc(struct line);
            d = d->li_next;
        }
        else
            nd = d = alloc(struct line);
        d->li_linenum = deck->li_linenum;
        d->li_line = copy(deck->li_line);
        if (deck->li_error)
            d->li_error = copy(deck->li_error);
        d->li_actual = inp_deckcopy(deck->li_actual);
        deck = deck->li_next;
    }
    return (nd);
}
