/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "ftedefs.h"
#include "ftehelp.h"
#include "hlpdefs.h"
#include "scedio.h"

#ifdef __STDC__
static int hcomp(struct comm**,struct comm**);
static void byemesg(void);
#else
static int hcomp();
static void byemesg();
#endif

#define SHORTHELPMSG "For a complete description of %s read the %s manual.\n"

/* Some C compilers can't handle strings longer than 256 characters */

#define AHELPMESG1 \
"For a complete description of %s read the %s manual.\n\
There are short introductions to both Spice 3 and Nutmeg available, in\n\
addition to the SPICE User's Guide and the manual pages.  If you don't\n"

#define AHELPMESG2 \
"have these documents ask your system manager for them.  Below is a short\n\
list of useful commands -- for a list of all commands type \"help all\",\n\
and for a short description of \"command\", type \"help command\".\n"


void
com_help(wl)

wordlist *wl;
{
    struct comm *c;
    struct comm *ccc[512];  /* Should be enough. */
    int numcoms, i;
    bool allflag = false, tmpmm;

    if (wl && eq(wl->wl_word, "all")) {
        allflag = true;
        wl = NULL;  /* XXX Probably right */
    }

    /* We want to use more mode whether "moremode" is set or not. */
    tmpmm = out_moremode;
    out_moremode = true;
    out_init();
    out_moremode = tmpmm;
    if (wl == NULL) {
            if (allflag)
                out_printf(SHORTHELPMSG, cp_program, cp_program);
            else {
                out_printf(AHELPMESG1, cp_program, cp_program);
                out_printf(AHELPMESG2, cp_program, cp_program);
            }

        /* Sort the commands */
        for (numcoms = 0; cp_coms[numcoms].co_func != NULL; numcoms++)
            ccc[numcoms] = &cp_coms[numcoms];
        qsort((char *) ccc, numcoms, sizeof (struct comm *),
#ifdef __STDC__
        (int(*)(const void*,const void*))hcomp);
#else
        hcomp);
#endif

        for (i = 0; i < numcoms; i++) {
            if ((ccc[i]->co_spiceonly && ft_nutmeg) || 
                    (ccc[i]->co_help == NULL) || 
                    (!allflag && !ccc[i]->co_major))
                continue;
            out_printf("%s ", ccc[i]->co_comname);
            out_printf(ccc[i]->co_help, cp_program);
            out_send("\n");
        }
    }
    else {
        while (wl != NULL) {
            for (c = &cp_coms[0]; c->co_func != NULL; c++)
                if (eq(wl->wl_word, c->co_comname)) {
                    out_printf("%s ", c->co_comname);
                    out_printf(c->co_help, cp_program);
                    if (c->co_spiceonly && ft_nutmeg)
                        out_send(
                        " (Not available in nutmeg)");
                    out_send("\n");
                    break;
                }
            if (c->co_func == NULL) {
                /* See if this is aliased. */
                struct alias *al;

                for (al = cp_aliases; al; al = al->al_next)
                    if (eq(al->al_name, wl->wl_word))
                        break;
                if (al == NULL)
                    out_printf("Sorry, no help for %s.\n", 
                        wl->wl_word);
                else {
                    out_printf("%s is aliased to ",
                        wl->wl_word);
                    /* Minor badness here... */
                    out_wlprint(al->al_text);
                    out_send("\n");
                }
            }
            wl = wl->wl_next;
        }
    }
    out_send("\n");
    return;
}


void
com_ahelp(wl)

wordlist *wl;
{

    int i, n;
    /* assert: number of commands must be less than 512 */
    struct comm *cc[512];
    int env = 0;
    struct comm *com;
    int level;
    char slevel[256];
    extern char *kw_level;

    if (wl) {
      com_help(wl);
      return;
    }

    out_init();

    /* determine environment */
    if (plot_list->pl_next) {   /* plots load */
      env |= E_HASPLOTS;
    } else {
      env |= E_NOPLOTS;
    }

    /* determine level */
    if (cp_getvar(kw_level, VT_STRING, slevel)) {
      switch (*slevel) {
        case 'b':   level = 1;
            break;
        case 'i':   level = 2;
            break;
        case 'a':   level = 4;
            break;
        default:    level = 1;
            break;
      }
    } else {
      level = 1;
    }

    /* Some C compilers can't handle strings longer than 256 characters */
    out_printf(AHELPMESG1, cp_program, cp_program);
    out_printf(AHELPMESG2, cp_program, cp_program);

    /* sort the commands */
    for (n = 0; cp_coms[n].co_func != (void (*)()) NULL; n++) {
      cc[n] = &cp_coms[n];
    }
    qsort((char *) cc, n, sizeof(struct comm *),
#ifdef __STDC__
        (int(*)(const void*,const void*))hcomp);
#else
        hcomp);
#endif

    /* filter the commands */
    for (i=0; i< n; i++) {
      com = cc[i];
      if ((com->co_env < (level << 13)) && (!(com->co_env & 4095) ||
        (env & com->co_env))) {
        if ((com->co_spiceonly && ft_nutmeg) ||
        (com->co_help == (char *) NULL)) {
          continue;
        }
        out_printf("%s ", com->co_comname);
        out_printf(com->co_help, cp_program);
        out_send("\n");
      }
    }

    out_send("\n");

    return;

}

void
com_ghelp(wl)
    wordlist *wl;
{
    char *opath, *path = Help_Path, buf[BSIZE_SP];
    int i;

    if (cp_getvar("helppath", VT_STRING, buf))
        path = buf;
        path = cp_tildexpand(opath = path);
    if (path == NULL || *path == '\0') {
        fprintf(cp_err, "Note: can't find help dir %s\n", opath);
        fprintf(cp_err, "Defaulting to old help.\n\n");
        com_help(wl);
        return;
    }
    if (cp_getvar("helpinitxpos", VT_NUM, (char *) &i))
        hlp_initxpos = i;
    if (cp_getvar("helpinitypos", VT_NUM, (char *) &i))
        hlp_initypos = i;
    hlp_main(path, wl);
    tfree(path);
    return;
}


static int
hcomp(c1, c2)

struct comm **c1, **c2;
{
    return (strcmp((*c1)->co_comname, (*c2)->co_comname));
}


/* ARGSUSED */
void
com_quit(wl)

wordlist *wl;
{
    struct circ *cc;
    struct plot *pl;
    int ncc = 0, npl = 0;
    char buf[64];
    char *zz = "Are you sure you want to quit (yes)? ";
    bool noask;

    (void) cp_getvar("noaskquit", VT_BOOL, (char *) &noask);
    cp_ccon(false);
    
    /* Make sure the guy really wants to quit. */
    if (!ft_nutmeg && !noask) {
        for (cc = ft_circuits; cc; cc = cc->ci_next)
            if (cc->ci_inprogress)
                ncc++;
        for (pl = plot_list; pl; pl = pl->pl_next)
            if (!pl->pl_written && pl->pl_hashtab)
                npl++;
        if (ncc || npl) {
            out_init();
            out_printf("Warning: ");
            if (ncc) {
                out_printf( 
            "the following simulation%s still in progress:\n",
                        (ncc > 1) ? "s are" : " is");
                for (cc = ft_circuits; cc; cc = cc->ci_next)
                    if (cc->ci_inprogress)
                        out_printf("\t%s\n", cc->ci_name);
            }
            if (npl) {
                if (ncc)
                    out_printf("and ");
                out_printf( 
                "the following plot%s been saved:\n",
                    (npl > 1) ? "s haven't" : " hasn't");
                for (pl = plot_list; pl; pl = pl->pl_next)
                    if (!pl->pl_written && pl->pl_hashtab)
                        out_printf("%s\t%s, %s\n",
                                pl->pl_typename,
                                pl->pl_title,
                                pl->pl_name);
            }
            (void) SCEDfgets(buf,BSIZE_SP,cp_in,zz);

            if ((*buf == 'y') || (*buf == 'Y') || (*buf == '\n'))
                byemesg();
            else {
                return;
            }
        }
        else
            byemesg();
    }
    else
        byemesg();

    exit(EXIT_NORMAL);
    /* NOTREACHED */
}


#ifdef SYSTEM_MAIL

/* ARGSUSED */
void
com_bug(wl)

wordlist *wl;
{
    char buf[BSIZE_SP];

    if (!Bug_Addr || !*Bug_Addr) {
        fprintf(cp_err, "Error: No address to send bug reports to.\n");
            return;
    }
    if (SCEDactive()) {
        ShowPrompt("Exit sced to send bug report.");
        return;
    }

    fprintf(cp_out, "Calling the mail program . . .(sending to %s)\n\n",
            Bug_Addr);
    fprintf(cp_out,
            "Please include the OS version number and machine architecture.\n");
    fprintf(cp_out,
            "If the problem is with a specific circuit, please include the\n");
    fprintf(cp_out, "input file.\n");

    (void) sprintf(buf, SYSTEM_MAIL, ft_sim->simulator,
            ft_sim->version, Bug_Addr);
    (void) system(buf);
    fprintf(cp_out, "Bug report sent.  Thank you.\n");
    return;
}

#else

/* ARGSUSED */
void
com_bug(wl)

wordlist *wl;
{
    char buf[BSIZE_SP], *s;
    FILE *fp;

    if (SCEDactive()) {
        ShowPrompt("Exit sced to send bug report.");
        return;
    }
    if (News_File && *News_File) {
        strcpy(buf,News_File);
        s = strrchr(buf,DIR_TERM);
        if (s != NULL) {
            s++;
            strcpy(s,"bug.log");
            if (access(buf,0)) {
                if ((fp = fopen(buf,"w")) != NULL) {
                    fprintf(fp,"SPICE and NUTMEG bug log file: %s\n",buf);
                    fclose(fp);
                }
            }
            inp_edit(buf);
        }
    }
}

#endif

void
com_version(wl)

wordlist *wl;
{
    char *s;

    out_init();
    if (!wl) {
        out_printf("Program: %s, version: %s\n", ft_sim->simulator,
                ft_sim->version);
            if (Spice_Notice && *Spice_Notice)
                out_printf("\t%s\n", Spice_Notice);
            if (Spice_Build_Date && *Spice_Build_Date)
                out_printf("Date built: %s\n", Spice_Build_Date);
    }
    else {
        s = wl_flatten(wl);
        if (!eq(ft_sim->version, s)) {
            fprintf(stderr,
        "Note: rawfile is version %s (current version is %s)\n",
                    wl->wl_word, ft_sim->version);
        }
        tfree(s);
    }
    return;
}


static void
byemesg()
{
    out_init();
    out_printf("%s-%s done\n", ft_sim->simulator, ft_sim->version);
    return;
}
