/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * This routine parses a number.
 */

#include "spice.h"
#include "ftedefs.h"

bool ft_strictnumparse = false;

/* Parse a number. This will handle things like 10M, etc... If the number
 * must not end before the end of the string, then whole is true.
 * If whole is false and there is more left to the number, the argument
 * is advanced to the end of the word. Returns NULL
 * if no number can be found or if there are trailing characters when
 * whole is true.
 * If ft_strictnumparse is true, and whole is false, the first of the
 * trailing characters must be a '_'.
 */

double *
ft_numparse(s, whole)

char **s;
bool whole;
{
    double mant, expo, tmpn;
    bool sign = false, exsign = false;
    static double num;
    char *string = *s;

    /* See if the number begins with + or -. */
    if (*string == '+')
        string++;
    else if (*string == '-') {
        string++;
        sign = true;
    }

    /* We don't want to recognise "P" as 0P, or .P as 0.0P... */
    if ((!isdigit(*string) && *string != '.') ||
            ((*string == '.') && !isdigit(string[1])))
        return (NULL);

    /* Now accumulate a number. Note ascii dependencies here... */
    if (isdigit(*string)) {
        mant = (*string++ - '0');
        while (isdigit(*string))
            mant = mant * 10.0 + (*string++ - '0');
    }
    else
        mant = 0.0;
    
    /* Now maybe a decimal point. */
    if (*string == '.') {
        string++;
        tmpn = 1.0;
        while (isdigit(*string))
            mant += (*string++ - '0') * (tmpn *= 0.1);
    }

    /* Now look for the scale factor or the exponent (can't have both). */
    expo = 1.0;
    switch (*string) {
        case 'e':
        case 'E':
            /* Parse another number. */
            string++;
            if (*string == '+') {
                string++;
            }
            else if (*string == '-') {
                exsign = true;
                string++;
            }
            if (isdigit(*string)) {
                expo = (*string++ - '0');
                while (isdigit(*string))
                    expo = expo * 10.0 + (*string++ - '0');
            }
            else
                expo = 0.0;
            if (*string == '.') {
                string++;
                tmpn = 1.0;
                while (isdigit(*string))
                    expo += (*string++ - '0') * (tmpn *= 0.1);
            }
            if (exsign)
                expo = pow(10.0,-expo);
            else
                expo = pow(10.0,expo);
            break;
        case 't':
        case 'T':
            expo = 1e12;
            string++;
            break;
        case 'g':
        case 'G':
            expo = 1e9;
            string++;
            break;
        case 'k':
        case 'K':
            expo = 1e3;
            string++;
            break;
        case 'u':
        case 'U':
            expo = 1e-6;
            string++;
            break;
        case 'n':
        case 'N':
            expo = 1e-9;
            string++;
            break;
        case 'p':
        case 'P':
            expo = 1e-12;
            string++;
            break;
        case 'f':
        case 'F':
            expo = 1e-15;
            string++;
            break;
        case 'm':
        case 'M':
            /* Can be either m, mil, or meg. */
            if (string[1] && string[2] && 
                ((string[1] == 'e') || (string[1] == 'E')) &&
                ((string[2] == 'g') || (string[2] == 'G'))) {
                expo = 1e6;
                string += 3;
            }
            else if (string[1] && string[2] && 
                ((string[1] == 'i') || (string[1] == 'I')) &&
                ((string[2] == 'l') || (string[2] == 'L'))) {
                expo = 1e-6;
                mant *= 25.4;
                string += 3;
            }
            else {
                expo = 1e-3;
                string++;
            }
            break;
    }

    if (whole && *string != '\0') {
        return (NULL);
    }
    else if (ft_strictnumparse && *string && isdigit(string[-1])) {
        if (*string == '_')
            while (isalpha(*string) || (*string == '_'))
                string++;
        else
            return (NULL);
    }
    else {
        while (isalpha(*string) || (*string == '_'))
            string++;
    }
    *s = string;
    if (sign)
        num = -mant * expo;
    else
        num = mant * expo;

    if (ft_parsedb)
        fprintf(cp_err, "numparse: got %e, left = %s\n", num, *s);
    return (&num);
}
