/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Table of available commands.  Note that they're sorted so that the commands
 * that appear in the spiceinit file are at the top.
 */

#include "spice.h"
#include "ftedefs.h"
#include "ftehelp.h"
#include "ftegraph.h"
#include "plotext.h"

/* Bool fields:     stringargs, spiceonly, major */

struct comm nutcp_coms[] = {
    { "let", com_let, false, false, true,
      { 040000, 040000, 040000, 040000 }, E_DEFHMASK, 0, LOTS,
      arg_let,
      "varname = expr : Assign vector variables." } ,
    { "define", com_define, false, false, true,
      { 010000, 040000, 040000, 040000 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[[func (args)] stuff] : Define a user-definable function." } ,
    { "set", com_set, false, false, true,
      { 020000, 020000, 020000, 020000 }, E_DEFHMASK, 0, LOTS,
      arg_set,
      "[option] [option = value] ... : Set a variable." } ,
    { "alias", com_alias, false, false, false,
      { 02, 04, 04, 04 }, E_ADVANCED, 0, LOTS,
      (int (*)()) NULL,
      "[[word] alias] : Define an alias." } ,
    { "deftype", com_dftype, false, false, false,
      { 0, 0, 0, 0 }, E_DEFHMASK, 3, LOTS,
      (int (*)()) NULL,
      "spec name pat ... : Redefine vector and plot types.\n" } ,
    { "plot", com_plot, false, false, true,
      { 041000, 041000, 041000, 041000 }, E_BEGINNING | E_HASPLOTS, 1, LOTS,
      arg_plot,
      "expr ... [vs expr] [xl xlo xhi] [yl ylo yhi] : Plot things." },


    /* SRW */
    { "graf", com_graf, false, false, true,
      { 040000, 040000, 040000, 040000 }, E_BEGINNING | E_HASPLOTS, 0, LOTS,
      (int (*)()) NULL,
      "[options] expr ... : Plot vectors." },
    { "usrset", com_usrset, false, false, false,
      { 020000, 020000, 020000, 020000 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      ": Display list of option keywords." } ,
    { "mplot", com_mplot, false, false, false,
      {1, 1, 1, 1 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      ": plot results from margin analysis." } ,
    { "setrdb", com_setrdb, false, false, false,
      { 0, 0, 0, 0 }, E_DEFHMASK, 1, 2,
      (int (*)()) NULL,
      "resource: value : set resources for X windows." } ,

    { "display", com_display, false, false, true,
      { 040000, 040000, 040000, 040000 }, E_BEGINNING, 0, LOTS,
      arg_display,
      ": Display vector status." } ,
    { "destroy", com_destroy, false, false, false,
      { 0400, 0400, 0400, 0400 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[plotname] ... : Throw away all the data in the plot." } ,
    { "setplot", com_splot, false, false, true,
      { 0400, 0, 0, 0 }, E_DEFHMASK, 0, 1,
      (int (*)()) NULL,
      "[plotname] : Change the current working plot." } ,
    { "xgraph", com_xgraph, false, false, true,
      { 1, 041000, 041000, 041000 }, E_DEFHMASK, 1, LOTS,
      (int (*)()) NULL,
      "file plotargs : Send plot to Xgraph-11." } ,
    { "hardcopy", com_hardcopy, false, false, true,
      { 1, 041000, 041000, 041000 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "file plotargs : Produce hardcopy plots." } ,
    { "asciiplot", com_asciiplot, false, false, true,
      { 041000, 041000, 041000, 041000 }, E_DEFHMASK, 1, LOTS,
      (int (*)()) NULL,
      "plotargs : Produce ascii plots." } ,
    { "write", com_write, false, false, true,
      { 1, 040000, 040000, 040000 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "file expr ... : Write data to a file." } ,
    { "compose", com_compose, false, false, false,
      { 0, 0, 0, 0 }, E_DEFHMASK, 2, LOTS,
      (int (*)()) NULL,
      "var parm=val ... : Compose a vector." } ,
    { "unlet", com_unlet, false, false, false,
      { 040000, 040000, 040000, 040000 }, E_DEFHMASK, 1, LOTS,
      (int (*)()) NULL,
      "varname ... : Undefine vectors." } ,
    { "print", com_print, false, false, true,
      { 040000, 040000, 040000, 040000 }, E_BEGINNING, 1, LOTS,
      arg_print,
      "[col] expr ... : Print vector values." } ,
    { "load", com_load, false, false, true,
      { 1, 1, 1, 1 }, E_BEGINNING | E_NOPLOTS, 1, LOTS,
      arg_load,
      "file ... : Load in data." } ,
    { "cross", com_cross, false, false, true,
      { 040000, 0, 040000, 040000 }, E_DEFHMASK, 2, LOTS,
      (int (*)()) NULL,
      "vecname number [ vector ... ] : Make a vector in a strange way." } ,
    { "undefine", com_undefine, false, false, false,
      { 010000, 010000, 010000, 010000 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[func ...] : Undefine a user-definable function." } ,

    { "edit", nutcom_edit, false, true, true,
      { 1, 0, 0, 0 }, E_DEFHMASK, 1, LOTS,
      (int (*)()) NULL,
      "filename : Edit a file." } ,

    { "fourier", com_fourier, false, false, true,
      { 0, 040000, 040000, 040000 }, E_DEFHMASK, 1, LOTS,
      (int (*)()) NULL,
      "fund_freq vector ... : Do a fourier analysis of some data." } ,

#ifdef notdef
    { "aspice", com_aspice, false, false, false,
      { 1, 1, 1, 1 }, E_DEFHMASK, 1, 2,
      (int (*)()) NULL,
      "file [outfile] : Run a spice job asynchronously." } ,
    { "jobs", com_jobs, false, false, false,
      { 0, 0, 0, 0 }, E_DEFHMASK, 0, 0,
      (int (*)()) NULL,
      ": Report on asynchronous spice jobs." } ,
    { "rspice", com_rspice, false, false, false,
      { 1, 1, 1, 1 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[input file] : Run a spice job remotely." } ,
#endif
    { "bug", com_bug, false, false, true,
      { 0, 0, 0, 0 }, E_DEFHMASK, 0, 0,
      (int (*)()) NULL,
      ": Report a %s bug." } ,
    { "newhelp", com_ahelp, false, false, true,
      { 010, 010, 010, 010 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[command name] ... : help." },
    { "tutorial", com_ghelp, false, false, true,
      { 023010, 023010, 023010, 023010 }, E_BEGINNING, 0, LOTS,
      (int (*)()) NULL,
      "[subject] ... : Hierarchical documentation browser." } ,
    { "help", com_ghelp, false, false, true,
      { 023010, 023010, 023010, 023010 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[subject] ... : Hierarchical documentation browser." } ,
    { "oldhelp", com_help, false, false, true,
      { 010, 010, 010, 010 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[command name] ... : Print help." } ,
    { "quit", com_quit, false, false, true,
      { 0, 0, 0, 0 }, E_BEGINNING, 0, 0,
      (int (*)()) NULL,
      ": Quit %s." } ,

    { "source", nutcom_source, false, false, true,
      { 1, 1, 1, 1 }, E_DEFHMASK, 1, LOTS,
      (int (*)()) NULL,
      "file : Source a %s file." } ,

    { "shift", com_shift, false, false, false,
      { 020000, 0, 0, 0 }, E_DEFHMASK, 0, 2,
      (int (*)()) NULL,
      "[var] [number] : Shift argv or the named list var to the left." } ,
    { "unset", com_unset, false, false, false,
      { 020000, 020000, 020000, 020000 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "varname ... : Unset a variable." } ,
    { "unalias", com_unalias, false, false, false,
      { 02, 02, 02, 02 }, E_DEFHMASK, 1, LOTS,
      (int (*)()) NULL,
      "word ... : Undefine an alias." } ,
    { "history", com_history, false, false, false,
      { 0, 0, 0, 0 }, E_DEFHMASK, 0, 2,
      (int (*)()) NULL,
      "[-r] [number] : Print command history." } ,
    { "echo", com_echo, false, false, false,
      { 1, 1, 1, 1 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[stuff ...] : Print stuff." } ,
    { "shell", com_shell, false, false, true,
      { 1, 1, 1, 1 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[args] : Fork a shell, or execute the command." } ,
    { "rusage", com_rusage, false, false, false,
      { 02000, 02000, 02000, 02000 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[resource ...] : Print current resource usage." } ,
    { "cd", com_chdir, false, false, false,
      { 1, 0, 0, 0 }, E_DEFHMASK, 0, 1,
      (int (*)()) NULL,
      "[directory] : Change working directory." } ,
    { "version", com_version, false, false, false,
      { 0, 0, 0, 0 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "[number] : Print the version number." } ,
    { "diff", com_diff, false, false, false,
      { 0400, 0400, 040000, 040000 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      "plotname plotname [vec ...] : 'diff' two plots." } ,
    { "rehash", com_rehash, false, false, false,
      { 0, 0, 0, 0 }, E_DEFHMASK, 0, 0,
      (int (*)()) NULL,
      ": Rebuild the unix command database." } ,
    { "cdump", com_cdump, false, false, false,
      { 0, 0, 0, 0 }, E_DEFHMASK, 0, 0,
      (int (*)()) NULL,
      ": Dump the current control structures." } ,
    { "settype", com_stype, false, false, false,
      { 0200000, 040000, 040000, 040000 }, E_DEFHMASK, 2, LOTS,
      (int (*)()) NULL,
      "type vec ... : Change the type of a vector." } ,
    { "strcmp", com_strcmp, false, false, false,
      { 0, 0, 0, 0 }, E_DEFHMASK, 3, 3,
      (int (*)()) NULL,
      "varname s1 s2 : Set $varname to strcmp(s1, s2)." } ,

    { 0, NULL, false, false, false, { 0, 0, 0, 0 }, E_DEFHMASK, 0, LOTS,
      (int (*)()) NULL,
      NULL }

} ;

