/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * For dealing with nutmeg input decks and command scripts
 */

#include "spice.h"
#include "ftedefs.h"
#include "fteinp.h"

void
nutcom_edit(wl)

wordlist *wl;
{
    int inter;

    inter = cp_interactive;
    cp_interactive = false;

    if (SCEDactive()) {
        ShowPrompt("Exit sced to edit.");
        return;
    }
    for ( ; wl; wl = wl->wl_next) {

        /* ignore -r, -n */
        if (eq(wl->wl_word,"-n"))
            continue;

        if (eq(wl->wl_word,"-r"))
            continue;

        (void)inp_edit(wl->wl_word);
    }
    cp_interactive = inter;
    return;
}


bool
inp_edit(filename)

char *filename;
{
    char buf[BSIZE_SP], buf2[BSIZE_SP], *editor;
    extern char *kw_editor;

    if (cp_getvar(kw_editor, VT_STRING, buf2)) {
        editor = buf2;
    }
    else {
        if (!(editor = getenv("EDITOR"))) {
            if (Def_Editor && *Def_Editor)
            editor = Def_Editor;
        else
            editor = "/usr/ucb/vi";
        }
    }
    (void) sprintf(buf, "%s %s", editor, filename);
    /* Note: probably don't want this if editor creates its
     * own window.
     */
    return (DevXterm(buf));
}


/* nutmeg SCED interface */
#ifdef notdef
void
nutcom_sced(wl)

wordlist *wl;
{
    FILE *fp,*errfp,*outfp,*errtmp,*outtmp;
    char *errmsg = "Error: no symbolic representation found.\n";
    char *filename,*errfile,*outfile;
    int inter;

    inter = cp_interactive;
    cp_interactive = false;
top:
    if (wl) {

        /* ignore -r, -n */
        if (eq(wl->wl_word,"-n")) {
            wl = wl->wl_next;
            goto top;
        }
        if (eq(wl->wl_word,"-r")) {
            wl = wl->wl_next;
            goto top;
        }
        filename = wl->wl_word;
    }
    else {
        filename = "noname";
    }
    errfile = smktemp("er");
    outfile = smktemp("ot");

    errfp = fopen(errfile,"w+");
    outfp = fopen(outfile,"w");
    if (errfp == NULL || outfp == NULL) {
        fprintf(cp_err,"Error: redirection failed.\n");
        return;
    }
    errtmp = cp_err;
    outtmp = cp_out;
    cp_err = cp_curerr = errfp;
    cp_out = cp_curout = outfp;
    (void)sced(filename);
    cp_curout = outtmp;
    cp_curerr = errtmp;
    cp_ioreset();
    unlink(errfile);
    unlink(outfile);
    tfree(errfile);
    tfree(outfile);

    cp_interactive = inter;
}
#endif


/* The routine to source a spice input deck. We read the deck in, take out
 * the front-end commands, and create a CKT structure. Also we filter out
 * the following cards: .save, .width, .four, .print, and .plot, to perform
 * after the run is over.
 */

void
inp_nutsource(fp, comfile, filename)

FILE *fp;
bool comfile;
char *filename;
{
    struct line *deck, *dd;
    bool commands = comfile;
    FILE *lastin, *lastout, *lasterr;
    
    inp_readall(fp, &deck, NULL);

    if (!deck)
        return;

    if (!deck->li_next)
        fprintf(cp_err, "Warning: no cards in deck...\n");

    /* Now save the IO context and start a new control set...  After
     * we are done with the source we'll put the old file descriptors 
     * back.  I guess we could use a FILE stack, but since this routine
     * is recursive anyway...
     */
    lastin = cp_curin;
    lastout = cp_curout;
    lasterr = cp_curerr;
    cp_curin = cp_in;
    cp_curout = cp_out;
    cp_curerr = cp_err;
    cp_pushcontrol();

    for (dd = deck->li_next; dd; dd = dd->li_next) {
        if ((dd->li_line[0] == '*') && (dd->li_line[1] != '#')) {
            continue;
        }
        if (!*dd->li_line ||
            (commands && (*dd->li_line == '#' ||
            *dd->li_line == '*'))) {
            /* So blank lines in com files don't get
             * considered as circuits.
             * Also kill comments in commands.
             */
            continue;
        }
        if (ciprefix(".control", dd->li_line)) {
            if (!comfile) {
                if (commands)
                    fprintf(cp_err, 
                    "Warning: redundant .control card\n");
                else
                    commands = true;
            }
            continue;
        }
        if (ciprefix(".endc", dd->li_line)) {
            if (!comfile) {
                if (commands)
                    commands = false;
                else
                    fprintf(cp_err, 
                    "Warning: misplaced .endc card\n");
            }
            continue;
        }
        if (commands || prefix("*#", dd->li_line)) {
            if (prefix("*#", dd->li_line))
                (void) cp_evloop(dd->li_line + 2);
            else
                (void) cp_evloop(dd->li_line);
        }
    }
    inp_deckfree(deck);

    /* Now reset everything.  Pop the control stack, and fix up the IO
     * as it was before the source.
     */
    cp_popcontrol();
    cp_curin = lastin;
    cp_curout = lastout;
    cp_curerr = lasterr;
    return;
}


void
nutcom_source(wl)

wordlist *wl;

{
    FILE *fp, *tp;
    char buf[BSIZE_SP];
    bool inter;
    char *tempfile = NULL, *fname;
    wordlist *owl, *ww, *wn;
    int i;

    owl = wl;
    wl = wl_copy(wl);
    ww = wl;
    for (; wl; wl = wn) {
        wn = wl->wl_next;
        if (*wl->wl_word != '-')
            continue;
        if (!strchr(wl->wl_word,'n') &&
            !strchr(wl->wl_word,'c') &&
            !strchr(wl->wl_word,'r'))
            continue;
        if (wl->wl_prev)
            wl->wl_prev->wl_next = wl->wl_next;
        if (wl->wl_next)
            wl->wl_next->wl_prev = wl->wl_prev;
        txfree(wl->wl_word);
        if (ww == wl)
            ww = wl->wl_next;
        txfree((char*)wl);
    }
    wl = ww;

    inter = cp_interactive;
    cp_interactive = false;
    if (wl && wl->wl_next) {
        /* There are several files -- put them into a temp file  */
        tempfile = smktemp("sp");
        if (!(fp = inp_pathopen(tempfile, "w+"))) {
            perror(tempfile);
            goto done;
        }
        while (wl) {
            if (!(tp = inp_pathopen(wl->wl_word, "r"))) {
                perror(wl->wl_word);
                goto done;
            }
            while ((i = fread(buf, 1, BSIZE_SP, tp)) > 0)
                (void) fwrite(buf, 1, i, fp);
            (void) fclose(tp);
            wl = wl->wl_next;
        }
        (void) fseek(fp, (long) 0, 0);
    }
    else {
        if (wl) {
            if (!(fp = inp_pathopen(wl->wl_word, "r"))) {
                perror(wl->wl_word);
                goto done;
            }
        }
        else
            fp = stdin;
    }

    if (tempfile)
        fname = NULL;
    else {
        if (wl)
            fname = wl->wl_word;
        else
            fname = NULL;
    }

    /* Don't print the title if this is a .spiceinit file. */
    if (substring(".spiceinit", owl->wl_word) ||
            substring("spice.rc", owl->wl_word))
        inp_nutsource(fp, true, fname);
    else
        inp_nutsource(fp,false,fname);
    if (fp != stdin)
        (void) fclose(fp);

done:
    cp_interactive = inter;

    if (tempfile) {
        (void) unlink(tempfile);
        txfree(tempfile);
    }
    wl_free(ww);
    return;
}
