/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Resource-related routines.
 */

#include "spice.h"
#include "ftedefs.h"

#ifdef HAVE_GETRUSAGE
#include <sys/time.h>
#include <sys/resource.h>
#else
#ifdef HAVE_TIMES
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>
#else
#ifdef HAVE_FTIME
#include <sys/types.h>
#include <sys/timeb.h>
#endif
#endif
#endif

#ifndef HAVE_GETRUSAGE
#ifndef HAVE_TIMES
#ifdef HAVE_FTIME
static struct timeb timebegin;
static struct timeb lasttime;
#endif
#endif
#endif 
static char *enddata;
static int origsec;
static int origusec;
static int lastsec;
static int lastusec;
static int lastusrsec;
static int lastusrusec;
static int lastsyssec;
static int lastsysusec;

#ifdef HAVE_DOSRLIMIT
extern char * _memavl();
#endif 

#ifdef __STDC__
static void printres(char*);
#else
static void printres();
#endif


void
com_rusage(wl)
wordlist *wl;
{
    /* Fill in the SPICE accounting structure... */
    char *c;

    out_send("\n");
    if (wl && (eq(wl->wl_word, "everything") || eq(wl->wl_word, "all"))) {
        printres((char *) NULL);
    }
    else if (wl) {
        for (; wl; wl = wl->wl_next) {
            c = copy(wl->wl_word);
            cp_unquote(c);
            printres(c);
            txfree(c);
        }
    }
    else {
        printres("elapsed");
        printres("totaltime");
        printres("space");
    }
    return;
}


/* Find out if the user is approaching his maximum data size. */

void
ft_ckspace()
{
#ifdef HAVE_GETRLIMIT
    struct rlimit rld;
    long lim;
    char *hi;

    getrlimit(RLIMIT_DATA, &rld);
    if (rld.rlim_cur == RLIM_INFINITY)
        return;
    hi = (char*)sbrk(0);
    lim = rld.rlim_max;
    if ((int) (hi - (char *) enddata) > rld.rlim_max * 0.9) {
        fprintf(cp_err, "Warning - approaching max data size: ");
        fprintf(cp_err, "cur size = %d, hard limit = %d.\n", 
                (int) hi, (int) lim);
    }
    else if ((int) (hi - (char *) enddata) > rld.rlim_cur * 0.9) {
        fprintf(cp_err, "Warning - approaching max data size: ");
        fprintf(cp_err, "cur size = %d, soft limit = %d.\n", 
                (int) hi, (int) rld.rlim_cur);
    }
#else
#ifdef HAVE_ULIMIT
    long lim;
    char *hi;

    lim = ulimit(3, 0L);
    hi = (char*)sbrk(0);
    if ((int) (hi - (char *) enddata) > lim * 0.9) {
        fprintf(cp_err, "Warning - approaching max data size: ");
        fprintf(cp_err, "cur size = %d, hard limit = %d.\n", 
                (int) hi, (int) lim);
    }
#else
#ifdef HAVE_DOSRLIMIT
    long lim;

    lim = (long)_memavl( );
    if ((long) (enddata - lim) > ((long) enddata) * 0.9) {
        fprintf(cp_err, "Warning - approaching max data size: ");
        fprintf(cp_err, "cur size = %ld, hard limit = %ld.\n", 
                (long) (enddata - lim), (long) enddata);
    }
#endif
#endif
#endif
}


void
rusage_init()

{
#ifdef HAVE_GETRUSAGE
    struct rusage ruse;
#ifdef HAVE_GETTIMEOFDAY
    struct timeval tv;
    struct timezone tz;

    (void) gettimeofday(&tv, &tz);
    lastsec = tv.tv_sec;
    lastusec = tv.tv_usec;
    origsec = lastsec;
    origusec = lastusec;
#endif
    (void) getrusage(RUSAGE_SELF, &ruse);
    lastusrsec = ruse.ru_utime.tv_sec;
    lastusrusec = ruse.ru_utime.tv_usec;
    lastsyssec = ruse.ru_stime.tv_sec;
    lastsysusec = ruse.ru_stime.tv_usec;
#else
#ifdef HAVE_TIMES
    struct tms dummy;

    origsec = times(&dummy);
    lastusrsec = dummy.tms_utime;
    lastsyssec = dummy.tms_stime;
    lastsec = origsec;
#else
#ifdef HAVE_FTIME
    ftime(&timebegin);      /* initialize time we started */
#endif
#endif
#endif

#ifdef HAVE_GETRLIMIT
    enddata = (char*)sbrk(0);
#else
#ifdef HAVE_ULIMIT
    enddata = (char*)sbrk(0);
#else
#ifdef HAVE_DOSRLIMIT
    enddata = _memavl( );
#endif
#endif
#endif
}


/* Print out one piece of resource usage information. */

static void
printres(name)

char *name;
{
    wordlist *kw, *tw;
    struct variable *v, *vv;
    long realt;
    unsigned buf[25];
    char *fmtf = "%-15s%-10.2f%s\n";
    char *fmtx = "%-15s%-10x%s\n";
    char *fmtd = "%-15s%-10d%s\n";

    if (!name || eq(name, "totaltime")) {

#ifdef HAVE_GETRUSAGE
        struct rusage ruse;
#ifdef HAVE_GETTIMEOFDAY
        struct timeval tv;
        struct timezone tz;

        (void) gettimeofday(&tv, &tz);
        out_printf(fmtf,"totaltime",
            (double)(tv.tv_sec - origsec) +
            (double)(tv.tv_usec - origusec)/1.0e6,
            "Total elapsed seconds");
#endif
        (void) getrusage(RUSAGE_SELF, &ruse);
#ifndef HAVE_GETTIMEOFDAY
        out_printf(fmtf,"totaltime",
#else
        out_printf(fmtf,"",
#endif
            (double)ruse.ru_utime.tv_sec +
            (double)ruse.ru_utime.tv_usec/1.0e6,
            "Total user cpu seconds");
        out_printf(fmtf,"",
            (double)ruse.ru_stime.tv_sec +
            (double)ruse.ru_stime.tv_usec/1.0e6,
            "Total system cpu seconds");
#else
#ifdef HAVE_TIMES
        struct tms ruse;

        realt = times(&ruse);
        out_printf(fmtf,"totaltime",
            (double)(realt - origsec)/HZ +
            (double)((realt - origsec)%HZ)/HZ,
            "Total elapsed seconds");

        out_printf(fmtf,"",
            (double)ruse.tms_utime/HZ +
            (double)(ruse.tms_utime%HZ)/HZ,
            "Total user cpu seconds");
        out_printf(fmtf,"",
            (double)ruse.tms_stime/HZ +
            (double)(ruse.tms_stime%HZ)/HZ,
            "Total system cpu seconds");
#else
#ifdef HAVE_FTIME
        struct timeb timenow;
        int sec, msec;

        ftime(&timenow);
        timediff(&timenow, &timebegin, &sec, &msec);
        out_printf(fmtf,"totaltime",(double)sec + (double)msec/1000,
            "Total run time seconds");
#endif
#endif
#endif
        if (name) return;
    }

    if (!name || eq(name, "elapsed")) {

#ifdef HAVE_GETRUSAGE
        struct rusage ruse;
#ifdef HAVE_GETTIMEOFDAY
        struct timeval tv;
        struct timezone tz;

        (void) gettimeofday(&tv, &tz);
        out_printf(fmtf,"elapsed",
            (double)(tv.tv_sec - lastsec) +
            (double)(tv.tv_usec - lastusec)/1.0e6,
            "Seconds since last call");
        lastsec = tv.tv_sec;
        lastusec = tv.tv_usec;
#endif
        (void) getrusage(RUSAGE_SELF, &ruse);
#ifndef HAVE_GETTIMEOFDAY
        out_printf(fmtf,"elapsed",
#else
        out_printf(fmtf,"",
#endif
            (double)(ruse.ru_utime.tv_sec - lastusrsec) +
            (double)(ruse.ru_utime.tv_usec - lastusrusec)/1.0e6,
            "User cpu seconds since last call");
        lastusrsec = ruse.ru_utime.tv_sec;
        lastusrusec = ruse.ru_utime.tv_usec;
        out_printf(fmtf,"",
            (double)(ruse.ru_stime.tv_sec - lastsyssec) +
            (double)(ruse.ru_stime.tv_usec - lastsysusec)/1.0e6,
            "System cpu seconds since last call");
        lastsyssec = ruse.ru_stime.tv_sec;
        lastsysusec = ruse.ru_stime.tv_usec;
#else
#ifdef HAVE_TIMES
        struct tms ruse;

        realt = times(&ruse);
        out_printf(fmtf,"elapsed",
            (double)(realt - lastsec)/HZ +
            (double)((realt - lastsec)%HZ)/HZ,
            "Seconds since last call");
        out_printf(fmtf,"",
            (double)(ruse.tms_utime - lastusrsec)/HZ +
            (double)((ruse.tms_utime - lastusrsec)%HZ)/HZ,
            "User cpu seconds since last call");
        out_printf(fmtf,"",
            (double)(ruse.tms_stime - lastsyssec)/HZ +
            (double)((ruse.tms_stime - lastsyssec)%HZ)/HZ,
            "System cpu seconds since last call");
        }
        lastusrsec = ruse.tms_utime;
        lastsyssec = ruse.tms_stime;
        lastsec = realt;
#else
#ifdef HAVE_FTIME
        struct timeb timenow;
        int sec, msec;

        ftime(&timenow);
        if (lasttime.time != 0 || lasttime.millitm != 0) {
            timediff(&timenow, &lasttime, &sec, &msec);
            out_printf(fmtf,"elapsed",(double)sec + (double)msec/1000,
                "Seconds since last call");
        }
        lasttime = timenow;
#endif
#endif
#endif
        if (name) return;
    } 

    if (!name || eq(name, "space")) {
#ifdef HAVE_GETRLIMIT
        struct rlimit rld;
        char *hi;

        getrlimit(RLIMIT_DATA, &rld);
        hi = (char*)sbrk(0);
        out_printf(fmtd,"space", (int) (hi - enddata),
            "Current data size");
        out_printf(fmtd,"", (unsigned)rld.rlim_max,"Hard data limit");
        out_printf(fmtd,"", (unsigned)rld.rlim_cur,"Soft data limit");
#else
#ifdef HAVE_ULIMIT
        long lim;
        char *hi;

        lim = ulimit(3, 0L);
        hi = (char*)sbrk(0);
        out_printf(fmtd,"space", (int) (hi - (char *) enddata),
            "Current data size");
        out_printf(fmtd,"", lim,"Data limit");
#else
#ifdef HAVE_DOSRLIMIT
        long lim;

        lim = (long)_memavl( );
        out_printf(fmtd,"space",((long)enddata - lim), "Current data size");
        out_printf(fmtd,"",lim, "Data limit");
#endif
#endif
#endif
        if (name) return;
    }

    if (!name || eq(name, "faults")) {
#ifdef HAVE_GETRUSAGE
        struct rusage ruse;

        (void) getrusage(RUSAGE_SELF, &ruse);
        out_printf(fmtd,"faults",ruse.ru_majflt,"Page Faults");
        out_printf("%-15s%-10d%s (vol %d + invol %d)\n","",
            ruse.ru_nvcsw + ruse.ru_nivcsw,
            "Context switches",ruse.ru_nvcsw,ruse.ru_nivcsw);
#endif
        if (name) return;
    }

    /* Now get all the spice resource stuff. */
    if (ft_curckt && ft_curckt->ci_ckt) {
        v = if_getstat(ft_curckt->ci_ckt, name, &kw);
        if (v) {
            tw = kw;
            vv = v;
            while (v) {
                switch (v->va_type) {
                    case VT_NUM:
                        out_printf(fmtd,tw->wl_word,v->va_num,v->va_name);
                        break;
                    case  VT_REAL:
                        out_printf(fmtf,tw->wl_word,v->va_real,v->va_name);
                        break;
                }
                tw = tw->wl_next;
                v = v->va_next;

            }
            wl_free(kw);
            va_free(vv);
        }
        else if (name)
            fprintf(cp_err,
                "Note: no resource usage information on %s,\n", name);
        return;
    }
    else
        fprintf(cp_err, "Note: no active circuit available\n");
    return;
}
