/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * The signal routines for spice 3 and nutmeg.
 */

#include "spice.h"
#include "ftedefs.h"
#include "iferrmsg.h"
#include "plotdev.h"
#include <setjmp.h>
#ifdef HAVE_SIGNAL
#include <signal.h>
#endif

#ifdef HAVE_SIGNAL
#ifdef HAVE_SETJMP

extern jmp_buf jbuf;

#ifdef __STDC__
static void killplot(void);
#else
static void killplot();
#endif

/* The signal handlers... SIGINT is the only one that gets reset (by
 * cshpar) so it is global. They are ifdef BSD because of the sigmask
 * stuff in sigstop. We set the interrupt flag and return if ft_setflag
 * is true.
 */

RETSIGTYPE
ft_sigintr()

{
    /* should never get here from sced */
    (void) signal( SIGINT, (RETSIGTYPE(*)())ft_sigintr );

    if (ft_intrpt)
        fprintf(cp_err, " Interrupt (ouch)\n");
    else {
        ft_intrpt = true;
    }
}


/* ARGSUSED */
RETSIGTYPE
sigfloat(sig, code)

{
    /* should never get here from sced */
    killplot();
    fperror("Error", code);
    rewind(cp_out);
    (void) signal( SIGFPE, (RETSIGTYPE(*)())sigfloat );
    longjmp(jbuf, 1);
}


static void
killplot()

/* To restore screen after an interrupt to a plot */
{
    int id;

    if (!currentgraph) return;
    switch (currentgraph->graphtype) {
        case GR_PLOT:
            currentgraph->plotdata =
                (*currentgraph->copydata)(currentgraph->plotdata);
        case GR_GRAF:
        case GR_MPLT:
            DevHalt();
            id = currentgraph->graphid;
            PopGraphContext();
            DestroyGraph(id);
            break;
    }
}


/* This should give a new prompt if cshpar is waiting for input.  */

#ifdef SIGTSTP

RETSIGTYPE
sigstop( )
{
    cp_ccon(false);
    (void) signal(SIGTSTP, SIG_DFL);
    (void) kill(getpid(), SIGTSTP); /* This should stop us */
}


RETSIGTYPE
sigcont( )
{
    (void) signal(SIGTSTP, (RETSIGTYPE(*)())sigstop);
    if (cp_cwait)
        longjmp(jbuf, 1);
}

#endif

/* Special signal handlers. */


RETSIGTYPE
sigill()

{
    while (currentgraph) {
        DevHalt();
        PopGraphContext();
    }
    fprintf(cp_err, "\ninternal error -- illegal instruction\n");
    fatal();
}


RETSIGTYPE
sigbus()

{
    while (currentgraph) {
        DevHalt();
        PopGraphContext();
    }
    fprintf(cp_err, "\ninternal error -- bus error\n");
    fatal();
}


RETSIGTYPE
sigsegv()

{
    while (currentgraph) {
        DevHalt();
        PopGraphContext();
    }
    fprintf(cp_err, "\ninternal error -- segmentation violation\n");
    fatal();
}


RETSIGTYPE
sig_sys()

{
    while (currentgraph) {
        DevHalt();
        PopGraphContext();
    }
    fprintf(cp_err, 
        "\ninternal error -- bad argument to system call\n");
    fatal();
}

#endif
#endif
