/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1986 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Toss the help window up on the screen, and deal with the graph...
 */

#include "spice.h"
#include "hlpdefs.h"

bool hlp_usex;


void
hlp_provide(top)

topic *top;
{
    toplink *res;
    topic *parent, *newtop;

    if (!top)
        return;

    top->xposition = top->yposition = 0;
    if (hlp_usex) {
        if (!hlp_xdisplay(top)) {
            fprintf(stderr, "Couldn't open X display.\n");
        }
        /* X does this asynchronously */
        return;
    }

    if (!hlp_tdisplay(top)) {
        fprintf(stderr, "Couldn't display text\n");
        return;
    }
    
    for (;;) {
        res = hlp_thandle(&parent);
        if (!res && !parent) {
            /* No more windows. */
            return;
        }
        if (res) {
            /* Create a new window... */
            if (!(newtop = hlp_read(res->keyword))) {
                fprintf(stderr, "Internal error: bad link\n");
                continue;
            }
            newtop->sibling = parent->lastborn;
            parent->lastborn = newtop;
            newtop->parent = parent;
            newtop->xposition = parent->xposition + 50;
            newtop->yposition = parent->yposition + 50;
            if (!hlp_tdisplay(newtop)) {
                fprintf(stderr, "Couldn't display\n"); 
                return; 
            }
        }
        else {
            /* Blow this one and its descendants away. */
            hlp_fixchildren(parent);
            hlp_killfamily(parent);
        }
    }
}


void
hlp_fixchildren(top)

topic *top;
{
    topic *tp;
    /* unlink top from the tree */

    if (top->parent) {
        if (top->parent->lastborn == top)
            top->parent->lastborn = top->sibling;
        else {
            for (tp = top->parent->lastborn;
                    tp->sibling; tp = tp->sibling)
                if (tp->sibling == top)
                    break;
            if (!tp->sibling) {
                fprintf(stderr, "bah...\n");
            }
            tp->sibling = tp->sibling->sibling;
        }
    }
}


void
hlp_killfamily(top)

topic *top;
{
    topic *tp;

    for (tp = top->lastborn; tp; tp = tp->sibling)
        hlp_killfamily(tp);
    if (hlp_usex)
        hlp_xkillwin(top);
    else
        hlp_tkillwin(top);
    hlp_free(top);
}
