/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1986 Wayne A. Christopher
         1994 Stephen R. Whiteley
****************************************************************************/

/*
 * The main entry point for the help system.
 */

#include "spice.h"
#include "hlpdefs.h"

char *hlp_directory;
int hlp_initxpos = START_XPOS;
int hlp_initypos = START_YPOS;

static struct sHlpEnt *HelpBase;

#if __STDC__
static void sortlist(toplink**);
static int  sortcmp(toplink**,toplink**);
static void tlfree(toplink*);
static FILE *db_open(char*,struct sHlpEnt**);
static void read_db(FILE*);
static struct sHlpEnt *find_entry(char*);
static char *my_fgets(char*,int,FILE*);
#else
static void sortlist();
static int  sortcmp();
static void tlfree();
static FILE *db_open();
static void read_db();
static struct sHlpEnt *find_entry();
static char *my_fgets();
#endif

#define prefix(x,y) ciprefix(x,y)


void
hlp_main(path, wl)

char *path;
wordlist *wl;
{
    topic *top = NULL;

    txfree(hlp_directory);
    hlp_directory = copy(path);

    if (wl == NULL) {
        if (!(top = hlp_read(NULL))) {
            fprintf(stderr, "Error: no top level topic\n");
            return;
        }
        hlp_provide(top);
    }
    while (wl) {
        if (!(top = hlp_read(wl->wl_word))) {
            if (wl->wl_word && *wl->wl_word)
                fprintf(stderr, "Error: No such topic: %s\n",
                    wl->wl_word);
            else
                fprintf(stderr, "Error: no top level topic\n");
            wl = wl->wl_next;
            continue;
        }
        hlp_provide(top);
        wl = wl->wl_next;
    }
}


topic *
hlp_read(word)

char *word;
{
    char buf[BSIZE_SP], *s, *t;
    int i, seealso, subtopics;
    FILE *fp;
    topic *top;
    toplink *tl;
    topic *alltopics = NULL;
    struct sHlpEnt *bb;
    wordlist *wl = NULL;

    if (!(fp = db_open(word,&bb))) {
        fprintf(stderr,"Error: no title for topic %s.\n", word);
        return (NULL);
    }
    top = alloc(topic);
    top->keyword = copy(word);
    top->title = copy(bb->title);
    top->subtopics = NULL;
    top->seealso = NULL;
    top->text = NULL;
    top->maxcols = 0;
    seealso = false;
    subtopics = false;
        
    while ((s = my_fgets(buf,BSIZE_SP,fp)) != NULL) {

        if (prefix(HLP_KEYWORD,s))
            break;

        if (prefix(HLP_SEEALSO,s)) {
            seealso = true;
            subtopics = false;
            tl = NULL;
            continue;
        }
        if (prefix(HLP_SUBTOPICS,s)) {
            subtopics = true;
            seealso = false;
            tl = NULL;
            continue;
        }

        if (seealso || subtopics) {

            if (*s == '#' || *s == '*')
                continue;
            while (isspace(*s))
                s++;
            if (!*s)
                continue;

            t = s + strlen(s)-1;
            while (*t <= ' ')
                *t-- = '\0';

            if (tl) {
                tl->next = alloc(toplink);
                tl = tl->next;
            }
            else
                tl = alloc(toplink);
            tl->keyword = copy(s);
            tl->next = NULL;
            tl->top = top;
            bb = find_entry(s);
            if (bb)
                tl->description = copy(bb->title);
            else
                tl->description = copy("<unknown>");

            if (seealso && !top->seealso)
                top->seealso = tl;
            else if (subtopics && !top->subtopics)
                top->subtopics = tl;
            continue;
        }
        if ((s = strchr(buf,'\n')) != NULL)
            *s = '\0';
        if (wl) {
            wl->wl_next = alloc(wordlist);
            wl->wl_next->wl_prev = wl;
            wl = wl->wl_next;
        }
        else {
            wl = alloc(wordlist);
            wl->wl_prev = NULL;
        }
        wl->wl_next = NULL;
        wl->wl_word = copy(buf);

        if (!top->text)
            top->text = wl;

        top->numlines++;
        if ((i = strlen(buf)) > top->maxcols)
            top->maxcols = i;
    }
    (void) fclose(fp);

    sortlist(&top->seealso);
    sortlist(&top->subtopics);

    return (top);
}


static void
sortlist(tlp)

toplink **tlp;
{
    toplink **vec, *tl;
    int num, i;

    for (num = 0,tl = *tlp; tl; tl = tl->next) num++;

    if (!num) return;

    vec = (toplink **) tmalloc(sizeof (toplink *) * num);
    for (tl = *tlp, i = 0; tl; tl = tl->next, i++)
        vec[i] = tl;
    (void) qsort((char *) vec, num, sizeof (toplink *),
#if __STDC__
        (int(*)(const void*,const void*))sortcmp);
#else
        sortcmp);
#endif
    *tlp = vec[0];
    for (i = 0; i < num - 1; i++)
        vec[i]->next = vec[i + 1];
    vec[i]->next = NULL;
    free(vec);
    return;
}


static int
sortcmp(tlp1, tlp2)

toplink **tlp1, **tlp2;
{
    return (strcmp((*tlp1)->description, (*tlp2)->description));
}


void
hlp_free(top)

topic *top;
{
    txfree(top->title);
    txfree(top->keyword);
    txfree(top->chartext);
    wl_free(top->text);
    tlfree(top->subtopics);
    tlfree(top->seealso);
    free((char*)top);
}


static void
tlfree(tl)

toplink *tl;
{
    toplink *nt = NULL;

    while (tl) {
        txfree(tl->description);
        txfree(tl->keyword);
        /* Don't free the button stuff... */
        nt = tl->next;
        free((char*)tl);
        tl = nt;
    }
}


static FILE *
db_open(word,pb)

char *word;
struct sHlpEnt **pb;
{
    char buf[BSIZE_SP];
    FILE *fp;
    struct sHlpEnt *bb;

    (void) sprintf(buf,"%s/%s",hlp_directory,DBFILE);
    cp_pathfix(buf);
    /* open in binary mode, so ftell/fseek work in DOS */
    fp = fopen(buf,"rb");
    if (!fp) {
        perror(buf);
        return (NULL);
    }
    if (!HelpBase)
        read_db(fp);

    if (!HelpBase)
        return (NULL);

    bb = find_entry(word);
    if (bb) {
        *pb = bb;
        fseek(fp,bb->offset,0);
    }
    else {
        fclose(fp);
        fp = NULL;
    }
    return (fp);
}


static void
read_db(fp)

FILE *fp;
{
    char buf[BSIZE_SP], *s, *t, *kw, *ti;
    struct sHlpEnt *bb;
    int keyword, title;

    while ((s = my_fgets(buf, BSIZE_SP, fp)) != NULL) {

        if (prefix(HLP_KEYWORD,buf)) {
            keyword = true;
            title = false;
            kw = NULL;
            continue;
        }
        if (prefix(HLP_TITLE,buf)) {
            title = true;
            keyword = false;
            ti = NULL;
            continue;
        }
        if (prefix(HLP_TEXT,buf)) {
            if (ti && kw) {

                if (HelpBase == NULL) {
                    HelpBase = alloc(struct sHlpEnt);
                    bb = HelpBase;
                }
                else {
                    bb->next = alloc(struct sHlpEnt);
                    bb = bb->next;
                }
                bb->next = NULL;
                bb->keyword = kw;
                bb->title = ti;
                bb->offset = ftell(fp);
                ti = NULL;
                kw = NULL;
                keyword = false;
                title = false;
            }
            continue;
        }

        if (keyword || title) {

            while (isspace(*s))
                s++;
            if (!*s)
                continue;

            t = s + strlen(s)-1;
            while (*t <= ' ')
                *t-- = '\0';

            if (keyword)
                kw = copy(s);
            else
                ti = copy(s);

        }
    }
}


static struct sHlpEnt *
find_entry(word)

char *word;
{
    struct sHlpEnt *bb;

    if (!word)
        return (HelpBase);

    for (bb = HelpBase; bb; bb = bb->next) {
        if (!strcmp(word,bb->keyword))
            return (bb);
    }
    return (NULL);
}


static char *
my_fgets(buf,size,fp)

char *buf;
int size;
FILE *fp;
{
    char *s;
    int i, c;
    /* works with DOS or UNIX */

    for (s = buf, i = size; i; s++, i--) {
        c = getc(fp);
        if (c == '\r')
            c = getc(fp);
        if (c == EOF) {
            *s = '\0';
            if (s == buf)
                return (NULL);
            return (buf);
        }
        *s = c;
        if (c == '\n') {
            *++s = '\0';
            return (buf);
        }
    }
    buf[size-1] = '\0';
    return (buf);
}
