/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1986 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "hlpdefs.h"

static topic *curtop;
static bool quitflag;

#ifdef __STDC__
static void putline(char*);
static int putstuff(toplink*,int);
#else
static void putline();
static int putstuff();
#endif

extern FILE *cp_in, *cp_out, *cp_err;
extern bool out_isatty;


bool
hlp_tdisplay(top)

topic *top;
{
    wordlist *wl;
    int i = 0;
    char *xx;

    curtop = top;

    out_init();
    out_printf("\n");
    if (cp_out == stdout) {
        fprintf(cp_out," ");
        fflush(cp_out);
        cprint(15,top->title);
        out_printf("\n");
    }
    else
        out_printf("%s\n",top->title);
    for (wl = top->text; wl; wl = wl->wl_next)
        putline(wl->wl_word);
    if (top->subtopics) {
        xx = "Sub-Topics:";
        if (cp_out == stdout) {
            fprintf(cp_out," ");
            fflush(cp_out);
            cprint(15,xx);
            out_printf("\n");
        }
        else
            out_printf("%s\n",xx);
        i = putstuff(top->subtopics, 0);
        out_printf("\n");
    }
    if (top->seealso) {
        xx = "See Also:";
        if (cp_out == stdout) {
            fprintf(cp_out," ");
            fflush(cp_out);
            cprint(15,xx);
            out_printf("\n");
        }
        else
            out_printf("%s\n",xx);
        (void) putstuff(top->seealso, i);
        out_printf("\n");
    }
    if (!top->subtopics && !top->seealso)
        out_printf("\n");
    return (true);
}


toplink *
hlp_thandle(parent)

topic **parent;
{
    char buf[BSIZE_SP], *s;
    char *mesg;
    toplink *tl;
    int num;

    quitflag = false;
    if (!curtop) {
        *parent = NULL;
        return (NULL);
    }

    if (SCEDactive()) {
        mesg =
            ((curtop->subtopics || curtop->seealso) ? "Selection? " : NULL);
    }
    else
        mesg = "Selection (`?' for help): ";
        
    for (;;) {
        if (!SCEDfgets(buf,sizeof(buf),cp_in,mesg)) {
            quitflag = true;
            *parent = NULL;
            return (NULL);
        }

        for (s = buf; *s && isspace(*s); s++)
            ;
        switch (*s) {
            case '?':
                fprintf(cp_out,
"\nType the number of a sub-topic or see also, or one of:\n\
\tr\tReprint the current topic\n\
\tp or CR\tReturn to the previous topic\n\
\tq\tQuit help\n\
\t?\tPrint this message\n\n");
                continue;

            case 'r':
                {
                char *ss;
                FILE *tmp;
                bool ttytmp;

                if ((ss = strchr(s,'\n')) != NULL)
                    *ss = '\0';

                if ((ss = strchr(s,'>')) == NULL) {
                    (void) hlp_tdisplay(curtop);
                    continue;
                }    
                tmp = cp_out;
                ss++;
                if (*ss == '>') {
                    ss++;
                    cp_out = fopen(ss,"a");
                }
                else
                    cp_out = fopen(ss,"w");
                if (cp_out) {
                    ttytmp = out_isatty;
                    out_isatty = isatty(fileno(cp_out));
                    (void) hlp_tdisplay(curtop);
                    out_isatty = ttytmp;
                    (void) fclose(cp_out);
                }
                else
                    perror(ss);
                cp_out = tmp;
                continue;
                }

            case 'q':
                quitflag = true;
                *parent = NULL;
                return (NULL);

            case 'p':
            case '\0':
            case '\n':
            case '\r':
                *parent = curtop;
                return (NULL);
        }
        if (!isdigit(*s)) {
            fprintf(cp_err, "Invalid command\n");
            continue;
        }
        num = atoi(s);
        if (num <= 0) {
            fprintf(cp_err, "Bad choice.\n");
            continue;
        }
        for (tl = curtop->subtopics; tl; tl = tl->next)
            if (--num == 0)
                break;
        if (num) {
            for (tl = curtop->seealso; tl; tl = tl->next)
                if (--num == 0)
                    break;
        }
        if (num) {
            fprintf(cp_err, "Bad choice.\n");
            continue;
        }
        *parent = curtop;
        return (tl);
    }
}


/* ARGSUSED */
void
hlp_tkillwin(top)

topic *top;
{
    if (curtop)
        curtop = curtop->parent;
    if (curtop && !quitflag)
        (void) hlp_tdisplay(curtop);
    return;
}



static void
putline(s)

/* This has to rip out the font changes from the lines... */
char *s;
{
    char buf[BSIZE_SP];
    int i = 0;

    while (*s) {
        if (((*s == '\033') && s[1]) ||
                ((*s == '_') && (s[1] == '\b')))
            s += 2;
        else
            buf[i++] = *s++;
    }
    buf[i] = '\0';
    out_printf("%s\n", buf);
    return;
}


static int
putstuff(tl, base)

/* Figure out the number of columns we can use.  Assume an entry like
 * nn) word -- add 5 characters to the width...
 */
toplink *tl;
int base;
{
    int width, maxwidth = 0, ncols, nrows, nbuts = 0, i, j, k;
    toplink *tt;

    if (SCEDactive())
        width = 72;
    else
        out_winsize(&width,NULL);
    width -= 4;
    for (tt = tl; tt; tt = tt->next) {
        if (strlen(tt->description) + 5 > maxwidth)
            maxwidth = strlen(tt->description) + 5;
        nbuts++;
    }
    ncols = width / maxwidth;
    if (!ncols) {
        fprintf(stderr, "Help, button too big!!\n");
        return (0);
    }
    if (ncols > nbuts)
        ncols = nbuts;
    maxwidth = width / ncols;
    nrows = nbuts / ncols;
    if (nrows * ncols < nbuts)
        nrows++;

    for (i = 0; i < nrows; i++) {
        for (tt = tl, j = 0; j < i; j++, tt = tt->next)
            ;
        for (j = 0; j < ncols; j++) {
            if (tt)
                out_printf("%2d) %-*s ", base + j * nrows + i +
                    1, maxwidth - 5, tt->description);
            for (k = 0; k < nrows; k++)
                if (tt)
                    tt = tt->next;
            
        }
        out_printf("\n");
    }
    return (nbuts);
}

