/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1990 Jeffery M. Hsu
         1994 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "hlpdefs.h"
#ifdef HAVE_SIGNAL
#include <signal.h>
#endif

#ifdef HAVE_X11

#include <X11/StringDefs.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Shell.h>

/* from x11.c */
#ifdef __STDC__
extern void ToTopOfSced(Widget,XtPointer,XEvent*,Boolean*);
#else
extern void ToTopOfSced();
#endif

#ifdef __STDC__
static void newtopic(Widget,XtPointer,XtPointer);
static void delete(Widget,XtPointer,XtPointer);
static void quit(Widget,XtPointer,XtPointer);
static void print(Widget,XtPointer,XtPointer);
static void save(Widget,XtPointer,XtPointer);
static void DoSave(Widget,XtPointer,XtPointer);
#else
static void newtopic();
static void delete();
static void quit();
static void print();
static void save();
static void DoSave();
#endif

#define hlpx_def_ht 350
#define hlpx_def_wd 660

bool hlp_killx;
static int hlp_cnt;

static char *sens_list[] = { "save", 0};

struct hlp_toplist {
    topic *top;
    struct hlp_toplist *next;
    struct hlp_toplist *prev;
};
static struct hlp_toplist *top_list;

bool
hlp_xdisplay(top)

/* Create a new window... */
topic *top;
{
    Widget button;
    Widget subbox;
    Widget seebox;
    Widget sublabel;
    Widget seelabel;
    toplink *tl;
    wordlist *wl;
    char *s;
    int cnt;
    widget_bag *w = &top->widgets;
    struct hlp_toplist *tw;

    if (!top->parent) {
        tw = alloc(struct hlp_toplist);
        tw->top = top;
        tw->next = top_list;
        if (tw->next)
            tw->next->prev = tw;
        top_list = tw;
    }

    if (hlp_cnt > 5)
        hlp_cnt = 0;
    if (!top->parent) {
        top->xposition = hlp_initxpos;
        top->yposition = hlp_initypos + hlp_cnt++ *2*Y_INCR;;
    }
    else {
        top->xposition = top->parent->xposition + X_INCR;
        top->yposition = top->parent->yposition + Y_INCR;
    }
    w->shell = XtVaAppCreateShell(NULL, "Jspice3", applicationShellWidgetClass,
        Xdisplay,
        XtNx, top->xposition,
        XtNy, top->yposition,
        NULL);
    XtOverrideTranslations(w->shell,
        XtParseTranslationTable("<Message>WM_PROTOCOLS: quit_action()"));

    if (SCEDactive())
        XtAddEventHandler(w->shell, VisibilityChangeMask,
            FALSE, ToTopOfSced, NULL);

    w->form = XtVaCreateManagedWidget("help", formWidgetClass,
        w->shell, NULL);

    w->butbox = XtVaCreateManagedWidget("buttonbox", boxWidgetClass,
        w->form,
        XtNtop, XtChainTop,
        XtNbottom, XtChainTop,
        XtNleft, XtChainLeft,
        XtNright, XtChainLeft,
        XtNwidth, 650,
        NULL);

    (void)XtVaCreateManagedWidget("titlelabel", labelWidgetClass,
        w->butbox,
        XtNlabel, top->title,
        XtNborderWidth, 2,
        NULL);

    button = XtVaCreateManagedWidget("quit", commandWidgetClass,
        w->butbox,
        XtNlabel, "quit help",
        NULL);
    XtAddCallback(button, XtNcallback, quit, top);

    button = XtVaCreateManagedWidget("delete", commandWidgetClass,
        w->butbox,
        XtNlabel, "delete window",
        NULL);
    XtAddCallback(button, XtNcallback, delete, top);

    button = XtVaCreateManagedWidget("print", commandWidgetClass,
        w->butbox,
        XtNlabel, "print",
        NULL);
    XtAddCallback(button, XtNcallback, print, top);

    button = XtVaCreateManagedWidget("save", commandWidgetClass,
        w->butbox,
        XtNlabel, "save",
        NULL);
    XtAddCallback(button, XtNcallback, save, top);

    for (cnt = 0, wl = top->text; wl; wl = wl->wl_next)
        cnt += strlen(wl->wl_word) + 2;
    top->chartext = tmalloc(cnt);
    top->chartext[0] = '\0';
    s = top->chartext;
    for (wl = top->text; wl; wl = wl->wl_next) {
        strcat(s, wl->wl_word);
        s += strlen(wl->wl_word);
        *s++ = '\n';
    }
    *s = '\0';

    w->text = XtVaCreateManagedWidget("helptext", asciiTextWidgetClass,
        w->form,
        XtNstring, top->chartext,
        XtNallowResize, True,
        XtNscrollHorizontal, XawtextScrollWhenNeeded,
        XtNscrollVertical, XawtextScrollWhenNeeded,
        XtNfromVert, w->butbox,
        XtNwidth, hlpx_def_wd,
        XtNheight, hlpx_def_ht,
        XtNbottom, XtChainBottom,
        XtNdisplayCaret, False,
        NULL);

    if (top->subtopics) {
        sublabel = XtVaCreateManagedWidget("sublabel", labelWidgetClass,
            w->form,
            XtNfromVert, w->text,
            XtNvertDistance, 8,
            XtNtop, XtChainBottom,
            XtNbottom, XtChainBottom,
            XtNborderWidth, 2,
            XtNlabel, "Subtopics: ",
            NULL);

        subbox = XtVaCreateManagedWidget("subbox", boxWidgetClass,
            w->form,
            XtNwidth, hlpx_def_wd - 20,
            XtNfromVert, sublabel,
            XtNtop, XtChainBottom,
            XtNbottom, XtChainBottom,
            XtNleft, XtChainLeft,
            XtNright, XtChainLeft,
            NULL);

        for (tl = top->subtopics; tl; tl = tl->next) {
            tl->buttontext = tl->description;
            if (!tl->buttontext)
                tl->buttontext = "<unknown>";

            tl->button = XtVaCreateManagedWidget(tl->buttontext,
                toggleWidgetClass,
                subbox,
                XtNlabel, tl->buttontext,
                NULL);
            XtAddCallback(tl->button, XtNcallback, newtopic, tl);
        }
    }

    if (top->seealso) {
        Widget fromvert = (top->subtopics ? subbox : w->text);

        seelabel = XtVaCreateManagedWidget("seelabel", labelWidgetClass,
            w->form,
            XtNfromVert, fromvert,
            XtNvertDistance, 8,
            XtNtop, XtChainBottom,
            XtNbottom, XtChainBottom,
            XtNborderWidth, 2,
            XtNlabel, "See also: ",
            NULL);

        seebox = XtVaCreateManagedWidget("seebox", boxWidgetClass,
            w->form,
            XtNwidth, hlpx_def_wd - 20,
            XtNfromVert, seelabel,
            XtNtop, XtChainBottom,
            XtNbottom, XtChainBottom,
            XtNleft, XtChainLeft,
            XtNright, XtChainLeft,
            NULL);

        for (tl = top->seealso; tl; tl = tl->next) {
            tl->buttontext = tl->description;
            if (!tl->buttontext)
                tl->buttontext = "<unknown>";

            tl->button = XtVaCreateManagedWidget(tl->buttontext,
                toggleWidgetClass,
                seebox,
                XtNlabel, tl->buttontext,
                NULL);
            XtAddCallback(tl->button, XtNcallback, newtopic, tl);
        }
    }
    XtRealizeWidget(w->shell);
    w->wm_delete = XInternAtom(Xdisplay, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(Xdisplay, XtWindow(w->shell), &w->wm_delete, 1);

    w->popup_sens_list = sens_list;
    return (true);
}


/* ARGSUSED */
static void
newtopic(w, client_data, call_data)

Widget w;
XtPointer client_data, call_data;
{
    topic *parent = ((toplink *) client_data)->top;
    toplink *result = (toplink *) client_data;
    topic *newtop;

    if (result->newtop) {
        delete(NULL, (XtPointer)result->newtop, NULL);
        result->newtop = NULL;
        return;
    }

    if (!(newtop = hlp_read(result->keyword))) {
        fprintf(stderr, "Database error: bad link\n");
        XtVaSetValues(result->button, XtNstate, 0, NULL);
        return;
    }

    result->newtop = newtop;
    newtop->sibling = parent->lastborn;
    parent->lastborn = newtop;
    newtop->parent = parent;
    newtop->xposition = parent->xposition + 50;
    newtop->yposition = parent->yposition + 50;
    if (!hlp_xdisplay(newtop)) {
        fprintf(stderr, "Couldn't open window\n");
        return;
    }
}


/* ARGSUSED */
static void
delete(w, client_data, call_data)

Widget w;
XtPointer client_data, call_data;
{
    topic *top = (topic *) client_data;
    struct hlp_toplist *tl;
    struct toplink *tp;

    if (!top->parent) {
        /* delete from top level topic list */
        for (tl = top_list; tl; tl = tl->next) {
            if (tl->top == top) {
                if (tl->prev)
                    tl->prev->next = tl->next;
                if (tl->next)
                    tl->next->prev = tl->prev;
                if (tl == top_list)
                    top_list = tl->next;
                free((char*)tl);
                break;
            }
        }
    }
    else {
        for (tp = top->parent->subtopics; tp; tp = tp->next) {
            if (tp->newtop == top) {
                tp->newtop = NULL;
                break;
            }
        }
        if (!tp) {
            for (tp = top->parent->seealso; tp; tp = tp->next) {
                if (tp->newtop == top) {
                    tp->newtop = NULL;
                    break;
                }
            }
        }
        if (tp)
            XtVaSetValues(tp->button, XtNstate, 0, NULL);
    }
    hlp_fixchildren(top);
    hlp_killfamily(top);
    if (!top_list) {
        hlp_cnt = 0;
#ifdef HAVE_SIGNAL
        if (hlp_killx)
            kill(0,SIGINT);
#endif
    }
}


/* ARGSUSED */
static void
quit(w, client_data, call_data)

Widget w;
XtPointer client_data, call_data;
{
    topic *top = (topic *) client_data;
    struct hlp_toplist *tl, *tn;

    for (tl = top_list; tl; tl = tn) {
        tn = tl->next;
        hlp_killfamily(tl->top);
        free((char*)tl);
    }
    hlp_cnt = 0;
    top_list = NULL;
    ScedESC();  /* pop out of help in sced */
#ifdef HAVE_SIGNAL
    if (hlp_killx)
        kill(0,SIGINT);
#endif
}


/* ARGSUSED */
static void
print(w, client_data, call_data)

Widget w;
XtPointer client_data, call_data;
{
    topic *top = (topic *) client_data;
    char *fname;
    char buf[BSIZE_SP], device[64];
    extern char *kw_hcopydev;

    fname = smktemp("hlp");
    if (XawAsciiSaveAsFile(XawTextGetSource(top->widgets.text), fname)) {
        if (!cp_getvar(kw_hcopydev, VT_STRING, device))
            *device = '\0';
        ft_makehardcopy(NULL, fname, "text", device, buf);
        PopUpMessage(buf, &top->widgets);

        if (*device);
            unlink(fname);
    }
    else
        PopUpMessage("Internal error: text not saved", &top->widgets);
    txfree(fname);
}


/* ARGSUSED */
static void
save(w, client_data, call_data)

Widget w;
XtPointer client_data, call_data;
{
    topic *top = (topic *) client_data;

    PopUpInput("", "save", DoSave, &top->widgets);
}

static void
DoSave(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    widget_bag *w = (widget_bag*)client_data;
    String fname;
    XawTextPosition begin, end;
    FILE *fp;
    String s, string;
    int len;
    String mesg;
    char buf[BSIZE_SP];

    XtVaGetValues(w->popup_text, XtNstring, &fname, NULL);
    if (fname && *fname) {
        strcpy(buf, fname);
        cp_pathfix(buf);
        fname = buf;
    }
    if (CheckFile(fname, W_OK, w) == NOGO)
        return;
    
    XawTextGetSelectionPos(w->text, &begin, &end);
    if (begin == end) {
        /* no selected text */
        if (!XawAsciiSaveAsFile(XawTextGetSource(w->text), fname)) {
            PopUpMessage("Unknown error, text not saved", w);
            return;
        }   
        mesg = "Text saved";
    }
    else {
        if ((fp = fopen(fname, "w")) == NULL) {
            PopUpMessage("Unknown error, block not saved", w);
            return;
        }
        string = XFetchBytes(Xdisplay, &len);
        for (s = string; len; s++,len--)
            putc(*s,fp);
        fclose(fp);
        XFree(string);
        mesg = "Selected block saved";
    }
    PopDownInput(caller, (XtPointer)w, call_data);
    PopUpMessage(mesg, w);
}


void
hlp_xkillwin(top)

topic *top;
{
    XtDestroyWidget(top->widgets.shell);
}

#else

/* ARGSUSED */ bool hlp_xdisplay(top) topic *top; { return (false); }
/* ARGSUSED */ void hlp_xkillwin(top) topic *top; { }

#endif
