/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "ftedefs.h" /* really don't need this without breakpoints */
#include "inpptree.h"
#include "iferrmsg.h"

#ifdef __STDC__
static void init_node(INPparseNode*,double,double);
static void PTsetBreaks(INPparseNode*,double);
static void PTinit(INPparseNode*,double,double);
static void free_parsenode(INPparseNode*);
#else /* stdc */
static void init_node();
static void PTsetBreaks();
static void PTinit();
static void free_parsenode();
#endif /* stdc */

extern double PTfudge_factor;

#define setbreaks

int
IFinit(tree,step,finaltime)

IFparseTree *tree;
double step, finaltime;
{
    struct INPparseTree *pt = (struct INPparseTree*)tree;
    int i;

    if (!tree)
        return;
    for (i = 0; i < tree->numVars; i++)
        init_node(pt->derivs[i],step,finaltime);
    init_node(pt->tree,step,finaltime);

    return (OK);
}


static void
init_node(p,step,finaltime)

INPparseNode *p;
double step, finaltime;
{
    if (!p) return;
    init_node(p->left,step,finaltime);
    init_node(p->right,step,finaltime);
    PTinit(p,step,finaltime);
}

#define V1    p->tranparms[0]
#define V2    p->tranparms[1]
#define TD    p->tranparms[2]
#define TR    p->tranparms[3]
#define TF    p->tranparms[4]
#define PW    p->tranparms[5]
#define PER   p->tranparms[6]

#define SPER  p->tranparms[2]
#define SDEL  p->tranparms[3]

#define VO    p->tranparms[0]
#define VA    p->tranparms[1]
#define FREQ  p->tranparms[2]
#define TDL   p->tranparms[3]
#define THETA p->tranparms[4]

#define TD1   p->tranparms[2]
#define TAU1  p->tranparms[3]
#define TD2   p->tranparms[4]
#define TAU2  p->tranparms[5]

#define FC    p->tranparms[2]
#define MDI   p->tranparms[3]
#define FS    p->tranparms[4]


#ifdef setbreaks

static void
PTsetBreaks(p,finaltime)

INPparseNode *p;
double finaltime;
{
    double time, time1;
    int i, n;

    switch (p->funcnum) {

        case PTF_tPULSE:

            time1 = TD;
            time = time1;
            while (time < finaltime) {
                if (time > 0)
                    CKTsetBreak(ft_curckt->ci_ckt,time);
                time += TR;
                if (time >= finaltime)
                    break;
                if (time > 0)
                    CKTsetBreak(ft_curckt->ci_ckt,time);
                if (PW != 0) {
                    time += PW;
                    if (time >= finaltime)
                        break;
                    if (time > 0)
                        CKTsetBreak(ft_curckt->ci_ckt,time);
                }
                if (TF != 0) {
                    time += TF;
                    if (time >= finaltime)
                        break;
                    if (time > 0)
                        CKTsetBreak(ft_curckt->ci_ckt,time);
                }
                time1 += PER;
                time = time1;
            }
            /* set for additional offsets */
            for (i = 7; i < p->numtrancoeffs; i++) {
                time1 = p->trancoeffs[i];
                time = time1;
                while (time < finaltime) {
                    if (time > 0)
                        CKTsetBreak(ft_curckt->ci_ckt,time);
                    time += TR;
                    if (time >= finaltime)
                        break;
                    if (time > 0)
                        CKTsetBreak(ft_curckt->ci_ckt,time);
                    if (PW != 0) {
                        time += PW;
                        if (time >= finaltime)
                            break;
                        if (time > 0)
                            CKTsetBreak(ft_curckt->ci_ckt,time);
                    }
                    if (TF != 0) {
                        time += TF;
                        if (time >= finaltime)
                            break;
                        if (time > 0)
                            CKTsetBreak(ft_curckt->ci_ckt,time);
                    }
                    time1 += PER;
                    time = time1;
                }
            }
            break;

        case PTF_tPWL:

            n = p->numtrancoeffs/2;
            for (i = 0; i < n; i++)
                CKTsetBreak(ft_curckt->ci_ckt,*(p->trancoeffs + i + i));
            break;
    }
}
#endif


static void
PTinit(p,step,finaltime)

INPparseNode *p;
double step, finaltime;
{
    if (p->type != PT_TFUNC)
        return;

    switch (p->funcnum) {

        case PTF_tPULSE:
            V1 = *p->trancoeffs;
            V2 = *(p->trancoeffs+1);
            TD = (p->numtrancoeffs >= 3) ?
                *(p->trancoeffs+2) : 0;
            TR = (p->numtrancoeffs >= 4) ?
                *(p->trancoeffs+3) : 0;
            TF = (p->numtrancoeffs >= 5) ?
                *(p->trancoeffs+4) : 0;
            PW = (p->numtrancoeffs >= 6) ?
                *(p->trancoeffs+5) : 0;
            PER = (p->numtrancoeffs >= 7) ?
                *(p->trancoeffs+6) : 0;
            if (TR <= 0) TR = step;
            if (TF > 0 && PW <= 0)
                PW = 0.0;
            else {
                if (TF <= 0) TF = step;
                if (PW <= 0) PW = finaltime + step;
            }
            if (PER <= 0) PER = finaltime + step;
            else if (PER < TR+PW+TF)
                PER = TR+PW+TF;
            p->trancache[0] = (V2 - V1)/TR;
            p->trancache[1] = (V1 - V2)/TF;
#ifdef setbreaks
            PTsetBreaks(p,finaltime);
#endif
            break;

#ifdef setbreaks
        case PTF_tPWL:
            PTsetBreaks(p,finaltime);
            break;
#endif

        case PTF_tSIN:
            VO = *p->trancoeffs;
            VA = *(p->trancoeffs+1);
            FREQ = (p->numtrancoeffs >= 3) ?
                *(p->trancoeffs+2) : 0;
            TDL = (p->numtrancoeffs >= 4) ?
                *(p->trancoeffs+3) : 0;
            THETA = (p->numtrancoeffs >= 5) ?
                *(p->trancoeffs+4) : 0;
            if (!FREQ) FREQ = 1/finaltime;
            break;

        case PTF_tSPULSE:
            V1 = *p->trancoeffs;
            V2 = *(p->trancoeffs+1);
            SPER = (p->numtrancoeffs >= 3) ?
                *(p->trancoeffs+2) : 0;
            SDEL = (p->numtrancoeffs >= 4) ?
                *(p->trancoeffs+3) : 0;
            THETA = (p->numtrancoeffs >= 5) ?
                *(p->trancoeffs+4) : 0;
            if (!SPER) SPER = finaltime;
            break;

        case PTF_tEXP:
            V1 = *p->trancoeffs;
            V2 = *(p->trancoeffs+1);
            TD1 = (p->numtrancoeffs >= 3) ?
                *(p->trancoeffs+2) : 0;
            TAU1 = (p->numtrancoeffs >= 4) ?
                *(p->trancoeffs+3) : 0;
            TD2 = (p->numtrancoeffs >= 5) ?
                *(p->trancoeffs+4) : 0;
            TAU2 = (p->numtrancoeffs >= 6) ?
                *(p->trancoeffs+5) : 0;
            if (!TAU2) TAU2 = step;
            if (!TD1)  TD1  = step;
            if (!TAU1) TAU1 = step;
            if (!TD2)  TD2  = TD1 + step;
            break;

        case PTF_tSFFM:
            VO = *p->trancoeffs;
            VA = *(p->trancoeffs+1);
            FC = (p->numtrancoeffs >= 3) ?
                *(p->trancoeffs+2) : 0;
            MDI = (p->numtrancoeffs >= 4) ?
                *(p->trancoeffs+3) : 0;
            FS = (p->numtrancoeffs >= 5) ?
                *(p->trancoeffs+4) : 0;
            if (!FC) FC = 1/finaltime;
            if (!FS) FS = 1/finaltime;
            break;
    }
}


int
IFeval(tree, gmin, result, vals, derivs, parms)

IFparseTree *tree;
double gmin;
double *result;
double *vals;
double *derivs;
double *parms;
{
    int i, err;
    INPparseTree *myTree = (INPparseTree*) tree;

/*
INPptPrint("calling PTeval, tree = ", myTree);
printf("values:");
for (i = 0; i < myTree->p.numVars; i++)
printf("\tvar%d = %lg\n", i, vals[i]);
*/

    PTfudge_factor = gmin;
    if ((err = (*myTree->tree->evfunc)(myTree->tree,result,vals,parms))
            != OK)
        return (err);

    for (i = 0; i < myTree->p.numVars; i++)
        if ((err = (*myTree->derivs[i]->evfunc)(myTree->derivs[i],
                &derivs[i], vals, parms)) != OK)
            return (err);

/*
printf("results: function = %lg\n", *result);
for (i = 0; i < myTree->p.numVars; i++)
printf("\td / d var%d = %lg\n", i, derivs[i]);
*/

    return (OK);
}


void
IFfree(tree)

IFparseTree *tree;
{
    struct INPparseTree *pt = (struct INPparseTree*)tree;
    int i;

    if (!tree)
        return;
    if (tree->line)
        txfree(tree->line);
    for (i = 0; i < tree->numVars; i++)
        if (tree->varTypes[i] == IF_STRING)
            txfree(tree->vars[i].sValue);
    txfree((char*)tree->varTypes);
    txfree((char*)tree->vars);
    for (i = 0; i < tree->numVars; i++)
        free_parsenode(pt->derivs[i]);
    txfree((char*)pt->derivs);
    free_parsenode(pt->tree);
    txfree((char*)tree);
}


static void
free_parsenode(p)

INPparseNode *p;
{
    if (!p) return;
    free_parsenode(p->left);
    free_parsenode(p->right);
    if (p->type == PT_TFUNC) {
        if (p->trancoeffs) txfree((char*)p->trancoeffs);
        if (p->tranparms) txfree((char*)p->tranparms);
        if (p->trancache) txfree((char*)p->trancache);
    }
    txfree((char*)p);
}
