/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

    /*
     * INPdevParse()
     *
     *  parse a given input according to the standard rules - look
     *  for the parameters given in the parmlists, In addition, 
     *  an optional leading numeric parameter is handled.
     */

#include "spice.h"
#include "inpdefs.h"
#include "iferrmsg.h"
#include "misc.h"


char *
INPdevParse(line,ckt,dev,fast,leading,waslead,tab)

char **line;     /* the line to parse */
GENERIC *ckt;    /* the circuit this device is a member of */
int dev;         /* the device type code to the device being parsed */
GENERIC *fast;   /* direct pointer to device being parsed */
double *leading; /* the optional leading numeric parameter */
int *waslead;    /* flag - 1 if leading double given, 0 otherwise */
INPtables *tab;
{
    int error;  /* int to store evaluate error return codes in */
    char *parm;
    char *errbuf;
    int i;
    IFvalue *val;
    IFdevice *device = ft_sim->devices[dev];

    /* check for leading value */
    *waslead = 0;
    *leading = INPevaluate(line,&error,1);

    if (error == 0)
        /* found a good leading number */
        *waslead = 1;
    else
        *leading = 0.0;


    while (**line != (char) 0) {
        error = INPgetTok(line,&parm,1);
        if (error)
            return (INPerror(error));

        for (i = 0; i < *device->numInstanceParms; i++) {
            if (!strcmp(parm,device->instanceParms[i].keyword)) {

                val = INPgetValue(ckt,line,
                        device->instanceParms[i].dataType,tab);
                if (!val) {
                    txfree(parm);
                    return (INPerror(E_PARMVAL));
                }
                error = (*(ft_sim->setInstanceParm))(ckt,fast,
                        device->instanceParms[i].id,
                        val,(IFvalue*)NULL);
                if (error) {
                    txfree(parm);
                    return (INPerror(error));
                }
                break;
            }
        }
        if (i == *device->numInstanceParms) {
            errbuf = tmalloc(strlen(parm)+25);
            (void)sprintf(errbuf,"Unknown parameter (%s)",parm);
            txfree(parm);
            return (errbuf);
        }
        txfree(parm);
    }
    return ((char *)NULL);
}
