/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

/* 
 *  provide the error message appropriate for the given error code
 */

#include "spice.h"
#include "ifsim.h"
#include "iferrmsg.h"
#include "misc.h"

extern IFfrontEnd *SPfrontEnd;

#define MSG(a) \
val = tmalloc(2+strlen(a));\
if(val) {\
    (void)strcpy(val,a);\
    break;\
}\
(*(SPfrontEnd->IFerror))(E_PANIC,nomem,(IFuid *)NULL);\
break;


char *
INPerror(type)

int type;
{

static char *from =         "%s detected in routine %s";
static char *unknownError = "Unknown error code %d";
static char *Pause =        "Pause requested";
static char *panic =        "Impossible error - can't occur";
static char *exists =       "Device already exists, existing one being used";
static char *nodev =        "No such device";
static char *nomod =        "No such model";
static char *noanal =       "No such analysis type";
static char *noterm =       "No such terminal on this device";
static char *badparm =      "No such parameter on this device";
static char *nomem =        "Out of Memory";
static char *nodecon =      "Warning: old connection replaced";
static char *unsupp =       "Action unsupported by this simulator";
static char *parmval =      "Parameter value is illegal";
static char *notempty =     "Can't delete - still referenced";
static char *nochange =     "Sorry, simulator can't handle that now";
static char *notfound =     "Not found";
static char *syntax =       "Parse error";

char *val;
char *tmp;

    switch(type) {
        default:
            if (errMsg) {
                val = tmalloc(strlen(errMsg)+5);
                strcpy(val,errMsg);
                txfree(errMsg);
                errMsg = NULL;
                break;
            }
            else {
                val = tmalloc(strlen(unknownError) +10);
                if (val) {
                    (void) sprintf(val,unknownError,type);
                    break;
                }
            }
            (*(SPfrontEnd->IFerror))(E_PANIC,nomem,(IFuid *)NULL);
            break;
        case E_PAUSE:
            MSG(Pause)
        case OK:
            return(NULL);
        case E_PANIC:
            MSG(panic)
        case E_EXISTS:
            MSG(exists)
        case E_NODEV:
            MSG(nodev)
        case E_NOMOD:
            MSG(nomod)
        case E_NOANAL:
            MSG(noanal)
        case E_NOTERM:
            MSG(noterm)
        case E_BADPARM:
            MSG(badparm)
        case E_NOMEM:
            MSG(nomem)
        case E_NODECON:
            MSG(nodecon)
        case E_UNSUPP:
            MSG(unsupp)
        case E_PARMVAL:
            MSG(parmval)
        case E_NOTEMPTY:
            MSG(notempty)
        case E_NOCHANGE:
            MSG(nochange)
        case E_NOTFOUND:
            MSG(notfound)
        case E_SYNTAX:
            if (errMsg) {
                val = tmalloc(strlen(syntax) + strlen(errMsg) + 5);
                (void) sprintf(val,"%s: %s",syntax,errMsg);
                txfree(errMsg);
                errMsg = NULL;
            }
            else {
                MSG(syntax)
            }
    }
    if (errRtn) {
        tmp = tmalloc(strlen(from) + strlen(val) + strlen(errRtn) + 5);
        sprintf(tmp, from, val ,errRtn);
        txfree(val);
        return (tmp);
    }
    else {
        return (val);
    }
    /*NOTREACHED*/
}
    
