/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "inpdefs.h"
#include "misc.h"

extern INPmodel *modtab;

char *
INPgetMod(ckt,name,model,tab)

GENERIC * ckt;
char *name;
INPmodel **model;
INPtables *tab;
{
    INPmodel *modtmp;
    IFvalue *val;
    int j;
    char *line;
    char *parm;
    char *err = NULL;
    char *temp;
    int error;

    for (modtmp = modtab; modtmp != (INPmodel *)NULL;
            modtmp = ((modtmp)->INPnextModel)) {
        if (strcmp((modtmp)->INPmodName,name))
            continue;

        /* found the model in question - now instantiate if necessary
         * and return an appropriate pointer to it
         */
        if (modtmp->INPmodType < 0) {
            /* illegal device type, so can't handle */
            *model = (INPmodel *)NULL;
            err = (char *)tmalloc((35+strlen(name)) * sizeof(char));
            (void) sprintf(err,
                    "Unknown device type for model %s.",name);
            return (err);
        }
        if (! ((modtmp)->INPmodUsed )) {
            /* not already defined, so create & give parameters */
            error = (*(ft_sim->newModel))( ckt,(modtmp)->INPmodType, 
                    &((modtmp)->INPmodfast), (modtmp)->INPmodName);
            if (error)
                return (INPerror(error));

            /* parameter isolation, identification, binding */
            line = ((modtmp)->INPmodLine)->line;
            INPgetTok(&line,&parm,1);     /* throw away '.model' */
            txfree(parm);
            INPgetTok(&line,&parm,1);     /* throw away 'modname' */
            txfree(parm);
            INPgetTok(&line,&parm,1);
            while (*parm != '\0') {
                for (j = 0;
                    j < (*(*(ft_sim->devices)[(modtmp)->INPmodType]).
                        numModelParms); j++) {
                    if (strcmp(parm,((*(ft_sim->devices) [ (modtmp)->
                            INPmodType ]).modelParms[j].keyword)) == 0) {
                        val = INPgetValue(ckt,&line,
                                ((*(ft_sim->devices)[(modtmp)->
                                INPmodType ]).modelParms[j].
                                dataType),tab);
                        error = (*(ft_sim->setModelParm))(ckt, 
                                ((modtmp)->INPmodfast),
                                (*(ft_sim->devices)[(modtmp)->INPmodType]).
                                modelParms[j].id,val,(IFvalue*)NULL);
                        if (error) { 
                            txfree(parm);
                            return (INPerror(error));
                        }
                        break;
                    } 
                }
                if (strcmp(parm,"level") == 0) {
                    /* just grab the level number and throw away */
                    /* since we already have that info from pass1 */
                    val = INPgetValue(ckt,&line,IF_REAL,tab);
                }
                else if (j >= 
                        (*(*(ft_sim->devices)[(modtmp)->INPmodType]).
                                numModelParms)) {
                    temp = (char *)tmalloc((40+strlen(parm)) * sizeof(char));
                    (void)sprintf(temp,
                        "Unrecognized parameter (%s) - ignored.", parm);
                    err = INPerrCat(err,temp);
                }
                txfree(parm);
                INPgetTok(&line,&parm,1);
            }
            txfree(parm);
            (modtmp)->INPmodUsed = 1;
            (modtmp)->INPmodLine->error = err;
        }
        *model = modtmp;
        return ((char *)NULL);
    }

    /* didn't find model - ERROR  - return model */
    *model = (INPmodel *)NULL;
    err = (char *)tmalloc((60+strlen(name)) * sizeof(char));
    (void) sprintf(err,
        "Unable to find definition of model %s - default assumed.",name);
    return (err);
}

