/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "inpdefs.h"
#include "iferrmsg.h"
#include "inpmacs.h"

/* pass 2 - Scan through the lines.  ".model" cards have been processed
 * in pass1 and are ignored here.
 */

void
INPpas2(ckt,data,tab,task)

GENERIC *ckt;
card *data;
INPtables *tab;
GENERIC *task;
{
    card *current;
    char c;
    char *groundname = "0";
    char *gname;
    char *thisline;
    GENERIC *gnode;
    IFdevice *dev;
    int i, j, error;  /* used by the macros defined above */

    error = INPgetTok(&groundname,&gname,1);
    if (error) data->error = INPerrCat(data->error,copy(
            "can't read internal ground node name!\n"));

    error = INPgndInsert(ckt,&gname,tab,&gnode);
    if (error && error != E_EXISTS)
        data->error = INPerrCat(data->error,copy(
            "can't insert internal ground node in symbol table!\n"));

    for (current = data; current != NULL; current = current->nextcard) {

        thisline = current->line;
        while (*thisline && ((*thisline == ' ') || (*thisline == '\t')))
            thisline++;
        c = *thisline;

        if (isalpha(c)) {
            if (isupper(c)) c = tolower(c);

            for (i = 0; i < ft_sim->numDevices; i++) {
                dev = ft_sim->devices[i];
                for (j = 0; j < *dev->numKeys; j++) {
                    if (dev->keys[j].key == c)
                        break;
                }
                if (j < *dev->numKeys) {
                    (*dev->parse)(i,ckt,tab,current);
                    break;
                }
            }
            if (i == ft_sim->numDevices) {
                LITERR("Unknown device type.")
            }
            continue;
        }

        switch (c) {

            case '*':   /* *<anything> - a comment - ignore */
            case 0:
                break;

            case '.':   /* .<something> Many possibilities */
                if (INP2dot(ckt,tab,current,task,gnode)) goto end;
                break;

            default:
                LITERR("Syntax error.")
                break;
        }
    }
end:
    INPtablFix(ckt);
    return;
}

