/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Stuff for the terminal and node symbol tables.
 * Defined: INPtabInit, INPinsert, INPtermInsert, INPtabEnd
 */

#include "spice.h"
#include "inpdefs.h"
#include "iferrmsg.h"
#include "misc.h"

#ifdef __STDC__
static int hash(char *, int);
#else /* stdc */
static int hash();
#endif /* stdc */


/* Initialize the symbol tables. */

INPtables *
INPtabInit(numlines)

int numlines;
{
    INPtables *tab;
    char *tmalloc();

    tab = (INPtables *)tmalloc(sizeof(INPtables));
    tab->INPsymtab = (struct INPtab **) tmalloc((numlines / 4 + 1) * 
                sizeof (struct INPtab *));
    tab->INPtermsymtab = (struct INPnTab **) tmalloc(numlines * 
                sizeof (struct INPnTab *));
    tab->INPsize = numlines / 4 + 1;
    tab->INPtermsize = numlines;
    return (tab);
}


/* insert 'token' into the terminal symbol table */
/* create a NEW NODE and return a pointer to it in *node */

int
INPtermInsert(ckt,token,tab,node)

GENERIC *ckt;
char **token;
INPtables *tab;
GENERIC**node;
{
    int key;
    int error;
    struct INPnTab *t;
    char *tmalloc();

    key = hash(*token, tab->INPtermsize);
    for (t = tab->INPtermsymtab[key]; t; t = t->t_next) {
        if (!strcmp(*token, t->t_ent)) {
            txfree(*token);
            *token = t->t_ent;
            if (node)
                *node = t->t_node;
            return (E_EXISTS);
        }
    }
    t = (struct INPnTab *) tmalloc(sizeof (struct INPnTab));
    if (t == (struct INPnTab*)NULL)
        return (E_NOMEM);
    error = (*(ft_sim->newNode))(ckt,&t->t_node,*token);
    if (error)
        return (error);
    if (node)
        *node = t->t_node;
    t->t_ent = *token;
    t->t_next = tab->INPtermsymtab[key];
    tab->INPtermsymtab[key] = t;
    return (OK);
}


/* insert 'token' into the terminal symbol table */
/* USE node as the node pointer */

/* ARGSUSED */
int
INPmkTerm(ckt,token,tab,node)

GENERIC *ckt;
char **token;
INPtables *tab;
GENERIC**node;
{
    int key;
    struct INPnTab *t;
    char *tmalloc();

    key = hash(*token, tab->INPtermsize);
    for (t = tab->INPtermsymtab[key]; t; t = t->t_next) {
        if (!strcmp(*token, t->t_ent)) {
            txfree(*token);
            *token = t->t_ent;
            if (node)
                *node = t->t_node;
            return (E_EXISTS);
        }
    }
    t = (struct INPnTab *) tmalloc(sizeof (struct INPnTab));
    if (t == (struct INPnTab*)NULL)
        return (E_NOMEM);
    t->t_node = *node ;
    t->t_ent = *token;
    t->t_next = tab->INPtermsymtab[key];
    tab->INPtermsymtab[key] = t;
    return (OK);
}


/* insert 'token' into the terminal symbol table as a name for ground*/

int
INPgndInsert(ckt,token,tab,node)

GENERIC *ckt;
char **token;
INPtables *tab;
GENERIC**node;
{
    int key;
    int error;
    struct INPnTab *t;
    char *tmalloc();

    key = hash(*token, tab->INPtermsize);
    for (t = tab->INPtermsymtab[key]; t; t = t->t_next) {
        if (!strcmp(*token, t->t_ent)) {
            txfree(*token);
            *token = t->t_ent;
            if (node)
                *node = t->t_node;
            return (E_EXISTS);
        }
    }
    t = (struct INPnTab *) tmalloc(sizeof (struct INPnTab));
    if (t == (struct INPnTab*)NULL)
        return(E_NOMEM);
    error = (*(ft_sim->groundNode))(ckt,&t->t_node,*token);
    if (error)
        return(error);
    if (node)
        *node = t->t_node;
    t->t_ent = *token;
    t->t_next = tab->INPtermsymtab[key];
    tab->INPtermsymtab[key] = t;
    return (OK);
}


/* insert 'token' into the symbol table */

int
INPinsert(token,tab)

char **token;
INPtables *tab;
{
    struct INPtab *t;
    int key;
    char *tmalloc();

    key = hash(*token, tab->INPsize);
    for (t = tab->INPsymtab[key]; t; t = t->t_next)
        if (!strcmp(*token, t->t_ent)) {
            txfree(*token);
            *token = t->t_ent;
            return (E_EXISTS);
        }
    t = (struct INPtab *) tmalloc(sizeof (struct INPtab));
    if (t == (struct INPtab*)NULL)
        return (E_NOMEM);
    t->t_ent = *token;
    t->t_next = tab->INPsymtab[key];
    tab->INPsymtab[key] = t;
    return (OK);
}


/* Free the space used by the symbol tables. */

void
INPtabEnd(tab)

INPtables *tab;
{
    struct INPtab *t, *lt;
    struct INPnTab *n, *ln;
    int i;
    /* SRW - used to not free t_ent, for whatever reason */

    for (i = 0; i < tab->INPsize; i++)
        for (t = tab->INPsymtab[i]; t; t = lt) {
            lt = t->t_next;
            txfree(t->t_ent);
            txfree((char*)t);
        }
    txfree((char*)tab->INPsymtab);
    for (i = 0; i < tab->INPtermsize; i++)
        for (n = tab->INPtermsymtab[i]; n; n = ln) {
            ln = n->t_next;
            txfree(n->t_ent);
            txfree((char*)n);
        }
    txfree((char*)tab->INPtermsymtab);
    txfree((char*)tab);
    return;
}


static int
hash(name, tsize)

char *name;
int tsize;
{
    char *s;
    int i = 0;

    for (s = name; *s; s++)
            i += *s;
    return (i % tsize);
}

