/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "cpdefs.h"
#define SHORT_SCEDIO
#include "scedio.h"

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_SIGNAL
#include <signal.h>
#endif
#ifdef HAVE_STAT
#include <sys/stat.h>
#endif

bool
tryexec(name, argv)

/* fork and execute */
char *name;
char *argv[];
{
#ifdef HAVE_FORK
    int status;
    int pid, j;
    RETSIGTYPE (*svint)( ), (*svquit)( ), (*svtstp)( );

    pid = fork( );
    if (pid == 0) {
        cp_fixdescriptors();
        (void) execv(name, argv);
        (void) _exit(120);  /* A random value. */
        /* NOTREACHED */
    }
    else {
        svint = signal(SIGINT, SIG_DFL);
        svquit = signal(SIGQUIT, SIG_DFL);
        svtstp = signal(SIGTSTP, SIG_DFL);
        do {
            j = wait(&status);
        } while (j != pid);
        (void) signal(SIGINT, svint);
        (void) signal(SIGQUIT, svquit);
        (void) signal(SIGTSTP, svtstp);
    }
    if (WIFEXITED(status) && WEXITSTATUS(status) == 120)
        return (false);
    return (true);
#else
    char *s;
    wordlist *wl;

    argv++;
    wl = alloc(wordlist);
    wl->wl_word = copy(name);
    if (*argv) {
        wl->wl_next = wl_build(argv);
        wl->wl_next->wl_prev = wl;
    }
    s = wl_flatten(wl);
    wl_free(wl);
    (void) system(s);
    putc('\n',stdout);
    tfree(s);
    return (true);

#endif
}


bool
is_exec(file,dir)

/* return true if the file can be executed */
char *file, *dir;
{
#ifdef MSDOS
    char *c;

    if ((c = index(file,'.')) == NULL)
        return (false);
    *c = '\0';
    c++;
    if (!cieq(c,"exe") && !cieq(c,"com") && !cieq(c,"bat"))
        return (false);
#else
    char buf[BSIZE_SP];
#ifdef HAVE_STAT
    struct stat stbuf;

    (void)sprintf(buf,"%s%c%s",dir,DIR_TERM,file);
    if (stat(buf,&stbuf) < 0)
        return (false);
    if (!(stbuf.st_mode & S_IFREG) || !(stbuf.st_mode & S_IXUSR))
        return (false);
#else
    (void)sprintf(buf,"%s%c%s",dir,DIR_TERM,file);
    if (access(buf,1))
        return (false);
#endif
#endif
    return (true);
}
