/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cpstd.h"

#define HTAB_SIZE 256

struct s_hashlist {
    char *h_name;
    void *h_what;
    struct s_hashlist *h_next;
};
typedef struct s_hashlist hashlist;

#define HASH(s,n)  {int j=0; n=0; while(s[j]) n+=s[j++]; n%=HTAB_SIZE;}

#define COMP(ss,tt) \
    {char *s=ss, *t=tt; while(*s && *s == *t){s++; t++;} i = *s - *t;}


void *
htab_init()
{
    hashlist **h;

    h = (hashlist **)tmalloc(HTAB_SIZE * sizeof(hashlist*));
    return ((void*)h);
}


void
htab_add(name,data,listp)

char *name;
void *data;
void *listp;
{
    hashlist **list = (hashlist **)listp;
    hashlist *h, *hh;
    int n, i;

    if (list == NULL)
        return;
    HASH(name,n);
    h = list[n];

    if (h == NULL) {
        h = list[n] = (hashlist*)tmalloc(sizeof(hashlist));
        h->h_name = copy(name);
        h->h_what = data;
    }
    else {
        for (hh = NULL; h; hh = h, h = h->h_next) {
            COMP(name,h->h_name);
            /*
            i = strcmp(name,h->h_name);
            */
            if (i < 0)
                continue;
            if (i == 0)
                return;
            if (hh) {
                hh->h_next = (hashlist*)tmalloc(sizeof(hashlist));
                hh = hh->h_next;
            }
            else {
                hh = (hashlist*)tmalloc(sizeof(hashlist));
                list[n] = hh;
            }
            hh->h_name = copy(name);
            hh->h_what = data;
            hh->h_next = h;
            return;
        }
        hh->h_next = (hashlist*)tmalloc(sizeof(hashlist));
        hh = hh->h_next;
        hh->h_name = copy(name);
        hh->h_what = data;
        hh->h_next = h;
    }
}


void
htab_delete(name,listp)

char *name;
void *listp;
{
    hashlist **list = (hashlist**)listp;
    hashlist *h, *hh;
    int n, i;

    if (list == NULL)
        return;
    HASH(name,n);
    h = list[n];

    if (h) {
        for (hh = NULL; h; hh = h, h = h->h_next) {
            COMP(name,h->h_name);
            /*
            i = strcmp(name,h->h_name);
            */
            if (i < 0)
                continue;
            if (i == 0) {
                if (hh)
                    hh->h_next = h->h_next;
                else
                    list[n] = h->h_next;
                txfree(h->h_name);
                txfree((char*)h);
                return;
            }
            return;
        }
    }
}


void *
htab_get(name,listp)

char *name;
void *listp;
{
    hashlist **list = (hashlist **)listp;
    hashlist *h, *hh;
    int n, i;

    if (list == NULL)
        return (NULL);
    HASH(name,n);
    h = list[n];

    if (h) {
        for (hh = NULL; h; hh = h, h = h->h_next) {
            COMP(name,h->h_name);
            /*
            i = strcmp(name,h->h_name);
            */
            if (i < 0)
                continue;
            if (i == 0) {
                return (h->h_what);
            }
            return (NULL);
        }
    }
    return (NULL);
}


void
htab_free(listp,freedata)

void *listp;
int freedata;
{
    hashlist **list = (hashlist **)listp;
    hashlist *h, *hh;
    int i;

    if (list == NULL)
        return;

    for (i = 0; i < HTAB_SIZE; i++) {
        for (h = list[i]; h; h = hh) {
            hh = h->h_next;
            txfree(h->h_name);
            if (freedata) txfree((char*)h->h_what);
            txfree((char*)h);
        }
        list[i] = NULL;
    }
}


void
htab_print(listp,datafmt)

void *listp;
char *datafmt;
{
    hashlist **list = (hashlist **)listp;
    hashlist *h;
    int i;

    if (list == NULL)
        return;
    if (datafmt == NULL)
        datafmt = "hash=%d name=%s\n";
    for (i = 0; i < HTAB_SIZE; i++) {
        for (h = list[i]; h; h = h->h_next) {
            fprintf(stderr, datafmt, i, h->h_name, h->h_what);
        }
    }
}


void *
htab_wl(listp)

void *listp;
{
    hashlist **list = (hashlist **)listp;
    hashlist *h;
    wordlist *wl, *wl0 = NULL;
    int i;

    if (list == NULL)
        return (NULL);
    for (i = 0; i < HTAB_SIZE; i++) {
        for (h = list[i]; h; h = h->h_next) {
            if (wl0 == NULL)
                wl0 = wl = (wordlist*)tmalloc(sizeof(wordlist));
            else {
                wl->wl_next = (wordlist*)tmalloc(sizeof(wordlist));
                wl->wl_next->wl_prev = wl;
                wl = wl->wl_next;
            }
            wl->wl_word = copy(h->h_name);
        }
    }
    return ((void*)wl0);
}


void *
htab_list(listp)

void *listp;
{
    hashlist **list = (hashlist **)listp;
    hashlist *h;
    wordlist *wl, *wl0 = NULL;
    int i;
    char **s;

    if (list == NULL)
        return (NULL);
    for (i = 0; i < HTAB_SIZE; i++) {
        for (h = list[i]; h; h = h->h_next) {
            if (wl0 == NULL)
                wl0 = wl = (wordlist*)tmalloc(sizeof(wordlist));
            else {
                wl->wl_next = (wordlist*)tmalloc(sizeof(wordlist));
                wl->wl_next->wl_prev = wl;
                wl = wl->wl_next;
            }
            s = (char**)tmalloc(2*sizeof(char*));
            s[0] = h->h_name;
            s[1] = (char*)h->h_what;
            wl->wl_word = (char*)s;
        }
    }
    return ((void*)wl0);
}


int
htab_empty(listp)

void *listp;
{
    hashlist **list = (hashlist **)listp;
    int i;

    if (list == NULL)
        return (1);
    for (i = 0; i < HTAB_SIZE; i++)
        if (list[i])
            return (0);
    return (1);
}
