/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors:  1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "plotdefs.h"

#ifndef MSDOS

static void
message(instr, outstr)

char *instr, *outstr;
{
    if (!outstr)
        fprintf(cp_out, "%s\n", instr);
    else
        strcpy(outstr, instr);
}


int
ft_makehardcopy(graph, fname, devtype, device, mesg)

GRAPH *graph;
char *fname, *devtype, *device, *mesg;
{
    bool printed = false;
    char buf[BSIZE_SP];
    char *sif = "%s saved in file \"%s\"";

    if (*device) {
        sprintf(buf, "Printing %s on the %s printer", fname, device);
        message(buf, mesg);

#ifdef SYSTEM_PLOT5LPR
        if (!strcmp(devtype, "plot5")) {
            (void) sprintf(buf, SYSTEM_PLOT5LPR, device, fname);
            (void) system(buf);
            printed = true;
        }
#endif

#ifdef SYSTEM_PSLPR
        if (!strcmp(devtype, "postscript")) {
            (void) sprintf(buf, SYSTEM_PSLPR, device, fname);
            (void) system(buf);
            printed = true;
        }
#endif
        if (!strcmp(devtype, "laser")) {
            (void) sprintf(buf, "PCL not supported yet");
            message(buf, mesg);
            printed = false;
        }
        if (!strcmp(devtype, "text")) {
            (void) sprintf(buf, "lpr -h -P%s %s", device, fname);
            (void) system(buf);
            printed = true;
        }
    }

    if (!printed) {
        if (!strcmp(devtype, "plot5")) {
            sprintf(buf, sif, "Unix \"plot(1)\"", fname);
        }
        else if (!strcmp(devtype, "postscript")) {
            sprintf(buf, sif, "Postscript image", fname);
        }
        else if (!strcmp(devtype, "laser")) {
            sprintf(buf, sif, "PCL image", fname);
        }
        else if (!strcmp(devtype, "text")) {
            sprintf(buf, sif, "Text", fname);
        }
        message(buf, mesg);
    }
    return (0);
}

#else /* MSDOS */

#include <dos.h>
#include <sys/stat.h>

/* MSDOS routines to make hardcopies */

#ifdef __STDC__
static void errmsg(GRAPH*,char*,char*);
static int  port_status(int);
static int  dump(GRAPH*,char*,long,FILE*);
static int  input(GRAPH*);
#else
static void errmsg();
static int  port_status();
static int  dump();
static int  input();
#endif

static char *hmsg = "hit any key to continue";


int
ft_makehardcopy(graph, fname, devtype, device, mesg)

/* returns 1 on failure */
GRAPH *graph;
char *fname, *devtype, *device, *mesg;
{
    int i, j, hc_port, er;
    char hc_device[16];
    char buf[BSIZE_SP];
    struct stat b;
    FILE *fp;

    /* fname is the name of the file which contains the plot data,
     * and must not be null.
     * devtype contains the name of the device type the data in fname are
     * formatted for.
     * device contains the name of the output device or file, defaults to
     * "prn" if null.
     */

    if (graph)
        PushGraphContext(graph);

    if (*device) {
        if (cieq(device,"prn") || cieq(device,"lpt1"))
            hc_port = 0;
        else if (cieq(device,"lpt2"))
            hc_port = 1;
        else if (cieq(device,"lpt3"))
            hc_port = 2;
        else if (cieq(device,"lpt4"))
            hc_port = 3;
        else
            hc_port = -1;

        for (i = 0; i < 8 && device[i] && device[i] != '.'; i++)
            hc_device[i] = device[i];
        if (device[i] == '.') {
            hc_device[i] = device[i];
            for (j = 0, i++; j < 3 && device[i]; i++, j++)
                hc_device[i] = device[i];
        }
        hc_device[i] = '\0';
    }
    else {
        strcpy(hc_device,"prn");
        hc_port = 0;
    }

    if (hc_port >= 0) {
        while (port_status(hc_port)) {
            errmsg(graph,"Error: Printer is not accessible.",
                "Hit any key to continue, q to abort:");
            if (input(graph) == 'q') {
                goto bad;
            }
        }
    }
    if (stat(fname,&b)) {
        sprintf(buf,"Error: File \"%s\", can't stat.\n",fname);
        errmsg(graph,buf,graph ? hmsg : NULL);
        if (graph)
            (void)input(graph);
        goto bad;
    }

    fp = fopen(fname,"rb");
    if (!fp) {
        sprintf(buf,"Error: File \"%s\" not found.\n",fname);
        errmsg(graph,buf,graph ? hmsg : NULL);
        if (graph)
            (void)input(graph);
        goto bad;
    }
    er = dump(graph,hc_device,(long) b.st_size,fp);
    fclose(fp);
    if (graph)
        PopGraphContext();
    return (er);

bad:

    if (eq(devtype, "plot5")) {
        sprintf(buf, "File \"%s\" is in the Unix ", fname);
        errmsg(graph,buf,"\"plot(1)\" format.\n");
    }
    else if (eq(devtype, "postscript")) {
        sprintf(buf, "File \"%s\" may be printed on a ", fname);
        errmsg(graph,buf,"postscript printer.\n");
    }
    else if (eq(devtype, "laser")) {
        sprintf(buf, "File \"%s\" may be printed on a ", fname);
        errmsg(graph,buf,"pcl laser printer.\n");
    }
    else if (eq(devtype, "text")) {
        sprintf(buf, "File \"%s\" may be printed on a ", fname);
        errmsg(graph,buf,"text printer.\n");
    }
    if (graph)
        PopGraphContext();
    return (1);
}


void
errmsg(graph, msg1, msg2)

GRAPH *graph;
char *msg1, *msg2;
{
    int x,y,w,h;
    int l,r,t,b;

    if (graph == NULL) {
        if (msg1)
            fprintf(cp_err,msg1);
        if (msg2)
            fprintf(cp_err,msg2);
        fflush(cp_err);
        return;
    }
    w = graph->fontwidth;
    h = graph->fontheight;
    x = graph->absolute.width/2;
    y = graph->absolute.height/2;
    l = x - 19*w;
    r = x + 19*w;
    t = y + 2*h;
    b = y - 2*h;

    DevSetColor(0);
    DevBox(l,b,r,t);
    DevSetColor(1);
    DevLine(l,b,r,b);
    DevLine(r,b,r,t);
    DevLine(r,t,l,t);
    DevLine(l,t,l,b);
    DevSetColor(2);
    if (msg1)
        DevText(msg1,l+w,t-2*h);
    if (msg2)
        DevText(msg2,l+w,t-3*h);
}


static int
port_status(which)

int which;
{
    int i;
    union REGS r;

    r.x.ax = 0x0200;
    r.x.dx = which;
    int86(0x17,&r,&r);
    i = r.x.ax;
    i >>= 8;
    /* bit 5 out of paper, bit 3 io error, bit 0 timed out */
    if (i & 0x29) return (1);
    return (0);
}


static int
dump(graph,device,size,fp)

/* dump size bytes from fp to parallel port */
GRAPH *graph;
char *device;
long size;
FILE *fp;
{
    int i;
    unsigned char *c;
    static char *buf;
    int bsize = 8192;
    FILE *op;
    union REGS r;

    if (!buf) buf = (char *) malloc(bsize);
    if (!buf) {
        errmsg(graph,"Error: Out of memory.\n",graph ? hmsg : NULL);
        if (graph)
            (void)input(graph);
        return (1);
    }

    op = fopen(device,"wb");
    if (!op) {
        sprintf(buf,"Error: Printer driver %s not found.\n",device);
        errmsg(graph,buf,graph ? hmsg : NULL);
        if (graph)
            (void)input(graph);
        return (1);
    }
    errmsg(graph,"Printing ...",NULL);

    if (bsize > size) bsize = size;

    /* put driver in raw mode */
    r.x.bx = fileno(op);
    r.x.dx = 0xa0;
    r.x.ax = 0x4401;
    int86(0x21,&r,&r);

    while (size >= bsize) {
        fread(buf,bsize,1,fp);
        for (i = 0; i < bsize; i++)
            putc(buf[i],op);
        size -= bsize;
    }
    if (size) {
        fread(buf,size,1,fp);
        for (i = 0; i < size; i++)
            putc(buf[i],op);
    }
    errmsg(graph,"Done",graph ? NULL : "\n");
    fclose(op);
    return (0);
}


static int
input(graph)

GRAPH *graph;
{
    REQUEST request;
    RESPONSE response;

    if (!graph) {
        response.reply.ch = getch() & 0xff;
        if (response.reply.ch == 0)
            response.reply.ch = 256 + getch();
        fprintf(cp_err,"\n");
    }
    else {
        request.option = char_option;
        request.fp = (FILE*)0;
        DevInput(&request, &response);
    }
    return (response.reply.ch);
}

#endif
