/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1994 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <errno.h>
#include "ftedefs.h"
#include "ftegraph.h"

#ifndef MSDOS

/* A special 'getc' so that we can deal with ^D properly. There is no way for
 * stdio to know if we have typed a ^D after some other characters, so
 * don't use buffering at all.
 *
 * Also, we can't use getc() from the keyboard with X, as it will block
 * twice, since X11_Input() blocks on the CR.
 */
int
cp_inchar(fp)

FILE *fp;
{
    char c;
    int i;

    if (cp_interactive) {
        do {
            i = read((int) fileno(fp), &c, 1);
        } while (i == -1 && errno == EINTR);

        if (i == 0 || c == '\004')
            return (EOF);
        else if (i == -1) {
            perror("read");
            return (EOF);
        }
    }
    else {
        c = getc(fp);
    }
    return ((int) c);
}

#else /* MSDOS */

#include <dos.h>

/* Input routines that support DOS command completion. Used in lexical.c
 * and complete.c.  TIOCSTI must be defined.
 * These routines are compiler specific.
 */

#if __NDPC__

static bool init_ioctl;
static bool init_alloc;

void
ioctl(fd,mode,c)

int fd;
int mode;
char *c;
{
    /* add character to input queue and maybe print it */
    if (mode != TIOCSTI) return;
    if (init_ioctl) {
        init_alloc = true;
        init_ioctl = false;
        if (!_iob[fd].curbuf->base) {
            _iob[fd].curbuf->base = (unsigned char *) tmalloc(BUFSIZ);
            _iob[fd].curbuf->lastwritable = _iob[fd].curbuf->base + BUFSIZ;
        }
        _iob[fd].curbuf->lastvalid = _iob[fd].curbuf->base;
        _iob[fd].curbuf->firstvalid = _iob[fd].curbuf->base;
        _iob[fd].oldnext = _iob[fd].next = _iob[fd].curbuf->base;
        _iob[fd].lastpos = _iob[fd].curpos = 0L;
    }
    if (fd == 0) {
        putc(*c,stdout);
        fflush(stdout);
    }
    *((_iob[fd].curbuf)->lastvalid)++ = *c;
}


int
cp_inchar(fp)

/* getc() that works with ioctl above, NDPC only */

FILE *fp;
{
    short c;

    if (!cp_interactive)
        return (getc(fp));

    if (!init_alloc && fp->next >= fp->curbuf->firstvalid &&
        fp->next < fp->curbuf->lastvalid)
        return *fp->next++;
    if (!init_alloc) {
        init_alloc = true;
        if (!fp->curbuf->base) {
            fp->curbuf->base = (unsigned char *) tmalloc(BUFSIZ);
            fp->curbuf->lastwritable = fp->curbuf->base + BUFSIZ;
        }
        fp->curbuf->firstvalid = fp->curbuf->base;
        fp->curbuf->lastvalid = fp->curbuf->base;
        fp->oldnext = fp->next = fp->curbuf->base;
        fp->lastpos = fp->curpos = 0L;
    }
    for (;;) {
        while (!(c = inkey$())) ;
        switch (c) {
            case 4:  /* ^D */
            case 27: /* ESC */
                init_ioctl = true;
            case 1:  /* ^A */
            case 26: /* ^Z */
                if (c == 26) c = EOF;
                *fp->curbuf->lastvalid++ = c;
                init_alloc = false;
                return *fp->next++;
            case 8:  /* ^H (backspace) */
                if (fp->curbuf->lastvalid >
                        fp->curbuf->firstvalid) {
                    fp->curbuf->lastvalid--;
                    if (!fileno(fp)) {
                        putc(c,stdout);
                        putc(' ',stdout);
                        putc(c,stdout);
                        fflush(stdout);
                    }
                }
                break;
            case 13: /* ENTER */
                *fp->curbuf->lastvalid++ = '\n';
                if (!fileno(fp)) {
                    putc('\n',stdout);
                    fflush(stdout);
                }
                init_alloc = false;
                return *fp->next++;
            case 21: /* ^U */
                while (fp->curbuf->lastvalid >
                        fp->curbuf->firstvalid) {
                    fp->curbuf->lastvalid--;
                    if (!fileno(fp)) {
                        putc(8,stdout);
                        putc(' ',stdout);
                        putc(8,stdout);
                    }
                }
                if (!fileno(fp)) {
                    fflush(stdout);
                }
                break;
            default:
                if (isprint(c)) {
                    *fp->curbuf->lastvalid++ = c;
                    if (!fileno(fp)) {
                        putc(c,stdout);
                        fflush(stdout);
                    }
                    break;
                }
        }
    }
}

#else
#if __GNUC__

static bool init_ioctl;
static bool init_alloc;


void
ioctl(fd,mode,c)

int fd;
int mode;
char *c;
{
    /* add character to input queue and maybe print it */
    if (mode != TIOCSTI) return;
    if (init_ioctl) {
        init_alloc = true;
        init_ioctl = false;
        if (!_iob[fd]._base) {
            _iob[fd]._base = tmalloc(BUFSIZ);
            _iob[fd]._bufsiz = BUFSIZ;
        }
        _iob[fd]._cnt = 0;
        _iob[fd]._ptr = _iob[fd]._base;
    }
    if (fd == 0) {
        putc(*c,stdout);
        fflush(stdout);
    }
    *(_iob[fd]._base + _iob[fd]._cnt) = *c;
    (_iob[fd]._cnt)++;
}


int
cp_inchar(fp)

/* getc() that works with ioctl above, GCC only */

FILE *fp;
{
    short c;

    if (!cp_interactive)
        return (getc(fp));

    if (!init_alloc && fp->_cnt > 0) {
        fp->_cnt--;
        return ((int)(*(unsigned char*)(fp)->_ptr++));
    }
    if (!init_alloc) {
        init_alloc = true;
        if (!fp->_base) {
            fp->_base = tmalloc(BUFSIZ);
            fp->_bufsiz = BUFSIZ;
        }
        fp->_cnt = 0;
        fp->_ptr = fp->_base;
    }
    for (;;) {
        c = getch() & 0xff;
        if (c == 0)
            c = 256 + getch();
        switch (c) {
            case 4:  /* ^D */
            case 27: /* ESC */
                init_ioctl = true;
            case 1:  /* ^A */
            case 26: /* ^Z */
                if (c == 26) c = EOF;
                *(fp->_base + fp->_cnt) = c;
                init_alloc = false;
                return ((int)(*(unsigned char*)(fp)->_ptr++));
            case 8:  /* ^H (backspace) */
                if (fp->_cnt) {
                    fp->_cnt--;
                    if (!fileno(fp)) {
                        putc(c,stdout);
                        putc(' ',stdout);
                        putc(c,stdout);
                        fflush(stdout);
                    }
                }
                break;
            case 13: /* ENTER */
                *(fp->_base + fp->_cnt) = '\n';
                if (!fileno(fp)) {
                    putc('\n',stdout);
                    fflush(stdout);
                }
                init_alloc = false;
                return ((int)(*(unsigned char*)(fp)->_ptr++));
            case 21: /* ^U */
                while (fp->_cnt) {
                    fp->_cnt--;
                    if (!fileno(fp)) {
                        putc(8,stdout);
                        putc(' ',stdout);
                        putc(8,stdout);
                    }
                }
                if (!fileno(fp)) {
                    fflush(stdout);
                }
                fp->_ptr = fp->_base;
                break;
            default:
                if (isprint(c)) {
                    *(fp->_base + fp->_cnt) = c;
                    fp->_cnt++;
                    if (!fileno(fp)) {
                        putc(c,stdout);
                        fflush(stdout);
                    }
                    break;
                }
        }
    }
}

#endif
#endif
#endif
