/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: UCB CAD Group
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "misc.h"
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifndef HAVE_POPEN
static char *tfile;
/* This is not reentrant! */

/* ARGSUSED */
FILE *
popen(cmd,mode)

const char *cmd, *mode;
{
    char buf[BSIZE_SP];
    FILE *fp;

    tfile = smktemp("po");
    fp = fopen(tfile,"w");
    if (fp != NULL) {
        (void)sprintf(buf, "%s > %s",cmd,tfile);
        (void)system(buf);
        (void)fclose(fp);
        fp = fopen(tfile,"r");
        return (fp);
    }
    return (NULL);
}


int
pclose(fp)

FILE *fp;
{
    if (fp)
        (void)fclose(fp);
    if (tfile) {
        (void)unlink(tfile);
        tfree(tfile);
    }
    return (0);
}   
#endif


#ifndef HAVE_PERROR

void
perror(str)

char *str;
{
    char *m = "error occurred\n";
    if (str && *str)
        (void)fprintf(stderr,"%s: %s",str,m);
    else
        (void)fprintf(stderr,"%s",m);
}
#endif


#ifndef HAVE_ACCESS
/* ARGSUSED */
int
access(pth,m)
char *pth;
int m;
{
    return (-1);
}
#endif

#ifndef HAVE_SYSTEM
int
system(str)
char *str;
{
    fprintf(stderr, "Can not execute: %s\n",str);
    return (-1);
}
#endif

#ifndef HAVE_UNLINK
/* ARGSUSED */
int
unlink(fn)
char *fn;
{
    return (-1);
}
#endif

#ifndef HAVE_GETENV
/* ARGSUSED */
char *
getenv(c)
char *c;
{
    return (NULL);
}
#endif

#ifndef HAVE_CLEARERR
/* ARGSUSED */
void
clearerr(fp)
FILE *fp;
{}
#endif

#ifndef HAVE_GETCWD
/* ARGSUSED */
char *
getcwd(buf,size)
char *buf;
int size;
{
    if (buf == NULL)
        buf = tmalloc(4);
    *buf = '\0';
    return (buf);
}
#endif

#ifndef HAVE_GETPID
int
getpid()
{
    return (100);
}
#endif


#ifndef HAVE_DUP2
#ifdef HAVE_FCNTL_H

int
dup2(oldd, newd)

int oldd, newd;
{
    close(newd);
#ifdef F_DUPFD
    (void) fcntl(oldd, F_DUPFD, newd);
#endif
    return 0;
}
#endif
#endif


/*
 * A more portable version of the standard "mktemp( )" function
 */
#ifndef TEMPFORMAT
#define TEMPFORMAT "temp%s%d"
#endif

char *
smktemp(id)

char *id;
{
    char rbuf[513];
    char *nbuf;
    static int num;

    if (num == 0) {
        num = getpid();
    }
    else
        num++;

    if (!id)
    id = "sp";

    sprintf(rbuf, TEMPFORMAT, id, num);
    nbuf = (char *) malloc(strlen(rbuf) + 1);
    strcpy(nbuf, rbuf);

    return (nbuf);
}


/*
 * Missing math functions
 */

#ifdef notused
double
logb(x)

double x;
{
    double y = 0.0;

    if (x != 0.0) {
        if (x < 0.0)
            x = - x;
        while (x > 2.0) {
            y += 1.0;
            x /= 2.0;
        }
        while (x < 1.0) {
            y -= 1.0;
            x *= 2.0;
        }
    }
    else
        y = 0.0;

    return (y);
}


double
scalb(x, n)

double x;
int n;
{
    double y, z = 1.0, k = 2.0;

    if (n < 0) {
        n = -n;
        k = 0.5;
    }

    if (x != 0.0)
        for (y = 1.0; n; n >>= 1) {
            y *= k;
            if (n & 1)
                z *= y;
        }

    return (x * z);
}
#endif


#ifndef HAVE_ERFC
double
erfc(x)
double x;
{
/* Returns the complementary error function with fractional error
 * everywhere less that 1.2e-7.  From Numerical Recipes in C.
 */
    double t, z, ans;

    z = fabs(x);
    t = 1.0/(1.0 + 0.5*x);
    ans = t*exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+t*(0.09678418+
        t*(-0.18628806+t*(0.27886807+t*(-1.13520398+t*(1.48851587+
        t*(-0.82215223+t*0.17087277)))))))));
    return (x >= 0 ? ans : 2.0 - ans);
}
#endif
