/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: UCB CAD Group
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * String functions
 */

#include "spice.h"
#include "misc.h"


/***************************************************************************/
/* true if p is a leading substring of s.
 */

int
prefix(p, s)

char *p, *s;
{
    if (p == NULL || s == NULL)
        return (false);
    while (*p && (*p == *s))
        p++, s++;
    if (!*p)
        return (true);
    else
        return (false);
}


/***************************************************************************/
/* Create a copy of a string.
 */

char *
copy(str)

char *str;
{
    char *p;
    
    if (str == NULL)
        return (NULL);
    p = tmalloc(strlen(str) + 1);
    (void) strcpy(p, str);
    return(p);
}


/***************************************************************************/
/* Determine whether sub is a substring of str.
 */

int
substring(sub, str)

char *str, *sub;
{
    char *s;

    if (str == NULL || sub == NULL)
        return (false);
    while (*str) {
        if (*str == *sub) {
            for (s = sub; *s; s++) {
                if (!*str || (*s != *str++))
                    break;
            }
            if (*s == '\0')
                return (true);
        }
        str++;
    }
    return (false);
}


/***************************************************************************/
/* Append one character to a string. Don't check for overflow.
 */

void
appendc(s, c)

char *s, c;
{
    if (s) {
        while (*s)
            s++;
        *s++ = c;
        *s = '\0';
    }
    return;
}


/***************************************************************************/
/* Try to identify an integer that begins a string. Stop when a non-
 * numeric character is reached.
 */

int
scannum(str)

char *str;
{
    int i = 0;

    if (str)
        while (isdigit(*str))
            i = i * 10 + *(str++) - '0';
    return(i);
}


/***************************************************************************/
/* Case insensitive str eq.
 */

int
cieq(p, s)

char *p, *s;
{
    if (p == NULL || s == NULL)
        return (false);
    while (*p) {
        if ((isupper(*p) ? tolower(*p) : *p) !=
            (isupper(*s) ? tolower(*s) : *s))
            return (false);
        p++;
        s++;
    }
    return (*s ? false : true);
}


/***************************************************************************/
/* Case insensitive prefix.
 */

int
ciprefix(p, s)

char *p, *s;
{
    if (p == NULL || s == NULL)
        return (false);
    while (*p) {
        if ((isupper(*p) ? tolower(*p) : *p) !=
            (isupper(*s) ? tolower(*s) : *s))
            return (false);
        p++;
        s++;
    }
    return (true);
}


/***************************************************************************/
/* Convert str to lower case.
 */

void
strtolower(str)

char *str;
{
    if (str)
        while (*str) {
            *str = tolower(*str);
            str++;
        }
}


/***************************************************************************/
/* Return a malloc'ed token, advance the pointer to next token.
 */

char *
gettok(s)

char **s;
{
    char buf[BSIZE_SP];
    int i = 0;

    if (s == NULL || *s == NULL)
        return (NULL);
    while (isspace(**s))
        (*s)++;
    if (!**s)
        return (NULL);
    while (**s && !isspace(**s))
        buf[i++] = *(*s)++;
    buf[i] = '\0';
    while (isspace(**s))
        (*s)++;
    return (copy(buf));
}


/***************************************************************************/
/* Copy token to caller's buffer dst, advance the pointer to next token.
 * Returns true if token is valid.
 */

int
copytok(dst,s)

char *dst,**s;
{
    if (s == NULL || *s == NULL)
        return (0);
    while (isspace(**s))
        (*s)++;
    if (!**s)
        return (0);
    while (**s && !isspace(**s))
        *dst++ = *(*s)++;
    *dst = '\0';
    while (isspace(**s))
        (*s)++;
    return (1);
}


/***************************************************************************/
/* Copy token to caller's buffer dst, advance the pointer to next token.
 * Leave a space after the copied token.
 * Returns true if token is valid.
 */

int
copytok1(dst,s)

char *dst,**s;
{
    if (s == NULL || *s == NULL)
        return (0);
    while (isspace(**s))
        (*s)++;
    if (!**s)
        return (0);
    while (**s && !isspace(**s))
        *dst++ = *(*s)++;
    *dst++ = ' ';
    *dst++ = '\0';
    while (isspace(**s))
        (*s)++;
    return (1);
}


/***************************************************************************/
/* Advance the pointer to next token.
 */

void
advtok(s)

char **s;
{
    int i = 0;

    if (s == NULL || *s == NULL)
        return;
    while (isspace(**s))
        (*s)++;
    if (!**s)
        return;
    while (**s && !isspace(**s))
        (*s)++;
    while (isspace(**s))
        (*s)++;
    return;
}


/***************************************************************************/
/* Compatibility routines.                                                 */
/***************************************************************************/
#ifndef HAVE_INDEX

#ifndef index

char *
index(s, c)

const char *s;
char c;
{
    while ((*s != c) && (*s != '\0'))
        s++;
    if (*s == '\0')
        return ((char *) 0);
    else
        return ((char*)s);
}

#endif


#ifndef rindex

char *
rindex(s, c)

const char *s;
char c;
{
    char *t;

    for (t = (char*)s; *t != '\0'; t++);
    while ((*t != c) && (t != (char*)s))
        t--;
    if (t == (char*)s)
        return ((char *) 0);
    else
        return (t);
}

#endif
#endif


/***************************************************************************/
#ifndef HAVE_BCOPY

#ifndef bcopy

void
bcopy(from, to, num)

char *from, *to;
int num;
{
    while (num-- > 0)
        *to++ = *from++;
    return;
}

#endif


#ifndef bzero

/* can't declare void here, because we've already used it in this file */
/* and haven't declared it void before the use */
int
bzero(ptr, num)

char *ptr;
int num;
{
    while (num-- > 0)
        *ptr++ = '\0';
    return (0);
}

#endif
#endif
/***************************************************************************/

/* This might not be around.  If not then forget about sorting. */

#ifndef HAVE_QSORT
#ifndef qsort
qsort() {}
#endif
#endif

/***************************************************************************/
