/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: UCB CAD Group
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Date and time utility functions
 */

#include "spice.h"
#include "misc.h"

#ifdef HAVE_GETTIMEOFDAY
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <sys/resource.h>
#else
#ifdef HAVE_TIMES
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>
#else
#ifdef HAVE_FTIME
/* default to ftime if we can't get real CPU times */
#include <sys/types.h>
#include <sys/timeb.h>
#endif
#endif
#endif

/* Return the date. Return value is static data. */

char *
datestring()

{

#if defined(HAVE_GETTIMEOFDAY) || defined(HAVE_TIMES)
    struct tm *tp;
    static char tbuf[40];
    char *ap;
    int i;
#ifdef HAVE_GETTIMEOFDAY
    struct timeval tv;
    struct timezone tz;

    (void) gettimeofday(&tv, &tz);
    tp = localtime(&tv.tv_sec);
    ap = asctime(tp);
#else
    long tloc;

    time(&tloc);
    tp = localtime(&tloc);
    ap = asctime(tp);
#endif

    strcpy(tbuf,ap);
    i = strlen(tbuf);
    tbuf[i - 1] = '\0';
    return (tbuf);

#else

    return ("today");

#endif
}

/* return time interval in seconds and milliseconds */

#ifndef HAVE_GETTIMEOFDAY
#ifndef HAVE_TIMES
#ifdef HAVE_FTIME

struct timeb timebegin;

void
timediff(now, begin, sec, msec)
struct timeb *now, *begin;
int *sec, *msec;
{
    *msec = now->millitm - begin->millitm;
    *sec = now->time - begin->time;
    if (*msec < 0) {
        *msec += 1000;
        (*sec)--;
    }
    return;
}

#endif
#endif
#endif


/* How many seconds have elapsed in running time. */

double
seconds()

{
#ifdef MSDOS
    unsigned long tics = (unsigned long)dos_tics();
    return ((double) tics*0.054945);  /* 1/18.2 */
#else
#ifdef HAVE_GETTIMEOFDAY
    struct timeval tv;
    struct timezone tz;

    (void) gettimeofday(&tv, &tz);
    return
        (tv.tv_sec + (double) tv.tv_usec / 1000000.0);
#else
#ifdef HAVE_FTIME
    struct timeb timenow;
    int sec, msec;

    ftime(&timenow);
    timediff(&timenow, &timebegin, &sec, &msec);
    return (sec + (double) msec / 1000.0);
#else
#ifdef HAVE_VMSHACK
    return ((double)clock()/(double)CLK_TCK);
#else /* unknown */
    /* don't know how to do this in general. */
    return (-1.0); /* Obvious error condition */
#endif /* !VMS */
#endif /* !FTIME */
#endif /* !BSD */
#endif /* DOS */

}
