/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "sperror.h"
#include "niext.h"

#ifdef __STDC__
static int ni_gearcof(CKTcircuit*);
#else
static int ni_gearcof();
#endif


int
NIcomCof(ckt)

CKTcircuit *ckt;
{
/*  this routine calculates the timestep-dependent terms used in the
 *  numerical integration.
 */ 

    if (ckt->CKTintegrateMethod == TRAPEZOIDAL) {

        if (ckt->CKTorder == 2) {
            ckt->CKTag[0] = 2.0/ckt->CKTdelta;
            ckt->CKTag[1] = 1.0;
            /*
            ckt->CKTag[0] = 1.0/ckt->CKTdelta/(1.0-0.5);
            ckt->CKTag[1] = 0.5/(1.0 - 0.5);
            above lines should have 'xmu' instead of .5 eventually
            (in all three places)
            */
            return (OK);
        }
        if (ckt->CKTorder == 1) {
            ckt->CKTag[0] = 1/ckt->CKTdelta;
            ckt->CKTag[1] = -ckt->CKTag[0];
            return (OK);
        }
        return(E_ORDER);
    }
    if (ckt->CKTintegrateMethod == GEAR) {
        return (ni_gearcof(ckt));
    }
    return (E_METHOD);
}


static int
ni_gearcof(ckt)

CKTcircuit* ckt;
{
    double mat[8][8];   /* matrix to compute the gear coefficients in */
    int i,j,k;          /* generic loop indicies */
    double arg;
    double arg1;

    if (ckt->CKTorder < 1 || ckt->CKTorder > 6)
        return (E_ORDER);

    bzero(ckt->CKTag,7*sizeof(double));
    ckt->CKTag[1] = -1/ckt->CKTdelta;
    /* first, set up the matrix */
    arg = 0;
    for (i = 0; i <= ckt->CKTorder; i++) { mat[0][i]=1; }
    for (i = 1; i <= ckt->CKTorder; i++) { mat[i][0]=0; }
    /* SPICE2 difference warning
     * the following block builds the corrector matrix
     * using (sum of h's)/h(final) instead of just (sum of h's)
     * because the h's are typically ~1e-10, so h^7 is an
     * underflow on many machines, but the ratio is ~1
     * and produces much better results
     */
    for (i = 1; i <= ckt->CKTorder; i++) {
        arg += ckt->CKTdeltaOld[i-1];
        arg1 = 1;
        for (j = 1; j <= ckt->CKTorder; j++) {
            arg1 *= arg/ckt->CKTdelta;
            mat[j][i] = arg1;
        }
    }
    /* lu decompose */
    /* weirdness warning! 
     * The following loop (and the first one after the forward
     * substitution comment) start at one instead of zero
     * because of a SPECIAL CASE - the first column is 1 0 0 ...
     * thus, the first iteration of both loops does nothing,
     * so it is skipped
     */
    for (i = 1; i <= ckt->CKTorder; i++) {
        for (j = i+1; j <= ckt->CKTorder; j++) {
            mat[j][i] /= mat[i][i];
            for (k = i+1; k <= ckt->CKTorder; k++) {
                mat[j][k] -= mat[j][i]*mat[i][k];
            }
        }
    }
    /* forward substitution */
    for (i = 1; i <= ckt->CKTorder; i++) {
        for (j = i+1; j <= ckt->CKTorder; j++) {
            ckt->CKTag[j] = ckt->CKTag[j]-mat[j][i]*ckt->CKTag[i];
        }
    }
    /* backward substitution */
    ckt->CKTag[ckt->CKTorder] /= mat[ckt->CKTorder][ckt->CKTorder];
    for (i = ckt->CKTorder-1; i >= 0; i--) {
        for (j = i+1; j <= ckt->CKTorder; j++) {
            ckt->CKTag[i] = ckt->CKTag[i] - mat[i][j]*ckt->CKTag[j];
        }
        ckt->CKTag[i] /= mat[i][i];
    }
    return (OK);
}
