/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

/*
 * NIconvTest(ckt)
 *  perform the convergence test - returns 1 if any of the 
 *  values in the old and new arrays have changed by more 
 *  than absTol + relTol*(max(old,new)), otherwise returns 0
 */

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "cktdefs.h"
#include "util.h"
#include "cktext.h"
#include "niext.h"

int
NIconvTest(ckt)

CKTcircuit *ckt;
{
    int i;         /* generic loop variable */
    int size;      /* size of the matrix */
    CKTnode *node; /* current matrix entry */
    double old, aold;
    double new, anew;
    double dd, tol;

    node = ckt->CKTnodes;
    size = spGetSize(ckt->CKTmatrix,1);
    for (i = 1; i <= size; i++) {

        node = node->next;

        new =  *((ckt->CKTrhs) + i);
        old =  *((ckt->CKTrhsOld) + i);

        if (new > 0)
            anew = new;
        else
            anew = -new;

        if (old > 0)
            aold = old;
        else
            aold = -old;

        dd = new - old;
        if (dd < 0)
            dd = -dd;

        if (node->type == 3) {
            tol =  ckt->CKTreltol*MAX(aold,anew) + ckt->CKTvoltTol;
            if (dd > tol) {
#ifdef STEPDEBUG
                printf(" non-convergence at node %s\n",CKTnodName(ckt,i));
#endif /* STEPDEBUG */
                ckt->CKTtroubleNode = i;
                return (1);
            }
        }
        else {
            tol =  ckt->CKTreltol*MAX(aold,anew) + ckt->CKTabstol;
            if (dd > tol) {
#ifdef STEPDEBUG
                printf(" non-convergence at node %s\n",CKTnodName(ckt,i));
#endif /* STEPDEBUG */
                return (1);
            }
        }
    }
    return (CKTconvTest(ckt));
}
