/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "util.h"
#include "sperror.h"
#include "niext.h"

#define CKALLOC(ptr,size,type) if(( ckt->ptr =\
    (type *) MALLOC((size)*sizeof(type))) == NULL) return(E_NOMEM);


int
NIinit(ckt)

/*
 *  Initialize the numerical iteration package.
 */
CKTcircuit  *ckt;
{
    int error;

    ckt->CKTniState = NIUNINITIALIZED;
    ckt->CKTmatrix = spCreate(0,1,&error);
    return (error);
}


void
NIprint(ckt)

/*
 *  Print the matrix in human readable form.
 */
CKTcircuit  *ckt;
{

    if (ckt->CKTmatrix &&
            spGetSize(ckt->CKTmatrix,1) > 0)
        spPrint(ckt->CKTmatrix,0,1,1);
    else
        printf("Matrix not found.\n");
}


int
NIreinit(ckt)

/*
 *  Perform reinitialization necessary for the numerical iteration
 *  package - the matrix has now been fully accessed once, so we know
 *  how big it is, so allocate RHS vector
 */
CKTcircuit *ckt;
{
    int size;
    int i;

    size = spGetSize(ckt->CKTmatrix,1);
    FREE(ckt->CKTrhs);
    FREE(ckt->CKTrhsOld);
    FREE(ckt->CKTrhsSpare);
    FREE(ckt->CKTirhs);
    FREE(ckt->CKTirhsOld);
    FREE(ckt->CKTirhsSpare);
    CKALLOC(CKTrhs,size+1,double);
    CKALLOC(CKTrhsOld,size+1,double);
    CKALLOC(CKTrhsSpare,size+1,double);
    CKALLOC(CKTirhs,size+1,double);
    CKALLOC(CKTirhsOld,size+1,double);
    CKALLOC(CKTirhsSpare,size+1,double);
    for (i = 0; i < 8; i++) {
        FREE(ckt->CKTsols[i]);
        CKALLOC(CKTsols[i],size+1,double);
    }
    ckt->CKTniState = NISHOULDREORDER | NIACSHOULDREORDER | NIPZSHOULDREORDER;
    return (0);
}


void
NIdestroy(ckt)

/*
 * delete the data structures allocated for numeric integration.
 */
CKTcircuit *ckt;
{
    spDestroy(ckt->CKTmatrix);
    ckt->CKTmatrix = NULL;
    FREE(ckt->CKTrhs);
    FREE(ckt->CKTrhsOld);
    FREE(ckt->CKTrhsSpare);
    FREE(ckt->CKTirhs);
    FREE(ckt->CKTirhsOld);
    FREE(ckt->CKTirhsSpare);
}
