/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1992 Stephen R. Whiteley
****************************************************************************/

    /* NIintegrate(ckt,geq,ceq,cap,qcap)
     *  integrate the specified capacitor - method and order in the
     *  ckt structure, ccap follows qcap.
     */

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "sperror.h"
#include "util.h"
#include "niext.h"

#define ccap qcap+1

int
NIintegrate(ckt,geq,ceq,cap,qcap)

CKTcircuit *ckt;
double *geq;
double *ceq;
double cap;
int qcap;
{
    int error;

    error = NIsum(ckt,ceq,qcap);
    if (error != OK)
        return (error);

    *(ckt->CKTstate0 + ccap) =
        ckt->CKTag[0] * *(ckt->CKTstate0 + qcap) + *ceq;

    *geq = ckt->CKTag[0] * cap;
    return(OK);
}


int
NIsum(ckt,ceq,qcap)

/* For linear inductors and capacitors, one can call this when
 * mode is MODEINITTRAN or MODEINITPRED, and save ceq, avoiding
 * the repeated and redundant calls to NIintegrate().  An accept
 * function can then finish the integration - see INDaccept().
 */

CKTcircuit *ckt;
double *ceq;
int qcap;
{
    static char *ordmsg = "Illegal integration order";
    static char *methodmsg = "Unknown integration method";
    double cq = 0;

    switch (ckt->CKTintegrateMethod) {

        case TRAPEZOIDAL:
            switch (ckt->CKTorder) {
                case 1:
                    cq = ckt->CKTag[1] * (*(ckt->CKTstate1+qcap));
                    break;
                case 2:
                    cq = - *(ckt->CKTstate1+ccap) * ckt->CKTag[1] - 
                            ckt->CKTag[0] * *(ckt->CKTstate1+qcap);
                    break;
                default:
                    errMsg = MALLOC(strlen(ordmsg)+1);
                    strcpy(errMsg,ordmsg);
                    return (E_ORDER);
            }
            break;

        case GEAR:
            switch (ckt->CKTorder) {

                case 6:
                    cq += ckt->CKTag[6]* *(ckt->CKTstate6+qcap);
                    /* fall through */
                case 5:
                    cq += ckt->CKTag[5]* *(ckt->CKTstate5+qcap);
                    /* fall through */
                case 4:
                    cq += ckt->CKTag[4]* *(ckt->CKTstate4+qcap);
                    /* fall through */
                case 3:
                    cq += ckt->CKTag[3]* *(ckt->CKTstate3+qcap);
                    /* fall through */
                case 2:
                    cq += ckt->CKTag[2]* *(ckt->CKTstate2+qcap);
                    /* fall through */
                case 1:
                    cq += ckt->CKTag[1]* *(ckt->CKTstate1+qcap);
                    break;

                default:
                    errMsg = MALLOC(strlen(ordmsg)+1);
                    strcpy(errMsg,ordmsg);
                    return (E_ORDER);

            }
            break;

        default:
            errMsg = MALLOC(strlen(methodmsg)+1);
            strcpy(errMsg,methodmsg);
            return (E_METHOD);
    }
    *ceq = cq;

    return (OK);
}
