/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Gary W. Ng
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "cktdefs.h"
#include "niext.h"

/*
 * NInzIter (ckt, posDrive, negDrive)
 *
 * This routine solves the adjoint system.  It assumes that the matrix has
 * already been loaded by a call to NIacIter, so it only alters the right
 * hand side vector.  The unit-valued current excitation is applied
 * between nodes posDrive and negDrive.
 */

void
NInzIter(ckt, posDrive, negDrive)

CKTcircuit *ckt;
int posDrive;
int negDrive;
{
    int i, size;

    /* clear out the right hand side vector */
    size = spGetSize(ckt->CKTmatrix,1);
    for (i = 0; i <= size; i++) {
        *((ckt->CKTrhs) + i) = 0.0;
        *((ckt->CKTirhs) + i) = 0.0;
    }

    *((ckt->CKTrhs) + posDrive) = 1.0;     /* apply unit current excitation */
    *((ckt->CKTrhs) + negDrive) = -1.0;
    spSolveTransposed(ckt->CKTmatrix,ckt->CKTrhs,
        ckt->CKTrhs,ckt->CKTirhs,ckt->CKTirhs);
    *ckt->CKTrhs = 0.0;
    *ckt->CKTirhs = 0.0;
}
