/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 UCB
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "plotdefs.h"

#ifdef __STDC__
static int gen_DatatoScreen(GRAPH*,double,double,int*,int*);
static int gen_Input(REQUEST*,RESPONSE*);
static int nodev();
static int nop();
#else
static int gen_DatatoScreen();
static int gen_Input();
static int nodev();
static int nop();
#endif


DISPDEVICE device[] = {

    {"error",
    false, false, 0, 0, 0, 0, 0, 0, nop, nop,
    nop, nop, nop, nop,
    nop, nop, nop, nop,
    nop, nop, nop,
    nop, nop, nop,
    nop, nop,
    nop, nop, nop, gen_Input,
    nop,},

#ifdef HAVE_X11
    {"X11",
    true, false, 0, 0, 1024, 864, 0, 0, X11_Init, X11_NewViewport,
    X11_Close, X11_Halt, X11_Clear, X11_Title,
    X11_Pixel, X11_Line, X11_Box, X11_Arc,
    X11_Polygon, X11_Text, X11_SetGhost,
    X11_DefineColor, X11_DefineLinestyle, X11_SetLinestyle,
    X11_SetColor, X11_Update,
    nodev, nodev, nodev, X11_Input,
    gen_DatatoScreen,},
#endif

#ifdef HAVE_MFB
    {"MFB",
    false, false, 0, 0, 1000, 1000, 0, 0, Mfb_Init, Mfb_NewViewport,
    Mfb_Close, Mfb_Halt, Mfb_Clear, nop,
    Mfb_Pixel, Mfb_Line, Mfb_Box, Mfb_Arc,
    Mfb_Polygon, Mfb_Text, Mfb_SetGhost,
    Mfb_DefineColor, Mfb_DefineLinestyle, Mfb_SetLinestyle,
    Mfb_SetColor, Mfb_Update,
    nodev, Mfb_MakeMenu, Mfb_MakeDialog, Mfb_Input,
    gen_DatatoScreen,},
#endif

    {"plot5",
    false, true, 0, 0, 1000, 1000, 0, 0, Plt5_Init, Plt5_NewViewport,
    Plt5_Close, Plt5_Halt, nop, nop,
    Plt5_Pixel, Plt5_Line, Plt5_Box, Plt5_Arc,
    Plt5_Polygon, Plt5_Text, nop,
    nop, nop, Plt5_SetLinestyle,
    Plt5_SetColor, Plt5_Update,
    nop, nop, nop, nop,
    gen_DatatoScreen,},

    {"postscript",
    false, true, 0, 0, 1000, 1000, 0, 0, PS_Init, PS_NewViewport,
    PS_Close, PS_Halt, nop, nop,
    PS_Pixel, PS_Line, PS_Box, PS_Arc,
    PS_Polygon, PS_Text, nop,
    nop, nop, PS_SetLinestyle,
    PS_SetColor, PS_Update,
    nop, nop, nop, nop,
    gen_DatatoScreen,},

    {"laser",
    false, true, 0, 0, 1000, 1000, 0, 0, HP_Init, HP_NewViewport,
    HP_Close, HP_Halt, nop, nop,
    HP_Pixel, HP_Line, HP_Box, HP_Arc,
    HP_Polygon, HP_Text, nop,
    nop, HP_DefineLinestyle, HP_SetLinestyle,
    nop, nop,
    nop, nop, nop, nop,
    gen_DatatoScreen,},

    {"wp",
    false, true, 0, 0, 1000, 1000, 0, 0, WP_Init, WP_NewViewport,
    WP_Close, WP_Halt, nop, nop,
    WP_Pixel, WP_Line, WP_Box, WP_Arc,
    WP_Polygon, WP_Text, nop,
    nop, nop, WP_SetLinestyle,
    nop, nop,
    nop, nop, nop, nop,
    gen_DatatoScreen,},

    {"printf",
    false, true, 0, 0, 24, 80, 0, 0, nodev, nodev,
    nodev, nodev, nodev, nodev,
    nodev, nodev, nodev, nodev,
    nodev, nodev, nodev,
    nodev, nodev, nodev,
    nodev, nodev,
    nodev, nodev, nodev, nodev,
    nodev,},

};

DISPDEVICE *dispdev;

#define XtNumber(arr)       (sizeof(arr) / sizeof(arr[0]))


DISPDEVICE *
FindDev(name)

char *name;
{
    int i;

    for (i=0; i < XtNumber(device); i++) {
        if (!strcmp(name, device[i].name)) {
            return (&device[i]);
        }
    }
    sprintf(ErrorMessage, "Can't find device %s.", name);
    internalerror(ErrorMessage);
    return (&device[0]);
}

int
DevXterm(string)

char *string;
{
    char buf[BSIZE_SP];

    if (dispdev->windows)
        (void) sprintf(buf, "xterm -e %s", string);
    else
        (void) sprintf(buf, "%s", string);

    return (system(buf));
}


void
DevGrHalt()

{
    int id;

    /* halt graphics if not windowed */
    if (!dispdev->windows &&
            currentgraph && currentgraph->graphtype != GR_SCED) {
        DevHalt();
        id = currentgraph->graphid;
        PopGraphContext();
        DestroyGraph(id);
    }
}


void
DevInit()

{
    char buf[128];
    extern char *kw_display;

/* note: do better determination */
/*
    dumb tradition that got passed on from gi_interface
    to do compile time determination
*/
#ifdef HAVE_MFB
#ifdef HAVE_X11
    /* determine display type */
    if (cp_display)
        dispdev = FindDev("X11");
    else
        dispdev = FindDev("MFB");
#else       /* MFB but not X */
    dispdev = FindDev("MFB");
#endif
#else
#ifdef HAVE_X11
    dispdev = FindDev("X11");
#else       /* no MFB nor X */
    externalerror(
    "no graphics packages; need to define either WANT_X11, or WANT_MFB");
    dispdev = FindDev("error");
#endif
#endif

    if ((*(dispdev->Init))()) {
        fprintf(cp_err,
        "Warning: can't initialize display device for graphics.\n");
        dispdev = FindDev("error");
    }
}


/* NewViewport is responsible for filling in graph->viewport */
int
DevNewViewport(pgraph)

GRAPH *pgraph;
{
    return (*(dispdev->NewViewport))(pgraph);
}


void
DevClose()

{
    (void)(*(dispdev->Close))();
}


void
DevHalt()

{
    (void)(*(dispdev->Halt))();
}


void
DevClear()

{
    (void)(*(dispdev->Clear))();
}


void
DevTitle(wintitle,icontitle)

char *wintitle, *icontitle;
{
    (void)(*(dispdev->Title))(wintitle,icontitle);
}


void
DevPixel(x, y)

int x, y;
{
    (void)(*(dispdev->Pixel))(x, y);
}


void
DevLine(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    (void)(*(dispdev->Line))(x1, y1, x2, y2);
}


void
DevBox(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    (void)(*(dispdev->Box))(x1, y1, x2, y2);
}


void
DevArc(x0, y0, radius, theta1, theta2)

int x0, y0, radius;
double theta1, theta2;
{
    (void)(*(dispdev->Arc))(x0, y0, radius, theta1, theta2);
}


void
DevPolygon(p)

POLYGON *p;
{
    (void)(*(dispdev->Polygon))(p);
}


void
DevText(text, x, y)

char *text;
int x, y;
{
    (void)(*(dispdev->Text))(text, x, y);
}


void
DevSetGhost(callback,x,y)

#ifdef __STDC__
void (*callback)(int,int,int,int);
#else
void (*callback)();
#endif
int x,y;
{
    (void)(*(dispdev->SetGhost))(callback, x, y);
}


void
DevDefineColor(colorid, red, green, blue)

int colorid;
double red, green, blue;
{
    (void)(*(dispdev->DefineColor))(colorid, red, green, blue);
}


void
DevDefineLinestyle(linestyleid, mask)

int linestyleid;
int mask;
{
    (void)(*(dispdev->DefineLinestyle))(linestyleid, mask);
}


void
DevSetLinestyle(linestyleid)

int linestyleid;
{
    (void)(*(dispdev->SetLinestyle))(linestyleid);
}


void
DevSetColor(colorid)

int colorid;
{
    (void)(*(dispdev->SetColor))(colorid);
}


void
DevUpdate()
{
    (void)(*(dispdev->Update))();
}


/* note: screen coordinates are relative to window
 * so need to add viewport offsets
 */
static int
gen_DatatoScreen(graph, x, y, screenx, screeny)

GRAPH *graph;
double x, y;
int *screenx, *screeny;
{
    double low, high;

    /* note: think this out---Is 1 part of the viewport? Do we handle
        this correctly? */

    /* have to handle several types of grids */

    /* note: we can't compensate for X's demented y-coordinate system here
        since the grid routines use DevLine w/o calling this routine */
    if ((graph->grid.gridtype == GRID_LOGLOG) ||
            (graph->grid.gridtype == GRID_YLOG)) {
        low = mylog10(graph->datawindow.ymin);
        high = mylog10(graph->datawindow.ymax);
        *screeny = rnd( (((mylog10(y) - low) / (high - low))
            * graph->viewport.height)
            + graph->viewportyoff );
    }
    else {
        *screeny = rnd( ((y - graph->datawindow.ymin) / graph->aspectratioy)
            + graph->viewportyoff );
    }

    if ((graph->grid.gridtype == GRID_LOGLOG) ||
            (graph->grid.gridtype == GRID_XLOG)) {
        low = mylog10(graph->datawindow.xmin);
        high = mylog10(graph->datawindow.xmax);
        *screenx = rnd( ((mylog10(x) - low) / (high - low))
            * graph->viewport.width
            + graph ->viewportxoff);
    }
    else {
        *screenx = rnd( (x - graph->datawindow.xmin) / graph->aspectratiox
            + graph ->viewportxoff);
    }
    return (0);
}


void
DevDatatoScreen(graph, x, y, screenx, screeny)

GRAPH *graph;
double x, y;
int *screenx, *screeny;
{
    (void)(*(dispdev->DatatoScreen))(graph, x, y, screenx, screeny);
}

#ifdef notdef
/*
void
NDCtoScreen(x0, y0, px, py)

double x0, y0;
int *px, *py;
{
    (void)(*(dispdev->NDCtoScreen))(x0, y0, px, py);
}
*/
#endif


void
DevInput(request, response)

REQUEST *request;
RESPONSE *response;
{
    (void)(*(dispdev->Input))(request, response);
}


static int
gen_Input(request, response)

REQUEST *request;
RESPONSE *response;
{
    switch (request->option) {
        case char_option:
            response->reply.ch = cp_inchar(request->fp);
            response->option = request->option;
            break;
        default:
            /* just ignore, since we don't want a million error messages */
            if (response)
                response->option = error_option;
            break;
    }
    return (0);
}


/* no operation, do nothing */
static int nop()
{
    return (1);  /* so NewViewport will fail */
}


static int
nodev()
{
    sprintf(ErrorMessage,
        "This operation is not defined for display type %s.",
        dispdev->name);
    internalerror(ErrorMessage);
    return (1);
}


void
SaveText(graph, text, x, y)

GRAPH *graph;
char *text;
int x, y;
{
    struct _keyed *keyed;

    keyed = (struct _keyed *) calloc(1, sizeof(struct _keyed));

    if (!graph->keyed) {
       graph->keyed = keyed;
    }
    else {
       keyed->next = graph->keyed;
       graph->keyed = keyed;
    }

    keyed->text = tmalloc(strlen(text) + 1);
    strcpy(keyed->text, text);

    keyed->x = x;
    keyed->y = y;

    keyed->colorindex = 1;
}


int
DevSwitch(devname)

/* if given name of a hardcopy device, finds it and switches devices
 * if given NULL, switches back
 */
char *devname;
{
    static DISPDEVICE *lastdev = NULL;

    if (devname != NULL) {
        if (lastdev != NULL) {
            internalerror("DevSwitch w/o changing back");
            return (1);
        }
        lastdev = dispdev;
        dispdev = FindDev(devname);
        if (!strcmp(dispdev->name, "error")) {
            internalerror("no hardcopy device");
            dispdev = lastdev;  /* undo */
            lastdev = NULL;
            return (1);
        }
        (*(dispdev->Init))();
    }
    else {
        (*(dispdev->Close))();
        dispdev = lastdev;
        lastdev = NULL;
    }
    return (0);
}


int
DevGetchar(fp)

FILE *fp;
{
    REQUEST request;
    RESPONSE response;

    request.option = char_option;
    request.fp = fp;
    DevInput(&request, &response);
    return(response.reply.ch);
}
