/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Author:  1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Hewlett-Packard Laserjet (and compatible) driver.
 */

#include "spice.h"
#include "plotdev.h"

static FILE *plotfile;

struct hptext {
    int x;
    int y;
    char *text;
    struct hptext *next;
};

struct hplaser {
    int bytpline;
    int maxx;
    int maxy;
    char *base;
    struct hptext *textlist;
    unsigned char linestyle;
    unsigned char linestyle_stored;
};

#define ror(x,n) ((x >> n)  | (x << (8-n)))
#define swap(a,b) {int t=a; a=b; b=t;}

/* upper left of image in pcl coordinates */
#define OFFSETX 0
#define OFFSETY 90

/* dots per inch (75, 100, 150, 300) */
#define RESOL  150

int
HP_Init()

{
    dispdev->numlinestyles = 1;
    dispdev->numcolors = 2;

    /* 8" X 10.5" drawable area */
    dispdev->width = 8*RESOL;
    dispdev->height = 10*RESOL + RESOL/2;
    return (0);
}


int
HP_NewViewport(graph)

/* devdep initially contains name of output file */
GRAPH *graph;
{
    struct hplaser *hp;
    int bpline;

    plotfile = fopen((char*)graph->devdep, "wb");
    if (!plotfile) {
        perror(graph->devdep);
        graph->devdep = (char *) NULL;
        return (1);
    }
    graph->devdep = tmalloc(sizeof(struct hplaser));
    hp = (struct hplaser*)graph->devdep;

    switch (graph->graphtype) {
        default:
        case GR_PLOT:
        case GR_GRAF:
        case GR_MPLT:
            /* square plotting area */
            graph->absolute.width = dispdev->width;
            graph->absolute.height = dispdev->width;
            break;
        case GR_SCED:
            /* full page plot */
            graph->absolute.width = dispdev->width;
            graph->absolute.height = dispdev->height;
            break;
    }
    /* bytes per line */
    bpline = (graph->absolute.width + 7)/8;

    hp->base = malloc(graph->absolute.height*bpline);
    if (hp->base == NULL) {
        internalerror("Not enough memory for raster conversion");
        return (1);
    }
    memset(hp->base,0,graph->absolute.height*bpline);

    graph->fontwidth = 15;
    graph->fontheight = 30;

    hp->bytpline = bpline;
    hp->maxx = graph->absolute.width - 1;
    hp->maxy = graph->absolute.height - 1;
    hp->linestyle = 0xff;
    hp->linestyle_stored = 0xcc;
    return (0);
}


int
HP_Close()

{
    return (0);
}


int
HP_Halt()

{
    struct hplaser *hp = (struct hplaser *)currentgraph->devdep;
    struct hptext *old;
    char *buf, *rgen, *c;
    int len, i;

    /* dump the stuff */

    /* reset printer
     * top margin 10
     * X cursor position OFFSETX
     * Y cursor position OFFSETY
     * resolution RESOL dpi
     * start raster graphics at current position
     */
    fprintf(plotfile,
        "\033E\033&l0E\033*p%dX\033*p%dY\033*t%dR\033*r1A",
        OFFSETX,OFFSETY,RESOL);

    buf = tmalloc(hp->bytpline+8);
    sprintf(buf,"\033*b%dW",hp->bytpline);
    len = strlen(buf);
    c = buf + len;

    rgen = hp->base;
    for (i = 0; i <= hp->maxy; i++) {
        memcpy(c,rgen,hp->bytpline);
        rgen += hp->bytpline;

        if (fwrite(buf,1,hp->bytpline+len,plotfile) != hp->bytpline+len) {
            break;
        }
        
    }
    /* end raster graphics */
    fprintf(plotfile,"\033*rB");

    /* now for the text */

    while (hp->textlist) {
        /* x position, y position, text */
        fprintf(plotfile,"\033*p%dX\033*p%dY%s",
            hp->textlist->x,hp->textlist->y,hp->textlist->text);
        txfree(hp->textlist->text);
        old = hp->textlist;
        hp->textlist = hp->textlist->next;
        txfree((char*)old);
    }

    /* form feed
     * reset printer
     */
    fprintf(plotfile,"\033*rB\014\033E");

    txfree(buf);
    txfree(hp->base);
    fclose(plotfile);
    return (0);
}


int
HP_Pixel(x, y)

int x, y;
{
    unsigned char c;
    struct hplaser *hp = (struct hplaser *)currentgraph->devdep;

    c = 0x80 >> (x & 7);
        *(hp->base + ((x >> 3) + (hp->maxy-y)*hp->bytpline)) |= c;
    return (0);
}


int
HP_Line(x1, y1, x2, y2)

int x1, y1, x2, y2;
{

    struct hplaser *hp = (struct hplaser *)currentgraph->devdep;
    int dx, dy, dy2, errterm = 0, next, lcnt;
    char *rgen;
    unsigned char cbuf, left, right;

    if (x2 < x1) {
        swap(x1,x2);
        swap(y1,y2);
    }
    dx = x2 - x1;

    next = hp->bytpline;
    dy = y1 - y2;

    lcnt = hp->maxy - y1;
    rgen = hp->base + (x1 >> 3) + lcnt*next;
    left = 0x80 >> ((lcnt+x1) & 7);
    right = hp->linestyle;

    if (y1 < y2) {
        next = -next;
        dy = -dy;
    }
    dy2 = dy;

    cbuf = 0x80 >> (x1 & 7);

    for (dy++; dy; dy--) {
        errterm += dx;
        if (errterm <= 0) {
            if (left & right) *rgen |= cbuf;
            rgen += next;
            left = ror(left,1);
            continue;
        }
        while (errterm > 0 && x1 != x2) {
            if (left & right) *rgen |= cbuf;
            left = ror(left,1);
            cbuf = ror(cbuf,1);
            if (cbuf & 0x80) rgen++;
            x1++;
            errterm -= dy2;
        }
        rgen += next;
        left = ror(left,1);
    }
    return (0);
}


int
HP_Box(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    HP_Line(x1,y1,x1,y2);
    HP_Line(x1,y2,x2,y2);
    HP_Line(x2,y2,x2,y1);
    HP_Line(x2,y1,x1,y1);
    return (0);
}


/* ARGSUSED */
int
HP_Arc(x0, y0, radius, theta1, theta2)

int x0, y0, radius;
double theta1, theta2;
{
    return (0);
}


int
HP_Polygon(p)

POLYGON *p;
{
    int i, n = 2*p->nvertices - 2;

    for (i = 0; i < n; i += 2)
        HP_Line(p->xy[i],p->xy[i+1],p->xy[i+2],p->xy[i+3]);

    return (0);
}


int
HP_Text(text, x, y)

char *text;
int x, y;
{
    struct hplaser *hp = (struct hplaser *)currentgraph->devdep;
    struct hptext *hpt;

    hpt = (struct hptext *)tmalloc(sizeof(struct hptext));
    hpt->x = x*(300/RESOL) + OFFSETX;
    hpt->y = (hp->maxy - y)*(300/RESOL) + OFFSETY;
    hpt->text = copy(text);
    hpt->next = hp->textlist;
    hp->textlist = hpt;
    return (0);
}


int
HP_DefineLinestyle(linestyleid,mask)

int linestyleid, mask;
{
    struct hplaser *hp = (struct hplaser *)currentgraph->devdep;

    if (linestyleid) {
        hp->linestyle_stored = mask;
        hp->linestyle = mask;
    }
}


int
HP_SetLinestyle(linestyleid)

int linestyleid;
{
    struct hplaser *hp = (struct hplaser *)currentgraph->devdep;
    
    if (linestyleid)
        hp->linestyle = hp->linestyle_stored;
    else
        hp->linestyle = 0xff;
    return (0);
}
