/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1988 Jeffrey M. Hsu
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "ftedefs.h"
#include "ftedebug.h"
#include "plotdefs.h"

extern struct dbcomm *dbiplot;
extern GRAPH *EchoGraph;  /* used/defined in newgraf.c */

#define XFACTOR 2       /* How much to expand the X scale during iplot. */
#define YFACTOR 1.5     /* How much to expand the Y scale during iplot. */

#ifdef __STDC__
static void iplot(struct dbcomm*,struct plot*);
static struct dvlist *get_dvecs(struct dbcomm*);
#else
static void iplot();
static struct dvlist *get_dvecs();
#endif


void
gr_iplot(plot)

/* Do some incremental plotting. 3 cases -- first, if length < IPOINTMIN,
 * don't do anything. Second, if length = IPOINTMIN, plot what we have
 * so far. Third, if length > IPOINTMIN, plot the last points and resize
 * if needed.
 * Note we don't check for pole / zero because they are of length 1.
 */
struct plot *plot;
{

    struct dbcomm *db;
    GRAPH *g;
    int gid;

    for (db = dbiplot; db; db = db->db_next) {
        if (db->db_type == DB_IPLOT) {

            gid = db->db_graphid;
            if (gid) {
                g = FindGraph(db->db_graphid);
                if (g) {
                    if (!EchoGraph)
                        EchoGraph = g;
                    PushGraphContext(g);
                }
            }
            iplot(db, plot);
            if (gid && g) PopGraphContext();
            if (!dispdev->windows)
                return;
        }
    }
}


static void
iplot(db, pl)

struct dbcomm *db;
struct plot *pl;
{
    struct dvlist *dl0, *dl;
    struct dvec *xs = pl->pl_scale;
    int len = xs->v_length;
    struct dvec *v;
    double *lims, dy;
    double start, stop;
    int j;
    bool changed = false;
    int yt;
    double xlims[2], ylims[2];
    GRAPH *graph;
    struct grinit gr;

    /* db is NULL if the graphics does not support more than one
     * window.
     */

    if (ft_grdb)
        fprintf(cp_err, "Entering iplot, len = %d\n\r", len);

    if (len < IPOINTMIN) {
        /* Nothing yet, initialize flag for errors */
        return;
    }

    if (len == IPOINTMIN || !db->db_graphid) {
        /* Draw the grid for the first time, and plot everything. */

        /* Error handling:  If an error occurs during plot inialization,
         * db_iteration is set to -1 (db_graphid remains 0), so this
         * routine just exits on future calls.  If an error occurs later,
         * db_iteration is set to -1, and the plot is "frozen".
         */

        if (len == IPOINTMIN)
            db->db_iteration = 0;

        if (db->db_iteration == -1)  /* error flag, set below */
            return;

        dl0 = get_dvecs(db);
        if (dl0 == NULL) {
            db->db_iteration = -1;
            return;
        }

        if (eq(db->db_nodename2,"graf")) {
            db->db_graphid = iplot_begin(dl0,pl);
            if (!db->db_graphid)
                db->db_iteration = -1;
            return;
        }

        lims = ft_minmax(xs, true);
        xlims[0] = lims[0];
        xlims[1] = lims[1];
        ylims[0] = 1;
        ylims[1] = -1;
        for (j = 0, dl = dl0; dl; j++, dl = dl->dl_next) {
            v = dl->dl_dvec;
            lims = ft_minmax(v, true);
            if (ylims[0] > ylims[1]) {
                ylims[0] = lims[0];
                ylims[1] = lims[1];
            }
            else {
                if (lims[0] < ylims[0]) ylims[0] = lims[0];
                if (lims[1] > ylims[1]) ylims[1] = lims[1];
            }
            if (ylims[0] == ylims[1]) {
                ylims[0] -= 1e-30;
                ylims[1] += 1e-30;
            }
        }
        if (ft_grdb)
            fprintf(cp_err, 
                "iplot: after %d, xlims = %G, %G, ylims = %G, %G\n\r",
                IPOINTMIN,
                xlims[0],
                xlims[1],
                ylims[0],
                ylims[1]);

        for (yt = dl0->dl_dvec->v_type, dl = dl0; dl; dl = dl->dl_next)
            if (dl->dl_dvec->v_type != yt) {
                yt = 0;
                break;
            }

        gr.xlims = xlims;
        gr.ylims = ylims;
        gr.xname = xs->v_name;
        gr.plotname = pl->pl_title;
        gr.hcopy = NULL;
        gr.nplots = j;
        gr.xdelta = 0.0;
        gr.ydelta = 0.0;
        gr.gridtype = GRID_LIN;
        gr.plottype = PLOT_LIN;
        gr.xlabel = xs->v_name;
        gr.ylabel = "";
        gr.xtype = xs->v_type;
        gr.ytype = yt;
        gr.pname = "";
        gr.commandline = "iplot";

        graph = gr_init(dl0,(char*)&gr);
        if (graph)
            db->db_graphid = graph->graphid;
        if (!db->db_graphid)
            db->db_iteration = -1;
        return;
    }

/* bad foobar here for ^C pause in DOS, have to restart graphics */

    /* error occurred */
    if (db->db_iteration == -1)
        return;

    if (eq(db->db_nodename2,"graf")) {
        if (iplot_point()) {
            /* error */
            db->db_iteration = -1;
        }
        return;
    }

    /* First see if we have to make the screen bigger */
    dy = (isreal(xs) ? xs->v_realdata[len - 1] :
            realpart(&xs->v_compdata[len - 1]));
    if (ft_grdb)
        fprintf(cp_err, "x = %G\n\r", dy);

    /* hack to keep from overshooting range */
    stop = pl->pl_stop;
    start = pl->pl_start;
    if (stop < start) {
        stop = pl->pl_start;
        start = pl->pl_stop;
    }

    while (dy < currentgraph->datawindow.xmin) {
        changed = true;
        if (ft_grdb)
            fprintf(cp_err, "resize: xlo %G -> %G\n\r",
                currentgraph->datawindow.xmin,
                currentgraph->datawindow.xmin -
                    (currentgraph->datawindow.xmax -
                    currentgraph->datawindow.xmin)* XFACTOR);

        currentgraph->datawindow.xmin -=
            (currentgraph->datawindow.xmax -
            currentgraph->datawindow.xmin) * XFACTOR;

        if (start != stop && currentgraph->datawindow.xmin < start) {
            currentgraph->datawindow.xmin = start;
            break;
        }
    }
    if (currentgraph->datawindow.xmax < currentgraph->datawindow.xmin)
        currentgraph->datawindow.xmax = currentgraph->datawindow.xmin;

    while (dy > currentgraph->datawindow.xmax) {
        changed = true;
        if (ft_grdb)
            fprintf(cp_err, "resize: xhi %G -> %G\n\r",
                currentgraph->datawindow.xmax,
                currentgraph->datawindow.xmax +
                    (currentgraph->datawindow.xmax - 
                    currentgraph->datawindow.xmin) * XFACTOR);

        currentgraph->datawindow.xmax +=
                (currentgraph->datawindow.xmax -
                currentgraph->datawindow.xmin) * XFACTOR;

        if (start != stop && currentgraph->datawindow.xmax > stop) {
            currentgraph->datawindow.xmax = stop;
            break;
        }
    }

    dl0 = (struct dvlist*)currentgraph->plotdata;
    for (dl = dl0; dl; dl = dl->dl_next) {
        v = dl->dl_dvec;
        dy = (isreal(v) ? v->v_realdata[len - 1] :
                realpart(&v->v_compdata[len - 1]));
        if (ft_grdb)
            fprintf(cp_err, "y = %G\n\r", dy);

        while (dy < currentgraph->datawindow.ymin) {
            changed = true;
            if (ft_grdb)
                fprintf(cp_err, "resize: ylo %G -> %G\n\r",
                    currentgraph->datawindow.ymin,
                    currentgraph->datawindow.ymin -
                        (currentgraph->datawindow.ymax - 
                        currentgraph->datawindow.ymin) * YFACTOR);
            currentgraph->datawindow.ymin -=
              (currentgraph->datawindow.ymax -
              currentgraph->datawindow.ymin) * YFACTOR;
        }

        if (currentgraph->datawindow.ymax <
            currentgraph->datawindow.ymin)
            currentgraph->datawindow.ymax =
                currentgraph->datawindow.ymin;

        while (dy > currentgraph->datawindow.ymax) {
            changed = true;
            if (ft_grdb)
                fprintf(cp_err, "resize: yhi %G -> %G\n\r",
                    currentgraph->datawindow.ymax,
                    currentgraph->datawindow.ymax +
                    (currentgraph->datawindow.ymax -
                    currentgraph->datawindow.ymin) * YFACTOR);

            currentgraph->datawindow.ymax +=
              (currentgraph->datawindow.ymax -
              currentgraph->datawindow.ymin) * YFACTOR;
        }
    }
    if (changed) {
        /* Redraw everything. */
        (*currentgraph->redraw)(currentgraph);
    }
    else {
        /* Just connect the last two points. This won't
         * be done with curve interpolation, so it might
         * look funny.
         */
        for (dl = dl0; dl; dl = dl->dl_next) {
            v = dl->dl_dvec;
            gr_point(v, 
                (isreal(xs) ? xs->v_realdata[len - 1] :
                realpart(&xs->v_compdata[len - 1])),
                (isreal(v) ? v->v_realdata[len - 1] :
                realpart(&v->v_compdata[len - 1])),
                (isreal(xs) ? xs->v_realdata[len - 2] :
                realpart(&xs->v_compdata[len - 2])),
                (isreal(v) ? v->v_realdata[len - 2] :
                realpart(&v->v_compdata[len - 2])),
                len - 1);
        }
    }
    DevUpdate();
    return;
}


static struct dvlist *
get_dvecs(db)

/* make up a list of dvec's to be iplotted */
struct dbcomm *db;
{
    struct dvec *v;
    struct dvlist *dl, *dl0 = NULL;
    struct dbcomm *dc;
    wordlist *wl0, *wl, wt;
    char *tmp;
    struct pnode *pn;

    wt.wl_prev = wt.wl_next = NULL;
again:
    for (dc = db; dc; dc = dc->db_also) {
        if (eq(dc->db_nodename1,"all")) {
            wl0 = (wordlist*)htab_list(plot_cur->pl_hashtab);
            for (wl = wl0; wl; wl = wl->wl_next) {
                if (dl0 == NULL) {
                    dl = dl0 = alloc(struct dvlist);
                }
                else { 
                    dl->dl_next = alloc(struct dvlist);
                    dl = dl->dl_next;
                }
                dl->dl_dvec = (struct dvec*)((char**)wl->wl_word)[1];
            }
            wl_free(wl0);
            continue;
        }
        v = vec_fromplot(dc->db_nodename1, plot_cur);
/* hook for expressions
        if (!v) {
            wt.wl_word = dc->db_nodename1;
            pn = ft_getpnames(&wt, false);
            if (!pn)
                continue;
            v = ft_evaluate(pn);
            inp_pnfree(pn);
        }
*/
        if (!v)
            continue;

        if (dl0 == NULL) {
            dl = dl0 = alloc(struct dvlist);
        }
        else {
            dl->dl_next = alloc(struct dvlist);
            dl = dl->dl_next;
        }
        dl->dl_dvec = v;
    }

    if (!dispdev->windows) {
        /* pick up any more dvecs */
        for (db = db->db_next; db; db = db->db_next) {
            if (db->db_type != DB_IPLOT)
                continue;
            goto again;
        }
    }

    return (dl0);
}


/*
 *  This gets called after iplotting is done.  We clear out the db_graphid
 *  fields.  Copy the dvecs, which we referenced by reference, so
 *  DestroyGraph gets to free its own copy.  Note:  This is a clear
 *  case for separating the linestyle and color fields from dvec.
 */

void
gr_end_iplot()
{

    struct dbcomm *db, *prev, *next;
    GRAPH *graph;
    struct dvec *dv;
    struct dvlist *ddv, *ddn;

    EchoGraph = NULL;

    /* redraw for the last time */
    for (db = dbiplot; db; db = db->db_next) {
        if (db->db_type == DB_IPLOT) {

            if (db->db_graphid) {
                graph = FindGraph(db->db_graphid);
                PushGraphContext(graph);
                if (eq(db->db_nodename2,"graf"))
                    iplot_end();
                else {
                    (*graph->redraw)(graph);
                    gr_end();
                }
                if (!dispdev->windows) {
                    db->db_graphid = 0;
                    if (graph->graphtype == GR_PLOT) {
                        /* the dveclist has not been copied */
                        for (ddv = (struct dvlist *)graph->plotdata; ddv;
                                ddv = ddn) {
                            ddn = ddv->dl_next;
                            txfree((char*)ddv);
                        }
                        graph->plotdata = (GRDATA)NULL;
                    }
                    DevHalt();
                    PopGraphContext();
                    DestroyGraph(graph->graphid);
                    return;
                }
                PopGraphContext();
            }
        }
    }

    /* Tricky stuff - if an iplot window is closed before the simulation
     * is complete, it is assumed that the debug is no longer wanted.
     * DestroyGraph simply sets the debug type to DB_DEADIPLOT (so that
     * no more plotting is done).  Below we finish killing the associated
     * graph, and unlink the dead iplot from the debug list.
     * After the simulation finishes, we can kill the window without
     * affecting the debug list, as the db_graphid is set to zero (below).
     */

    prev = NULL;
    for (db = dbiplot; db; prev = db, db = next) {
        next = db->db_next;
        if (db->db_type == DB_DEADIPLOT) {
            if (db->db_graphid) {
                /* should always be true */

                graph = FindGraph(db->db_graphid);
                if (graph->graphtype == GR_PLOT) {
                    /* the dveclist has not been copied */
                    for (ddv = (struct dvlist *)graph->plotdata; ddv;
                            ddv = ddn) {
                        ddn = ddv->dl_next;
                        txfree((char*)ddv);
                    }
                    graph->plotdata = (GRDATA)NULL;
                }
                DestroyGraph(db->db_graphid);
                if (prev)
                    prev->db_next = next;
                else
                    dbiplot = next;
                dbfree(db);
            }
        }
        else if (db->db_type == DB_IPLOT) {

            if (db->db_graphid) {

                graph = FindGraph(db->db_graphid);
                if (graph->graphtype == GR_PLOT) {
                    /* get private copy of dvecs, free list */
                    ddv = (struct dvlist*)graph->plotdata;
                    graph->plotdata = (*graph->copydata)(graph->plotdata);
                    for ( ; ddv; ddv = ddn) {
                        ddn = ddv->dl_next;
                        txfree((char*)ddv);
                    }
                }
                db->db_graphid = 0;
            }
            else {
                /* warn that this wasn't plotted */
                fprintf(cp_err, "Warning: iplot %d was not executed.\n",
                db->db_number);
            }
        }
    }
    return;
}

