/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 UCB
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "plotdev.h"

static FILE *plotfile;

#define putsi(a)    putc((char) (a), plotfile); \
            putc((char) ((a) >> 8), plotfile)

#define SOLID 0
static char *linestyle[] = { "solid", "dotted", "longdashed", "shortdashed",
        "dotdashed" };
static int currentlinestyle = SOLID;


int
Plt5_Init()
{
    dispdev->numlinestyles = 4;
    dispdev->numcolors = 2;

    /* arbitrary drawable area */
    dispdev->width = 800;
    dispdev->height = 1050;

    return(0);
}


int
Plt5_NewViewport(graph)

GRAPH *graph;
{
    plotfile = fopen((char*)graph->devdep, "wb");
    if (!plotfile) {
        graph->devdep = (char *) NULL;
        perror(graph->devdep);
        return (1);
    }

    switch (graph->graphtype) {
        default:
        case GR_PLOT:
        case GR_GRAF:
        case GR_MPLT:
            /* square plotting area */
            graph->absolute.width = dispdev->width;
            graph->absolute.height = dispdev->width;
            break;
        case GR_SCED:
            /* full page plot */
            graph->absolute.width = dispdev->width;
            graph->absolute.height = dispdev->height;
            break;
    }

    putc('s', plotfile);
    putsi(0);
    putsi(0);
    putsi(graph->absolute.width);
    putsi(graph->absolute.height);

    /* reasonable values, used in gr_ for placement */
    graph->fontwidth = 12;
    graph->fontheight = 24;

    /* set to NULL so graphdb doesn't incorrectly de-allocate it */
    graph->devdep = (char *) NULL;

    return(0);
}


int
Plt5_Close()

{
    return (0);
}


int
Plt5_Halt()
{
    if (plotfile)
        fclose(plotfile);
    return (0);
}


/* ARGSUSED */
int
Plt5_Pixel(x, y)

int x, y;
{
    putc('p', plotfile);
    putsi(x);
    putsi(y);
    return (0);
}


int
Plt5_Line(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    putc('l', plotfile);
    putsi(x1);
    putsi(y1);
    putsi(x2);
    putsi(y2);
    return (0);
}


/* ARGSUSED */
int
Plt5_Box(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    Plt5_Line(x1,y1,x1,y2);
    Plt5_Line(x1,y2,x2,y2);
    Plt5_Line(x2,y2,x2,y1);
    Plt5_Line(x2,y1,x1,y1);
    return (0);
}


/* ARGSUSED */
int
Plt5_Arc(x0, y0, radius, theta1, theta2)

int x0, y0, radius;
double theta1, theta2;
{
    int xs, ys, xf, yf;

    xs = radius*sin(theta1);
    ys = radius*cos(theta1);
    xf = radius*sin(theta2);
    yf = radius*cos(theta2);
    putc('a', plotfile);
    putsi(x0);
    putsi(y0);
    putsi(xs);
    putsi(ys);
    putsi(xf);
    putsi(yf);
    return (0);
}


/* ARGSUSED */
int
Plt5_Polygon(p)

POLYGON *p;
{
    return (0);
}


int
Plt5_Text(text, x, y)
char *text;
int x, y;
{
    int savedlstyle;

    /* set linestyle to solid
     * or may get funny color text on some plotters
     */
    savedlstyle = currentlinestyle;
    Plt5_SetLinestyle(SOLID);

    /* move to (x, y) */
    putc('m', plotfile);
    putsi(x);
    putsi(y);

    /* use the label option */
    fprintf(plotfile, "t%s\n", text);

    /* restore old linestyle */
    Plt5_SetLinestyle(savedlstyle);
    return (0);
}


int
Plt5_SetLinestyle(linestyleid)

int linestyleid;
{
    if (linestyleid < 0) {
        internalerror("bad linestyleid");
        return (1);
    }
    if (linestyleid > dispdev->numlinestyles)
        linestyleid = (linestyleid % dispdev->numlinestyles) + 1;

    putc('f', plotfile);
    fprintf(plotfile, "%s\n", linestyle[linestyleid]);
    currentlinestyle = linestyleid;
    return (0);
}


/* ARGSUSED */
int
Plt5_SetColor(colorid)

int colorid;
{
    return (0);
}


int
Plt5_Update()

{
    fflush(plotfile);
    return (0);
}

