/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1992 Stephen R. Whiteley
****************************************************************************/

/*
 * WordPerfect driver
 */

#include "spice.h"
#include "plotdefs.h"

static FILE *plotfile;

#define putshort(s,fp) putc(((unsigned char)s),fp); putc(((s)>>8),fp)

/* these scale, make them user adjustable some day */
#define WP_fontwidth 100
#define WP_fontheight 180


int
WP_Init()
{
    double w, h;
    extern char *kw_wpboxw, *kw_wpboxh;

    dispdev->numlinestyles = 1;
    dispdev->numcolors = 2;

    if (!cp_getvar(kw_wpboxw,VT_REAL,(char*)&w))
        w = 4.0;
    if (!cp_getvar(kw_wpboxh,VT_REAL,(char*)&h))
        h = 4.0;

    /* 1200 units per inch */
    dispdev->width = w*1200;
    dispdev->height = h*1200;

    return (0);
}


int
WP_NewViewport(graph)

/* devdep initially contains name of output file */
GRAPH *graph;
{
    plotfile = fopen((char*)graph->devdep, "wb");
    if (!plotfile) {
        perror(graph->devdep);
        graph->devdep = (char *) NULL;
        return (1);
    }

    graph->absolute.width = dispdev->width;
    graph->absolute.height = dispdev->height;

    /* Header */

    putc(0xff,plotfile);
    putc('W',plotfile);
    putc('P',plotfile);
    putc('C',plotfile);

    putc(0x10,plotfile);
    putc(0,plotfile);
    putc(0,plotfile);
    putc(0,plotfile);

    putc(1,plotfile);
    putc(0x16,plotfile);
    putc(1,plotfile);
    putc(0,plotfile);

    putc(0,plotfile);
    putc(0,plotfile);
    putc(0,plotfile);
    putc(0,plotfile);

    /* Start WPG record */

    putc(15,plotfile);
    putc(6,plotfile);
    putc(1,plotfile);
    putc(0,plotfile);
    putshort(dispdev->width,plotfile);
    putshort(dispdev->height,plotfile);

    /* Text attributes record */

    putc(13,plotfile);
    putc(22,plotfile);
    putshort(WP_fontwidth,plotfile);
    putshort(WP_fontheight,plotfile);
    putshort(0,plotfile);
    putshort(0,plotfile);
    putshort(0,plotfile);
    putshort(0,plotfile);
    putshort(0,plotfile);
    putshort(0x1950,plotfile);
    putc(0,plotfile);
    putc(0,plotfile);
    putc(0,plotfile);
    putc(0,plotfile);
    putshort(0,plotfile);

    /* Line attributes record */

    putc(2,plotfile);
    putc(4,plotfile);
    putc(1,plotfile);
    putc(0,plotfile);
    putshort(8,plotfile);

    /* reasonable values, used in gr_ for placement */
    graph->fontwidth = WP_fontwidth;
    graph->fontheight = WP_fontheight;

    return (0);
}


int
WP_Close()

{
    return (0);
}


int
WP_Halt()
{
    putc(16,plotfile);
    putc(0,plotfile);
    fclose(plotfile);
    return (0);
}


/* ARGSUSED */
int
WP_Pixel(x, y)

int x, y;
{

}


int
WP_Line(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    /* Line record */

    putc(5,plotfile);
    putc(8,plotfile);
    putshort(x1,plotfile);
    putshort(y1,plotfile);
    putshort(x2,plotfile);
    putshort(y2,plotfile);

    return (0);
}


int
WP_Box(x1, y1, x2, y2)
int x1, y1, x2, y2;
{
    /* poly line record */

    putc(6,plotfile);
    putc(22,plotfile);
    putshort(5,plotfile);
    putshort(x1,plotfile);
    putshort(y1,plotfile);
    putshort(x1,plotfile);
    putshort(y2,plotfile);
    putshort(x2,plotfile);
    putshort(y2,plotfile);
    putshort(x2,plotfile);
    putshort(y1,plotfile);
    putshort(x1,plotfile);
    putshort(y1,plotfile);

    /*
    WP_Line(x1,y1,x1,y2);
    WP_Line(x1,y2,x2,y2);
    WP_Line(x2,y2,x2,y1);
    WP_Line(x2,y1,x1,y1);
    */
    return (0);
}


/* ARGSUSED */
int
WP_Arc(x0, y0, radius, theta1, theta2)

int x0, y0, radius;
double theta1, theta2;
{
    return (0);
}


/* ARGSUSED */
int
WP_Polygon(p)
POLYGON *p;
{
    int i = p->nvertices;
    int *v = (int*)p->xy;

    /* poly line record */

    putc(6,plotfile);
    putc(0xff,plotfile);
    putshort(4*i + 2,plotfile);
    putshort(i,plotfile);
    while (i) {
        putshort(*v,plotfile);
        v++;
        putshort(*v,plotfile);
        v++;
        i--;
    }
}


int
WP_Text(text, x, y)

char *text;
int x, y;
{
    int j;

    j = strlen(text);

    /* Text record */
    putc(12,plotfile);
    putc(j+6,plotfile);
    putshort(j,plotfile);
    putshort(x,plotfile);
    putshort(y,plotfile);
    while (*text) {
        putc(*text,plotfile);
        text++;
    }
    return (0);
}


int
WP_SetLinestyle(linestyleid)

int linestyleid;
{
    return (0);
}


/* ARGSUSED */
int
WP_SetColor(colorid)
{
    return (0);
}


int
WP_Update()

{
    fflush(plotfile);
    return (0);
}

