/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Author: 1995 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_X11
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Shell.h>

#include "x11util.h"

#ifdef __STDC__
static void ToTop(Widget, XtPointer, XEvent*, Boolean*);
#else
static void ToTop();
#endif
            
static char *oneLineTranslations =
    "<Key>Return: cr_action()\n<Key>Delete: delete-next-character()";
static char *butTranslations = "<Btn1Up>: notify() reset()";
static char *wmTranslations = "<Message>WM_PROTOCOLS: quit_action()";


void
PopUpInput(initial_str, action_str, action_callback, w)

String initial_str, action_str;
void (*action_callback)();
widget_bag *w;
{
    Widget form;
    Widget label;
    Widget cancel;
    Widget action;
    Dimension width, b_width, d;
    char **p;

    w->popup = XtVaCreatePopupShell("popup", transientShellWidgetClass,
        w->shell, NULL);
    XtOverrideTranslations(w->popup, XtParseTranslationTable(wmTranslations));
    XtAddEventHandler(w->popup, VisibilityChangeMask,
        False, ToTop, (XtPointer)w);

    form = XtVaCreateManagedWidget("form", formWidgetClass, w->popup,
        NULL);

    label = XtVaCreateManagedWidget("label", labelWidgetClass,
        form,
        XtNlabel, "Enter filename:                             ",
        XtNleft, XtChainLeft,
        XtNright, XtChainLeft,
        XtNresizable, TRUE,
        XtNborderWidth, 0,
        NULL);

    w->popup_text = XtVaCreateManagedWidget("text", asciiTextWidgetClass,
        form,
        XtNfromVert, label,
        XtNleft, XtChainLeft,
        XtNright, XtChainRight,
        XtNeditType, XawtextEdit,
        XtNresizable, TRUE,
        XtNresize, XawtextResizeWidth,
        XtNstring, initial_str,
        NULL);
    XtOverrideTranslations(w->popup_text,
        XtParseTranslationTable(oneLineTranslations));

    action = XtVaCreateManagedWidget("action", commandWidgetClass,
        form,
        XtNlabel, action_str,
        XtNfromVert, w->popup_text,
        XtNleft, XtChainLeft,
        XtNright, XtChainLeft,
        NULL);
    XtAddCallback(action, XtNcallback, action_callback, (XtPointer)w);
    XtOverrideTranslations(action, XtParseTranslationTable(butTranslations));

    cancel = XtVaCreateManagedWidget("cancel", commandWidgetClass,
        form,
        XtNlabel, "Cancel",
        XtNfromVert, w->popup_text,
        XtNfromHoriz, action,
        XtNleft, XtChainLeft,
        XtNright, XtChainLeft,
        NULL);
    XtAddCallback(cancel, XtNcallback, PopDownInput, (XtPointer)w);

    XtRealizeWidget(w->popup);
    XSetWMProtocols(Xdisplay, XtWindow(w->popup), &w->wm_delete, 1);
    CenterWidgetOnWidget(w->popup,w->shell);

    XtVaGetValues(form,
        XtNwidth, &width,
        XtNborderWidth, &b_width,
        XtNdefaultDistance, &d,
        NULL);
    width -= 2*(b_width + d + 1);
    XtVaSetValues(w->popup_text, XtNwidth, width, NULL);

    XtPopup(w->popup, XtGrabNone);
    XtSetKeyboardFocus(w->popup, w->popup_text);
    XtSetKeyboardFocus(w->shell, w->popup);
    for (p = w->popup_sens_list; p && *p; p++)
        XtSetSensitive(XtNameToWidget(w->butbox, *p), False);
}


/* ARGSUSED */
void
PopDownInput(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    widget_bag *w = (widget_bag*)client_data;
    char **p;

    XtSetKeyboardFocus(w->shell, None);
    XtPopdown(w->popup);
    for (p = w->popup_sens_list; p && *p; p++)
        XtSetSensitive(XtNameToWidget(w->butbox, *p), True);
    w->popup = NULL;
}


void
PopUpMessage(message_str, w)

String message_str;
widget_bag *w;
{
    Widget popup;
    Widget form;
    Widget label;
    Widget cancel;
    XColor visualcolor, exactcolor;

    popup = XtVaCreatePopupShell("popup_m", transientShellWidgetClass,
        w->shell, NULL);
    XtOverrideTranslations(popup, XtParseTranslationTable(wmTranslations));
    XtAddEventHandler(popup, VisibilityChangeMask,
        False, ToTop, (XtPointer)w);

    form = XtVaCreateManagedWidget("form_m", formWidgetClass, popup,
        NULL);

    cancel = XtVaCreateManagedWidget("cancel_m", commandWidgetClass,
        form,
        XtNlabel, "OK",
        XtNleft, XtChainLeft,
        XtNright, XtChainLeft,
        NULL);

    XAllocNamedColor(Xdisplay,
        DefaultColormap(Xdisplay, DefaultScreen(Xdisplay)),
        "red", &exactcolor, &visualcolor);
    label = XtVaCreateManagedWidget("label_m", labelWidgetClass,
        form,
        XtNlabel, message_str,
        XtNleft, XtChainLeft,
        XtNright, XtChainLeft,
        XtNfromHoriz, cancel,
        XtNresizable, TRUE,
        XtNborderWidth, 2,
        XtNborderColor, visualcolor.pixel,
        NULL);

    w->message = popup;
    XtAddCallback(cancel, XtNcallback, PopDownMessage, (XtPointer)w);
    XtAddEventHandler(form, KeyPressMask, False,
        (XtEventHandler)PopDownMessage, (XtPointer)w);
    XtRealizeWidget(popup);
    XSetWMProtocols(Xdisplay, XtWindow(popup), &w->wm_delete, 1);
    CenterWidgetOnWidget(popup, w->shell);
    if (w->popup)
        XtSetSensitive(w->popup, False);
    XtSetSensitive(w->form, False);
    XtSetKeyboardFocus(w->shell, form);
    XtPopup(popup, XtGrabExclusive);
}


/* ARGSUSED */
void
PopDownMessage(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    widget_bag *w = (widget_bag*)client_data;

    if (w->popup) {
        XtSetKeyboardFocus(w->shell, w->popup);
        XtPopdown(w->message);
        XtSetSensitive(w->popup, True);
    }
    else {
        XtSetKeyboardFocus(w->shell, None);
        XtPopdown(w->message);
    }
    XtSetSensitive(w->form, True);
}


void
PopUpErr(message_str, w)
    
String message_str;
widget_bag *w;
{   
    Widget popup;
    Widget form;
    Widget label;
    Widget cancel;
    XColor visualcolor, exactcolor;
    
    popup = XtVaCreatePopupShell("popup_e", transientShellWidgetClass,
        w->shell, NULL);
    XtOverrideTranslations(popup, XtParseTranslationTable(wmTranslations));
    XtAddEventHandler(popup, VisibilityChangeMask,
        False, ToTop, (XtPointer)w);

    form = XtVaCreateManagedWidget("form_e", formWidgetClass, popup,
        NULL);
    
    cancel = XtVaCreateManagedWidget("cancel_e", commandWidgetClass,
        form,
        XtNlabel, "OK",
        XtNleft, XtChainLeft,
        XtNright, XtChainLeft,
        NULL);

    XAllocNamedColor(Xdisplay,
        DefaultColormap(Xdisplay, DefaultScreen(Xdisplay)),
        "red", &exactcolor, &visualcolor);
    label = XtVaCreateManagedWidget("text_e", asciiTextWidgetClass,
        form,
        XtNstring, message_str,
        XtNleft, XtChainLeft,
        XtNright, XtChainRight,
        XtNbottom, XtChainBottom,
        XtNscrollHorizontal, XawtextScrollWhenNeeded,
        XtNscrollVertical, XawtextScrollWhenNeeded,
        XtNwidth, 500,
        XtNheight, 200,
        XtNfromHoriz, cancel,
        XtNdisplayCaret, False,
        XtNborderWidth, 2,
        XtNborderColor, visualcolor.pixel,
        NULL);
    
    XtAddCallback(cancel, XtNcallback, PopDownErr, (XtPointer)popup);
    XtAddEventHandler(form, KeyPressMask, False, (XtEventHandler)PopDownErr,
        (XtPointer)popup);
    XtRealizeWidget(popup);
    XSetWMProtocols(Xdisplay, XtWindow(popup), &w->wm_delete, 1);
    CenterWidgetOnWidget(popup, w->shell);
    XtPopup(popup, XtGrabNone);
}   


/* ARGSUSED */
void
PopDownErr(caller, client_data, call_data)
    
Widget caller;
XtPointer client_data, call_data;
{   
    XtPopdown((Widget)client_data);
} 


void
CenterWidgetOnWidget(sub, master)

Widget sub, master;
{
    Dimension width, height, b_width;
    Position x, y, max_x, max_y;

    XtVaGetValues(master,
        XtNx, &x,
        XtNy, &y,
        XtNwidth, &width,
        XtNheight, &height,
        XtNborderWidth, &b_width,
        NULL);

    width += 2*b_width;
    height += 2*b_width;
    height += 2*b_width;
    x += (Position)width/2;
    y += (Position)height/2;

    XtVaGetValues(sub,
        XtNwidth, &width,
        XtNheight, &height,
        XtNborderWidth, &b_width,
        NULL);

    width += 2*b_width;
    height += 2*b_width;
    x -= (Position)width/2;
    if (x < 0) x = 0;
    y -= (Position)height/2;
    if (y < 0) y = 0;
    if (y > (max_y = (Position) (XtScreen(sub)->height - height))) y = max_y;

    XtVaSetValues(sub, XtNx, x, XtNy, y, NULL);
}


check_type
CheckFile(fname, mode, w)

char *fname;
int mode;
widget_bag *w;
{
    char buf[512];
    struct stat st;
    FILE *fp;
    char *msg = "Error: can't %s file %s";

    /* check filename */
    if (!fname) return (NOGO);
    while (isspace(*fname)) fname++;
    if (!*fname) return (NOGO);

    if (!access(fname, F_OK)) {
        /* named file exists */
        if (stat(fname, &st))
            return (NOGO); /* shouldn't happen */

        if ((st.st_mode&S_IFMT) != S_IFLNK &&
                (st.st_mode&S_IFMT) != S_IFREG) goto bad;
            /* not a simple file or symbolic link */
    }

    switch (mode) {
    case R_OK:
        if (access(fname,R_OK)) {
            /* can't open for reading */
            if (!access(fname,F_OK)) goto bad; /* it exists, so error */
            return (NO_EXIST);
        }
        return (READ_OK);
    case W_OK:
        if (access(fname,W_OK)) {
            /* can't open for writing */
            if (!access(fname,F_OK)) goto bad; /* it exists, so error */
            fp = fopen(fname,"w");
            if (!fp) goto bad;
            fclose(fp);
        }
        return (WRITE_OK);
    }
bad:
    sprintf(buf, msg, (mode == R_OK) ? "read" : "write", fname);
    PopUpMessage(buf, w);
    return (NOGO);
}


/* ARGSUSED */
static void        
ToTop(caller, clientdata, event, ctd)
            
Widget caller;
XtPointer clientdata;
XEvent *event;
Boolean *ctd;
{       
    /*  prevent windows from disappearing */
    widget_bag *w = (widget_bag*)clientdata;

    XVisibilityEvent *vev = (XVisibilityEvent*)event;
    XWindowChanges xv;
    
    if (vev->state == VisibilityFullyObscured) {
        xv.sibling = XtWindow(w->shell);
        xv.stack_mode = Above;
        XReconfigureWMWindow(vev->display, vev->window,
            DefaultScreen(vev->display), CWSibling|CWStackMode, &xv);
    }       
}

#else
int x11_util_dummy;
#endif
