/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * SCED basic commands.
 */

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"



/***********************************************************************
 *
 * Commands to switch menus, undo.
 * 
 *
 ***********************************************************************/

extern char *MenuUNDO;


void
Basic(LookedAhead)

int *LookedAhead;
{
    *LookedAhead = False;
    Parameters.kpMenu = BASICMENU;
    FixMenuPrefix(BasicMenu);
    ShowCommandMenu();
    Parameters.kpColor = DrawingColor;
}


void
Undo()

{
    MenuSelect(MenuUNDO);
    ShowPrompt("Sorry, but it's too late.");
    MenuDeselect(MenuUNDO);
}



/***********************************************************************
 *
 * Commands for setting screen colors and grid parameters.
 * 
 *
 ***********************************************************************/

extern char *MenuBGRND;
extern char *MenuHLITE;
extern char *MenuRGB;
extern char *MenuPLUSR,*MenuMINSR;
extern char *MenuPLUSG,*MenuMINSG;
extern char *MenuPLUSB,*MenuMINSB;
extern char *MenuGRID;
extern char *MenuMENU;
extern char *MenuPRMPT;
extern char *MenuMORE;
extern char *MenuHELP;
extern char *MenuUPDAT;
extern char *MenuSUBCT;

#ifdef __STDC__
static void subct_color(int*);
#else
static void subct_color();
#endif


void
Color(LookedAhead)

int *LookedAhead;
{
    int i = 0;
    char OldMenu;

    AmbiguityMenu[i].mEntry = MenuHELP;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = "     ";
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuUPDAT;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = "     ";
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuBGRND;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuHLITE;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuMENU;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuPRMPT;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuMORE;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuGRID;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuSUBCT;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = "     ";
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuPLUSR;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuMINSR;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuPLUSG;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuMINSG;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuPLUSB;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuMINSB;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = "     ";
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = MenuRGB;
    AmbiguityMenu[i].mActive = False;
    i++;
    AmbiguityMenu[i].mEntry = NULL;
    OldMenu = Parameters.kpMenu;
    Parameters.kpMenu = AMBIGUITYMENU;
    FixMenuPrefix(AmbiguityMenu);
    ShowMenu(AmbiguityMenu);
    ShowPrompt("Hit ESC to exit.");

    loop {
        switch (PointLoop(LookedAhead)) {
            case PL_CMD:
                if (Matching(MenuSUBCT)) {
                    *LookedAhead = False;
                    subct_color(LookedAhead);
                    if (*LookedAhead == True && !Matching(MenuSUBCT))
                        goto quit;
                    continue;
                }
                if (Matching(MenuHELP)) {
                    SCEDhelp();
                    *LookedAhead = False;
                    continue;
                }
            case PL_ESC:
                goto quit;
            case PL_PCW:
                continue;
        }
    }
quit:
    Parameters.kpMenu = OldMenu;
    ShowCommandMenu();
}


void
Bgrnd(LookedAhead)

int *LookedAhead;
{
    MenuSelect(MenuBGRND);
    Parameters.kpColor = BackgroundColor;
    ShowPrompt("Set RGB of background color.");
    (void)PointColor(LookedAhead);
    Parameters.kpColor = DrawingColor;
    ErasePrompt();
    MenuDeselect(MenuBGRND);
}


void
Hlite(LookedAhead)

int *LookedAhead;
{
    MenuSelect(MenuHLITE);
    Parameters.kpColor = HighlightingColor;
    ShowPrompt("Set RGB of highlighting color.");
    (void)PointColor(LookedAhead);
    Parameters.kpColor = DrawingColor;
    ErasePrompt();
    MenuDeselect(MenuHLITE);
}


static void
subct_color(LookedAhead)

int *LookedAhead;
{
    MenuSelect(MenuSUBCT);
    Parameters.kpColor = InstanceBBColor;
    ShowPrompt(
        "Set RGB of unexpanded subcircuit box (point in window for next).");
    switch (PointColor(LookedAhead)) {
        case PL_ESC:
        case PL_CMD:
            goto quit;
    }
    Parameters.kpColor = InstanceNameColor;
    ShowPrompt("Set RGB of unexpanded subcircuit label.");
    (void)PointColor(LookedAhead);
quit:
    Parameters.kpColor = DrawingColor;
    ErasePrompt();
    MenuDeselect(MenuSUBCT);
}


void
RGB()

{
    MenuSelect(MenuRGB);
    ErasePrompt();
    ShowRGB();
    MenuDeselect(MenuRGB);
}


void
AlterColor(RGB,PlusOrMinus)

char RGB;
char PlusOrMinus;
{
    int *which;
    char *menu;
    double R,G,B;
    double M = FBMAXINTENSITY;

    switch (RGB) {
        case 'r':
        case 'R':
            if (PlusOrMinus == '+') 
                menu = MenuPLUSR;
            else
                menu = MenuMINSR;
            which = &ColorTable[Parameters.kpColor].klR;
            break;
        case 'g':
        case 'G':
            if (PlusOrMinus == '+') 
                menu = MenuPLUSG;
            else
                menu = MenuMINSG;
            which = &ColorTable[Parameters.kpColor].klG;
            break;
        case 'b':
        case 'B':
            if (PlusOrMinus == '+') 
                menu = MenuPLUSB;
            else
                menu = MenuMINSB;
            which = &ColorTable[Parameters.kpColor].klB;
            break;
        default:
            return;
    }
    MenuSelect(menu);
    if (PlusOrMinus == '+') {
        if (*which == FBMAXINTENSITY) *which = 0;
        else if (((*which) += 10) > FBMAXINTENSITY)
            *which = FBMAXINTENSITY;
    }
    else {
        if (*which == 0) *which = FBMAXINTENSITY;
        else if (((*which) -= 10) < 0) *which = 0;
    }
    R = ColorTable[Parameters.kpColor].klR/M;
    G = ColorTable[Parameters.kpColor].klG/M;
    B = ColorTable[Parameters.kpColor].klB/M;
    DevDefineColor(Parameters.kpColor,R,G,B);

    ShowRGB();
    MenuDeselect(menu);
}


void
SetGrid(LookedAhead)

int *LookedAhead;
{
    char *TypeIn;
    double d;
    int OldParameter = Parameters.kpShowGridInLargeViewport;
    int i;

    MenuSelect(MenuGRID);
    if (Parameters.kpMenu == AMBIGUITYMENU) {
        sprintf(TypeOut,"Hex code for grid line style, 0 for point grid (%x)?",
            Parameters.kpGridLineStyle);
        ShowPrompt(TypeOut);
        TypeIn = FBEdit(NULL);
        if (TypeIn == NULL) goto quit;
        if (sscanf(TypeIn,"%x",&i) == 1 && i >= 0 && i < 256)
            Parameters.kpGridLineStyle = i;

        sprintf(TypeOut,"Show grid above layout geometries (%c)?",
            Parameters.kpGridOnTop ? 'Y' : 'N');
        ShowPrompt(TypeOut);
        TypeIn = FBEdit(NULL);
        if (TypeIn == NULL) goto quit;
        if (*TypeIn == 'n' Or *TypeIn == 'N')
            Parameters.kpGridOnTop = False;
        elif (*TypeIn == 'y' Or *TypeIn == 'Y')
            Parameters.kpGridOnTop = True;

        sprintf(TypeOut,"Show grid in coarse viewport of a split screen (%c)?",
            Parameters.kpShowGridInLargeViewport ? 'Y' : 'N');
        ShowPrompt(TypeOut);
        TypeIn = FBEdit(NULL);
        if (TypeIn == NULL) goto quit;
        if (*TypeIn == 'n' Or *TypeIn == 'N')
            Parameters.kpShowGridInLargeViewport = False;
        elif (*TypeIn == 'y' Or *TypeIn == 'Y')
            Parameters.kpShowGridInLargeViewport = True;

        Parameters.kpColor = FineGridColor;
        ShowPrompt(
            "Set RGB for fine grid lines (point in window for next).");
        switch (PointColor(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                goto quit;
        }

        Parameters.kpColor = CoarseGridColor;
        ShowPrompt("Set RGB for coarse grid lines.");
        switch (PointColor(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                goto quit;
        }
    }
    sprintf(TypeOut,
        "Enter new grid interval (currently %g, CR to turn grid %s): ",
        (double)Parameters.kpGrid/RESOLUTION,
        Parameters.kpGridDisplayed ? "off" : "on" );
    ShowPrompt(TypeOut);
    TypeIn = FBEdit(NULL);
    if (TypeIn != NULL) {
        if (!strlen(TypeIn)) {
            Parameters.kpGridDisplayed ^= 1;
            ErasePrompt();
        }
        elif (sscanf(TypeIn,"%lg",&d) == 1 And d > 0) {
            Parameters.kpGrid = RESOLUTION*d;
            if (Parameters.kpGrid == 0)
                Parameters.kpGrid = 1;
            sprintf(TypeOut,"Grid line spacing now %g.",
                (double)Parameters.kpGrid/RESOLUTION);
            ShowPrompt(TypeOut);
        }
        else
            goto quit;
    }
quit:
    if (Parameters.kpShowGridInLargeViewport Or OldParameter)
        RedisplaySCED(currentgraph);
    else
        ShowFineViewport();
    ErasePrompt();
    Parameters.kpColor = DrawingColor;
    MenuDeselect(MenuGRID);
}


void
MenuParams(LookedAhead)

int *LookedAhead;
{
    MenuSelect(MenuMENU);

    Parameters.kpColor = MenuTextColor;
    ShowPrompt("Set RGB for menu text (point in window for next).");
    switch (PointColor(LookedAhead)) {
        case PL_ESC:
        case PL_CMD:
            goto quit;
    }

    Parameters.kpColor = MenuTextPrefixColor;
    ShowPrompt("Set RGB for menu prefix (point in window for next).");
    switch (PointColor(LookedAhead)) {
        case PL_ESC:
        case PL_CMD:
            goto quit;
    }

    Parameters.kpColor = MenuSelectedColor;
    ShowPrompt(
        "Set RGB for selected menu text (point in window for next).");
    switch (PointColor(LookedAhead)) {
        case PL_ESC:
        case PL_CMD:
            goto quit;
    }

    Parameters.kpColor = MenuSelectedPrefixColor;
    ShowPrompt(
        "Set RGB for selected prefix (point in window for next).");
    switch (PointColor(LookedAhead)) {
        case PL_ESC:
        case PL_CMD:
            goto quit;
    }

    Parameters.kpColor = MenuHighlightingColor;
    ShowPrompt("Set RGB for menu highlighting.");
    (void)PointColor(LookedAhead);

quit:
    ErasePrompt();
    Parameters.kpColor = DrawingColor;
    MenuDeselect(MenuMENU);
}


void
MenuPrompt(LookedAhead)

int *LookedAhead;
{
    MenuSelect(MenuPRMPT);
    Parameters.kpColor = PromptTextColor;
    ShowPrompt("Set RGB for message text.");
    (void)PointColor(LookedAhead);
    Parameters.kpColor = DrawingColor;
    ErasePrompt();
    MenuDeselect(MenuPRMPT);
}


void
MenuMore(LookedAhead)

int *LookedAhead;
{
    MenuSelect(MenuMORE);
    Parameters.kpColor = MoreTextColor;
    ShowPromptWithColor("Set RGB for \"more\" window text.",
        MoreTextColor);
    (void)PointColor(LookedAhead);
    Parameters.kpColor = DrawingColor;
    ErasePrompt();
    MenuDeselect(MenuMORE);
}


void
ShowRGB()

{
    int r,g,b;

    r = ColorTable[Parameters.kpColor].klR;
    g = ColorTable[Parameters.kpColor].klG;
    b = ColorTable[Parameters.kpColor].klB;

    OutlineText(Parameters.kpLastCursorColumn,gi_numtextrows,gi_numtextcols,
        gi_numtextrows,FILL,ERASE,0);

    sprintf(TypeOut,"  (%d,%d,%d)",r,g,b);

    DevSetColor(PromptTextColor);
    FBText(ROW_COLUMN,gi_numtextrows,Parameters.kpLastCursorColumn,TypeOut);
}



/***********************************************************************
 *
 * Edit, save, and write commands.
 * Manage symbol files.
 *
 ***********************************************************************/

extern char *MenuEDIT;
extern char *MenuSAVE;


void
Edit(Ready,Center,Modified)

int Ready,Center,Modified;
{
    /*
     * Ready == True if the name of the cell to be edited is
     * already in Parameters.kpCellName.
     * Center == True if the cell should be automatically centered.
     * Pop does NOT want it to be--it wants it to be exactly where it was
     * when it was pushed into.
     */
    char *TypeIn,*CellName,*NextCellName();
    char cbuf[128];
    struct s *TempCellDesc;
    FILE *fp, *gp;

    MenuSelect(MenuEDIT);

    TempCellDesc = Parameters.kpCellDesc;

    if (Not Ready){
        ShowPrompt("Cell?");
        CellName = FBEdit(NULL);
        if (CellName == NULL) {
            ErasePrompt();
            MenuDeselect(MenuEDIT);
            return;
        }
        while (isspace(*CellName)) CellName++;
        if (*CellName == '\0') {
            /* get next cellname from argument list */
            CellName = NextCellName();
            if (CellName == NULL) {
                if (Parameters.kpCellName == NULL 
                    Or *Parameters.kpCellName == '\0')
                    CellName = DEFAULT_EDIT_FILE;
                else
                    CellName = Parameters.kpCellName;
            }
        }
        else {
            strcpy(cbuf,CellName);
            CellName = cbuf;
        }

        fp = POpen(CellName,"r",(char **)NULL);
        if (fp) {
            if (!IsSymFile(fp)) {

                /* If model library is called by name, and
                 * the file is not in the CWD, drop a copy
                 * of the system model library into the CWD.
                 */
                if (!strcmp(CellName,ModelLibrary)) {
                    if (access(ModelLibrary,0)) {
                        int c;

                        rewind(fp);
                        gp = fopen(ModelLibrary,"w");
                        if (gp) {
                            while ((c = getc(fp)) != EOF)
                                putc(c,gp);
                            fclose(gp);
                        }
                    }
                }
                fclose(fp);

                if (!dispdev->windows) {
                    DevHalt();
                    (void)inp_edit(CellName);

                    if (substring(ModelLibrary,CellName)) {
                        ModelLibraryClose();
                        ModelLibraryOpen(ModelLibrary);
                    }

                    DevNewViewport(currentgraph);
                    MenuDeselect(MenuEDIT);
                    RedisplaySCED(currentgraph);
                    sprintf(TypeOut,"Editing: %s",Parameters.kpCellName);
                    ShowPrompt(TypeOut);
                }
                else {
                    (void)inp_edit(CellName);
                    if (substring(ModelLibrary,CellName)) {
                        ModelLibraryClose();
                        ModelLibraryOpen(ModelLibrary);
                    }
                    MenuDeselect(MenuEDIT);
                    ErasePrompt();
                }
                return;
            }
            else
                fclose(fp);
        }

        if (Parameters.kpModified && Parameters.kpCellName[0] != EOS
                && Parameters.kpCellDesc) {
            sprintf(TypeOut,
                "You've modified %s.  Do you want to save it (y)?",
                Parameters.kpCellName);
            ShowPrompt(TypeOut);
            TypeIn = FBEdit(NULL);
            if (TypeIn == NULL) {
                ErasePrompt();
                MenuDeselect(MenuEDIT);
                return;
            }
            if (TypeIn[0] != 'n' && TypeIn[0] != 'N')
                WriteCell();
        }

        ClearContext();

        if (!strcmp(Parameters.kpCellName,CellName)) {
            if (TempCellDesc) {
                if (Not CDClose(TempCellDesc))
                    MallocFailed();
            }
        }
        else
            strcpy(Parameters.kpCellName,CellName);
    }

    /* strip off any path prefix */
    if (FixCellName(Parameters.kpCellName)) {
        ShowPromptAndWait("Warning: internal path overflow");
    }

    SQClear(); /* clear selset queue */
    GPclearGraf(); /* clear graf references */
    if (Parameters.kpShowTerminals)
        DisplayTerminals(ERASE);
    ShowPrompt("Building database.  Please wait.");
    if (Not CDOpen(Parameters.kpCellName,&Parameters.kpCellDesc,'w')) {
        CDClose(Parameters.kpCellDesc);
        Parameters.kpCellDesc = NULL;
        DefaultWindows();
        EraseLargeCoarseViewport();
        Redisplay(View->kvCoarseWindow);
        sprintf(TypeOut,"Can't edit cell %s. MORE",Parameters.kpCellName); 
        ShowPrompt(TypeOut);
        (void)FBGetchar(ERASE);
        sprintf(TypeOut,"%s",CDStatusString);
        ShowPrompt(TypeOut);
        Parameters.kpCellName[0] = EOS;
    }
    else {
        ConnectRecursive(Parameters.kpCellDesc);
        if (TempCellDesc && (TempCellDesc != Parameters.kpCellDesc))
            if (Not CDClose(TempCellDesc))
                MallocFailed();
        strcpy(Parameters.kpTopName,Parameters.kpCellName);
        Parameters.kpTopDesc = Parameters.kpCellDesc;

        if (CDStatusInt == CDNEWSYMBOL)
            DefaultWindows();
        elif (Center)
            CenterFullView();
        EraseLargeCoarseViewport();

        /*
         * CDUnmark is expensive and has two purposes!
         * First, we make sure that all objects begin with a 0 info field.
         * We will therefore see EVERYTHING in the symbol during redisplay.
         * Secondly, the bounding boxes are recomputed and therefore
         * propogated correctly through the hierarchy.
         */
        CDUnmark(Parameters.kpCellDesc);
        Redisplay(View->kvCoarseWindow);
        if (CDStatusInt == CDNEWSYMBOL) {
            sprintf(TypeOut,"Editing new cell: %s",Parameters.kpCellName);
            ShowPrompt(TypeOut);
        }
        else {
            sprintf(TypeOut,"Editing: %s",Parameters.kpCellName);
            ShowPrompt(TypeOut);
            if (Parameters.kpShowTerminals)
                DisplayTerminals(DISPLAY);
        }
        Parameters.kpModified = Modified;
        TitleWindow();
    }
    SaveLastView();
    MenuDeselect(MenuEDIT);
}


void
TitleWindow()

{
    char Title[80];
 
    sprintf(Title,"Circuit: %s",Parameters.kpCellName);
    DevTitle(Title,"sced");
}
 

void
Save()

{
    FILE *fp;

    ShowPrompt("Saving circuit.    Please wait.");
    if (CDUpdate(Parameters.kpCellDesc,(char *)NULL)) {
        Parameters.kpModified = False;
        if (Parameters.kpCellDesc == Parameters.kpTopDesc) {
            fp = POpen(Parameters.kpCellName,"a",(char **)NULL);
            if (fp) {
                DumpSpiceFile(fp);
                fclose(fp);
            }
        }
        else if (Not CDReflect(Parameters.kpCellDesc))
            MallocFailed();
        ShowPrompt("Current circuit has been saved.");
    }
    else
        ShowPrompt("Can't save circuit.");
}


void
WriteCell()

{
    struct prpty *PDesc;
    char *TypeIn,NewName[128],*t;
    FILE *fp;

    MenuSelect(MenuSAVE);
    if (Parameters.kpCellName[0] == '\0') {
        ShowPrompt("There isn't anything to write.");
        MenuDeselect(MenuSAVE);
        return;
    }
    ShowPrompt("Circuit name?");
    TypeIn = FBEdit(Parameters.kpCellName);
    if (TypeIn == NULL Or *TypeIn == '\0' Or *TypeIn == '\n') {
        MenuDeselect(MenuSAVE);
        ErasePrompt();
        return;
    }
    t = NewName;
    while (isspace(*TypeIn)) TypeIn++;
    if (*TypeIn == '\0') {
        MenuDeselect(MenuSAVE);
        ErasePrompt();
        return;
    }
    while (*TypeIn && !isspace(*TypeIn))
        *t++ = *TypeIn++;
    *t = '\0';

    ShowPrompt("Writing circuit.    Please wait.");

    if (strcmp(Parameters.kpCellName,NewName)) {
        /* new name given */
        TypeIn = Parameters.kpCellDesc->sName;
        Parameters.kpCellDesc->sName = NewName;

        /* change the subcircuit's reference name */
        PDesc = Parameters.kpCellDesc->sPrptyList;
        for (; PDesc; PDesc = PDesc->prpty_Succ)
            if (PDesc->prpty_Value == P_NAME) break;
        if (PDesc && PDesc->prpty_Data->p_name.subname) {
            t = PDesc->prpty_Data->p_name.subname;
            PDesc->prpty_Data->p_name.subname = NewName;
        }

        if (Not CDUpdate(Parameters.kpCellDesc,NewName)) {
            ShowPrompt("Can't save circuit.");
            MenuDeselect(MenuSAVE); 
            Parameters.kpCellDesc->sName = TypeIn;
            return;
        }
        if (PDesc && PDesc->prpty_Data->p_name.subname) {
            PDesc->prpty_Data->p_name.subname = t;
        }
        Parameters.kpModified = False;
        Parameters.kpCellDesc->sName = TypeIn;
        if (Parameters.kpCellDesc == Parameters.kpTopDesc) {
            fp = POpen(NewName,"a",(char **)NULL);
            if (fp) {
                DumpSpiceFile(fp);
                fclose(fp);
            }
        }
        UpdateParent(NewName);
        ShowPrompt("New circuit has been saved.");
    }
    else {
        if (Not CDUpdate(Parameters.kpCellDesc,(char *)NULL)) {
            ShowPrompt("Can't save circuit.");
            MenuDeselect(MenuSAVE); 
            return;
        }
        if (Parameters.kpCellDesc == Parameters.kpTopDesc) {
            fp = POpen(Parameters.kpCellName,"a",(char **)NULL);
            if (fp) {
                DumpSpiceFile(fp);
                fclose(fp);
            }
        }
        ShowPrompt("Current circuit has been saved.");
        Parameters.kpModified = False;
        if (Not CDReflect(Parameters.kpCellDesc))
            MallocFailed();
    }
    MenuDeselect(MenuSAVE);
}


static void
newtok(p,t)

char **p, *t;
{
    char *s = *p;

    while (isspace(*s)) s++;
    while (*s && !isspace(*s)) *t++ = *s++;
    *t = '\0';
    while (isspace(*s)) s++;
    *p = s;
}


int
FixCellName(cname)

/* Strip off any path prefix and add to search path it not already
 * there.
 */
char *cname;
{
    char *c, *p;
    char tok[81];
    int i, ret = 0;

    c = strrchr(cname,DIR_TERM);
    if (!c)
        return (0);
    *c = '\0';

    p = PGetPath();
    while (*p) {
        newtok(&p,tok);
#ifdef MSDOS
        if (!stricmp(tok,cname))
#else
        if (!strcmp(tok,cname))
#endif
            goto done;
    }
    ret = PAppendPath(cname);

done:
    for (i = 0, c++; *c; i++, c++)
        cname[i] = *c;
    cname[i] = '\0';
    return (ret);
}



/***********************************************************************
 *
 * Rdraw and view commands.
 * Expnd command - expand subcells.
 * Update the screen.
 *
 ***********************************************************************/

extern char *MenuRDRAW;
extern char *MenuVIEW;
extern char *MenuEXPND;


void
Rdraw()

{
    MenuSelect(MenuRDRAW);
    RedisplaySCED(currentgraph);
    MenuDeselect(MenuRDRAW);
}


void
ShowFull()

{
    MenuSelect(MenuVIEW);
    SaveLastView();
    EraseLargeCoarseViewport();
    CenterFullView();
    Redisplay(View->kvCoarseWindow);
    MenuDeselect(MenuVIEW);
}


void
Expand()

{
    extern char *MenuEXPND;

    if (Parameters.kpExpandInstances) {
        MenuDeselect(MenuEXPND);
        Parameters.kpExpandInstances = False;
    }
    else {
        MenuSelect(MenuEXPND);
        Parameters.kpExpandInstances = True;
    }
    EraseLargeCoarseViewport();
    Redisplay(View->kvCoarseWindow);
}
