/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/* Enhanced property handling for SCED */


#include "spice.h"
#include "sced.h" 

#ifdef __STDC__
static void prpty_free(struct prpty*);
#else
static void prpty_free();
#endif

#define umalloc(u) (union u*)tmalloc(sizeof(union u))

void
CDProperty(SymbolDesc,Pointer,Property)

struct s *SymbolDesc;
struct o *Pointer;
struct prpty **Property;
{
    /*
     * Return info field of object.
     * If Pointer == NULL, object is symbol itself. 
     */
    if(Pointer == NULL)
        *Property = SymbolDesc->sPrptyList;
    else
        *Property = Pointer->oPrptyList; 
}


int
CDCopyProperty(SymbolDesc,Pointer,PDesc)

struct s *SymbolDesc;
struct o *Pointer;
struct prpty *PDesc;
{
    struct prpty *prptyDesc;
    union prp_data *p_data;

    prptyDesc = alloc(prpty);

    if (PDesc->prpty_String)
        prptyDesc->prpty_String = CopyString(PDesc->prpty_String);

    else if (PDesc->prpty_Data) {
        switch (PDesc->prpty_Value) {
            case P_MODEL:
            case P_VALUE:
            case P_INITC:
            case P_OTHER:
                prptyDesc->prpty_Data = (union prp_data*)
                    HYcopy((struct hprlist*)PDesc->prpty_Data);
                break;
            default:
                p_data = umalloc(prp_data);
                *p_data = *PDesc->prpty_Data;
                if (PDesc->prpty_Value == P_NAME) {
                    p_data->p_name.name = CopyString(p_data->p_name.name);
                    if (p_data->p_name.subname)
                        p_data->p_name.subname =
                            CopyString(p_data->p_name.subname);
                }
                prptyDesc->prpty_Data = p_data;
        }
    }
    if (PDesc->prpty_Info)
        prptyDesc->prpty_Info = CopyString(PDesc->prpty_Info);

    prptyDesc->prpty_Value = PDesc->prpty_Value;

    if (Pointer == NULL) {
        prptyDesc->prpty_Succ = SymbolDesc->sPrptyList;
        SymbolDesc->sPrptyList = prptyDesc;
    }
    else {
        prptyDesc->prpty_Succ = Pointer->oPrptyList;
        Pointer->oPrptyList = prptyDesc; 
    }
    return (True);
}


int
CDAddProperty(SymbolDesc,Pointer,Value,String)

struct s *SymbolDesc;
struct o *Pointer;
int Value;
char *String;
{
    char *cp, name[128], subname[128];
    struct prpty *prptyDesc;
    union prp_data *p_data;
    int i, indx1, indx2;
    int X, Y, X1, Y1;
    double coeff;

    if (String == NULL)
        goto bad;

    prptyDesc = alloc(prpty);

    switch (Value) {
        case P_NODE:
            if (Pointer && Pointer->oType == CDWIRE) {
                if (sscanf(String,"%d",&indx1) != 1)
                    goto bad;
                indx2 = 0;
                X = Y = 0;
            }
            else if (sscanf(String,"%d %d %d %d",&indx1,&indx2,&X,&Y) != 4)
                goto bad;
            p_data = umalloc(prp_data);
            p_data->p_node.enode = indx1;
            p_data->p_node.inode = indx2;
            p_data->p_node.x = X;
            p_data->p_node.y = Y;
            prptyDesc->prpty_Data = p_data;
            break;
        case P_MUT:
            if (sscanf(String,"%d %d %d %d %lf",&X,&Y,&X1,&Y1,&coeff) != 5)
                goto bad;
            p_data = umalloc(prp_data);
            p_data->p_mut.x1 = X;
            p_data->p_mut.y1 = Y;
            p_data->p_mut.x2 = X1;
            p_data->p_mut.y2 = Y1;
            p_data->p_mut.coeff = coeff;
            prptyDesc->prpty_Data = p_data;
            break;
        case P_BRANCH:
            if (sscanf(String,"%d %d %d %d",&X,&Y,&X1,&Y1) != 4)
                goto bad;
            p_data = umalloc(prp_data);
            p_data->p_branch.x = X;
            p_data->p_branch.y = Y;
            p_data->p_branch.r1 = X1;
            p_data->p_branch.r2 = Y1;
            prptyDesc->prpty_Data = p_data;
            break;
        case P_NAME:
            i = sscanf(String,"%s %d %s",name,&indx1,subname);
            if (i < 2)
                goto bad;
            p_data = umalloc(prp_data);
            p_data->p_name.name = CopyString(name);
            p_data->p_name.num = indx1;
            if (i == 3)
                p_data->p_name.subname = CopyString(subname);
            prptyDesc->prpty_Data = p_data;
            break;
        case P_MODEL:
        case P_VALUE:
        case P_INITC:
        case P_OTHER:
            prptyDesc->prpty_Data = (union prp_data*)HYfromascii(String);
            break;
        default:
            prptyDesc->prpty_String = CopyString(String);

            /* we can't allow semicolons because of CIF */
            for (cp = prptyDesc->prpty_String; *cp; cp++)
                if (*cp == ';') *cp = ' ';
    }
    prptyDesc->prpty_Value = Value;

    if (Pointer == NULL) {
        prptyDesc->prpty_Succ = SymbolDesc->sPrptyList;
        SymbolDesc->sPrptyList = prptyDesc;
    }
    else {
        prptyDesc->prpty_Succ = Pointer->oPrptyList;
        Pointer->oPrptyList = prptyDesc; 
    }
    return (True);
bad:
    return (False);
}


int
CDRemoveProperty(SymbolDesc,Pointer,Value)

struct s *SymbolDesc;
struct o *Pointer;
int Value;
{
    struct prpty *prptyDesc;
    struct prpty *prptyHead;
    struct prpty *prptyCopy;
    struct prpty *prptyTemp;

    if (Pointer == NULL)
        prptyHead = prptyDesc = SymbolDesc->sPrptyList;
    else
        prptyHead = prptyDesc = Pointer->oPrptyList; 

    for (prptyCopy = NULL; prptyDesc != NULL;
        prptyCopy = prptyDesc,prptyDesc = prptyTemp) {
        prptyTemp = prptyDesc->prpty_Succ;

        if (prptyDesc->prpty_Value == Value) {
            if (prptyCopy == NULL)
                prptyHead = prptyDesc->prpty_Succ;
            else
                prptyCopy->prpty_Succ = prptyDesc->prpty_Succ;
            prpty_free(prptyDesc);
            prptyDesc = prptyCopy;
        }
    }
    if (Pointer == NULL)
        SymbolDesc->sPrptyList = prptyHead;
    else
        Pointer->oPrptyList = prptyHead;
    return (True);
}


static void
prpty_free(PDesc)

struct prpty *PDesc;
{
    if (PDesc->prpty_String)
        free(PDesc->prpty_String);
    if (PDesc->prpty_Data) {
        switch (PDesc->prpty_Value) {
            case P_MODEL:
            case P_VALUE:
            case P_INITC:
            case P_OTHER:
                HYfree((struct hprlist*)PDesc->prpty_Data);
                break;
            case P_NAME:
                if (PDesc->prpty_Data->p_name.name)
                    free(PDesc->prpty_Data->p_name.name);
                if (PDesc->prpty_Data->p_name.subname)
                    free(PDesc->prpty_Data->p_name.subname);
            default:
                free(PDesc->prpty_Data);
        }
    }
    if (PDesc->prpty_Info)
        free(PDesc->prpty_Info);
    free(PDesc);
}


void
CDPrptyListFree(PDesc)

struct prpty *PDesc;
{
    struct prpty *pd;

    for (; PDesc; PDesc = pd) {
        pd = PDesc->prpty_Succ;
        prpty_free(PDesc);
    }
}


void
CDPrptyListPrint(FileDesc,PDesc)

FILE *FileDesc;
struct prpty *PDesc;
{
    union prp_data *p_data;
    char *s;

    while (PDesc) {
        if (PDesc->prpty_String)
            fprintf(FileDesc,"5 %d %s;\n",PDesc->prpty_Value,
                PDesc->prpty_String);

        else if (PDesc->prpty_Data) {
            p_data = PDesc->prpty_Data;

            switch (PDesc->prpty_Value) {
                case P_NODE:
                    fprintf(FileDesc,"5 %d %d %d %d %d;\n",
                        PDesc->prpty_Value,
                        p_data->p_node.enode,p_data->p_node.inode,
                        p_data->p_node.x,p_data->p_node.y);
                    break;
                case P_MUT:
                    fprintf(FileDesc,"5 %d %d %d %d %d %g;\n",
                        PDesc->prpty_Value,
                        p_data->p_mut.x1,p_data->p_mut.y1,
                        p_data->p_mut.x2,p_data->p_mut.y2,
                        p_data->p_mut.coeff);
                    break;
                case P_BRANCH:
                    fprintf(FileDesc,"5 %d %d %d %d %d;\n",
                        PDesc->prpty_Value,
                        p_data->p_branch.x,p_data->p_branch.y,
                        p_data->p_branch.r1,p_data->p_branch.r2);
                    break;
                case P_NAME:
                    if (p_data->p_name.subname)
                        fprintf(FileDesc,"5 %d %s %d %s;\n",
                            PDesc->prpty_Value,
                            p_data->p_name.name,p_data->p_name.num,
                            p_data->p_name.subname);
                    else
                        fprintf(FileDesc,"5 %d %s %d;\n",
                            PDesc->prpty_Value,
                            p_data->p_name.name,p_data->p_name.num);
                    break;
                case P_VALUE:
                case P_MODEL:
                case P_INITC:
                case P_OTHER:
                    s = HYtoascii((struct hprlist*)p_data);
                    if (s) {
                        fprintf(FileDesc,"5 %d %s;\n",PDesc->prpty_Value,s);
                        txfree(s);
                    }
                    break;
            }
        }
        PDesc = PDesc->prpty_Succ;
    }
}
