/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * SCED copy and move selection operators.
 */

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"

#define MOVE 1
#define COPY 0

extern char *MenuCOPY;
extern char *MenuMOVE;
extern char *MenuUNDO;

#ifdef __STDC__
static void restore_color(void);
static void show_device(struct s*,struct ka*);
static void free_path(struct p*);
static void show_move(struct ka*,struct ka*);
static int  copy_OK(long,long);
static void do_copy(long,long,long,long,struct ka*,int);
static char fix_xform(int);
static void do_copy_call(long,long,long,long,struct o*,struct o**);
#else
static void restore_color();
static void show_device();
static void free_path();
static void show_move();
static int  copy_OK();
static void do_copy();
static char fix_xform();
static void do_copy_call();
#endif


void
Copy(LookedAhead)

int *LookedAhead;
{
    struct ka BB;
    long RefX,RefY,CopX,CopY;
    int i;
    int Undo = False;
    int FirstTime = True;
    int GotOne = False;
    int Modified = 0;

    MenuSelect(MenuCOPY);

    if (SelectQHead != NULL)
        GotOne = True;
top:

    if (Not GotOne) {
        ShowPrompt("Point to object to copy.");
        switch (PointLoop(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
            case PL_UND:
                goto quit;
            case PL_PCW:
                BB.kaLeft = BB.kaRight = SCursor.kcRawX;
                BB.kaBottom = BB.kaTop = SCursor.kcRawY;
                Selection(&BB);
                if (SelectQHead == NULL)
                goto top;
        }
    }

Next:
    ShowPrompt("Point to the reference point.");

    switch (PointLoop(LookedAhead)) {
        case PL_ESC:
        case PL_CMD:
            goto quit;
        case PL_UND:
            if (Not GotOne) {
                SQComputeBB();
                SQDesel((char*)NULL);
                EraseBox(&SelectQBB);
                Redisplay(&SelectQBB);
                goto top;
            }
            goto quit;
    }

    restore_color();
    FBSetRubberBanding('m');
    RefX = SCursor.kcX;
    RefY = SCursor.kcY;
    ShowPrompt("Point to locations where the selected items will be copied.");

    loop {
        switch (PointLoop(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                goto quit;
            case PL_UND:
                MenuSelect(MenuUNDO);
                if (FirstTime) {
                    MenuDeselect(MenuUNDO);
                    FBSetRubberBanding(0);
                    goto Next;
                }
                if (Undo == False) {
                    SQRestore(True);
                    EraseBox(&BB);
                    Redisplay(&BB);
                    Modified--;
                    Undo = True;
                }
                else {
                    do_copy(RefX,RefY,CopX,CopY,&BB,COPY);
                    Redisplay(&BB);
                    Modified++;
                    Undo = False;
                }
                MenuDeselect(MenuUNDO);
                continue;
            case PL_PCW:
                CopX = SCursor.kcX;
                CopY = SCursor.kcY;
                i = copy_OK(CopX-RefX,CopY-RefY);
                if (i == -1) goto quit;
                if (i == 0)  continue;
                SQRestore(False);
                do_copy(RefX,RefY,CopX,CopY,&BB,COPY);
                FirstTime = False;
                Undo = False;
                Modified++;
                Redisplay(&BB);
                continue;
        }
    }

quit:
    FBSetRubberBanding(0);
    SQRestore(False);
    if (Not GotOne And AreTypesInQ((char*)NULL)) {
        SQComputeBB();
        SQDesel((char*)NULL);
        EraseBox(&SelectQBB);
        Redisplay(&SelectQBB);
    }
    if (Modified)
        Parameters.kpModified = True;
    ErasePrompt();
    MenuDeselect(MenuUNDO);
    MenuDeselect(MenuCOPY);
}


void
Move(LookedAhead)

int *LookedAhead;
{
    struct ka OldBB,NewBB;
    long LastRefX,LastRefY;
    long RefX,RefY,MovX,MovY;
    int  FirstTime = True;
    int GotOne = False;
    int Modified = 0;
    int Undo = False;

    MenuSelect(MenuMOVE);

    if (SelectQHead != NULL)
        GotOne = True;

    loop {
top:
        if (Not GotOne) {
            ShowPrompt("Point to object to move.");
            switch (PointLoop(LookedAhead)) {
                case PL_ESC:
                case PL_CMD:
                    goto quit;
                case PL_UND:
                    if (FirstTime) goto quit;
                    MenuSelect(MenuUNDO);
                    if (Undo == False) {
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(ERASE);
                        UndoReferenceTransform();
                        HYundoTransform();
                        SQRestore(True);
                        /* restored objects have Info = SQ_NEWSEL */
                        show_move(&NewBB,&OldBB);
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(DISPLAY);
                        Modified--;
                        Undo = True;
                        MenuDeselect(MenuUNDO);
                        break;
                    }
                    else {
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(ERASE);
                        /* should have only Info = SQ_NEW objects here */
                        do_copy(RefX,RefY,MovX,MovY,&NewBB,MOVE);
                        SQDesel((char*)NULL);
                        show_move(&OldBB,&NewBB);
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(DISPLAY);
                        Modified++;
                        Undo = False;
                        MenuDeselect(MenuUNDO);
                        continue;
                    }
                case PL_PCW:
                    SelectTypes((char*)NULL);
                    if (Not AreTypesInQ((char*)NULL))
                        goto top;
            }
        }

next:
        ShowPrompt("Point to the reference point.");
        switch (PointLoop(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                goto quit;
            case PL_UND:
                MenuSelect(MenuUNDO);
                if (Not GotOne) {
                    SQComputeBB();
                    SQDesel((char*)NULL);
                    /* newly selected objects deleted, restored objects
                     * have Info = SQ_NEW.
                     */
                    EraseBox(&SelectQBB);
                    Redisplay(&SelectQBB);
                }
                else {
                    if (FirstTime) {
                        MenuDeselect(MenuUNDO);
                        goto quit;
                    }
                    if (Undo == False) {
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(ERASE);
                        UndoReferenceTransform();
                        HYundoTransform();
                        SQRestore(True);
                        /* restored objects have Info = SQ_NEWSEL */
                        show_move(&OldBB,&NewBB);
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(DISPLAY);
                        Modified--;
                        Undo = True;
                    }
                    else {
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(ERASE);
                        /* should have only Info = SQ_NEWSEL objects here */
                        do_copy(RefX,RefY,MovX,MovY,&NewBB,MOVE);
                        show_move(&OldBB,&NewBB);
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(DISPLAY);
                        Modified++;
                        Undo = False;
                    }
                }
                MenuDeselect(MenuUNDO);
                goto top;
        }

        restore_color();
        FBSetRubberBanding('m');
        LastRefX = RefX;
        LastRefY = RefY;
        RefX = SCursor.kcX;
        RefY = SCursor.kcY;
        ShowPrompt("Point to where the selected items will be moved.");

        switch (PointLoop(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                goto quit;
            case PL_UND:
                MenuSelect(MenuUNDO);
                RefX = LastRefX;
                RefY = LastRefY;
                MenuDeselect(MenuUNDO);
                FBSetRubberBanding(0);
                goto next;
        }

        FBSetRubberBanding(0);
        SQComputeBB();
        OldBB = SelectQBB;
        MovX = SCursor.kcX;
        MovY = SCursor.kcY;
        SQRestore(False);
        if (Parameters.kpShowTerminals) DisplayTerminals(ERASE);
        do_copy(RefX,RefY,MovX,MovY,&NewBB,MOVE);
        FirstTime = False;
        Modified++;
        Undo = False;
        if (Not GotOne)
            SQDesel((char*)NULL);
        show_move(&OldBB,&NewBB);
        if (Parameters.kpShowTerminals) DisplayTerminals(DISPLAY);
    }
quit:
    FBSetRubberBanding(0);
    ClearReferenceUndoFlags();
    HYclearUndoFlags();
    SQRestore(False);
    if (Not GotOne And AreTypesInQ((char*)NULL)) {
        SQComputeBB();
        SQDesel((char*)NULL);
        EraseBox(&SelectQBB);
        Redisplay(&SelectQBB);
    }
    if (Modified)
        Parameters.kpModified = True;
    ErasePrompt();
    MenuDeselect(MenuUNDO);
    MenuDeselect(MenuMOVE);
}


void
ShowGhost(NewX,NewY,RefX,RefY)

/* Called from rubber banding routine. */
long NewX,NewY,RefX,RefY;
{
    struct ks *SQDesc;
    struct p *Path;
    struct ka BB;
    struct s *MasterDesc;
    int  Layer;
    int NumX,NumY,Int1,Int2;
    long Width;
    long TFold[9], TFnew[9], TFtemp[9];
    long DX,DY;
    char *MasterName;


    TCurrent(TFold);
    TPush();
    TIdentity();
    SetNewTransform(RefX,RefY,NewX,NewY);
    TCurrent(TFnew);
    TPop();

    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc) {
        if (SQDesc->ksPointer->oInfo != SQ_OLDSEL &&
            SQDesc->ksPointer->oInfo != SQ_NEWSEL) continue;
        /* prevent drawing duplicate entries twice */
        SQDesc->ksPointer->oInfo += 20;

        switch (SQDesc->ksPointer->oType) {

            case CDSYMBOLCALL:
                CDCall(SQDesc->ksPointer,&MasterName,&NumX,&DX,&NumY,&DY);
                if (IsCellInLib(MasterName)) {

                    if (OpenCell(MasterName,&MasterDesc)) break;

                    SetTransform(SQDesc->ksPointer);
                    SetNewTransform(RefX,RefY,NewX,NewY);
                    TPremultiply();

                    for (Int1 = NumY-1; Int1 >= 0; --Int1) {
                        for (Int2 = 0; Int2 < NumX; ++Int2 ){
                            TPush();
                            TIdentity();
                            TTranslate(Int2*DX,Int1*DY);
                            TPremultiply();
                            ShowTransformed(MasterDesc,View->kvCoarseWindow,
                                TFold); 
                            TPop();
                        }
                    }
                    TPop();
                    break;
                }

            case CDLABEL:
                if (SQDesc->ksPointer->oType == CDLABEL)
                    BBLabel(View->kvCoarseWindow,SQDesc->ksPointer,&BB);
                else
                    CDStatusInt = CDBB(Parameters.kpCellDesc,
                        SQDesc->ksPointer,&BB.kaLeft,
                        &BB.kaBottom,&BB.kaRight,&BB.kaTop);
                TLoadCurrent(TFnew);
                TPoint(&BB.kaLeft,&BB.kaBottom);
                TPoint(&BB.kaRight,&BB.kaTop);
                TLoadCurrent(TFold);
                ShowEmptyBox(HighlightingColor,&BB);
                break;

            case CDWIRE:
                CDWire(SQDesc->ksPointer,&Layer,&Width,&Path);
                TLoadCurrent(TFnew);
                CopyPathWithXForm(&Path);
                TLoadCurrent(TFold);
                ShowWire(HighlightingColor,Width,Path);
                free_path(Path);
                break;

            case CDPOLYGON:
                CDPolygon(SQDesc->ksPointer,&Layer,&Path);
                TLoadCurrent(TFnew);
                CopyPathWithXForm(&Path);
                TLoadCurrent(TFold);
                ShowPath(HighlightingColor,Path,True);
                free_path(Path);
                break;
        }
    }
    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc)
        if (SQDesc->ksPointer->oInfo > 20)
            SQDesc->ksPointer->oInfo -= 20;
}


void
CopyPathWithXForm(Path)

struct p **Path;
{
    struct p *OldPair,*NewPair;
    /*
     * Copy Path with transform and return pointer to new path
     */
    OldPair = *Path;
    if ((NewPair = alloc(p)) == NULL)
        MallocFailed();
    *Path = NewPair;
    NewPair->pX = OldPair->pX; 
    NewPair->pY = OldPair->pY; 
    TPoint(&NewPair->pX,&NewPair->pY);
    NewPair->pSucc = NULL;
    OldPair = OldPair->pSucc;
    while(OldPair != NULL) {
        if ((NewPair->pSucc = alloc(p)) == NULL)
            MallocFailed();
        NewPair = NewPair->pSucc;
        NewPair->pX = OldPair->pX; 
        NewPair->pY = OldPair->pY; 
        NewPair->pSucc = NULL;
        TPoint(&NewPair->pX,&NewPair->pY);
        OldPair = OldPair->pSucc;
    }
}


static void
restore_color()

/* repaint with drawing color */
{
    struct ks *SQDesc;
    struct p *Path;
    struct ka BB;
    struct s *MasterDesc;
    int  Layer;
    int NumX,NumY,Int1,Int2;
    long Width;
    long DX,DY;
    char *MasterName;

    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc) {
        if (SQDesc->ksPointer->oInfo != SQ_OLDSEL &&
            SQDesc->ksPointer->oInfo != SQ_NEWSEL) continue;

        switch (SQDesc->ksPointer->oType) {

            case CDSYMBOLCALL:
                CDCall(SQDesc->ksPointer,&MasterName,&NumX,&DX,&NumY,&DY);
                if (IsCellInLib(MasterName)) {

                    if (OpenCell(MasterName,&MasterDesc)) break;

                    SetTransform(SQDesc->ksPointer);
                    TPremultiply();

                    for (Int1 = NumY-1; Int1 >= 0; --Int1) {
                        for (Int2 = 0; Int2 < NumX; ++Int2 ){
                            TPush();
                            TIdentity();
                            TTranslate(Int2*DX,Int1*DY);
                            TPremultiply();
                            show_device(MasterDesc,View->kvCoarseWindow);
                            TPop();
                        }
                    }
                    TPop();
                    break;
                }

            case CDLABEL:
                if (SQDesc->ksPointer->oType == CDLABEL)
                    BBLabel(View->kvCoarseWindow,SQDesc->ksPointer,&BB);
                else
                    CDStatusInt = CDBB(Parameters.kpCellDesc,
                        SQDesc->ksPointer,&BB.kaLeft,
                        &BB.kaBottom,&BB.kaRight,&BB.kaTop);
                ShowEmptyBox(DrawingColor,&BB);
                break;

            case CDWIRE:
                CDWire(SQDesc->ksPointer,&Layer,&Width,&Path);
                ShowWire(DrawingColor,Width,Path);
                break;

            case CDPOLYGON:
                CDPolygon(SQDesc->ksPointer,&Layer,&Path);
                ShowPath(DrawingColor,Path,True);
                break;
        }
    }
}


static void
show_device(CellDesc,AOI)

struct s *CellDesc;
struct ka *AOI;
{
    struct g *GenDesc;
    struct o *Pointer;
    struct p *Path;
    int Layer;
    long Width;

    if (Not CDInitGen(CellDesc,1,AOI->kaLeft,AOI->kaBottom,
        AOI->kaRight,AOI->kaTop,&GenDesc)) MallocFailed();

    loop {
        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE) continue;

        switch (Pointer->oType) {

            case CDLABEL:
                /*
                BBLabel(View->kvCoarseWindow,Pointer,&BB);
                ShowEmptyBox(BackgroundColor,&BB);
                */
                break;

            case CDWIRE:
                CDWire(Pointer,&Layer,&Width,&Path);
                ShowWire(DrawingColor,Width,Path);
                break;

            case CDPOLYGON:
                CDPolygon(Pointer,&Layer,&Path);
                ShowPath(DrawingColor,Path,True);
                break;
        }
    }
}


void
ShowTransformed(CellDesc,AOI,TFold)

struct s *CellDesc;
struct ka *AOI;
long *TFold;
{
    struct g *GenDesc;
    struct o *Pointer;
    struct p *Path;
    struct ka BB;
    int Layer;
    long Width;
    long TFtmp[9];

    if (Not CDInitGen(CellDesc,1,AOI->kaLeft,AOI->kaBottom,
        AOI->kaRight,AOI->kaTop,&GenDesc)) MallocFailed();

    TCurrent(TFtmp);
    loop {
        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE) continue;

        switch (Pointer->oType) {

            case CDLABEL:
                BBLabel(View->kvCoarseWindow,Pointer,&BB);
                TPoint(&BB.kaLeft,&BB.kaBottom);
                TPoint(&BB.kaRight,&BB.kaTop);
                TLoadCurrent(TFold);
                ShowEmptyBox(HighlightingColor,&BB);
                TLoadCurrent(TFtmp);
                break;

            case CDWIRE:
                CDWire(Pointer,&Layer,&Width,&Path);
                CopyPathWithXForm(&Path);
                TLoadCurrent(TFold);
                ShowWire(HighlightingColor,Width,Path);
                TLoadCurrent(TFtmp);
                free_path(Path);
                break;

            case CDPOLYGON:
                CDPolygon(Pointer,&Layer,&Path);
                CopyPathWithXForm(&Path);
                TLoadCurrent(TFold);
                ShowPath(HighlightingColor,Path,True);
                TLoadCurrent(TFtmp);
                free_path(Path);
                break;
        }
    }
}


static void
free_path(Path)

struct p *Path;
{
    struct p *Next;

    for (; Path; Path = Next) {
        Next = Path->pSucc;
        free(Path);
    }
}


static void
show_move(OBB,NBB)

struct ka *OBB,*NBB;
{
    long L1,R1,B1,T1,L2,R2,B2,T2;
    struct ka BB;

    L1 = NBB->kaLeft;
    R1 = NBB->kaRight;
    B1 = NBB->kaBottom;
    T1 = NBB->kaTop;
    L2 = OBB->kaLeft;
    R2 = OBB->kaRight;
    B2 = OBB->kaBottom;
    T2 = OBB->kaTop;
    if (L1 > R1) SwapInts(L1,R1);
    if (B1 > T1) SwapInts(B1,T1);
    if (L2 > R2) SwapInts(L2,R2);
    if (B2 > T2) SwapInts(B2,T2);
    if (L1 > R2 Or L2 > R1 Or B1 > T2 Or B2 > T1) {
        /*
         * Old BB is not within the new BB
         * So, redisplay twice.
         */
        EraseBox(OBB);
        Redisplay(OBB);
        EraseBox(NBB);
        Redisplay(NBB);
        return;
    }
    /*
     * Old BB intersects new BB
     */
    BB.kaLeft   = Min(L1,L2);
    BB.kaRight  = Max(R1,R2);
    BB.kaBottom = Min(B1,B2);
    BB.kaTop    = Max(T1,T2);
    EraseBox(&BB);
    Redisplay(&BB);
}


static int
copy_OK(X,Y)

long X,Y;
{
    char *TypeIn;

    if (Not Parameters.kpMX And Not Parameters.kpMY And
        Parameters.kpRotationAngle == 0 And
        X == 0 And Y == 0 ) {

        ShowPrompt(
         "This will copy objects directly over themselves.  Continue(N)?");
        TypeIn = FBEdit(NULL);
        if (TypeIn == NULL) return (-1);
        if ((TypeIn[0] != 'Y' And TypeIn[0] != 'y'))
            return (0);
    }
    return (1);
}


static void
do_copy(RefX,RefY,NewX,NewY,NewBB,MoveOrCopy)

long RefX,RefY,NewX,NewY;
struct ka *NewBB;
int MoveOrCopy;
{
    struct ks *SQDesc;
    struct p *Path;
    struct ka BB;
    struct o *Pointer;
    struct prpty *PDesc;
    long X,Y,Length,Width;
    int  Layer;
    char *Label;
    char Xform;

    BB.kaLeft = BB.kaRight = NewX;
    BB.kaTop = BB.kaBottom = NewY;
    NewBB->kaLeft = NewBB->kaRight = NewX;
    NewBB->kaTop = NewBB->kaBottom = NewY;

    if (MoveOrCopy == COPY) {
        /* identify mutual inductors */
        PDesc = Parameters.kpCellDesc->sPrptyList;
        for (; PDesc; PDesc = PDesc->prpty_Succ) {
            if (PDesc->prpty_Value != P_MUT) continue;
            if (MutSelected(PDesc)) {
                CDCopyProperty(Parameters.kpCellDesc,(struct o*)NULL,
                    PDesc);
                Parameters.kpCellDesc->sPrptyList->prpty_Info = "new";
            }
        }
    }

    TPush();
    TIdentity();
    SetNewTransform(RefX,RefY,NewX,NewY);

    if (MoveOrCopy == COPY) {
        /* transform mutual inductors */
        PDesc = Parameters.kpCellDesc->sPrptyList;
        for (; PDesc; PDesc = PDesc->prpty_Succ) {
            if (PDesc->prpty_Value != P_MUT) break;
            if (PDesc->prpty_Info && !strcmp(PDesc->prpty_Info,"new")) {
                TPoint((long*)&PDesc->prpty_Data->p_mut.x1,
                    (long*)&PDesc->prpty_Data->p_mut.y1);
                TPoint((long*)&PDesc->prpty_Data->p_mut.x2,
                    (long*)&PDesc->prpty_Data->p_mut.y2);
                continue;
            }
            break;
        }
    }

    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc) {
        if (SQDesc->ksPointer->oInfo == SQ_GONE) continue;

        switch (SQDesc->ksPointer->oType) {
            case CDLABEL:

                CDLabel(SQDesc->ksPointer,&Layer,&Label,&X,&Y,&Xform);
                TPoint(&X,&Y);

                if (Not CDMakeLabel(Parameters.kpCellDesc,Layer,Label,
                    X,Y,fix_xform(Xform),&Pointer)) MallocFailed();
                break;

            case CDWIRE:
                CDWire(SQDesc->ksPointer,&Layer,&Width,&Path);
                CopyPathWithXForm(&Path);
                if (Not CDMakeWire(Parameters.kpCellDesc,Layer,Width,Path,
                    &Pointer)) MallocFailed();
                AssignWireProperties(Pointer);
                break;

            case CDPOLYGON:
                CDPolygon(SQDesc->ksPointer,&Layer,&Path);
                CopyPathWithXForm(&Path);
                if (Not CDMakePolygon(Parameters.kpCellDesc,Layer,Path,
                    &Pointer)) MallocFailed();
                break;

            case CDSYMBOLCALL:
                do_copy_call(RefX,RefY,NewX,NewY,SQDesc->ksPointer,&Pointer);
                break;

            default:
                continue;
        }

        if (SQDesc->ksPointer->oType == CDLABEL)
            BBLabel(View->kvCoarseWindow,Pointer,&BB);
        else
            CDStatusInt = CDBB(Parameters.kpCellDesc,Pointer,&BB.kaLeft,
                &BB.kaBottom,&BB.kaRight,&BB.kaTop);
        if (MoveOrCopy == MOVE)
            Pointer->oInfo = SQ_NEWSEL;
        else
            Pointer->oInfo = SQ_NEW;
        SQInsert(Pointer);

        if (BB.kaLeft   < NewBB->kaLeft)   NewBB->kaLeft   = BB.kaLeft;
        if (BB.kaRight  > NewBB->kaRight)  NewBB->kaRight  = BB.kaRight;
        if (BB.kaBottom < NewBB->kaBottom) NewBB->kaBottom = BB.kaBottom;
        if (BB.kaTop    > NewBB->kaTop)    NewBB->kaTop    = BB.kaTop;

        if (MoveOrCopy == MOVE) {
            TransformReferences(SQDesc->ksPointer);
            HYtransform(SQDesc->ksPointer,Pointer);
            SQDesc->ksPointer->oInfo = SQ_GONE;
        }
    }
    TPop();
}


static char
fix_xform(Xform)

char Xform;
{
    long TF[9];

    TPush();
    TIdentity();
    if (Parameters.kpMX) TMX();
    if (Parameters.kpMY) TMY();
    if (Parameters.kpRotationAngle != 0) {
        if (Parameters.kpRotationAngle == 180)   TRotate(-1L,0L);
        elif (Parameters.kpRotationAngle == 90)  TRotate(0L,1L);
        elif (Parameters.kpRotationAngle == 270) TRotate(0L,-1L);
    }
    if (Xform & 4) TMY();
    if (Xform & 8) TMX();
    Xform &= 3;
    if (Xform != 0) {
        if (Xform == 1)   TRotate(0L,1L);
        elif (Xform == 2) TRotate(-1L,0L);
        elif (Xform == 3) TRotate(0L,-1L);
    }
    TCurrent(TF);
    Xform = SetXform(TF);
    TPop();
    return (Xform);
}


static void
do_copy_call(RefX,RefY,NewX,NewY,OPointer,NPointer)

long RefX,RefY,NewX,NewY;
struct o *OPointer;
struct o **NPointer;
{
    char *SymbolName;
    int NumX,NumY;
    long DX,DY;
    long TF[9];
    struct o *Pointer;
    struct prpty *PDesc;

    CDCall(OPointer,&SymbolName,&NumX,&DX,&NumY,&DY);
    /* only possible error is CDMALLOCFAILED */
    if (Not CDBeginMakeCall(Parameters.kpCellDesc,SymbolName,
        NumX,DX,NumY,DY,&Pointer))
        MallocFailed();

    SetTransform(OPointer);
    SetNewTransform(RefX,RefY,NewX,NewY);
    TCurrent(TF);
    AddResultingTransform(Pointer,TF);
    if (Not CDEndMakeCall(Parameters.kpCellDesc,Pointer))
        MallocFailed();

    /* remove any default user properties */
    CDRemoveProperty(Parameters.kpCellDesc,Pointer,P_VALUE);
    CDRemoveProperty(Parameters.kpCellDesc,Pointer,P_MODEL);
    CDRemoveProperty(Parameters.kpCellDesc,Pointer,P_INITC);
    CDRemoveProperty(Parameters.kpCellDesc,Pointer,P_OTHER);

    /* add the user properties */
    for (PDesc = OPointer->oPrptyList; PDesc; PDesc = PDesc->prpty_Succ) {
        if (PDesc->prpty_Value == P_VALUE ||
            PDesc->prpty_Value == P_MODEL ||
            PDesc->prpty_Value == P_INITC ||
            PDesc->prpty_Value == P_OTHER) {
            CDCopyProperty(Parameters.kpCellDesc,Pointer,PDesc);
        }
    }

    *NPointer = Pointer;
    TPop();
}


void
SetNewTransform(RefX,RefY,NewX,NewY)

long RefX,RefY,NewX,NewY;
{
    TTranslate(-RefX,-RefY);
    if (Parameters.kpMX) TMX();
    if (Parameters.kpMY) TMY();
    if (Parameters.kpRotationAngle != 0) {
        if (Parameters.kpRotationAngle == 90)    TRotate(0L,1L);
        elif (Parameters.kpRotationAngle == 180) TRotate(-1L,0L);
        elif (Parameters.kpRotationAngle == 270) TRotate(0L,-1L);
    }
    TTranslate(NewX,NewY);
}


void
AddResultingTransform(Pointer,TF)

struct o *Pointer;
long *TF;
{
    long A,B,C,D,TX,TY;
    int ret;
    /*
     * Take the transformation defined in TF and add
     * it to the symbol call currently being created.
     *
     *                  | a    c    0  |
     * Transform = TM = | b    d    0  |
     *                  | TX   TY   1  |
     *
     * A = TM[0][0] = TF[0];
     * B = TM[1][0] = TF[3];
     * C = TM[0][1] = TF[1];
     * D = TM[1][1] = TF[4];
     * TX = TM[2][0] = TF[6];
     * TY = TM[2][1] = TF[7];
     */
    A = TF[0]; B = TF[3]; C = TF[1]; D = TF[4];
    TX = TF[6]; TY = TF[7];
    if (A == 0 And B == 1 And C == 1 And D == 0) {
        /* MX R 0 -1 T tx,ty */
        ret = CDT(Pointer,CDMIRRORX,0L,0L);
        ret &= CDT(Pointer,CDROTATE,0L,-1L);
        ret &= CDT(Pointer,CDTRANSLATE,TX,TY);
    }
    elif (A == 0 And B == -1 And C == -1 And D == 0) {
        /* MX R 0 1 T tx,ty */
        ret = CDT(Pointer,CDMIRRORX,0L,0L);
        ret &= CDT(Pointer,CDROTATE,0L,1L);
        ret &= CDT(Pointer,CDTRANSLATE,TX,TY);
    }
    elif (A == 0 And B == 1 And C == -1 And D == 0) {
        /* R 0 -1 T tx,ty */
        ret = CDT(Pointer,CDROTATE,0L,-1L);
        ret &= CDT(Pointer,CDTRANSLATE,TX,TY);
    }
    elif (A == 0 And B == -1 And C == 1 And D == 0) {
        /* R 0 1 T tx,ty */
        ret = CDT(Pointer,CDROTATE,0L,1L);
        ret &= CDT(Pointer,CDTRANSLATE,TX,TY);
    }
    elif (A == 1 And B == 0 And C == 0 And D == 1) {
        /* T tx,ty */
        ret = CDT(Pointer,CDTRANSLATE,TX,TY);
    }
    elif (A == -1 And B == 0 And C == 0 And D == -1) {
        /* R -1 0 T tx,ty */
        ret = CDT(Pointer,CDROTATE,-1L,0L);
        ret &= CDT(Pointer,CDTRANSLATE,TX,TY);
    }
    elif (A == -1 And B == 0 And C == 0 And D == 1) {
        /* MX T tx,ty */
        ret = CDT(Pointer,CDMIRRORX,0L,0L);
        ret &= CDT(Pointer,CDTRANSLATE,TX,TY);
    }
    else{
        /* MY T tx,ty */
        ret = CDT(Pointer,CDMIRRORY,0L,0L);
        ret &= CDT(Pointer,CDTRANSLATE,TX,TY);
    }
    if (Not ret)
        MallocFailed();
}

