/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * SCED break selection operator.
 */

#include "spice.h"
#include "sced.h"

extern char *MenuBREAK;
extern char *MenuUNDO;
#ifdef __STDC__
static int  do_break(long,long,int);
static int  break_wire(long,long,struct o*,int);
static struct p *intersect_y(struct p*,struct p*,long);
static struct p *intersect_x(struct p*,struct p*,long);
#else
static int  do_break();
static int  break_wire();
static struct p *intersect_y();
static struct p *intersect_x();
#endif


void
Break(LookedAhead)

int *LookedAhead;
{
    int GotOne = False;
    int Undo = False;
    int FirstTime = True;
    int Modified = 0;
    struct ka BB, OldSelectQBB;
    long OldX, OldY;
    char Types[4];

    MenuSelect(MenuBREAK);

    Types[0] = CDWIRE;
    Types[1] = '\0';

    if (AreTypesInQ(Types))
        GotOne = True;
top:

    if (Not GotOne) {
        ShowPrompt("Point to wire to break.");
        switch (PointLoop(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                goto quit;
            case PL_UND:
                if (FirstTime) goto quit;
                MenuSelect(MenuUNDO);
                SQRestore(True);
                EraseBox(&OldSelectQBB);
                Redisplay(&OldSelectQBB);
                Modified--;
                Undo = True;
                MenuDeselect(MenuUNDO);
                break;
        case PL_PCW:
                SelectTypes(Types);
                if (Not AreTypesInQ(Types))
                    goto top;;
        }
    }

    ShowPrompt("Point on the break line.");

next:
    switch (PointLoop(LookedAhead)) {
        case PL_ESC:
        case PL_CMD:
            goto quit;
        case PL_UND:
            MenuSelect(MenuUNDO);
            if (Not GotOne) {
                if (Undo == False) {
                    SQDesel(Types);
                    EraseBox(&OldSelectQBB);
                    Redisplay(&OldSelectQBB);
                    MenuDeselect(MenuUNDO);
                }
                else {
                    (void) do_break(OldX,OldY,True);
                    EraseBox(&OldSelectQBB);
                    Redisplay(&OldSelectQBB);
                    Modified++;
                    Undo = False;
                }
                goto top;
            }
            if (FirstTime)
                goto quit;
            if (Undo == False) {
                MenuSelect(MenuBREAK);
                ShowPrompt("Point on the break line.");
                SQRestore(True);
                EraseBox(&OldSelectQBB);
                Redisplay(&OldSelectQBB);
                Modified--;
                Undo = True;
            }
            else {
                (void) do_break(OldX,OldY,True);
                EraseBox(&OldSelectQBB);
                Redisplay(&OldSelectQBB);
                ErasePrompt();
                MenuDeselect(MenuBREAK);
                Modified++;
                Undo = False;
            }
            MenuDeselect(MenuUNDO);
            goto next;
        case PL_PCW:
            if (Not FirstTime And GotOne And Not Undo) goto next;
            if (do_break(SCursor.kcX,SCursor.kcY,False)) {
                SQRestore(False);
                do_break(SCursor.kcX,SCursor.kcY,True);
                OldX = SCursor.kcX;
                OldY = SCursor.kcY;
                OldSelectQBB = SelectQBB;
                EraseBox(&OldSelectQBB);
                Redisplay(&OldSelectQBB);
                FirstTime = False;
                Modified++;
                Undo = False;
                if (Not GotOne)
                    goto top;
                ErasePrompt();
                MenuDeselect(MenuBREAK);
                goto next;
            }
    }

quit:
    SQRestore(False);
    if (Not GotOne And AreTypesInQ(Types)) {
        SQComputeBB();
        SQDesel(Types);
        EraseBox(&SelectQBB);
        Redisplay(&SelectQBB);
    }
    if (Modified)
        Parameters.kpModified = True;
    ErasePrompt();
    MenuDeselect(MenuUNDO);
    MenuDeselect(MenuBREAK);
}


struct p *
CopyPath(path)

struct p *path;
{
    struct p *pcopy,*pc = NULL,*pp = path;

    if (path == NULL) return (NULL);
    pcopy = pc = alloc(p);
    pc->pX = pp->pX;
    pc->pY = pp->pY;
    pp = pp->pSucc;
    while (pp) {
        pc->pSucc = alloc(p);
        pc = pc->pSucc;
        pc->pX = pp->pX;
        pc->pY = pp->pY;
        pp = pp->pSucc;
    }
    pc->pSucc = NULL;
    return (pcopy);
}


static int
do_break(RefX,RefY,Flag)

long RefX,RefY;
int Flag;
{
    struct ks *SQDesc;
    int DidBreak = False;

    /* If Flag is False, returns True if a break would be performed,
     * but operation is not done.
     */
    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc) {
        if (SQDesc->ksPointer->oInfo == SQ_GONE) continue;
        if (SQDesc->ksPointer->oType == CDWIRE) {
            DidBreak = break_wire(RefX,RefY,SQDesc->ksPointer,Flag);
            continue;
        }
    }
    return (DidBreak);
}


static int
break_wire(RefX,RefY,Pointer,Flag)

long RefX,RefY;
struct o *Pointer;
int Flag;
{
    struct p *pp,*px,*pTemp,*pNew,*pInt;
    struct o *NewPointer;
    struct ka BB;
    int VertCut;
    long *InPath();

    if (InPath(20,((struct w *)Pointer->oRep)->wPath,RefX,RefY) != NULL) {
        if (Not Flag) return (True);
        px = CopyPath(((struct w *)Pointer->oRep)->wPath);
        for (pNew = pp = px; pp And pp->pSucc; pp = pTemp) {
            pTemp = pp->pSucc;
            if (
                ( pp->pX == pp->pSucc->pX && RefX == pp->pX And
                ((pp->pY <= RefY And pp->pSucc->pY > RefY) Or
                (pp->pY >= RefY And pp->pSucc->pY < RefY)) )
                Or
                ( pp->pY == pp->pSucc->pY && RefY == pp->pY And
                ((pp->pX <= RefX And pp->pSucc->pX > RefX) Or
                (pp->pX >= RefX And pp->pSucc->pX < RefX)) )) {

                pp->pSucc = NULL;
                pInt = alloc(p);
                if (pInt == NULL) MallocFailed();
                pInt->pX = RefX;
                pInt->pY = RefY;

                if (pp->pX != pInt->pX Or pp->pY != pInt->pY) {
                    pp->pSucc = pInt;
                    pInt->pSucc = NULL;
                    pInt = CopyPath(pInt);
                }
                if (Not CDMakeWire(Parameters.kpCellDesc,Pointer->oLayer,
                    ((struct w *)Pointer->oRep)->wWidth,pNew,&NewPointer))
                    MallocFailed();
                AssignWireProperties(NewPointer);
                NewPointer->oInfo = SQ_NEW;
                SQInsert(NewPointer);
                pInt->pSucc = pTemp;
                pNew = pInt;
            }
        }
    }
    else {

        CDStatusInt = CDBB(Parameters.kpCellDesc,Pointer,&BB.kaLeft,
            &BB.kaBottom,&BB.kaRight,&BB.kaTop);

        VertCut = False;
        if (Parameters.kpRotationAngle == 0 ||
            Parameters.kpRotationAngle == 180) VertCut = True;

        if (VertCut && (RefX <= BB.kaLeft || RefX >= BB.kaRight))
            return (False);
        if (!VertCut && (RefY <= BB.kaBottom Or RefY >= BB.kaTop))
            return (False);

        if (Not Flag) return (True);
        px = CopyPath(((struct w *)Pointer->oRep)->wPath);
        for (pNew = pp = px; pp And pp->pSucc; pp = pTemp) {
            pTemp = pp->pSucc;
            if (
                ( VertCut And
                ((pp->pX <= RefX And pp->pSucc->pX > RefX) Or
                   (pp->pX >= RefX And pp->pSucc->pX < RefX)) )
                Or
                ( !VertCut And
                ((pp->pY <= RefY And pp->pSucc->pY > RefY) Or
                   (pp->pY >= RefY And pp->pSucc->pY < RefY)) )) {

                pp->pSucc = NULL;

                if (VertCut)
                   pInt = intersect_y(pp,pTemp,RefX);
                else
                    pInt = intersect_x(pp,pTemp,RefY);

                if (pp->pX != pInt->pX Or pp->pY != pInt->pY) {
                    pp->pSucc = pInt;
                    pInt->pSucc = NULL;
                    pInt = CopyPath(pInt);
                }
                if (Not CDMakeWire(Parameters.kpCellDesc,Pointer->oLayer,
                    ((struct w *)Pointer->oRep)->wWidth,
                    pNew,&NewPointer)) MallocFailed();
                AssignWireProperties(NewPointer);
                NewPointer->oInfo = SQ_NEW;
                SQInsert(NewPointer);
                pInt->pSucc = pTemp;
                pNew = pInt;
            }
        }
    }
    if (Not CDMakeWire(Parameters.kpCellDesc,Pointer->oLayer,
        ((struct  w *)Pointer->oRep)->wWidth,pNew,&NewPointer))
        MallocFailed();
    AssignWireProperties(NewPointer);
    NewPointer->oInfo = SQ_NEW;
    SQInsert(NewPointer);
    Pointer->oInfo = SQ_GONE;
    return (True);
}


static struct p *
intersect_y(p1,p2,X)

struct p *p1,*p2;
long X;
{
    struct p *pp;

    pp = alloc(p);
    pp->pX = X;
    pp->pY = p1->pY + ((X - p1->pX)*(p2->pY - p1->pY))/(p2->pX - p1->pX);
    return (pp);
}


static struct p *
intersect_x(p1,p2,Y)

struct p *p1,*p2;
long Y;
{
    struct p *pp;

    pp = alloc(p);
    pp->pX = p1->pX + ((Y - p1->pY)*(p2->pX - p1->pX))/(p2->pY - p1->pY);
    pp->pY = Y;
    return (pp);
}

