/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"
#include <time.h>

#include <sys/types.h>
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#ifndef direct
#define direct dirent
#endif
#else
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#endif


/***********************************************************************
 *
 * Dir command.
 * Show a directory listing of symbol files.
 *
 ***********************************************************************/


char *
CopyString(s)

char *s;
{
    char *t;

    if (s) {
        t = tmalloc(strlen(s)+1);
        if (t == NULL) MallocFailed();
        strcpy(t,s);
        return (t);
    }
    return (s);
}

extern char *MenuDIR;

struct list {
    char *l_word;
    struct list *l_next;
};

#ifdef __STDC__
static void format_lines(struct list*);
static void display_lines(struct list*);
static void list_free(struct list*);
static void list_sort(struct list*);
static int  lcomp(char**,char**);
static struct list *symfiles(char*);
#else
static void format_lines();
static void display_lines();
static void list_free();
static void list_sort();
static int  lcomp();
static struct list *symfiles();
#endif


void
Dir()

{
    char *path;
    char *c,*s;
    struct list *wl,*wx,*wx0 = NULL;
    char buf[BSIZE_SP];


    MenuSelect(MenuDIR);
    path = PGetPath();

    for (s = path;;) {

        while (isspace(*s)) s++;
        if (*s == '\0')
            break;

        c = buf;
        while (!isspace(*s) && *s != '\0')
            *c++ = *s++;
        *c = '\0';
        /* buf contains the directory path */

        wl = symfiles(buf);

        list_sort(wl);
        format_lines(wl);
        if (wx0 == NULL)
            wx = wx0 = alloc(list);
        else {
            wx->l_next = alloc(list);
            wx = wx->l_next;
        }
        if (wx == NULL) MallocFailed();
        strcat(buf," :");
        if (!wl)
            strcat(buf, "\n  no symbol files found\n ");
        wx->l_word = CopyString(buf);
        wx->l_next = wl;
        while (wx->l_next) wx = wx->l_next;
    }
    display_lines(wx0);
    list_free(wx0);
    if (!dispdev->windows)
        RedisplaySCED(currentgraph);
    MenuDeselect(MenuDIR);
}


static void
format_lines(li)

/* combine the words into columns */
struct list *li;
{
    struct list *wl;
    int width, colw, nfiles = 0;
    int i, j, k, fpc;
    char *buf, *b, **ptr;

    if (li == NULL) return;
    if (!dispdev->windows) {
        width = View->kvLargeCoarseViewport->kaWidth - 8;
        width /= gi_fntwidth;
    }
    else
        width = 70;
    buf = tmalloc(width+1);

    colw = 0;
    for (wl = li; wl; wl = wl->l_next) {
        i = strlen(wl->l_word);
        if (i > colw) colw = i;
        nfiles++;
    }
    colw += 2;
    width /= colw;  /* width = number of columns */
    fpc = nfiles/width + 1;  /* fpc = max number files/column */
    ptr = (char**)tmalloc(nfiles*sizeof(char*));
    for (i = 0, wl = li; wl; i++, wl = wl->l_next) {
        *(ptr+i) = wl->l_word;
    }

    wl = li;
    for (i = 0; i < fpc; i++) {
        b = buf;
        for (j = 0; j < width; j++) {
            k = j*fpc + i;
            if (k >= nfiles)
                break;
            sprintf(b,"%-*s", colw, *(ptr+k));
            b = buf + strlen(buf);
        }
        free(wl->l_word);
        wl->l_word = CopyString(buf);
        if (!wl->l_next) {
            wl->l_next = alloc(list);
            if (!wl->l_next)
                MallocFailed();
            wl->l_next->l_next = NULL;
        }
        wl = wl->l_next;
    }
    wl->l_word = CopyString(" ");
    if (wl->l_next)
        list_free(wl->l_next);
    wl->l_next = NULL;

    free((char*)ptr);
    free(buf);
}


static void
display_lines(wl)

/* print the list, using MORE mode */
struct list *wl;
{
    (void)EnableMore(True);
    for (; wl; wl = wl->l_next)
        if (MoreLine(wl->l_word)) break;

    (void)EnableMore(False);
    if (!dispdev->windows) {
        ShowPrompt("Hit any key to continue.");
        (void)FBGetchar(ERASE);
        ErasePrompt();
    }
}


static void
list_free(wl)

struct list *wl;
{
    struct list *nw;

    for (; wl; wl = nw) {
        nw = wl->l_next;
        free(wl->l_word);
        tfree(wl);
    }
    return;
}


static void
list_sort(wl)

/* alphabetically sort the list */
struct list *wl;
{
    int i = 0;
    struct list *ww;
    char **stuff;

    for (i = 0, ww = wl; ww; i++, ww = ww->l_next) ;
    if (i < 2)
        return;
    stuff = (char **) tmalloc(i * sizeof (char *));
    if (stuff == NULL) MallocFailed();
    for (i = 0, ww = wl; ww; i++, ww = ww->l_next)
        stuff[i] = ww->l_word;
    qsort((char *) stuff, i, sizeof (char *),
#if __STDC__
        (int(*)(const void*,const void*))lcomp);
#else
        lcomp);
#endif
    for (i = 0, ww = wl; ww; i++, ww = ww->l_next)
        ww->l_word = stuff[i];
    free(stuff);
    return;
}


static int
lcomp(s, t)

char **s, **t;
{
    return (strcmp(*s,*t));
}


static struct list *
symfiles(dir)

char *dir;
{
    DIR *wdir;
    struct direct *de;
    struct list *wl, *wl0 = NULL;
    char buf[BSIZE_SP];
    FILE *fp;
    int i;

    if (!(wdir = opendir(dir)))
        return (NULL);

    while ((de = readdir(wdir)) != NULL) {

        sprintf(buf,"%s%c%s",dir,DIR_TERM,de->d_name);
        fp = fopen(buf,"r");
        if (fp == NULL)
            continue;

        /* is it a symbol file? */

        for (i = 0; i <7; i++)
            buf[i] = getc(fp);
        buf[i] = '\0';
        fclose(fp);

        if (strcmp(buf,"(Symbol"))
            continue;

        if (wl0 == NULL) {
            wl = wl0 = alloc(list);
        }
        else {
            wl->l_next = alloc(list);
            wl = wl->l_next;
        }
        wl->l_word = CopyString(de->d_name);
        wl->l_next = NULL;
    }
    closedir(wdir);
    return (wl0);
}
