/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * SCED display control code.
 */

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"



/***********************************************************************
 *
 * Window manipulation menu commands.
 *
 ***********************************************************************/

extern char *MenuPAN;
extern char *MenuZOOM;
extern char *MenuWINDO;
extern char *MenuLAST;

void
Pan(LookedAhead)

int *LookedAhead;
{
    MenuSelect(MenuPAN);
    ShowPrompt("Point to center of new window.");
    loop {
        switch (PointLoopSafe(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                MenuDeselect(MenuPAN);
                ErasePrompt();
                return;
            case PL_PCW:
                SaveLastView();
                InitCoarseWindow(SCursor.kcX,SCursor.kcY,
                    (long)View->kvCoarseWindow->kaWidth);
                InitFineWindow(SCursor.kcX, SCursor.kcY);
                EraseLargeCoarseViewport();
                Redisplay(View->kvCoarseWindow);
        }
    }
}


void
Zoom(LookedAhead)

int *LookedAhead;
{
    char *TypeIn;
    long NewWindowWidth;
    long X,Y;

    MenuSelect(MenuZOOM);
    sprintf(TypeOut,
        "Enter window width (now %g) or RETURN for %s display: ",
        View->kvCoarseWindow->kaWidth/RESOLUTION,
        View->kvControl == SPLITSCREEN ? "full screen" : "split screen");
    ShowPrompt(TypeOut);
    TypeIn = FBEdit(NULL);
    if (TypeIn == NULL) { /* esc entered */
        MenuDeselect(MenuZOOM);
        ErasePrompt();
        return;
    }
    SaveLastView();
    if ((sscanf(TypeIn,"%ld",&NewWindowWidth) < 1) Or NewWindowWidth <= 0) {
        if (View->kvControl != FULLSCREEN) {
            View->kvControl = FULLSCREEN;
            Parameters.kpRedisplayControl = COARSEVIEWPORTONLY;
            NewWindowWidth = View->kvFineWindow->kaWidth;
            View->kvCoarseWindow->kaX = View->kvFineWindow->kaX;
            View->kvCoarseWindow->kaY = View->kvFineWindow->kaY;
            InitCoarseWindow(View->kvCoarseWindow->kaX,
                View->kvCoarseWindow->kaY,NewWindowWidth);
            SetPositioning();
        }
        else {
            View->kvControl = SPLITSCREEN;
            Parameters.kpRedisplayControl = SPLITSCREEN;
            X = View->kvCoarseWindow->kaX;
            Y = View->kvCoarseWindow->kaY;
            NewWindowWidth = View->kvCoarseWindow->kaWidth;
            SetPositioning();
            CenterFullView();
            if (NewWindowWidth > View->kvCoarseWindow->kaWidth)
                NewWindowWidth = 3*View->kvCoarseWindow->kaWidth/4;
            View->kvFineWindow->kaWidth = NewWindowWidth;
            InitFineWindow(X,Y);
        }
    }
    else {
        NewWindowWidth *= RESOLUTION;
        if (View->kvControl == FULLSCREEN)
            InitCoarseWindow(View->kvCoarseWindow->kaX,
                View->kvCoarseWindow->kaY,NewWindowWidth);
        else
            /* zoom from fine window position */
            InitCoarseWindow(View->kvFineWindow->kaX,
                View->kvFineWindow->kaY,NewWindowWidth);
        SetPositioning();
    }
    EraseLargeCoarseViewport();
    if (View->kvControl == FULLSCREEN)
        Redisplay(View->kvCoarseWindow);
    else {
        /* make sure fine window is shown, if out of coarse clip range */
        Parameters.kpRedisplayControl = COARSEVIEWPORTONLY;
        Redisplay(View->kvCoarseWindow);
        ShowEmptyBox(HighlightingColor,View->kvFineWindow);
        Parameters.kpRedisplayControl = FINEVIEWPORTONLY;
        Redisplay(View->kvFineWindow);
        Parameters.kpRedisplayControl = SPLITSCREEN;
    }
    MenuDeselect(MenuZOOM);
    *LookedAhead = False;
    ErasePrompt();
    return;
}


void
Windo(LookedAhead)

int *LookedAhead;
{
    long NewWindowWidth, Width, Height, Tmp;
    MenuSelect(MenuWINDO);
    loop{
        ShowPrompt("Point to endpoints of diagonal.");
        switch (PointLoopSafe(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                goto quit;
            case PL_PCW:
                break;
        }
        FBSetRubberBanding('R');
        switch (PointLoopSafe(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                FBSetRubberBanding(0);
                goto quit;
            case PL_PCW:
                FBSetRubberBanding(0);
                break;
        }
        SaveLastView();
        /*
         * Coarse window height is smaller than its width.
         * When user points to endpoints of diagonal to define new window,
         * he expects to see everything in this window after redisplay.
         * To make this happen, we have to EXPAND the width enough so that
         * the height that InitCoarseWindow computes is equal to the height
         * the user wants.
         * The following lines compute the multiplier that expands the
         * width by the right amount.
         */

        Width = SCursor.kcX - SCursor.kcPredX;
        Height = SCursor.kcY - SCursor.kcPredY;
        if (Width < 0) Width = -Width;
        if (Height < 0) Height = -Height;

        Tmp = Height * View->kvCoarseViewport->kaWidth/
            View->kvCoarseViewport->kaHeight;
        NewWindowWidth = Max(Width,Tmp);

        Height /= 2;
        Width /= 2;

        if (NewWindowWidth <= 0)    NewWindowWidth = RESOLUTION;
        InitCoarseWindow( Min(SCursor.kcX,SCursor.kcPredX)+Width,
            Min(SCursor.kcY,SCursor.kcPredY)+Height,NewWindowWidth);

        InitFineWindow(Min(SCursor.kcX,SCursor.kcPredX)+Width,
            Min(SCursor.kcY,SCursor.kcPredY)+Height);

        SetPositioning();
        EraseLargeCoarseViewport();
        Redisplay(View->kvCoarseWindow);
    }
quit:
    MenuDeselect(MenuWINDO);
    ErasePrompt();
}


void
LastView()

{
    int  i;
    struct kw Last;
    struct kw *Tmp;
    char OldMenu;
    int  EscWasReturned = 0;

    Last.kwLastWindowX           = View->kvCoarseWindow->kaX;
    Last.kwLastWindowY           = View->kvCoarseWindow->kaY;
    Last.kwLastFineWindowX       = View->kvFineWindow->kaX;
    Last.kwLastFineWindowY       = View->kvFineWindow->kaY;
    Last.kwLastWindowWidth       = View->kvCoarseWindow->kaWidth;
    Last.kwLastFineWindowWidth   = View->kvFineWindow->kaWidth;
    (Last.kwName)[5]             = View->kvFineViewportOnBottom;
    (Last.kwName)[6]             = View->kvControl;

    i = 0;
    for (Tmp = Parameters.kpWindowStack; Tmp; Tmp = Tmp->kwNext)
        AmbiguityMenu[i++].mEntry = Tmp->kwName;
    AmbiguityMenu[i].mEntry = NULL;
    FixMenuPrefix(AmbiguityMenu);

    if (i > 1) {
        /* Show ambiguity menu. */
        OldMenu = Parameters.kpMenu;
        Parameters.kpMenu = AMBIGUITYMENU;
        ShowMenu(AmbiguityMenu);

        /* Which viewport is user interested in? */
        ShowPrompt("Point to the name of the desired view.");
        loop {
            int dummy;
            switch (PointLoopSafe(&dummy)) {
                case PL_ESC:
                    EscWasReturned = 1;
                    break;
                case PL_CMD:
                    break;
                case PL_PCW:
                    ShowPrompt("You aren't pointing at the menu.");
                    continue;
            }
            break;
        }
        Parameters.kpMenu = OldMenu;
        ShowCommandMenu();
        if (EscWasReturned) goto quit;

        /* find new view */
        for (Tmp = Parameters.kpWindowStack; Tmp; Tmp = Tmp->kwNext)
            if (strncmp(Tmp->kwName,Parameters.kpCommand,
                MenuViewport.kaRight-MenuViewport.kaLeft+1) == 0){
                break;
            }
    }
    else
        Tmp = Parameters.kpWindowStack;

    /* The status of the fine window positioning is stored in
     * the name field, as is full screen mode flag.
     */
    View->kvFineViewportOnBottom = (Tmp->kwName)[5];
    View->kvControl = (Tmp->kwName)[6];

    InitViewport(); 

    /* change to new viewport */
    InitCoarseWindow(Tmp->kwLastWindowX,Tmp->kwLastWindowY,
        Tmp->kwLastWindowWidth);

    SetPositioning();

    InitFineWindow(Tmp->kwLastFineWindowX,Tmp->kwLastFineWindowY);

    Parameters.kpWindowStack->kwLastWindowX         = Last.kwLastWindowX;        
    Parameters.kpWindowStack->kwLastWindowY         = Last.kwLastWindowY;        
    Parameters.kpWindowStack->kwLastFineWindowX     = Last.kwLastFineWindowX;    
    Parameters.kpWindowStack->kwLastFineWindowY     = Last.kwLastFineWindowY;    
    Parameters.kpWindowStack->kwLastWindowWidth     = Last.kwLastWindowWidth;    
    Parameters.kpWindowStack->kwLastFineWindowWidth = Last.kwLastFineWindowWidth;
    Parameters.kpWindowStack->kwExpand              = Last.kwExpand;
    Parameters.kpWindowStack->kwExpandFineOnly      = Last.kwExpandFineOnly;
    (Parameters.kpWindowStack->kwName)[5]           = (Last.kwName)[5];          
    (Parameters.kpWindowStack->kwName)[6]           = (Last.kwName)[6];          

    EraseLargeCoarseViewport();
    Redisplay(View->kvCoarseWindow);
quit:
    MenuDeselect(MenuLAST);
    ErasePrompt();
}


void
SaveLastView()

{
    /*
     * Parameters.kpWindowStack is always the last view.
     */
    if (Parameters.kpWindowStack == NULL) {
        if ((Parameters.kpWindowStack = alloc(kw)) == NULL)
            MallocFailed();
        strcpy(Parameters.kpWindowStack->kwName,"prev");
        Parameters.kpWindowStack->kwNext = NULL;
    }
    Parameters.kpWindowStack->kwLastWindowX     = View->kvCoarseWindow->kaX;
    Parameters.kpWindowStack->kwLastWindowY     = View->kvCoarseWindow->kaY;
    Parameters.kpWindowStack->kwLastFineWindowX = View->kvFineWindow->kaX;
    Parameters.kpWindowStack->kwLastFineWindowY = View->kvFineWindow->kaY;
    Parameters.kpWindowStack->kwLastWindowWidth = 
        View->kvCoarseWindow->kaWidth;
    Parameters.kpWindowStack->kwLastFineWindowWidth =
        View->kvFineWindow->kaWidth;
    (Parameters.kpWindowStack->kwName)[5] = View->kvFineViewportOnBottom;
    (Parameters.kpWindowStack->kwName)[6] = View->kvControl;
}


void
SaveViewOnStack()

{
    int i = 0;
    char c, buf[32];
    struct kw *New, *Tmp;

    New = alloc(kw);
    if (New == NULL) MallocFailed();

    New->kwLastWindowX         = View->kvCoarseWindow->kaX;
    New->kwLastWindowY         = View->kvCoarseWindow->kaY;
    New->kwLastFineWindowX     = View->kvFineWindow->kaX;
    New->kwLastFineWindowY     = View->kvFineWindow->kaY;
    New->kwLastWindowWidth     = View->kvCoarseWindow->kaWidth;
    New->kwLastFineWindowWidth = View->kvFineWindow->kaWidth;
    (New->kwName)[5]           = View->kvFineViewportOnBottom;
    (New->kwName)[6]           = View->kvControl;

    if (Parameters.kpWindowStack == NULL)
        SaveLastView();

    New->kwNext = NULL;
    for (Tmp = Parameters.kpWindowStack; Tmp->kwNext; i++,Tmp = Tmp->kwNext) ;
    Tmp->kwNext = New;
    sprintf(buf,"Current view assigned to: %c",'A'+i);
    ShowPrompt(buf);
    New->kwName[0] = 'a'+ i;
    New->kwName[1] = '\0';
}



/***********************************************************************
 *
 * Redisplay control code.
 *
 ***********************************************************************/

static int DisplayDepth, HierarchyLevel;

#ifdef __STDC__
static void redisplay_layer(struct s*,struct ka*);
#else
static void redisplay_layer();
#endif

#define ka_copy(BB1,BB2) BB1.kaLeft = BB2->kaLeft; \
    BB1.kaRight = BB2->kaRight; \
    BB1.kaBottom = BB2->kaBottom; \
    BB1.kaTop = BB2->kaTop; \
    TPoint(&BB1.kaLeft,&BB1.kaBottom); \
    TPoint(&BB1.kaRight,&BB1.kaTop); \
    if(BB.kaLeft > BB.kaRight) \
        SwapInts(BB1.kaLeft,BB1.kaRight); \
    if(BB.kaBottom > BB.kaTop) \
        SwapInts(BB1.kaBottom,BB1.kaTop);


void
Redisplay(AOI)

struct ka *AOI;
{
    struct s *CellDesc = Parameters.kpTopDesc;
   
    SetCurrentAOI(AOI);

    if (CurrentAOI.aInCoarse == 0 And CurrentAOI.aInFine == 0)
        goto quit;

    Parameters.kpNumGeometries = 0;

    if (!Parameters.kpGridOnTop)
        ShowGrid();

    if (Parameters.kpCellDesc != NULL) {

        TPush();
        TLoad();

        DisplayDepth = 0;
        HierarchyLevel = 0;
        if (CellDesc)
            redisplay_layer(CellDesc,AOI);

        TPop();
    }

    if (Parameters.kpGridOnTop)
        ShowGrid();

    /*
     * Show fine positioning window in coarse window.
     */
    DevSetColor(HighlightingColor);
    if (Parameters.kpRedisplayControl != COARSEVIEWPORTONLY) {

        if (Parameters.kpRedisplayControl == SPLITSCREEN)
            ShowEmptyBox(HighlightingColor,View->kvFineWindow);
        OutlineBox(View->kvFineViewport);
    }
    else if (!Parameters.kpDoingHardcopy &&
            View->kvControl == FULLSCREEN)
        OutlineBox(View->kvLargeCoarseViewport);

    /*
     * Highlight selected objects.
     */
    SQShow();

    /* SRW ** show markers */
    ShowMarker(DISPLAY,HighlightingColor,0L,0L,0,0,0);

quit:
    SetCurrentAOI(View->kvCoarseWindow);
}


static void
redisplay_layer(CellDesc,AOI)

struct s *CellDesc;
struct ka *AOI;
{
    struct g *GenDesc;
    struct o *Pointer;
    struct p *Path;
    struct ka BB;
    struct ka MasterBB;
    struct s *MasterDesc;
    char Xform;
    char *MasterName;
    char *TypeIn;
    char *Label;
    int NumX,NumY;
    int Int1,Int2;
    int Layer;
    int Expand;
    long X,Y;
    long Width;
    long DX,DY;

    if (Not CDInitGen(CellDesc,1,AOI->kaLeft,AOI->kaBottom,
        AOI->kaRight,AOI->kaTop,&GenDesc)) MallocFailed();
    loop {
        /* Test for user interrupt */
        if (Parameters.kpSIGINTERRUPT) return;

        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE) continue;

        ++Parameters.kpNumGeometries;

        if (Pointer->oType == CDWIRE) {
            CDWire(Pointer,&Layer,&Width,&Path);
            if (Pointer->oInfo == SQ_INCMPLT)
                ShowPath(DrawingColor,Path,False);
            else
                ShowWire(DrawingColor,Width,Path);
            continue;
        }
        if (Pointer->oType == CDPOLYGON) {
            if (Parameters.kpShowDots) {
                CDPolygon(Pointer,&Layer,&Path);
                if (Pointer->oInfo == SQ_INCMPLT)
                    ShowPath(DrawingColor,Path,False);
                else
                    ShowPolygon(DrawingColor,Path);
            }
            continue;
        }
        if (Pointer->oType == CDLABEL) {
            CDLabel(Pointer,&Layer,&Label,&X,&Y,&Xform);
            ShowLabel(DrawingColor,Label,X,Y,Xform,True);
            continue;
        }
    }
    if (TCheck()) return;

    if (Not CDInitGen(CellDesc,0,AOI->kaLeft,AOI->kaBottom,AOI->kaRight,
        AOI->kaTop,&GenDesc)) MallocFailed();

    loop {
        /* Test for user interrupt */
        if (Parameters.kpSIGINTERRUPT) break;

        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE) continue;
        CDCall(Pointer,&MasterName,&NumX,&DX,&NumY,&DY);

        if (IsCellInLib(MasterName)) {

            if (OpenCell(MasterName,&MasterDesc)) break;

            if (Not CDBB(MasterDesc,(struct o *)NULL,&MasterBB.kaLeft,
                &MasterBB.kaBottom,&MasterBB.kaRight,&MasterBB.kaTop))
                MallocFailed();

            SetTransform(Pointer);
            TPremultiply();

            for (Int1 = NumY-1; Int1 >= 0; --Int1) {
                for (Int2 = 0; Int2 < NumX; ++Int2 ){
                    TPush();
                    TIdentity();
                    TTranslate(Int2*DX,Int1*DY);
                    TPremultiply();
                    /*
                     * Box test instance array element with AOI.
                     */
                    ka_copy(BB,(&MasterBB));
                    if (Not (BB.kaLeft > AOI->kaRight Or 
                        BB.kaRight  < AOI->kaLeft Or
                        BB.kaBottom > AOI->kaTop Or
                        BB.kaTop    < AOI->kaBottom))
                        redisplay_layer(MasterDesc,AOI); 
                    TPop();
                }
            }
            TPop();
        }
    }

    if (Not CDInitGen(CellDesc,0,AOI->kaLeft,AOI->kaBottom,AOI->kaRight,
        AOI->kaTop,&GenDesc)) MallocFailed();

    loop {
        /* Test for user interrupt */
        if (Parameters.kpSIGINTERRUPT) break;

        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE) continue;
        CDCall(Pointer,&MasterName,&NumX,&DX,&NumY,&DY);
        if (IsCellInLib(MasterName)) continue;

        if (OpenCell(MasterName,&MasterDesc)) break;

        if (CellDesc != Parameters.kpCellDesc ||
            Parameters.kpExpandInstances) {

            if (Not CDBB(MasterDesc,(struct o *)NULL,&MasterBB.kaLeft,
                &MasterBB.kaBottom,&MasterBB.kaRight,&MasterBB.kaTop))
                MallocFailed();

            SetTransform(Pointer);
            TPremultiply();

            for (Int1 = NumY-1; Int1 >= 0; --Int1) {
                for (Int2 = 0; Int2 < NumX; ++Int2 ){
                    TPush();
                    TIdentity();
                    TTranslate(Int2*DX,Int1*DY);
                    TPremultiply();
                    /*
                     * Box test instance array element with AOI.
                     */
                    ka_copy(BB,(&MasterBB));
                    if (Not (BB.kaLeft > AOI->kaRight Or 
                        BB.kaRight  < AOI->kaLeft Or
                        BB.kaBottom > AOI->kaTop Or
                        BB.kaTop    < AOI->kaBottom))
                        redisplay_layer(MasterDesc,AOI); 
                    TPop();
                }
            }
            TPop();
        }
        else {
            /*
             * Unexpanded form of an instance array is its BB with the
             * name of its master shown in its center. 
             * If it is not 1 by 1, tell the user its dimensions. 
             * CDBB will always return True if Pointer != NULL.
             */
            CDStatusInt = CDBB(CellDesc,Pointer,&BB.kaLeft,&BB.kaBottom,
                &BB.kaRight,&BB.kaTop);

            if (BB.kaRight < BB.kaLeft) SwapInts(BB.kaLeft,BB.kaRight);
            if (BB.kaTop < BB.kaBottom) SwapInts(BB.kaBottom,BB.kaTop);

            /*
             * Outline BB of Instance.
             */
            ShowEmptyBox(InstanceBBColor,&BB);

            if (NumX != 1 Or NumY != 1)
                sprintf(TypeOut,"%d/%d %s",NumX,NumY,MasterName);
            else sprintf(TypeOut,"%s",MasterName);
            ShowLabel(InstanceNameColor,TypeOut,BB.kaLeft,
                BB.kaBottom + (BB.kaTop-BB.kaBottom)/2,0,True);
        }
    }
}


void
RedisplayAfterInterrupt()

{
    TInit();
    ShowPrompt("Interrupted.  Type Ctrl A to abort.");
    InitSignals();
    ShowCommandMenu();
}


int
TCheck()

{
    char *TypeIn;

    if (TFull()) {
        ShowPrompt("Cell hierarchy is too deep. MORE");
        (void)FBGetchar(ERASE);
        ShowPrompt("Probably you have a recursive hierarchy.");
        TInit();
        return (True);
    }
    return (False);
}


void
SetTransform(Pointer)

struct o *Pointer;
{
    struct t *TGen;
    long X,Y;
    char Type;

    TPush();
    TIdentity();
    CDInitTGen(Pointer,&TGen);
    loop {
        CDTGen(&TGen,&Type,&X,&Y);
        if (TGen == NULL) break;

        if   (Type == CDROTATE)    TRotate(X,Y);
        elif (Type == CDTRANSLATE) TTranslate(X,Y);
        elif (Type == CDMIRRORX)   TMX();
        elif (Type == CDMIRRORY)   TMY();
    }
}

