/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Interface to Spice for output graphics.
 */

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"
#include "scedio.h"

/* selected trace hypertext entries */
static struct hyEnt *GPoint[9];

/* an iplot is active if set */
static int DoingIplot;
int IplotStatusChanged;

#ifdef __STDC__
static char *hent_string(struct ka*,int,struct hyEnt**);
static wordlist *GP_wl(void);
static int  select_trace(void);
static void show_mark(int,int,int);
static int  trans_mark(int);
static char *get_subname(struct hyEnt*);
static struct hyEnt *select_point(struct s*,struct ka*,int);
static struct hyEnt *alloc_ent(long,long,struct o*,int,int);
static struct hyEnt *subckt_point(struct o*,char*,struct ka*,int);
static char *nodestring(struct hyEnt*);
static int  get_enode(int,struct o*);
static int  find_wire_node(struct o*);
static void display_graf_references(int);
static int  nodenum(struct hyEnt*);
#else
static char *hent_string();
static wordlist *GP_wl();
static int  select_trace();
static void show_mark();
static int  trans_mark();
static char *get_subname();
static struct hyEnt *select_point();
static struct hyEnt *alloc_ent();
static struct hyEnt *subckt_point();
static char *nodestring();
static int  get_enode();
static int  find_wire_node();
static void display_graf_references();
static int  nodenum();
#endif

/* imports from spice */
#ifdef __STDC__
extern wordlist *cp_lexer(char*);
extern void com_graf(wordlist*);
extern void wl_free(wordlist*);
#else
extern wordlist *cp_lexer();
extern void com_graf();
extern void wl_free();
#endif


/***********************************************************************
 *
 * Routines for displaying spice output graphically.
 *
 ***********************************************************************/

extern char *MenuGRAF;
extern char *MenuIPLOT;
extern char *MenuEDIT;


void
ShowOutput(LookedAhead)

int *LookedAhead;
{
    char buf[128],*name;
    char *s, *t;
    wordlist *wl, opt;
    int i;

    if (Parameters.kpCellDesc != Parameters.kpTopDesc) {
        ShowPrompt("Pop to top level first.");
        return;
    }
    MenuSelect(MenuGRAF);
    ShowPrompt("Select nodes/branches, graf button to plot, undo clears all.");
    ConnectRecursive(Parameters.kpCellDesc);

    display_graf_references(DISPLAY);

    loop {
        switch (PointLoop(LookedAhead)) {
            case PL_UND:
                display_graf_references(ERASE);
                GPclearGraf();
                ShowPrompt("All traces cleared (can't be undone).");
                continue;
            case PL_ESC:
                goto quit;
            case PL_PCW:
                if (select_trace()) continue;
                ShowPromptAndWait("Limit eight traces.");
                continue;
            case PL_CMD:
                if (Matching(MenuGRAF)) {
                    opt.wl_next = GP_wl();
                    opt.wl_word = "-t";
                    opt.wl_prev = NULL;
                    com_graf(&opt);
                    wl_free(opt.wl_next);
                    if (!dispdev->windows && !dispdev->hardcopy)
                        RedisplaySCED(currentgraph);
                    MenuSelect(MenuGRAF);
                    (void)CheckError();
                    ShowPrompt("Select nodes/branches or press ESC to quit.");
                    *LookedAhead = False;
                }
                else if (Matching(MenuEDIT)) {
                    MenuSelect(MenuEDIT);
                    opt.wl_next = GP_wl();
                    opt.wl_word = "-t";
                    opt.wl_prev = NULL;
                    s = wl_flatten(&opt);
                    wl_free(opt.wl_next);
                    ShowPrompt("graf");
                    t = FBEdit(s);
                    ErasePrompt();
                    MenuDeselect(MenuEDIT);
                    free(s);
                    if (t) {
                        wl = cp_lexer(t);
                        com_graf(wl);
                        wl_free(wl);
                        if (!dispdev->windows && !dispdev->hardcopy)
                            RedisplaySCED(currentgraph);
                        MenuSelect(MenuGRAF);
                        (void)CheckError();
                    }
                    *LookedAhead = False;
                }
                else
                    goto quit;
        }
    }
quit:
    display_graf_references(ERASE);
    MenuDeselect(MenuGRAF);
}


void
DoIplot()

{
    wordlist wx,*wl;

    if (DoingIplot) {
        MenuDeselect(MenuIPLOT);
        DoingIplot = False;
        ShowPrompt("No plotting during simulation.");
    }
    else {
        MenuSelect(MenuIPLOT);
        DoingIplot = True;
        ShowPrompt("Will generate plot while simulating.");
    }
    IplotStatusChanged = True;
}


void
SetIplot()

{
    wordlist wl;

    if (!DoingIplot) return;
    wl.wl_prev = NULL;
    wl.wl_next = GP_wl();
    wl.wl_word = "graf";
    if (wl.wl_next) {
        (*ft_sced->sc_setiplot)(&wl);
        wl_free(wl.wl_next);
    }
    IplotStatusChanged = False;
}


void
ClearIplot()

{
    wordlist wl;

    if (!DoingIplot) return;
    wl.wl_next = NULL;
    wl.wl_word = "iplot";
    (*ft_sced->sc_clriplot)(&wl);
    IplotStatusChanged = False;
}


char *
GPstring(typemask,hent)

/* return a string naming the node/branch just pointed to */
int typemask;        /* return these types only */
struct hyEnt **hent; /* point descriptor, if found */
{
    struct ka AOI;
    int Delta;

    if (!Parameters.kpPointCoarseWindow)
        return (NULL);
    if (SCursor.kcInFine == True)
        Delta = 5.0/View->kvFineRatio;
    else
        Delta = 5.0/View->kvCoarseRatio;

    AOI.kaLeft = SCursor.kcX - Delta;
    AOI.kaRight = SCursor.kcX + Delta;
    AOI.kaBottom = SCursor.kcY - Delta;
    AOI.kaTop = SCursor.kcY + Delta;

    return (hent_string(&AOI,typemask,hent));
}


char *
GPgetString(hent)

/* return the string associated with hent, if successful
 * update entries in hent
 */
struct hyEnt *hent;
{
    char *name, buf[128];
    struct ka AOI;
    struct hyEnt *newent;
    struct parent *p, *pnext;
    int Delta = 5; /* some slop, should already be centered */
    char *s;

    if (!hent)
        return (NULL);

    AOI.kaLeft = hent->hyX - Delta;
    AOI.kaRight = hent->hyX + Delta;
    AOI.kaBottom = hent->hyY - Delta;
    AOI.kaTop = hent->hyY + Delta;

    s = hent_string(&AOI,hent->hyRefType,&newent);
    if (s) {
        for (p = hent->hyParent; p; p = pnext) {
            pnext = p->pNext;
            txfree((char*)p);
        }
        *hent = *newent;
        txfree((char*)newent);
    }
    return (s);
}


static char *
hent_string(AOI,typemask,hret)

/* return a string naming the node/branch just referenced */
struct ka *AOI;      /* region to search */
int typemask;        /* return these types only */
struct hyEnt **hret; /* returned point descriptor */
{
    struct hyEnt *h;
    struct parent *p, *pnext;
    char *name, buf[128];

    *hret = NULL;
    h = select_point(Parameters.kpCellDesc,AOI,typemask);
    if (h == NULL) return (NULL);
    name = get_subname(h);
    if (name == NULL)
        return (NULL);
    if (h->hyRefType == HY_NODE)
        sprintf(buf,"v(%s)",name);
    else {
        strcpy(buf,name);
    }
    *hret = h;
    free(name);
    return (CopyString(buf));
}


static wordlist *
GP_wl()

/* make up command list */
{
    wordlist *wl,*wl1 = NULL;
    int i;
    char *name;

    for (i = 0; i < 8 && GPoint[i]; i++) {
        name = get_subname(GPoint[i]);
        if (GPoint[i]->hyRefType == HY_NODE)
            sprintf(TypeOut,"v(%s)",name);
        else {
            strcpy(TypeOut,name);
        }
        free(name);
        if (wl1 == NULL)
            wl = wl1 = cp_lexer(TypeOut);
        else {
            wl->wl_next = cp_lexer(TypeOut);
            wl = wl->wl_next;
        }
    }
    if (wl1)
        wl->wl_next = NULL;
    return (wl1);
}


int
GPdeleteGrafRef(hent)

/* returns true if hent is a graf reference,
 * in which case it is deleted
 */
struct hyEnt *hent;
{
    struct parent *p, *pnext;
    int i;

    if (!hent)
        return (False);
    for (i = 0; i < 8; i++)
        if (GPoint[i] == hent)
            break;
    if (i == 8)
        return (False);
    for (p = hent->hyParent; p; p = pnext) {
        pnext = p->pNext;
        txfree((char*)p);
    }
    txfree((char*)hent);
    for ( ; i < 8; i++)
        GPoint[i] = GPoint[i+1];
    GPoint[i] = NULL;
    return (True);
}


void
GPclearGraf()
{
    while (GPoint[0]) {
        HYdel(GPoint[0]);
        (void)GPdeleteGrafRef(GPoint[0]);
    }
    IplotStatusChanged = True;
}


static int
select_trace()

/* return False if too many traces selected */
{
    struct ka AOI;
    int i, j, Delta;
    struct hyEnt *h;
    struct parent *p, *pnext;
    char *s1, *s2;

    if (SCursor.kcInFine == True)
        Delta = 5.0/View->kvFineRatio;
    else
        Delta = 5.0/View->kvCoarseRatio;

    AOI.kaLeft = SCursor.kcX - Delta;
    AOI.kaRight = SCursor.kcX + Delta;
    AOI.kaBottom = SCursor.kcY - Delta;
    AOI.kaTop = SCursor.kcY + Delta;

    h = select_point(Parameters.kpCellDesc,&AOI,HY_NODE|HY_BRAN);
    if (h == NULL) return (True);

    /* is it already selected? */
    for (i = 0; i < 8; i++) {
        if (!GPoint[i])
            continue;

        if (h->hyRefType == GPoint[i]->hyRefType) {
            s1 = get_subname(h);
            s2 = get_subname(GPoint[i]);
            j = strcmp(s1,s2);
            free(s1);
            free(s2);
            if (!j)
                break;
        }
    }
    if (i < 8) {

        show_mark(i,ERASE,True);
        /* erase renumbered marks */
        for (j = i+1; j < 8 && GPoint[j]; j++)
            show_mark(j,ERASE,False);
        HYdel(GPoint[i]);
        (void)GPdeleteGrafRef(GPoint[i]);
        /* redraw renumbered marks */
        for (; i < 8 && GPoint[i]; i++)
            show_mark(i,DISPLAY,False);
        for (p = h->hyParent; p; p = pnext) {
            pnext = p->pNext;
            free(p);
        }
        free(h);
    }
    else {

        for (i = 0; i < 8; i++)
            if (!GPoint[i]) break;

        if (i == 8) {
            for (p = h->hyParent; p; p = pnext) {
                pnext = p->pNext;
                free(p);
            }
            free(h);
            return (False);
        }

        GPoint[i] = h;
        HYadd(h);

        show_mark(i,DISPLAY,True);
    }
    IplotStatusChanged = True;
    return (True);
}


static void
show_mark(i,DorE,prompt)

int i;
int DorE;
int prompt;
{
    long x, y;
    struct parent *p;
    int orient;

    x = GPoint[i]->hyX;
    y = GPoint[i]->hyY;

    for (p = GPoint[i]->hyParent; p; p = p->pNext) {
        SetTransform(p->pPointer);
        TPremultiply();
    }

    TPoint(&x,&y);
    orient = trans_mark(i);

    for (p = GPoint[i]->hyParent; p; p = p->pNext) {
        TPop();
    }

    if (DorE == DISPLAY) {
        ShowMarker(DISPLAY,HighlightingColor,x,y,300,i+2,orient);
        if (prompt) {
            if (GPoint[i]->hyRefType == HY_NODE)
                sprintf(TypeOut,"Trace %d: v(%s)",i+1,get_subname(GPoint[i]));
            else
                sprintf(TypeOut,"Trace %d: %s",i+1,get_subname(GPoint[i]));
            ShowPrompt(TypeOut);
        }
    }
    else {
        if (prompt) {
            ShowPrompt("Trace deleted.");
        }
        ShowMarker(ERASE,HighlightingColor,x,y,300,i+2,orient);
    }
}


static int
trans_mark(i)
int i;
{
    long X,Y,X1,Y1;
    int orient;

    switch (GPoint[i]->hyOrient) {
        default:
        case MARK_NONE:
            return (MARK_NONE);
        case MARK_DN:
            X = 0;
            Y = -1;
            break;
        case MARK_RT:
            X = 1;
            Y = 0;
            break;
        case MARK_UP:
            X = 0;
            Y = 1;
            break;
        case MARK_LT:
            X = -1;
            Y = 0;
            break;
    }
    X1 = 0;
    Y1 = 0;
    TPoint(&X,&Y);
    TPoint(&X1,&Y1);
    X -= X1;
    Y -= Y1;

    if (X == 0) {
        if (Y == 1)
            orient = MARK_UP;
        else
            orient = MARK_DN;
    }
    else {
        if (X == 1)
            orient = MARK_RT;
        else
            orient = MARK_LT;
    }
    return (orient);
}


static char *
get_subname(hent)

struct hyEnt *hent;
{
    char buf[256];
    char *c;
    struct parent *p;

    if (hent->hyRefType == HY_NODE) {
        if (nodenum(hent) == 0) {
            /* ground node */
            return (CopyString("0"));
        }
        return (nodestring(hent));
    }

    /* device */
    c = GetName(hent->hyPointer->oPrptyList);
    sprintf(buf,"%c",*c);
    p = hent->hyParent;
    if (p)
        strcat(buf,":");
    for ( ; p; p = p->pNext) {
        c = GetName(p->pPointer->oPrptyList);
        strcat(buf,c+1);
        strcat(buf,":");
    }
    c = GetName(hent->hyPointer->oPrptyList);
    strcat(buf,c+1);
    if (hent->hyRefType == HY_BRAN)
        strcat(buf,"#branch");
    strtolower(buf);
    return (CopyString(buf));
}


static struct hyEnt *
select_point(CellDesc,AOI,mask)

struct s *CellDesc;
struct ka *AOI;
int mask;
{
    struct g *GenDesc;
    struct o *Pointer;
    struct p *Path;
    struct prpty *PDesc;
    struct ka InvAOI;
    struct hyEnt *h;
    char *name,**b;
    int  i,Layer;
    long X,Y,Width,*xy;
    long X1,Y1;


    InvAOI = *AOI;
    TInverse();
    TInversePoint(&InvAOI.kaLeft,&InvAOI.kaBottom);
    TInversePoint(&InvAOI.kaRight,&InvAOI.kaTop);

    /* first check the wires */
    if (mask & HY_NODE) {

        if (Not CDInitGen(CellDesc,1,AOI->kaLeft,AOI->kaBottom,
            AOI->kaRight,AOI->kaTop,&GenDesc)) MallocFailed();
        loop {
            CDGen(CellDesc,GenDesc,&Pointer);
            if (Pointer == NULL) break;
            if (Pointer->oInfo == SQ_GONE) continue;
            if (Pointer->oType != CDWIRE) continue;
            CDWire(Pointer,&Layer,&Width,&Path);

            if ((xy = InPath((int)(AOI->kaRight - AOI->kaLeft)/2,
                Path,(InvAOI.kaRight + InvAOI.kaLeft)/2,
                (InvAOI.kaTop + InvAOI.kaBottom)/2)) == NULL)
                continue;
            tfree(GenDesc);
            return (alloc_ent(*xy, *(xy+1), Pointer, HY_NODE, MARK_NONE));
        }
    }

    /* no wire selected, try the devices */

    if (Not CDInitGen(CellDesc,0,AOI->kaLeft,AOI->kaBottom,
        AOI->kaRight,AOI->kaTop,&GenDesc)) MallocFailed();
    loop {
        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE) continue;

        /* first check the nodes */
        if (mask & HY_NODE) {
            PDesc = Pointer->oPrptyList;
            for (; PDesc; PDesc = PDesc->prpty_Succ) {
                if (PDesc->prpty_Value != P_NODE) continue;
                X = PDesc->prpty_Data->p_node.x;
                Y = PDesc->prpty_Data->p_node.y;

                if (InBox(X,Y,&InvAOI)) {
                    tfree(GenDesc);
                    return (alloc_ent(X, Y, Pointer, HY_NODE, MARK_NONE));
                }
            }
        }
        /* no matching nodes */

        name = ((struct c *)Pointer->oRep)->cMaster->mName;

        if (!IsCellInLib(name)) {
            /* a subcircuit */
            h = subckt_point(Pointer,name,AOI,mask);
            if (h) {
                tfree(GenDesc);
                return (h);
            }
            else
                continue;
        }

        PDesc = Pointer->oPrptyList;
        for ( ; PDesc; PDesc = PDesc->prpty_Succ)
            if (PDesc->prpty_Value == P_BRANCH) break;
        if (PDesc == NULL) {
            /* not a "branch" device */
            if (mask & HY_DEVN) {
                /* just want device's name */
                CDBB(CellDesc,Pointer,&X,&Y,&X1,&Y1);
                X = (X+X1)/2;
                Y = (Y+Y1)/2;
                tfree(GenDesc);
                return (alloc_ent(X, Y, Pointer, HY_DEVN, MARK_NONE));
            }
            else
                continue;
        }

        /* is a "branch" device */
        X = PDesc->prpty_Data->p_branch.x;
        Y = PDesc->prpty_Data->p_branch.y;
        if (InBox(X,Y,&InvAOI)) {
            /* pointing at the "branch" area */

            if (mask & HY_BRAN) {
                X1 = PDesc->prpty_Data->p_branch.r1;
                Y1 = PDesc->prpty_Data->p_branch.r2;
                if (X1 == 0) {
                    if (Y1 == 1)
                        i = MARK_UP;
                    else
                        i = MARK_DN;
                }
                else {
                    if (X1 == 1)
                        i = MARK_RT;
                    else
                        i = MARK_LT;
                }
                tfree(GenDesc);
                return (alloc_ent(X, Y, Pointer, HY_BRAN, i));
            }
        }
        if (mask & HY_DEVN) {
            tfree(GenDesc);
            return (alloc_ent(X, Y, Pointer, HY_DEVN, MARK_NONE));
        }
    }
    return (NULL);
}


static struct hyEnt *
alloc_ent(x,y,pointer,reftype,orient)

long x,y;
struct o *pointer;
int reftype;
int orient;
{
    struct hyEnt *h = (struct hyEnt *)tmalloc(sizeof(struct hyEnt));

    h->hyX = x;
    h->hyY = y;
    h->hyPointer = pointer;
    h->hyRefType = reftype;
    h->hyOrient = orient;
    return (h);
}


static struct hyEnt *
subckt_point(pointer,name,AOI,mask)

struct o *pointer;
char *name;
struct ka *AOI;
int mask;
{
    struct s *NewDesc;
    struct hyEnt *h = NULL;
    struct parent *p;

    if (!OpenCell(name,&NewDesc)) {
        SetTransform(pointer);
        TPremultiply();
        h = select_point(NewDesc,AOI,mask);
        TPop();
    }

    if (h) {
        p = (struct parent *)tmalloc(sizeof(struct parent));
        p->pPointer = pointer;
        p->pNext = h->hyParent;
        h->hyParent = p;
    }
    return (h);
}


static char *
nodestring(h)

struct hyEnt *h;
{
    struct parent *p, *pp, *pTmp;
    char buf[256], *c;
    int node, nnode;

    pp = NULL;

    /* copy new reversed list */
    pTmp = NULL;
    for (p = h->hyParent; p; p = p->pNext) {
        pp = (struct parent *)tmalloc(sizeof(struct parent));
        pp->pPointer = p->pPointer;
        pp->pNext = pTmp;
        pTmp = pp;
    }

    /* walk back up node naming hierarchy */
    node = nodenum(h);
    for (; pp; pp = p) {
        p = pp->pNext;
        nnode = get_enode(node,pp->pPointer);
        if (nnode == -1)
            break;
        node = nnode;
        free(pp);
    }

    /* reverse the list */
    pTmp = NULL;
    for (; pp; pp = p) {
        p = pp->pNext;
        pp->pNext = pTmp;
        pTmp = pp;
    }
    pp = pTmp;

    /* create text and free */
    *buf = '\0';
    for (; pp; pp = p) {
        p = pp->pNext;
        c = GetName(pp->pPointer->oPrptyList);
        strcat(buf,c+1);
        strcat(buf,":");
        free(pp);
    }
    strtolower(buf);
    sprintf(buf + strlen(buf),"%d",node);
    return (CopyString(buf));
}


static int
get_enode(Node,Pointer)

/* If the selected node is tied to a node in the parent
 * subcircuit, return the parent's node, otherwise
 * return -1.
 */
int Node;
struct o *Pointer;
{
    struct prpty *PDesc;
    struct s *CDesc;
    char *name;
    int inode = -1;

    name = ((struct c *)Pointer->oRep)->cMaster->mName;
    if (OpenCell(name,&CDesc))
        return (-1);

    PDesc = CDesc->sPrptyList;
    for ( ; PDesc; PDesc = PDesc->prpty_Succ) {
        if (PDesc->prpty_Value != P_NODE) continue;
        if (Node == PDesc->prpty_Data->p_node.enode) {
            inode = PDesc->prpty_Data->p_node.inode;
            break;
        }
    }
    if (inode == -1)
        return (-1);

    PDesc = Pointer->oPrptyList;
    for ( ; PDesc; PDesc = PDesc->prpty_Succ) {
        if (PDesc->prpty_Value != P_NODE) continue;
        if (inode == PDesc->prpty_Data->p_node.inode)
            return (PDesc->prpty_Data->p_node.enode);
    }
    return (-1);
}


static int
find_wire_node(Pointer)

/* return the node number of the wire in Pointer */
struct o *Pointer;
{
    struct prpty *PDesc;

    PDesc = Pointer->oPrptyList;
    for (; PDesc; PDesc = PDesc->prpty_Succ) {
        if (PDesc->prpty_Value != P_NODE) continue;
        return (PDesc->prpty_Data->p_node.enode);
    }
    return (-1);
}


static void
display_graf_references(DisplayOrErase)

char DisplayOrErase;
{
    int i;

    for (i = 0; i < 8 && GPoint[i]; i++) {
        show_mark(i,DisplayOrErase,False);
    }
}


static int
nodenum(h)

struct hyEnt *h;
{
    struct prpty *PDesc;
    long X, Y;

    if (h->hyRefType != HY_NODE)
        return (-1);
    if (h->hyPointer->oType == CDWIRE)
        return (find_wire_node(h->hyPointer));

    PDesc = h->hyPointer->oPrptyList;
    for (; PDesc; PDesc = PDesc->prpty_Succ) {
        if (PDesc->prpty_Value != P_NODE) continue;
        X = PDesc->prpty_Data->p_node.x;
        Y = PDesc->prpty_Data->p_node.y;
        if (h->hyX == X && h->hyY == Y) {
            return (PDesc->prpty_Data->p_node.enode);
        }
    }
    return (-1);
}
