/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * SCED grid and line drawing routines.
 */

#include "spice.h"
#include "sced.h"


void
ShowGrid()

{
    /*
     * Show grid for Window within AOI.
     * Window maps to Viewport.
     */
    long YInL,XInL;
    long XInP,YInP;
    long Left,Top,Right,Bottom;
    long Step;

    if (!Parameters.kpGridDisplayed Or Parameters.kpGrid <= 0)
        return;

    DevDefineLinestyle(1,Parameters.kpGridLineStyle);
    DevSetLinestyle(1);

    /*
     *
     * The grid is computed relative to (0,0).
     *
     */
    Step = Parameters.kpGrid*5;

    if (
        (Parameters.kpRedisplayControl != FINEVIEWPORTONLY And
        (View->kvCoarseViewport == View->kvLargeCoarseViewport Or
        Parameters.kpShowGridInLargeViewport))
        And
        Parameters.kpGrid*View->kvCoarseRatio
        > Parameters.kpPointingThreshold And CurrentAOI.aInCoarse) {

        Top = View->kvCoarseWindow->kaTop -
            (View->kvCoarseWindow->kaTop % Step);

        if (View->kvCoarseWindow->kaTop >= 0)
            Top += Step;

        Bottom = View->kvCoarseWindow->kaBottom;

        Left = View->kvCoarseWindow->kaLeft -
            (View->kvCoarseWindow->kaLeft % Step);

        if (View->kvCoarseWindow->kaLeft < 0)
            Left -= Step;
        
        Right = View->kvCoarseWindow->kaRight;

        if (Parameters.kpGridLineStyle == 0) {

            /* point grid */
            for (YInL = Top; YInL >= Bottom; YInL -= Parameters.kpGrid) {
                YInP = .5+(YInL - View->kvCoarseWindow->kaBottom)*
                    View->kvCoarseRatio;
                YInP += View->kvCoarseViewport->kaBottom;
                if (YInP < CurrentAOI.aBC) break;
                if (YInP > CurrentAOI.aTC) continue;
                for (XInL = Left; XInL <= Right; XInL += Parameters.kpGrid) {
                    XInP = .5+(XInL - View->kvCoarseWindow->kaLeft)*
                        View->kvCoarseRatio;
                    XInP += View->kvCoarseViewport->kaLeft;
                    if (XInP > CurrentAOI.aRC) break;
                    if (XInP < CurrentAOI.aLC) continue;
                    if ((Top-YInL) % Step == 0 || (XInL-Left) % Step == 0)
                        DevSetColor(CoarseGridColor);
                    else
                        DevSetColor(FineGridColor);

                    DevPixel((int)XInP,(int)YInP);
                }
            }
        }
        else {

            /* Horizontal grid lines */
            for (YInL = Top; YInL >= Bottom; YInL -= Parameters.kpGrid) {

                YInP = .5+(YInL - View->kvCoarseWindow->kaBottom)*
                    View->kvCoarseRatio;
                YInP += View->kvCoarseViewport->kaBottom;
                if (YInP < CurrentAOI.aBC) break;
                if (YInP > CurrentAOI.aTC) continue;

                if ((Top-YInL) % Step == 0)
                    DevSetColor(CoarseGridColor);
                else
                    DevSetColor(FineGridColor);

                DevLine((int)CurrentAOI.aLC,(int)YInP,
                    (int)CurrentAOI.aRC,(int)YInP);
            
            }

            /* Vertical grid lines */
            for (XInL = Left; XInL <= Right; XInL += Parameters.kpGrid) {

                XInP = .5+(XInL - View->kvCoarseWindow->kaLeft)*
                    View->kvCoarseRatio;
                XInP += View->kvCoarseViewport->kaLeft;
                if (XInP > CurrentAOI.aRC) break;
                if (XInP < CurrentAOI.aLC) continue;

                if ((XInL-Left) % Step == 0)
                    DevSetColor(CoarseGridColor);
                else
                    DevSetColor(FineGridColor);

                DevLine((int)XInP,(int)CurrentAOI.aBC,
                    (int)XInP,(int)CurrentAOI.aTC);
            
            }
        }
    }

    if (Parameters.kpRedisplayControl != COARSEVIEWPORTONLY And
        Parameters.kpGrid*View->kvFineRatio
            > Parameters.kpPointingThreshold And CurrentAOI.aInFine) {


        Top = View->kvFineWindow->kaTop -
            (View->kvFineWindow->kaTop % Step);

        if (View->kvFineWindow->kaTop >= 0)
            Top += Step;

        Bottom = View->kvFineWindow->kaBottom;

        Left = View->kvFineWindow->kaLeft -
            (View->kvFineWindow->kaLeft % Step);

        if (View->kvFineWindow->kaLeft < 0)
            Left -= Step;

        Right = View->kvFineWindow->kaRight;

        if (Parameters.kpGridLineStyle == 0) {

            /* point grid */
            for (YInL = Top; YInL >= Bottom; YInL -= Parameters.kpGrid) {
                YInP = .5+(YInL - View->kvFineWindow->kaBottom)*
                    View->kvFineRatio;
                YInP += View->kvFineViewport->kaBottom;
                if (YInP < CurrentAOI.aBF) break;
                if (YInP > CurrentAOI.aTF) continue;
                for (XInL = Left; XInL <= Right; XInL += Parameters.kpGrid) {
                    XInP = .5+(XInL - View->kvFineWindow->kaLeft)*
                        View->kvFineRatio;
                    XInP += View->kvFineViewport->kaLeft;
                    if (XInP > CurrentAOI.aRF) break;
                    if (XInP < CurrentAOI.aLF) continue;
                    if ((Top-YInL) % Step == 0 || (XInL-Left) % Step == 0)
                        DevSetColor(CoarseGridColor);
                    else
                        DevSetColor(FineGridColor);

                    DevPixel((int)XInP,(int)YInP);
                }
            }
        }
        else {
            /* Horizontal grid lines */
            for (YInL = Top; YInL >= Bottom; YInL -= Parameters.kpGrid) {

                YInP = .5+(YInL - View->kvFineWindow->kaBottom)*
                    View->kvFineRatio;
                YInP += View->kvFineViewport->kaBottom;
                if (YInP < CurrentAOI.aBF) break;
                if (YInP > CurrentAOI.aTF) continue;

                if ((Top-YInL) % Step == 0)
                       DevSetColor(CoarseGridColor);
                else
                       DevSetColor(FineGridColor);
                DevLine((int)CurrentAOI.aLF,(int)YInP,
                    (int)CurrentAOI.aRF,(int)YInP);
            }

            /* Vertical grid lines */
            for (XInL = Left; XInL <= Right; XInL += Parameters.kpGrid){

                XInP = .5+(XInL - View->kvFineWindow->kaLeft)*
                    View->kvFineRatio;
                XInP += View->kvFineViewport->kaLeft;
                if (XInP > CurrentAOI.aRF) break;
                if (XInP < CurrentAOI.aLF) continue;

                if ((XInL-Left) % Step == 0)
                    DevSetColor(CoarseGridColor);
                else
                    DevSetColor(FineGridColor);
                DevLine((int)XInP,(int)CurrentAOI.aBF,
                    (int)XInP,(int)CurrentAOI.aTF);
            }
        }
    }
    DevSetLinestyle(0);
}


void
ShowLine(Color,X1,Y1,X2,Y2)

int Color;
long X1,Y1,X2,Y2;
{
    long X1P,Y1P,X2P,Y2P;

    TPoint(&X1,&Y1);
    TPoint(&X2,&Y2);

    DevSetColor(Color);

    if (Parameters.kpRedisplayControl != FINEVIEWPORTONLY And
        CurrentAOI.aInCoarse) {

        CoarseLToP(X1,Y1,X1P,Y1P);
        CoarseLToP(X2,Y2,X2P,Y2P);
        if (Not clip_line(&X1P,&Y1P,&X2P,&Y2P,CurrentAOI.aLC,
            CurrentAOI.aBC,CurrentAOI.aRC,CurrentAOI.aTC)){

            DevLine((int)X1P,(int)Y1P,(int)X2P,(int)Y2P);
        }
    }

    if (Parameters.kpRedisplayControl != COARSEVIEWPORTONLY And
        CurrentAOI.aInFine) {

        FineLToP(X1,Y1,X1P,Y1P);
        FineLToP(X2,Y2,X2P,Y2P);
        if (Not clip_line(&X1P,&Y1P,&X2P,&Y2P,CurrentAOI.aLF,
            CurrentAOI.aBF,CurrentAOI.aRF,CurrentAOI.aTF)) {

            DevLine((int)X1P,(int)Y1P,(int)X2P,(int)Y2P);
        }
    }
}


void
ShowManhattanLine(Color,X1,Y1,X2,Y2)

int Color;
long X1,Y1,X2,Y2;
{ 
    long X1P,Y1P,X2P,Y2P;

    DevSetColor(Color);

    if (Parameters.kpRedisplayControl != FINEVIEWPORTONLY And
        CurrentAOI.aInCoarse) {

        CoarseLToP(X1,Y1,X1P,Y1P);
        CoarseLToP(X2,Y2,X2P,Y2P);

        if (X1P == X2P) {
            if (X1P < CurrentAOI.aLC Or X1P > CurrentAOI.aRC) return;
            if (Y1P > Y2P) SwapInts(Y1P,Y2P);
            if (Y1P > CurrentAOI.aTC Or Y2P < CurrentAOI.aBC) return;
            if (Y1P < CurrentAOI.aBC) Y1P = CurrentAOI.aBC;
            if (Y2P > CurrentAOI.aTC) Y2P = CurrentAOI.aTC;
        }
        elif (Y1P == Y2P) {
            if (Y1P < CurrentAOI.aBC Or Y1P > CurrentAOI.aTC) return;
            if (X1P > X2P) SwapInts(X1P,X2P);
            if (X1P > CurrentAOI.aRC Or X2P < CurrentAOI.aLC) return;
            if (X1P < CurrentAOI.aLC) X1P = CurrentAOI.aLC;
            if (X2P > CurrentAOI.aRC) X2P = CurrentAOI.aRC;
        }
        else return;

        DevLine((int)X1P,(int)Y1P,(int)X2P,(int)Y2P);
    }
    if (Parameters.kpRedisplayControl != COARSEVIEWPORTONLY And
        CurrentAOI.aInFine) {

        FineLToP(X1,Y1,X1P,Y1P);
        FineLToP(X2,Y2,X2P,Y2P);

        if (X1P == X2P) {
            if (X1P < CurrentAOI.aLF Or X1P > CurrentAOI.aRF) return;
            if (Y1P > Y2P) SwapInts(Y1P,Y2P);
            if (Y1P > CurrentAOI.aTF Or Y2P < CurrentAOI.aBF) return;
            if (Y1P < CurrentAOI.aBF) Y1P = CurrentAOI.aBF;
            if (Y2P > CurrentAOI.aTF) Y2P = CurrentAOI.aTF;
        }
        elif (Y1P == Y2P) {
            if (Y1P < CurrentAOI.aBF Or Y1P > CurrentAOI.aTF) return;
            if (X1P > X2P) SwapInts(X1P,X2P);
            if (X1P > CurrentAOI.aRF Or X2P < CurrentAOI.aLF) return;
            if (X1P < CurrentAOI.aLF) X1P = CurrentAOI.aLF;
            if (X2P > CurrentAOI.aRF) X2P = CurrentAOI.aRF;
        }
        else return;

        DevLine((int)X1P,(int)Y1P,(int)X2P,(int)Y2P);
    }
}
