/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"
#include "plotext.h"   /* for ft_hardcopy() */
#include <time.h>

/***********************************************************************
 *
 * Hcopy command.
 * HP Laserjet hardcopy support.
 *
 ***********************************************************************/

#ifdef __STDC__
static void hcpy_display(GRAPH*);
static void init_window(long,long,long,long);
static void init_hc_viewport(void);
#else
static void hcpy_display();
static void init_window();
static void init_hc_viewport();
#endif

extern char *MenuHCOPY;


void
Hcopy()

{
    MenuSelect(MenuHCOPY);
    Parameters.kpDoingHardcopy = True;

    currentgraph->redraw = hcpy_display;
    ft_hardcopy(NULL, currentgraph, True, NULL, False);
    currentgraph->redraw = RedisplaySCED;

    Parameters.kpDoingHardcopy = False;
    InitViewport();
    CenterFullView();
    RedisplaySCED(currentgraph);
    MenuDeselect(MenuHCOPY);
}


static void
hcpy_display(graph)

GRAPH* graph;
{
    int i;
    long L,B,R,T;
    long VPmaxX,VPmaxY,MidX,Wid,Hei;
    double Ratio;
    int scale,Margin;
    int Leg;
    char OldControl;
    char buf[64];

    PushGraphContext(graph);
    OldControl = Parameters.kpRedisplayControl;
    Parameters.kpRedisplayControl = COARSEVIEWPORTONLY;

    init_hc_viewport();

    if (Not CDBB(Parameters.kpCellDesc,(struct o *)NULL,&L,&B,&R,&T))
        MallocFailed();

if (!strcmp(dispdev->name,"wp")) {
    init_window((R+L)/2,(T+B)/2,(R-L)+(R-L)/10,((T-B)+(T-B)/10)/2);

    Redisplay(View->kvCoarseWindow);
    Parameters.kpRedisplayControl = OldControl;
    PopGraphContext();
    return;
}

    /* scale for text */
    scale = 2400/graph->absolute.width;
    if      (scale == 1) Parameters.kpHardcopyTextScale = 4; /* 300dpi */
    else if (scale == 2) Parameters.kpHardcopyTextScale = 2; /* 150dpi */
    else                 Parameters.kpHardcopyTextScale = 1; /* 75,100dpi */

    TPush();
    TIdentity();
    if ((R-L) > (T-B)) {
        TRotate(0L,1L);
        TPoint(&L,&B);
        TPoint(&R,&T);
        if (L > R) SwapInts(L,R);
        if (B > T) SwapInts(T,B);
    }
    else TIdentity();
    TPremultiply();
    TStore();

    Leg = 3*graph->fontheight;
    VPmaxY = graph->absolute.height - Leg - 1;
    VPmaxX = graph->absolute.width - 1;

    if (((double)(T-B))/(R-L) >
            ((double)VPmaxY)/VPmaxX) {

        Ratio = ((double)(T-B))/VPmaxY;
        Hei = (long)(.5 + (double)(T-B + Leg*Ratio)/2);
        Wid = VPmaxX*Ratio;
    }
    else {
        Ratio = ((double)(R-L))/VPmaxX;
        Hei = (VPmaxY+Leg)*Ratio/2;
        Wid = R-L;
        T = (T+B + Ratio*VPmaxY)/2;
    }
    init_window((R+L)/2,T - Hei,Wid,Hei);

    Redisplay(View->kvCoarseWindow);

    sprintf(buf,"Circuit: %s",Parameters.kpCellName);
    DevText(buf,0,currentgraph->fontheight*3/2);
    DevText(datestring(),0,currentgraph->fontheight/2);

    TPop();
    TStore();
    Parameters.kpRedisplayControl = OldControl;
    PopGraphContext();
}


static void
init_window(X,Y,Width,HalfHeight)

long X,Y,Width,HalfHeight;
{

    View->kvCoarseWindow->kaX = X;
    View->kvCoarseWindow->kaY = Y; 
    View->kvCoarseWindow->kaLeft = X - Width/2; 
    View->kvCoarseWindow->kaRight = X + Width/2; 
    View->kvCoarseWindow->kaBottom = Y - HalfHeight;
    View->kvCoarseWindow->kaTop = Y + HalfHeight;
    View->kvCoarseWindow->kaWidth = Width;
    View->kvCoarseWindow->kaHeight =
        View->kvCoarseWindow->kaTop - View->kvCoarseWindow->kaBottom;
    View->kvCoarseRatio =
        View->kvCoarseViewport->kaWidth/View->kvCoarseWindow->kaWidth;
}

static void
init_hc_viewport()
{
    struct ka *Vl;

    Vl = View->kvLargeCoarseViewport;
    Vl->kaLeft   = 0;
    Vl->kaBottom = 0;
    Vl->kaRight  = gi_maxx;
    Vl->kaTop    = gi_maxy;
    Vl->kaWidth  = Vl->kaRight - Vl->kaLeft;
    Vl->kaHeight = Vl->kaTop - Vl->kaBottom;
}

