/***************************************************************************
JSPICE3 adaptation of Spice3f4 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1994 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"

/* reference table head */
#define hyBase ((struct hyEnt**)Parameters.kpCellDesc->sHY)

char *
HYtostr(h)

/* return an expanded string with up to date entries */
struct hprlist *h;
{
    struct hprlist *hh;
    char buf[BSIZE_SP], *s;

    buf[0] = '\0';
    for (hh = h; hh; hh = hh->hlNext) {
        s = HYgetentry(hh);
        if (!s)
            strcat(buf,"<unknown>");
        else {
            strcat(buf,s);
            txfree(s);
        }
    }
    return (copy(buf));
}


char *
HYtoascii(h)

/* return an expanded string with ascii encoded entries */
struct hprlist *h;
{
    struct hprlist *hh;
    char buf[BSIZE_SP];
    char tbuf[BSIZE_SP];

    buf[0] = '\0';
    for (hh = h; hh; hh = hh->hlNext) {
        if (hh->hlRefType == HY_TEXT) {
            strcat(buf,hh->hlText);
        }
        else {
            sprintf(tbuf,"(||%d:%d %d||)",
                hh->hlRefType,hh->hlEnt->hyX,hh->hlEnt->hyY);
            strcat(buf,tbuf);
        }
    }
    return (copy(buf));
}


struct hprlist *
HYfromascii(str)

/* return a hypertext list from ascii encoded string */
char *str;
{
    struct hprlist *hh, *h0 = NULL;
    char *s, *t, *tt;
    int i;

    s = str;
    for (;;) {
        for (t = s;
            *t && (*t != '(' || *(t+1) != '|' || *(t+2) != '|'); t++) ;
        if (!*t)
            break;
        tt = t - 1;
        if (tt >= s) {
            tt++;
            if (!h0)
                h0 = hh = alloc(hprlist);
            else {
                hh->hlNext = alloc(hprlist);
                hh = hh->hlNext;
            }
            hh->hlRefType = HY_TEXT;
            hh->hlText = tmalloc(tt - s + 1);
            strncpy(hh->hlText,s,tt-s);
            hh->hlText[tt-s] = '\0';
        }
        if (!h0)
            h0 = hh = alloc(hprlist);
        else {
            hh->hlNext = alloc(hprlist);
            hh = hh->hlNext;
        }
        /* just store coords in hyEnt, will fill in rest later */
        hh->hlEnt = alloc(hyEnt);
        HYadd(hh->hlEnt);
        sscanf(t+3,"%d:%d %d",
            &hh->hlRefType,&hh->hlEnt->hyX,&hh->hlEnt->hyY);
        hh->hlEnt->hyRefType = hh->hlRefType;
        s = t+3;
        while (*s && *s != ')') s++;
        if (!*s)
            break;
        s++;
    }
    if (*s) {
        if (!h0)
            h0 = hh = alloc(hprlist);
        else {
            hh->hlNext = alloc(hprlist);
            hh = hh->hlNext;
        }
        hh->hlRefType = HY_TEXT;
        hh->hlText = copy(s);
    }
    return (h0);
}


char *
HYgetentry(h)

/* expand a hypertext entry */
struct hprlist *h;
{
    switch (h->hlRefType) {
        case HY_TEXT:
            return (copy(h->hlText));
        case HY_NODE:
        case HY_BRAN:
        case HY_DEVN:
            if (h->hlEnt->hyRefType == HY_BOGUS)
                return (NULL);
            return (GPgetString(h->hlEnt));
        default:
            return (NULL);
    }
}


struct hprlist *
HYcopy(h)

/* copy a hypertext list */
struct hprlist *h;
{
    struct hprlist *hh, *h0 = NULL, *h1;

    for (hh = h; hh; hh = hh->hlNext) {
        if (!h0)
            h0 = h1 = alloc(hprlist);
        else {
            h1->hlNext = alloc(hprlist);
            h1 = h1->hlNext;
        }
        h1->hlRefType = hh->hlRefType;
        if (hh->hlRefType == HY_TEXT) {
            if (hh->hlText)
                h1->hlText = copy(hh->hlText);
        }
        else if (hh->hlEnt) {
            h1->hlEnt = HYcopyEnt(hh->hlEnt);
            HYadd(h1->hlEnt);
        }
    }
    return (h0);
}


struct hyEnt *
HYcopyEnt(hent)

struct hyEnt *hent;
{
    struct hyEnt *newh;
    struct parent *p, *pp;

    newh = alloc(hyEnt);
    if (!hent) {
        newh->hyRefType = HY_BOGUS;
        return (newh);
    }
    *newh = *hent;
    newh->hyParent = NULL;

    for (p = hent->hyParent; p; p = p->pNext) {
        if (newh->hyParent == NULL)
            newh->hyParent = pp = alloc(parent);
        else {
            pp->pNext = alloc(parent);
            pp = pp->pNext;
        }
        pp->pPointer = p->pPointer;
        pp->pOldPointer = p->pOldPointer;
    }
    return (newh);
}


void
HYfree(h)

/* free a hypertext list */
struct hprlist *h;
{
    struct hprlist *hh;
    struct parent *p, *pnext;

    for ( ; h; h = hh) {
        hh = h->hlNext;
        txfree(h->hlText);
        if (h->hlEnt) {
            HYdel(h->hlEnt);
            for (p = h->hlEnt->hyParent; p; p = pnext) {
                pnext = p->pNext;
                txfree((char*)p);
            }
            txfree((char*)h->hlEnt);
        }
        txfree((char*)h);
    }
}


/*********************************************************************
 *
 * Hypertext input editor.
 *
 *********************************************************************/

/* cursor height */
#define CURHT 1

struct sHtxt {
    int c;                 /* character, if type == HY_TEXT */
    int type;              /* one of HY_... TEXT, NODE, BRAN, DEVN, END */
    struct hyEnt *ent;     /* point struct for NODE, BRAN, DEVN */
    char *str;             /* string for NODE, BRAN, DEVN */
};

/* static storage for edit params (allows redraw) */
struct sEdit {
    bool active;           /* true when editing */
    int x,y;               /* lower left coords of string */
    int fg, bg, cc, hc;    /* fore, back, cursor, and highlight colors */
    int col;               /* current cursor column */
    int cwid;              /* current cursor width in cols */
    struct sHtxt txt[256]; /* storage for input */
};
static struct sEdit HY;

#define DRAW 1
#define UNDRAW 0

#ifdef __STDC__
static bool hyeditor(void);
static void hydelcol(int,int);
static void hysetcol(int);
static void hytext(int,int);
static void hycursor(int);
static struct hprlist *hylist(void);
static struct sHtxt *input(void);
#else
static bool hyeditor();
static void hydelcol();
static void hysetcol();
static void hytext();
static void hycursor();
static struct hprlist *hylist();
static struct sHtxt *input();
#endif


char *
KbEdit(s,x,y,bg,fg,cc)

/* String editing */
/*
 * s       Initial string to edit.
 * x,y     Lower left coordinates.
 * bg,fg   Background and foreground colors.
 * cc      Color at cursor location.
 */
char *s;
int x, y, bg, fg, cc;
{
    int i;
    static char editbuf[256];
    char *t, *tt, *end;

    HY.bg = bg;
    HY.fg = fg;
    HY.cc = cc;
    HY.hc = cc;
    HY.x = x;
    HY.y = y;

    i = 0;
    for (t = s; t && *t; t++) {
        HY.txt[i].c = *t;
        HY.txt[i].type = HY_TEXT;
        i++;
        if (i == 255)
            break;
    }
    HY.txt[i].type = HY_END;
    if (hyeditor())
        return (NULL);
    t = editbuf;
    end = t + 255;
    for (i = 0; i < 256; i++) {
        if (HY.txt[i].type == HY_END)
            break;
        if (HY.txt[i].type == HY_TEXT) {
            *t++ = HY.txt[i].c;
        }
        else {
            tt = HY.txt[i].str;
            if (tt) {
                while (*tt != '\0') {
                    *t++ = *tt++;
                    if (t >= end)
                        break;
                }
            }
        }
        if (t >= end)
            break;
    }
    *t = '\0';
    hydelcol(0,256);
    return (editbuf);
}


struct hprlist *
HYedit(h,x,y,bg,fg,cc,hc)

/* Hypertext list editing */
/*
 * h       Initial hypertext list to edit.
 * x,y     Lower left coordinates.
 * bg,fg   Background and foreground colors.
 * cc      Color at cursor location.
 * hc      Highlighting color for hypertext references.
 */
struct hprlist *h;
int x, y, bg, fg, cc, hc;
{
    struct hprlist *hh;
    char *s;
    int i, j, k;

    HY.bg = bg;
    HY.fg = fg;
    HY.cc = cc;
    HY.hc = hc;
    HY.x = x;
    HY.y = y;

    for (i = 0,hh = h; hh; hh = hh->hlNext) {
        if (hh->hlRefType == HY_TEXT) {
            k = strlen(hh->hlText);
            for (j = 0; j < k; j++) {
                HY.txt[i].c = hh->hlText[j];
                HY.txt[i].type = HY_TEXT;
                HY.txt[i].ent = NULL;
                HY.txt[i].str = NULL;
                i++;
            }
        }
        else {
            HY.txt[i].type = hh->hlRefType;
            HY.txt[i].c = '\0';
            s = HYgetentry(hh);
            if (s)
                HY.txt[i].str = s;
            else
                HY.txt[i].str = copy("<unknown>");
            HY.txt[i].ent = HYcopyEnt(hh->hlEnt);
            i++;
        }
        if (i >= 256) {
            i = 255;
            break;
        }
    }
    HY.txt[i].type = HY_END;
    if (hyeditor())
        return (NULL);
    return (hylist());
}


void
KbRepaint(x,y)

/* redraw function */
int x,y;
{
    int tmp;

    if (!HY.active) return;
    HY.x = x;
    HY.y = y;
    hycursor(DRAW);
    tmp = HY.col;
    HY.col = 0;
    hytext(DRAW,-1);
    HY.col = tmp;
}


void
KbCursor(x,y)

/* draw a text cursor, like the editor */
int x,y;
{
    DevBox(x,y,x+currentgraph->fontwidth,y+CURHT);
}


static bool
hyeditor()

/* hypertext editor core */
{
    struct sHtxt *hret;
    int i;

    HY.active = true;
    HY.col = 0;

    hysetcol(0);
    hycursor(DRAW);
    hytext(DRAW,-1);

    for (;;) {
        hret = input();

        if (hret->type == HY_TEXT) {
            if (hret->c == '\r')
                break;
            if (hret->c == ARROW_RT) {
                /* right arrow */
                if (HY.txt[HY.col].type == HY_END) continue;
                hycursor(UNDRAW);
                hytext(DRAW,1);
                hysetcol(HY.col+1);
                hycursor(DRAW);
                hytext(DRAW,1);
                continue;
            }
            if (hret->c == ARROW_LT) {
                /* left arrow */
                if (HY.col <= 0) continue;
                hycursor(UNDRAW);
                hytext(DRAW,1);
                hysetcol(HY.col-1);
                hycursor(DRAW);
                hytext(DRAW,1);
                continue;
            }
            if (hret->c == '\b') {
                /* backspace */
                if (HY.col == 0) continue;
                hycursor(UNDRAW);
                hysetcol(HY.col-1);
                hytext(UNDRAW,-1);
                hydelcol(HY.col,1);
                for (i = HY.col; ; i++) {
                    HY.txt[i] = HY.txt[i+1];
                    if (HY.txt[i].type == HY_END)
                        break;
                }
                hytext(DRAW,-1);
                hysetcol(HY.col);
                hycursor(DRAW);
                hytext(DRAW,1);
                continue;
            }
            if (hret->c == 339 || hret->c == 127) {
                /* DEL, 339 in DOS */
                if (HY.txt[HY.col].type == HY_END) continue;
                hycursor(UNDRAW);
                hytext(UNDRAW,-1);
                hydelcol(HY.col,1);
                for (i = HY.col; ; i++) {
                    HY.txt[i] = HY.txt[i+1];
                    if (HY.txt[i].type == HY_END)
                        break;
                }
                hytext(DRAW,-1);
                hysetcol(HY.col);
                hycursor(DRAW);
                hytext(DRAW,1);
                continue;
            }
            if (hret->c == '\025' ||
                    hret->c == '\030' ||
                    hret->c == '\033') {
                /* '\025':^U, '\030':^X, '\033':ESC */
                hycursor(UNDRAW);
                HY.col = 0;
                HY.cwid = 1;
                hytext(UNDRAW,-1);
                hydelcol(0,256);
                HY.txt[0].type = HY_END;

                if (hret->c == '\033') {
                    HY.active = false;
                    return (true);
                }
                hycursor(DRAW);
                continue;
            }
            if (hret->c > 255)
                continue;
            if (((char) hret->c < ' ') || ((char) hret->c > '~'))
                continue;
        }
        hycursor(UNDRAW);
        hytext(UNDRAW,-1);
        for (i = HY.col; i < 256; i++) {
            if (HY.txt[i].type == HY_END)
                break;
        }
        i++;
        if (i >= 256) {
            i = 255;
            hydelcol(254,2);
            HY.txt[254].type = HY_END;
        }
        while (i > HY.col) {
            HY.txt[i] = HY.txt[i-1];
            i--;
        }
        HY.txt[HY.col] = *hret;
        hytext(DRAW,-1);
        hysetcol(HY.col+1);
        hycursor(DRAW);
        hytext(DRAW,1);
    }
    hycursor(UNDRAW);
    hytext(DRAW,1);
    HY.active = false;
    return (false);
}


static void
hydelcol(col,num)

/* delete the current logical column */
int col,num;
{
    struct parent *p, *pnext;

    if (col < 0 || col > 255)
        return;
    if (num < 0)
        return;
    num += col;
    if (num > 256)
        num = 256;

    for ( ; col < num; col++) {
        tfree(HY.txt[col].str);
        HY.txt[col].type = HY_END;
        if (HY.txt[col].ent) {
            for (p = HY.txt[col].ent->hyParent; p; p = pnext) {
                pnext = p->pNext;
                txfree((char*)p);
            }
            tfree(HY.txt[col].ent);
        }
        HY.txt[col].c = 0;
    }
}


static void
hysetcol(col)

/* set the current logical column */
int col;
{
    if (col < 0 || col > 255)
        return;
    switch (HY.txt[col].type) {
        case HY_TEXT:
        case HY_END:
            HY.cwid = 1;
            break;
        case HY_NODE:
        case HY_BRAN:
        case HY_DEVN:
            if (HY.txt[col].str)
                HY.cwid = strlen(HY.txt[col].str);
            else
                HY.cwid = 1;
            break;
        default:
            /* points at garbage, make it an end record */
            HY.txt[col].type = HY_END;
            HY.cwid = 1;
    }
    HY.col = col;
}


static void
hytext(draw,ncols)

/* display ncols of hypertext string, starting at the current
 * logical column, if ncols is < 0, show to end
 */
int draw, ncols;
{
    int i, k, realcol, endcol;
    char buf[BSIZE_SP];

    realcol = 0;
    for (i = 0; i < HY.col; i++) {
        if (HY.txt[i].type == HY_END)
            break;
        if (HY.txt[i].type == HY_TEXT)
            realcol++;
        else
            realcol += strlen(HY.txt[i].str);
    }
    if (i < HY.col)
        return;

    k = 0;
    if (ncols < 0)
        endcol = 255;
    else {
        endcol = i + ncols;
        if (endcol > 255)
            endcol = 255;
    }
    for (;;) {
        if (HY.txt[i].type == HY_END || i == endcol) {
            if (k) {
                buf[k] = '\0';
                if (draw)
                    DevSetColor(HY.fg);
                else
                    DevSetColor(HY.bg);
                DevText(buf,HY.x + realcol*currentgraph->fontwidth,HY.y);
            }
            break;
        }
        if (HY.txt[i].type == HY_TEXT) {
            buf[k] = HY.txt[i].c;
            k++;
        }
        else {
            if (k) {
                buf[k] = '\0';
                if (draw)
                    DevSetColor(HY.fg);
                else
                    DevSetColor(HY.bg);
                DevText(buf,HY.x + realcol*currentgraph->fontwidth,HY.y);
                realcol += k;
                k = 0;
            }
            if (draw)
                DevSetColor(HY.hc);
            else
                DevSetColor(HY.bg);
            DevText(HY.txt[i].str,
                HY.x + realcol*currentgraph->fontwidth,HY.y);
            realcol += strlen(HY.txt[i].str);
        }
        i++;
    }
}


static void
hycursor(draw)

/* erase or draw the cursor */
int draw;
{
    int i, x, y, w, realcol;

    realcol = 0;
    for (i = 0; i < HY.col; i++) {
        if (HY.txt[i].type == HY_END) {
            realcol++;
            break;
        }
        if (HY.txt[i].type == HY_TEXT)
            realcol++;
        else
            realcol += strlen(HY.txt[i].str);
    }
    x = HY.x + realcol * currentgraph->fontwidth;
    y = HY.y;
    w = HY.cwid * currentgraph->fontwidth;
    if (draw)
        DevSetColor(HY.cc);
    else
        DevSetColor(HY.bg);
    DevBox(x,y,x + w,y + CURHT);
}


static struct hprlist *
hylist()

/* return hprlist from input queue */
{
    struct hprlist *h, *h0 = NULL;
    int i, k;
    char buf[BSIZE_SP];

    k = 0;
    for (i = 0; ; i++) {

        if (HY.txt[i].type == HY_END || i == 255) {
            if (k) {
                buf[k] = '\0';
                if (!h0)
                    h0 = h = alloc(hprlist);
                else {
                    h->hlNext = alloc(hprlist);
                    h = h->hlNext;
                }
                h->hlRefType = HY_TEXT;
                h->hlText = copy(buf);
            }
            if (i == 0) {
                /* No text, return something to distinguish
                 * from ESC.
                 */
                h0 = alloc(hprlist);
                h0->hlRefType = HY_TEXT;
                h0->hlText = copy("");
            }
            break;
        }
        if (HY.txt[i].type == HY_TEXT) {
            buf[k] = HY.txt[i].c;
            k++;
        }
        else {
            if (k) {
                buf[k] = '\0';
                k = 0;
                if (!h0)
                    h0 = h = alloc(hprlist);
                else {
                    h->hlNext = alloc(hprlist);
                    h = h->hlNext;
                }
                h->hlRefType = HY_TEXT;
                h->hlText = copy(buf);
            }
            if (!h0)
                h0 = h = alloc(hprlist);
            else {
                h->hlNext = alloc(hprlist);
                h = h->hlNext;
            }
            h->hlRefType = HY.txt[i].type;
            h->hlEnt = HYcopyEnt(HY.txt[i].ent);
            HYadd(h->hlEnt);
        }
    }
    hydelcol(0,256);
    return (h0);
}


static struct sHtxt *
input()

/* get character or point return */
{
    static struct sHtxt hret;
    REQUEST request;
    RESPONSE response;
    char *GPstring();
    struct hyEnt *ent;

    request.option = point_option;
    request.fp = (FILE*)0;
    for (;;) {
        DevInput(&request, &response);
        if (response.option == char_option) {
            hret.c = response.reply.ch;
            hret.type = HY_TEXT;
            hret.ent = NULL;
            hret.str = NULL;
            return (&hret);
        }
        if (currentgraph && currentgraph->graphtype == GR_SCED) {
            ButtonPress(response.reply.button,response.x,response.y);
            /* if button1, get string from sced */
            if (response.reply.button == 1) {
                hret.str = GPstring(HY_ALL,&ent);
                if (!hret.str)
                    continue;
                hret.ent = ent;
                hret.type = ent->hyRefType;
                hret.c = 0;
                return (&hret);
            }
        }
    }
}


/***********************************************************************
 *
 * Routines for updating saved hypertext references if the
 * underlying element is moved, and otherwise maintaining
 * the table of references.
 *
 ***********************************************************************/


void
HYinit()

/* Make sure all references have a device pointer attached.
 * Called after entire circuit has been read in.
 */
{
    int i;
    struct hyEnt **hent;
    char *s;

    if (hyBase == NULL)
        return;

    for (hent = hyBase; *hent; hent++) {
        if ((*hent)->hyPointer == NULL && (
                (*hent)->hyRefType == HY_NODE ||
                (*hent)->hyRefType == HY_BRAN ||
                (*hent)->hyRefType == HY_DEVN)) {
            s = GPgetString(*hent);
            txfree(s);
        }
    }
}


void
HYadd(ent)

/* add ent to the hypertext database */
struct hyEnt *ent;
{
    int i;

    if (hyBase == NULL) {
        /* DEC barfs
        hyBase = (struct hyEnt **)
        */
        Parameters.kpCellDesc->sHY =

            tmalloc(2*sizeof(struct hyEnt*));
        hyBase[0] = ent;
        hyBase[1] = NULL;
        return;
    }
    for (i = 0; hyBase[i]; i++) {
        if (hyBase[i] == ent)
            return;
    }

    /* DEC barfs
    hyBase = (struct hyEnt**)
    */
    Parameters.kpCellDesc->sHY =

        trealloc((char*)hyBase,(i+2)*sizeof(struct hyEnt*));
    hyBase[i] = ent;
    hyBase[i+1] = NULL;
}


void
HYdel(ent)

/* delete ent from the hypertext database */
struct hyEnt *ent;
{
    int i;

    if (hyBase == NULL)
        return;
    for (i = 0; hyBase[i]; i++) {
        if (hyBase[i] == ent)
            break;
    }
    if (hyBase[i] == NULL)
        return;
    for ( ; hyBase[i]; i++)
        hyBase[i] = hyBase[i+1];
}


void
HYtransform(Pointer,NewPointer)

/* Fix references attached to objects being moved.  Note that this only
 * has to be done for top level references, as the parent list supplies
 * the transforms for lower level references.  Set a flag so that this
 * can be undone.
 */
struct o *Pointer,*NewPointer;
{
    struct hyEnt **hent;
    long X,Y,X1,Y1;

    if (!hyBase)
        return;

    for (hent = hyBase; *hent; hent++) {

        if (Pointer == (*hent)->hyPointer) {
            (*hent)->hyOldX = (*hent)->hyX;
            (*hent)->hyOldY = (*hent)->hyY;
            (*hent)->hyOldPointer = (*hent)->hyPointer;
            TPoint(&((*hent)->hyX),&((*hent)->hyY));
            (*hent)->hyPointer = NewPointer;
            (*hent)->hyUndo = True;

            switch ((*hent)->hyOrient) {
                default:
                case MARK_NONE:
                    continue;
                case MARK_DN:
                    X = 0;
                    Y = -1;
                    break;
                case MARK_RT:
                    X = 1;
                    Y = 0;
                    break;
                case MARK_UP:
                    X = 0;
                    Y = 1;
                    break;
                case MARK_LT:
                    X = -1;
                    Y = 0;
                    break;
            }
            X1 = 0;
            Y1 = 0;
            TPoint(&X,&Y);
            TPoint(&X1,&Y1);
            X -= X1;
            Y -= Y1;
            (*hent)->hyOldOrient = (*hent)->hyOrient;

            if (X == 0) {
                if (Y == 1)
                    (*hent)->hyOrient = MARK_UP;
                else
                    (*hent)->hyOrient = MARK_DN;
            }
            else {
                if (X == 1)
                    (*hent)->hyOrient = MARK_RT;
                else
                    (*hent)->hyOrient = MARK_LT;
            }
        }
    }
}


void
HYtransformStretch(Pointer,NewPointer,RefX,RefY,NewX,NewY)

/* Fix references attached to wires that are being stretched. */
struct o *Pointer,*NewPointer;
long RefX,RefY,NewX,NewY;
{
    struct hyEnt **hent;
    struct p *p,*q,*ptmp;
    int i;
    double x1,y1,d,d0;

    if (Pointer->oType != CDWIRE)
        return;
    if (!hyBase)
        return;

    for (hent = hyBase; *hent; hent++) {

        if ((*hent)->hyPointer != Pointer)
            continue;
        p = ((struct w *)Pointer->oRep)->wPath;
        for (; p; p = q) {
            q = p->pSucc;
            if (q == NULL) break;
            ptmp = q->pSucc;
            q->pSucc = NULL;
            if (InPath(20,p,(*hent)->hyX,(*hent)->hyY) == NULL) {
                q->pSucc = ptmp;
                continue;
            }
            if (RefX == p->pX && RefY == p->pY) {
                x1 = q->pX - (*hent)->hyX;
                y1 = q->pY - (*hent)->hyY;
            }
            else if (RefX == q->pX && RefY == q->pY) {
                x1 = (*hent)->hyX - p->pX;
                y1 = (*hent)->hyY - p->pY;
            }
            else {
                q->pSucc = ptmp;
                return;
            }
            d  = x1*x1 + y1*y1;
            x1 = q->pX - p->pX;
            y1 = q->pY - p->pY;
            d0 = x1*x1 + y1*y1;
            d = sqrt(d/d0);
            (*hent)->hyOldX = (*hent)->hyX;
            (*hent)->hyOldY = (*hent)->hyY;
            (*hent)->hyOldPointer = (*hent)->hyPointer;
            (*hent)->hyX += (NewX-RefX)*d;
            (*hent)->hyY += (NewY-RefY)*d;
            (*hent)->hyPointer = NewPointer;
            (*hent)->hyUndo = True;
            q->pSucc = ptmp;
        }
    }
}


void
HYundoTransform()

{
    struct hyEnt **hent;

    if (!hyBase)
        return;

    for (hent = hyBase; *hent; hent++) {

        if ((*hent)->hyUndo) {
            (*hent)->hyX = (*hent)->hyOldX;
            (*hent)->hyY = (*hent)->hyOldY;
            (*hent)->hyPointer = (*hent)->hyOldPointer;
            (*hent)->hyOrient = (*hent)->hyOldOrient;
            (*hent)->hyUndo = False;
        }
    }
}


void
HYclearUndoFlags()

{
    struct hyEnt **hent;

    if (!hyBase)
        return;

    for (hent = hyBase; *hent; hent++)
        (*hent)->hyUndo = False;
}


void
HYdeleteReference(Pointer)

/* The object referenced by Pointer is being deleted.
 * Null the references.
 */
struct o *Pointer;
{
    struct parent *p, *pnext;
    struct hyEnt **hent, **h0;

    if (!hyBase)
        return;

    for (hent = hyBase; *hent; hent = h0) {
        h0 = hent + 1;

        if ((*hent)->hyPointer != Pointer) {

            for (p = (*hent)->hyParent; p; p = p->pNext)
                if (p->pPointer == Pointer)
                    break;
            if (!p)
                continue;
        }
        for (p = (*hent)->hyParent; p; p = pnext) {
            pnext = p->pNext;
            free(p);
        }
        (*hent)->hyParent = NULL;
        (*hent)->hyRefType = HY_BOGUS;
        (*hent)->hyPointer = NULL;
        (*hent)->hyOldPointer = NULL;
        (*hent)->hyOrient = MARK_NONE;
        (*hent)->hyUndo = False;
        if (GPdeleteGrafRef(*hent)) {
            /* *hent was freed */
            for ( ; *hent; hent++)
                *hent = *(hent+1);
            h0--;
        }
        /* *hent references nothing, user has to fix */
    }
}
