/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Initialize SCED. 
 */

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"

struct kv *View;
struct ka MenuViewport;
struct kp Parameters;
struct kl *ColorTable;
struct kc SCursor;
struct a  CurrentAOI;
char TypeOut[200];

#ifdef __STDC__
static void correct_cell_view(long*,long*,long*,long*);
static int  get_window_coords(long*,long*,int);
#else
static void correct_cell_view();
static int  get_window_coords();
#endif


char *
Init()

{
    GRAPH *graph;

    /*
     * Initialize for default values
     */
    if (ColorTable == NULL) {

        Parameters.kpGrid = 2*RESOLUTION;

        Parameters.kpGridDisplayed = True;
        Parameters.kpGridOnTop = False;
        Parameters.kpShowGridInLargeViewport = False;
        Parameters.kpGridLineStyle = 0xcc;
        Parameters.kpCursorShape = -1;

        ColorTable = (struct kl *)malloc(16*sizeof(struct kl));
        if (ColorTable == NULL)
            MallocFailed();

        ColorTable[BackgroundColor].klR = 0;
        ColorTable[BackgroundColor].klG = 0;
        ColorTable[BackgroundColor].klB = 0;

        ColorTable[HighlightingColor].klR = 255;
        ColorTable[HighlightingColor].klG = 255;
        ColorTable[HighlightingColor].klB = 255;

        ColorTable[MenuTextColor].klR = 175;
        ColorTable[MenuTextColor].klG = 200;
        ColorTable[MenuTextColor].klB = 210;

        ColorTable[MenuTextPrefixColor].klR = 125;
        ColorTable[MenuTextPrefixColor].klG = 255;
        ColorTable[MenuTextPrefixColor].klB = 85;

        ColorTable[MenuSelectedColor].klR = 170;
        ColorTable[MenuSelectedColor].klG = 210;
        ColorTable[MenuSelectedColor].klB = 200;

        ColorTable[MenuSelectedPrefixColor].klR = 240;
        ColorTable[MenuSelectedPrefixColor].klG = 240;
        ColorTable[MenuSelectedPrefixColor].klB = 155;

        ColorTable[MenuHighlightingColor].klR = 50;
        ColorTable[MenuHighlightingColor].klG = 90;
        ColorTable[MenuHighlightingColor].klB = 200;

        ColorTable[PromptTextColor].klR = 255;
        ColorTable[PromptTextColor].klG = 230;
        ColorTable[PromptTextColor].klB = 145;

        ColorTable[MoreTextColor].klR = 220;
        ColorTable[MoreTextColor].klG = 210;
        ColorTable[MoreTextColor].klB = 215;

        ColorTable[CoarseGridColor].klR = 150;
        ColorTable[CoarseGridColor].klG = 205;
        ColorTable[CoarseGridColor].klB = 245;

        ColorTable[FineGridColor].klR = 65;
        ColorTable[FineGridColor].klG = 125;
        ColorTable[FineGridColor].klB = 205;

        ColorTable[InstanceBBColor].klR = 200;
        ColorTable[InstanceBBColor].klG = 200;
        ColorTable[InstanceBBColor].klB = 200;

        ColorTable[InstanceNameColor].klR = 240;
        ColorTable[InstanceNameColor].klG = 240;
        ColorTable[InstanceNameColor].klB = 155;

        ColorTable[ExtraColor1].klR = 255;
        ColorTable[ExtraColor1].klG = 255;
        ColorTable[ExtraColor1].klB = 255;

        ColorTable[DrawingColor].klR = 180;
        ColorTable[DrawingColor].klG = 60;
        ColorTable[DrawingColor].klB = 0;

        ColorTable[ExtraColor2].klR = 255;
        ColorTable[ExtraColor2].klG = 255;
        ColorTable[ExtraColor2].klB = 255;


        if (CDInit()) {
            /* can't call MallocFailed() yet */
            fprintf(stderr,"CD memory allocation failure.\n");
            return (NULL);
        }
        Readin();
        CDSetLayer(1,'S',"CED");
    }

    if (!(graph = NewGraph())) {
        return (NULL);
    }
    graph->graphtype = GR_SCED;

    if (DevNewViewport(graph) == 1) {
        DestroyGraph(graph->graphid);
        return (NULL);
    }
    /* don't draw until later */
    graph->redraw = RedisplaySCED;

    /* remember to pop on leaving sced */
    PushGraphContext(graph);

    DevClear();

    InitParameters();
    InitSignals();
    return ((char*)graph);
}


void
InitVLT()

{
    int i;
    double R,G,B;
    double M = FBMAXINTENSITY;

    for (i = 0; i < 16; i++) {
        R = ColorTable[i].klR/M;
        G = ColorTable[i].klG/M;
        B = ColorTable[i].klB/M;
        DevDefineColor(i,R,G,B);
    }
}


void
InitParameters()

{
    if (Parameters.kpCellName == NULL) {
        Parameters.kpCellName = malloc(80);
        if (Parameters.kpCellName == NULL) MallocFailed();
        Parameters.kpTopName = malloc(80);
        if (Parameters.kpTopName == NULL) MallocFailed();
        Parameters.kpCommand = malloc(80);
        if (Parameters.kpCommand == NULL) MallocFailed();
        Parameters.kpTopDesc                  = NULL;
        Parameters.kpCellDesc                 = NULL;
        Parameters.kpEnableSelectQRedisplay   = True;
        Parameters.kpRotationAngle            = 0;
        Parameters.kpSelectTypes[0]           = CDSYMBOLCALL;
        Parameters.kpSelectTypes[1]           = CDPOLYGON;
        Parameters.kpSelectTypes[2]           = CDROUNDFLASH;
        Parameters.kpSelectTypes[3]           = CDLABEL;
        Parameters.kpSelectTypes[4]           = CDWIRE;
        Parameters.kpSelectTypes[5]           = CDBOX;
        Parameters.kpSelectTypes[6]           = EOS;
        Parameters.kpNumX                     = 1;
        Parameters.kpNumY                     = 1;
        Parameters.kpDX                       = 0;
        Parameters.kpDY                       = 0;
        Parameters.kpLastCursorColumn         = 1;

        /*
         * The fine window has a FIXED width and height. 
         * What should it be?
         * PointingThreshold is the minimum value of ViewportWidth/WindowWidth
         * such that it is still comfortable to point with lambda precision.
         * If PointingThreshold == 7 and FineViewportWidth ~ 467 pixels, 
         * FineWindowWidth ~ 60 lambda.
         * Most people have prefered 6 or 7.
         */
        Parameters.kpPointingThreshold = 6;

        View = alloc(kv);
        if (View == NULL) MallocFailed();
        View->kvFineWindow = alloc(ka);
        if (View->kvFineWindow == NULL) MallocFailed();
        View->kvFineViewport = alloc(ka);
        if (View->kvFineViewport == NULL) MallocFailed();
        View->kvLargeCoarseViewport = alloc(ka);
        if (View->kvLargeCoarseViewport == NULL) MallocFailed();
        View->kvSmallCoarseViewport = alloc(ka);
        if (View->kvSmallCoarseViewport == NULL) MallocFailed();
        View->kvCoarseWindow = alloc(ka);
        if (View->kvCoarseWindow == NULL) MallocFailed();

        Parameters.kpModified = False;
        View->kvFineViewportOnBottom = True;
    }    
    SCursor.kcX = SCursor.kcPredX = 0;
    SCursor.kcY = SCursor.kcPredY = 0;
    /* can start out in FULLSCREEN or SPLITSCREEN mode */
    View->kvControl = FULLSCREEN;
    TInit();
    TStore();
    InitViewport();
    DefaultWindows();
    SQInit();
}


void
DefaultWindows()

{
    int DefWidth;

    View->kvCoarseViewport = View->kvLargeCoarseViewport;
    if (View->kvControl == FULLSCREEN)
        DefWidth = 100;
    else
        DefWidth = 1000;
    InitCoarseWindow((long)0,(long)0,(long)DefWidth*RESOLUTION);

    /* default fine window */
    View->kvFineWindow->kaWidth = View->kvFineViewport->kaWidth/
        Parameters.kpPointingThreshold*RESOLUTION;

    View->kvFineWindow->kaHeight = View->kvFineWindow->kaWidth*
        View->kvFineViewport->kaHeight/View->kvFineViewport->kaWidth;


    SetPositioning();
    SaveLastView();
}


void
InitCoarseWindow(X,Y,Width)

long X,Y,Width;
{
    long diff;
    int Layer;
    double Wid2;
    struct ka *Cw;

    Wid2 = Width/2;
    diff = (long)((Wid2*View->kvCoarseViewport->kaHeight)/
        View->kvCoarseViewport->kaWidth);

    Cw = View->kvCoarseWindow;
    Cw->kaX      = X;
    Cw->kaY      = Y; 
    Cw->kaLeft   = Cw->kaX - Wid2; 
    Cw->kaRight  = Cw->kaX + Wid2; 
    Cw->kaBottom = Cw->kaY - diff;
    Cw->kaTop    = Cw->kaY + diff;
    Cw->kaWidth  = Cw->kaRight - Cw->kaLeft;
    Cw->kaHeight = Cw->kaTop - Cw->kaBottom;

    View->kvCoarseRatio = View->kvCoarseViewport->kaWidth/Cw->kaWidth;
}


void
InitFineWindow(X,Y)

long X,Y;
{
    /* Width and height are fixed. */
    long diff;
    int Layer;
    double Wid2;
    struct ka *Fw;

    Wid2 = View->kvFineWindow->kaWidth/2;
    diff = (long)((Wid2*View->kvFineViewport->kaHeight)/
        View->kvFineViewport->kaWidth);

    Fw = View->kvFineWindow;
    Fw->kaX      = X;
    Fw->kaY      = Y;
    Fw->kaLeft   = Fw->kaX - Wid2; 
    Fw->kaRight  = Fw->kaX + Wid2; 
    Fw->kaBottom = Fw->kaY - diff;
    Fw->kaTop    = Fw->kaY + diff; 
    Fw->kaWidth  = Fw->kaRight - Fw->kaLeft;
    Fw->kaHeight = Fw->kaTop - Fw->kaBottom;

    View->kvFineRatio = View->kvFineViewport->kaWidth/Fw->kaWidth;
}


void
SetPositioning()

{
    long X,Y,Width;

    if (View->kvControl != FULLSCREEN And
        View->kvCoarseRatio < Parameters.kpPointingThreshold) {
        /*
         * Split--enable fine positioning.
         */
        Parameters.kpRedisplayControl = SPLITSCREEN;
        View->kvCoarseViewport = View->kvSmallCoarseViewport;
    }
    else {
        /*
         * Don't split--don't enable fine positioning.
         */
        Parameters.kpRedisplayControl = COARSEVIEWPORTONLY;
        View->kvCoarseViewport = View->kvLargeCoarseViewport;
    }
    Width = View->kvCoarseWindow->kaWidth;
    X     = View->kvCoarseWindow->kaX;
    Y     = View->kvCoarseWindow->kaY;
    InitCoarseWindow(X,Y,Width);

    /* it is enough to just check the width */
    if (Width <= View->kvFineWindow->kaWidth) {
        View->kvFineWindow->kaWidth = Width - Width/4;
        View->kvFineWindow->kaHeight = View->kvFineWindow->kaWidth*
            (View->kvFineViewport->kaHeight/View->kvFineWindow->kaWidth);
    }
    InitFineWindow(X,Y);
}


void
InitViewport()

{
    struct ka *Vl, *Vs, *Vf;

    MenuViewport.kaLeft         = 1;
    MenuViewport.kaBottom       = gi_numtextrows - 1;
    MenuViewport.kaRight        = 11;
    MenuViewport.kaTop          = 1;
    /* save column width and height in X,Y integer fields */
    MenuViewport.kaX            = 5;
    MenuViewport.kaY            = MenuViewport.kaBottom -
                                      MenuViewport.kaTop + 1;

    /*
     * The COARSE viewport is LARGE if fine-positioning isn't enabled.
     * If it is, the LARGE COARSE viewport is split into a SMALL COARSE 
     * viewport and the FINE viewport.
     */
    if (View->kvFineViewportOnBottom) {

        Vl = View->kvLargeCoarseViewport;
        Vl->kaLeft   = gi_fntwidth*MenuViewport.kaRight + gi_fntwidth/2;
        Vl->kaBottom = 2*gi_fntheight;
        Vl->kaRight  = gi_maxx;
        Vl->kaTop    = gi_maxy;
        Vl->kaWidth  = Vl->kaRight - Vl->kaLeft;
        Vl->kaHeight = Vl->kaTop - Vl->kaBottom;

        Vs = View->kvSmallCoarseViewport;
        Vs->kaLeft   = Vl->kaLeft; 
        Vs->kaBottom = Vl->kaBottom + ((Vl->kaTop - Vl->kaBottom)*5)/12;
        Vs->kaRight  = Vl->kaRight; 
        Vs->kaTop    = Vl->kaTop;
        Vs->kaWidth  = Vs->kaRight - Vs->kaLeft;
        Vs->kaHeight = Vs->kaTop - Vs->kaBottom;
    
        Vf = View->kvFineViewport;
        Vf->kaLeft   = Vs->kaLeft; 
        Vf->kaBottom = Vl->kaBottom; 
        Vf->kaRight  = Vs->kaRight; 
        Vf->kaTop    = Vs->kaBottom - 1;
        Vf->kaWidth  = Vf->kaRight - Vf->kaLeft;
        Vf->kaHeight = Vf->kaTop - Vf->kaBottom;
    }
    else {

        Vl = View->kvLargeCoarseViewport;
        Vl->kaLeft   = gi_fntwidth*MenuViewport.kaRight + gi_fntwidth/2;
        Vl->kaBottom = 2*gi_fntheight;
        Vl->kaRight  = gi_maxx;
        Vl->kaTop    = gi_maxy;
        Vl->kaWidth  = Vl->kaRight - Vl->kaLeft;
        Vl->kaHeight = Vl->kaTop - Vl->kaBottom;

        Vs = View->kvSmallCoarseViewport;
        Vs->kaLeft   = Vl->kaLeft; 
        Vs->kaBottom = Vl->kaBottom;
        Vs->kaRight  = (gi_maxx - Vl->kaLeft)/2 + Vl->kaLeft;
        Vs->kaTop    = Vl->kaTop; 
        Vs->kaWidth  = Vs->kaRight - Vs->kaLeft;
        Vs->kaHeight = Vs->kaTop - Vs->kaBottom;
      
        Vf = View->kvFineViewport;
        Vf->kaLeft   = Vs->kaRight + 1; 
        Vf->kaBottom = Vl->kaBottom; 
        Vf->kaRight  = Vl->kaRight; 
        Vf->kaTop    = Vs->kaTop;
        Vf->kaWidth  = Vf->kaRight - Vf->kaLeft;
        Vf->kaHeight = Vf->kaTop - Vf->kaBottom;
    }
}


void
SetCurrentAOI(Window)

struct ka *Window;
{
    CoarseLToP(Window->kaLeft,Window->kaBottom,CurrentAOI.aLC,CurrentAOI.aBC);
    CoarseLToP(Window->kaRight,Window->kaTop,CurrentAOI.aRC,CurrentAOI.aTC);
    if (CurrentAOI.aLC <= View->kvCoarseViewport->kaRight And
        CurrentAOI.aRC >= View->kvCoarseViewport->kaLeft And
        CurrentAOI.aBC <= View->kvCoarseViewport->kaTop And
        CurrentAOI.aTC >= View->kvCoarseViewport->kaBottom) {
        ClipVP(View->kvCoarseViewport,CurrentAOI.aLC,CurrentAOI.aBC);
        ClipVP(View->kvCoarseViewport,CurrentAOI.aRC,CurrentAOI.aTC);
        CurrentAOI.aInCoarse = True;
    }
    else
        CurrentAOI.aInCoarse = False;

    FineLToP(Window->kaLeft,Window->kaBottom,CurrentAOI.aLF,CurrentAOI.aBF);
    FineLToP(Window->kaRight,Window->kaTop,CurrentAOI.aRF,CurrentAOI.aTF);
    if (CurrentAOI.aLF <= View->kvFineViewport->kaRight And
        CurrentAOI.aRF >= View->kvFineViewport->kaLeft And
        CurrentAOI.aBF <= View->kvFineViewport->kaTop And
        CurrentAOI.aTF >= View->kvFineViewport->kaBottom) {
        ClipVP(View->kvFineViewport,CurrentAOI.aLF,CurrentAOI.aBF);
        ClipVP(View->kvFineViewport,CurrentAOI.aRF,CurrentAOI.aTF);
        CurrentAOI.aInFine = True;
    }
    else
        CurrentAOI.aInFine = False;
}


void
CenterFullView()

{
    long L,B,R,T;
    long X,Y,Width,Height;
    double Cratio,Vratio;

    if (Not CDBB(Parameters.kpCellDesc,(struct o *)NULL,&L,&B,&R,&T))
        MallocFailed();
    /*
     * Window cell automatically.  How slick.
     * Leave some white space around the perimeter.
     * Check for a null (empty) cell.
     */
    if (L == R Or B == T Or (L == CDINFINITY And B == CDINFINITY)) {
        DefaultWindows();
        return;
    }
    Vratio = View->kvCoarseViewport->kaWidth/View->kvCoarseViewport->kaHeight;

    Width = R - L;
    Height = T - B;
    X = L + Width/2;
    Y = B + Height/2;
    Cratio = (double) Width/Height;

    if (Cratio < Vratio)
        Width = Height*Vratio;
    else
        Height = Width/Vratio;

    InitCoarseWindow(X,Y,Width);
    L = X - Width/2;
    R = X + Width/2;
    B = Y - Height/2;
    T = Y + Height/2;
/*
    correct_cell_view(&L,&B,&R,&T);
*/
    
    Width = R - L;
    Height = T - B;
    X = L + Width/2;
    Y = B + Height/2;
    Cratio = (double) Width/Height;
    if (Cratio > Vratio)
        Width *= 1.1;
    else
        Width = 1.1*Height*Vratio;

    InitCoarseWindow(X,Y,Width);

    SetPositioning();
}


static void
correct_cell_view(L,B,R,T)

/* Take into account labels that may extend beyond geometries. */
long *L,*B,*R,*T;
{
    struct g *GenDesc;
    struct o *Pointer;
    struct ka LBB;
    int Info;
    char Type;
    long Width,Height;

    Width = *R - *L;
    Height = *T - *B;
    if (Not CDInitGen(Parameters.kpCellDesc,1,*L,*B,
        *R,*T,&GenDesc)) MallocFailed();

    loop {
        CDGen(Parameters.kpCellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;

        if (Pointer->oType != CDLABEL) continue;
        if (Pointer->oInfo == SQ_GONE) continue;

        BBLabel(View->kvCoarseWindow,Pointer,&LBB);

        if (LBB.kaLeft < *L) { 
            Width *= (double)(LBB.kaRight - LBB.kaLeft)/(LBB.kaRight - *L);
            *L = *R - Width;
        }
        if (LBB.kaRight > *R) {
            Width *= (double)(LBB.kaRight - LBB.kaLeft)/(*R - LBB.kaLeft);
            *R =  *L + Width;
        }
        if (LBB.kaBottom < *B) {
            Height *= (double)(LBB.kaTop - LBB.kaBottom)/(LBB.kaTop - *B);
            *B = *T - Height;
        }
        if (LBB.kaTop > *T) {
            Height *= (double)(LBB.kaTop - LBB.kaBottom)/(*T - LBB.kaBottom);
            *T = *B + Height;
        }
    }
}


void
UpdateSCED()

{
    FILE *fp;
    int i;
    char *c;
    extern char *MenuUPDAT;

    MenuSelect(MenuUPDAT);
    fp = fopen(INIT_FILE,"w");
    if (fp == NULL) {
        ShowPrompt("Can't open file.");
        return;
    }
    fprintf(fp,"SCED init file\n");
    for (i = 0; i < 16; i++) {
        fprintf(fp,"%d %3d %3d %3d\n",i,
            ColorTable[i].klR,ColorTable[i].klG,ColorTable[i].klB);
    }
    fprintf(fp,"%d %d %d %d\n",Parameters.kpGridDisplayed,
        Parameters.kpGridOnTop,
        Parameters.kpShowGridInLargeViewport,
        Parameters.kpGridLineStyle);
    c = PGetPath();
    if (c && *c)
        fprintf(fp,"%s\n",c);
    fclose(fp);
    ShowPrompt("Init file has been saved.");
    MenuDeselect(MenuUPDAT);
}


void
Readin()

{
    FILE *fp;
    int i,d1,d2,d3,d4;

    fp = POpen(INIT_FILE,"r",(char**)NULL);
    if (fp == NULL)
        return;
    (void)fgets(TypeOut,80,fp);
    if (TypeOut[0] != 'S'|| TypeOut[1] != 'C')
        goto error;
    for (i = 0; i < 16; i++) {
        if (fscanf(fp,"%d %d %d %d\n",&d1,&d2,&d3,&d4) != 4)
            goto error;
        if (d1 < 0 || d1 > 15)
            goto  error;
        if (d2 < 0 || d2 > 255)
            goto error;
        if (d3 < 0 || d3 > 255)
            goto error;
        if (d4 < 0 || d4 > 255)
            goto error;
        ColorTable[d1].klR = d2;
        ColorTable[d1].klG = d3;
        ColorTable[d1].klB = d4;
    }
    if (fscanf(fp,"%d %d %d %d\n",&d1,&d2,&d3,&d4) != 4)
        goto error;
    Parameters.kpGridDisplayed = d1;
    Parameters.kpGridOnTop = d2;
    Parameters.kpShowGridInLargeViewport= d3;
    Parameters.kpGridLineStyle = d4;
    if (fgets(TypeOut,80,fp) != NULL)
        SetCDPath(TypeOut);
error:
    fclose(fp);
    return;
}


void
RedisplaySCED(graph)

/* Redisplay what is on the screen.
 * This is called after a resizing, or after the screen has been
 * seriously messed with.  We reinititialize everything, as the
 * parameters may have changed.  RedrawPrompt should take care of
 * anything on the prompt line, including a return from FBGetchar()
 * in the DISPLAY mode.  KbRepaint() redraws the editing input
 * if there was any.
 */
GRAPH* graph;
{
    /* establish current graph so default graphic calls will work right */
    PushGraphContext(graph);

    DevClear();

    InitViewport();
    InitCoarseWindow(View->kvCoarseWindow->kaX,View->kvCoarseWindow->kaY,
        (long)View->kvCoarseWindow->kaWidth);
    InitFineWindow(View->kvFineWindow->kaX,View->kvFineWindow->kaY);
    InitVLT();
    ShowCommandMenu();
    if (!RepaintMore())
        Redisplay(View->kvCoarseWindow);
    RedrawPrompt();
    KbRepaint(Parameters.kpLastCursorColumn*gi_fntwidth,
        gi_maxy - gi_fntheight*gi_numtextrows);
    PopGraphContext();
}


/* ARGSUSED */
void
ResizeSCED(graph)

GRAPH* graph;
{
    /* do nothing */
}



/********************************************************************
 *
 * Some rendering interfaces
 *
 ********************************************************************/

#ifdef __STDC__
static void ghost_line(int,int,int,int);
static void ghost_line_snap(int,int,int,int);
static void ghost_box(int,int,int,int);
static void ghost_box_snap(int,int,int,int);
static void ghost_wires(int,int,int,int);
static void ghost_move(int,int,int,int);
static void ghost_place(int,int,int,int);
#else
static void ghost_line();
static void ghost_line_snap();
static void ghost_box();
static void ghost_box_snap();
static void ghost_wires();
static void ghost_move();
static void ghost_place();
#endif



short
FBGetchar(DisplayOrErase)

int DisplayOrErase;
{
    int x,y;
    short c;
    char s[2];

    if (DisplayOrErase == ERASE)
        return ((short)DevGetchar(NULL));

    y = gi_maxy - gi_numtextrows*gi_fntheight;
    x = Parameters.kpLastCursorColumn*gi_fntwidth;
    DevSetColor(HighlightingColor);
    KbCursor(x,y);
    c = (short)DevGetchar(NULL);
    s[0] = (char)(c & 0xff);
    s[1] = '\0';
    DevSetColor(PromptTextColor);
    DevText(s,x,y);
    AppendToOldPrompt(s[0]);
    DevSetColor(BackgroundColor);
    KbCursor(x,y);
    return (c);
}


char *
FBEdit(string)

char *string;
{
    int x,y;

    y = gi_maxy - gi_numtextrows*gi_fntheight;
    x = Parameters.kpLastCursorColumn*gi_fntwidth;
    return (KbEdit(string,x,y,BackgroundColor,PromptTextColor,
        HighlightingColor));
}


struct hprlist *
FBHyEdit(htxt)

struct hprlist *htxt;
{
    int x,y;

    y = gi_maxy - gi_numtextrows*gi_fntheight;
    x = Parameters.kpLastCursorColumn*gi_fntwidth;
    return (HYedit(htxt,x,y,BackgroundColor,PromptTextColor,
        HighlightingColor,HighlightingColor));
}


void
FBText(Mode,RowOrX,ColumnOrY,String)

char Mode;
int RowOrX,ColumnOrY;
char *String;
{
    int i;
    if(Mode == ROW_COLUMN) {
        i = gi_maxy - RowOrX * gi_fntheight;
        RowOrX = (ColumnOrY - 1) * gi_fntwidth;
        ColumnOrY = i;
    }
    DevText(String,RowOrX,ColumnOrY);
}


void
FBPolygon(Pixel,Type,xy,n)

long *xy;
int Pixel,n;
char Type;
{
    /*
     * Type == FILL means fill
     * Type == OUTLINE means outline
     */
    int j,n2,X,Y,Z,T;
    POLYGON poly;

    DevSetColor(Pixel);

    poly.xy = xy;
    while (NextPolygon(&poly)) {
        n = poly.nvertices;

        if (Type == OUTLINE) {
            n2 = n + n;
            j = 2;
            X = xy[0];
            Y = xy[1];
            while (j < n2) {
                Z = xy[j++];
                T = xy[j++];
                DevLine(X,Y,Z,T);
                X = Z;
                Y = T;
            }
        }
        else {
            if (n < 2) return;
            DevPolygon(&poly);
        }
    }
}


void
FBPolygonClip(coord,n,window)

long *coord;
int *n;
struct ka *window;
{
    POLYGON poly;

    poly.nvertices = *n;
    if (*n < 3) return;
    poly.xy = coord;
    PolygonClip(&poly,window->kaLeft,window->kaBottom,
        window->kaRight,window->kaTop);
}


/* Warning: 16 bit ints will not work below */
void
FBSetRubberBanding(mode)

/* if mode is lower case will clip to grid point */
char mode;
{
    /* need some hokum for button 3 box */
#ifdef __STDC__
    static void (*func)(int,int,int,int);
#else
    static void (*func)();
#endif
    static int X, Y;
    static int lastR;

    switch (mode) {
        case 'l':
            func = ghost_line_snap;
            X = (int)SCursor.kcX;
            Y = (int)SCursor.kcY;
            DevSetGhost(func,X,Y);
            return;

        case 'L':
            func = ghost_line;
            X = (int)SCursor.kcRawX;
            Y = (int)SCursor.kcRawY;
            DevSetGhost(func,X,Y);
            return;

        case 'r':
            func = ghost_box_snap;
            X = (int)SCursor.kcX;
            Y = (int)SCursor.kcY;
            DevSetGhost(func,X,Y);
            return;

        case 'R':
            if (func) {
                lastR = True;
                DevSetGhost(ghost_box,
                    (int)SCursor.kcRawX,(int)SCursor.kcRawY);
                return;
            }
            func = ghost_box;
            X = (int)SCursor.kcRawX;
            Y = (int)SCursor.kcRawY;
            DevSetGhost(func,X,Y);
            return;

        case 's':
            func = ghost_wires;
            X = (int)SCursor.kcX;
            Y = (int)SCursor.kcY;
            DevSetGhost(func,X,Y);
            return;

        case 'm':
            func = ghost_move;
            X = (int)SCursor.kcX;
            Y = (int)SCursor.kcY;
            DevSetGhost(func,X,Y);
            return;

        case 'p':
            func = ghost_place;
            X = (int)SCursor.kcX;
            Y = (int)SCursor.kcY;
            DevSetGhost(func,X,Y);
            return;

        case 0:
            if (lastR) {
                DevSetGhost(func,X,Y);
            }
            else {
                DevSetGhost(NULL,0,0);
#ifdef __STDC__
                func = (void(*)(int,int,int,int))NULL;
#else
                func = (void(*)())NULL;
#endif
            }
            lastR = False;
            return;
    }
}


static void
ghost_line(x,y,refx,refy)

int x, y, refx, refy;
{
    if (!get_window_coords((long*)&x,(long*)&y,False))
        return;
    ShowLine(HighlightingColor,refx,refy,x,y);
    DevSetColor(DrawingColor);
}


static void
ghost_line_snap(x,y,refx,refy)

int x, y, refx, refy;
{
    if (!get_window_coords((long*)&x,(long*)&y,True))
        return;
    ShowLine(HighlightingColor,refx,refy,x,y);
    DevSetColor(DrawingColor);
}


static void
ghost_box(x,y,refx,refy)

int x, y, refx, refy;
{
    struct ka BB;

    if (!get_window_coords((long*)&x,(long*)&y,False))
        return;
    BB.kaLeft = refx;
    BB.kaBottom = refy;
    BB.kaRight = x;
    BB.kaTop = y;
    ShowEmptyBox(HighlightingColor,&BB);
    DevSetColor(DrawingColor);
}


static void
ghost_box_snap(x,y,refx,refy)

int x, y, refx, refy;
{
    struct ka BB;

    if (!get_window_coords((long*)&x,(long*)&y,True))
        return;
    BB.kaLeft = refx;
    BB.kaBottom = refy;
    BB.kaRight = x;
    BB.kaTop = y;
    ShowEmptyBox(HighlightingColor,&BB);
    DevSetColor(DrawingColor);
}


static void
ghost_wires(x,y,refx,refy)

int x, y, refx, refy;
{
    if (!get_window_coords((long*)&x,(long*)&y,True))
        return;
    ShowStretch(x,y,refx,refy);
    DevSetColor(DrawingColor);
}


static void
ghost_move(x,y,refx,refy)

int x, y, refx, refy;
{
    if (!get_window_coords((long*)&x,(long*)&y,True))
        return;
    ShowGhost(x,y,refx,refy);
    DevSetColor(DrawingColor);
}


static void
ghost_place(x,y,refx,refy)

int x, y, refx, refy;
{
    if (!get_window_coords((long*)&x,(long*)&y,True))
        return;
    ShowNewInstance(x,y,refx,refy);
    DevSetColor(DrawingColor);
}


static int
get_window_coords(X,Y,Clip)

/* replace X,Y (in pixel coords) with the window coords if pointing
 * in the Coarse or Fine viewports, and return True.  Otherwise
 * return False.  Clip to grid if Clip is True.
 */
long *X,*Y;
int Clip;
{
    if (InBox(*X,*Y,View->kvCoarseViewport)) {
        PToL(View->kvCoarseWindow,X,Y);
        if (Clip)
            ClipToGridPoint(X,Y);
        return (True);
    }
    if (InBox(*X,*Y,View->kvFineViewport)) {
        PToL(View->kvFineWindow,X,Y);
        if (Clip)
            ClipToGridPoint(X,Y);
        return (True);
    }
    return (False);
}
