/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"

/***********************************************************************
 *
 * Routines for "More" text presentation.
 *
 *
 ***********************************************************************/

 
static int UsePopup;    /* don't use MORE, use a popup instead */
static int MoreOn;      /* set while in MORE mode */
static int DidInitMore; /* initialized */
static int MoreShowed;  /* nonzero if some characters in output */
static struct ka MoreBB;
static char *MoreBuffer, *MoreLinePtr;
static FILE *tmpfp;
static char *tmpfn;
 
#ifdef __STDC__
static void init_more(void);
static int  more_prompt(void);
#else
static void init_more();
static int  more_prompt();
#endif


void
MoreText(Left,Bottom,Right,Top,Textfile,color)

int Left;
int Bottom;
int Right;
int Top;
FILE *Textfile;
int color;
{
    /* this is not called under X */
    char cbuf[200];        /* 200 chars per line max */
    int c,i;
    int linecount;
    int done = 0;
    int nlines;
    int nchars;
    int controlchar;
    
    /* test to be sure of window area */
    if (Top < Bottom) {
        i = Top;
        Top = Bottom;
        Bottom = i;
    }
    if (Right < Left) {
        i = Right;
        Right = Left;
        Left = i;
    }

    Left   += 2;
    Right  -= 2;
    Bottom += 2;
    Top    -= 2;
    /* calculate parameters */
    nlines = (Top - Bottom)/gi_fntheight;
    nchars = (Right - Left)/gi_fntwidth;
    if (nchars > 200)
        nchars = 200;

    if (nlines <= 0) return;
    linecount = 1;
    DevSetColor(BackgroundColor);
    DevBox(Left - 2,Bottom - 2,Right + 2,Top + 2);
    DevSetColor(color);
    while (! done) {
        i = 0;
        controlchar = 0;
        while (i < nchars && (c = getc(Textfile)) != '\n' && c != EOF) {
            if (c == 9) { /* tab */
                cbuf[i++] = ' ';
                while (i < nchars && (i % 8)) cbuf[i++] = ' ';
            }
            else if (c < ' ') {
                if (controlchar == 0) {
                    cbuf[i++] = '^';
                    controlchar = 1;
                    ungetc(c,Textfile);
                }
                else {
                    cbuf[i++] = c + '@';
                    controlchar = 0;
                }
            }
            else if (c <= '~')
                cbuf[i++] = c;
        }
        cbuf[i] = '\0';
        if (c == EOF) done = 1;
        DevText(cbuf,Left,Top - (linecount) * gi_fntheight);
        if (done || ++linecount >= nlines) {
            linecount = 1;
            DevSetColor(MenuTextColor);
            if(done)
                DevText("-DONE-",Left,Bottom);
            else
                DevText("-MORE- (^D to exit)",Left,Bottom);
            DevSetColor(color);
            c = (char) DevGetchar(NULL);
            if(c == 4) 
                done = 1;
            DevSetColor(BackgroundColor);
            DevBox(Left - 2,Bottom - 2,Right + 2,Top + 2);
            DevSetColor(color);
        }
    }
}


int
MoreLine(line)

char *line;
{
    char *s, c, *next;
 
    if (!line) return (False);
    if (!MoreOn) return (False);
    if (UsePopup) {
        if (!tmpfp) {
            tmpfn = smktemp("sct");
            tmpfp = fopen(tmpfn, "w+");
            if (!tmpfp) {
                UsePopup = False;
                free(tmpfn);
                tmpfn = NULL;
            }
        }
        if (tmpfp) {
            if (*(line + strlen(line) - 1) != '\n')
                fprintf(tmpfp, "%s\n", line);
            else
                fprintf(tmpfp, line);
            MoreShowed = True;
            return (False);
        }
    }

    if (!DidInitMore) init_more();
    if (!DidInitMore) return (False);
 
    for (s = line; s && *s; s = next) {
        if ((next = strchr(s,'\n')) != NULL) {
            *next = '\0';
            next++;
        }
        while (strlen(s) > MoreBB.kaX) {
            c = *(s + MoreBB.kaX);
            *(s + MoreBB.kaX) = '\0';
            strcpy(MoreLinePtr,s);
            MoreLinePtr += gi_numtextcols+1;
            MoreBB.kaY -= gi_fntheight;
            if (MoreBB.kaY < MoreBB.kaBottom) {
                MoreShowed += MorePageDisplay();
                if (more_prompt()) return (True);
            }
            *(s + MoreBB.kaX) = c;
            s += MoreBB.kaX;
            if (!*s) return (False);
        }
        strcpy(MoreLinePtr,s);
        MoreLinePtr += gi_numtextcols+1;
        MoreBB.kaY -= gi_fntheight;
        if (MoreBB.kaY < MoreBB.kaBottom) {
            MoreShowed += MorePageDisplay();
            if (more_prompt()) return (True);
        }
    }
    return (False);
}


int
EnableMore(On)

int On;
{
    if (On) {
        MoreOn = True;
        if (dispdev->windows) {
            UsePopup = true;
            MoreShowed = False;
        }
    }
    else {
        MoreOn = False;
        if (!UsePopup) {
            if (!DidInitMore) return (0);
            MoreShowed += MorePageDisplay();
            DidInitMore = False;
            free(MoreBuffer);
            MoreBuffer = NULL;
        }
        else {
            MorePageDisplay();
        }
    }
    return (MoreShowed);
}


int
RepaintMore()

{
    if (!MoreOn || UsePopup)
        return (False);
    (void)MorePageDisplay();
    return (True);
}


int
MorePageDisplay()

/* returns False if nothing to print */
{
    char *s, *t;
    int Y;

    if (UsePopup) {
        if (tmpfp) {
            fflush(tmpfp);
            rewind(tmpfp);
            PopUpErrMessage(tmpfp);
            fclose(tmpfp);
            unlink(tmpfn);
            free(tmpfn);
            tmpfp = NULL;
            tmpfn = NULL;
        }
        return (MoreShowed);
    }

    if (MoreLinePtr == MoreBuffer)
        return (False);

    /* return if only blank lines */
    for (s = MoreBuffer; s < MoreLinePtr; s += gi_numtextcols+1) {
        t = s;
        while (isspace(*t)) t++;
        if (*t == '\0')
            continue;
        else
            break;
    }
    if (*t == '\0')
        return (False);

    EraseLargeCoarseViewport();
    DevSetColor(MoreTextColor);
    Y = MoreBB.kaTop - gi_fntheight;
    for (s = MoreBuffer; s < MoreLinePtr; s += gi_numtextcols+1) {
        FBText(PIXEL_COORDINATE,(int)MoreBB.kaLeft,Y,s);
        Y -= gi_fntheight;
    }
    return (True);
}


static void
init_more()

{
    int i;

    MoreBB.kaLeft = View->kvLargeCoarseViewport->kaLeft + 7;
    MoreBB.kaBottom = View->kvLargeCoarseViewport->kaBottom +
        gi_fntheight + 1;
    MoreBB.kaRight = View->kvLargeCoarseViewport->kaRight - 1;
    i = View->kvLargeCoarseViewport->kaTop - 1 - MoreBB.kaBottom;
    i /= gi_fntheight;
    MoreBB.kaTop = MoreBB.kaBottom + i*gi_fntheight;
    MoreBB.kaX = (MoreBB.kaRight - MoreBB.kaLeft)/gi_fntwidth;
    MoreBB.kaY = MoreBB.kaTop - gi_fntheight;

    if (MoreBuffer == NULL) {
        MoreBuffer = malloc(gi_numtextrows*(gi_numtextcols+1));
        if (MoreBuffer == NULL) {
            ShowPrompt("Error, Out of memory!");
            return;
        }
    }
    MoreLinePtr = MoreBuffer;
    for (i = 0; i < gi_numtextrows; i++) {
        *MoreLinePtr = '\0';
        MoreLinePtr += gi_numtextcols+1;
    }
    MoreLinePtr = MoreBuffer;
 
    DidInitMore = True;
    MoreShowed = 0;
}


static int
more_prompt()

{
    int i;
    char *s, InChar;

    DevSetColor(MenuTextColor);
    FBText(PIXEL_COORDINATE,(int)MoreBB.kaLeft,
        (int)MoreBB.kaBottom - gi_fntheight,"      --More--");
    InChar = FBGetchar(ERASE);

    MoreBB.kaY = MoreBB.kaTop - gi_fntheight;
    s = MoreBuffer;
    for (i = 0; i < gi_numtextrows; i++) {
        *s = '\0';
        s += gi_numtextcols+1;
    }
    MoreLinePtr = MoreBuffer;

    /* can't do this */
    /*
    if (InChar == ESCAPE || InChar == 'q')
        return (True);
    */
    return (False);
}
