/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * SCED displayable object management code.
 */

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"



/***********************************************************************
 *
 * Routine for deleting objects.
 * 
 *
 ***********************************************************************/

extern char *MenuDELET;
extern char *MenuUNDO;
#ifdef __STDC__
static void do_del(int,struct ka*);
#else
static void do_del();
#endif


void
Del(LookedAhead)

int *LookedAhead;
{
    struct ks *SQDesc;
    int Info = SQ_GONE;

    MenuSelect(MenuDELET);
    if (SelectQHead == NULL) {
        ShowPrompt("There are no selected objects to delete."); 
        MenuDeselect(MenuDELET);
        return;
    }
    SQComputeBB();
    /* take care of instance markers */
    SelectQBB.kaRight  += 600;
    SelectQBB.kaLeft   -= 600;
    SelectQBB.kaTop    += 600;
    SelectQBB.kaBottom -= 600;

    do_del(SQ_GONE,&SelectQBB);
    MenuDeselect(MenuDELET);

top:
    switch (PointLoop(LookedAhead)) {
        case PL_UND:
            MenuSelect(MenuUNDO);
            if (Info == SQ_GONE) Info = SQ_OLDSEL;
            else Info = SQ_GONE;
            do_del(Info,&SelectQBB);
            MenuDeselect(MenuUNDO);
        case PL_ESC:
        case PL_PCW:
            goto top;
        case PL_CMD:
            break;
    }
    if (Info == SQ_GONE) {
        for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc) {
            CDDelete(Parameters.kpCellDesc,SQDesc->ksPointer);
        }
        SQClear();
        SelectQBB.kaLeft = SelectQBB.kaBottom =
            SelectQBB.kaRight = SelectQBB.kaTop = 0;
        Parameters.kpModified = True;
    }
    else
        SQComputeBB();
}


static void
do_del(Info,BB)

int Info;
struct ka *BB;
{
    struct ks *SQDesc;
    /*
     * Change Info field and redisplay.
     */
    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc)
        SQDesc->ksPointer->oInfo = Info;

    EraseBox(BB);
    Redisplay(BB);
    DevUpdate();
}



/***********************************************************************
 *
 * Routines for managing wires.
 * 
 *
 ***********************************************************************/

extern char *MenuWIRES;
extern char *MenuUNDO;
#ifdef __STDC__
static void allocate_wire(int,struct p*,long,struct ka*,struct o**,int);
static void record_wire(struct o*);
static void zap_last_point(struct p**);
static void append_point(long,long,struct p**);
static void last_path_point(long*,long*,struct p*);
static struct p *allocate_path(long,long);
#else
static void allocate_wire();
static void record_wire();
static void zap_last_point();
static void append_point();
static void last_path_point();
static struct p *allocate_path();
#endif


void
Wires(LookedAhead)

/*  Wires command - create wires. */
int *LookedAhead;
{
    struct p *Path;
    struct ka BB,OldBB;
    struct o *Pointer = NULL,*OldPointer = NULL;
    long Width;
    long X,Y,FX,FY;
    int NumVertices = 0;
    int ExpectFirstPoint = True;
    int Modified = 0;
    int Undo = False;

    MenuSelect(MenuWIRES);
    Path = NULL;

    Width = 0;

    ShowPrompt("Point to reference points.");
    loop {
        switch (PointLoop(LookedAhead)) {
            case PL_UND:
                MenuSelect(MenuUNDO);
                if (NumVertices == 0) {
                    if (OldPointer == NULL) {
                        MenuDeselect(MenuUNDO);
                        goto quit;
                    }
                    if (Undo == False) {
                        OldPointer->oInfo = SQ_GONE;
                        Modified--;
                        Undo = True;
                    }
                    else {
                        OldPointer->oInfo = SQ_NEW;
                        Modified++;
                        Undo = False;
                    }
                    EraseBox(&OldBB);
                    Redisplay(&OldBB);
                    MenuDeselect(MenuUNDO);
                    continue;
                }
                if (NumVertices > 1)
                    zap_last_point(&Path);
                CDDelete(Parameters.kpCellDesc,Pointer);
                EraseBox(&BB);
                Redisplay(&BB);
                FBSetRubberBanding(0);
                if (NumVertices == 1) {
                    ExpectFirstPoint = True;
                }
                else {
                    long tmpX, tmpY;

                    allocate_wire(1,Path,Width,&BB,&Pointer,5);
                    ShowWire(DrawingColor,Width,Path);
                    last_path_point(&X,&Y,Path);
                    tmpX = SCursor.kcX;
                    tmpY = SCursor.kcY;
                    SCursor.kcX = X;
                    SCursor.kcY = Y;
                    FBSetRubberBanding('l');
                    SCursor.kcX = tmpX;
                    SCursor.kcY = tmpY;
                }
                NumVertices--;
                MenuDeselect(MenuUNDO);
                continue;
            case PL_ESC:
                if (Not ExpectFirstPoint) {
                    CDDelete(Parameters.kpCellDesc,Pointer);
                    Pointer = NULL;
                    EraseBox(&BB);
                    Redisplay(&BB);
                }
                goto quit;
            case PL_CMD:
                if (Not ExpectFirstPoint) {
                    if (NumVertices == 1) {
                        CDDelete(Parameters.kpCellDesc,Pointer);
                    }
                    else {
                        Modified++;
                        record_wire(Pointer);
                    }
                    Pointer = NULL;
                    EraseBox(&BB);
                    Redisplay(&BB);
                }
                goto quit;
        }
        if (ExpectFirstPoint) {
            NumVertices++;
            ExpectFirstPoint = False;
            FX = X = SCursor.kcX;
            FY = Y = SCursor.kcY;
            Path = allocate_path(X,Y);
            allocate_wire(1,Path,Width,&BB,&Pointer,SQ_INCMPLT);
            ShowWire(DrawingColor,Width,Path);
            FBSetRubberBanding('l');
            continue;
        }
        if (SCursor.kcX == X And SCursor.kcY == Y) {
            if (NumVertices == 1) {
                /* click twice to exit */
                CDDelete(Parameters.kpCellDesc,Pointer);
                Pointer = NULL;
                goto quit;
            }
            else {
                Pointer->oInfo = SQ_NEW;
                record_wire(OldPointer);
                OldPointer = Pointer;
                Pointer = NULL;
                OldBB = BB;
                Modified++;
                FBSetRubberBanding(0);
            }
            NumVertices = 0;
            ExpectFirstPoint = True;
            EraseBox(&BB);
            Redisplay(&BB);
            Undo = False;
        }
        else { 
            X = SCursor.kcX;
            Y = SCursor.kcY;
            append_point(X,Y,&Path);
            CDDelete(Parameters.kpCellDesc,Pointer);
            allocate_wire(1,Path,Width,&BB,&Pointer,SQ_INCMPLT);
            FBSetRubberBanding(0);
            FBSetRubberBanding('l');
            ShowWire(DrawingColor,Width,Path);
            NumVertices++;
        }
    }
quit:
    record_wire(OldPointer);
    record_wire(Pointer);
    if (Modified)
        Parameters.kpModified = True;
    ErasePrompt();
    FBSetRubberBanding(0);
    MenuDeselect(MenuWIRES);
}


/* ARGSUSED */
void
ShowWire(Color,Width,Path)

/* Display a wire. */
int Color;
long Width;
struct p *Path;
{
    ShowPath(Color,Path,0);
}


int
IsManhattan(X1,Y1,X2,Y2)

/* return True if coordinates are Manhattan. */
long X1,Y1,X2,Y2;
{
    if (X1 == X2 Or Y1 == Y2)
        return (True);
    else return (False);
}


static void
allocate_wire(Layer,Path,Width,BB,Pointer,Info)

/* Create a new wire. */
int Layer;
struct p *Path;
long Width;
struct ka *BB;
struct o **Pointer;
int Info;
{
    if (Not CDMakeWire(Parameters.kpCellDesc,
        Layer,Width,Path,Pointer)) MallocFailed();
    (*Pointer)->oInfo = Info;
    CDStatusInt = CDBB(Parameters.kpCellDesc,*Pointer,
        &BB->kaLeft,&BB->kaBottom,&BB->kaRight,&BB->kaTop);
    OversizeBox(BB,Width);
}


static void
record_wire(Pointer)

/* Check the Info field, and delete if conditionally deleted.
 * Otherwise set Info to SQ_OLD, and add default properties to wire.
 */
struct o *Pointer;
{
    int Info;

    if (Pointer == NULL) return;
    if (Pointer->oInfo == SQ_GONE)
        CDDelete(Parameters.kpCellDesc,Pointer);
    else {
        Pointer->oInfo = SQ_OLD;
        if (Pointer->oPrptyList == NULL)
            AssignWireProperties(Pointer);
    }
}


static void
zap_last_point(Path)

/* Create a new path with the last point removed.  Delete the old path. */
struct p **Path;
{
    struct p *OldPair;
    struct p *NewPair;

    OldPair = *Path;
    if (OldPair == NULL Or OldPair->pSucc == NULL)
        return;
    if ((NewPair = alloc(p)) == NULL)
        MallocFailed();
    *Path = NewPair;
    NewPair->pX = OldPair->pX; 
    NewPair->pY = OldPair->pY; 
    NewPair->pSucc = NULL;
    OldPair = OldPair->pSucc;
    while (OldPair->pSucc != NULL) {
        if ((NewPair->pSucc = alloc(p)) == NULL)
            MallocFailed();
        NewPair = NewPair->pSucc;
        NewPair->pX = OldPair->pX; 
        NewPair->pY = OldPair->pY; 
        NewPair->pSucc = NULL;
        OldPair = OldPair->pSucc;
    }
}


static void
append_point(X,Y,Path)

/* Create a new path with an added point.  Delete the old path. */
struct p **Path;
long X,Y;
{
    struct p *OldPair;
    struct p *NewPair;

    if (Path == NULL || *Path == NULL)
        return;
    OldPair = *Path;
    if ((NewPair = alloc(p)) == NULL)
        MallocFailed();
    *Path = NewPair;
    NewPair->pX = OldPair->pX; 
    NewPair->pY = OldPair->pY; 
    NewPair->pSucc = NULL;
    OldPair = OldPair->pSucc;
    while (OldPair != NULL) {
        if ((NewPair->pSucc = alloc(p)) == NULL)
            MallocFailed();
        NewPair = NewPair->pSucc;
        NewPair->pX = OldPair->pX; 
        NewPair->pY = OldPair->pY; 
        NewPair->pSucc = NULL;
        OldPair = OldPair->pSucc;
    }
    /* Append the new reference point */
    if ((NewPair->pSucc = alloc(p)) == NULL)
        MallocFailed();
    NewPair = NewPair->pSucc;
    NewPair->pX = X;
    NewPair->pY = Y;
    NewPair->pSucc = NULL;
}


static void
last_path_point(X,Y,Path)

/* Return in pointers the last point in path. */
long *X,*Y;
struct p *Path;
{
    struct p *p;

    for (p = Path; p && p->pSucc; p = p->pSucc) ;
    if (p) {
        *X = p->pX;
        *Y = p->pY;
    }
}


static struct p *
allocate_path(X,Y)

/* Allocate an element of a path list. */
long X,Y;
{
    struct p *Path;

    if ((Path = alloc(p)) == NULL) MallocFailed();
    Path->pX = X;
    Path->pY = Y;
    Path->pSucc = NULL;
    return (Path);
}


/***********************************************************************
 *
 * Routines for rendering line paths and polygons.
 * 
 *
 ***********************************************************************/


void
ShowPath(Color,Path,Terminate)

/* display a line path. */
int Color;
struct p *Path;
int Terminate;     /* If True, the path is closed */
{
    struct p *Pair;
    long firstX,firstY,X,Y,lastX,lastY;

    Pair = Path;
    firstX = lastX = Pair->pX;
    firstY = lastY = Pair->pY;
    Pair = Pair->pSucc;
    while (Pair != NULL) {
        X = Pair->pX;
        Y = Pair->pY;
        ShowLine(Color,lastX,lastY,X,Y);
        lastX = X;
        lastY = Y;
        Pair = Pair->pSucc;
    }
    if (Terminate)
        ShowLine(Color,firstX,firstY,lastX,lastY);
}


void
ShowPolygon(Color,Path)

/* Display a polygon. */
int Color;
struct p *Path;
{
    struct p *Pair;
    struct ka Fw, Cw;
    static long PolygonBuffer[POLYGONBUFSIZE];
    char oldRedisplay;
    int i,n,nfine,ncoarse,n2;
    long X,Y;


    if (Color == HighlightingColor) {
        ShowPath(Color,Path,True);
        return;
    }

    for (n = 0,Pair = Path; Pair != NULL; n++,Pair = Pair->pSucc) ;
    n = Min(MAXPOLYGONVERTICES,n);
    n2 = n << 1;
    ncoarse = nfine = n;

    if (Parameters.kpRedisplayControl != FINEVIEWPORTONLY And
        CurrentAOI.aInCoarse) {

        for (i = 0,Pair = Path; i < n2; i += 2,Pair = Pair->pSucc) {
            X = Pair->pX;
            Y = Pair->pY;
            TPoint(&X,&Y);
            CoarseLToP(X,Y,PolygonBuffer[i],PolygonBuffer[i+1]);
        }

        FBPolygonClip(PolygonBuffer,&ncoarse,(struct ka*)&CurrentAOI.aLC);
        FBPolygon(Color,FILL,PolygonBuffer,ncoarse);
    }

    if (Parameters.kpRedisplayControl != COARSEVIEWPORTONLY And
        CurrentAOI.aInFine) {

        for (i = 0,Pair = Path; i < n2; i += 2,Pair = Pair->pSucc) {
            X = Pair->pX;
            Y = Pair->pY;
            TPoint(&X,&Y);
            FineLToP(X,Y,PolygonBuffer[i],PolygonBuffer[i+1]);
        }

        FBPolygonClip(PolygonBuffer,&nfine,(struct ka*)&CurrentAOI.aLF);
        FBPolygon(Color,FILL,PolygonBuffer,nfine);
    }

}


/***********************************************************************
 *
 * Routines for managing rectangles.
 * 
 *
 ***********************************************************************/

#define ka_copy(BB1,BB2) BB1.kaLeft = BB2->kaLeft; \
    BB1.kaRight = BB2->kaRight; \
    BB1.kaBottom = BB2->kaBottom; \
    BB1.kaTop = BB2->kaTop; \
    TPoint(&BB1.kaLeft,&BB1.kaBottom); \
    TPoint(&BB1.kaRight,&BB1.kaTop); \
    if (BB1.kaLeft > BB1.kaRight) \
        SwapInts(BB1.kaLeft,BB1.kaRight); \
    if (BB1.kaBottom > BB1.kaTop) \
        SwapInts(BB1.kaBottom,BB1.kaTop);


int
InBox(X,Y,AOI)

/* return True if X,Y in AOI */
long X,Y;
struct ka *AOI;
{
    if (((AOI->kaLeft <= X And X <= AOI->kaRight) Or
        (AOI->kaRight <= X And X <= AOI->kaLeft)) And
        ((AOI->kaBottom <= Y And Y <= AOI->kaTop) Or
        (AOI->kaTop <= Y And Y <= AOI->kaBottom)))
        return (True);
    return (False);
}


void
OversizeBox(BB,Delta)

/* expand BB by Delta */
struct ka *BB;
int Delta;
{
    BB->kaTop += Delta;
    BB->kaRight += Delta;
    BB->kaBottom -= Delta;
    BB->kaLeft -= Delta;
}


void
OutlineBox(AOI)

/* Draw a line box. AOI is pixel coordinates. */
struct ka *AOI;
{
    int x1,y1,x2,y2;
    x1 = AOI->kaLeft;
    y1 = AOI->kaBottom;
    x2 = AOI->kaRight;
    y2 = AOI->kaTop;
    DevLine(x1,y1,x2,y1);
    DevLine(x2,y1,x2,y2);
    DevLine(x2,y2,x1,y2);
    DevLine(x1,y2,x1,y1);
}


void
ShowEmptyBox(Color,boxBB)

/* Draw a line box (window coordinates). */
int Color;
struct ka *boxBB;
{
    struct ka BB,BB1;
    int ShowLeft   = True;
    int ShowBottom = True;
    int ShowRight  = True;
    int ShowTop    = True;

    ka_copy(BB,boxBB);

    if (Parameters.kpRedisplayControl != FINEVIEWPORTONLY And
        CurrentAOI.aInCoarse) {

        CoarseLToP(BB.kaLeft,BB.kaBottom,BB1.kaLeft,BB1.kaBottom);
        CoarseLToP(BB.kaRight,BB.kaTop,BB1.kaRight,BB1.kaTop);

        if (BB1.kaLeft <= CurrentAOI.aRC And BB1.kaRight >= CurrentAOI.aLC And
            BB1.kaBottom <= CurrentAOI.aTC And BB1.kaTop >= CurrentAOI.aBC) {

            if (BB1.kaLeft < CurrentAOI.aLC) {
                BB1.kaLeft = CurrentAOI.aLC;
                ShowLeft = False;
            }
            if (BB1.kaBottom < CurrentAOI.aBC) {
                BB1.kaBottom = CurrentAOI.aBC;
                ShowBottom = False;
            }
            if (BB1.kaRight > CurrentAOI.aRC) {
                BB1.kaRight = CurrentAOI.aRC;
                ShowRight = False;
            }
            if (BB1.kaTop > CurrentAOI.aTC) {
                BB1.kaTop = CurrentAOI.aTC;
                ShowTop = False;
            }
            DevSetColor(Color);

            if (ShowTop)
                DevLine((int)BB1.kaLeft,(int)BB1.kaTop,
                    (int)BB1.kaRight,(int)BB1.kaTop);
            if (ShowRight)
                DevLine((int)BB1.kaRight,(int)BB1.kaTop,
                    (int)BB1.kaRight,(int)BB1.kaBottom);
            if (ShowBottom)
                DevLine((int)BB1.kaRight,(int)BB1.kaBottom,
                    (int)BB1.kaLeft,(int)BB1.kaBottom);
            if (ShowLeft)
                DevLine((int)BB1.kaLeft,(int)BB1.kaBottom,
                    (int)BB1.kaLeft,(int)BB1.kaTop);
        }
    }

    ShowLeft   = True;
    ShowBottom = True;
    ShowRight  = True;
    ShowTop    = True;

    if (Parameters.kpRedisplayControl != COARSEVIEWPORTONLY And
        CurrentAOI.aInFine) {

        FineLToP(BB.kaLeft,BB.kaBottom,BB1.kaLeft,BB1.kaBottom);
        FineLToP(BB.kaRight,BB.kaTop,BB1.kaRight,BB1.kaTop);

        if (BB1.kaLeft <= CurrentAOI.aRF And BB1.kaRight >= CurrentAOI.aLF And
            BB1.kaBottom <= CurrentAOI.aTF And BB1.kaTop >= CurrentAOI.aBF) {

            if (BB1.kaLeft < CurrentAOI.aLF) {
                BB1.kaLeft = CurrentAOI.aLF;
                ShowLeft = False;
            }
            if (BB1.kaBottom < CurrentAOI.aBF) {
                BB1.kaBottom = CurrentAOI.aBF;
                ShowBottom = False;
            }
            if (BB1.kaRight > CurrentAOI.aRF) {
                BB1.kaRight = CurrentAOI.aRF;
                ShowRight = False;
            }
            if (BB1.kaTop > CurrentAOI.aTF) {
                BB1.kaTop = CurrentAOI.aTF;
                ShowTop = False;
            }
            DevSetColor(Color);

            if (ShowTop)
                DevLine((int)BB1.kaLeft,(int)BB1.kaTop,
                    (int)BB1.kaRight,(int)BB1.kaTop);
            if (ShowRight)
                DevLine((int)BB1.kaRight,(int)BB1.kaTop,
                    (int)BB1.kaRight,(int)BB1.kaBottom);
            if (ShowBottom)
                DevLine((int)BB1.kaRight,(int)BB1.kaBottom,
                    (int)BB1.kaLeft,(int)BB1.kaBottom);
            if (ShowLeft)
                DevLine((int)BB1.kaLeft,(int)BB1.kaBottom,
                    (int)BB1.kaLeft,(int)BB1.kaTop);
        }
    }
}


void
EraseBox(boxBB)

/* Erase area in boxBB (window coordinates). */
struct ka *boxBB;
{
    struct ka BB,BB1;

    SetCurrentAOI(boxBB);
    ka_copy(BB,boxBB);

    if (Parameters.kpRedisplayControl != FINEVIEWPORTONLY And
        CurrentAOI.aInCoarse) {

        CoarseLToP(BB.kaLeft,BB.kaBottom,BB1.kaLeft,BB1.kaBottom);
        CoarseLToP(BB.kaRight,BB.kaTop,BB1.kaRight,BB1.kaTop);

        if (BB1.kaLeft <= CurrentAOI.aRC And BB1.kaRight >= CurrentAOI.aLC And
            BB1.kaBottom <= CurrentAOI.aTC And BB1.kaTop >= CurrentAOI.aBC) {

            if (BB1.kaLeft < CurrentAOI.aLC)
                BB1.kaLeft = CurrentAOI.aLC;
            if (BB1.kaBottom < CurrentAOI.aBC)
                BB1.kaBottom = CurrentAOI.aBC;
            if (BB1.kaRight > CurrentAOI.aRC)
                BB1.kaRight = CurrentAOI.aRC;
            if (BB1.kaTop > CurrentAOI.aTC)
                BB1.kaTop = CurrentAOI.aTC;
            FBEraseBox(BB1.kaLeft,BB1.kaBottom,
                BB1.kaRight,BB1.kaTop);
        }
    }

    if (Parameters.kpRedisplayControl != COARSEVIEWPORTONLY And
        CurrentAOI.aInFine) {

        FineLToP(BB.kaLeft,BB.kaBottom,BB1.kaLeft,BB1.kaBottom);
        FineLToP(BB.kaRight,BB.kaTop,BB1.kaRight,BB1.kaTop);

        if (BB1.kaLeft <= CurrentAOI.aRF And BB1.kaRight >= CurrentAOI.aLF And
            BB1.kaBottom <= CurrentAOI.aTF And BB1.kaTop >= CurrentAOI.aBF) {

            if (BB1.kaLeft < CurrentAOI.aLF)
                BB1.kaLeft = CurrentAOI.aLF;
            if (BB1.kaBottom < CurrentAOI.aBF)
                BB1.kaBottom = CurrentAOI.aBF;
            if (BB1.kaRight > CurrentAOI.aRF)
                BB1.kaRight = CurrentAOI.aRF;
            if (BB1.kaTop > CurrentAOI.aTF)
                BB1.kaTop = CurrentAOI.aTF;
            FBEraseBox(BB1.kaLeft,BB1.kaBottom,
                BB1.kaRight,BB1.kaTop);
        }
    }
}


/***********************************************************************
 *
 * Strch command.
 * Move wire vertices around.
 *
 ***********************************************************************/

extern char *MenuSTRCH;
extern char *MenuUNDO;
#ifdef __STDC__
static void restore_stretch(void);
static void do_stretch_path(long,long,long,long,struct ka*,int);
static struct p *get_nearest_vertex(struct p*,long,long);
static void set_ref_to_vertex(long*,long*);
#else
static void restore_stretch();
static void do_stretch_path();
static struct p *get_nearest_vertex();
static void set_ref_to_vertex();
#endif


void
StretchPath(LookedAhead)

/* Strch command, move vertices. */
int *LookedAhead;
{
    struct ka NBB,OBB;
    long RefX,RefY,LastRefX,LastRefY,MapX,MapY;
    long RefTmpX, RefTmpY;
    int FirstTime = True;
    int GotOne = False;
    int Undo = False;
    int Modified = 0;
    int Pt;
    char Types[2];

    MenuSelect(MenuSTRCH);

    Types[0] = CDWIRE;
    Types[1] = '\0';

    if (AreTypesInQ(Types))
        GotOne = True;

    loop {
top:
        if (Not GotOne) {
            ShowPrompt("Point at wire to stretch");

            switch (PointLoop(LookedAhead)) {
                case PL_ESC:
                case PL_CMD:
                    goto quit;
                case PL_UND:
                    if (FirstTime) goto quit;
                    if (Undo  == False) {
                        MenuSelect(MenuUNDO);
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(ERASE);
                        UndoReferenceTransform();
                        HYundoTransform();
                        SQRestore(True);
                        /* restored objects have Info = SQ_NEWSEL */
                        Modified--;
                        Undo = True;
                        NBB = OBB;
                        EraseBox(&OBB);
                        Redisplay(&OBB);
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(DISPLAY);
                        MenuDeselect(MenuUNDO);
                        break;
                    }
                    else {
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(ERASE);
                        /* should have only Info = SQ_NEW objects here */
                        do_stretch_path(RefX,RefY,MapX,MapY,&NBB,False);
                        Modified++;
                        EraseBox(&NBB);
                        Redisplay(&NBB);
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(DISPLAY);
                        Undo = False;
                        MenuDeselect(MenuUNDO);
                        continue;
                    }
                case PL_PCW:
                    SelectTypes(Types);
                    if (Not AreTypesInQ(Types))
                        goto top;
                    SQComputeBB();
                    NBB = SelectQBB;
            }
        }

next:
        ShowPrompt("Point to the vertex.");

        switch (PointLoop(LookedAhead)) {
            case PL_ESC:
            case PL_CMD:
                goto quit;
            case PL_UND:
                MenuSelect(MenuUNDO);
                if (Not GotOne)
                     SQDesel(Types);
                    /* newly selected objects deleted, restored objects
                     * have Info = SQ_NEW.
                     */
                else {
                    if (FirstTime) {
                        MenuDeselect(MenuUNDO);
                        goto quit;
                    }
                    if (Undo == False) {
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(ERASE);
                        UndoReferenceTransform();
                        HYundoTransform();
                        SQRestore(True);
                        /* restored objects have Info = SQ_NEWSEL */
                        Modified--;
                        Undo = True;
                    }
                    else {
                        if (Parameters.kpShowTerminals)
                            DisplayTerminals(ERASE);
                        /* should have only Info = SQ_NEWSEL objects here */
                        do_stretch_path(RefX,RefY,MapX,MapY,&NBB,True);
                        Modified++;
                        Undo = False;
                    }
                }
                EraseBox(&NBB);
                Redisplay(&NBB);
                if (GotOne && Parameters.kpShowTerminals)
                    DisplayTerminals(DISPLAY);
                MenuDeselect(MenuUNDO);
                goto top;
        }

        LastRefX = RefX;
        LastRefY = RefY;
        RefX = SCursor.kcX;
        RefY = SCursor.kcY;
        set_ref_to_vertex(&RefX,&RefY);
        ShowPrompt("Point to where it should stretch.");

        RefTmpX = SCursor.kcX;
        RefTmpY = SCursor.kcY;
        SCursor.kcX = RefX;
        SCursor.kcY = RefY;
        restore_stretch();
        FBSetRubberBanding('s');
        SCursor.kcX = RefTmpX;
        SCursor.kcY = RefTmpY;
        Pt = PointLoop(LookedAhead);
        FBSetRubberBanding(0);

        switch (Pt) {
            case PL_ESC:
            case PL_CMD:
                goto quit;
            case PL_UND:
                MenuSelect(MenuUNDO);
                RefX = LastRefX;
                RefY = LastRefY;
                MenuDeselect(MenuUNDO);
                goto next;
        }

        SQRestore(False);
        /* should have only Info = SQ_OLDSEL objects here */
        FirstTime = False;

        MapX = SCursor.kcX;
        MapY = SCursor.kcY;

        if (Parameters.kpShowTerminals) DisplayTerminals(ERASE);
        do_stretch_path(RefX,RefY,MapX,MapY,&NBB,GotOne);
        EraseBox(&NBB);
        Redisplay(&NBB);
        if (Parameters.kpShowTerminals) DisplayTerminals(DISPLAY);
        OBB = NBB;
        Modified++;
        Undo = False;
    }

quit:
    ClearReferenceUndoFlags();
    HYclearUndoFlags();
    SQRestore(False);
    if (Not GotOne And AreTypesInQ(Types)) {
        SQComputeBB();
        SQDesel(Types);
        EraseBox(&SelectQBB);
        Redisplay(&SelectQBB);
    }
    if (Modified)
        Parameters.kpModified = True;
    ErasePrompt();
    MenuDeselect(MenuUNDO);
    MenuDeselect(MenuSTRCH);
}


void
ShowStretch(MapX,MapY,RefX,RefY)

/* Called from rubber banding routine. */
long RefX,RefY,MapX,MapY;
{
    struct ks *SQDesc, *SQDesc1;
    struct p *Path, *pp;
    int Layer;
    long Width;

    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc) {
        if (SQDesc->ksPointer->oInfo != SQ_OLDSEL &&
            SQDesc->ksPointer->oInfo != SQ_NEWSEL) continue;
        if (SQDesc->ksPointer->oType != CDWIRE) continue;

        /* draw only once if in queue more than once */
        for (SQDesc1 = SQDesc->ksSucc; SQDesc1; SQDesc1 = SQDesc1->ksSucc)
            if (SQDesc1->ksPointer == SQDesc->ksPointer) break;

        if (SQDesc1 == NULL) {
            CDWire(SQDesc->ksPointer,&Layer,&Width,&Path);
            pp = get_nearest_vertex(Path,RefX,RefY);
            pp->pX += MapX - RefX;
            pp->pY += MapY - RefY;
            ShowWire(HighlightingColor,Width,Path);
            pp->pX -= MapX - RefX;
            pp->pY -= MapY - RefY;
        }
    }
}

static void
restore_stretch()

/* Repaint objects to be stretched. */
{
    struct ks *SQDesc;
    struct p *Path;
    int Layer;
    long Width;

    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc) {
        if (SQDesc->ksPointer->oInfo == SQ_GONE) continue;
        if (SQDesc->ksPointer->oType != CDWIRE) continue;

        CDWire(SQDesc->ksPointer,&Layer,&Width,&Path);
        ShowWire(DrawingColor,Width,Path);
    }
}


static void
do_stretch_path(RefX,RefY,MapX,MapY,NBB,SelectNew)

/* Perform the stretch. */
long RefX,RefY,MapX,MapY;
struct ka *NBB;
int SelectNew;
{
    struct ks *SQDesc;
    struct ka OBB,BB;
    struct o *Pointer;
    struct p *Path, *pp;
    int Layer;
    long Width;

    NBB->kaLeft   = CDINFINITY;
    NBB->kaRight  = -CDINFINITY;
    NBB->kaBottom = CDINFINITY;
    NBB->kaTop    = -CDINFINITY;

    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc) {
        if (SQDesc->ksPointer->oInfo == SQ_GONE) continue;
        if (SQDesc->ksPointer->oType != CDWIRE) continue;

        CDWire(SQDesc->ksPointer,&Layer,&Width,&Path);
        CDStatusInt = CDBB(Parameters.kpCellDesc,SQDesc->ksPointer,
            &OBB.kaLeft,&OBB.kaBottom,&OBB.kaRight,&OBB.kaTop);

        Path = CopyPath(Path);
        pp = get_nearest_vertex(Path,RefX,RefY);
        pp->pX += MapX - RefX;
        pp->pY += MapY - RefY;

        if (Not CDMakeWire(Parameters.kpCellDesc,Layer,Width,Path,
            &Pointer)) MallocFailed();
        AssignWireProperties(Pointer);

        if (SelectNew)
            Pointer->oInfo = SQ_NEWSEL;
        else
            Pointer->oInfo = SQ_NEW;

        SQInsert(Pointer);

        CDStatusInt = CDBB(Parameters.kpCellDesc,Pointer,
            &BB.kaLeft,&BB.kaBottom,&BB.kaRight,&BB.kaTop);

        if (OBB.kaLeft   < BB.kaLeft)   BB.kaLeft   = OBB.kaLeft;
        if (OBB.kaRight  > BB.kaRight)  BB.kaRight  = OBB.kaRight;
        if (OBB.kaBottom < BB.kaBottom) BB.kaBottom = OBB.kaBottom;
        if (OBB.kaTop    > BB.kaTop)    BB.kaTop    = OBB.kaTop;

        if (BB.kaLeft   < NBB->kaLeft)   NBB->kaLeft   = BB.kaLeft;
        if (BB.kaRight  > NBB->kaRight)  NBB->kaRight  = BB.kaRight;
        if (BB.kaBottom < NBB->kaBottom) NBB->kaBottom = BB.kaBottom;
        if (BB.kaTop    > NBB->kaTop)    NBB->kaTop    = BB.kaTop;

        TPush();
        TIdentity();
        TTranslate(MapX-RefX,MapY-RefY);
        TransformReferencePoint(SQDesc->ksPointer,RefX,RefY);
        HYtransformStretch(SQDesc->ksPointer,Pointer,RefX,RefY,MapX,MapY);
        TPop();
        SQDesc->ksPointer->oInfo = SQ_GONE;
    }
}


static struct p *
get_nearest_vertex(Path,X,Y)

/* return the path vertex nearest X,Y. */
struct p *Path;
long X,Y;
{
    double dx,dy,d,mind;
    struct p *p;
    int i,indx = 0;

    mind = 1e30;
    for (p = Path,i = 0; p; p = p->pSucc,i++) {
        dx = p->pX - X;
        dy = p->pY - Y;
        d = dx*dx + dy*dy;
        if (d < mind) {
            mind = d;
            indx = i;
        }
    }
    for (p = Path, i = 0; i < indx; p = p->pSucc,i++) ;
    return (p);
}


static void
set_ref_to_vertex(X,Y)

/* Return in pointers the vertex closest to the given coordinates. */
long *X,*Y;
{
    struct ks *SQDesc;
    struct o *Pointer;
    struct p *p,*Path;
    double dx,dy,d,mind;
    int i,indx;


    mind = 1e30;
    for (SQDesc = SelectQHead; SQDesc; SQDesc = SQDesc->ksSucc) {

        if (SQDesc->ksPointer->oInfo == SQ_GONE) continue;
        if (SQDesc->ksPointer->oType != CDWIRE) continue;
        Path = ((struct w *)SQDesc->ksPointer->oRep)->wPath;

        for (p = Path,i = 0; p; p = p->pSucc,i++) {
            dx = p->pX - *X;
            dy = p->pY - *Y;
            d = dx*dx + dy*dy;
            if (d < mind) {
                mind = d;
                indx = i;
                Pointer = SQDesc->ksPointer;
            }
        }
    }
    Path = ((struct w *)Pointer->oRep)->wPath;
    for (p = Path, i = 0; i < indx; p = p->pSucc,i++) ;
    *X = p->pX;
    *Y = p->pY;
}


/***********************************************************************
 *
 * Dot command.
 * Place dots at connections.
 *
 ***********************************************************************/

extern char *MenuDOTS;
extern char *MenuUNDO;

static char DotP[] =  {0,60, 40,40, 60,0, 40,-40, 0,-60, -40,-40,
                      -60,0, -40,40, 0,60};

static struct p *DotList;

struct pp {
    int ppX, ppY;
    int ppType;
    struct pp *ppSucc;
};
static struct pp *VertexList;

struct clist {
    char *name;
    struct s *cd;
    struct clist *next;
};
static struct clist *CList;

#ifdef __STDC__
static void clear_dots(struct s*);
static void create_dots(struct s*);
static void save_vertex(int,int,int);
static void save_dot(int,int);
static void add_dots(struct s*);
static void list_cells(void);
static void mlist(struct m*);
static struct s *add_to_list(char*);
#else
static void clear_dots();
static void create_dots();
static void save_vertex();
static void save_dot();
static void add_dots();
static void list_cells();
static void mlist();
static struct s *add_to_list();
#endif


void
Dots(LookedAhead)

int *LookedAhead;
{
    struct clist *s;

    MenuSelect(MenuDOTS);
    Parameters.kpShowDots = True;
    list_cells();
    for (s = CList; s; s = s->next)
        create_dots(s->cd);
    EraseLargeCoarseViewport();
    Redisplay(View->kvCoarseWindow);

    loop {
        switch (PointLoop(LookedAhead)) {
            case PL_CMD:
                if (Matching(MenuDOTS))
                    Parameters.kpCommand[0] = '\0';
            case PL_ESC:
            case PL_UND:
                goto quit;
            case PL_PCW:
                break;
        }
    }
quit:
    for (s = CList; s; s = s->next)
        clear_dots(s->cd);
    for (; CList; CList = s) {
        s = CList->next;
        txfree((char*)CList);
    }
    CList = NULL;
    MenuDeselect(MenuDOTS);
    Parameters.kpShowDots = False;
    EraseLargeCoarseViewport();
    Redisplay(View->kvCoarseWindow);
}


static void
clear_dots(CellDesc)

struct s *CellDesc;
{
    struct g *GenDesc;
    struct o *Pointer;

    if (Not CDInitGen(CellDesc,1,-CDINFINITY,-CDINFINITY,
        CDINFINITY,CDINFINITY,&GenDesc)) MallocFailed();

    loop {
        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE Or Pointer->oInfo == SQ_INCMPLT)
            continue;
        if (Pointer->oType == CDPOLYGON) {
            CDDelete(CellDesc,Pointer);
        }
    }
}

static void
create_dots(CellDesc)

struct s *CellDesc;
{
    struct g *GenDesc;
    struct o *Pointer;
    struct p *p;
    struct prpty *pd;

    /* Process wires */

    if (Not CDInitGen(CellDesc,1,-CDINFINITY,-CDINFINITY,
        CDINFINITY,CDINFINITY,&GenDesc)) MallocFailed();

    loop {
        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE Or Pointer->oInfo == SQ_INCMPLT)
            continue;
        if (Pointer->oType == CDWIRE) {
            p = ((struct w *)Pointer->oRep)->wPath;
            save_vertex(p->pX,p->pY,0);
            for (p = p->pSucc; p && p->pSucc; p = p->pSucc)
                save_vertex(p->pX,p->pY,1);
            if (p)
                save_vertex(p->pX,p->pY,0);
        }
    }

    /* Process devices */

    if (Not CDInitGen(CellDesc,0,-CDINFINITY,-CDINFINITY,CDINFINITY,
        CDINFINITY,&GenDesc)) MallocFailed();

    loop {
        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE) continue;

        pd = Pointer->oPrptyList;
        for (; pd; pd = pd->prpty_Succ) {
            if (pd->prpty_Value != P_NODE)
                continue;
            save_vertex(pd->prpty_Data->p_node.x,
                pd->prpty_Data->p_node.y,0);
        }
    }
    add_dots(CellDesc);
}


static void
save_vertex(x,y,type)

/* Consider saving point in vertex list; if it is already there,
 * update the type field or add it to the dot list.
 */
int x, y, type;
{
    struct pp *p;

    for (p = VertexList; p; p = p->ppSucc) {
        if (p->ppX == x && p->ppY == y) {
            /* already there, if both types are zero,
             * update the stored entry to type 1 (so as
             * not to mark two abutting line segments)
             */
            if (p->ppType == 0 && type == 0)
                p->ppType = 1;
            else
                save_dot(x,y);
            return;
        }
    }
    p = alloc(pp);
    if (p == NULL) MallocFailed();
    p->ppX = x;
    p->ppY = y;
    p->ppType = type;
    p->ppSucc = VertexList;
    VertexList = p;
}


static void
save_dot(x,y)

/* Save point in the dot list; if it is already
 * there, just return.
 */
int x, y;
{
    struct p *p;

    for (p = DotList; p; p = p->pSucc) {
        if (p->pX == x && p->pY == y)
            return;
    }
    p = alloc(p);
    if (p == NULL) MallocFailed();
    p->pX = x;
    p->pY = y;
    p->pSucc = DotList;
    DotList = p;
}


static void
add_dots(CellDesc)

/* Process and clear the dot and vertex lists. */
struct s *CellDesc;
{
    struct p *p, *pd, *Path;
    struct o *Pointer;
    int i;

    for (pd = DotList; pd; pd = pd->pSucc) {

        p = Path = alloc(p);
        if (p == NULL) MallocFailed();
        p->pX = pd->pX + DotP[0];
        p->pY = pd->pY + DotP[1];
        for (i = 1; i < 9; i++) {
            p->pSucc = alloc(p);
            p = p->pSucc;
            if (p == NULL) MallocFailed();
            p->pX = pd->pX + DotP[2*i];
            p->pY = pd->pY + DotP[2*i+1];
        }
        p->pSucc = NULL;
        if (!CDMakePolygon(CellDesc,1,Path,&Pointer))
            MallocFailed();
    }
    for (; DotList; DotList = p) {
        p = DotList->pSucc;
        txfree((char*)DotList);
    }
    for (; VertexList; VertexList = (struct pp*)p) {
        p = (struct p*)VertexList->ppSucc;
        txfree((char*)VertexList);
    }
    DotList = NULL;
    VertexList = NULL;
}


static void
list_cells()

/* get a list of cells in hierarchy */
{
    CList = alloc(clist);
    if (CList == NULL)
        MallocFailed();
    CList->name = Parameters.kpCellName;
    CList->cd = Parameters.kpCellDesc;
    CList->next = NULL;
    mlist(Parameters.kpCellDesc->sMasterList);
}


static void
mlist(md)

/* walk the master list and recursively add subcircuits */
struct m *md;
{
    struct s *cd;

    for (; md; md = md->mSucc) {
        if (!IsCellInLib(md->mName)) {
            cd = add_to_list(md->mName);
            if (cd) {
                mlist(cd->sMasterList);
            }
        }
    }
}


static struct s *
add_to_list(name)

/* add the cell if not already there, return descr of new addition */
char *name;
{
    struct clist *s;
    struct s *cd;

    for (s = CList; s; s = s->next) {
        if (!strcmp(s->name,name))
            return (NULL);
    }
    if (!CDOpen(name,&cd,'r'))
        return (NULL);
    s = alloc(clist);
    if (s == NULL)
        MallocFailed();
    s->name = name;
    s->cd = cd;
    s->next = CList;
    CList = s;
    return (cd);
}
