/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

#include <stdio.h>
#include <ctype.h>

static char *path;

/* OpenDevice() is user supplied */
#ifdef __STDC__
extern FILE *OpenDevice(char*);
extern char *copy(char*);
#else
extern FILE *OpenDevice();
extern char *copy();
#endif


int
PSetPath(string)

char *string;
{
    char buf[2048];

    if (path) {
        free(path);
        path = 0;
    }
    if (string) {
        strcpy(buf,string);
        cp_pathfix(buf);
        path = copy(buf);
    }
    return (0);
}


int
PAppendPath(string)

char *string;
{
    char buf[2048], *s, *trealloc();
    int len;

    if (!path)
        return (PSetPath(string));
    strcpy(buf,string);
    cp_pathfix(buf);
    len = strlen(path);
    path = trealloc(path,len + strlen(buf) + 2);
    s = path + len;
    *s = ' ';
    strcpy(s+1,buf);
    return (0);
}


char *
PGetPath()

{
    return (path);
}


FILE *
POpen(file, mode, prealname)

char *file;            /* Name of the file to be opened. */
char *mode;            /* The file mode, as given to fopen. */
char **prealname;      /* Pointer to a location that will be filled
                        * in with the address of the real name of
                        * the file that was successfully opened.
                        * If NULL, then nothing is stored.
                        */
{
    char buf[512], buf1[512];
    char *p, *q;
    FILE *f;

    strcpy(buf,file);
    cp_pathfix(buf);
    file = buf;

    /*
     * The function OpenDevice() must be supplied externally.  The intended
     * purpose is to support multi-symbol files (libraries).  OpenDevice()
     * would check to see if "file" is included in the library, and if so,
     * returns a file pointer suitably offset.  Otherwise NULL is returned.
     * Note that libraries are only opened for reading.
     */
    if (!strcmp(mode,"r")) {
        f = OpenDevice(file);
        if (f) {
            if (prealname)
                *prealname = copy(file);
            return (f);
        }
    }

    if (!path || !*path || file[0] == '/' 
#ifdef MSDOS
        || file[0] == '\\' || file[1] == ':'
#endif
        ) {
        if (prealname)
            *prealname = copy(file);
        return (fopen(file,mode));
    }

    /* Last, but not least, try going through the path. */

    p = path;
    while (*p != '\0') {
        while (isspace(*p)) p++;
        q = buf1;
        while ((*p != '\0') && !isspace(*p))
            *q++ = *p++;
        *q++ = '/';
        strcpy(q, file);
        cp_pathfix(buf1);
        f = fopen(buf1, mode);
        if (f != NULL) {
            if (prealname)
                *prealname = copy(buf1);
            return (f);
        }
    }
    return (NULL);
}
