/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * JSPICE3 interface to SCED.
 */

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"
#include "scedio.h"
#include <setjmp.h>
#include <signal.h>


/* imports from spice */
extern int ft_nutmeg;
extern int ft_intrpt;
#ifdef __STDC__
extern int  cp_getvar(char*,int,char*);
extern wordlist *cp_lexer(char*);
extern void wl_free(wordlist*);
extern int  cp_evloop(char*);
extern void cp_pushcontrol(void);
extern void cp_popcontrol(void);
extern wordlist *GetAnalysisFromDeck(void);
#else
extern int  cp_getvar();
extern wordlist *cp_lexer();
extern void wl_free();
extern int  cp_evloop();
extern void cp_pushcontrol();
extern void cp_popcontrol();
extern wordlist *GetAnalysisFromDeck();
#endif


/***********************************************************************
 *
 * Menu definition package.
 *
 ***********************************************************************/

MENU *BasicMenu;
MENU *DeviceMenu;
MENU AmbiguityMenu[DefMenuSize];

char *Menu0;                    
char *Menu90;                
char *Menu180;            
char *Menu270;            
char *MenuMINSB;    
char *MenuMINSG;    
char *MenuMINSR;    
char *MenuPLUSB;    
char *MenuPLUSG;    
char *MenuPLUSR;    
char *MenuABORT;    
char *MenuAREA;        
char *MenuARRAY;        
char *MenuBGRND;    
char *MenuBREAK;    
char *MenuCOLOR;       
char *MenuCOPY;       
char *MenuCRSYM;    
char *MenuDELET;    
char *MenuDESEL;    
char *MenuDIR;            
char *MenuDOTS;            
char *MenuEDIT;        
char *MenuEXPND;        
char *MenuFLATN;    
char *MenuGRAF;        
char *MenuGRID;        
char *MenuHCOPY;
char *MenuHELP;
char *MenuHLITE;    
char *MenuIPLOT;    
char *MenuLABEL;    
char *MenuLAST;        
char *MenuMENU;        
char *MenuMORE;        
char *MenuMOVE;
char *MenuMUTUL;
char *MenuMX;               
char *MenuMY;
char *MenuNEWSC;
char *MenuPAN;            
char *MenuPLACE;            
char *MenuPOP;            
char *MenuPRMPT;
char *MenuPRPTY;
char *MenuPUSH;        
char *MenuRDRAW;    
char *MenuRGB;            
char *MenuRUN;            
char *MenuSAVE;        
char *MenuSELEC;    
char *MenuSHOW;        
char *MenuSTRCH;                
char *MenuSUBCT;                
char *MenuTERMS;        
char *MenuUNDO;
char *MenuUPDAT;
char *MenuVIEW;        
char *MenuWINDO;    
char *MenuWIRES;    
char *MenuZOOM;
static char *MenuSpace;


void
InitMenus()

{

    Menu0        = "0    ";
    Menu90       = "90   ";
    Menu180      = "180  ";
    Menu270      = "270  ";
    MenuMINSB    = "-b   ";
    MenuMINSG    = "-g   ";
    MenuMINSR    = "-r   ";
    MenuPLUSB    = "+b   ";
    MenuPLUSG    = "+g   ";
    MenuPLUSR    = "+r   ";
    MenuABORT    = "quit ";
    MenuAREA     = "area ";
    MenuARRAY    = "array";
    MenuBGRND    = "bgrnd";
    MenuBREAK    = "break";
    MenuCOLOR    = "color";
    MenuCOPY     = "copy ";
    MenuCRSYM    = "crsym";
    MenuDELET    = "delet";
    MenuDESEL    = "desel";
    MenuDIR      = "dir  ";
    MenuDOTS     = "dots ";
    MenuEDIT     = "edit ";
    MenuEXPND    = "expnd";
    MenuFLATN    = "flatn";
    MenuGRAF     = "graf ";
    MenuGRID     = "grid ";
    MenuHCOPY    = "hcopy";
    MenuHELP     = "help";
    MenuHLITE    = "hlite";
    MenuIPLOT    = "iplot";
    MenuLABEL    = "label";
    MenuLAST     = "last ";
    MenuMENU     = "menu ";
    MenuMORE     = "more ";
    MenuMOVE     = "move ";
    MenuMUTUL    = "mutul";
    MenuMX       = "mx   ";
    MenuMY       = "my   ";
    MenuNEWSC    = "newsc";
    MenuPAN      = "pan  ";
    MenuPLACE    = "place";
    MenuPOP      = "pop  ";
    MenuPRMPT    = "prmpt";
    MenuPRPTY    = "prpty";
    MenuPUSH     = "push ";
    MenuRDRAW    = "rdraw";
    MenuRGB      = "rgb  ";
    MenuRUN      = "run  ";
    MenuSAVE     = "save ";
    MenuSELEC    = "selec";
    MenuSTRCH    = "strch";
    MenuSUBCT    = "subct";
    MenuTERMS    = "terms";
    MenuUNDO     = "undo ";
    MenuUPDAT    = "updat";
    MenuVIEW     = "view ";
    MenuWINDO    = "windo";
    MenuWIRES    = "wires";
    MenuZOOM     = "zoom ";
    MenuSpace    = "     ";

    InitBasicMenu();
    InitDeviceMenu();
}


void
InitBasicMenu()

{
    int i = 0, NumBasicMenu = 53;
    int nrows = MenuViewport.kaY;

    if (BasicMenu == NULL)
        BasicMenu = (MENU *) malloc(NumBasicMenu*sizeof(MENU));
    if (BasicMenu == NULL) MallocFailed();
    memset(BasicMenu,0,NumBasicMenu*sizeof(MENU));

    BasicMenu[i++].mEntry = MenuHELP;
    BasicMenu[i++].mEntry = MenuCOLOR;
    BasicMenu[i++].mEntry = MenuEDIT;
    BasicMenu[i++].mEntry = MenuDIR;
    BasicMenu[i++].mEntry = MenuSAVE;

    BasicMenu[i++].mEntry = MenuABORT;
    BasicMenu[i++].mEntry = MenuSpace;
    BasicMenu[i++].mEntry = MenuPAN;
    BasicMenu[i++].mEntry = MenuZOOM;
    BasicMenu[i++].mEntry = MenuWINDO;

    BasicMenu[i++].mEntry = MenuVIEW;
    BasicMenu[i++].mEntry = MenuLAST;
    BasicMenu[i++].mEntry = MenuEXPND;
    BasicMenu[i++].mEntry = MenuGRID;
    BasicMenu[i++].mEntry = MenuSpace;

    BasicMenu[i++].mEntry = MenuAREA;
    BasicMenu[i++].mEntry = MenuSELEC;
    BasicMenu[i++].mEntry = MenuDESEL;
    BasicMenu[i++].mEntry = MenuMOVE;
    BasicMenu[i++].mEntry = MenuCOPY;

    BasicMenu[i++].mEntry = MenuPUSH;
    BasicMenu[i++].mEntry = MenuPOP;
    if (i < nrows) BasicMenu[i++].mEntry = MenuSpace;
    BasicMenu[i++].mEntry = MenuSTRCH;
    BasicMenu[i++].mEntry = MenuBREAK;

    BasicMenu[i++].mEntry = MenuCRSYM;
    BasicMenu[i++].mEntry = MenuFLATN;
    BasicMenu[i++].mEntry = MenuDELET;
    if (i < nrows) BasicMenu[i++].mEntry = MenuSpace;
    BasicMenu[i++].mEntry = MenuUNDO;

    BasicMenu[i++].mEntry = MenuMX;
    BasicMenu[i++].mEntry = MenuMY;
    BasicMenu[i++].mEntry = Menu0;
    if (i < nrows) BasicMenu[i++].mEntry = MenuSpace;
    BasicMenu[i++].mEntry = MenuWIRES;

    BasicMenu[i++].mEntry = MenuLABEL;
    BasicMenu[i++].mEntry = MenuDOTS;
    if (i < nrows) BasicMenu[i++].mEntry = MenuSpace;
    BasicMenu[i++].mEntry = MenuTERMS;
    BasicMenu[i++].mEntry = MenuSUBCT;

    BasicMenu[i++].mEntry = MenuMUTUL;
    BasicMenu[i++].mEntry = MenuPRPTY;
    if (i < nrows) BasicMenu[i++].mEntry = MenuSpace;
    BasicMenu[i++].mEntry = MenuRUN;
    if (i < nrows) BasicMenu[i++].mEntry = MenuSpace;

    BasicMenu[i++].mEntry = MenuGRAF;
    BasicMenu[i++].mEntry = MenuIPLOT;
    BasicMenu[i++].mEntry = MenuHCOPY;
    if (i < nrows) BasicMenu[i++].mEntry = MenuSpace;
    BasicMenu[i++].mEntry = MenuNEWSC;

    BasicMenu[i++].mEntry = MenuSpace;
    BasicMenu[i++].mEntry = NULL;
}



/***********************************************************************
 *
 * Top level entry point and dispatch code.
 *
 ***********************************************************************/

/* set if circuit sourced, so we don't do it again in spice */
static int SourceDone;
static enum Active SCEDisActive;

static jmp_buf errjbuf;

/* from spice */
extern FILE *cp_err;

#ifdef __STDC__
static int  sced_main(void);
static void do_update(void);
static RETSIGTYPE sced_interrupt();
static void fix_spice_deck(struct line*);
#else
static int  sced_main();
static void do_update();
static RETSIGTYPE sced_interrupt();
static void fix_spice_deck();
#endif


int
sced(nameptr)

/* main routine, called from com_sced in input.c
 * return values:
 * -1  "fatal" error, such as malloc failure
 * 0   ok, circuit sourced
 * 1   ok, circuit not sourced
 * 2   user quit without saving file
 * 3   bad file or file name
 */
char **nameptr;
{
    FILE *fp;
    char *c, *name;
    int i;
    GRAPH *graph;
    RETSIGTYPE (*sc_intr)();
    RETSIGTYPE (*sc_float)();

    if (SCEDisActive) {
        ShowPromptAndWait("SCED is already running!");
        return (-1);
    }
    name = *nameptr;
    SourceDone = False;
    SetPathFromSpice();
    if (name && *name) {

        fp = POpen(name,"r",(char **)NULL);
        if (fp) {
            if (!IsSymFile(fp)) return (3);
            fclose(fp);
        }
    }
    else
        return (3);

    graph = (GRAPH*)Init();     /* push graph context on success */
    if (graph == NULL)
        return (-1);

    sc_intr = signal(SIGINT,SIG_IGN);
    sc_float = signal(SIGFPE,SIG_IGN);

    /* clear out iplots */
    {wordlist wl; wl.wl_next = NULL; wl.wl_word = "iplot";
    (*ft_sced->sc_clriplot)(&wl);}

    if (!strcmp(Parameters.kpCellName,name) && Parameters.kpCellDesc) {
        if (Not CDClose(Parameters.kpCellDesc)) MallocFailed();
        Parameters.kpCellDesc = NULL;
    }
    else
        strcpy(Parameters.kpCellName,name);
    LibraryOpen(DeviceLibrary);
    ModelLibraryOpen(ModelLibrary);
    if (BasicMenu == NULL)
        InitMenus();
    if (setjmp(errjbuf) !=  1) {
        SCEDisActive = SCED_active;
        i = sced_main();
        DevHalt();
        PopGraphContext();
        DestroyGraph(graph->graphid);
        signal(SIGINT,sc_intr);
        signal(SIGFPE,sc_float);
        *nameptr = Parameters.kpCellName;
        SCEDisActive = SCED_inactive;
        return (i);
    }
    else {
        ClearContext();
        PopGraphContext();
        DestroyGraph(graph->graphid);
        signal(SIGINT,sc_intr);
        signal(SIGFPE,sc_float);
        SCEDisActive = SCED_inactive;
        return (-1);  /* malloc failed */
    }
}


static int
sced_main()

{
    int i, LookedAhead = False;
    char *TypeIn;

    InitVLT();
    Basic(&LookedAhead);
    Edit(True,True,False);
    loop {
        InitSignals();
        if (Not LookedAhead)
            Point();
        else
            LookedAhead = False;
        if (Parameters.kpCellName[0] == EOS) {
            /*
             * The CD package doesn't check the integrity of symbol
             * descriptors and now Parameters.kpCellDesc == NULL. 
             * Except for a few commands, each command invokes the
             * CD package.  If such a command is invoked now, the
             * CD package will crash and thus SCED will.  So, force user
             * to select a cell to edit before invoking any such command.
             */
            if (Not (Matching(MenuEDIT) Or
                Matching(MenuDIR) Or Matching(MenuABORT))) {
                ShowPrompt("You haven't selected a cell to edit.");
                continue;
            }
        }
        if (SafeCmds(&LookedAhead)) continue;

        if (Matching(MenuABORT)) {
            i = CheckSaved();
            if (i == -1) continue;
            ClearContext();
            if (i == 0) {
                if (SourceDone) return (0);
                return (1);
            }
            return (2);
        }

        if (Matching(MenuAREA))  { Area(&LookedAhead);           continue; }
        if (Matching(MenuBREAK)) { Break(&LookedAhead);          continue; }
        if (Matching(MenuCOPY))  { Copy(&LookedAhead);           continue; }
        if (Matching(MenuCRSYM)) { NewSymbol();                  continue; }
        if (Matching(MenuDELET)) { Del(&LookedAhead);            continue; }
        if (Matching(MenuDESEL)) { Desel();                      continue; }
        if (Matching(MenuDOTS))  { Dots(&LookedAhead);           continue; }
        if (Matching(MenuEDIT))  { SourceDone = False;
                                   Edit(False,True,False);       continue; }
        if (Matching(MenuFLATN)) { Flatten(&LookedAhead);        continue; }
        if (Matching(MenuGRAF))  { ShowOutput(&LookedAhead);     continue; }
        if (Matching(MenuHELP))  { SCEDhelp();                   continue; }
        if (Matching(MenuLABEL)) { Label(&LookedAhead);          continue; }
        if (Matching(MenuMOVE))  { Move(&LookedAhead);           continue; }
        if (Matching(MenuMUTUL)) { ShowMutual(&LookedAhead);     continue; }
        if (Matching(MenuNEWSC)) { NewSubckt();                  continue; }
        if (Matching(MenuPLACE)) { Place(&LookedAhead);          continue; }
        if (Matching(MenuPOP))   { Pop();                        continue; }
        if (Matching(MenuPRPTY)) { ShowProperties(&LookedAhead); continue; }
        if (Matching(MenuPUSH))  { Push(&LookedAhead);           continue; }
        if (Matching(MenuRUN))   { RunSpice();                   continue; }
        if (Matching(MenuSAVE))  { SourceDone = False;
                                   WriteCell();                  continue; }
        if (Matching(MenuSELEC)) { Sel(&LookedAhead);            continue; }
        if (Matching(MenuSTRCH)) { StretchPath(&LookedAhead);    continue; }
        if (Matching(MenuSUBCT)) { Subcircuit(&LookedAhead);     continue; }
        if (Matching(MenuUNDO))  { Undo();                       continue; }
        if (Matching(MenuWIRES)) { Wires(&LookedAhead);          continue; }
    }
}


int
SafeCmds(LookedAhead)

int *LookedAhead;
{

    if (Matching(Menu0))     { Rotat0();                    return (True); }
    if (Matching(Menu90))    { Rotat90();                   return (True); }
    if (Matching(Menu180))   { Rotat180();                  return (True); }
    if (Matching(Menu270))   { Rotat270();                  return (True); }
    if (Matching(MenuBGRND)) { Bgrnd(LookedAhead);          return (True); }
    if (Matching(MenuCOLOR)) { Color(LookedAhead);          return (True); }
    if (Matching(MenuDIR))   { Dir();                       return (True); }
    if (Matching(MenuEXPND)) { Expand();                    return (True); }
    if (Matching(MenuGRID))  { SetGrid(LookedAhead);        return (True); }
    if (Matching(MenuHCOPY)) { Hcopy();                     return (True); }
    if (Matching(MenuHLITE)) { Hlite(LookedAhead);          return (True); }
    if (Matching(MenuIPLOT)) { DoIplot();                   return (True); }
    if (Matching(MenuLAST))  { LastView();                  return (True); }
    if (Matching(MenuMENU))  { MenuParams(LookedAhead);     return (True); }
    if (Matching(MenuMINSB)) { AlterColor('b','-');         return (True); }
    if (Matching(MenuMINSG)) { AlterColor('g','-');         return (True); }
    if (Matching(MenuMINSR)) { AlterColor('r','-');         return (True); }
    if (Matching(MenuMORE))  { MenuMore(LookedAhead);       return (True); }
    if (Matching(MenuMX))    { MX();                        return (True); }
    if (Matching(MenuMY))    { MY();                        return (True); }
    if (Matching(MenuPAN))   { Pan(LookedAhead);            return (True); }
    if (Matching(MenuPLUSB)) { AlterColor('b','+');         return (True); }
    if (Matching(MenuPLUSG)) { AlterColor('g','+');         return (True); }
    if (Matching(MenuPLUSR)) { AlterColor('r','+');         return (True); }
    if (Matching(MenuPRMPT)) { MenuPrompt(LookedAhead);     return (True); }
    if (Matching(MenuRDRAW)) { Rdraw();                     return (True); }
    if (Matching(MenuRGB))   { RGB();                       return (True); }
    if (Matching(MenuTERMS)) { ShowTerminals();             return (True); }
    if (Matching(MenuUPDAT)) { do_update();                 return (True); }
    if (Matching(MenuVIEW))  { ShowFull();                  return (True); }
    if (Matching(MenuWINDO)) { Windo(LookedAhead);          return (True); }
    if (Matching(MenuZOOM))  { Zoom(LookedAhead);           return (True); }

    return (False);
}

static char *PathFromFile;

static void
do_update()

/* update the init file */
{
    /* set path to what was in file for stand-alone sced */
    PSetPath(PathFromFile);
    UpdateSCED();
    SetPathFromSpice();
}



/***********************************************************************
 *
 * Routines for running spice analysis.
 *
 ***********************************************************************/


void
RunSpice()

{
    char *TypeIn;
    wordlist *wl;
    struct line *deck;
    static char Rcmd[64];
    RETSIGTYPE (*intrpt)();

    if (Parameters.kpCellDesc != Parameters.kpTopDesc) {
        ShowPrompt("Pop to top level first.");
        return;
    }
    MenuSelect(MenuRUN);


    if ((*ft_sced->sc_inprogress)()) {
        ShowPrompt("Resume run in progress? [y] ");
        if ((TypeIn = FBEdit(NULL)) == NULL) {
            MenuDeselect(MenuRUN);
            ErasePrompt();
            return;
        }
        if (*TypeIn != 'n' && *TypeIn != 'N') {
            if (IplotStatusChanged) {
                ClearIplot();
                SetIplot();
            }
            intrpt = signal(SIGINT,sced_interrupt);
            (*ft_sced->sc_resume)(NULL);
            (void)signal(SIGINT,intrpt);
            if (!dispdev->windows && !dispdev->hardcopy)
                if ((*ft_sced->sc_isiplot)())
                    RedisplaySCED(currentgraph);
            (void)CheckError();
            if (!(*ft_sced->sc_inprogress)()) {
                ShowPrompt("Analysis complete.");
                ClearIplot();
            }
            MenuDeselect(MenuRUN);
            return;
        }
        
    }

    ClearIplot();
    deck = MakeSpiceDeck();
    fix_spice_deck(deck);
    if (SourceDone)
        (*ft_sced->sc_cktclear)();
    (*ft_sced->sc_source)(deck,False,Parameters.kpCellName);
    (void)CheckError();
    SourceDone = True;

    wl = GetAnalysisFromDeck();
    if (wl == NULL) {
        ShowPrompt("Enter analysis command : ");
        if ((TypeIn = FBEdit(Rcmd)) != NULL) {
            wl = cp_lexer(TypeIn);
            strncpy(Rcmd,TypeIn,64);
            Rcmd[63] = '\0';
        }
        else
            ErasePrompt();
    }
    if (wl) {
        if (wl->wl_word && *wl->wl_word) {
            ShowPrompt("Performing analysis, please wait.");
            SetIplot();
            intrpt = signal(SIGINT,sced_interrupt);
            (*ft_sced->sc_dosim)(wl->wl_word,wl->wl_next);
            (void)signal(SIGINT,intrpt);
            if (!dispdev->windows && !dispdev->hardcopy)
                if ((*ft_sced->sc_isiplot)())
                    RedisplaySCED(currentgraph);
            (void)CheckError();
            if (!(*ft_sced->sc_inprogress)()) {
                ShowPrompt("Analysis complete.");
                ClearIplot();
            }
        }
        wl_free(wl);
    }
    MenuDeselect(MenuRUN);
}


static RETSIGTYPE
sced_interrupt()

{
    (void)signal(SIGINT,sced_interrupt);
    ft_intrpt = True;
}


static void
fix_spice_deck(deck)

/* Concatenate continuation lines for spice. */
struct line *deck;
{
    struct line *d,*d0,*d1,*dn;
    char *s;

    for (d = deck; d; d = dn) {
        dn = d->li_next;
        if (*d->li_line == '+') {
            /* continuation line */
            d->li_next = NULL;
            s = d0->li_line;
            d0->li_line = AllocateNewString(s,d->li_line+1);
            d0->li_next = dn;
            if (d0->li_actual == NULL) {
                d0->li_actual = AllocateLine(s);
                d0->li_actual->li_linenum = d0->li_linenum;
                d0->li_actual->li_next = d;
            }
            else {
                for (d1 = d0->li_actual; d1->li_next; d1 = d1->li_next) ;
                d1->li_next = d;
            }
            continue;
        }
        d0 = d;
    }
}



/***********************************************************************
 *
 * Routines to check for error conditions.
 *
 ***********************************************************************/

int
CheckSaved()

/* Determine if the circuit, and parent (if any) have been saved,
 * and query user on intentions.
 *  return values:
 *  0   saved, quit
 *  1   not saved, quit anyway
 *  -1  not saved, don't quit
 */
{
    int i,j;
    char *TypeIn;

    MenuSelect(MenuABORT);
    i = CheckModified();
    j = (Parameters.kpModified And Parameters.kpCellName[0]);
    if (i And j)
        TypeIn  =
        "Both parent and current circuit have changed.  Quit anyway? (n): ";
    elif (i And Not j)
        TypeIn  = "Parent circuit has been modified.  Quit anyway? (n): ";
    elif (Not i And j)
        TypeIn  = "Current circuit has been modified.  Quit anyway? (n): ";
    else {
        MenuDeselect(MenuABORT);
        return (0);
    }
    ShowPrompt(TypeIn);
    TypeIn = FBEdit(NULL);
    if (TypeIn != NULL And (TypeIn[0] == 'Y' Or TypeIn[0] == 'y')){
        MenuDeselect(MenuABORT);
        return (1);
    }
    ErasePrompt();
    MenuDeselect(MenuABORT);
    return (-1);
}


int
CheckError()

/* find out if spice has put anything in the error output file.  If
 * so, display it and return True.
 */
{
    int c;
    struct ka *AOI,BB;

    rewind(cp_err);
    if ((c = getc(cp_err)) == EOF) return (False);
    ungetc(c,cp_err);
    if (!PopUpErrMessage(cp_err)) {
        AOI = View->kvLargeCoarseViewport;
        FBEraseBox(AOI->kaLeft,AOI->kaBottom,
            AOI->kaRight,AOI->kaTop);
        MoreText((int)AOI->kaLeft-1,(int)AOI->kaBottom-1,
            (int)AOI->kaRight-1,(int)AOI->kaTop-1,cp_err,MoreTextColor);
        RedisplaySCED(currentgraph);
    }
    rewind(cp_err);
#ifdef HAVE_NO_FTRUNCATE
    /* have to add something here */
#else
    ftruncate(fileno(cp_err),0);
#endif
    return (True);
}


int
IsSymFile(fp)

/* Check if the file just opened is a symbol file. */
FILE *fp;
{
    char buf[12];

    (void)fgets(buf,8,fp);

    if (buf[0] != '(') return (False);
    if (buf[1] != 'S') return (False);
    if (buf[2] != 'y') return (False);
    if (buf[3] != 'm') return (False);
    if (buf[4] != 'b') return (False);
    if (buf[5] != 'o') return (False);
    if (buf[6] != 'l') return (False);
    return (True);
}



/***********************************************************************
 *
 * Routines to set the search path as obtained from spice, and get
 * device name for haedcopy.
 *
 ***********************************************************************/


void
SetPathFromSpice()

/* Set the CD path to the spice search path. */
{
    char buf[80], *t;
    struct variable *v;
    extern char *kw_sourcepath;
    int len;

    TypeOut[0] = '\0';

    if (!cp_getvar(kw_sourcepath,VT_LIST,(char *)&v)) {
        CDPath(".");
        return;
    }
    len = 256;
    t = tmalloc(len);
    switch (v->va_type) {
        case VT_STRING:
            sprintf(t,"%s",v->va_string);
            break;
        case VT_NUM:
            sprintf(t,"%d",v->va_num);
            break;
        case VT_REAL:
            sprintf(t,"%lg",v->va_real);
            break;
    }
    for (v = v->va_next; v; v = v->va_next) {
        switch (v->va_type) {
            case VT_STRING:
                sprintf(buf," %s",v->va_string);
                break;
            case VT_NUM:
                sprintf(buf," %d",v->va_num);
                break;
            case VT_REAL:
                sprintf(buf," %lg",v->va_real);
                break;
        }
        if (strlen(t) + strlen(buf) >= len)
            t = trealloc(t, len += 256);
        strcat(t,buf);
    }
    /* append the path to the sced system files */
    if (strlen(t) + strlen(Sced_Path) > len)
        t = trealloc(t, len += 256);
    strcat(t," ");
    strcat(t,Sced_Path);
    CDPath(t);
    txfree(t);
}


void
SetCDPath(Path)

char *Path;
{
    if (PathFromFile != NULL)
        free(PathFromFile);
    PathFromFile = CopyString(Path);
}


char *
GetHardCopyDevice()

/* Return the name of the device used for output as set in spice. */
{
    struct variable *v;
    extern char *kw_hcopydev;

    if (!cp_getvar(kw_hcopydev, VT_STRING, (char *)&v))
        return ("prn");
    return (v->va_string);
}



/***********************************************************************
 *
 * Routines needed to satisfy references in SCED library, for error and
 * signal handling.
 *
 ***********************************************************************/

void
InitSignals()
{
    ft_intrpt = false; 
    vec_gc();
}

char *
NextCellName()
{return (NULL);}


void
MallocFailed()

{
    char *cp;
    char *mktemp();
    cp = mktemp("scXXXXXX");
    DevHalt();
    fprintf(stderr,"OUT OF MEMORY.  SCED is DEAD!\n");

    if (Parameters.kpCellName[0] != EOS) {
        if (Not CDUpdate(Parameters.kpCellDesc,cp))
            fprintf(stderr,"\007Sorry.  Couldn't save current circuit.\n");
        else
            fprintf(stderr,"\007Current circuit saved in %s.\n",cp);
    }
    longjmp(errjbuf,1);
}



/***********************************************************************
 *
 * Routines which allocate memory from spice, etc.
 *
 ***********************************************************************/

extern char *copy();


struct line *
AllocateLine(string)

/* create and initialize a line structure */
char *string;
{
    struct line *l;

    /* allocate memory in spice */

    l = alloc(line);
    l->li_line = copy(string);
    return (l);
}


char *
AllocateNewString(string,buf)

/* Create a new concatenated string from a previous string and buffer.
 */
char *string;
char *buf;
{
    char *c;

    c = tmalloc(strlen(string) + strlen(buf) + 2);
    sprintf(c,"%s %s",string,buf);
    return (c);
}



/***********************************************************************
 *
 * Routines to execute spice commands, or enter a shell;
 * 
 *
 ***********************************************************************/

#ifdef __STDC__
static int  do_spice_cmd(char*);
#else
static int  do_spice_cmd();
#endif

/* imports from spice */
extern int  cp_evloop();
extern void cp_pushcontrol();
extern void cp_popcontrol();
extern FILE *cp_out;
extern FILE *cp_in;


void
ShowProcess(cp)

char *cp;
{
    char buffer[80];
    int i, gicalls;

    if (!cp || !strlen(cp)) {
        if (!dispdev->windows) {
#ifdef MSDOS
            cp = "command";
#else
            cp = "csh";
#endif
            DevHalt();
            system(cp);
            DevNewViewport(currentgraph);
            RedisplaySCED(currentgraph);
        }
        else
            system("xterm &");
        return;
    }
    gicalls = numgraphcxsw;
    i = do_spice_cmd(cp);

    if (!CheckError()) {
        if (!dispdev->windows) {
            /* CheckError() redraws screen if error */
            if (i) {
                /* Cmd showed some text */
                ShowPrompt("Hit any key to continue.");
                (void)FBGetchar(ERASE);
                ErasePrompt();
            }
            if (gicalls != numgraphcxsw)
                InitVLT();
            if (i || gicalls != numgraphcxsw)
                RedisplaySCED(currentgraph);
        }
    }
}


static int
do_spice_cmd(Cmd)

char *Cmd;
{
    *TypeOut = '\0';
    (void)EnableMore(True);
    cp_pushcontrol();
    cp_evloop(Cmd);
    cp_popcontrol();
    return (EnableMore(False));
}


int
SCEDline(string)

/* Called from output.c in spice, redirects output to
 * graphics window.
 */
char *string;
{
    if (SCEDisActive) {
        if (strlen(TypeOut) + strlen(string) < 200)
            strcat(TypeOut,string);
        else
            strcat(TypeOut,"\n");
        if (strchr(TypeOut,'\n') == NULL) return;
        if (MoreLine(TypeOut))
            EnableMore(False);
        *TypeOut = '\0';
        return (True);
    }
    return (False);
}


void
SCEDfflush()

{
    if (!SCEDisActive) {
        fflush(cp_out);
        return;
    }
    strcat(TypeOut,"\n");
    if (MoreLine(TypeOut))
        EnableMore(False);
    *TypeOut = '\0';
}


char *
SCEDfgets(s,n,fp,prompt)

/* replacement for fgets() for use in spice */
char *s;
int n;
FILE *fp;
char *prompt;
{
    char *c;

    if (!SCEDisActive) {
        if (prompt != NULL) {
            (void)fprintf(cp_out,prompt);
            (void)fflush(cp_out);
        }
        c = fgets(s,n,fp);
        if (c == NULL) {
            clearerr(fp);
        }
        return (c);
    }
    if (!prompt) {
        ErasePrompt();
        return (NULL);
    }
    ShowPrompt(prompt);
    MorePageDisplay();
    c = FBEdit(NULL);
    ErasePrompt();
    if (c == NULL || *c == '\0' || *c == '\n') return (NULL);
    strncpy(s,c,n);
    return (s);
}


enum Active
SCEDactive()

{
    return (SCEDisActive);
}



/***********************************************************************
 *
 * Help interface.
 * 
 *
 ***********************************************************************/

void
SCEDhelp()

{
    int dummy;
    char cmd[32];

    MenuSelect(MenuHELP);
    ShowPrompt("Point to menu item for help (ESC to exit help).");
    SCEDisActive = SCED_inhelp;
    loop {
        switch (PointLoopSafe(&dummy)) {
            case PL_ESC:
                MenuDeselect(MenuHELP);
                if (!dispdev->windows)
                    RedisplaySCED(currentgraph);
                else
                    ErasePrompt();
                SCEDisActive = SCED_active;
                return;
            case PL_PCW:
                continue;
            case PL_CMD:
                if (Matching(MenuMINSR) ||
                    Matching(MenuPLUSR) ||
                    Matching(MenuMINSG) ||
                    Matching(MenuPLUSG) ||
                    Matching(MenuMINSB) ||
                    Matching(MenuPLUSB))
                    strcpy(Parameters.kpCommand,"chgcolor");

                else

                if (Matching(Menu0)   ||
                    Matching(Menu90)  ||
                    Matching(Menu180) ||
                    Matching(Menu270))
                    strcpy(Parameters.kpCommand,"rotate");

                else

                if (Matching(MenuMX) ||
                    Matching(MenuMY))
                    strcpy(Parameters.kpCommand,"mirror");

                else

                if (Matching(MenuSpace))
                    continue;

                sprintf(cmd,"help sced:%s",Parameters.kpCommand);
                (void)do_spice_cmd(cmd);
        }
    }
}
