/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1986 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Routines to query and alter devices.
 */

#include "spice.h"
#include "ftedefs.h"
#include "spfteext.h"

#ifdef __STDC__
static wordlist *devexpand(char*);
static void parseline(wordlist*,wordlist**,wordlist**);
#else
static wordlist *devexpand();
static void parseline();
#endif


/* Display various device parameters.  The syntax of this command is 
 *   show [devicelist] [, parmlist]
 * where devicelist can be "all", device names, or glob-matched subnames
 * using wildcards *,?,[].  The parms are names of parameters that should
 * be valid for all the named devices, or "all".  Defaults to all, all.
 */

void
com_show(wl)

wordlist *wl;
{
    wordlist *devs, *parms, *keyword;
    wordlist *tw, *pw, *kw, *ww, *w0;
    struct variable *v, *vv;
    int len;
    char *neww;

    if (!ft_curckt) {
        fprintf(cp_err, "Error: no circuit loaded\n");
        return;
    }
        parseline(wl,&devs,&parms);

    if (parms == NULL) {
        parms = alloc(wordlist);
        parms->wl_word = copy("all");
    }

    if (devs == NULL)
        devs = cp_cctowl(ft_curckt->ci_devices);

    wl_sort(devs);
    out_init();

    for (tw = devs; tw; tw = tw->wl_next) {
        out_printf("\n%s:\n", tw->wl_word);

        for (pw = parms; pw; pw = pw->wl_next) {

            keyword = NULL;
            neww = copy(tw->wl_word); /* address put in symtab */
            vv = if_getparam((char*)ft_curckt->ci_ckt, &neww,
                pw->wl_word, 0, &keyword);

            for (v = vv, kw = keyword; v; v = v->va_next, kw = kw->wl_next) {
                len = 14 + strlen(kw->wl_word);
                out_printf("  %s =",kw->wl_word);
                w0 = cp_varwl(v);
                for (ww = w0; ww; ww = ww->wl_next) {
                    len += strlen(ww->wl_word) + 1;
                    out_printf(" %s", ww->wl_word);
                }
                wl_free(w0);

                for (; len < 40; len++)
                    out_send(" ");
                out_printf(" %s\n",v->va_name);
            }
            wl_free(keyword);
            va_free(vv);
        }
    }
    out_send("\n");

    wl_free(devs);
    wl_free(parms);
}


/* Alter a device parameter.  The syntax here is
 *   alter devicelist : parmname value [parmname value] ...
 * where devicelist is as above, parmname is the name of the desired parm
 * and value is a string, numeric, or bool value.
 */

void
com_alter(wl)

wordlist *wl;
{
    wordlist *devs, *parms, *tw;

    if (!ft_curckt) {
        fprintf(cp_err, "Error: no circuit loaded\n");
        return;
    }

    parseline(wl,&devs,&parms);

    /*
        for (tw = devs; tw; tw = tw->wl_next)
                if_alter(tw->wl_word,parms);
    */

    wl_free(devs);
    wl_free(parms);
}


static void
parseline(line,dlist,plist)
wordlist *line,**dlist,**plist;
{
    wordlist *tw, *tl, *t0;
    char *c = NULL;

    *plist = NULL;
    *dlist = NULL;
    if (!line)
        return;
    tl = tw = wl_copy(line);

    while (tw) {
        if ((c = strchr(tw->wl_word,',')) != NULL) break;
        tw = tw->wl_next;
    }
    if (c) {
        if (*(c+1) == '\0') {
            *c = '\0';
            *plist = tw->wl_next;
            if (tw->wl_next)
                tw->wl_next->wl_prev = NULL;
            tw->wl_next = NULL;
            if (*tw->wl_word == '\0') {
                if (tw->wl_prev)
                    tw->wl_prev->wl_next = NULL;
                if (tl == tw)
                    tl = NULL;
                tfree(tw);
            }
        }
        else if (c == tw->wl_word) {
            tw->wl_word = copy(c+1);
            tfree(c);
            *plist = tw;
            if (tw->wl_prev)
                tw->wl_prev->wl_next = NULL;
            tw->wl_prev = NULL;
        }
        else {
            *plist = alloc(wordlist);
            (*plist)->wl_word = copy(c+1);
            (*plist)->wl_next = tw->wl_next;
            if (tw->wl_next)
                tw->wl_next->wl_prev = *plist;
            *c = '\0';
            tw->wl_next = NULL;
        }
    }

    /* Now expand the devicelist... */
    t0 = tl;
    for (tw = NULL; tl; tl = tl->wl_next) {
        inp_casefix(tl->wl_word);
        tw = wl_append(tw, devexpand(tl->wl_word));
    }
    wl_free(t0);
    *dlist = tw;
}


/* Given a device name, possibly with wildcards, return the matches. */

static wordlist *
devexpand(name)

char *name;
{
    wordlist *wl, *d, *devices, *tw;

    if (strchr(name, '*') || strchr(name, '[') || strchr(name, '?')) {
        devices = cp_cctowl(ft_curckt->ci_devices);
        d = devices;
        for (wl = NULL; d; d = d->wl_next) {
            if (cp_globmatch(name, d->wl_word)) {
                tw = alloc(struct wordlist);
                if (wl) {
                    wl->wl_prev = tw;
                    tw->wl_next = wl;
                    wl = tw;
                }
                else
                    wl = tw;
                wl->wl_word = copy(d->wl_word);
            }
        }
        wl_free(devices);
    }
    else if (eq(name, "all")) {
        wl = cp_cctowl(ft_curckt->ci_devices);
    }
    else {
        wl = alloc(struct wordlist);
        wl->wl_word = copy(name);
    }
    wl_sort(wl);
    return (wl);
}

