/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Stuff for dealing with spice input decks and command scripts, and
 * the listing routines.
 */

#include "spice.h"
#ifdef HAVE_STAT
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include "ftedefs.h"
#include "inpdefs.h"
#include "fteinp.h"
#include "spfteext.h"

#ifdef __STDC__
static char *upper(char*);
static void spsource(FILE*,bool,bool,char*);
static void decksource(struct line*,bool,bool,char*);
static bool is_ckt(struct line*);
static wordlist *get_speccmds(struct line*);
static wordlist *get_controls(struct line*,bool);
static void varsub(char**);
static wordlist *wl_separate(char*);
static char *flatten(wordlist*);
#else
static char *upper();
static void spsource();
static void decksource();
static bool is_ckt();
static wordlist *get_speccmds();
static wordlist *get_controls();
static void varsub();
static wordlist *wl_separate();
static char *flatten();
#endif


/* Do a listing. Use is listing [expanded] [logical] [physical] [deck] */

void
com_listing(wl)

wordlist *wl;
{
    int type = LS_LOGICAL;
    bool expand = false;
    char *s;

    if (ft_curckt) {
        while (wl) {
            s = wl->wl_word;
            switch (*s) {
                case 'l':
                case 'L':
                    type = LS_LOGICAL;
                    break;
                case 'p':
                case 'P':
                    type = LS_PHYSICAL;
                    break;
                case 'd':
                case 'D':
                    type = LS_DECK;
                    break;
                case 'e':
                case 'E':
                    expand = true;
                    break;
                default:
                    fprintf(cp_err,
                    "Error: bad listing type %s\n", s);
            }
            wl = wl->wl_next;
        }
        inp_list(cp_out, expand ? ft_curckt->ci_deck :
                ft_curckt->ci_origdeck, ft_curckt->ci_options,
                type);
    }
    else
        fprintf(cp_err, "Error: no circuit loaded.\n");
    return;
}


static char *
upper(string)

register char *string;
{

    static char buf[BSIZE_SP];
    register char *s;

    if (string) {
        strncpy(buf, string, BSIZE_SP - 1);
        buf[BSIZE_SP - 1] = 0;
        inp_casefix(buf);
    }
    else {
        strcpy(buf, "<null>");
    }

    return buf;

}


/* Provide an input listing on the specified file of the given
 * card deck.  The listing should be of either LS_PHYSICAL or LS_LOGICAL
 * or LS_DECK lines as specified by the type parameter.
 */

void
inp_list(file, deck, extras, type)

FILE *file;
struct line *deck, *extras;
int type;
{
    struct line *here;
    struct line *there;
    struct line *tmp, *next;
    bool renumber;
    bool useout = (file == cp_out);
    int i = 1;
    extern char *kw_renumber;

    if (useout)
        out_init();

    if (type != LS_DECK)
        if (useout) {
            out_printf("\t%s\n", ft_curckt->ci_name);
            if (ft_curckt->ci_contblk)
                out_printf("\tBound codeblock: %s\n\n",ft_curckt->ci_contblk);
            else
                out_send("\n");
        }
        else {
            fprintf(cp_out, "\t%s\n", ft_curckt->ci_name);
            if (ft_curckt->ci_contblk)
                fprintf(cp_out,
                    "\tBound codeblock: %s\n\n",ft_curckt->ci_contblk);
            else
                fputc('\n',cp_out);
        }


    (void) cp_getvar(kw_renumber, VT_BOOL, (char *) &renumber);
    if (type == LS_LOGICAL) {
top1:
        for (here = deck; here; here = here->li_next) {
            if (renumber)
                here->li_linenum = i;
            i++;
            if (ciprefix(".end", here->li_line) &&
                    !isalpha(here->li_line[4]))
                continue;
            if (*here->li_line != '*') {
                if (useout) {
                    out_printf("%6d : %s\n",
                            here->li_linenum,
                            upper(here->li_line));
                }
                else
                    fprintf(file, "%6d : %s\n",
                            here->li_linenum,
                            upper(here->li_line));
                if (here->li_error) {
                    if (useout) {
                        out_printf("%s\n",
                                here->li_error);
                    }
                    else
                        fprintf(file, "%s\n",
                            here->li_error, file);
                }
            }
        }
        if (extras) {
            deck = extras;
            extras = NULL;
            goto top1;
        }
        if (useout) {
            out_printf("%6d : .end\n", i);
        }
        else
            fprintf(file, "%6d : .end\n", i);
    }
    else if ((type == LS_PHYSICAL) || (type == LS_DECK)) {
top2:
        for (here = deck; here; here = here->li_next) {
            if ((here->li_actual == NULL) || (here == deck)) {
                if (renumber)
                    here->li_linenum = i;
                i++;
                if (ciprefix(".end", here->li_line) &&
                        !isalpha(here->li_line[4]))
                    continue;
                if (type == LS_PHYSICAL) {
                    if (useout) {
                        out_printf("%6d : %s\n",
                            here->li_linenum,
                            upper(here->li_line));
                    }
                    else
                        fprintf(file, "%6d : %s\n",
                            here->li_linenum,
                            upper(here->li_line));
                }
                else {
                    if (useout)
                        out_printf("%s\n",
                            upper(here->li_line));
                    else
                        fprintf(file, "%s\n",
                            upper(here->li_line));
                }
                if (here->li_error && (type == LS_PHYSICAL)) {
                    if (useout)
                        out_printf("%s\n",
                            here->li_error);
                    else
                        fprintf(file, "%s\n",
                            here->li_error);
                }
            }
            else {
                for (there = here->li_actual; there;
                        there = there->li_next) {
                    there->li_linenum = i++;
                    if (ciprefix(".end", here->li_line) &&
                            isalpha(here->li_line[4]))
                        continue;
                    if (type == LS_PHYSICAL) {
                        if (useout) {
                            out_printf("%6d : %s\n",
                                there->li_linenum,
                                upper(there->li_line));
                        }
                        else
                            fprintf(file, "%6d : %s\n",
                                there->li_linenum,
                                upper(there->li_line));
                    }
                    else {
                        if (useout)
                            out_printf("%s\n",
                                upper(there->li_line));
                        else
                            fprintf(file, "%s\n",
                                upper(there->li_line));
                    }
                    if (there->li_error && 
                            (type == LS_PHYSICAL)) {
                        if (useout)
                            out_printf("%s\n",
                            there->li_error);
                        else
                            fprintf(file, "%s\n",
                            there->li_error);
                    }
                }
                here->li_linenum = i;
            }
        }
        if (extras) {
            deck = extras;
            extras = NULL;
            goto top2;
        }
        if (type == LS_PHYSICAL) {
            if (useout) {
                out_printf("%6d : .end\n", i);
            }
            else
                fprintf(file, "%6d : .end\n", i);
        }
        else {
            if (useout)
                out_printf(".end\n");
            else
                fprintf(file, ".end\n");
        }
    }
    else
        fprintf(cp_err, "inp_list: Internal Error: bad type %d\n",type);
    return;
}


/* The routine to source a spice input deck. We read the deck in, take out
 * the front-end commands, and create a CKT structure. Also we filter out
 * the following cards: .save, .width, .four, .print, and .plot, to perform
 * after the run is over.
 */
    
void
inp_spsource(fp, comfile, filename)

FILE *fp;
bool comfile;
char *filename;
{
    spsource(fp, comfile, comfile, filename);
}


static void
spsource(fp, nospice, nocmds, filename)

FILE *fp;
bool nospice, nocmds;
char *filename;
{
    struct line *deck;
    wordlist wl;
    char *buf;
    
    if ((buf = readline(fp)) == NULL)
        return;
    if (prefix("(Symbo",buf)) {
        /* file contains symbolic layout information */
        txfree(buf);
        while ((buf = readline(fp)) != NULL) {
            if (eq(buf,"E\n")) {
                txfree(buf);
                break;
            }
            txfree(buf);
        }
        if ((buf = readline(fp)) == NULL)
            return;
    }
    if (prefix(".check",buf)) {
        /* operating range analysis file */
        wl.wl_word = filename;
        wl.wl_next = wl.wl_prev = NULL;
        txfree(buf);
        ft_check(&wl,fp);
        return;
    }
#ifdef HAVE_STIM
    if (prefix(".stim",buf)) {
        /* stimulus description file */
        txfree(buf);
        MB_run(fp,1,filename);
        return;
    }
#endif

    inp_readall(fp,&deck,buf);
    if (deck)
        decksource(deck, nospice, nocmds, filename);
}

void
inp_decksource(deck, comfile, filename)

struct line *deck;
bool comfile;
char *filename;
{
    decksource(deck,comfile,comfile,filename);
}


static void
decksource(deck, nospice, nocmds, filename)

/* Source a deck.  If nospice is true, execute the commands in the deck
 * without saving the spice text, if any.  If nocmds is true, save the
 * spice text but discard the commands.  If both flags are true, then
 * execute all commands up until a .endc line, the deck is assumed to
 * come from an init file.  If neither flag is true, source the spice
 * text and execute the commands.
 */
struct line *deck;
bool nospice, nocmds;
char *filename;
{
    wordlist *wl, *controls;
    FILE *lastin, *lastout, *lasterr;

    if (!deck)
        return;

    controls = get_controls(deck,nospice && nocmds);
    if (!nospice && is_ckt(deck)) {
        out_init();
        if (!ft_servermode && deck->li_line && *deck->li_line)
            out_printf("\nCircuit: %s\n\n", deck->li_line);
        if (inp_spdeck(deck,filename)) {
            wl_free(controls);
            return;
        }
    }
    else
        inp_deckfree(deck);

    /* Now do the commands */
    if (controls) {
        if (!nocmds || nospice) {

            lastin = cp_curin;
            lastout = cp_curout;
            lasterr = cp_curerr;
            cp_curin = cp_in;
            cp_curout = cp_out;
            cp_curerr = cp_err;
            cp_pushcontrol();

            for (wl = controls; wl; wl = wl->wl_next)
                (void) cp_evloop(wl->wl_word);

            cp_popcontrol();
            cp_curin = lastin;
            cp_curout = lastout;
            cp_curerr = lasterr;
        }
        wl_free(controls);
    }
    return;
}


static bool
is_ckt(deck)

/* return true if there is a circuit */
struct line *deck;
{
    struct line *l;
    char *s;

    for (l = deck->li_next; l; l = l->li_next) {
        s = l->li_line;
        if (!s) {
            continue;
        }
        while (isspace(*s))
            s++;
        if (!*s || *s == '*' || *s == '#') {
            continue;
        }
        /* found something real */
        return (true);
    }
    return (false);
}


bool
inp_spdeck(deck,filename)

struct line *deck;
char *filename;
{
    struct line *dd, *realdeck, *options;
    wordlist *end;
    char *tt;
    bool nosubckts = false;
    extern char *kw_nosubckt;

    realdeck = inp_deckcopy(deck);
    tt = copy(deck->li_line);

    /* substitute for shell variables in spice text */
    for (dd = deck->li_next; dd; dd = dd->li_next) {
        if (*dd->li_line == '*') continue;
        if (strchr(dd->li_line,'$'))
            varsub(&dd->li_line);
    }
    end = get_speccmds(deck);
    options = inp_getopts(deck);

    if (deck->li_next) {

        /* Now expand subcircuit macros. Note that we have to 
         * fix the case before we do this but after we 
         * deal with the commands.
         */
        if (!cp_getvar(kw_nosubckt, VT_BOOL, (char *)&nosubckts)) {
            dd = inp_subcktexpand(deck->li_next);
            if (dd == (struct line *)NULL) {
                inp_deckfree(deck);
                inp_deckfree(realdeck);
                inp_deckfree(options);
                wl_free(end);
                tfree(tt);
                fprintf(cp_err,
            "Error during subcircuit expansion, circuit not loaded.\n");
                return (true);;
            }
            deck->li_next = dd;
        }
        deck->li_actual = realdeck;
        inp_dodeck(deck, tt, end, false, options, copy(filename));
    }
    else {
        fprintf(cp_err, "Error: no lines in input\n");
        tfree(tt);
        wl_free(end);
        inp_deckfree(options);
        inp_deckfree(realdeck);
        inp_deckfree(deck);
        return (true);
    }
    return (false);
}


static wordlist *
get_speccmds(deck)

struct line *deck;
{
    struct line *dd, *ld;
    wordlist *wl, *end = NULL;
    char *s;

    ld = deck;
    for (dd = deck->li_next; dd; dd = ld->li_next) {
        if (dd->li_line[0] == '*') {
            ld = dd;
            continue;
        }
        s = dd->li_line;
        while (isspace(*s)) s++;
        inp_casefix(dd->li_line);

        if (prefix(".width",s) || prefix(".four",s) ||
                prefix(".plot",s) || prefix(".print",s) ||
                prefix(".save",s)) {
            if (end) {
                end->wl_next = alloc(struct wordlist);
                end->wl_next->wl_prev = end;
                end = end->wl_next;
            }
            else
                wl = end = alloc(struct wordlist);
            end->wl_word = copy(dd->li_line);
            ld->li_next = dd->li_next;
            txfree(dd->li_line);
            txfree((char*)dd);
        }
        else
            ld = dd;
    }
    return (end);
}
        

static wordlist *
get_controls(deck,allcmds)

/* strip the control lines out of the deck and return as a wordlist */
struct line *deck;
bool allcmds;
{
    bool commands = allcmds;
    struct line *dd, *ld;
    char *s;
    wordlist *wl, *controls = NULL;

    ld = deck;
    for (dd = deck->li_next; dd; dd = ld->li_next) {
        if (ciprefix(".control", dd->li_line)) {
            ld->li_next = dd->li_next;
            txfree(dd->li_line);
            txfree((char*)dd);
            if (commands && !allcmds)
                fprintf(cp_err,
                    "Warning: redundant .control card\n");
            commands = true;
        }
        else if (ciprefix(".endc", dd->li_line)) {
            ld->li_next = dd->li_next;
            txfree(dd->li_line);
            txfree((char*)dd);
            if (commands)
                commands = false;
            else
                fprintf(cp_err, 
                "Warning: misplaced .endc card\n");
        }
        else if (!*dd->li_line ||
            (commands && (*dd->li_line == '#' || *dd->li_line == '*'))) {
            /* So blank lines in com files don't get
             * considered as circuits.
             * Also kill comments in commands.
             */
            ld->li_next = dd->li_next;
            txfree(dd->li_line);
            txfree((char*)dd);
        }
        else if (commands || prefix("*#", dd->li_line)) {
            wl = alloc(struct wordlist);
            if (controls) {
                wl->wl_next = controls;
                controls->wl_prev = wl;
                controls = wl;
            }
            else
                controls = wl;
            if (prefix("*#", dd->li_line))
                wl->wl_word = copy(dd->li_line + 2);
            else
                wl->wl_word = copy(dd->li_line);
            if (commands) {
                for (;;) {
                    /* look for lines concatenated with \ */
                    s = wl->wl_word + strlen(wl->wl_word) - 1;
                    while (isspace(*s)) s--;
                    if (*s == '\\') {
                        *s = '\0';
                        ld->li_next = dd->li_next;
                        txfree(dd->li_line);
                        txfree((char*)dd);
                        dd = ld->li_next;
                        s = tmalloc(
                            strlen(wl->wl_word) + strlen(dd->li_line) + 1);
                        strcpy(s,wl->wl_word);
                        strcat(s,dd->li_line);
                        txfree(wl->wl_word);
                        wl->wl_word = s;
                        continue;
                    }
                    break;
                }
            }
            ld->li_next = dd->li_next;
            txfree(dd->li_line);
            txfree((char*)dd);
        }
        else
            ld = dd;
    }
    if (!controls)
        return (NULL);
    return (wl_reverse(controls));
}


/* New Feature: substitute for $variables in the spice deck as it is
 * read.  The variables must have been defined previously, and (presently)
 * not in the same deck, i.e., not in control statements of the same file.
 */


static void
varsub(str)

/* replace the $variables */
char **str;
{
    wordlist *wl;

    wl = wl_separate(*str);
    cp_variablesubst(&wl);
    txfree(*str);
    *str = flatten(wl);
    wl_free(wl);
}


#define VALIDCHARS "_<#?@.()[]&"


static wordlist *
wl_separate(str)

/* separate out the $variables into a linked wordlist */
char *str;
{
    char *c, *d;
    wordlist *wl, *wl0 = NULL;
    int lev1, lev2;

    c = str;
    while (isspace(*c)) c++;

    while ((d = strchr(c,'$')) != NULL) {
        if (d != c) {
            if (wl0 == NULL) {
                wl = wl0 = alloc(wordlist);
            }
            else {
                wl->wl_next = alloc(wordlist);
                wl->wl_next->wl_prev = wl;
                wl = wl->wl_next;
            }
            wl->wl_word = tmalloc((d-c+1)*sizeof(char));
            strncpy(wl->wl_word,c,d-c);
            wl->wl_word[d-c] = '\0';
            c = wl->wl_word + (d-c-1);
            while(isspace(*c)) *c-- = '\0';
        }
        c = d;
        if (*++d == '\0') break;
        if (*++d == '\0') break;
        if (wl0 == NULL) {
            wl = wl0 = alloc(wordlist);
        }
        else {
            wl->wl_next = alloc(wordlist);
            wl->wl_next->wl_prev = wl;
            wl = wl->wl_next;
        }
        if (*d &&
            (isalphanum(*d) || strchr(VALIDCHARS, *d) || *d == cp_dol)) {
            d++;
        }
        lev1 = lev2 = 0;
        for ( ; *d; d++) {
            if (isalphanum(*d))
                continue;
            if (*d == '[') {
                lev1++;
                continue;
            }
            if (*d == '(') {
                lev2++;
                continue;
            }
            if (*d == ']') {
                if (lev1) lev1--;
                continue;
            }
            if (*d == ')') {
                if (lev2) lev2--;
                continue;
            }
            if (strchr(VALIDCHARS, *d))
                continue;
            if (lev1 || lev2)
                continue;
            break;
        }
        wl->wl_word = tmalloc((d-c+1)*sizeof(char));
        strncpy(wl->wl_word,c,d-c);
        wl->wl_word[d-c] = '\0';
        while (*d && isspace(*d)) d++;
        c = d;
    }
    if (*c) {
        if (wl0 == NULL) {
            wl = wl0 = alloc(wordlist);
        }
        else {
            wl->wl_next = alloc(wordlist);
            wl->wl_next->wl_prev = wl;
            wl = wl->wl_next;
        }
        wl->wl_word = copy(c);
    }
    return (wl0);
}


static char *
flatten(wl)

/* like wl_flatten(), but join tokens with leading '%' */
wordlist *wl;
{
    wordlist *ww;
    int len;
    char *str, *s;

    len = 1;
    for (ww = wl; ww; ww = ww->wl_next)
        len += strlen(ww->wl_word) + 1;
    str = tmalloc(len);
    *str = '\0';
    for (ww = wl; ww; ww = ww->wl_next) {
        s = ww->wl_word;
        if (*s == '%') {
            while (*s == '%') s++;
            strcat(str,s);
            continue;
        }
        if (ww != wl)
            strcat(str," ");
        strcat(str,s);
    }
    return (str);
}


/* This routine is cut in half here because com_rset has to do what follows
 * also. End is the list of commands we execute when the job is finished:
 * we only bother with this if we might be running in batch mode, since
 * it isn't much use otherwise.
 */

void
inp_dodeck(deck, tt, end, reuse, options, filename)

struct line *deck;
char *tt;
wordlist *end;
bool reuse;
struct line *options;
char *filename;
{
    struct circ *ct;
    struct line *dd;
    char *ckt, *s;
    INPtables *tab;
    struct variable *eev = NULL;
    wordlist *wl;
    bool noparse, ii;
    extern char *kw_noparse;

    /* First throw away any old error messages there might be and fix
     * the case of the lines.
     */
    for (dd = deck; dd; dd = dd->li_next) {
        if (dd->li_error) {
            tfree(dd->li_error);
            dd->li_error = NULL;
        }
    }
    for (dd = options; dd; dd = dd->li_next) {
        if (dd->li_error) {
            tfree(dd->li_error);
            dd->li_error = NULL;
        }
    }
    if (reuse) {
        ct = ft_curckt;
    }
    else {
        if (ft_curckt) {
            ft_curckt->ci_devices = cp_kwswitch(CT_DEVNAMES, 
                    (char *) NULL);
            ft_curckt->ci_nodes = cp_kwswitch(CT_NODENAMES, 
                    (char *) NULL);
        }
        ft_curckt = ct = alloc(struct circ);
    }
    (void) cp_getvar(kw_noparse, VT_BOOL, (char *) &noparse);
    if (!noparse)
        ckt = if_inpdeck(deck, (char**)&tab);
    else
        ckt = NULL;

    for (dd = deck; dd; dd = dd->li_next)
        if (dd->li_error) {
            if (cp_out != stdout)
                fprintf(cp_err, "Warning, line %d : %s\n%s\n",
                    dd->li_linenum, dd->li_line, dd->li_error);
            else
                out_printf("Warning, line %d : %s\n%s\n",
                    dd->li_linenum, dd->li_line, dd->li_error);
        }

    /* Add this circuit to the circuit list. If reuse is true then 
     * use the ft_curckt structure.
     */

    if (!reuse) {
        for (dd = deck->li_next; dd; dd = dd->li_next)
            if_setndnames(dd->li_line);

        /* Be sure that ci_devices and ci_nodes are valid */
        ft_curckt->ci_devices = cp_kwswitch(CT_DEVNAMES, 
                (char *) NULL);
        (void) cp_kwswitch(CT_DEVNAMES, ft_curckt->ci_devices);
        ft_curckt->ci_nodes = cp_kwswitch(CT_NODENAMES, (char *) NULL);
        (void) cp_kwswitch(CT_NODENAMES, ft_curckt->ci_nodes);
        ft_newcirc(ct);
        /* ft_setccirc(); */ ft_curckt = ct;
    }
    ct->ci_name = tt;
    ct->ci_deck = deck;
    ct->ci_options = options;
    if (deck->li_actual)
        ct->ci_origdeck = deck->li_actual;
    else
        ct->ci_origdeck = ct->ci_deck;
    ct->ci_ckt = ckt;
    ct->ci_symtab = (char*)tab;
    ct->ci_inprogress = false;
    ct->ci_runonce = false;
    ct->ci_commands = end;
    if (filename)
        ct->ci_filename = filename;
    else
        tfree(ct->ci_filename);

    if (!noparse) {
        for (; options; options = options->li_next) {
            INP2dot(ct->ci_ckt,(INPtables*)ct->ci_symtab,
                (card*)options,NULL,NULL);
            for (s = options->li_line; *s && !isspace(*s); s++)
                ;
            ii = cp_interactive;
            cp_interactive = false;
            wl = cp_lexer(s);
            cp_interactive = ii;
            if (!wl || !wl->wl_word || !*wl->wl_word)
                continue;
            if (eev)
                eev->va_next = cp_setparse(wl);
            else
                ct->ci_vars = eev = cp_setparse(wl);
            wl_free(wl);
            while (eev->va_next)
                eev = eev->va_next;

            if (options->li_error)
                out_printf("Warning, line %d : %s\n%s\n",
                    options->li_linenum, options->li_line, options->li_error);
        }
    }

    cp_addkword(CT_CKTNAMES, tt);
    return;
}


/* Edit and re-load the current input deck.  In this version, the previous
 * circuit (before editing) is kept, but is no longer associated with the
 * original file name, unless "-r" option is given. "-n" supresses source
 * after edit;
 */

void
com_edit(wl)

wordlist *wl;
{

    char *filename;
    FILE *fp;
    bool permfile, nosource = false, reuse = false;
    struct circ *old;
    extern char *kw_editor;
    char ed[BSIZE_SP];
    int rval;

    if (SCEDactive()) {
        ShowPrompt("Exit sced to edit.");
        return;
    }

top:
    if (wl) {

        /* supress sourcing if "-n" given */
        /* reuse ckt structure if "-r" given */
        if (eq(wl->wl_word,"-n")) {
            nosource = true;
            wl = wl->wl_next;
            goto top;
        }
        if (eq(wl->wl_word,"-r")) {
            reuse = true;
            wl = wl->wl_next;
            goto top;
        }
        if (wl->wl_next && eq(wl->wl_next->wl_word,"-n"))
            nosource = true;
        if (wl->wl_next && eq(wl->wl_next->wl_word,"-r"))
            reuse = true;
        filename = wl->wl_word;
        permfile = true;
    }
    else {
        if (ft_curckt && ft_curckt->ci_filename) {
            /* know the circuit and file */
            filename = ft_curckt->ci_filename;
            permfile = true;
        }
        else {
            /* work with temp file */
            filename = smktemp("sp");
            permfile = false;
        }
        if (ft_curckt && !ft_curckt->ci_filename) {
            /* Circuit came from file which was subsequently modified.
             * Dump circuit listing into temp file.
             */
            if (!(fp = fopen(filename, "w"))) {
                perror(filename);
                return;
            }
            inp_list(fp, ft_curckt->ci_origdeck ? ft_curckt->ci_origdeck :
                ft_curckt->ci_deck, ft_curckt->ci_options, LS_DECK);
            fprintf(cp_err,
        "Warning: editing a temporary file -- circuit not saved\n");
            (void) fclose(fp);
        }
        else if (!ft_curckt) {
            /* No current circuit, user must create (using temp file) */
            if (!(fp = fopen(filename, "w"))) {
                perror(filename);
                return;
            }
            fprintf(fp, "SPICE 3 test deck\n");
            (void) fclose(fp);
        }
    }

    rval = cp_getvar(kw_editor, VT_STRING, ed);
    if ((!rval || cieq(ed, "xeditor")) && !xeditor(filename)) {
        if (!permfile) {
            (void) unlink(filename);
            txfree(filename);
        }
        return;
    }
    else {

#ifdef HAVE_STAT
        /* supress source if file is not written */
        int i, j;
        struct stat st1, st2;

        if (permfile) {
            i = stat(filename,&st1);
            if (inp_edit(filename))
                return;
            j = stat(filename,&st2);
            if (j == -1)
                return;
            if (!i && st1.st_mtime == st2.st_mtime) {
                /* file was not modified */
                for (old = ft_circuits; old; old = old->ci_next) {
                    if (old->ci_filename && eq(old->ci_filename,filename))
                        /* already sourced */
                        return;
                }
            }
        }
        else
#endif
        if (inp_edit(filename)) return;
    }
    if (!nosource)
        inp_srcedit(filename, permfile, reuse);
}


void
inp_srcedit(filename, permfile, reuse)

char *filename;
bool permfile, reuse;
{
    FILE *fp;
    struct circ *old;
    bool inter;

    inter = cp_interactive;
    cp_interactive = false;
    if (!(fp = inp_pathopen(filename, "r"))) {
        perror(filename);
        goto ret;
    }
    if (reuse) if_cktclear();
    inp_spsource(fp, false, permfile ? filename : (char *) NULL);
    (void) fclose(fp);

    if (!permfile) {
        (void) unlink(filename);
        txfree(filename);
    }

    /* Check for other loaded circuits from the same file.  If found,
     * free and null the filename entry, as the file has been edited
     * and therefore does not correspond to the saved circuit struct.
     */
    if (permfile)
        for (old = ft_circuits; old; old = old->ci_next) {
            if (old != ft_curckt && old->ci_filename &&
                eq(old->ci_filename,filename))
                tfree(old->ci_filename);
        }
ret:
    cp_interactive = inter;
    return;
}


/* SCED interface */
void
com_sced(wl)

wordlist *wl;
{
    FILE *fp,*errfp,*outfp,*errtmp,*outtmp;
    struct circ *old;
    bool inter, nosource = false, reuse = false;
    char *errmsg = "Error: no symbolic representation found.\n";
    char *filename,*errfile,*outfile;
    int i;

    inter = cp_interactive;
    cp_interactive = false;
top:
    if (wl) {

        /* supress sourcing if "-n" given */
        /* reuse ckt structure if "-r" given */
        if (eq(wl->wl_word,"-n")) {
            nosource = true;
            wl = wl->wl_next;
            goto top;
        }
        if (eq(wl->wl_word,"-r")) {
            reuse = true;
            wl = wl->wl_next;
            goto top;
        }
        if (wl->wl_next && eq(wl->wl_next->wl_word,"-n"))
            nosource = true;
        if (wl->wl_next && eq(wl->wl_next->wl_word,"-r"))
            reuse = true;
        filename = wl->wl_word;
    }
    else {
        if (ft_curckt && ft_curckt->ci_filename) {
            /* know the circuit and symbolic file */
            filename = ft_curckt->ci_filename;
        }
        else {
            if (ft_curckt) {
                fprintf(cp_err,errmsg);
                goto ret;
            }
            else {
                filename = "noname";
            }
        }
    }
    errfile = smktemp("er");
    outfile = smktemp("ot");

    errfp = fopen(errfile,"w+");
    outfp = fopen(outfile,"w");
    if (errfp == NULL || outfp == NULL) {
        fprintf(cp_err,"Error: redirection failed.\n");
        return;
    }
    errtmp = cp_err;
    outtmp = cp_out;
    cp_err = cp_curerr = errfp;
    cp_out = cp_curout = outfp;


#ifdef HAVE_STAT
/* supress source if file is not written */
{
    int k, j;
    struct stat st1, st2;
    char *f1;

    f1 = filename;
    k = stat(filename,&st1);
    i = sced(&filename);
    if (i == 1) {
        j = stat(filename,&st2);
        if (j == -1)
            i = 2;
        if (!k && eq(f1,filename) && st1.st_mtime == st2.st_mtime) {
            /* file was not modified */
            for (old = ft_circuits; old; old = old->ci_next) {
                if (old->ci_filename && eq(old->ci_filename,filename)) {
                    /* already sourced */
                    i = 2;
                    break;
                }
            }
        }
    }
}
#else
    i = sced(&filename);
#endif
    cp_curout = outtmp;
    cp_curerr = errtmp;
    cp_ioreset();
    unlink(errfile);
    unlink(outfile);
    tfree(errfile);
    tfree(outfile);

    if (i == 3) {
        fprintf(cp_err,errmsg);
        goto ret;
    }
    if (i == 2 || nosource)
        goto ret;

    if (i == 1) {
        if (!(fp = inp_pathopen(filename, "r"))) {
            perror(filename);
            goto ret;
        }
        if (reuse) if_cktclear();
        fprintf(cp_err,"Sourcing the spice text from %s\n",filename);
        fprintf(cp_err,
"Warning: this text may not accurately reflect subcircuit changes\n\
made after %s was saved in sced\n",filename);
        inp_spsource(fp,false,filename);
        (void) fclose(fp);
    }

    /* Check for other loaded circuits from the same file.  If found,
     * free and null the filename entry, as the file has been edited
     * and therefore does not correspond to the saved circuit struct.
     */
    for (old = ft_circuits; old; old = old->ci_next) {
        if (old != ft_curckt)
            if (old->ci_filename && eq(old->ci_filename,filename))
                tfree(old->ci_filename);
    }
ret:
    cp_interactive = inter;
    return;
}


void
com_source(wl)

wordlist *wl;
{
    FILE *fp, *tp;
    char buf[BSIZE_SP];
    bool inter;
    char *tempfile = NULL, *fname;
    wordlist *owl, *ww, *wn;
    int i;
    bool nospice = false, nocmds = false, reuse = false;

    owl = wl;
    wl = wl_copy(wl);
    ww = wl;
    for (; wl; wl = wn) {
        wn = wl->wl_next;
        if (*wl->wl_word != '-')
            continue;
        if (strchr(wl->wl_word,'n'))
            nospice = true;
        else if (strchr(wl->wl_word,'c'))
            nocmds = true;
        else if (strchr(wl->wl_word,'r'))
            reuse = true;
        else
            continue;
        if (wl->wl_prev)
            wl->wl_prev->wl_next = wl->wl_next;
        if (wl->wl_next)
            wl->wl_next->wl_prev = wl->wl_prev;
        txfree(wl->wl_word);
        if (ww == wl)
            ww = wl->wl_next;
        txfree((char*)wl);
    }
    wl = ww;

    inter = cp_interactive;
    cp_interactive = false;
    if (wl && wl->wl_next) {
        /* There are several files -- put them into a temp file  */
        tempfile = smktemp("sp");
        if (!(fp = inp_pathopen(tempfile, "w+"))) {
            perror(tempfile);
            goto done;
        }
        while (wl) {
            if (!(tp = inp_pathopen(wl->wl_word, "r"))) {
                perror(wl->wl_word);
                (void) fclose(fp);
                goto done;
            }
            while ((i = fread(buf, 1, BSIZE_SP, tp)) > 0)
                (void) fwrite(buf, 1, i, fp);
            (void) fclose(tp);
            wl = wl->wl_next;
        }
        (void) fseek(fp, (long) 0, 0);
    }
    else {
        if (wl) {
            if (!(fp = inp_pathopen(wl->wl_word, "r"))) {
                perror(wl->wl_word);
                goto done;
            }
        }
        else
            fp = stdin;
    }
    if (reuse && !nospice)
        if_cktclear();

    if (tempfile)
        fname = NULL;
    else {
        if (wl)
            fname = wl->wl_word;
        else
            fname = NULL;
    }

    /* Don't print the title if this is a .spiceinit file. */
    if (substring(".spiceinit", owl->wl_word) ||
            substring("spice.rc", owl->wl_word))
        spsource(fp, true, true, fname);
    else
        spsource(fp,nospice,nocmds,fname);
    if (fp != stdin)
        (void) fclose(fp);

done:
    cp_interactive = inter;

    if (tempfile) {
        (void) unlink(tempfile);
        txfree(tempfile);
    }
    wl_free(ww);
    return;
}


void
inp_source(file)

char *file;
{
    wordlist *wl;

    wl = cp_lexer(file);
    com_source(wl);
    wl_free(wl);
    return;
}
