/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "ftedefs.h"
#include "spfteext.h"

/* Interpolate all the vectors in a plot to a linear time scale, which
 * we determine by looking at the transient parameters in the CKT struct.
 */

#ifdef __STDC__
static void lincopy(struct dvec*,double*,int,struct dvec*);
#else
static void lincopy();
#endif


void
com_linearize(wl)

wordlist *wl;
{
    double tstart, tstop, tstep, d;
    struct plot *new, *old;
    struct dvec *newtime, *v;
    struct dvec *oldtime;
    wordlist *wl0, *tl;
    int len, i;
    char buf[BSIZE_SP];

    if (!plot_cur || !plot_cur->pl_scale) {
        fprintf(cp_err, "Error: no current plot or no scale\n");
        return;
    }
    if (!isreal(plot_cur->pl_scale)) {
        fprintf(cp_err, "Error: non-real time scale for %s\n",
                plot_cur->pl_typename);
        return;
    }
    if (!ciprefix("tran", plot_cur->pl_typename)) {
        fprintf(cp_err, "Error: plot must be a transient analysis\n");
        return;
    }

    tstart = plot_cur->pl_start;
    tstop = plot_cur->pl_stop;
    tstep = plot_cur->pl_step;

    if (((tstop - tstart) * tstep <= 0.0) || ((tstop - tstart) < tstep)) {
        fprintf(cp_err, 
         "Error: bad parameters -- start = %G, stop = %G, step = %G\n",
                tstart, tstop, tstep);
        return;
    }

    old = plot_cur;
    oldtime = old->pl_scale;
    new = plot_alloc("transient");
    (void) sprintf(buf, "%s (linearized)", old->pl_name);
    new->pl_name = copy(buf);
    new->pl_title = copy(old->pl_title);
    new->pl_date = copy(old->pl_date);
    new->pl_next = plot_list;
    new->pl_start = tstart;
    new->pl_stop = tstop;
    new->pl_step = tstep;
    plot_new(new);
    plot_setcur(new->pl_typename);
    plot_list = new;
    len = (tstop - tstart) / tstep + 1.5;
    newtime = alloc(struct dvec);
    newtime->v_name = copy(oldtime->v_name);
    newtime->v_type = oldtime->v_type;
    newtime->v_flags = oldtime->v_flags;
    newtime->v_length = len;
    newtime->v_plot = new;
    newtime->v_realdata = (double *) tmalloc(len * sizeof (double));
    for (i = 0, d = tstart; i < len; i++, d += tstep)
        newtime->v_realdata[i] = d;
    vec_newperm(newtime); /* set to scale */

    if (wl && !cieq(wl->wl_word,"all")) {
        while (wl) {
            v = vec_fromplot(wl->wl_word, old);
            if (!v) {
                fprintf(cp_err, "Error: no such vector %s\n",
                        wl->wl_word);
                continue;
            }
            lincopy(v, newtime->v_realdata, len, oldtime);
            wl = wl->wl_next;
        }
    }
    else {
        wl0 = (wordlist*)htab_list(old->pl_hashtab);
        for (tl = wl0; tl; tl = tl->wl_next) {
            v = (struct dvec *)((char**)tl->wl_word)[1];
            if (v == old->pl_scale)
                continue;
            lincopy(v, newtime->v_realdata, len, oldtime);
        }
        wl_free(wl0);
    }
    return;
}


static void
lincopy(ov, newscale, newlen, oldscale)

struct dvec *ov, *oldscale;
double *newscale;
int newlen;
{
    struct dvec *v;
    double *nd;

    if (!isreal(ov)) {
        fprintf(cp_err, "Warning: %s is not real\n", ov->v_name);
        return;
    }
    if (ov->v_length < oldscale->v_length) {
        fprintf(cp_err, "Warning: %s is too short\n", ov->v_name);
        return;
    }
    v = alloc(struct dvec);
    v->v_name = copy(ov->v_name);
    v->v_type = ov->v_type;
    v->v_flags = ov->v_flags;
    v->v_length = newlen;

    nd = (double *) tmalloc(newlen * sizeof (double));
    if (!ft_interpolate(ov->v_realdata, nd, oldscale->v_realdata,
            oldscale->v_length, newscale, newlen, 1)) {
        fprintf(cp_err, "Error: can't interpolate %s\n", ov->v_name);
        return;
    }
    v->v_realdata = nd;
    vec_newperm(v);
    return;
}
