/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1993 Stephen R. Whiteley
****************************************************************************/

/*
 * Circuit simulation commands.
 */

#include "spice.h"
#include "ftedefs.h"
#include "fteinp.h"
#include "spfteext.h"
#include "scedio.h"

/* Routines for the commands op, tran, ac, dc, listing, device, state,
 * resume, stop, trace, run, end.  Op, tran, ac, and dc cause the action
 * to be performed immediately, and run causes whatever actions were
 * present in the deck to be carried out. End has the effect of stopping
 * any simulations in progress, as opposed to ending te input deck as
 * the .end card does.
 */

static FILE *rawfileFp;
static bool rawfileBinary;


void
com_scirc(wl)

wordlist *wl;
{
    struct circ *p;
    int i, j = 0;
    char buf[BSIZE_SP];
        char *xx = "Type the number of the desired circuit: ";

    if (ft_circuits == NULL) {
        fprintf(cp_err, "Error: there aren't any circuits loaded.\n");
        return;
    }
    if (wl == NULL) {
        for (p = ft_circuits; p; p = p->ci_next) {
            if (ft_curckt == p)
                out_printf("Current %-4d%s\n", ++j, p->ci_name);
            else
                out_printf("        %-4d%s\n", ++j, p->ci_name);
        }
        out_printf("\n");
        (void) SCEDfgets(buf,BSIZE_SP,cp_in,xx);

        clearerr(cp_in);
        if ((sscanf(buf, " %d ", &i) != 1) || (i < 0) || (i > j))
            return;
        for (p = ft_circuits; --i > 0; p = p->ci_next);
    }
    else {
        for (p = ft_circuits; p; p = p->ci_next)
            if (ciprefix(wl->wl_word, p->ci_name))
                break;
        if (p == NULL) {
            fprintf(cp_err, "Warning: no such circuit \"%s\"\n",
                    wl->wl_word);
            return;
        }
        out_printf("%s\n", p->ci_name);
    }
    if (ft_curckt) {
        /* Actually this can't be false */
        ft_curckt->ci_devices = 
            cp_kwswitch(CT_DEVNAMES, p->ci_devices);
        ft_curckt->ci_nodes = cp_kwswitch(CT_NODENAMES, p->ci_nodes);
    }
    ft_curckt = p;
    return;
}


void
com_pz(wl)

wordlist *wl;
{
    ft_dosim("pz", wl);
    return;
}


void
com_op(wl)

wordlist *wl;
{
    ft_dosim("op", wl);
    return;
}


void
com_dc(wl)

wordlist *wl;
{
    ft_dosim("dc", wl);
    return;
}


void
com_ac(wl)

wordlist *wl;
{
    ft_dosim("ac", wl);
    return;
}


void
com_tf(wl)

wordlist *wl;
{
    ft_dosim("tf", wl);
    return;
}


void
com_tran(wl)

wordlist *wl;
{
    ft_dosim("tran", wl);
    return;
}


void
com_sens(wl)

wordlist *wl;
{
    ft_dosim("sens", wl);
    return;
}


void
com_disto(wl)

wordlist *wl;
{
    ft_dosim("disto", wl);
    return;
}


void
com_noise(wl)

wordlist *wl;
{
    ft_dosim("noise", wl);
    return;
}


void
ft_dosim(what, wl)

char *what;
wordlist *wl;
{
    bool dofile = false;
    char buf[BSIZE_SP];
    struct circ *ct;
    int err = 0;
    bool ascii = true;
    extern char *kw_filetype;

    /* give sced access to loop command */
    if (eq(what,"loop")) {
        com_loop(wl);
        return;
    }

    if (ft_servermode || (eq(what, "run") && wl))
        dofile = true;
    
    if (cp_getvar(kw_filetype, VT_STRING, buf)) {
        if (eq(buf, "binary"))
            ascii = false;
#ifdef notdef
        else if (eq(buf, "oct"))
            oct = true;
#endif
        else if (!eq(buf, "ascii"))
            fprintf(cp_err, "Warning: strange file type %s\n", buf);
    }

    if (!ft_curckt) {
        fprintf(cp_err, "Error: there aren't any circuits loaded.\n");
        return;
    }
    else if (ft_curckt->ci_ckt == NULL) { /* Set noparse? */
        fprintf(cp_err, "Error: circuit not parsed.\n");
        return;
    }
    for (ct = ft_circuits; ct; ct = ct->ci_next)
        if (ct->ci_inprogress && (ct != ft_curckt)) {
            fprintf(cp_err, 
                "Warning: losing old state for circuit '%s'\n", ct->ci_name);
            ct->ci_inprogress = false;
        }
    if (ft_curckt->ci_inprogress && !eq(what, "resume")) {
        ft_setflag = true;
        fprintf(cp_err, "Resuming run in progress.\n");
        com_resume((wordlist *) NULL);
        return;
    }
    if (ft_curckt->ci_runonce)
        com_rset((wordlist *) NULL);

    ft_intrpt = false;
    if (dofile) {
        if (ft_servermode)
            rawfileFp = cp_out;
        else if (!wl || !wl->wl_word || !*wl->wl_word) {
            if (ft_rawfile)
                if (!(rawfileFp = fopen(ft_rawfile, "w"))) {
                    perror(wl->wl_word);
                    return;
                }
            else
                rawfileFp = cp_out;
        }
        else if (!(rawfileFp = fopen(wl->wl_word, "w"))) {
            perror(wl->wl_word);
            return;
        }
        rawfileBinary = !ascii;
    }
    else {
        rawfileFp = NULL;
    }

    ft_curckt->ci_inprogress = true;
    err = if_run(what, wl);
    if (err == 1) {
        /* The circuit was interrupted somewhere. */
        DevGrHalt();
        fprintf(cp_err, "%s: simulation interrupted\n", what);
    }
    else if (err == 2) {
        DevGrHalt();
        ft_curckt->ci_inprogress = false;
        fprintf(cp_err, "%s: simulation aborted\n", what);
    }
    else
        ft_curckt->ci_inprogress = false;
    
    if (rawfileFp && rawfileFp != cp_out) {
        (void) fclose(rawfileFp);
        rawfileFp = NULL;
    }
    ft_curckt->ci_runonce = true;
}


/* Continue a simulation. If there is non in progress, this is the
 * equivalent of "run".
 */

/* ARGSUSED */
void
com_resume(wl)

wordlist *wl;
{
    int err;

    if (!ft_curckt) return;
    if (ft_curckt->ci_inprogress == false) {
        fprintf(cp_err, "Note: run starting\n");
        com_run((wordlist *) NULL);
        return;
    }
    ft_curckt->ci_inprogress = true;

    err = if_run("resume", (wordlist *) NULL);
    if (err == 1) {
        /* The circuit was interrupted somewhere. */
        DevGrHalt();
        fprintf(cp_err, "simulation interrupted\n");
    }
    else if (err == 2) {
        DevGrHalt();
        ft_curckt->ci_inprogress = false;
        fprintf(cp_err, "simulation aborted\n");
    }
    else
        ft_curckt->ci_inprogress = false;

    if (rawfileFp && rawfileFp != cp_out) {
        (void) fclose(rawfileFp);
        rawfileFp = NULL;
    }
    ft_curckt->ci_runonce = true;
}


void
com_run(wl)

/* Usage is run [filename]
 */
wordlist *wl;
{
    ft_dosim("run", wl);
    return;
}


void
ft_dorun(file)

char *file;
{
    static wordlist wl = { NULL, NULL, NULL } ;

    wl.wl_word = file;
    if (file)
        ft_dosim("run", &wl);
    else
        ft_dosim("run", (wordlist *) NULL);
    return;
}


/* ARGSUSED */
void
com_rset(wl)

/* Throw out the circuit struct and recreate it from the original deck.
 * The shell variables are updated.
 */
wordlist *wl;
{
    struct line *deck, *dd;
    char *filename, *cblock;

    if (ft_curckt == NULL) {
        fprintf(cp_err, "Error: there is no circuit loaded.\n");
        return;
    }

    /*
    (*(ft_sim->deleteTask))(ft_curckt->ci_ckt,ft_curckt->ci_defTask);
    if_cktfree(ft_curckt->ci_ckt, (char*)ft_curckt->ci_symtab);
    inp_dodeck(ft_curckt->ci_deck, ft_curckt->ci_name, (wordlist *) NULL,
            true, ft_curckt->ci_options, ft_curckt->ci_filename);
    ft_curckt->ci_runonce = false;
    */

    dd = ft_curckt->ci_deck;
    deck = dd->li_actual;
    dd->li_actual = NULL;
    ft_curckt->ci_origdeck = NULL;
    filename = ft_curckt->ci_filename;
    ft_curckt->ci_filename = NULL;
    cblock = ft_curckt->ci_contblk;
    ft_curckt->ci_contblk = NULL;
    if_cktclear();
    inp_spdeck(deck,filename);
    ft_curckt->ci_contblk = cblock;
    txfree(filename);
}


/* ARGSUSED */
bool
ft_getOutReq(fpp, plotp, binp, name, title)

FILE **fpp;
struct plot **plotp;
bool *binp;
char *name, *title;
{
    /*struct plot *pl;*/

    if (rawfileFp) {
        *fpp = rawfileFp;
        *binp = rawfileBinary;
        return (true);
    }
    else {
        return (false);
    }
}


void
com_free(wl)

/* free command - free the current plot, circuit */
wordlist *wl;
{
    char s[8];
    char *xx1 = "Delete all plots (no) ? ";
    char *xx2 = "Delete current plot (yes) ? ";
    char *xx3 = "Delete all circuits (no) ? ";
    char *xx4 = "Delete current circuit (yes) ? ";
    bool yes = false;
    bool plots = false;
    bool circuits = false;
    bool all = false;
    wordlist *ww;


    for (ww = wl; ww; ww = ww->wl_next) {
        if (ciprefix("p",ww->wl_word)) {
            plots = true;
            continue;
        }
        if (ciprefix("c",ww->wl_word)) {
            circuits = true;
            continue;
        }
        if (ciprefix("y",ww->wl_word)) {
            yes = true;
            continue;
        }
        if (ciprefix("a",ww->wl_word)) {
            all = true;
            continue;
        }
    }
    if (!plots && !circuits) {
        plots = true;
        circuits = true;
    }
    vec_gc();

    if (plots) {
        if (!plot_cur || eq(plot_cur->pl_name,"constants"))
            (void) fprintf(cp_err,"No plot to delete.\n");
        else {
            if (all) {
                if (!yes)
                    (void) SCEDfgets(s,8,cp_in,xx1);
                if (yes || *s == 'y' || *s == 'Y')
                    while (plot_cur && !eq(plot_cur->pl_name,"constants"))
                        com_destroy((wordlist *)NULL);
            }
            else {
                if (!yes)
                    (void) SCEDfgets(s,8,cp_in,xx2);
                if (yes || *s == 'y' || *s == 'Y' || *s == '\n')
                    com_destroy((wordlist *)NULL);
            }
        }
    }

    if (circuits) {
        if (!ft_curckt) {
            (void) fprintf(cp_err,"No circuit to delete.\n");
            return;
        }
        else {
            if (all) {
                if (!yes)
                    (void) SCEDfgets(s,8,cp_in,xx3);
                if (yes || *s == 'y' || *s == 'Y')
                    while (ft_curckt)
                        if_cktclear();
            }
            else {
                if (!yes)
                    (void) SCEDfgets(s,8,cp_in,xx4);
                if (yes || *s == 'y' || *s == 'Y' || *s == '\n')
                    if_cktclear();
            }
        }
    }
}
